/***************************************************************************
                        chunk.cpp  -  BSP room chunk
                             -------------------
    begin                : Thu Feb 27 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoBSP, chunk.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "bsp.h"
#include "chunk.h"

#include <neochunkio/chunktype.h>
#include <neochunkio/scenenode.h>
#include <neochunkio/boundingvolume.h>
#include <neochunkio/vertexbuffer.h>
#include <neochunkio/polygonbuffer.h>
#include <neochunkio/skydome.h>
#include <neochunkio/skybox.h>
#include <neochunkio/vector3d.h>
#include <neochunkio/stdstring.h>
#include <neochunkio/integer.h>
#include <neochunkio/float.h>
#include <neochunkio/bool.h>

#include <neoengine/aabb.h>
#include <neoengine/plane.h>
#include <neoengine/logstream.h>

#include <vector>

using namespace std;
using namespace NeoEngine;
using namespace NeoChunkIO;


#ifdef HAVE_NEOCHUNKIO


namespace NeoBSP
{


BSPVisClusterChunk::~BSPVisClusterChunk()
{
	delete m_pkCluster;
}


int BSPVisClusterChunk::GetSize( bool bIncludeHeader ) const
{
	int iSize = ComplexChunk::GetSize( bIncludeHeader );

	iSize += 4 + 4 + m_pkCluster->m_iClusterCount * m_pkCluster->m_iRowLength;

	return iSize;
}


int BSPVisClusterChunk::ReadData( File *pkFile, ChunkIO::CHUNKIOMODE eMode, unsigned int uiBytes )
{
	int iOffset = pkFile->Tellg();

	m_pkCluster = new BSPVisCluster;

	*pkFile >> m_pkCluster->m_iClusterCount >> m_pkCluster->m_iRowLength;

	m_pkCluster->m_pucData  = new unsigned char[ m_pkCluster->m_iClusterCount * m_pkCluster->m_iRowLength ];

	if( eMode == ChunkIO::BINARY )
		pkFile->Read( m_pkCluster->m_pucData, m_pkCluster->m_iClusterCount * m_pkCluster->m_iRowLength );
	else
	{
		//Viscluster data doesn't make much sense in ASCII mode though...
		unsigned char *pucLoop = m_pkCluster->m_pucData;

		for( unsigned int uiLoop = 0; uiLoop < (unsigned int)( m_pkCluster->m_iClusterCount * m_pkCluster->m_iRowLength ); ++uiLoop, ++pucLoop )
			*pkFile >> *pucLoop;
	}

	return( !*pkFile ? -1 : ( (int)pkFile->Tellg() - iOffset ) );
}


int BSPVisClusterChunk::WriteData( File *pkFile, ChunkIO::CHUNKIOMODE eMode, unsigned int uiLevel )
{
	int iOffset = pkFile->Tellp();


	if( eMode == ChunkIO::BINARY )
	{
		*pkFile << m_pkCluster->m_iClusterCount << m_pkCluster->m_iRowLength;

		pkFile->Write( m_pkCluster->m_pucData, m_pkCluster->m_iClusterCount * m_pkCluster->m_iRowLength );
	}
	else
	{
		//Viscluster data doesn't make much sense in ASCII mode though...
		*pkFile << ' ' << m_pkCluster->m_iClusterCount << ' ' << m_pkCluster->m_iRowLength << hex << showbase << showpos;

		unsigned char *pucLoop = m_pkCluster->m_pucData;
		unsigned int uiLoop;

		for( uiLoop = 0; uiLoop < (unsigned int)( m_pkCluster->m_iClusterCount * m_pkCluster->m_iRowLength ); ++uiLoop, ++pucLoop )
		{
			if( !( uiLoop % 10 ) )
				*pkFile << endl << string( uiLevel * 2, ' ' );

			*pkFile << *pucLoop;
		}

		if( !( --uiLoop % 10 ) )
			*pkFile << endl << string( uiLevel * 2, ' ' );

		*pkFile << dec << noshowbase << noshowpos;
	}

	return( !*pkFile ? -1 : ( (int)pkFile->Tellp() - iOffset ) );
}






BSPNodeChunk::~BSPNodeChunk()
{
	if( m_pkNode )
	{
		m_pkNode->m_pkFront = 0;
		m_pkNode->m_pkBack  = 0;

		delete m_pkNode;
	}

	delete [] m_piVertexBuffers;
}


int BSPNodeChunk::ParseData( unsigned int uiFlags, FileManager *pkFileManager )
{
	Chunk *pkChunk;

	m_pkNode = new BSPNode;


	if( !( pkChunk = FindChunk( "id", ChunkType::INTEGER, -1 ) ) )
		neolog << LogLevel( WARNING ) << "*** Invalid BSP " << ( ( m_usType == ChunkType::BSPNODE ) ? "node" : "leaf" ) << " chunk: No ID" << endl;
	else
		m_pkNode->m_iID = dynamic_cast< IntChunk* >( pkChunk )->m_iData;


	if( ( pkChunk = FindChunk( "bound", ChunkType::BOUNDINGVOLUME, -1 ) ) )
	{
		if( pkChunk->GetType() != ChunkType::AABB )
			neolog << LogLevel( WARNING ) << "*** Invalid BSP " << ( ( m_usType == ChunkType::BSPNODE ) ? "node" : "leaf" ) << " chunk: Unsupported bounding volume type" << endl;
		else
		{
			BoundingVolumeChunk *pkAABBChunk = dynamic_cast< BoundingVolumeChunk* >( pkChunk );

			m_pkNode->m_pkAABB = dynamic_cast< AABB* >( pkAABBChunk->m_pkVolume );

			pkAABBChunk->m_pkVolume = 0;
		}
	}


	if( m_usType == ChunkType::BSPLEAF )
	{
		m_pkNode->m_iType = BSPNode::LEAF;


		if( ( pkChunk = FindChunk( "vis", ChunkType::INTEGER, -1 ) ) )
			m_pkNode->m_iVisCluster = dynamic_cast< IntChunk* >( pkChunk )->m_iData;


		//Find all polygon buffers
		vector<Chunk*> vpkBufChunks;

		FindChunks( "", ChunkType::POLYGONBUFFER, &vpkBufChunks, false );

		unsigned int uiEnd = vpkBufChunks.size();

		if( uiEnd )
		{
			m_pkNode->m_vpkPolygons.resize( uiEnd );

			m_piVertexBuffers = new int[ uiEnd ];

			memset( m_piVertexBuffers, 0, sizeof( int ) * uiEnd );

			for( unsigned int uiBuf = 0; uiBuf < uiEnd; ++uiBuf )
			{
				m_pkNode->m_vpkPolygons[ uiBuf ] = dynamic_cast< PolygonBufferChunk* >( vpkBufChunks[ uiBuf ] )->m_pkBuffer;

				if( ( pkChunk == vpkBufChunks[ uiBuf ]->FindChunk( "vertexbuffer", ChunkType::INTEGER, -1 ) ) )
					m_piVertexBuffers[ uiBuf ] = dynamic_cast< IntChunk* >( pkChunk )->m_iData;
			}
		}
	}
	else
	{
		m_pkNode->m_iType       = BSPNode::NODE;
		m_pkNode->m_pkPartition = new Plane;

		if( !( pkChunk = FindChunk( "planedist", ChunkType::FLOAT, -1 ) ) )
			neolog << LogLevel( WARNING ) << "*** Invalid BSP " << ( ( m_usType == ChunkType::BSPNODE ) ? "node" : "leaf" ) << " chunk: No split plane distance" << endl;
		else
			m_pkNode->m_pkPartition->m_fDistance = dynamic_cast< FloatChunk* >( pkChunk )->m_fData;


		if( !( pkChunk = FindChunk( "planenormal", ChunkType::VECTOR, -1 ) ) )
			neolog << LogLevel( WARNING ) << "*** Invalid BSP " << ( ( m_usType == ChunkType::BSPNODE ) ? "node" : "leaf" ) << " chunk: No split plane normal" << endl;
		else
			m_pkNode->m_pkPartition->m_kNormal = dynamic_cast< VectorChunk* >( pkChunk )->m_kData;

		if( !( pkChunk = FindChunk( "front", ChunkType::INTEGER, -1 ) ) )
			neolog << LogLevel( WARNING ) << "*** Invalid BSP " << ( ( m_usType == ChunkType::BSPNODE ) ? "node" : "leaf" ) << " chunk: No front node ID" << endl;
		else
			m_pkNode->m_pkFront = (BSPNode*)dynamic_cast< IntChunk* >( pkChunk )->m_iData;

		if( !( pkChunk = FindChunk( "back", ChunkType::INTEGER, -1 ) ) )
			neolog << LogLevel( WARNING ) << "*** Invalid BSP " << ( ( m_usType == ChunkType::BSPNODE ) ? "node" : "leaf" ) << " chunk: No back node ID" << endl;
		else
			m_pkNode->m_pkBack = (BSPNode*)dynamic_cast< IntChunk* >( pkChunk )->m_iData;
	}

	return 1;
}





int BSPRoomChunk::ParseData( unsigned int uiFlags, FileManager *pkFileManager )
{
	RoomChunk::ParseData( uiFlags, pkFileManager );

	BSPRoom *pkRoom = new BSPRoom;
	m_pkRoom          = pkRoom;

	Chunk *pkChunk;

	if( ( pkChunk = FindChunk( "name", ChunkType::STRING, -1 ) ) )
		pkRoom->SetName( dynamic_cast< StringChunk* >( pkChunk )->m_strData );


	//Find vertex buffer chunks
	vector<Chunk*> vpkVBufChunks;

	FindChunks( "", ChunkType::VERTEXBUFFER, &vpkVBufChunks, false );
			
	vector<Chunk*>::iterator ppkChunk = vpkVBufChunks.begin();
	vector<Chunk*>::iterator ppkEnd   = vpkVBufChunks.end();

	if( ppkChunk == ppkEnd )
	{
		neolog << LogLevel( ERROR ) << "*** Invalid BSP room chunk: No vertex buffers" << endl;
		return -1;
	}

	pkRoom->m_vpkVertexBuffers.resize( vpkVBufChunks.size() );
	
	for( unsigned int uiVertexBuffer = 0; ppkChunk != ppkEnd; ++ppkChunk, ++uiVertexBuffer )
		pkRoom->m_vpkVertexBuffers[ uiVertexBuffer ] = dynamic_cast< VertexBufferChunk* >( *ppkChunk )->m_pkBuffer;


	if( ( pkChunk = FindChunk( "", ChunkType::BSPVISCLUSTER, -1 ) ) )
	{
		BSPVisClusterChunk *pkVisChunk = dynamic_cast< BSPVisClusterChunk* >( pkChunk );

		pkRoom->m_pkVisCluster  = pkVisChunk->m_pkCluster;
		pkVisChunk->m_pkCluster = 0;
	}


	vector< Chunk* > vpkChunks;
	vector< BSPNodeChunk* > vpkNodeChunks;
	vector< BSPNodeChunk* > vpkLeafChunks;

	FindChunks( "", ChunkType::BSPNODE, &vpkChunks, false );
	
	unsigned int uiNode;
	unsigned int uiStartNode = 0;
	unsigned int uiNumNodes  = vpkChunks.size();

	vpkNodeChunks.resize( uiNumNodes );

	for( uiNode = 0; uiNode < uiNumNodes; ++uiNode )
		vpkNodeChunks[ uiNode ] = dynamic_cast< BSPNodeChunk* >( vpkChunks[ uiNode ] );


	vpkChunks.clear();


	FindChunks( "", ChunkType::BSPLEAF, &vpkChunks, false );

	unsigned int uiLeaf;
	unsigned int uiStartLeaf = 0;
	unsigned int uiNumLeaves = vpkChunks.size();

	vpkLeafChunks.resize( uiNumLeaves );

	for( uiLeaf = 0; uiLeaf < uiNumLeaves; ++uiLeaf )
		vpkLeafChunks[ uiLeaf ] = dynamic_cast< BSPNodeChunk* >( vpkChunks[ uiLeaf ] );

	vpkChunks.clear();


	for( uiNode = 0; uiNode < uiNumNodes; ++uiNode )
	{
		BSPNode *pkNode = vpkNodeChunks[ uiNode ]->m_pkNode;

		if( pkNode->m_iID == 0 )
		{
			pkRoom->m_pkRootNode = pkNode;

			vpkNodeChunks[ uiNode ]->m_iLoadFlag++;
		}

		int iFront = (int)pkNode->m_pkFront; //Ugly, but works. see parsing of bsp node chunk above
		int iBack  = (int)pkNode->m_pkBack;

		pkNode->m_pkFront = 0;
		pkNode->m_pkBack  = 0;

		unsigned int i;

		for( i = uiStartNode; i < uiNumNodes; ++i )
		{
			BSPNode *pkThisNode = vpkNodeChunks[ i ]->m_pkNode;

			if( pkThisNode->m_iID == iFront )
			{
				pkNode->m_pkFront = pkThisNode;

				vpkNodeChunks[ i ]->m_iLoadFlag++;

				if( i == uiStartNode )
					++uiStartNode;
			}
			else if( pkThisNode->m_iID == iBack )
			{
				pkNode->m_pkBack = pkThisNode;

				vpkNodeChunks[ i ]->m_iLoadFlag++;

				if( i == uiStartNode )
					++uiStartNode;
			}

			if( pkNode->m_pkFront && pkNode->m_pkBack )
				break;
		}

		if( pkNode->m_pkFront && pkNode->m_pkBack )
			continue;

		for( i = uiStartLeaf; i < uiNumLeaves; ++i )
		{
			BSPNode *pkThisNode = vpkLeafChunks[ i ]->m_pkNode;

			if( pkThisNode->m_iID == iFront )
			{
				pkNode->m_pkFront = pkThisNode;

				vpkLeafChunks[ i ]->m_iLoadFlag++;

				if( i == uiStartLeaf )
					++uiStartLeaf;
			}
			else if( pkThisNode->m_iID == iBack )
			{
				pkNode->m_pkBack = pkThisNode;

				vpkLeafChunks[ i ]->m_iLoadFlag++;

				if( i == uiStartLeaf )
					++uiStartLeaf;
			}

			if( pkNode->m_pkFront && pkNode->m_pkBack )
				break;
		}

		if( !pkNode->m_pkFront )
			neolog << LogLevel( ERROR ) << "*** Invalid BSP room chunk: Failed to locate front node [" << iFront << "]" << endl;

		if( !pkNode->m_pkBack )
			neolog << LogLevel( ERROR ) << "*** Invalid BSP room chunk: Failed to locate back node [" << iBack << "]" << endl;
	}


	for( uiNode = 0; uiNode < uiNumNodes; ++uiNode )
	{
		if( !vpkNodeChunks[ uiNode ]->m_iLoadFlag )
			neolog << LogLevel( WARNING ) << "*** Unreferenced node [" << vpkNodeChunks[ uiNode ]->m_pkNode->m_iID << "]" << endl;
		else
			vpkNodeChunks[ uiNode ]->m_pkNode = 0;
	}


	for( uiLeaf = 0; uiLeaf < uiNumLeaves; ++uiLeaf )
	{
		BSPNode *pkThisNode = vpkLeafChunks[ uiLeaf ]->m_pkNode;

		if( !vpkLeafChunks[ uiLeaf ]->m_iLoadFlag )
		{
			int iPolys = 0;
			for( unsigned int iBuffer = 0; iBuffer < pkThisNode->m_vpkPolygons.size(); ++iBuffer )
				iPolys += pkThisNode->m_vpkPolygons[iBuffer]->GetNumElements();

			pkThisNode->m_pkFront = 0;
			pkThisNode->m_pkBack  = 0;

			neolog << LogLevel( WARNING ) << "*** Unreferenced leaf [" << vpkLeafChunks[ uiLeaf ]->m_pkNode->m_iID << "] with [" << iPolys << "] polygons" << endl;
		}
		else
		{
			pkRoom->m_vpkLeaves.push_back( pkThisNode );

			for( unsigned int uiBuffer = 0; uiBuffer < pkThisNode->m_vpkPolygons.size(); ++uiBuffer )
				pkThisNode->m_vpkPolygons[ uiBuffer ]->m_pkVertexBuffer = pkRoom->m_vpkVertexBuffers[ vpkLeafChunks[ uiLeaf ]->m_piVertexBuffers[ uiBuffer ] ];

			vpkLeafChunks[ uiLeaf ]->m_pkNode = 0;
		}
	}


	//Find scene nodes
	vector<Chunk*> vpkSceneNodeChunks;

	FindChunks( "", ChunkType::SCENENODE, &vpkSceneNodeChunks, false );

	unsigned int uiEnd = vpkSceneNodeChunks.size();

	for( uiNode = 0; uiNode < uiEnd; ++uiNode )
	{
		SceneNodeChunk *pkNodeChunk = dynamic_cast< SceneNodeChunk* >( vpkSceneNodeChunks[ uiNode ] );

		bool bGlobal = false;

		if( ( pkChunk = pkNodeChunk->FindChunk( "global", ChunkType::BOOL, -1 ) ) )
			bGlobal = dynamic_cast< BoolChunk* >( pkChunk )->m_bData;

		if( bGlobal )
			pkRoom->AttachGlobalNode( pkNodeChunk->m_pkNode );
		else
			pkRoom->AttachNode( pkNodeChunk->m_pkNode );

		pkNodeChunk->m_pkNode = 0;
	}

	return 1;
}


#endif


};

