/***************************************************************************
                 animlib.cpp  -  Animation library file loader
                             -------------------
    begin                : Mon May 5 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoChunkIO, animlib.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "animlib.h"
#include "chunkio.h"
#include "chunktype.h"
#include "submeshanimation.h"
#include "skeletonanimation.h"
#include "nodeanimation.h"

#include <neoengine/file.h>
#include <neoengine/logstream.h>

using namespace std;
using namespace NeoEngine;

namespace NeoChunkIO
{


AnimationLibrary::~AnimationLibrary()
{
	{
		vector< SubMeshAnimation* >::iterator ppkAnim = m_vpkSubMeshAnimations.begin();
		vector< SubMeshAnimation* >::iterator ppkEnd  = m_vpkSubMeshAnimations.end();

		for( ; ppkAnim != ppkEnd; ++ppkAnim )
			delete (*ppkAnim);
	}

	{
		vector< NodeAnimation* >::iterator ppkAnim = m_vpkNodeAnimations.begin();
		vector< NodeAnimation* >::iterator ppkEnd  = m_vpkNodeAnimations.end();

		for( ; ppkAnim != ppkEnd; ++ppkAnim )
			delete (*ppkAnim);
	}

	{
		vector< SkeletonAnimation* >::iterator ppkAnim = m_vpkSkeletonAnimations.begin();
		vector< SkeletonAnimation* >::iterator ppkEnd  = m_vpkSkeletonAnimations.end();

		for( ; ppkAnim != ppkEnd; ++ppkAnim )
			delete (*ppkAnim);
	}
}


bool AnimationLibrary::LoadNode( unsigned int uiFlags )
{
	if( !m_pkFile || !m_pkFile->IsValid() )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to load animation lib: Invalid file" << endl;
		return false;
	}

	char   szStr[5];
	string strMode;

	if( !m_pkFile->DetermineByteOrder( 0x494e414e ) )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to load animation lib: Failed to determine byte order, possible corrupt file" << endl;
		return false;
	}

    m_pkFile->Read( szStr, 4 ); szStr[4] = 0;

	if( string( szStr ) != "NANI" )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to load animation lib: Invalid ID" << endl;
		return false;
	}

	//Read mode string
	m_pkFile->Read( szStr, 4 );

	ChunkIO *pkIO = 0;

	if( string( szStr ) == "!bin" )
	{
		pkIO = new ChunkIO( ChunkIO::BINARY );
		m_pkFile->SetBinary( true );
	}
	else if( string( szStr ) == "!txt" )
	{
		pkIO = new ChunkIO( ChunkIO::ASCII );
		m_pkFile->SetBinary( false );
	}
	else
	{
		neolog << LogLevel( ERROR ) << "*** Unable to load animation lib: Unsupported chunk file mode [" << strMode << "]" << endl;
		return false;
	}

	*m_pkFile >> pkIO->m_iMajorVersion >> pkIO->m_iMinorVersion;

	if( !( pkIO->m_iMajorVersion == ChunkIO::MAJORVERSIONREQUIRED ) || ( pkIO->m_iMinorVersion < ChunkIO::MINORVERSIONREQUIRED ) )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to load animation lib: Invalid chunk format version " << pkIO->m_iMajorVersion << "." << pkIO->m_iMinorVersion << endl;
		return false;
	}


	Chunk *pkChunk = 0;

	vector< SubMeshAnimationChunk* >  vpkSubMeshChunks;
	vector< SkeletonAnimationChunk* > vpkSkeletonChunks;
	vector< NodeAnimationChunk* >     vpkNodeChunks;
	
	while( ( pkChunk = pkIO->ReadChunk( m_pkFile )  ) )
	{
		if( pkChunk->GetType() == ChunkType::SUBMESHANIMATION )
			vpkSubMeshChunks.push_back( dynamic_cast< SubMeshAnimationChunk* >( pkChunk ) );
		else if( pkChunk->GetType() == ChunkType::NODEANIMATION )
			vpkNodeChunks.push_back( dynamic_cast< NodeAnimationChunk* >( pkChunk ) );
		else if( pkChunk->GetType() == ChunkType::SKELETONANIMATION )
			vpkSkeletonChunks.push_back( dynamic_cast< SkeletonAnimationChunk* >( pkChunk ) );
		else
		{
			neolog << LogLevel( ERROR ) << "*** Invalid animation lib: Unsupported toplevel chunk type [" << pkChunk->GetType() << "] [" << pkChunk->GetTypeAsString() << "]" << endl;
			delete pkChunk;
		}
	}


	unsigned int uiParseFlags = 0;

	{
		vector< SubMeshAnimationChunk* >::iterator ppkChunk = vpkSubMeshChunks.begin();
		vector< SubMeshAnimationChunk* >::iterator ppkEnd   = vpkSubMeshChunks.end();

		//Now parse chunks
		for( ; ppkChunk != ppkEnd; ++ppkChunk )
		{
			if( pkIO->ParseChunk( *ppkChunk, uiParseFlags, m_pkFileManager ) >= 0 )
			{
				m_vpkSubMeshAnimations.push_back( (*ppkChunk)->m_pkAnim );

				(*ppkChunk)->m_pkAnim = 0;
			}

			delete( *ppkChunk );
		}
	}

	{
		vector< NodeAnimationChunk* >::iterator ppkChunk = vpkNodeChunks.begin();
		vector< NodeAnimationChunk* >::iterator ppkEnd   = vpkNodeChunks.end();

		//Now parse chunks
		for( ; ppkChunk != ppkEnd; ++ppkChunk )
		{
			if( pkIO->ParseChunk( *ppkChunk, uiParseFlags, m_pkFileManager ) >= 0 )
			{
				m_vpkNodeAnimations.push_back( (*ppkChunk)->m_pkAnim );

				(*ppkChunk)->m_pkAnim = 0;
			}

			delete( *ppkChunk );
		}
	}

	{
		vector< SkeletonAnimationChunk* >::iterator ppkChunk = vpkSkeletonChunks.begin();
		vector< SkeletonAnimationChunk* >::iterator ppkEnd   = vpkSkeletonChunks.end();

		//Now parse chunks
		for( ; ppkChunk != ppkEnd; ++ppkChunk )
		{
			if( pkIO->ParseChunk( *ppkChunk, uiParseFlags, m_pkFileManager ) >= 0 )
			{
				m_vpkSkeletonAnimations.push_back( (*ppkChunk)->m_pkAnim );

				(*ppkChunk)->m_pkAnim = 0;
			}

			delete( *ppkChunk );
		}
	}

	delete pkIO;

	return true;
}



}; // namespace NeoChunkIO


