/***************************************************************************
                  chunkio.cpp  -  I/O routines for chunk files
                             -------------------
    begin                : Tue Feb 25 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoChunkIO, chunkio.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "chunkio.h"
#include "chunk.h"
#include "factory.h"
#include "complex.h"

#include <neoengine/file.h>
#include <neoengine/logstream.h>

using namespace std;
using namespace NeoEngine;


namespace NeoChunkIO
{


Chunk *ChunkIO::ReadChunk( File *pkFile )
{
	Chunk *pkChunk = 0;

	switch( m_eMode ) {

	case ChunkIO::BINARY:
	{
		unsigned short usType = 0;
		string strID;

		*pkFile >> usType >> strID;

		if( !usType || !( pkChunk = ChunkFactory::CreateChunk( usType, strID ) ) )
			return 0;

		int iEnd = 0;

		if( pkChunk->IsComplex() )
		{
			int   iChunkSize = 0;
			short sSubChunks = 0;
		
			//Read chunk size
			*pkFile >> iChunkSize >> dynamic_cast< ComplexChunk* >( pkChunk )->m_usVersion >> sSubChunks;

			iEnd = (int)pkFile->Tellg() + iChunkSize - 4;
		
			//Read each subchunk
			for( short sSubChunk = 0; sSubChunk < sSubChunks; ++sSubChunk )
			{
				Chunk *pkSubChunk = ReadChunk( pkFile );

				if( pkSubChunk )
					pkChunk->AttachChunk( pkSubChunk );
			}
		}
		//else
			//iEnd = pkFile->Tell() + pkChunk->GetSize(); //Strings can't know it's size

		//Read remainer data
		if( pkChunk->ReadData( pkFile, m_eMode, iEnd ) < 0 )
		{
			neolog << LogLevel( ERROR ) << "*** Unable to read chunk data" << endl;
			delete pkChunk;
			return 0;
		}

		if( pkChunk->IsComplex() && ( (int)pkFile->Tellg() != iEnd ) )
		{
			neolog << LogLevel( ERROR ) << "*** [" << ( iEnd - (int)pkFile->Tellg() ) << "] bytes of unread data for chunk type [" << pkChunk->GetType() << "] [" << pkChunk->GetTypeAsString() << "]" << endl;
			pkFile->Seekg( iEnd, ios_base::beg );
		}

		break;
	}


	case ChunkIO::ASCII:
	default:
	{
		//
		// Chunk header is on format (no newlines allowed inside header). Version only for complex chunks
		//   type ( id )
		//   type ( id ) : version
		//
		// Optional subchunk block (if no subchunks, _no_ {} blocks)
		//   { subchunk } { subchunk } ...
		//
		// Data header is on format
		//   [
		//
		// Data terminator is on format
		//   ]
		//
		// Example chunks
		//
		//   node ( mynode ) []
		//
		//   node ( mynodewithsubchunk ) : 1
		//   {
		//      quat ( rot ) [ 0.0 0.0 0.0 1.0 ] 
		//   } []

		string strType;
		string strSep;
		string strID;

		if( !( *pkFile >> strType >> strSep ) )
			break; //EOF
		
		if( strSep != "()" )
			*pkFile >> strID;
		
		if( strID.length() && strID[0] != ')' )
			*pkFile >> strSep;
		else
			strID = "";

		pkChunk = ChunkFactory::CreateChunk( strType, strID );

		if( !pkChunk )
			return 0;

		if( pkChunk->IsComplex() )
			*pkFile >> strSep >> dynamic_cast< ComplexChunk* >( pkChunk )->m_usVersion;

		while( true )
		{
			if( !( *pkFile >> strSep ) )
			{
				assert( !"Invalid chunk format" );
				break;
			}

			if( pkChunk->IsComplex() && ( strSep[0] == '{' ) )
			{
				//Read subchunks
				Chunk *pkSubChunk = ReadChunk( pkFile );

				if( pkSubChunk )
					pkChunk->AttachChunk( pkSubChunk );
		
				*pkFile >> strSep;
			}
			else if( strSep[0] == '[' )
			{
				pkChunk->ReadData( pkFile, ChunkIO::ASCII, 0 );

				if( ( strSep.length() < 2 ) || ( strSep[1] != ']' ) )
					*pkFile >> strSep;

				break;
			}
			else
			{
				assert( !"Invalid chunk format" );
				break;
			}
		}

		break;
	}

	}

	return pkChunk;
}


void ChunkIO::WriteChunk( Chunk *pkChunk, File *pkFile, unsigned int uiLevel )
{
	switch( m_eMode ) {

	case BINARY:
	{
		*pkFile << pkChunk->GetType() << pkChunk->GetID();

		if( pkChunk->IsComplex() )
		{
			//Write chunk size
			*pkFile << pkChunk->GetSize( false ) << dynamic_cast< ComplexChunk* >( pkChunk )->m_usVersion << (unsigned short)pkChunk->GetSubChunks().size();

			vector< Chunk* >::const_iterator ppkChunk = pkChunk->GetSubChunks().begin();
			vector< Chunk* >::const_iterator ppkEnd   = pkChunk->GetSubChunks().end();
			
			for( ; ppkChunk != ppkEnd; ++ppkChunk )
				WriteChunk( *ppkChunk, pkFile );
		}

		pkChunk->WriteData( pkFile, m_eMode, 0 );

		break;
	}

	default:
	case ASCII:
	{
		if( pkChunk->IsComplex() )
			*pkFile << string( uiLevel * 2, ' ' );

		*pkFile << pkChunk->GetTypeAsString() << " ( " << pkChunk->GetID() << " ) ";

		if( pkChunk->IsComplex() )
		{
			//Write version
			*pkFile << ": " << dynamic_cast< ComplexChunk* >( pkChunk )->m_usVersion << endl;

			//Write subchunks
			vector< Chunk* >::const_iterator ppkChunk = pkChunk->GetSubChunks().begin();
			vector< Chunk* >::const_iterator ppkEnd   = pkChunk->GetSubChunks().end();
			
			for( ; ppkChunk != ppkEnd; ++ppkChunk )
			{
				*pkFile << string( uiLevel * 2, ' ' );

				if( (*ppkChunk)->IsComplex() )
					*pkFile << "{" << endl;
				else
					*pkFile << "{ ";

				WriteChunk( *ppkChunk, pkFile, uiLevel + 1 );

				if( (*ppkChunk)->IsComplex() )
					*pkFile << string( uiLevel * 2, ' ' ) << "}" << endl;
				else
					*pkFile << " }" << endl;
			}

			*pkFile << string( uiLevel * 2, ' ' );
		}

		*pkFile << "[ ";

		int iDataBytes = pkChunk->WriteData( pkFile, m_eMode, uiLevel + 1 );

		if( pkChunk->IsComplex() )
		{
			if( iDataBytes > 0 )
				*pkFile << string( uiLevel * 2, ' ' );
	
			*pkFile << "]" << endl;
		}
		else
			*pkFile << " ]";

		break;
	}

	}
}


int ChunkIO::ParseChunk( Chunk *pkChunk, unsigned int uiFlags, FileManager *pkFileManager )
{
	return( pkChunk->m_bParse ? pkChunk->ParseRecursive( uiFlags, pkFileManager ) : 1 );
}


}; // namespace NeoChunkIO

