/***************************************************************************
                      factory.cpp  -  Chunk factory
                             -------------------
    begin                : Tue Feb 25 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoChunkIO, factory.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "factory.h"
#include "chunktype.h"
#include "bone.h"
#include "bool.h"
#include "color.h"
#include "float.h"
#include "material.h"
#include "matrix.h"
#include "nodeanimation.h"
#include "nodekeyframe.h"
#include "quaternion.h"
#include "scenenode.h"
#include "short.h"
#include "skeleton.h"
#include "skeletonanimation.h"
#include "submeshanimation.h"
#include "submeshkeyframe.h"
#include "vertexbuffer.h"
#include "integer.h"
#include "polygonbuffer.h"
#include "polygonstrip.h"
#include "stdstring.h"
#include "texlayer.h"
#include "texmatrixgen.h"
#include "vector3d.h"
#include "skinvertexbuffer.h"
#include "boundingvolume.h"
#include "skin.h"
#include "submesh.h"
#include "mesh.h"
#include "meshentity.h"
#include "room.h"
#include "skydome.h"
#include "skybox.h"
#include "light.h"
#include "technique.h"
#include "shader.h"
#include "notify.h"

using namespace std;
using namespace NeoEngine;

namespace NeoChunkIO
{


HashTable< ChunkFactoryCallback  >    *ChunkFactory::s_pkStringTypeCallbacks = 0;
std::vector< ChunkFactoryCallback* >  *ChunkFactory::s_pvpkShortTypeCallbacks = 0;
bool                                   ChunkFactory::s_bHaveDefaultTypes     = false;




void ChunkFactory::RegisterDefaultTypes()
{
	if( s_bHaveDefaultTypes )
		return;

	s_bHaveDefaultTypes = true;

	neolog << LogLevel( INFO ) << "Registering default chunk types" << endl;

	//Register all default types
	RegisterChunkType( ChunkType::STRING,             "str",             StringChunk::Allocator            );
	RegisterChunkType( ChunkType::INTEGER,            "int",             IntChunk::Allocator               );
	RegisterChunkType( ChunkType::FLOAT,              "flt",             FloatChunk::Allocator             );
	RegisterChunkType( ChunkType::SHORT,              "sht",             ShortChunk::Allocator             );
	RegisterChunkType( ChunkType::BOOL,               "bool",            BoolChunk::Allocator              );
	RegisterChunkType( ChunkType::VECTOR,             "vec",             VectorChunk::Allocator            );
	RegisterChunkType( ChunkType::QUATERNION,         "quat",            QuaternionChunk::Allocator        );
	RegisterChunkType( ChunkType::MATRIX,             "mat",             MatrixChunk::Allocator            );
	RegisterChunkType( ChunkType::COLOR,              "col",             ColorChunk::Allocator             );

	RegisterChunkType( ChunkType::SCENENODE,          "node",            SceneNodeChunk::Allocator         );
	RegisterChunkType( ChunkType::ANIMATEDNODE,       "animnode",        SceneNodeChunk::Allocator         );
	RegisterChunkType( ChunkType::PHYSICSNODE,        "physnode",        SceneNodeChunk::Allocator         );
	RegisterChunkType( ChunkType::MASSPARTICLENODE,   "massnode",        SceneNodeChunk::Allocator         );
	RegisterChunkType( ChunkType::RIGIDBODYNODE,      "rigidnode",       SceneNodeChunk::Allocator         );

	RegisterChunkType( ChunkType::BONE,               "bone",            BoneChunk::Allocator              );
	RegisterChunkType( ChunkType::SKELETON,           "skeleton",        SkeletonChunk::Allocator          );

	RegisterChunkType( ChunkType::NODEKEYFRAME,       "nodekey",         NodeKeyframeChunk::Allocator      );
	RegisterChunkType( ChunkType::SUBMESHKEYFRAME,    "submeshkey",      SubMeshKeyframeChunk::Allocator   );

	RegisterChunkType( ChunkType::NODEANIMATION,      "nodeanim",        NodeAnimationChunk::Allocator     );
	RegisterChunkType( ChunkType::SUBMESHANIMATION,   "submeshanim",     SubMeshAnimationChunk::Allocator  );
	RegisterChunkType( ChunkType::SKELETONANIMATION,  "skeletonanim",    SkeletonAnimationChunk::Allocator );

	RegisterChunkType( ChunkType::MATERIAL,           "material",        MaterialChunk::Allocator          );
	RegisterChunkType( ChunkType::TEXTURELAYER,       "texlayer",        TextureLayerChunk::Allocator      );
	RegisterChunkType( ChunkType::TEXMATRIXGEN,       "texmatrixgen",    TextureMatrixGenChunk::Allocator  );
	RegisterChunkType( ChunkType::TECHNIQUE,          "technique",       TechniqueChunk::Allocator         );
	RegisterChunkType( ChunkType::SHADER,              "shader",         ShaderChunk::Allocator           );
	RegisterChunkType( ChunkType::PASS,               "pass",            TechniquePassChunk::Allocator     );

	RegisterChunkType( ChunkType::VERTEXBUFFER,       "vertexbuf",       VertexBufferChunk::Allocator      );
	RegisterChunkType( ChunkType::POLYGONBUFFER,      "polygonbuf",      PolygonBufferChunk::Allocator     );
	RegisterChunkType( ChunkType::POLYGONSTRIPBUFFER, "stripbuf",        PolygonStripChunk::Allocator      );
	RegisterChunkType( ChunkType::SKINVERTEXBUFFER,   "skinvertexbuf",   SkinVertexBufferChunk::Allocator  );

	RegisterChunkType( ChunkType::AABB,               "aabb",            BoundingVolumeChunk::Allocator    );
	RegisterChunkType( ChunkType::OBB,                "obb",             BoundingVolumeChunk::Allocator    );
	RegisterChunkType( ChunkType::SPHERE,             "sphere",          BoundingVolumeChunk::Allocator    );

	RegisterChunkType( ChunkType::SKIN,               "skin",            SkinChunk::Allocator              );

	RegisterChunkType( ChunkType::SUBMESH,            "submesh",         SubMeshChunk::Allocator           );
	RegisterChunkType( ChunkType::ANIMATEDSUBMESH,    "animatedsubmesh", SubMeshChunk::Allocator           );
	RegisterChunkType( ChunkType::PATCHSUBMESH,       "patchsubmesh",    SubMeshChunk::Allocator           );
	RegisterChunkType( ChunkType::SKELETALSUBMESH,    "skeletalsubmesh", SubMeshChunk::Allocator           );

	//Old type name first
	RegisterChunkType( ChunkType::MESH,               "meshblp",         MeshChunk::Allocator              );
	RegisterChunkType( ChunkType::MESH,               "mesh",            MeshChunk::Allocator              );
	RegisterChunkType( ChunkType::MESHENTITY,         "meshinstance",    MeshEntityChunk::Allocator        );
	RegisterChunkType( ChunkType::MESHENTITY,         "meshentity",      MeshEntityChunk::Allocator        );

	RegisterChunkType( ChunkType::SKYDOME,            "skydome",         SkyDomeChunk::Allocator           );
	RegisterChunkType( ChunkType::SKYBOX,             "skybox",          SkyBoxChunk::Allocator            );

	RegisterChunkType( ChunkType::LIGHT,              "light",           LightChunk::Allocator             );

	RegisterChunkType( ChunkType::NOTIFY,             "notify",          NotifyChunk::Allocator            );
}


void ChunkFactory::DeregisterAllTypes()
{
	vector< ChunkFactoryCallback* > vpkStringCallbacks;

	if( s_pkStringTypeCallbacks )
	{
		s_pkStringTypeCallbacks->GetAllNodeData( &vpkStringCallbacks );

		vector< ChunkFactoryCallback* >::iterator ppkCallback = vpkStringCallbacks.begin();
		vector< ChunkFactoryCallback* >::iterator ppkEnd      = vpkStringCallbacks.end();

		for( ; ppkCallback != ppkEnd; ++ppkCallback )
			delete( *ppkCallback );

		s_pkStringTypeCallbacks->Clear();
	}

	if( s_pvpkShortTypeCallbacks )
	{
		vector< ChunkFactoryCallback* >::iterator ppkCallback = s_pvpkShortTypeCallbacks->begin();
		vector< ChunkFactoryCallback* >::iterator ppkEnd      = s_pvpkShortTypeCallbacks->end();

		for( ; ppkCallback != ppkEnd; ++ppkCallback )
			delete( *ppkCallback );

		s_pvpkShortTypeCallbacks->clear();
	}

	s_bHaveDefaultTypes = false;
}


void ChunkFactory::RegisterChunkType( unsigned short usType, const HashString &rstrType, CreateChunkCallback pfAllocator )
{
	if( !s_pkStringTypeCallbacks )
		s_pkStringTypeCallbacks = new HashTable< ChunkFactoryCallback >();

	if( !s_pvpkShortTypeCallbacks )
		s_pvpkShortTypeCallbacks = new vector< ChunkFactoryCallback* >;

	//Delete previous type
	ChunkFactoryCallback *pkCallback = new ChunkFactoryCallback;

	pkCallback->m_usType      = usType;
	pkCallback->m_strType     = rstrType;
	pkCallback->m_pfAllocator = pfAllocator;

	delete s_pkStringTypeCallbacks->Insert( rstrType, pkCallback );

	pkCallback = new ChunkFactoryCallback;

	pkCallback->m_usType      = usType;
	pkCallback->m_strType     = rstrType;
	pkCallback->m_pfAllocator = pfAllocator;

	vector< ChunkFactoryCallback* >::iterator ppkCallback = s_pvpkShortTypeCallbacks->begin();
	vector< ChunkFactoryCallback* >::iterator ppkEnd      = s_pvpkShortTypeCallbacks->end();

	for( ; ppkCallback != ppkEnd; ++ppkCallback )
	{
		if( (*ppkCallback)->m_usType == usType )
		{
			delete( *ppkCallback );
			*ppkCallback = pkCallback;
			break;
		}
	}

	if( ppkCallback == ppkEnd )
		s_pvpkShortTypeCallbacks->push_back( pkCallback );
}


Chunk *ChunkFactory::CreateChunk( unsigned short usType, const HashString &rstrID )
{
	//Locate type in vector
	vector< ChunkFactoryCallback* >::iterator ppkAllocator = s_pvpkShortTypeCallbacks->begin();
	vector< ChunkFactoryCallback* >::iterator ppkEnd       = s_pvpkShortTypeCallbacks->end();

	for( ; ppkAllocator != ppkEnd; ++ppkAllocator )
	{
		if( (*ppkAllocator)->m_usType == usType )
			return (*ppkAllocator)->m_pfAllocator( usType, (*ppkAllocator)->m_strType, rstrID );
	}

	neolog << LogLevel( WARNING ) << "*** Unknown chunk type [" << usType << "]" << endl;

	return 0;
}


Chunk *ChunkFactory::CreateChunk( const HashString &rstrType, const HashString &rstrID )
{
	ChunkFactoryCallback *pkCallback = s_pkStringTypeCallbacks->Find( rstrType );

	if( pkCallback )
		return pkCallback->m_pfAllocator( pkCallback->m_usType, rstrType, rstrID );

	neolog << LogLevel( WARNING ) << "*** Unknown chunk type [" << rstrType << "]" << endl;

	return 0;
}



}; // namespace NeoChunkIO

