/***************************************************************************
                    polygonbuffer.cpp  -  Polygonbuffer chunk
                             -------------------
    begin                : Wed Feb 26 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoChunkIO, polygonbuffer.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "polygonbuffer.h"
#include "chunktype.h"
#include "chunkio.h"
#include "integer.h"
#include "stdstring.h"
#include "bool.h"
#include "polygonstrip.h"
#include "material.h"
#include "vertexbuffer.h"

#include <neoengine/render.h>
#include <neoengine/file.h>
#include <neoengine/strutil.h>
#include <neoengine/logstream.h>

using namespace std;
using namespace NeoEngine;


namespace NeoChunkIO
{


int PolygonBufferChunk::GetSize( bool bIncludeHeader ) const
{
	int iSize = ComplexChunk::GetSize( bIncludeHeader );

	iSize += sizeof( unsigned short ) * 3 * m_pkBuffer->GetNumElements();

	return iSize;
}


int PolygonBufferChunk::ReadData( File *pkFile, ChunkIO::CHUNKIOMODE eMode, unsigned int uiBytes )
{
	int iOffset = pkFile->Tellg();

	unsigned int uiType        = Buffer::NORMAL;
	unsigned int uiNumPolygons = 0;

	//It is safe to get primitive subchunks, since they need no processing
	Chunk *pkChunk;

	if( !( pkChunk = FindChunk( "num", ChunkType::INTEGER, -1 ) ) )
	{
		neolog << LogLevel( ERROR ) << "*** Invalid polygonbuffer chunk: no num subchunk" << endl;
		return -1;
	}
	else
		uiNumPolygons = (unsigned int)dynamic_cast< IntChunk* >( pkChunk )->m_iData;


	if( ( pkChunk = FindChunk( "type", ChunkType::STRING, -1 ) ) )
	{
		//Parse string
		vector<string> vstrFlags;

		Explode( " \t", dynamic_cast< StringChunk* >( pkChunk )->m_strData, &vstrFlags );

		unsigned int uiFlags = 0;

		vector<string>::iterator pstrFlag = vstrFlags.begin();
		vector<string>::iterator pstrEnd  = vstrFlags.end();

		for( ; pstrFlag != pstrEnd; ++pstrFlag )
		{
			//Strip whitespace
			unsigned int uiStart = pstrFlag->find_first_not_of( " \t" );
			unsigned int uiEnd   = pstrFlag->find_last_not_of( " \t" ) + 1;

			*pstrFlag = pstrFlag->substr( uiStart, uiEnd - uiStart );

			//Ignore empty strings
			if( !pstrFlag->length() )
				continue;

			if( *pstrFlag == "dynamic" )
				uiFlags |= Buffer::DYNAMIC;
			else if( *pstrFlag == "static" )
				uiFlags |= Buffer::STATIC;
			else if( *pstrFlag == "readpri" )
				uiFlags |= Buffer::READPRIORITIZED;
			else if( *pstrFlag == "writepri" )
				uiFlags |= Buffer::WRITEPRIORITIZED;
			// ### FIXME: Add support for loading noreadwrite buffers (read to temporary buffer)
			//else if( *pstrFlag == "noreadwrite" )
			//	uiFlags |= Buffer::NOREADWRITE;
			else if( *pstrFlag == "norender" )
				uiFlags |= Buffer::NORENDER;
		}

		uiType = uiFlags;
	}


	//Create buffer
	if( Core::Get()->GetRenderDevice() )
		m_pkBuffer = Core::Get()->GetRenderDevice()->CreatePolygonBuffer( uiType, uiNumPolygons );
	else
		m_pkBuffer = new PolygonBuffer( uiType, uiNumPolygons );
	
	m_pkBuffer->Lock( Buffer::WRITE );

	NeoEngine::Polygon *pkPoly = m_pkBuffer->GetPolygon();

	if( ( pkFile->GetByteOrder() == Core::BYTEORDER_SYSTEM ) && ( eMode == ChunkIO::BINARY ) )
		pkFile->Read( pkPoly, 6 * uiNumPolygons );
	else for( unsigned int uiPoly = 0; uiPoly < uiNumPolygons; ++uiPoly, ++pkPoly )
		*pkFile >> pkPoly->v[0] >> pkPoly->v[1] >> pkPoly->v[2];

	m_pkBuffer->Unlock();
	
	return( !*pkFile ? -1 : ( (int)pkFile->Tellg() - iOffset ) );
}


int PolygonBufferChunk::ParseData( unsigned int uiFlags, FileManager *pkFileManager )
{
	//Locate attached strips
	vector< Chunk* > vpkStripChunks;

	FindChunks( "", ChunkType::POLYGONSTRIPBUFFER, &vpkStripChunks, false );

	if( vpkStripChunks.size() )
	{
		if( vpkStripChunks.size() > 1 )
			neolog << LogLevel( WARNING ) << "*** Invalid polygonstrip chunk: Multiple strip buffers, grabbing first strip only" << endl;

		PolygonStripChunk *pkStripChunk = dynamic_cast< PolygonStripChunk* >( vpkStripChunks[0] );

		m_pkBuffer->m_pkStrip = pkStripChunk->m_pkStrip;
			
		pkStripChunk->m_pkStrip = 0;
	}
	else
	{
		//Locate flag chunk
		Chunk *pkFlagChunk = FindChunk( "stripify", ChunkType::BOOL, -1 );

		if( ( pkFlagChunk && dynamic_cast< BoolChunk* >( pkFlagChunk )->m_bData ) || ( uiFlags & ChunkIO::STRIPIFYALL ) )
			m_pkBuffer->Stripify();
	}

	Chunk *pkChunk;

	//See if attached material
	if( ( pkChunk = FindChunk( "", ChunkType::MATERIAL, -1 ) ) )
		m_pkBuffer->m_pkMaterial = dynamic_cast< MaterialChunk* >( pkChunk )->m_pkMaterial;

	//See if attached vertex buffer
	if( ( pkChunk = FindChunk( "", ChunkType::VERTEXBUFFER, -1 ) ) )
		m_pkBuffer->m_pkVertexBuffer = dynamic_cast< VertexBufferChunk* >( pkChunk )->m_pkBuffer;

	return 1;
}


int PolygonBufferChunk::WriteData( File *pkFile, ChunkIO::CHUNKIOMODE eMode, unsigned int uiLevel )
{
	int iOffset = pkFile->Tellp();

	unsigned int uiNumPolygons = m_pkBuffer->GetNumElements();

	m_pkBuffer->Lock( Buffer::READ );

	NeoEngine::Polygon *pkPoly = m_pkBuffer->GetPolygon();

	if( eMode == ChunkIO::BINARY )
	{
		for( unsigned int uiPoly = 0; uiPoly < uiNumPolygons; ++uiPoly, ++pkPoly )
			*pkFile << pkPoly->v[0] << pkPoly->v[1] << pkPoly->v[2];
	}
	else
	{
		++uiLevel;

		string strSpacer( uiLevel * 2, ' ' );

		for( unsigned int uiPoly = 0; uiPoly < uiNumPolygons; ++uiPoly, ++pkPoly )
			*pkFile << endl << strSpacer << pkPoly->v[0] <<  " " << pkPoly->v[1] << " " << pkPoly->v[2];

		*pkFile << endl << string( ( uiLevel - 3 ) * 2, ' '  );
	}

	m_pkBuffer->Unlock();
	
	return( !*pkFile ? -1 : ( (int)pkFile->Tellp() - iOffset ) );
}


void PolygonBufferChunk::SetNumChunk( unsigned int uiNumElements )
{
	AttachChunk( new IntChunk( (int)uiNumElements, "num" ), true );
}


void PolygonBufferChunk::SetTypeChunk( unsigned int uiType )
{
	AttachChunk( new StringChunk( Buffer::GetTypeAsString( uiType ), "type" ), true );
}



}; // namespace NeoChunkIO

