/***************************************************************************
                    polygonstrip.cpp  -  Polygon strip chunk
                             -------------------
    begin                : Wed Feb 26 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoChunkIO, polygonstrip.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "polygonstrip.h"
#include "chunktype.h"
#include "integer.h"
#include "stdstring.h"

#include <neoengine/render.h>
#include <neoengine/file.h>
#include <neoengine/strutil.h>
#include <neoengine/logstream.h>



using namespace std;
using namespace NeoEngine;

namespace NeoChunkIO
{


int PolygonStripChunk::GetSize( bool bIncludeHeader ) const
{
	int iSize = ComplexChunk::GetSize( bIncludeHeader );

	iSize += ( m_pkStrip->GetNumElements() + 2 ) * sizeof( unsigned short );

	return iSize;
}


int PolygonStripChunk::ReadData( File *pkFile, ChunkIO::CHUNKIOMODE eMode, unsigned int uiBytes )
{
	int iOffset = pkFile->Tellg();

	unsigned int uiType        = Buffer::NORMAL;
	unsigned int uiNumPolygons = 0;


	//It is safe to get primitive subchunks, since they need no processing
	Chunk *pkChunk;

	if( !( pkChunk = FindChunk( "num", ChunkType::INTEGER, -1 ) ) )
	{
		neolog << LogLevel( ERROR ) << "*** Invalid polygonstrip chunk: no num subchunk" << endl;
		return -1;
	}
	else
		uiNumPolygons = dynamic_cast< IntChunk* >( pkChunk )->m_iData;

	if( ( pkChunk = FindChunk( "type", ChunkType::STRING, -1 ) ) )
	{
		//Parse string
		vector<string> vstrFlags;

		Explode( " \t", dynamic_cast< StringChunk* >( pkChunk )->m_strData, &vstrFlags );

		unsigned int uiFlags = 0;

		vector<string>::iterator pstrFlag = vstrFlags.begin();
		vector<string>::iterator pstrEnd  = vstrFlags.end();

		for( ; pstrFlag != pstrEnd; ++pstrFlag )
		{
			//Strip whitespace
			unsigned int uiStart = pstrFlag->find_first_not_of( " \t" );
			unsigned int uiEnd   = pstrFlag->find_last_not_of( " \t" ) + 1;

			*pstrFlag = pstrFlag->substr( uiStart, uiEnd - uiStart );

			//Ignore empty strings
			if( !pstrFlag->length() )
				continue;

			if( *pstrFlag == "dynamic" )
				uiFlags |= Buffer::DYNAMIC;
			else if( *pstrFlag == "STATIC" )
				uiFlags |= Buffer::STATIC;
			else if( *pstrFlag == "readpri" )
				uiFlags |= Buffer::READPRIORITIZED;
			else if( *pstrFlag == "writepri" )
				uiFlags |= Buffer::WRITEPRIORITIZED;
			// ### FIXME: Add support for loading noreadwrite buffers (read to temporary buffer)
			//else if( *pstrFlag == "noreadwrite" )
			//	uiFlags |= Buffer::NOREADWRITE;
			else if( *pstrFlag == "norender" )
				uiFlags |= Buffer::NORENDER;
		}

		uiType = uiFlags;
	}


	//Create buffer
	if( Core::Get()->GetRenderDevice() )
		m_pkStrip = Core::Get()->GetRenderDevice()->CreatePolygonStripBuffer( uiType, uiNumPolygons );
	else
		m_pkStrip = new PolygonStripBuffer( uiType, uiNumPolygons );

	m_pkStrip->Lock( Buffer::WRITE );

	unsigned short *pusDest = m_pkStrip->GetIndices();

	if( ( pkFile->GetByteOrder() == Core::BYTEORDER_SYSTEM ) && ( eMode == ChunkIO::BINARY ) )
		pkFile->Read( pusDest, 2 * ( uiNumPolygons + 2 ) );
	else for( unsigned int uiIndex = 0; uiIndex < ( uiNumPolygons + 2 ); ++uiIndex, ++pusDest )
		*pkFile >> *pusDest;

	m_pkStrip->Unlock();
		
	return( !*pkFile ? -1 : ( (int)pkFile->Tellg() - iOffset ) );
}


int PolygonStripChunk::WriteData( File *pkFile, ChunkIO::CHUNKIOMODE eMode, unsigned int uiLevel )
{
	int iOffset = pkFile->Tellp();

	m_pkStrip->Lock( Buffer::READ );

	unsigned short *pusSrc = m_pkStrip->GetIndices();

	if( eMode == ChunkIO::BINARY )
	{
		for( unsigned int uiIndex = 0; uiIndex < ( m_pkStrip->GetNumElements() + 2 ); ++uiIndex, ++pusSrc )
			*pkFile << *pusSrc;
	}
	else if( eMode == ChunkIO::ASCII )
	{
		++uiLevel;

		string strSpacer( uiLevel * 2, ' ' );

		for( unsigned int uiIndex = 0; uiIndex < ( m_pkStrip->GetNumElements() + 2 ); ++uiIndex, ++pusSrc )
		{
			if( !( uiIndex % 10 ) )
				*pkFile << endl << strSpacer;
			else
				*pkFile << " ";

			*pkFile << *pusSrc;
		}

		*pkFile << endl << string( ( uiLevel - 3 ) * 2, ' ' );
	}	
	
	m_pkStrip->Unlock();

	return( !*pkFile ? -1 : ( (int)pkFile->Tellp() - iOffset ) );
}


void PolygonStripChunk::SetNumChunk( unsigned int uiNumElements )
{
	AttachChunk( new IntChunk( (int)uiNumElements, "num" ), true );
}


void PolygonStripChunk::SetTypeChunk( unsigned int uiType )
{
	AttachChunk( new StringChunk( Buffer::GetTypeAsString( uiType ), "type" ), true );
}


};
