/***************************************************************************
                        scene.cpp  -  Scene file loader
                             -------------------
    begin                : Thu Nov 1 2001
    copyright            : (C) 2001 by Mattias Jansson
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoChunkIO, scene.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "scene.h"
#include "chunkio.h"
#include "chunktype.h"
#include "room.h"
#include "scenenode.h"
#include "mesh.h"

#include <neoengine/file.h>
#include <neoengine/room.h>
#include <neoengine/logstream.h>
#include <neoengine/util.h>


using namespace std;
using namespace NeoEngine;


namespace NeoChunkIO
{



Scene::Scene( FileManager *pkFileManager ) :
	LoadableEntity( pkFileManager ),
	SceneNode( "scene" )
{
}


Scene::~Scene()
{
	vector<Room*>::iterator ppkRoom = m_vpkRooms.begin();
	vector<Room*>::iterator ppkEnd  = m_vpkRooms.end();

	for( ; ppkRoom != ppkEnd; ++ppkRoom )
		delete( *ppkRoom );
}


MeshPtr Scene::GetMesh( const HashString &rstrName )
{
	for( unsigned int i = 0; i < m_vpkMeshes.size(); ++i )
		if( rstrName == m_vpkMeshes[i]->GetName() )
			return m_vpkMeshes[i];

	return 0;
}


bool Scene::LoadNode( unsigned int uiFlags )
{
	m_uiLoadFlags = uiFlags;

	if( !m_pkFile || !m_pkFile->IsValid() )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to load scene: Invalid file" << endl;
		return false;
	}

	SetName( m_pkFile->GetName().substr( 0, m_pkFile->GetName().find_last_of( "." ) ) + "_scene" );

	if( !m_pkFile->DetermineByteOrder( 0x4543534e ) )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to load scene: Failed to determine byte order, possible corrupt file" << endl;
		return false;
	}

	char szStr[5]; szStr[4] = 0;

	m_pkFile->Read( szStr, 4 );

	if( string( szStr ) != "NSCE" )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to load scene: Invalid ID" << endl;
		return false;
	}

	ChunkIO *pkIO = new ChunkIO;

	//Read mode string
	m_pkFile->Read( szStr, 4 );

	if( string( szStr ) == "!bin" )
	{
		pkIO->m_eMode = ChunkIO::BINARY;
		m_pkFile->SetBinary( true );
	}
	else if( string( szStr ) == "!txt" )
	{
		pkIO->m_eMode = ChunkIO::ASCII;
		m_pkFile->SetBinary( false );
	}
	else
	{
		neolog << LogLevel( ERROR ) << "*** Unable to load scene: Unknown chunk file mode [" << szStr << "]" << endl;
		return false;
	}


	*m_pkFile >> pkIO->m_iMajorVersion >> pkIO->m_iMinorVersion;

	if( !( pkIO->m_iMajorVersion == ChunkIO::MAJORVERSIONREQUIRED ) || ( pkIO->m_iMinorVersion < ChunkIO::MINORVERSIONREQUIRED ) )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to load scene: Invalid chunk format version " << pkIO->m_iMajorVersion << "." << pkIO->m_iMinorVersion << endl;
		return false;
	}

	vector<Chunk*> vpkChunks;
	Chunk         *pkChunk = 0;

	//Read chunk tree
	while( ( pkChunk = pkIO->ReadChunk( m_pkFile ) ) )
		vpkChunks.push_back( pkChunk );


	unsigned int uiParseFlags = 0;

	if( uiFlags & STRIPIFYALL )
		uiParseFlags |= ChunkIO::STRIPIFYALL;

	//Now parse chunk tree
	vector<Chunk*>::iterator ppkChunk = vpkChunks.begin();
	vector<Chunk*>::iterator ppkEnd   = vpkChunks.end();

	for( ; ppkChunk != ppkEnd; )
	{
		if( pkIO->ParseChunk( *ppkChunk, uiParseFlags, m_pkFileManager ) < 0 )
		{
			delete( *ppkChunk );
			vpkChunks.erase( ppkChunk );
			ppkEnd = vpkChunks.end();
		}
		else
			++ppkChunk;
	}


	//Locate all meshes
	ppkChunk = vpkChunks.begin();

	for( ; ppkChunk != ppkEnd; ++ppkChunk )
	{
		if( (*ppkChunk)->FindChunk( "", ChunkType::MESH, 0 ) )
		{
			MeshPtr pkMesh = dynamic_cast< MeshChunk* >( *ppkChunk )->m_pkMesh;

			if( pkMesh )
				m_vpkMeshes.push_back( pkMesh );
		}
	}			


	//Locate all scene nodes
	ppkChunk = vpkChunks.begin();

	for( ; ppkChunk != ppkEnd; ++ppkChunk )
	{
		if( (*ppkChunk)->FindChunk( "", ChunkType::SCENENODE, 0 ) )
		{
			SceneNodeChunk *pkNodeChunk = dynamic_cast< SceneNodeChunk* >( *ppkChunk );

			//Attach to us so nodes can locate eachother in hierarchy
			if( pkNodeChunk->m_pkNode )
				AttachNode( pkNodeChunk->m_pkNode );

			//Don't delete object
			pkNodeChunk->m_pkNode = 0;
		}
	}



	//Locate all rooms
	vector<Chunk*> vpkRoomChunks;

	ppkChunk = vpkChunks.begin();

	for( ; ppkChunk != ppkEnd; ++ppkChunk )
	{
		if( (*ppkChunk)->FindChunk( "", ChunkType::ROOM, 0 ) )
			vpkRoomChunks.push_back( *ppkChunk );
	}

	vector<Chunk*>::iterator ppkRoomChunk = vpkRoomChunks.begin();
	vector<Chunk*>::iterator ppkRoomEnd   = vpkRoomChunks.end();

	for( ; ppkRoomChunk != ppkRoomEnd; ++ppkRoomChunk )
	{
		RoomChunk *pkRoomChunk = dynamic_cast< RoomChunk* >( *ppkRoomChunk );

		Room *pkRoom = pkRoomChunk->m_pkRoom;

		pkRoomChunk->m_pkRoom = 0;

		if( pkRoom )
			m_vpkRooms.push_back( pkRoom );
	}


	ppkChunk = vpkChunks.begin();

	for( ; ppkChunk != ppkEnd; ++ppkChunk )
		delete( *ppkChunk );

	delete pkIO;

	return true;
}


};

