/***************************************************************************
                      scenenode.cpp  -  Scenenode chunks
                             -------------------
    begin                : Wed Feb 26 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoChunkIO, scenenode.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "scenenode.h"
#include "chunktype.h"
#include "float.h"
#include "stdstring.h"
#include "vector3d.h"
#include "quaternion.h"
#include "matrix.h"
#include "nodeanimation.h"
#include "sceneentity.h"

#include <neoengine/scenenode.h>
#include <neoengine/animatednode.h>
#include <neoengine/physicsnode.h>
#include <neoengine/massparticle.h>
#include <neoengine/rigidbody.h>

using namespace std;
using namespace NeoEngine;

namespace NeoChunkIO
{

SceneNodeChunk::~SceneNodeChunk()
{
	delete m_pkNode;
}


int SceneNodeChunk::ParseData( unsigned int uiFlags, FileManager *pkFileManager )
{
	Chunk *pkChunk;

	if( m_usType == ChunkType::SCENENODE )
		m_pkNode = new SceneNode;
	else if( m_usType == ChunkType::ANIMATEDNODE )
		m_pkNode = new AnimatedNode;
	else if( m_usType == ChunkType::PHYSICSNODE )
		m_pkNode = new PhysicsNode;
	else if( m_usType == ChunkType::MASSPARTICLENODE )
		m_pkNode = new MassParticle;
	else if( m_usType == ChunkType::RIGIDBODYNODE )
		m_pkNode = new RigidBody;

	if( ( pkChunk = FindChunk( "name", ChunkType::STRING, -1 ) ) )
		m_pkNode->SetName( dynamic_cast< StringChunk* >( pkChunk )->m_strData );

	if( ( pkChunk = FindChunk( "pos", ChunkType::VECTOR, -1 ) ) )
		m_pkNode->SetTranslation( dynamic_cast< VectorChunk* >( pkChunk )->m_kData );

	if( ( pkChunk = FindChunk( "rot", ChunkType::QUATERNION, -1 ) ) )
		m_pkNode->SetRotation( dynamic_cast< QuaternionChunk* >( pkChunk )->m_kData );
	else if( ( pkChunk = FindChunk( "rot", ChunkType::MATRIX, -1 ) ) )
		m_pkNode->SetRotation( Quaternion( dynamic_cast< MatrixChunk* >( pkChunk )->m_kData ) );

	if( ( pkChunk = FindChunk( "scale", ChunkType::FLOAT, -1 ) ) )
		m_pkNode->SetScaling( dynamic_cast< FloatChunk* >( pkChunk )->m_fData );

/* NOT SUPPORTED YET
	if( ( pkChunk = FindChunk( "parent", Chunk::STRING, -1 ) ) && m_pkObject->GetParent() )
	{
		//Locate parent node
		std::string &rstrName = *dynamic_cast< ChunkTarget<string>* >( pkChunk )->m_pkObject;

		SceneNode *pkParent = m_pkObject->GetParent()->GetByName( rstrName, HierarchyNode<SceneNode>::BREADTH_FIRST );

		if( pkParent )
			pkParent->AttachNode( m_pkObject ), m_pkObject = 0;
		else
			Core::Get()->Log( Log::LOGLEVEL_WARNING, "[ChunkTarget<SceneNode>::ParseData] unable to locate parent node [%s] for node [%s]\n", rstrName.c_str(), m_pkObject->GetName().c_str() );
	}
*/

	//Get attached nodes
	{
		vector<Chunk*> vpkNodeChunks;

		FindChunks( "", ChunkType::SCENENODE, &vpkNodeChunks, false );

		vector<Chunk*>::iterator ppkChunk = vpkNodeChunks.begin();
		vector<Chunk*>::iterator ppkEnd   = vpkNodeChunks.end();

		for( ; ppkChunk != ppkEnd; ++ppkChunk )
		{
			SceneNodeChunk *pkNodeChunk = dynamic_cast< SceneNodeChunk* >( *ppkChunk );

			m_pkNode->AttachNode( pkNodeChunk->m_pkNode, false );

			pkNodeChunk->m_pkNode = 0;
		}
	}

	if( m_usType == ChunkType::ANIMATEDNODE )
	{
		//Try to find animation chunks
		vector<Chunk*> vpkAnimChunks;

		FindChunks( "", ChunkType::NODEANIMATION, &vpkAnimChunks, false );

		AnimatedNode *pkNode = dynamic_cast< AnimatedNode* >( m_pkNode );

		vector<Chunk*>::iterator ppkChunk = vpkAnimChunks.begin();
		vector<Chunk*>::iterator ppkEnd   = vpkAnimChunks.end();

		for( ; ppkChunk != ppkEnd; ++ppkChunk )
		{
			NodeAnimationChunk *pkAnimChunk = dynamic_cast< NodeAnimationChunk* >( *ppkChunk );

			pkNode->AddAnimation( pkAnimChunk->m_pkAnim );

			pkAnimChunk->m_pkAnim = 0;
		}
	}
	else
	{
		if( m_usType == ChunkType::RIGIDBODYNODE )
		{
			if( ( pkChunk = FindChunk( "invtensor", ChunkType::MATRIX, -1 ) ) )
				dynamic_cast< RigidBody* >( m_pkNode )->SetInvBodyInertiaTensor( dynamic_cast< MatrixChunk* >( pkChunk )->m_kData );
		}
		
		if( ( m_usType == ChunkType::MASSPARTICLENODE ) || ( m_usType == ChunkType::RIGIDBODYNODE ) )
		{
			if( ( pkChunk = FindChunk( "mass", ChunkType::FLOAT, -1 ) ) )
				dynamic_cast< MassParticle* >( m_pkNode )->SetMass( dynamic_cast< FloatChunk* >( pkChunk )->m_fData );
			else
				dynamic_cast< MassParticle* >( m_pkNode )->SetMass( 1.0f );
		}

		if( ( m_usType == ChunkType::PHYSICSNODE ) || ( m_usType == ChunkType::MASSPARTICLENODE ) || ( m_usType == ChunkType::MASSPARTICLENODE ) )
		{
			if( ( pkChunk = FindChunk( "vel", ChunkType::VECTOR, -1 ) ) )
				dynamic_cast< PhysicsNode* >( m_pkNode )->SetVelocity( dynamic_cast< VectorChunk* >( pkChunk )->m_kData );
		}
	}

	//Find entity subchunk if any
	if( ( pkChunk = FindChunk( "", ChunkType::SCENEENTITY, -1 ) ) )
	{
		SceneEntityChunk *pkEntityChunk = dynamic_cast< SceneEntityChunk* >( pkChunk );

		m_pkNode->SetEntity( pkEntityChunk->m_pkEntity );

		pkEntityChunk->m_pkEntity = 0;
	}

	//Make all node types findable with a FindChunk search for scene nodes
	m_usFindType = ChunkType::SCENENODE;

	return 1;
}


}; // namespace NeoChunkIO

