/***************************************************************************
                     vertexbuffer.cpp  -  Vertexbuffer chunk
                             -------------------
    begin                : Wed Feb 26 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoChunkIO, vertexbuffer.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "vertexbuffer.h"
#include "chunktype.h"
#include "integer.h"
#include "stdstring.h"

#include <neoengine/vertexdecl.h>
#include <neoengine/render.h>
#include <neoengine/file.h>
#include <neoengine/strutil.h>
#include <neoengine/logstream.h>

#include <vector>
#include <string>


using namespace std;
using namespace NeoEngine;


namespace NeoChunkIO
{


int VertexBufferChunk::GetSize( bool bIncludeHeader ) const
{
	int iSize = ComplexChunk::GetSize( bIncludeHeader );

	iSize += m_pkBuffer->GetVertexSize() * m_pkBuffer->GetNumElements();

	return iSize;
}


int VertexBufferChunk::ReadData( File *pkFile, ChunkIO::CHUNKIOMODE eMode, unsigned int uiBytes )
{
	int iOffset = pkFile->Tellg();

	unsigned int uiNumVertices = 0;
	unsigned int uiType        = Buffer::NORMAL;
	bool         bHaveColor    = false;

	vector< VertexElement* > vpkElements;

	//It is safe to get primitive subchunks, since they need no processing
	Chunk *pkChunk;

	if( !( pkChunk = FindChunk( "num", ChunkType::INTEGER, -1 ) ) )
	{
		neolog << LogLevel( ERROR ) << "*** Invalid vertex buffer chunk: no num subchunk" << endl;
		return -1;
	}
	else
		uiNumVertices = dynamic_cast< IntChunk* >( pkChunk )->m_iData;

	if( !( pkChunk = FindChunk( "format", ChunkType::STRING, -1 ) ) )
	{
		neolog << LogLevel( ERROR ) << "*** Invalid vertex buffer chunk: no format subchunk" << endl;
		return -1;
	}
	else
	{
		if( m_usVersion == 1 )
		{
			//Parse string
			vector<string> vstrFlags;

			Explode( " \t", dynamic_cast< StringChunk* >( pkChunk )->m_strData, &vstrFlags );

			unsigned int uiOffset = 0;

			vector<string>::iterator pstrFlag = vstrFlags.begin();
			vector<string>::iterator pstrEnd  = vstrFlags.end();

			for( ; pstrFlag != pstrEnd; ++pstrFlag )
			{
				//Strip whitespace
				unsigned int uiStart = pstrFlag->find_first_not_of( " \t" );
				unsigned int uiEnd   = pstrFlag->find_last_not_of( " \t" ) + 1;

				*pstrFlag = pstrFlag->substr( uiStart, uiEnd - uiStart );

				//Ignore empty strings
				if( !pstrFlag->length() )
					continue;

				if( *pstrFlag == "xyz" ) { vpkElements.push_back( new VertexElement( VertexElement::FLOAT3, VertexElement::POSITION, 0, uiOffset ) ); uiOffset += 12; } 
				else if( *pstrFlag == "normal" ) { vpkElements.push_back( new VertexElement( VertexElement::FLOAT3, VertexElement::NORMAL, 0, uiOffset ) ); uiOffset += 12; }
				else if( *pstrFlag == "color" ) { vpkElements.push_back( new VertexElement( VertexElement::COLOR32, VertexElement::DIFFUSECOLOR, 0, uiOffset ) ); uiOffset += 4; bHaveColor = true; }
				else if( *pstrFlag == "tangent" ) { vpkElements.push_back( new VertexElement( VertexElement::COLOR32, VertexElement::TANGENT, 0, uiOffset ) ); uiOffset += 4; }
				else
				{
					//Parse texcoord def
					vector<string> vstrArgs;

					Explode( "(,)", *pstrFlag, &vstrArgs );

					if( vstrArgs[0] == "tex" )
					{
						vpkElements.push_back( new VertexElement( VertexElement::FLOAT + atoi( vstrArgs[2].c_str() ) - 1, VertexElement::TEXCOORD, atoi( vstrArgs[1].c_str() ), uiOffset ) );
						uiOffset += 4 * atoi( vstrArgs[2].c_str() );
					}
				}
			}
		}
		else
		{
			//Parse string
			vector<string> vstrFlags;

			Explode( ",", dynamic_cast< StringChunk* >( pkChunk )->m_strData, &vstrFlags );

			unsigned int uiOffset = 0;

			vector<string>::iterator pstrFlag = vstrFlags.begin();
			vector<string>::iterator pstrEnd  = vstrFlags.end();

			for( ; pstrFlag != pstrEnd; ++pstrFlag )
			{
				//Strip whitespace
				*pstrFlag = Strip( *pstrFlag );

				//Ignore empty strings
				if( !pstrFlag->length() )
					continue;

				//element declaration is on format "type( usage )"
				vector<string> vstrArgs;

				Explode( "(:)", *pstrFlag, &vstrArgs );

				assert( vstrArgs.size() >= 2 );

				vstrArgs[0] = Strip( vstrArgs[0] );

				unsigned int uiType = VertexElement::FLOAT3;

				if( vstrArgs[0] == "float" )
					uiType = VertexElement::FLOAT;
				else if( vstrArgs[0] == "float2" )
					uiType = VertexElement::FLOAT2;
				else if( vstrArgs[0] == "float3" )
					uiType = VertexElement::FLOAT3;
				else if( vstrArgs[0] == "float4" )
					uiType = VertexElement::FLOAT4;
				else if( vstrArgs[0] == "color32" )
				{
					uiType = VertexElement::COLOR32;
					bHaveColor = true;
				}
				else
				{
					neolog << LogLevel( ERROR ) << "*** invalid vertex element type [" << vstrArgs[0] << "]" << endl;
					return false;
				}

				unsigned int uiUsage = VertexElement::POSITION;

				vstrArgs[1] = Strip( vstrArgs[1] );

				if( vstrArgs[1] == "position" )
					uiUsage = VertexElement::POSITION;
				else if( vstrArgs[1] == "normal" )
					uiUsage = VertexElement::NORMAL;
				else if( vstrArgs[1] == "color" )
					uiUsage = VertexElement::DIFFUSECOLOR;
				else if( vstrArgs[1] == "texcoord" )
					uiUsage = VertexElement::TEXCOORD;
				else if( vstrArgs[1] == "tangent" )
					uiUsage = VertexElement::TANGENT;
				else if( vstrArgs[1] == "binormal" )
					uiUsage = VertexElement::BINORMAL;
				else
				{
					neolog << LogLevel( ERROR ) << "*** invalid vertex element usage [" << vstrArgs[1] << "]" << endl;
					return false;
				}

				unsigned int uiIndex = 0;

				if( ( vstrArgs.size() >= 3 ) && vstrArgs[2].length() )
				{
					vstrArgs[2] = Strip( vstrArgs[2] );

					uiIndex = atoi( vstrArgs[2].c_str() );
				}

				vpkElements.push_back( new VertexElement( uiType, uiUsage, uiIndex, uiOffset ) );

				uiOffset += ( uiType < VertexElement::COLOR32 ) ? ( ( uiType + 1 ) * 4 ) : 4;
			}
		}
	}

	VertexDeclaration kDecl( vpkElements.size() );

	vector< VertexElement* >::iterator ppkElement    = vpkElements.begin();
	vector< VertexElement* >::iterator ppkElementEnd = vpkElements.end();

	for( unsigned int uiElement = 0; ppkElement != ppkElementEnd; ++ppkElement, ++uiElement )
	{
		kDecl.m_pkElements[ uiElement ] = *(*ppkElement);
		delete *ppkElement;
	}

	if( !kDecl.m_uiNumElements )
	{
		neolog << LogLevel( ERROR ) << "*** Invalid vertex buffer chunk: no valid format declaration" << endl;
		return -1;
	}

	if( ( pkChunk = FindChunk( "type", ChunkType::STRING, -1 ) ) )
	{
		//Parse string
		vector<string> vstrFlags;

		Explode( " \t", dynamic_cast< StringChunk* >( pkChunk )->m_strData, &vstrFlags );

		unsigned int uiFlags = 0;

		vector<string>::iterator pstrFlag = vstrFlags.begin();
		vector<string>::iterator pstrEnd  = vstrFlags.end();

		for( ; pstrFlag != pstrEnd; ++pstrFlag )
		{
			//Strip whitespace
			unsigned int uiStart = pstrFlag->find_first_not_of( " \t" );
			unsigned int uiEnd   = pstrFlag->find_last_not_of( " \t" ) + 1;

			*pstrFlag = pstrFlag->substr( uiStart, uiEnd - uiStart );

			//Ignore empty strings
			if( !pstrFlag->length() )
				continue;

			if( *pstrFlag == "dynamic" )
				uiFlags |= Buffer::DYNAMIC;
			else if( *pstrFlag == "static" )
				uiFlags |= Buffer::STATIC;
			else if( *pstrFlag == "readpri" )
				uiFlags |= Buffer::READPRIORITIZED;
			else if( *pstrFlag == "writepri" )
				uiFlags |= Buffer::WRITEPRIORITIZED;
			// ### FIXME: Add support for loading noreadwrite buffers (read to temporary buffer)
			//else if( *pstrFlag == "noreadwrite" )
			//	uiFlags |= Buffer::NOREADWRITE;
			else if( *pstrFlag == "norender" )
				uiFlags |= Buffer::NORENDER;
		}

		uiType = uiFlags;
	}

	//Create buffer
	if( Core::Get()->GetRenderDevice() )
		m_pkBuffer = Core::Get()->GetRenderDevice()->CreateVertexBuffer( uiType, uiNumVertices, &kDecl );
	else
		m_pkBuffer = new VertexBuffer( uiType, uiNumVertices, &kDecl );
	
	m_pkBuffer->Lock( Buffer::WRITE );

	unsigned char *pucData = (unsigned char*)m_pkBuffer->GetVertex();

	// ### FIXME: For faster loading when elements of color32 type, load as a single chunk and then
	//            loop all elements, for each color32 element loop all vertices and swizzle data
	//            to render device order (only for devices with other order than chunk format)
	if( ( pkFile->GetByteOrder() == Core::BYTEORDER_SYSTEM ) && ( eMode == ChunkIO::BINARY ) && !bHaveColor )
		pkFile->Read( pucData, m_pkBuffer->GetVertexSize() * uiNumVertices );
	else for( unsigned int uiVertex = 0; uiVertex < uiNumVertices; ++uiVertex )
	{
		for( unsigned int uiElement = 0; uiElement < kDecl.m_uiNumElements; ++uiElement )
		{
			if( kDecl.m_pkElements[ uiElement ].m_uiType == VertexElement::COLOR32 )
			{
				unsigned int uiColor;
				*pkFile >> showbase >> hex >> uiColor >> noshowbase >> dec;
				*(Color32*)pucData = uiColor;
				pucData += 4;
			}
			else
			{
				for( unsigned int uiFloat = 0; uiFloat < ( kDecl.m_pkElements[ uiElement ].m_uiType + 1 ); ++uiFloat, pucData += 4 )
					*pkFile >> *((float*)pucData);
			}
		}
	}

	m_pkBuffer->Unlock();

	return( !*pkFile ? -1 : ( (int)pkFile->Tellg() - iOffset ) );
}


int VertexBufferChunk::WriteData( File *pkFile, ChunkIO::CHUNKIOMODE eMode, unsigned int uiLevel )
{
	int iOffset = pkFile->Tellp();

	unsigned int uiNumVertices = m_pkBuffer->GetNumElements();
	const VertexDeclaration &kDecl = *m_pkBuffer->GetVertexFormat();

	m_pkBuffer->Lock( Buffer::READ );

	unsigned char *pucData = (unsigned char*)m_pkBuffer->GetVertex();

	for( unsigned int uiVertex = 0; uiVertex < uiNumVertices; ++uiVertex )
	{
		if( eMode == ChunkIO::ASCII )
			*pkFile << endl << string( uiLevel * 2, ' ' );

		for( unsigned int uiElement = 0; uiElement < kDecl.m_uiNumElements; ++uiElement )
		{
			if( kDecl.m_pkElements[ uiElement ].m_uiType == VertexElement::COLOR32 )
			{
				if( eMode == ChunkIO::ASCII )
					*pkFile << hex << showbase;

				Color32 &kColor = *(Color32*)pucData;
				unsigned int uiColor = kColor[0] | ( kColor[1] << 8 ) | ( kColor[2] << 16 ) | ( kColor[3] << 24 );
				*pkFile << uiColor;
				pucData += 4;

				if( eMode == ChunkIO::ASCII )
					*pkFile << dec << noshowbase << " ";
			}
			else
			{
				if( eMode == ChunkIO::BINARY )
				{
					for( unsigned int uiFloat = 0; uiFloat < ( kDecl.m_pkElements[ uiElement ].m_uiType + 1 ); ++uiFloat, pucData += 4 )
						*pkFile << *(float*)pucData;
				}
				else
				{
					for( unsigned int uiFloat = 0; uiFloat < ( kDecl.m_pkElements[ uiElement ].m_uiType + 1 ); ++uiFloat, pucData += 4 )
						*pkFile << *(float*)pucData << " ";
				}
			}
		}
	}

	if( eMode == ChunkIO::ASCII )
		*pkFile << endl << string( ( uiLevel - 3 ) * 2, ' ' );

	m_pkBuffer->Unlock();

	return( !*pkFile ? -1 : ( (int)pkFile->Tellp() - iOffset ) );
}


void VertexBufferChunk::SetNumChunk( unsigned int uiNumElements )
{
	AttachChunk( new IntChunk( (int)uiNumElements, "num" ), true );
}


void VertexBufferChunk::SetTypeChunk( unsigned int uiType )
{
	AttachChunk( new StringChunk( Buffer::GetTypeAsString( uiType ), "type" ), true );
}


void VertexBufferChunk::SetFormatChunk( const VertexDeclaration *pkFormat )
{
	string strFormat;

	for( unsigned int uiElement = 0; uiElement < pkFormat->m_uiNumElements; ++uiElement )
	{
		const VertexElement *pkElement = &pkFormat->m_pkElements[ uiElement ];

		if( pkElement->m_uiType == VertexElement::FLOAT )
			strFormat += "float( ";
		else if( pkElement->m_uiType == VertexElement::FLOAT2 )
			strFormat += "float2( ";
		else if( pkElement->m_uiType == VertexElement::FLOAT3 )
			strFormat += "float3( ";
		else if( pkElement->m_uiType == VertexElement::FLOAT4 )
			strFormat += "float4( ";
		else if( pkElement->m_uiType == VertexElement::COLOR32 )
			strFormat += "color32( ";

		if( pkElement->m_uiUsage == VertexElement::POSITION )
			strFormat += "position";
		else if( pkElement->m_uiUsage == VertexElement::NORMAL )
			strFormat += "normal";
		else if( pkElement->m_uiUsage == VertexElement::DIFFUSECOLOR )
			strFormat += "color";
		else if( pkElement->m_uiUsage == VertexElement::TEXCOORD )
		{
			strFormat += "texcoord : ";

			char szStr[32]; sprintf( szStr, "%d", pkElement->m_uiIndex );

			strFormat += szStr;
		}
		else if( pkElement->m_uiUsage == VertexElement::TANGENT )
			strFormat += "tangent";
		else if( pkElement->m_uiUsage == VertexElement::BINORMAL )
			strFormat += "binormal";

		strFormat += " ), ";
	}

	strFormat.erase( strFormat.end() - 1 );
	strFormat.erase( strFormat.end() - 1 );

	StringChunk *pkChunk = new StringChunk( strFormat, "format" );

	AttachChunk( pkChunk, true );
}


}; // namespace NeoChunkIO
