/***************************************************************************
                          sound.cpp  -  aRts output
                             -------------------
    begin                : Sun Jun 06 2004
    copyright            : (C) 2004 by Adam Butcher
    email                : adam.j.butcher@virgin.net
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoDevALSA, sound.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2004
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "sound.h"

#include <neoengine/file.h>
#include <neoengine/filemanager.h>
#include <neoengine/logstream.h>

#include <artsc.h>
#include <pthread.h>
#include <errno.h>

using namespace NeoEngine;
using namespace std;


namespace NeoArts
{


NeoEngine::Sound *AudioDevice::LoadSound( const std::string &rstrFilename )
{
	vector< SoundCodec* >::iterator ppkCodec    = m_vpkCodecs.begin();
	vector< SoundCodec* >::iterator ppkCodecEnd = m_vpkCodecs.end();

	for( ; ppkCodec != ppkCodecEnd; ++ppkCodec )
	{
		string strFilename = rstrFilename + ".";

		const vector<string> &rvstrExtensions = (*ppkCodec)->GetExtensions();

		vector<string>::const_iterator pstrExt    = rvstrExtensions.begin();
		vector<string>::const_iterator pstrExtEnd = rvstrExtensions.end();

		for( ; pstrExt != pstrExtEnd; ++pstrExt )
		{
			File        *pkFile   = m_pkFileManager->GetByName( strFilename + (*pstrExt) );
			SoundStream *pkStream = 0;

			if( pkFile && pkFile->Open() && (*ppkCodec)->IsType( pkFile ) && ( pkStream = (*ppkCodec)->GetStream( pkFile ) ) )
				return new Sound( this, rstrFilename, pkStream );
			else
				delete pkFile;
		}
	}

	neolog << LogLevel( WARNING ) << "*** unable to load sound [" << rstrFilename << "]" << endl;

	return 0;
}



Sound::Sound( AudioDevice *pkDevice, const std::string &rstrName, SoundStream *pkStream ) :
	NeoEngine::Sound( rstrName, pkStream ),
	m_pkDevice( pkDevice ),
	m_pkThread( 0 ),
	m_bPaused( false )
{
}


Sound::~Sound()
{
	delete m_pkThread;
}


void Sound::Play( bool bLoop )
{
	m_bLoop   = bLoop;
	
	if( m_pkThread )
	{
		if( m_pkThread->m_bTerminated )
			delete m_pkThread, m_pkThread = 0;
		else
			return;
	}

	m_pkThread = new Thread( this, 0 );
}


void *Sound::ThreadEntry( Thread *pkThread, void *pData )
{
	pkThread->m_bStarted = true;

	arts_stream_t stream;
	
	unsigned int uiRate = m_pkStream->GetSamplesPerSecond();
	unsigned int uiBitsPerSample = m_pkStream->GetBitsPerSample();
	unsigned int uiChannelCount = m_pkStream->GetNumChannels();
	
	if( !m_bPaused )
		m_pkStream->Reset();
		
	m_bPaused = false;
	
	m_pkDevice->Lock();

	stream = arts_play_stream( uiRate, uiBitsPerSample, uiChannelCount, "neodevarts" );
	arts_stream_set( stream, ARTS_P_BLOCKING, 1 );
	arts_stream_set( stream, ARTS_P_PACKET_SETTINGS, 0x0003000A );

	m_pkDevice->Unlock();
		
	unsigned int uiBufferSize = arts_stream_get( stream, ARTS_P_PACKET_SIZE );
	
	unsigned char *pucBuf       = new unsigned char[ uiBufferSize ];
	unsigned int   uiBytes      = m_pkStream->GetSize() - m_pkStream->GetPosition();

	neolog << LogLevel( DEBUG ) << "starting sound playing [" << GetName() << "]" << endl;
	
	while( uiBytes && !pkThread->m_bTerminate )
	{
		unsigned int uiDecoded = m_pkStream->Decode( pucBuf,
				uiBytes<uiBufferSize? uiBytes:uiBufferSize );
		unsigned int uiBuffer = 0;

		do
		{
			int uiFull;

			m_pkDevice->Lock();

			uiFull = arts_stream_get( stream, ARTS_P_BUFFER_SPACE );
			if( uiFull >= uiDecoded )
				uiBuffer = arts_write( stream, pucBuf, uiDecoded );
			
			m_pkDevice->Unlock();

			if( ! uiBuffer )
				usleep( 10000 );
		}
		while( ! uiBuffer );

		if( uiBytes < uiDecoded ) 
			uiBytes = 0;
		else
			uiBytes -= uiDecoded;
		
		if( !uiBytes && m_bLoop )
		{
			m_pkStream->Reset();
			uiBytes = m_pkStream->GetSize();
		}
	}
	
	neolog << LogLevel( DEBUG ) << "finished sound playing [" << GetName() << "]" << endl;

	m_pkDevice->Lock();
	arts_close_stream( stream );
	m_pkDevice->Unlock();

	delete [] pucBuf;
	
	pkThread->m_bTerminated = true;
	
	return 0;
}


void Sound::Stop()
{
	if( m_pkThread )
		delete m_pkThread, m_pkThread = 0;
		
	m_pkStream->Reset();
	
	m_bPaused = false;
}


void Sound::Pause()
{
	if( m_pkThread )
		delete m_pkThread, m_pkThread = 0;
		
	m_bPaused = true;
}


};


