/***************************************************************************
                 buffermanager.cpp  -  Buffer allocation manager
                             -------------------
    begin                : Fri Mar 27 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoDevD3D9, buffermanager.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "device.h"
#include "buffermanager.h"
#include "bufferregion.h"

#include <neoengine/logstream.h>

#include <assert.h>

using namespace std;
using namespace NeoEngine;


#ifndef NED3DBUFFERMANAGER_DEBUGALLOC
//Uncomment to get debug messages during allocation/deallocation
//#  define NED3DBUFFERMANAGER_DEBUGALLOC
#endif

#ifndef NED3DBUFFERMANAGER_FORCECHECK
//Uncomment to force check of buffer integrity after each alloc/dealloc
//#  define NED3DBUFFERMANAGER_FORCECHECK
#endif


namespace NeoD3D9
{


bool Device::SetStorageSize( unsigned int uiVertexStorageSize, unsigned int uiPolygonStorageSize )
{
	if( m_kWindow.m_hWnd )
		return false; //Already open

	m_pkVertexBufferManager->SetSize(  uiVertexStorageSize );
	m_pkPolygonBufferManager->SetSize( uiPolygonStorageSize );

	return true;
}





BufferManager::BufferManager( unsigned int uiSize )
{
	m_uiSize       = uiSize;
	m_pkRootRegion = 0;
	m_pkTailRegion = 0;
}


BufferManager::~BufferManager()
{
	BufferRegion *pkRegion = m_pkRootRegion;
	
	BufferRegion *pkNextRegion = 0;
	BufferRegion *pkLastRegion = 0;

	while( pkRegion )
	{
		pkLastRegion = pkRegion;
		pkNextRegion = pkRegion->m_pkPostRegion;

		delete pkRegion;

		pkRegion = pkNextRegion;
	}

	assert( pkLastRegion == m_pkTailRegion );
}


void BufferManager::SetSize( unsigned int uiSize )
{
	m_uiSize = uiSize;
}


BufferRegion *BufferManager::Alloc( unsigned int uiSize, unsigned int uiRequiredAlign )
{
	if( !uiSize )
		return 0;

	//Request extra space to be sure we can align properly
	unsigned int uiRequestedSize = uiSize;

	uiSize = uiSize + uiRequiredAlign;

#ifdef NED3DBUFFERMANAGER_DEBUGALLOC
	neolog << LogLevel( DEBUG ) << "[BufferManager::Alloc] allocating [" << uiSize << "] bytes" << endl;
#endif

	BufferRegion *pkRegion = m_kFreeRegions.BestFit( uiSize );

	if( pkRegion )
	{
		//We got best fit free region

		assert( pkRegion->m_eState == BufferRegion::FREE );

		// ### Can we do this during best fit search, or implement a DeleteAndInsertAgain method?
		m_kFreeRegions.Delete( pkRegion );

		if( pkRegion->m_uiRealSize > uiSize )
		{
			//Splice
			BufferRegion *pkNewRegion = new BufferRegion;

			pkNewRegion->m_pkPreRegion  = pkRegion->m_pkPreRegion;
			pkNewRegion->m_pkPostRegion = pkRegion;
			pkRegion->m_pkPreRegion     = pkNewRegion;

			if( pkNewRegion->m_pkPreRegion )
				pkNewRegion->m_pkPreRegion->m_pkPostRegion = pkNewRegion;

			pkNewRegion->m_uiRealOffset = pkRegion->m_uiRealOffset;
			pkNewRegion->m_uiRealSize   = uiSize;

			pkRegion->m_uiRealOffset    += uiSize;
			pkRegion->m_uiRealSize      -= uiSize;
			pkRegion->m_uiRegionOffset   = pkRegion->m_uiRealOffset;
			pkRegion->m_uiRegionSize     = pkRegion->m_uiRealSize;

			//Reinsert empty region
			m_kFreeRegions.Insert( pkRegion );

			//Set new root region if needed
			//If pkRegion is tail region, it will still be tail
			if( pkRegion == m_pkRootRegion )
				m_pkRootRegion = pkNewRegion;

			pkRegion = pkNewRegion;
		}
		else
			pkRegion->m_eState = BufferRegion::ALLOCATED;

		pkRegion->Align( uiRequiredAlign, uiRequestedSize );
	}
	else
	{
		//Allocate new region

		if( ( ( m_pkTailRegion ? m_pkTailRegion->m_uiRealOffset : 0 ) + uiSize ) > m_uiSize )
		{
			neolog << LogLevel( ERROR ) << "*** Unable to allocate buffer region: Out of memory" << endl;
			return 0;
		}

		pkRegion = new BufferRegion;

		pkRegion->m_pkPreRegion  = m_pkTailRegion;
		pkRegion->m_uiRealOffset = m_pkTailRegion ? ( m_pkTailRegion->m_uiRealOffset + m_pkTailRegion->m_uiRealSize ) : 0;
		pkRegion->m_uiRealSize   = uiSize;

		pkRegion->Align( uiRequiredAlign, uiRequestedSize );

		if( !m_pkRootRegion )
		{
			//If there is no root, there can be no tail :)
			assert( !m_pkTailRegion );
		
			m_pkRootRegion = pkRegion;
		}
		else
		{
			//If there exists a root, there must be a tail :)
			assert( m_pkTailRegion );

			m_pkTailRegion->m_pkPostRegion = pkRegion;
		}

		m_pkTailRegion = pkRegion;
	}

	return pkRegion;
}


void BufferManager::Free( BufferRegion *pkRegion )
{
	if( !pkRegion )
		return;

	if( pkRegion->m_eState != BufferRegion::ALLOCATED )
	{
		neolog << LogLevel( ERROR ) << "*** attempt to free buffer region that has already been freed" << endl;
		return;
	}

	//Do some basic sanity checks
	assert( ( pkRegion->m_pkPreRegion  || ( pkRegion == m_pkRootRegion ) ) && ( !pkRegion->m_pkPreRegion  || ( pkRegion != m_pkRootRegion ) ) );
	assert( ( pkRegion->m_pkPostRegion || ( pkRegion == m_pkTailRegion ) ) && ( !pkRegion->m_pkPostRegion || ( pkRegion != m_pkTailRegion ) ) );

	if( pkRegion == m_pkTailRegion )
	{
		if( pkRegion->m_pkPreRegion )
		{
			if( pkRegion->m_pkPreRegion->m_eState == BufferRegion::FREE )
			{
				m_kFreeRegions.Delete( pkRegion->m_pkPreRegion );

				m_pkTailRegion = pkRegion->m_pkPreRegion->m_pkPreRegion;

				if( m_pkTailRegion )
				{
					assert( m_pkTailRegion->m_eState == BufferRegion::ALLOCATED );
					m_pkTailRegion->m_pkPostRegion = 0;
				}
				else
					m_pkRootRegion = 0;

				delete pkRegion->m_pkPreRegion;
			}
			else
			{
				m_pkTailRegion = pkRegion->m_pkPreRegion;
				m_pkTailRegion->m_pkPostRegion = 0;
			}
		}
		else
		{
			assert( pkRegion == m_pkRootRegion );
			m_pkRootRegion = m_pkTailRegion = 0;
		}

		delete pkRegion;
	}
	else
	{
		if( pkRegion->m_pkPostRegion->m_eState == BufferRegion::FREE )
		{
			m_kFreeRegions.Delete( pkRegion->m_pkPostRegion );

			BufferRegion *pkOldPostRegion = pkRegion->m_pkPostRegion;

			pkRegion->m_uiRealSize   += pkRegion->m_pkPostRegion->m_uiRealSize;
			pkRegion->m_pkPostRegion  = pkRegion->m_pkPostRegion->m_pkPostRegion;

			assert( pkRegion->m_pkPostRegion );
			assert( pkRegion->m_pkPostRegion->m_eState == BufferRegion::ALLOCATED );

			pkRegion->m_pkPostRegion->m_pkPreRegion = pkRegion;

			delete pkOldPostRegion;
		}

		if( pkRegion->m_pkPreRegion && ( pkRegion->m_pkPreRegion->m_eState == BufferRegion::FREE ) )
		{
			m_kFreeRegions.Delete( pkRegion->m_pkPreRegion );

			pkRegion->m_pkPreRegion->m_uiRealSize   += pkRegion->m_uiRealSize;
			pkRegion->m_pkPreRegion->m_pkPostRegion  = pkRegion->m_pkPostRegion;

			if( pkRegion->m_pkPostRegion )
			{
				assert( pkRegion->m_pkPostRegion->m_eState == BufferRegion::ALLOCATED );

				pkRegion->m_pkPostRegion->m_pkPreRegion = pkRegion->m_pkPreRegion;
			}

			m_kFreeRegions.Insert( pkRegion->m_pkPreRegion );

			delete pkRegion;
		}
		else
		{
			pkRegion->m_eState = BufferRegion::FREE;

			m_kFreeRegions.Insert( pkRegion );
		}
	}
}


void BufferManager::CheckMemoryIntegrity()
{
	m_kFreeRegions.CheckTreeIntegrity();

	BufferRegion *pkRegion = m_pkRootRegion;

	unsigned int uiNumFreeRegions = 0;

	while( pkRegion )
	{
		assert( ( pkRegion->m_eState == BufferRegion::ALLOCATED ) || ( pkRegion->m_eState == BufferRegion::FREE ) );

		assert( !pkRegion->m_pkPostRegion || ( pkRegion->m_pkPostRegion->m_pkPreRegion == pkRegion ) );

		if( pkRegion->m_eState == BufferRegion::FREE )
			++uiNumFreeRegions;

		pkRegion = pkRegion->m_pkPostRegion;
	}

	vector< BufferRegion* > vpkFreeRegions;

	m_kFreeRegions.GetAllData( &vpkFreeRegions );

	unsigned int uiNumFreeRegionsTree = vpkFreeRegions.size();

	vector< BufferRegion* >::iterator ppkRegion    = vpkFreeRegions.begin();
	vector< BufferRegion* >::iterator ppkRegionEnd = vpkFreeRegions.end();

	for( ; ppkRegion != ppkRegionEnd; ++ppkRegion )
	{
		pkRegion = *ppkRegion;

		assert( pkRegion->m_eState == BufferRegion::FREE );

		assert( pkRegion->m_pkPostRegion );

		assert( pkRegion->m_pkPostRegion->m_pkPreRegion == pkRegion );
		assert( pkRegion->m_pkPostRegion->m_eState == BufferRegion::ALLOCATED );

		if( pkRegion->m_pkPreRegion )
		{
			assert( pkRegion->m_pkPreRegion->m_pkPostRegion == pkRegion );
			assert( pkRegion->m_pkPreRegion->m_eState == BufferRegion::ALLOCATED );
		}

		BufferRegion *pkSearchRegion = m_pkRootRegion;

		while( pkSearchRegion )
		{
			if( pkSearchRegion == pkRegion )
				break;

			pkSearchRegion = pkSearchRegion->m_pkPostRegion;
		}

		assert( pkSearchRegion );
	}

	assert( uiNumFreeRegions == uiNumFreeRegionsTree );
}


void BufferManager::DumpAllocationData()
{

}


}; // namespace NeoD3D9



