/***************************************************************************
                     intialize.cpp  -  Initialize Direct3D
                             -------------------
    begin                : Sun Mar 9 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoDevD3D9, initialize.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "device.h"
#include "textureunit.h"
#include "zbufferstate.h"
#include "buffermanager.h"
#include "rendertarget.h"
#include "framebuffertarget.h"
#include "pixelbuffertarget.h"
#include "shader.h"

#include <neoengine/nemath.h>
#include <neoengine/logstream.h>

#include <d3d9types.h>

using namespace std;
using namespace NeoEngine;


namespace NeoD3D9
{


//Helper method

bool CheckDepthFormat( LPDIRECT3D9 pD3D, unsigned int uiAdapter, D3DDEVTYPE eDevType, D3DFORMAT eDepthFormat, D3DFORMAT eAdapterFormat, D3DFORMAT eBackBufferFormat )
{
	//Verify that the depth format exists.
	HRESULT hr = pD3D->CheckDeviceFormat( uiAdapter, eDevType, eAdapterFormat, D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE, eDepthFormat );

	if( FAILED( hr ) )
		return false;

	//Verify that the depth format is compatible.
	hr = pD3D->CheckDepthStencilMatch( uiAdapter, eDevType, eAdapterFormat, eBackBufferFormat, eDepthFormat );

	if( FAILED( hr ) )
		return false;

	return true;
}




bool Device::Initialize( const RenderCaps &rkCaps )
{
	if( !m_pkD3D && !( m_pkD3D = Direct3DCreate9( D3D_SDK_VERSION ) ) )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to initialize rendering backend: Unable to create Direct3D 9 object" << endl;
		return false;
	}

	HRESULT               hResult;
	D3DPRESENT_PARAMETERS kPresentParams;
	D3DCAPS9              kCaps;
	DWORD                 lCreateFlags      = 0;
	bool                  bFullscreen       = m_kWindow.m_kCaps.IsSet( RenderCaps::FULLSCREEN ) ? TRUE : FALSE;
	unsigned int          uiAdapter         = !m_kWindow.m_uiAdapter ? D3DADAPTER_DEFAULT : m_kWindow.m_uiAdapter;
	D3DDEVTYPE            eDevType          = D3DDEVTYPE_HAL;
	D3DFORMAT             eDepthFormat      = D3DFMT_UNKNOWN;
	D3DFORMAT             eDisplayFormat    = D3DFMT_UNKNOWN;

	m_eFormat = D3DFMT_UNKNOWN;

	kPresentParams.BackBufferWidth  = bFullscreen ? m_kWindow.m_kResolution.m_uiWidth  : 0;
	kPresentParams.BackBufferHeight = bFullscreen ? m_kWindow.m_kResolution.m_uiHeight : 0;
	
	bool bGotFormat = false;

	if( bFullscreen )
	{
		if( m_kWindow.m_kResolution.m_uiBPP > 16 )
		{
			if( D3D_OK == ( hResult = m_pkD3D->CheckDeviceType( uiAdapter, eDevType, D3DFMT_X8R8G8B8, D3DFMT_A8R8G8B8, FALSE ) ) )
			{
				neolog << LogLevel( DEBUG ) << "[A8R8G8B8] back buffer format" << endl;
				m_eFormat         = D3DFMT_A8R8G8B8;
				eDisplayFormat    = D3DFMT_X8R8G8B8;
			}
			else if( D3D_OK == ( hResult = m_pkD3D->CheckDeviceType( uiAdapter, eDevType, D3DFMT_X8R8G8B8, D3DFMT_X8R8G8B8, FALSE ) ) )
			{
				neolog << LogLevel( DEBUG ) << "[X8R8G8B8] back buffer format" << endl;
				m_eFormat         = D3DFMT_X8R8G8B8;
				eDisplayFormat    = D3DFMT_X8R8G8B8;
			}
			else
			{
				neolog << LogLevel( WARNING ) << "*** Unable to use X8R8G8B8 format for " << m_kWindow.m_kResolution.m_uiBPP << " bpp, fallback to 16 bpp" << endl;
				m_kWindow.m_kResolution.m_uiBPP = 16;
			}
		}

		if( m_kWindow.m_kResolution.m_uiBPP == 16 )
		{
			if( D3D_OK == ( hResult = m_pkD3D->CheckDeviceType( uiAdapter, eDevType, D3DFMT_R5G6B5, D3DFMT_R5G6B5, FALSE ) ) )
			{
				neolog << LogLevel( DEBUG ) << "[R5G6B5] back buffer format" << endl;
				m_eFormat         = D3DFMT_R5G6B5;
				eDisplayFormat    = D3DFMT_R5G6B5;
			}
			else if( D3D_OK == ( hResult = m_pkD3D->CheckDeviceType( uiAdapter, eDevType, D3DFMT_X1R5G5B5, D3DFMT_A1R5G5B5, FALSE ) ) )
			{
				neolog << LogLevel( DEBUG ) << "[A1R5G5B5] back buffer format" << endl;
				m_eFormat         = D3DFMT_A1R5G5B5;
				eDisplayFormat    = D3DFMT_X1R5G5B5;
			}
			else if( D3D_OK == ( hResult = m_pkD3D->CheckDeviceType( uiAdapter, eDevType, D3DFMT_X1R5G5B5, D3DFMT_X1R5G5B5, FALSE ) ) )
			{
				neolog << LogLevel( DEBUG ) << "[X1R5G5B5] back buffer format" << endl;
				m_eFormat         = D3DFMT_X1R5G5B5;
				eDisplayFormat    = D3DFMT_X1R5G5B5;
			}
			else
			{
				neolog << LogLevel( ERROR ) << "*** Unable to initialize rendering backend: unable to find compatible back buffer format" << endl;
				return false;
			}
		}
	}
	else
	{
		D3DDISPLAYMODE kMode;

		m_pkD3D->GetAdapterDisplayMode( uiAdapter, &kMode );

		m_eFormat      = kMode.Format;
		eDisplayFormat = kMode.Format;
	}

	//Do the depth-stencil format dance
	if( m_kWindow.m_kCaps.IsSet( RenderCaps::STENCILBUFFER ) )
	{
		if( CheckDepthFormat( m_pkD3D, uiAdapter, eDevType, D3DFMT_D24S8, eDisplayFormat, m_eFormat ) )
		{
			neolog << LogLevel( DEBUG ) << "[D24S8] depth/stencil buffer format" << endl;
			eDepthFormat = D3DFMT_D24S8;
		}
		else if( CheckDepthFormat( m_pkD3D, uiAdapter, eDevType, D3DFMT_D24X4S4, eDisplayFormat, m_eFormat ) )
		{
			neolog << LogLevel( DEBUG ) << "[D24X4S4] depth/stencil buffer format" << endl;
			eDepthFormat = D3DFMT_D24X4S4;
		}
		else
		{
			neolog << LogLevel( WARNING ) << "*** Stencil buffer requested, no suitable format supported" << endl;
			m_kWindow.m_kCaps.Reset( RenderCaps::STENCILBUFFER );
		}
	}

	if( eDepthFormat == D3DFMT_UNKNOWN )
	{
		D3DFORMAT aeDepthList[3][3] = { { D3DFMT_D16, D3DFMT_D24X8, D3DFMT_D32 }, { D3DFMT_D24X8, D3DFMT_D32, D3DFMT_D16 }, { D3DFMT_D32, D3DFMT_D24X8, D3DFMT_D16 } };
		string strDepthList[3][3]   = { { "D16", "D24X8", "D32" }, { "D24X8", "D32", "D16" }, { "D32", "D24X8", "D16" } };

		//Default to 24bits depth buffer
		int iOffset = 1;

		if( m_kWindow.m_kResolution.m_uiDepthBits == 16 )
			iOffset = 0;
		else if( m_kWindow.m_kResolution.m_uiDepthBits == 32 )
			iOffset = 2;

		if( CheckDepthFormat( m_pkD3D, uiAdapter, eDevType, aeDepthList[iOffset][0], eDisplayFormat, m_eFormat ) )
		{
			neolog << LogLevel( DEBUG ) << "[" << strDepthList[iOffset][0] << "] depth buffer format" << endl;
			eDepthFormat = aeDepthList[iOffset][0];
		}
		else if( CheckDepthFormat( m_pkD3D, uiAdapter, eDevType, aeDepthList[iOffset][1], eDisplayFormat, m_eFormat ) )
		{
			neolog << LogLevel( DEBUG ) << "[" << strDepthList[iOffset][1] << "] depth buffer format" << endl;
			eDepthFormat = aeDepthList[iOffset][1];
		}
		else if( CheckDepthFormat( m_pkD3D, uiAdapter, eDevType, aeDepthList[iOffset][2], eDisplayFormat, m_eFormat ) )
		{
			neolog << LogLevel( DEBUG ) << "[" << strDepthList[iOffset][2] << "] depth buffer format" << endl;
			eDepthFormat = aeDepthList[iOffset][2];
		}
		else
		{
			neolog << LogLevel( ERROR ) << "*** Unable to initialize rendering backend: unable to find compatible depth buffer format" << endl;
			return false;
		}
	}

	if( D3D_OK != ( hResult = m_pkD3D->GetDeviceCaps( uiAdapter, eDevType, &kCaps ) ) )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to initialize rendering backend: Failed to get device caps" << endl;
		return false;
	}

	if( eDevType == D3DDEVTYPE_HAL )
	{
		if( kCaps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT )
		{
			lCreateFlags |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
			neolog << LogLevel( INFO ) << "[*] Hardware vertex processing" << endl;

			if( kCaps.DevCaps & D3DDEVCAPS_PUREDEVICE )
			{
				lCreateFlags |= D3DCREATE_PUREDEVICE;
				neolog << LogLevel( INFO ) << "[*] Pure device" << endl;
			}
		}
		else 
		{
			lCreateFlags |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;
			neolog << LogLevel( INFO ) << "[*] Software vertex processing" << endl;
		}
	}
	else
	{
		lCreateFlags |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;
		neolog << LogLevel( INFO ) << "[*] Software vertex processing [REF]" << endl;
	}

	kPresentParams.BackBufferFormat           = m_eFormat;
	kPresentParams.BackBufferCount            = 1;
	kPresentParams.MultiSampleType            = D3DMULTISAMPLE_NONE;
	kPresentParams.MultiSampleQuality         = 0;
	kPresentParams.SwapEffect                 = D3DSWAPEFFECT_DISCARD;
	kPresentParams.hDeviceWindow              = (HWND)m_kWindow.m_hWnd;
	kPresentParams.Windowed                   = !bFullscreen;
	kPresentParams.EnableAutoDepthStencil     = TRUE;
	kPresentParams.AutoDepthStencilFormat     = eDepthFormat;
	kPresentParams.Flags                      = D3DPRESENTFLAG_DISCARD_DEPTHSTENCIL;
	kPresentParams.FullScreen_RefreshRateInHz = ( bFullscreen && ( m_kWindow.m_kResolution.m_uiRefreshRate != 0 ) ) ? m_kWindow.m_kResolution.m_uiRefreshRate : D3DPRESENT_RATE_DEFAULT;
	kPresentParams.PresentationInterval       = D3DPRESENT_INTERVAL_IMMEDIATE;

	if( D3D_OK != ( hResult = m_pkD3D->CreateDevice( uiAdapter, eDevType, (HWND)m_kWindow.m_hWnd, lCreateFlags, &kPresentParams, &m_pkDevice ) ) )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to initialize rendering backend: Failed to create device" << endl;
		return false;
	}


	D3DCAPS9 kDevCaps;

	m_pkDevice->GetDeviceCaps( &kDevCaps );

	if( kDevCaps.TextureCaps & D3DPTEXTURECAPS_CUBEMAP )
		m_kWindow.m_kCaps.Set( RenderCaps::CUBEMAP );

	//Setup texture units
	m_uiNumTMUs = ( kCaps.MaxTextureBlendStages < kCaps.MaxSimultaneousTextures ) ? kCaps.MaxTextureBlendStages : kCaps.MaxSimultaneousTextures;
	m_ppkTMU    = new TextureUnit*[ m_uiNumTMUs ];

	for( unsigned int uiTMU = 0; uiTMU < m_uiNumTMUs; ++uiTMU )
		m_ppkTMU[ uiTMU ] = new TextureUnit( this, uiTMU );

	if( m_uiNumTMUs > 1 )
		m_kWindow.m_kCaps.Set( RenderCaps::MULTITEXTURE );

	m_kWindow.m_kCaps.Set( RenderCaps::MULTITEXTURE_GROUP | ( ( m_uiNumTMUs << RenderCaps::MULTITEXTURE_UNITS_SHIFT ) & RenderCaps::MULTITEXTURE_UNITS_MASK ) );

	// Always supported (?)
	m_kWindow.m_kCaps.Set( RenderCaps::BLENDMODE_FACTOR );

	neolog << LogLevel( INFO ) << "[" << m_uiNumTMUs << "] texture units setup" << endl
	                           << "[" << kCaps.MaxTextureWidth << "x" << kCaps.MaxTextureHeight << "] max texture size, [" << kCaps.MaxTextureAspectRatio << "] max aspect ratio" << endl;

	m_pkDevice->SetRenderState( D3DRS_CULLMODE,              D3DCULL_CW                                           );
	m_pkDevice->SetRenderState( D3DRS_ALPHABLENDENABLE,      TRUE                                                 );
	m_pkDevice->SetRenderState( D3DRS_COLORVERTEX,           FALSE                                                );
	m_pkDevice->SetRenderState( D3DRS_AMBIENT,               D3DCOLOR_COLORVALUE( 0.0f, 0.0f, 0.0f, 0.0f )        );
	m_pkDevice->SetRenderState( D3DRS_AMBIENTMATERIALSOURCE, D3DMCS_MATERIAL                                      );
	m_pkDevice->SetRenderState( D3DRS_DIFFUSEMATERIALSOURCE, D3DMCS_MATERIAL                                      );
	m_pkDevice->SetRenderState( D3DRS_STENCILENABLE,         m_kWindow.m_kCaps.IsSet( RenderCaps::STENCILBUFFER ) );
	m_pkDevice->SetRenderState( D3DRS_SPECULARENABLE,        FALSE                                                );
	m_pkDevice->SetRenderState( D3DRS_LOCALVIEWER,           TRUE                                                 );
	m_pkDevice->SetRenderState( D3DRS_COLORWRITEENABLE,      D3DCOLORWRITEENABLE_ALPHA | D3DCOLORWRITEENABLE_BLUE | D3DCOLORWRITEENABLE_GREEN | D3DCOLORWRITEENABLE_RED );

	m_pkZBufferState = new ZBufferState( this );


	unsigned int uiMajorVer = ( kDevCaps.VertexShaderVersion >> 8 ) & 0xFF;
	unsigned int uiMinorVer = kDevCaps.VertexShaderVersion & 0xFF;

	if( uiMajorVer > 0 )
	{
		if( uiMajorVer >=2 )
		{
			if( ( uiMajorVer > 2 ) || ( uiMinorVer > 0 ) )
				D3D9Shader::s_eVSTarget = Shader::NEPT_DX9_VS_2_X;
			else
				D3D9Shader::s_eVSTarget = Shader::NEPT_DX9_VS_2_0;
		}
		else if( uiMinorVer >= 1 )
			D3D9Shader::s_eVSTarget = Shader::NEPT_DX9_VS_1_1;
	}

	if( D3D9Shader::s_eVSTarget != Shader::NOTSUPPORTED )
	{
		m_kWindow.m_kCaps.Set( RenderCaps::VERTEXSHADER );
		neolog << LogLevel( INFO ) << "[" << Shader::GetTargetAsString( D3D9Shader::s_eVSTarget ) << "] supported" << endl;
	}

	uiMajorVer = ( kDevCaps.PixelShaderVersion >> 8 ) & 0xFF;
	uiMinorVer = kDevCaps.PixelShaderVersion & 0xFF;

	if( uiMajorVer > 0 )
	{
		neolog << LogLevel( INFO );

		if( uiMajorVer >=2 )
		{
			if( ( uiMajorVer > 2 ) || ( uiMinorVer > 0 ) )
				D3D9Shader::s_eFSTarget = Shader::NEPT_DX9_PS_2_X;
			else
				D3D9Shader::s_eFSTarget = Shader::NEPT_DX9_PS_2_0;
		}
		else if( uiMinorVer >= 4 )
			D3D9Shader::s_eFSTarget = Shader::NEPT_DX9_PS_1_4;
		else if( uiMinorVer >= 3 )
			D3D9Shader::s_eFSTarget = Shader::NEPT_DX9_PS_1_3;
		else if( uiMinorVer >= 2 )
			D3D9Shader::s_eFSTarget = Shader::NEPT_DX9_PS_1_2;
		else if( uiMinorVer >= 1 )
			D3D9Shader::s_eFSTarget = Shader::NEPT_DX9_PS_1_1;
	}

	if( D3D9Shader::s_eFSTarget != Shader::NOTSUPPORTED )
	{
		m_kWindow.m_kCaps.Set( RenderCaps::FRAGMENTSHADER );
		neolog << LogLevel( INFO ) << "[" << Shader::GetTargetAsString( D3D9Shader::s_eFSTarget ) << "] supported" << endl;
	}


	SetViewport( 0, 0, m_kWindow.m_kResolution.m_uiWidth, m_kWindow.m_kResolution.m_uiHeight );

	m_kPerspectiveProjection.BuildProjection( m_kViewport );
	m_kOrthographicProjection.BuildProjection( m_kViewport );

	SetPerspectiveProjection();


	m_pkDevice->Clear( 0, 0, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER | D3DCLEAR_STENCIL, D3DCOLOR_ARGB( 0, 0, 0, 0 ), 1.0f, 0 );


	ShowWindow( (HWND)m_kWindow.m_hWnd, SW_SHOW );
	UpdateWindow( (HWND)m_kWindow.m_hWnd );


	//Allocate vertex and polygon buffer storage
	hResult = m_pkDevice->CreateVertexBuffer( m_pkVertexBufferManager->GetSize(), D3DUSAGE_DYNAMIC, 0, D3DPOOL_DEFAULT, &m_pkVertexBufferStore, 0 );

	if( hResult != D3D_OK )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to initialize rendering backend: Failed to create backing store vertex buffer of [" << m_pkVertexBufferManager->GetSize() << "] bytes" << endl;
		return false;
	}
	else
		neolog << LogLevel( INFO ) << "Allocated [" << m_pkVertexBufferManager->GetSize() << "] bytes in vertex buffer backing store" << endl;

	hResult = m_pkDevice->CreateIndexBuffer( m_pkPolygonBufferManager->GetSize(), D3DUSAGE_DYNAMIC, D3DFMT_INDEX16, D3DPOOL_DEFAULT, &m_pkPolygonBufferStore, 0 );

	if( hResult != D3D_OK )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to initialize rendering backend: Failed to create backing store vertex buffer of [" << m_pkPolygonBufferManager->GetSize() << "] bytes" << endl;
		return false;
	}
	else
		neolog << LogLevel( INFO ) << "Allocated [" << m_pkPolygonBufferManager->GetSize() << "] bytes in index buffer backing store" << endl;


	m_pkDevice->SetIndices( m_pkPolygonBufferStore );

	neolog << LogLevel( INFO ) << "Maximum vertex index    [" << kCaps.MaxVertexIndex    << "]" << endl
	                           << "Maximum primitive count [" << kCaps.MaxPrimitiveCount << "]" << endl;

	m_uiMaxLights = ( kCaps.MaxActiveLights == -1 ) ? 8 : kCaps.MaxActiveLights;
	neolog << LogLevel( INFO ) << "[" << m_uiMaxLights << "] max lights" << endl;

	SetRenderTarget( FRAMEBUFFER );

	if( !m_pkCurTarget || !m_pkCurTarget->IsActive() )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to initialize rendering backend: Failed to activate frame buffer render target" << endl;
		return false;
	}

	if( !m_pkDefaultMaterial )
		m_pkDefaultMaterial = new Material( "__defmat", Core::Get()->GetMaterialManager() );

	return true;
}


}; // namespace NeoD3D9

