/***************************************************************************
             pixelbuffer.cpp  -  Direct3D pixel buffer implementation
                             -------------------
    begin                : Fri May 23 2003
    copyright            : (C) 2003 by Keaton Mullis
    email                : kmullis@zianet.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoDevD3D9, pixelbuffer.cpp

 The Initial Developer of the Original Code is Keaton Mullis.
 Portions created by Keaton Mullis are Copyright (C) 2003
 Keaton Mullis. All Rights Reserved.

 Contributors: Mattias Jansson (mattias@realityrift.com)

 ***************************************************************************/

#define __NEED_VECTOR_INT

#include "pixelbuffer.h"
#include "pixelbuffertarget.h"
#include "device.h"
#include "texture.h"

#include <neoengine/logstream.h>

using namespace std;
using namespace NeoEngine;


namespace NeoD3D9
{


/*! Free buffer IDs */
vector< int > g_vuiFreePBufferID;

/*! Last buffer ID */
unsigned int  g_uiNextPBufferID = 0;


NeoEngine::PixelBuffer *Device::CreatePixelBuffer( unsigned int uiWidth, unsigned int uiHeight, unsigned int uiBPP, Texture::TEXTURETYPE eTextureType )
{
	if( m_kWindow.m_kCaps.IsSet( RenderCaps::PIXELBUFFER ) )
	{
		PixelBuffer *pkPBuffer = new PixelBuffer( this );

		if( pkPBuffer->Open( uiWidth, uiHeight, uiBPP, eTextureType ) )
		{
			neolog << LogLevel( INFO ) << "Created pixel buffer [" << uiWidth << "x" << uiHeight << "]" << endl;
			return pkPBuffer;
		}

		neolog << LogLevel( ERROR ) << "*** Unable to create pixel buffer: unable to open buffer [" << uiWidth << "x" << uiHeight << "x" << uiBPP << "]" << endl;

		delete pkPBuffer;
	}

	return 0;
}





PixelBuffer::PixelBuffer( Device *pkDevice ) :
	m_pkSurface( 0 ),
	m_pkBackBuffer( 0 ),
	m_pkDevice( pkDevice ),
	m_eCurCubemapFace( Texture::CUBEFACE_FRONT )
{
	Deactivate();

	if( g_vuiFreePBufferID.size() )
	{
		m_uiID = *( g_vuiFreePBufferID.end() - 1 );
		g_vuiFreePBufferID.pop_back();
	}
	else
		m_uiID = g_uiNextPBufferID++;

	m_pkDevice = pkDevice;

	m_pkDevice->m_vpkPixelBuffers.push_back( this );
}


PixelBuffer::~PixelBuffer()
{
	Close();
	
	g_vuiFreePBufferID.push_back( m_uiID );

	vector< PixelBuffer* >::iterator ppkBuffer    = m_pkDevice->m_vpkPixelBuffers.begin();
	vector< PixelBuffer* >::iterator ppkBufferEnd = m_pkDevice->m_vpkPixelBuffers.end();

	for( ; ppkBuffer != ppkBufferEnd; ++ppkBuffer )
	{
		if( *ppkBuffer == this )
		{
			m_pkDevice->m_vpkPixelBuffers.erase( ppkBuffer );
			break;
		}
	}
}


bool PixelBuffer::Open( unsigned int uiWidth, unsigned int uiHeight, unsigned int uiBPP, Texture::TEXTURETYPE eTextureType )
{
	m_uiWidth   = uiWidth;
	m_uiHeight  = uiHeight;
	m_uiBPP     = uiBPP;

	float fExp    = float( log( float( m_uiWidth ) ) ) / float( log( 2.0f ) );
	float fFrac   = fExp - float( int( fExp ) );

	if( ( fFrac > 0.0001f ) && ( fFrac < 0.9999f ) )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to create pixelbuffer: width [" << m_uiWidth << "] not a power of 2, exponent [" << fExp << "]" << endl;
		return false;
	}

	fExp  = float( log( float( m_uiHeight ) ) ) / float( log( 2.0f ) );
	fFrac = fExp - float( int( fExp ) );	

	if( ( fFrac > 0.0001f ) && ( fFrac < 0.9999f ) )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to create pixelbuffer: height [" << m_uiHeight << "] not a power of 2, exponent [" << fExp << "]" << endl;
		return false;
	}


	Texture *pkD3DTexture = new Texture( m_pkDevice, "__pixelbuffer", eTextureType, Texture::RGB, 0, 1, 0, 0 );

	m_pkTexture = pkD3DTexture;


	if( eTextureType == Texture::TEX2D )
	{
		if( FAILED( m_pkDevice->m_pkDevice->CreateTexture( m_uiWidth, m_uiHeight, 1, D3DUSAGE_RENDERTARGET, m_pkDevice->m_eFormat, D3DPOOL_DEFAULT, (IDirect3DTexture9 **)&pkD3DTexture->m_pkD3DTexture, NULL ) ) )
		{
			neolog << LogLevel( WARNING ) << "*** Unable to open pixel buffer: Unable to create pixel buffer texture" << endl;
			return false;
		}

		if( FAILED( ((IDirect3DTexture9 *)pkD3DTexture->m_pkD3DTexture)->GetSurfaceLevel( 0, &m_pkSurface ) ) )
		{
			neolog << LogLevel( WARNING ) << "*** Unable to open pixel buffer: Unable to get surface" << endl;
			return false;
		}
	}
	else if( eTextureType == Texture::CUBEMAP )
	{
		if( m_uiWidth != m_uiHeight )
		{
			neolog << LogLevel( WARNING ) << "*** Pixelbuffer cubemap must be square, height set to [" << m_uiWidth << "]" << endl;

			m_uiHeight = m_uiWidth;
		}

		if( FAILED( m_pkDevice->m_pkDevice->CreateCubeTexture( m_uiWidth, 1, D3DUSAGE_RENDERTARGET, m_pkDevice->m_eFormat, D3DPOOL_DEFAULT, (IDirect3DCubeTexture9 **)&pkD3DTexture->m_pkD3DTexture, NULL ) ) )
		{
			neolog << LogLevel( WARNING ) << "*** Unable to open pixel buffer: Unable to create pixel buffer cube texture" << endl;
			return false;
		}

		if( FAILED( ((IDirect3DCubeTexture9 *)pkD3DTexture->m_pkD3DTexture)->GetCubeMapSurface( Texture::s_aeD3DCubeFaces[ m_eCurCubemapFace ], 0, &m_pkSurface ) ) )
		{
			neolog << LogLevel( WARNING ) << "*** Unable to open pixel buffer: Unable to get surface" << endl;
			return false;
		}
	}
	else
	{
		neolog << LogLevel( ERROR ) << "*** Unable to create pixelbuffer: unsupported texture type [" << (int)eTextureType << "]" << endl;
		return false;
	}

	return true;
}

void PixelBuffer::Close()
{
	if( m_pkDevice->m_pkPixelBufferTarget && ( m_pkDevice->m_pkPixelBufferTarget->m_pkCurBuffer == this ) )
		m_pkDevice->m_pkPixelBufferTarget->m_pkCurBuffer = 0;

	Deactivate();

	if( m_pkSurface )
	{
		m_pkSurface->Release();
		m_pkSurface = 0;
	}
}

void PixelBuffer::Activate()
{
	if( IsActive() )
		return;

	if( FAILED( m_pkDevice->m_pkDevice->GetBackBuffer( 0, 0, D3DBACKBUFFER_TYPE_MONO, &m_pkBackBuffer ) ) )
	{
		neolog << LogLevel( WARNING ) << "*** Unable to activate pixel buffer: Unable to get back buffer" << endl;
		return;
	}

	if( FAILED( m_pkDevice->m_pkDevice->SetRenderTarget( 0, m_pkSurface ) ) )
	{
		neolog << LogLevel( WARNING ) << "*** Unable to activate pixel buffer: Failed to set render target" << endl;
		return;
	}

	m_kViewport = m_pkDevice->GetViewport();

	m_pkDevice->SetViewport( 0, 0, m_uiWidth, m_uiHeight );

	Activator::Activate();
}


void PixelBuffer::Deactivate()
{
	if( !IsActive() )
		return;
	
	if( m_pkBackBuffer )
	{
		if( FAILED( m_pkDevice->m_pkDevice->SetRenderTarget( 0, m_pkBackBuffer ) ) )
		{
			neolog << LogLevel( WARNING ) << "*** Unable to dectivate pixel buffer: Failed to reset render target" << endl;
			return;
		}

		m_pkBackBuffer->Release();
		m_pkBackBuffer = 0;

		m_pkDevice->SetViewport( m_kViewport.m_iX, m_kViewport.m_iY, m_kViewport.m_iWidth, m_kViewport.m_iHeight );
	}

	Activator::Deactivate();
}


void PixelBuffer::SetCubeMapFace( Texture::TEXTURECUBEMAPFACE eCubeMapFace )
{
	if( m_eCurCubemapFace == eCubeMapFace )
		return;

	m_eCurCubemapFace = eCubeMapFace;

	if( m_pkTexture->GetType() == Texture::CUBEMAP )
	{
		if( m_pkSurface )
		{
			m_pkSurface->Release();
			m_pkSurface = 0;
		}

		if( FAILED( ((IDirect3DCubeTexture9 *)((NeoD3D9::Texture*)(NeoEngine::Texture*)m_pkTexture)->m_pkD3DTexture)->GetCubeMapSurface( Texture::s_aeD3DCubeFaces[ m_eCurCubemapFace ], 0, &m_pkSurface ) ) )
			neolog << LogLevel( WARNING ) << "*** Unable to set pixelbuffer cubemap face: Unable to get surface" << endl;

		if( IsActive() )
		{
			if( FAILED( m_pkDevice->m_pkDevice->SetRenderTarget( 0, m_pkSurface ) ) )
			{
				neolog << LogLevel( WARNING ) << "*** Unable to set pixelbuffer cubemap face: Failed to set render target" << endl;
				return;
			}
		}
	}
}


};

