/***************************************************************************
                      polygonbuffer.cpp  -  Polygon buffer
                             -------------------
    begin                : Sat Mar 15 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoDevD3D9, polygonbuffer.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/


#include "device.h"
#include "polygonbuffer.h"
#include "buffermanager.h"
#include "bufferregion.h"

#include <neoengine/logstream.h>

using namespace std;
using namespace NeoEngine;


namespace NeoD3D9
{


PolygonBufferPtr Device::CreatePolygonBuffer( unsigned int uiType, unsigned int uiNumPolygons, const NeoEngine::Polygon *pkData, bool bStripify )
{
	if( uiType & Buffer::NORENDER )
		return new NeoEngine::PolygonBuffer( uiType, uiNumPolygons, pkData, bStripify );

	//We don't care about NOREADWRITE buffers, infact, it will make us crash since we always need to be able to lock buffer
	uiType &= ~Buffer::NOREADWRITE;

	return new PolygonBufferD3D( this, m_pkPolygonBufferManager, uiType, uiNumPolygons, pkData, bStripify );
}





PolygonBufferD3D::PolygonBufferD3D( Device *pkDevice, BufferManager *pkBufferManager, unsigned int uiType, unsigned int uiNumPolygons, const NeoEngine::Polygon *pkData, bool bStripify ) :
	PolygonBuffer( uiType, 0, 0 )
{
	m_pkD3DBuffer = 0;
	m_pkDevice    = pkDevice;
	m_pkManager   = pkBufferManager;
	m_pkRegion    = 0;

	AllocatePolygons( uiNumPolygons, pkData, bStripify );
}


PolygonBufferD3D::~PolygonBufferD3D()
{
	m_pkManager->Free( m_pkRegion );

	//We don't want base buffer class deallocating pointer
	m_pkBuffer = 0;
}


void PolygonBufferD3D::AllocatePolygons( unsigned int uiNumPolygons, const NeoEngine::Polygon *pkData, bool bStripify )
{
	if( m_uiNumAllocated == uiNumPolygons )
	{
		m_uiNumCurrent = m_uiNumAllocated;

		if( !pkData )
			return;

		Lock( Buffer::WRITE );

		LoadPolygonData( pkData, bStripify );

		Unlock();

		return;
	}

	m_pkManager->Free( m_pkRegion );

	delete [] m_pkNormals, m_pkNormals = 0;
	delete [] m_pkEdges, m_pkEdges = 0;

	m_uiNumAllocated = uiNumPolygons;
	m_uiNumCurrent   = m_uiNumAllocated;
	m_pkD3DBuffer    = m_pkDevice->GetPolygonBufferStore();
	m_pkRegion       = 0;
	m_pkBuffer       = 0;

	if( !m_uiNumAllocated )
		return;

	m_pkRegion = m_pkManager->Alloc( m_uiNumAllocated * sizeof( unsigned short ) * 3, sizeof( unsigned short ) * 3 );

	if( pkData )
	{
		Lock( Buffer::WRITE );

		LoadPolygonData( pkData, bStripify );

		Unlock();
	}
}


void PolygonBufferD3D::ReleaseLock()
{
#ifdef NED3DPOLYGONBUFFER_USEMINMAXINDEX
	if( m_uiLock & WRITE )
	{
		unsigned short *pusIndex = (unsigned short*)m_pkPolygons;

		m_uiMinIndex = 0xFFFF;
		m_uiMaxIndex = 0;

		for( unsigned short uiLoop = 0; uiLoop < ( m_uiNumCurrent * 3 ); ++uiLoop, ++pusIndex )
		{
			if( m_uiMinIndex > *pusIndex )
				m_uiMinIndex = *pusIndex;
			if( m_uiMaxIndex < *pusIndex )
				m_uiMaxIndex = *pusIndex;
		}

		++m_uiMaxIndex;

		assert( m_uiMinIndex < m_uiMaxIndex );
	}
#endif

#ifdef _DEBUG
	HRESULT hResult;
	if( D3D_OK != ( hResult = m_pkD3DBuffer->Unlock() ) )
		neolog << LogLevel( ERROR ) << "*** unable to unlock D3D backing store index buffer [" << (unsigned int)hResult << "]" << endl;
#else
	m_pkD3DBuffer->Unlock();
#endif

	m_pkBuffer = 0;

	PolygonBuffer::ReleaseLock();
}


bool PolygonBufferD3D::AcquireLock()
{
	DWORD dwFlags = 0;

	if( m_uiLock == READ )
	{
		dwFlags |= D3DLOCK_NOOVERWRITE;
		dwFlags |= D3DLOCK_READONLY;
	}

#ifdef _DEBUG
	HRESULT hResult;
	if( D3D_OK != ( hResult = m_pkD3DBuffer->Lock( m_pkRegion->m_uiRegionOffset, m_pkRegion->m_uiRegionSize, (void**)&m_pkBuffer, dwFlags ) ) )
		neolog << LogLevel( ERROR ) << "*** unable to lock D3D backing store index buffer [" << (unsigned int)hResult << "]" << endl;
#else
	m_pkD3DBuffer->Lock( m_pkRegion->m_uiRegionOffset, m_pkRegion->m_uiRegionSize, (void**)&m_pkBuffer, dwFlags );
#endif

	return PolygonBuffer::AcquireLock();
}


}; // namespace NeoD3D9


