/***************************************************************************
                     shader.cpp  -  Pipeline shaders
                             -------------------
    begin       : Wed May 7 2003
    copyright   : (C) 2003 by Reality Rift Studios
    email       : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoDevD3D9, shader.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 Contributors: Keaton Mullis (kmullis@zianet.com)

 ***************************************************************************/

#include "shader.h"
#include "texture.h"

#include <neoengine/light.h>
#include <neoengine/scenenode.h>
#include <neoengine/logstream.h>

#include <d3dx9.h>

using namespace std;
using namespace NeoEngine;


namespace NeoD3D9
{


NeoEngine::ShaderPtr Device::CreateShader( Shader::SHADERTYPE eType )
{
	if( eType == Shader::VERTEXSHADER ) switch( D3D9Shader::s_eVSTarget )
	{
		case Shader::NEPT_DX9_VS_1_1:
		case Shader::NEPT_DX9_VS_2_0:
		case Shader::NEPT_DX9_VS_2_X:
			return new D3D9Shader( this, Shader::VERTEXSHADER, Core::Get()->GetFileManager() );

		case Shader::NOTSUPPORTED:
			break;

		default:
			neolog << LogLevel( ERROR ) << "*** Unable to create vertex shader: Unknown native vertex shader target [" << D3D9Shader::s_eVSTarget << "]" << endl;
			break;
	}
	else if( eType == Shader::FRAGMENTSHADER ) switch( D3D9Shader::s_eFSTarget )
	{
		case Shader::NEPT_DX9_PS_1_1:
		case Shader::NEPT_DX9_PS_1_2:
		case Shader::NEPT_DX9_PS_1_3:
		case Shader::NEPT_DX9_PS_1_4:
		case Shader::NEPT_DX9_PS_2_0:
		case Shader::NEPT_DX9_PS_2_X:
			return new D3D9Shader( this, Shader::FRAGMENTSHADER, Core::Get()->GetFileManager() );

		case Shader::NOTSUPPORTED:
			break;

		default:
			neolog << LogLevel( ERROR ) << "*** Unable to create fragment shader: Unknown native fragment shader target [" << D3D9Shader::s_eFSTarget << "]" << endl;
			break;
	}
	else
		neolog << LogLevel( ERROR ) << "*** Unable to create shader: Unknown type [" << eType << "]" << endl;

	return 0;
}


NeoEngine::ShaderPtr Device::GetShader( NeoEngine::Shader::SHADERTYPE eType, const std::string &rstrName )
{
	return (NeoEngine::Shader*)( ( eType == Shader::VERTEXSHADER ) ? D3D9Shader::s_kVSTable : D3D9Shader::s_kFSTable ).Find( rstrName );
}



ShaderTable D3D9Shader::s_kVSTable;
ShaderTable D3D9Shader::s_kFSTable;


D3D9Shader::D3D9Shader( Device *pkDevice, SHADERTYPE eType, FileManager *pkFileManager ) : 
	NeoEngine::Shader( eType, pkFileManager ),
	m_pkDevice( pkDevice ), 
	m_eType( eType ),
	m_bCompiled( false ),
	m_pkVertexShader( 0 ),
	m_pkPixelShader( 0 )
{
}


D3D9Shader::~D3D9Shader()
{
	if( m_pkDevice->m_pkVertexShader == this )
	{
		m_pkDevice->m_pkDevice->SetVertexShader( (IDirect3DVertexShader9*)0 );
		m_pkDevice->m_pkVertexShader = 0;
	}

	if( m_pkDevice->m_pkFragmentShader == this )
	{
		m_pkDevice->m_pkDevice->SetPixelShader( (IDirect3DPixelShader9*)0 );
		m_pkDevice->m_pkFragmentShader = 0;
	}

	if( m_pkVertexShader )
		m_pkVertexShader->Release();

	if( m_pkPixelShader )
		m_pkPixelShader->Release();

	( ( GetType() == VERTEXSHADER ) ? s_kVSTable : s_kFSTable ).Delete( m_strName, (D3DShader*)this );
}


void D3D9Shader::SetName( const HashString &rstrName )
{
	if( rstrName == GetName() )
		return;

	if( m_bCompiled )
	{
		ShaderTable &rkTable = ( ( GetType() == VERTEXSHADER ) ? s_kVSTable : s_kFSTable );
		
		rkTable.Delete( m_strName, (D3DShader*)this );

		m_strName = rstrName;

		rkTable.Insert( m_strName, (D3DShader*)this );
	}
	else
		m_strName = rstrName;
}


bool D3D9Shader::Compile( const string *pstrSource )
{
	if( m_bCompiled && !pstrSource )
		return true;

	if( m_pkVertexShader )
		m_pkVertexShader->Release(), m_pkVertexShader = 0;

	if( m_pkPixelShader )
		m_pkPixelShader->Release(), m_pkPixelShader = 0;

	ClearParams();

	m_bCompiled = false;

	if( pstrSource )
		m_strSource = *pstrSource;

	LPD3DXBUFFER pkBuffer = 0;
	LPD3DXBUFFER pkErr = 0;
	HRESULT hRes;

	if( FAILED( ( hRes = D3DXAssembleShader( m_strSource.c_str(), m_strSource.size() + 1, 0, 0, 0, &pkBuffer, &pkErr ) ) ) )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to compile shader: Assembler failed";

		if( hRes == D3DERR_INVALIDCALL )
			neolog << " [D3DERR_INVALIDCALL]";
		else if( hRes == D3DXERR_INVALIDDATA )
			neolog << " [D3DXERR_INVALIDDATA]";
		else
			neolog << " [UNKNOWN ERROR CODE]";

		neolog << endl << (char*)pkErr->GetBufferPointer() << endl;

		return false;
	}

	if( pkErr ) pkErr->Release();

	if( m_eType == VERTEXSHADER )
	{
		if( FAILED( m_pkDevice->m_pkDevice->CreateVertexShader( (DWORD*)pkBuffer->GetBufferPointer(), &m_pkVertexShader ) ) )
		{
			neolog << LogLevel( ERROR ) << "*** Unable to compile vertex shader: Unable to create vertex shader" << endl;
			if( pkBuffer ) pkBuffer->Release();
			return false;
		}
	}
	else
	{
		if( FAILED( m_pkDevice->m_pkDevice->CreatePixelShader( (DWORD*)pkBuffer->GetBufferPointer(), &m_pkPixelShader ) ) )
		{
			neolog << LogLevel( ERROR ) << "*** Unable to compile fragment shader: Unable to create pixel shader" << endl;
			if( pkBuffer ) pkBuffer->Release();
			return false;
		}
	}

	if( pkBuffer ) pkBuffer->Release();

	m_bCompiled = true;

	ParseParams();

	return true;
}


void D3D9Shader::BindParameterVS( ShaderParam *pkParam, RenderPrimitive *pkPrimitive, const Matrix &rkView )
{
	static float afIdentity[4] = { 0.0f, 0.0f, 0.0f, 1.0f };
	float afVal[4] = { 0.0f, 0.0f, 0.0f, 1.0f };
	unsigned int uiLight = pkParam->m_uiLight + m_pkDevice->m_uiCurLight;

	switch( pkParam->m_eType )
	{
		case ShaderParam::PPT_SYSTEM:
		{
			switch( pkParam->m_Value.m_eSystemID )
			{
				case ShaderParam::PSP_MODEL_MATRIX:
				{
					m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, &pkPrimitive->m_kModelMatrix.m_aafMatrix[0][0], 4 );
					break;
				}

				case ShaderParam::PSP_VIEW_MATRIX:
				{
					m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, &rkView.m_aafMatrix[0][0], 4 );
					break;
				}

				case ShaderParam::PSP_PROJECTION_MATRIX:
				{
					m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, &m_pkDevice->GetProjectionMatrix().m_aafMatrix[0][0], 4 );
					break;
				}

				case ShaderParam::PSP_MODELVIEW_MATRIX:
				{
					Matrix kMat = rkView * pkPrimitive->m_kModelMatrix;
					
					m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, &kMat.m_aafMatrix[0][0], 4 );
					break;
				}

				case ShaderParam::PSP_MODELVIEWPROJ_MATRIX:
				{
					Matrix kMat( ( m_pkDevice->GetProjectionMatrix() * rkView ) * pkPrimitive->m_kModelMatrix );

					m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, &kMat.m_aafMatrix[0][0], 4 );
					break;
				}

				case ShaderParam::PSP_INV_MODEL_MATRIX:
				{
					m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, &pkPrimitive->m_kInvModelMatrix.m_aafMatrix[0][0], 4 );
					break;
				}

				case ShaderParam::PSP_LIGHT_POSITION:
				{
					if( ( uiLight < m_pkDevice->m_uiNumLights ) && m_pkDevice->m_ppkLights[ uiLight ]->GetSceneNode() )
						*((Vector3d*)afVal) = m_pkDevice->m_ppkLights[ uiLight ]->GetSceneNode()->GetWorldTranslation();
					m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, afVal, 1 );
					break;
				}

				case ShaderParam::PSP_LIGHT_DIRECTION:
				{
					if( ( uiLight < m_pkDevice->m_uiNumLights ) && m_pkDevice->m_ppkLights[ uiLight ]->GetSceneNode() )
						*((Vector3d*)afVal) = -m_pkDevice->m_ppkLights[ uiLight ]->GetSceneNode()->GetWorldTransform().GetColumn(2);
					m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, afVal, 1 );
					break;
				}

				case ShaderParam::PSP_LIGHT_POSITION_OBJECTSPACE:
				{
					if( ( uiLight < m_pkDevice->m_uiNumLights ) && m_pkDevice->m_ppkLights[ uiLight ]->GetSceneNode() )
						*((Vector3d*)afVal) = pkPrimitive->m_kInvModelMatrix * m_pkDevice->m_ppkLights[ uiLight ]->GetSceneNode()->GetWorldTranslation();
					m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, afVal, 1 );
					break;
				}

				case ShaderParam::PSP_LIGHT_DIRECTION_OBJECTSPACE:
				{
					if( ( uiLight < m_pkDevice->m_uiNumLights ) && m_pkDevice->m_ppkLights[ uiLight ]->GetSceneNode() )
						*((Vector3d*)afVal) = pkPrimitive->m_kInvModelMatrix * (-m_pkDevice->m_ppkLights[ uiLight ]->GetSceneNode()->GetWorldTransform().GetColumn(2));
					m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, afVal, 1 );
					break;
				}

				case ShaderParam::PSP_LIGHT_AMBIENT_COLOR:
				{
					if( uiLight < m_pkDevice->m_uiNumLights )
						m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, m_pkDevice->m_ppkLights[ uiLight ]->m_kAmbient, 1 );
					else
						m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, afIdentity, 1 );
					break;
				}

				case ShaderParam::PSP_LIGHT_DIFFUSE_COLOR:
				{
					if( uiLight < m_pkDevice->m_uiNumLights )
						m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, m_pkDevice->m_ppkLights[ uiLight ]->m_kDiffuse, 1 );
					else
						m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, afIdentity, 1 );
					break;
				}

				case ShaderParam::PSP_LIGHT_SPECULAR_COLOR:
				{
					if( uiLight < m_pkDevice->m_uiNumLights )
						m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, m_pkDevice->m_ppkLights[ uiLight ]->m_kSpecular, 1 );
					else
						m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, afIdentity, 1 );
					break;
				}

				case ShaderParam::PSP_LIGHT_ATTENUATION:
				{
					if( uiLight < m_pkDevice->m_uiNumLights )
					{
						afVal[0] = m_pkDevice->m_ppkLights[ uiLight ]->m_fConstantAttenuation;
						afVal[1] = m_pkDevice->m_ppkLights[ uiLight ]->m_fLinearAttenuation;
						afVal[2] = m_pkDevice->m_ppkLights[ uiLight ]->m_fQuadraticAttenuation;
						afVal[3] = 1.0f;
						m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, afVal, 1 );
					}
					else
						m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, afIdentity, 1 );
					break;
				}

				case ShaderParam::PSP_LIGHT_REP_ATTENUATION:
				{
					if( uiLight < m_pkDevice->m_uiNumLights )
					{
						afVal[0] = 1.0f / m_pkDevice->m_ppkLights[ uiLight ]->m_fConstantAttenuation;
						afVal[1] = 1.0f / m_pkDevice->m_ppkLights[ uiLight ]->m_fLinearAttenuation;
						afVal[2] = 1.0f / m_pkDevice->m_ppkLights[ uiLight ]->m_fQuadraticAttenuation;
						afVal[3] = 1.0f;
						m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, afVal, 1 );
					}
					else
						m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, afIdentity, 1 );
					break;
				}

				case ShaderParam::PSP_TOTAL_LIGHT_AMBIENT_COLOR:
				{
					m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, (const float*)m_pkDevice->m_kTotalAmbientLight, 1 );
					break;
				}

				case ShaderParam::PSP_TOTAL_AMBIENTLIGHT_AMBIENT_COLOR:
				{
					m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, (const float*)m_pkDevice->m_kAmbientLight, 1 );
					break;
				}

				case ShaderParam::PSP_EYE_POSITION_OBJECTSPACE:
				{
					*((Vector3d*)afVal) = ( rkView * pkPrimitive->m_kInvModelMatrix ).GetColumn(3);
					m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, afVal, 1 );
					break;
				}

				default:
				{
					neolog << LogLevel( ERROR ) << "*** Unable to bind system parameter: Unsupported type" << endl;
					break;
				}
			}

			break;
		}

		case ShaderParam::PPT_FLOAT4:
		{
			m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, pkParam->m_Value.m_afFloat, 1 );
			break;
		}

		case ShaderParam::PPT_VECTOR:
		{
			float afVal[4] = { pkParam->m_Value.m_pkVector->x, pkParam->m_Value.m_pkVector->y, pkParam->m_Value.m_pkVector->z, 1.0f };
			m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, afVal, 1 );
			break;
		}

		case ShaderParam::PPT_COLOR:
		{
			m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, (const float*)pkParam->m_Value.m_pkColor, 1 );
			break;
		}

		case ShaderParam::PPT_QUAT:
		{
			m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, (const float*)pkParam->m_Value.m_pkQuaternion, 1 );
			break;
		}

		case ShaderParam::PPT_MATRIX:
		{
			m_pkDevice->m_pkDevice->SetVertexShaderConstantF( pkParam->m_uiIndex, &pkParam->m_Value.m_pkMatrix->m_aafMatrix[0][0], 4 );			
			break;
		}

		default:
		{
			neolog << LogLevel( ERROR ) << "*** Unable to bind parameter: Unsupported type" << endl;
			break;
		}
	}
}


void D3D9Shader::BindParameterPS( ShaderParam *pkParam, RenderPrimitive *pkPrimitive, const Matrix &rkView )
{
	static float afIdentity[4] = { 0.0f, 0.0f, 0.0f, 1.0f };
	float afVal[4] = { 0.0f, 0.0f, 0.0f, 1.0f };
	unsigned int uiLight = pkParam->m_uiLight + m_pkDevice->m_uiCurLight;

	switch( pkParam->m_eType )
	{
		case ShaderParam::PPT_SYSTEM:
		{
			switch( pkParam->m_Value.m_eSystemID )
			{
				case ShaderParam::PSP_MODEL_MATRIX:
				{
					m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, &pkPrimitive->m_kModelMatrix.m_aafMatrix[0][0], 4 );
					break;
				}

				case ShaderParam::PSP_VIEW_MATRIX:
				{
					m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, &rkView.m_aafMatrix[0][0], 4 );
					break;
				}

				case ShaderParam::PSP_PROJECTION_MATRIX:
				{
					m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, &m_pkDevice->GetProjectionMatrix().m_aafMatrix[0][0], 4 );
					break;
				}

				case ShaderParam::PSP_MODELVIEW_MATRIX:
				{
					Matrix kMat = rkView * pkPrimitive->m_kModelMatrix;
					
					m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, &kMat.m_aafMatrix[0][0], 4 );
					break;
				}

				case ShaderParam::PSP_MODELVIEWPROJ_MATRIX:
				{
					Matrix kMat( ( m_pkDevice->GetProjectionMatrix() * rkView ) * pkPrimitive->m_kModelMatrix );

					m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, &kMat.m_aafMatrix[0][0], 4 );
					break;
				}

				case ShaderParam::PSP_INV_MODEL_MATRIX:
				{
					m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, &pkPrimitive->m_kInvModelMatrix.m_aafMatrix[0][0], 4 );
					break;
				}

				case ShaderParam::PSP_LIGHT_POSITION:
				{
					if( ( uiLight < m_pkDevice->m_uiNumLights ) && m_pkDevice->m_ppkLights[ uiLight ]->GetSceneNode() )
						*((Vector3d*)afVal) = m_pkDevice->m_ppkLights[ uiLight ]->GetSceneNode()->GetWorldTranslation();
					m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, afVal, 1 );
					break;
				}

				case ShaderParam::PSP_LIGHT_DIRECTION:
				{
					if( ( uiLight < m_pkDevice->m_uiNumLights ) && m_pkDevice->m_ppkLights[ uiLight ]->GetSceneNode() )
						*((Vector3d*)afVal) = -m_pkDevice->m_ppkLights[ uiLight ]->GetSceneNode()->GetWorldTransform().GetColumn(2);
					m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, afVal, 1 );
					break;
				}

				case ShaderParam::PSP_LIGHT_POSITION_OBJECTSPACE:
				{
					if( ( uiLight < m_pkDevice->m_uiNumLights ) && m_pkDevice->m_ppkLights[ uiLight ]->GetSceneNode() )
						*((Vector3d*)afVal) = pkPrimitive->m_kInvModelMatrix * m_pkDevice->m_ppkLights[ uiLight ]->GetSceneNode()->GetWorldTranslation();
					m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, afVal, 1 );
					break;
				}

				case ShaderParam::PSP_LIGHT_DIRECTION_OBJECTSPACE:
				{
					if( ( uiLight < m_pkDevice->m_uiNumLights ) && m_pkDevice->m_ppkLights[ uiLight ]->GetSceneNode() )
						*((Vector3d*)afVal) = pkPrimitive->m_kInvModelMatrix * (-m_pkDevice->m_ppkLights[ uiLight ]->GetSceneNode()->GetWorldTransform().GetColumn(2));
					m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, afVal, 1 );
					break;
				}

				case ShaderParam::PSP_LIGHT_AMBIENT_COLOR:
				{
					if( uiLight < m_pkDevice->m_uiNumLights )
						m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, m_pkDevice->m_ppkLights[ uiLight ]->m_kAmbient, 1 );
					else
						m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, afIdentity, 1 );
					break;
				}

				case ShaderParam::PSP_LIGHT_DIFFUSE_COLOR:
				{
					if( uiLight < m_pkDevice->m_uiNumLights )
						m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, m_pkDevice->m_ppkLights[ uiLight ]->m_kDiffuse, 1 );
					else
						m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, afIdentity, 1 );
					break;
				}

				case ShaderParam::PSP_LIGHT_SPECULAR_COLOR:
				{
					if( uiLight < m_pkDevice->m_uiNumLights )
						m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, m_pkDevice->m_ppkLights[ uiLight ]->m_kSpecular, 1 );
					else
						m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, afIdentity, 1 );
					break;
				}

				case ShaderParam::PSP_LIGHT_ATTENUATION:
				{
					if( uiLight < m_pkDevice->m_uiNumLights )
					{
						afVal[0] = m_pkDevice->m_ppkLights[ uiLight ]->m_fConstantAttenuation;
						afVal[1] = m_pkDevice->m_ppkLights[ uiLight ]->m_fLinearAttenuation;
						afVal[2] = m_pkDevice->m_ppkLights[ uiLight ]->m_fQuadraticAttenuation;
						afVal[3] = 1.0f;
						m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, afVal, 1 );
					}
					else
						m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, afIdentity, 1 );
					break;
				}

				case ShaderParam::PSP_LIGHT_REP_ATTENUATION:
				{
					if( uiLight < m_pkDevice->m_uiNumLights )
					{
						afVal[0] = 1.0f / m_pkDevice->m_ppkLights[ uiLight ]->m_fConstantAttenuation;
						afVal[1] = 1.0f / m_pkDevice->m_ppkLights[ uiLight ]->m_fLinearAttenuation;
						afVal[2] = 1.0f / m_pkDevice->m_ppkLights[ uiLight ]->m_fQuadraticAttenuation;
						afVal[3] = 1.0f;
						m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, afVal, 1 );
					}
					else
						m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, afIdentity, 1 );
					break;
				}

				case ShaderParam::PSP_TOTAL_LIGHT_AMBIENT_COLOR:
				{
					m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, (const float*)m_pkDevice->m_kTotalAmbientLight, 1 );
					break;
				}

				case ShaderParam::PSP_TOTAL_AMBIENTLIGHT_AMBIENT_COLOR:
				{
					m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, (const float*)m_pkDevice->m_kAmbientLight, 1 );
					break;
				}

				case ShaderParam::PSP_EYE_POSITION_OBJECTSPACE:
				{
					*((Vector3d*)afVal) = ( rkView * pkPrimitive->m_kInvModelMatrix ).GetColumn(3);
					m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, afVal, 1 );
					break;
				}

				default:
				{
					neolog << LogLevel( ERROR ) << "*** Unable to bind system parameter: Unsupported type" << endl;
					break;
				}
			}

			break;
		}

		case ShaderParam::PPT_FLOAT4:
		{
			m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, pkParam->m_Value.m_afFloat, 1 );
			break;
		}

		case ShaderParam::PPT_VECTOR:
		{
			float afVal[4] = { pkParam->m_Value.m_pkVector->x, pkParam->m_Value.m_pkVector->y, pkParam->m_Value.m_pkVector->z, 1.0f };
			m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, afVal, 1 );
			break;
		}

		case ShaderParam::PPT_COLOR:
		{
			m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, (const float*)pkParam->m_Value.m_pkColor, 1 );
			break;
		}

		case ShaderParam::PPT_QUAT:
		{
			m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, (const float*)pkParam->m_Value.m_pkQuaternion, 1 );
			break;
		}

		case ShaderParam::PPT_MATRIX:
		{
			m_pkDevice->m_pkDevice->SetPixelShaderConstantF( pkParam->m_uiIndex, &pkParam->m_Value.m_pkMatrix->m_aafMatrix[0][0], 4 );			
			break;
		}

		default:
		{
			neolog << LogLevel( ERROR ) << "*** Unable to bind parameter: Unsupported type" << endl;
			break;
		}
	}
}


bool D3D9Shader::Bind( RenderPrimitive *pkPrimitive, const Matrix &rkView )
{
	if( !m_bCompiled )
		return false;

	if( m_eType == VERTEXSHADER )
	{
		if( FAILED( m_pkDevice->m_pkDevice->SetVertexShader( m_pkVertexShader ) ) )
			return false;

		//Bind all params
		vector< ShaderParam* >::iterator ppkParam = m_vpkParams.begin();
		vector< ShaderParam* >::iterator ppkEnd   = m_vpkParams.end();

		for( ; ppkParam != ppkEnd; ++ppkParam )
			BindParameterVS( *ppkParam, pkPrimitive, rkView );
	}
	else
	{
		if( FAILED( m_pkDevice->m_pkDevice->SetPixelShader( m_pkPixelShader ) ) )
			return false;

		//Bind all params
		vector< ShaderParam* >::iterator ppkParam = m_vpkParams.begin();
		vector< ShaderParam* >::iterator ppkEnd   = m_vpkParams.end();

		for( ; ppkParam != ppkEnd; ++ppkParam )
			BindParameterPS( *ppkParam, pkPrimitive, rkView );
	}

	//We don't use updated params for now, clear to avoid stockpiling pointers
	m_vpkUpdatedParams.clear();

	return true;
}


bool D3D9Shader::Unbind()
{
	if( m_eType == VERTEXSHADER )
	{
		if( FAILED( m_pkDevice->m_pkDevice->SetVertexShader( (IDirect3DVertexShader9*)0 ) ) )
			return false;
	}
	else
	{
		if( FAILED( m_pkDevice->m_pkDevice->SetPixelShader( (IDirect3DPixelShader9*)0 ) ) )
			return false;
	}

	return true;
}


bool D3D9Shader::BindLightParams( RenderPrimitive *pkPrimitive )
{
	//Bind all params
	vector< ShaderParam* >::iterator ppkParam = m_vpkParams.begin();
	vector< ShaderParam* >::iterator ppkEnd   = m_vpkParams.end();

	const Matrix &rkView = m_pkDevice->GetViewMatrix();

	for( ; ppkParam != ppkEnd; ++ppkParam )
	{
		if( ( (*ppkParam)->m_eType == ShaderParam::PPT_SYSTEM ) && ( (*ppkParam)->m_Value.m_eSystemID & ShaderParam::LIGHTPARAM ) )
		{
			if( m_eShaderType == Shader::VERTEXSHADER )
				BindParameterVS( *ppkParam, pkPrimitive, rkView );
			else
				BindParameterPS( *ppkParam, pkPrimitive, rkView );
		}
	}

	return true;
}


};
