/***************************************************************************
                texture.cpp  -  Direct3D texture implementation
                             -------------------
    begin                : Sat Mar 15 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoDevD3D9, texture.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "texture.h"
#include "device.h"

#include <neoengine/file.h>
#include <neoengine/logstream.h>

using namespace std;
using namespace NeoEngine;


namespace NeoD3D9
{


inline int lowpow( int n )
{
   int m;

   for( m = 1; m <= n; m <<= 1 );

   m >>= 1;

   return m;
}


NeoEngine::TexturePtr Device::GetTexture( const std::string &rstrName )
{
	const vector< NeoEngine::Texture* > *pvpkTextures = m_pkTexturePool->Find( rstrName );

	if( pvpkTextures && pvpkTextures->size() )
		return *(pvpkTextures->begin());

	return 0;
}


NeoEngine::TexturePtr Device::LoadTexture( const std::string &rstrFilename, Texture::TEXTURETYPE eTextureType, Texture::TEXTUREFORMAT eTextureFormat, unsigned int uiFlags, unsigned int uiFiltering, unsigned int uiMaxAnisotropy )
{
	//try to find texture
	{
		const vector< NeoEngine::Texture* > *pvpkTextures = m_pkTexturePool->Find( rstrFilename );

		if( pvpkTextures && pvpkTextures->size() )
		{
			vector< NeoEngine::Texture* >::const_iterator ppkTexture = pvpkTextures->begin();
			vector< NeoEngine::Texture* >::const_iterator ppkEnd     = pvpkTextures->end();

			for( ; ppkTexture != ppkEnd; ++ppkTexture )
				if( ( (*ppkTexture)->GetType() == eTextureType ) && ( ( eTextureFormat == Texture::DEFAULT ) || ( (*ppkTexture)->GetFormat() == eTextureFormat ) ) )
					return *ppkTexture;
		}
	}

	Texture *pkTexture = new Texture( this, rstrFilename, eTextureType, eTextureFormat, uiFiltering, uiMaxAnisotropy, m_pkTexturePool, m_pkFileManager );

	vector< ImageCodec* >::iterator ppkCodec    = m_vpkCodecs.begin();
	vector< ImageCodec* >::iterator ppkCodecEnd = m_vpkCodecs.end();

	for( ; ppkCodec != ppkCodecEnd; ++ppkCodec )
	{
		string strFilename = rstrFilename + ".";

		const vector<string> &rvstrExtensions = (*ppkCodec)->GetExtensions();

		vector<string>::const_iterator pstrExt    = rvstrExtensions.begin();
		vector<string>::const_iterator pstrExtEnd = rvstrExtensions.end();

		for( ; pstrExt != pstrExtEnd; ++pstrExt )
		{
			string strThisFile = strFilename + (*pstrExt);

			if( pkTexture->Load( strThisFile, uiFlags ) )
				return pkTexture;
		}
	}

	neolog << LogLevel( WARNING ) << "*** unable to load texture: file not found or not valid [" << rstrFilename << "]" << endl;

	delete pkTexture;

	return 0;
}






NeoEngine::TexturePtr Device::CreateTexture( const std::string &rstrName, Texture::TEXTURETYPE eTextureType, Texture::TEXTUREFORMAT eTextureFormat )
{
	return new Texture( this, rstrName, eTextureType, eTextureFormat, m_uiDefaultTextureFilter, m_uiDefaultMaxAnisotropy, m_pkTexturePool, m_pkFileManager );
}




int Texture::s_iIDCount = 0;

const D3DCUBEMAP_FACES Texture::s_aeD3DCubeFaces[6] = {
			
	D3DCUBEMAP_FACE_POSITIVE_X,
	D3DCUBEMAP_FACE_NEGATIVE_X,
	D3DCUBEMAP_FACE_POSITIVE_Y,
	D3DCUBEMAP_FACE_NEGATIVE_Y,
	D3DCUBEMAP_FACE_POSITIVE_Z,
	D3DCUBEMAP_FACE_NEGATIVE_Z

};


Texture::Texture( Device *pkDevice, const std::string &rstrName, TEXTURETYPE eTextureType, TEXTUREFORMAT eTextureFormat, unsigned int uiFiltering, unsigned int uiMaxAnisotropy, TexturePool *pkTexturePool, FileManager *pkFileManager ) :
	NeoEngine::Texture( rstrName, pkTexturePool ),
	LoadableEntity( pkFileManager ),
	m_pkDevice( pkDevice ),
	m_pkD3DTexture( 0 )
{
	m_iID             = ++s_iIDCount;
	m_eType           = eTextureType;
	m_eFormat         = eTextureFormat;
	m_uiFiltering     = uiFiltering;
	m_uiMaxAnisotropy = uiMaxAnisotropy;
}


Texture::~Texture()
{
	if( m_pkD3DTexture )
		m_pkD3DTexture->Release(), m_pkD3DTexture = 0;

	neolog << LogLevel( DEBUG ) << "Deallocated texture [" << GetName() << "]" << endl;
}


bool Texture::LoadNode( unsigned int uiFlags )
{
	bool bRet = true;

	string strName = m_pkFile->GetName();

	strName = strName.substr( 0, strName.find_last_of( '.' ) );

	if( GetName() != strName )
		SetName( strName );

	//Read and create new texture from image data
	ImageData *pkImageData = m_pkDevice->LoadImageData( m_pkFile );
	
	if( !pkImageData )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to load texture: unable to load image data from file [" << m_pkFile->GetName() << "]" << endl;
		return false;
	}	

	if( !UploadImage( pkImageData, m_eType, m_eFormat, uiFlags, m_uiFiltering, m_uiMaxAnisotropy ) )
		bRet = false;

	m_pkDevice->FreeImageData( pkImageData );

	return bRet;
}


bool Texture::UploadImage( ImageData *pkImageData, TEXTURETYPE eTextureType, TEXTUREFORMAT eTextureFormat, unsigned int uiFlags, unsigned int uiFiltering, unsigned int uiMaxAnisotropy )
{
	if( m_pkD3DTexture )
		m_pkD3DTexture->Release(), m_pkD3DTexture = 0;

	m_iWidth          = pkImageData->m_iWidth;
	m_iHeight         = ( eTextureType == Texture::CUBEMAP ) ? pkImageData->m_iHeight / 6 : pkImageData->m_iHeight;
	m_eFormat         = pkImageData->m_eFormat;
	m_uiFiltering     = uiFiltering ? uiFiltering : m_pkDevice->m_uiDefaultTextureFilter;
	m_uiMaxAnisotropy = uiMaxAnisotropy;
	m_eType           = eTextureType;

	int iLevels = 0;
	int iComp   = 0;

	if( uiFlags & Texture::NOMIPMAPS )
		iLevels = 1;

	if( m_pkD3DTexture )
		m_pkD3DTexture->Release(), m_pkD3DTexture = 0;


	D3DFORMAT     eInternalFormat;
	TEXTUREFORMAT eDataFormat;
	int r,g,b,a;
	int iSize;

	switch( m_eFormat )
	{
		case Texture::ALPHA:
		{
			eInternalFormat    = D3DFMT_A8;
			eDataFormat        = Texture::ALPHA;
			r = g = b          = 0;
			a                  = 8;
			iComp              = 1;
			
			m_iInternalFormat  = Texture::INTERNALFORMAT_ALPHA8;
			
			break;
		}

		case Texture::RGB:
		{
			eInternalFormat    = D3DFMT_R8G8B8;
			eDataFormat        = Texture::RGB;
			r = g = b          = 8;
			a                  = 0;
			iComp              = 3;
			
			m_iInternalFormat  = Texture::INTERNALFORMAT_RGB888;

			break;
		}

		case Texture::RGBA:
		{
			eInternalFormat    = D3DFMT_A8R8G8B8;
			eDataFormat        = Texture::RGBA;
			r = g = b = a      = 8;
			iComp              = 4;

			m_iInternalFormat  = Texture::INTERNALFORMAT_RGBA8888;

			break;
		}
		
		default:
		{
			neolog << LogLevel( ERROR ) << "*** Unable to upload texture: invalid texture format [" << (int)m_eFormat << "]" << endl;
			return false;
		}
	}

	if( ( eTextureFormat != Texture::DEFAULT ) && ( eTextureFormat != m_eFormat ) )
	{
		switch( eTextureFormat )
		{
			case Texture::ALPHA:
			{
				eInternalFormat    = D3DFMT_A8;
				m_iInternalFormat  = Texture::INTERNALFORMAT_ALPHA8;
				r = g = b          = 0;
				a                  = 8;
				
				break;
			}

			case Texture::RGB:
			{
				eInternalFormat    = D3DFMT_R8G8B8;
				m_iInternalFormat  = Texture::INTERNALFORMAT_RGB888;
				r = g = b          = 8;
				a                  = 0;

				break;
			}

			case Texture::RGBA:
			{
				eInternalFormat    = D3DFMT_A8R8G8B8;
				m_iInternalFormat  = Texture::INTERNALFORMAT_RGBA8888;
				r = g = b = a      = 8;

				break;
			}

			default:
			{
				neolog << LogLevel( WARNING ) << "*** Unable to upload texture: invalid requested texture format [" << (int)eTextureFormat << "]" << endl;
				return false;
			}
		}
	}

	int iNewWidth = lowpow( m_iWidth );
	int iNewHeight = ( m_eType == Texture::CUBEMAP ) ? iNewWidth : lowpow( m_iHeight );

	unsigned char *pucSavedImage = 0;

	if( ( m_iWidth != iNewWidth ) || ( m_iHeight != iNewHeight ) )
	{
		//Resize image
		unsigned char *pucImage = new unsigned char[ iComp * iNewWidth * iNewHeight * ( ( eTextureType == Texture::CUBEMAP ) ? 6 : 1 ) ];

		ImageData::ScaleImage( iComp, pkImageData->m_iWidth, pkImageData->m_iHeight, pkImageData->m_pucData, iNewWidth, iNewHeight, pucImage );

		pucSavedImage = pkImageData->m_pucData;

		pkImageData->m_pucData = pucImage;
	}

	iSize = iNewWidth * iNewHeight * iComp;

	IDirect3DTexture9     *pkSysTex     = 0;
	IDirect3DCubeTexture9 *pkSysCubeTex = 0;

	while( !pkSysTex && !pkSysCubeTex )
	{
		if( m_eType == Texture::TEX2D )
			m_pkDevice->m_pkDevice->CreateTexture( iNewWidth, iNewHeight, 1, 0, eInternalFormat, D3DPOOL_SYSTEMMEM, &pkSysTex, 0 );
		else if( m_eType == Texture::CUBEMAP )
			m_pkDevice->m_pkDevice->CreateCubeTexture( iNewWidth, 1, 0, eInternalFormat, D3DPOOL_SYSTEMMEM, &pkSysCubeTex, 0 );
		else
		{
			neolog << LogLevel( ERROR ) << "*** Unable to upload image: invalid texture type [" << (int)m_eType << "]" << endl;
			return false;
		}

		if( !pkSysTex && !pkSysCubeTex )
		{
			switch( eInternalFormat )
			{
				case D3DFMT_L8:
					eInternalFormat = D3DFMT_A8;
					break;

				case D3DFMT_A8:
					eInternalFormat = D3DFMT_A8R8G8B8;
					r = g = b        = 8;
					break;

				case D3DFMT_R8G8B8:
					eInternalFormat = D3DFMT_X8R8G8B8;
					a                = 8;
					break;

				case D3DFMT_X8R8G8B8:
					eInternalFormat = D3DFMT_A8R8G8B8;
					break;

				default:
				case D3DFMT_A8R8G8B8:
					neolog << LogLevel( ERROR ) << "*** Unable to upload texture: unable to find suitable conversion format" << endl;
					return false;
			}
		}
	}


	D3DLOCKED_RECT kRect;
	HRESULT        hResult;

	if( m_eType == Texture::TEX2D )
	{
		pkSysTex->LockRect( 0, &kRect, 0, 0/*D3DLOCK_DISCARD*/ );

		//Format conversion
		ConvertFormat( eDataFormat, eInternalFormat, pkImageData->m_pucData, iNewWidth, iNewHeight, (unsigned char*)kRect.pBits, kRect.Pitch );

		pkSysTex->UnlockRect( 0 );

		m_pkDevice->m_pkDevice->CreateTexture( iNewWidth, iNewHeight, 1, 0, eInternalFormat, D3DPOOL_DEFAULT, (IDirect3DTexture9**)&m_pkD3DTexture, 0 );

		hResult = m_pkDevice->m_pkDevice->UpdateTexture( pkSysTex, m_pkD3DTexture );

		pkSysTex->Release();
	}
	else if( m_eType == Texture::CUBEMAP )
	{
		IDirect3DSurface9 *pkSysSurface = 0;

		for( int iFace = 0; iFace < 6; ++iFace )
		{
			pkSysSurface = 0;

			pkSysCubeTex->GetCubeMapSurface( s_aeD3DCubeFaces[iFace], 0, &pkSysSurface );

			pkSysSurface->LockRect( &kRect, 0, 0 );

			ConvertFormat( eDataFormat, eInternalFormat, pkImageData->m_pucData + iSize * iFace, iNewWidth, iNewHeight, (unsigned char*)kRect.pBits, kRect.Pitch );

			pkSysSurface->UnlockRect();
		}

		m_pkDevice->m_pkDevice->CreateCubeTexture( iNewWidth, 1, 0, eInternalFormat, D3DPOOL_DEFAULT, (IDirect3DCubeTexture9**)&m_pkD3DTexture, 0 );

		hResult = m_pkDevice->m_pkDevice->UpdateTexture( pkSysCubeTex, m_pkD3DTexture );

		pkSysCubeTex->Release();
	}
	else
	{
		neolog << LogLevel( ERROR ) << "*** Unable to upload image: invalid texture type [" << (int)m_eType << "]" << endl;
		return false;
	}


	if( hResult != D3D_OK )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to upload texture: unable to update texture [" << hex << (unsigned int)hResult << dec << "]" << endl;

		m_pkD3DTexture->Release(), m_pkD3DTexture = 0;

		return false;
	}

	if( !( uiFlags & Texture::NOMIPMAPS ) /*&& ( m_eType != Texture::CUBEMAP ) */ )
		m_pkD3DTexture->GenerateMipSubLevels();

	if( uiFlags & Texture::NOMIPMAPS )
		m_uiFiltering &= 0xFF00FFFF; // Mask away mipmap filter to none

	neolog << LogLevel( DEBUG ) << "Texture [" << GetName() << "] " << m_iWidth << "x" << m_iHeight << " -> [" << iNewWidth << "x" << iNewHeight << "] ARGB" << a << r << g << b << ( ( eTextureType != Texture::CUBEMAP ) ? "" : " x 6" ) << endl;

	m_iWidth  = iNewWidth;
	m_iHeight = iNewHeight;
	m_fWidth  = (float)m_iWidth;
	m_fHeight = (float)m_iHeight;
	m_fRatio  = ( m_iWidth > m_iHeight ) ? (float)m_iWidth / (float)m_iHeight : (float)m_iHeight / (float)m_iWidth;

	if( pucSavedImage )
	{
		delete [] pkImageData->m_pucData;

		pkImageData->m_pucData = pucSavedImage;
	}

	return true;
}


bool Texture::ConvertFormat( TEXTUREFORMAT eSrcFormat, D3DFORMAT eDestFormat, unsigned char *pucSrc, int iWidth, int iHeight, unsigned char *pucDest, unsigned int uiDestPitch )
{
	switch( eSrcFormat )
	{
		case Texture::ALPHA:
		{
			switch( eDestFormat )
			{
				case D3DFMT_A8:
				{
					for( int y = 0; y < iHeight; ++y )
					{
						fmemcpy( pucDest, pucSrc, iWidth );

						pucSrc  += iWidth;
						pucDest += uiDestPitch;
					}

					return true;
				}

				case D3DFMT_A8R8G8B8:
				{
					for( int y = 0; y < iHeight; ++y )
					{
						unsigned char *pucLineDest = pucDest;

						for( int x = 0; x < iWidth; ++x )
						{
							*pucDest++ = 255;
							*pucDest++ = 255;
							*pucDest++ = 255;
							*pucDest++ = *pucSrc++;
						}

						pucDest = pucLineDest + uiDestPitch;
					}

					return true;
				}

				default:
					break;
			}

			break;
		}

		case Texture::RGB:
		{
			switch( eDestFormat )
			{
				case D3DFMT_R8G8B8:
				{
					for( int y = 0; y < iHeight; ++y )
					{
						unsigned char *pucLineDest = pucDest;

						for( int x = 0; x < iWidth; ++x )
						{
							*pucDest++ = *( pucSrc + 2 );
							*pucDest++ = *( pucSrc + 1 );
							*pucDest++ = *( pucSrc     );

							pucSrc += 3;
						}

						pucDest = pucLineDest + uiDestPitch;
					}

					return true;
				}

				case D3DFMT_X8R8G8B8:
				case D3DFMT_A8R8G8B8:
				{
					for( int y = 0; y < iHeight; ++y )
					{
						unsigned char *pucLineDest = pucDest;

						for( int x = 0; x < iWidth; ++x )
						{
							*pucDest++ = *( pucSrc + 2 );
							*pucDest++ = *( pucSrc + 1 );
							*pucDest++ = *( pucSrc     );
							*pucDest++ = 255;

							pucSrc += 3;
						}

						pucDest = pucLineDest + uiDestPitch;
					}

					return true;
				}

				default:
					break;
			}

			break;
		}

		case Texture::RGBA:
		{
			switch( eDestFormat )
			{
				case D3DFMT_A8R8G8B8:
				{
					for( int y = 0; y < iHeight; ++y )
					{
						unsigned char *pucLineDest = pucDest;

						for( int x = 0; x < iWidth; ++x )
						{
							*pucDest++ = *( pucSrc + 2 );
							*pucDest++ = *( pucSrc + 1 );
							*pucDest++ = *( pucSrc     );

							pucSrc += 3;

							*pucDest++ = *pucSrc++;
						}

						pucDest = pucLineDest + uiDestPitch;
					}

					return true;
				}

				default:
					break;
			}

			break;
		}

		default:
			break;
	}

	return false;
}


}; // namespace NeoD3D9

