/***************************************************************************
                   textureunit.cpp  -  Texture unit management
                             -------------------
    begin                : Sun Mar 16 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoDevD3D9, textureunit.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "textureunit.h"

using namespace NeoEngine;


namespace NeoD3D9
{


unsigned int TextureUnit::s_auiFilterMode[4] = { D3DTEXF_NONE, D3DTEXF_POINT, D3DTEXF_LINEAR, D3DTEXF_ANISOTROPIC };



TextureUnit::TextureUnit( Device *pkDevice, unsigned int uiID ) :
	m_pkDevice( pkDevice ),
	m_uiID( uiID ),
	m_pkCurTexture( 0 ),
	m_uiUV( uiID ),
	m_bGotTransform( false ),
	m_uiUVAddress( TextureLayer::WRAP ),
	m_uiMaxAnisotropy( 1 )
{
	m_aucFilter[0] = Texture::LINEAR;
	m_aucFilter[1] = Texture::LINEAR;
	m_aucFilter[2] = Texture::NEAREST;

	m_pkDevice->m_pkDevice->SetSamplerState( m_uiID, D3DSAMP_MINFILTER, s_auiFilterMode[ m_aucFilter[0] ] );
	m_pkDevice->m_pkDevice->SetSamplerState( m_uiID, D3DSAMP_MAGFILTER, s_auiFilterMode[ m_aucFilter[1] ] );
	m_pkDevice->m_pkDevice->SetSamplerState( m_uiID, D3DSAMP_MIPFILTER, s_auiFilterMode[ m_aucFilter[2] ] );

	m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_TEXTURE   );
	m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_CURRENT   );

	m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_ALPHAARG1, D3DTA_TEXTURE   );
	m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_ALPHAARG2, D3DTA_CURRENT   );

	if( m_uiID == 0 )
		m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP,   D3DTOP_MODULATE );
	else
		m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP,   D3DTOP_DISABLE  );

	m_eTransformState = D3DTRANSFORMSTATETYPE( D3DTS_TEXTURE0 + m_uiID );

	m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE );
}


TextureUnit::~TextureUnit()
{
}


void TextureUnit::Disable()
{
	m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP,   D3DTOP_DISABLE );
}


void TextureUnit::SetTransform( Matrix *pkMatrix, unsigned int uiDim )
{
	if( !pkMatrix && !m_bGotTransform )
		return;
	
	//D3DTSS_TCI_CAMERASPACEREFLECTIONVECTOR
	//D3DTSS_TCI_CAMERASPACENORMAL
	if( m_uiUV & 0xFFFF0000 )
		uiDim = 3;

	if( pkMatrix )
	{
		pkMatrix->Transpose();

		// Move translation component down according to dimension of texture coord
		if( uiDim == 1 )
			pkMatrix->m_aafMatrix[1][0] = pkMatrix->m_aafMatrix[3][0];
		else if( uiDim == 2 )
		{
			pkMatrix->m_aafMatrix[2][0] = pkMatrix->m_aafMatrix[3][0];
			pkMatrix->m_aafMatrix[2][1] = pkMatrix->m_aafMatrix[3][1];
		}

		m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT1 + ( uiDim - 1 ) );
		m_pkDevice->m_pkDevice->SetTransform( m_eTransformState, (D3DMATRIX*)pkMatrix );
	}
	else
		m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE );

	m_bGotTransform = ( pkMatrix != 0 );
}


bool TextureUnit::SetBlendMode( unsigned int uiSrcMode, unsigned int uiDstMode, float fFactor )
{
	//We don't want to be called on primary texture unit
	assert( m_uiID );

	m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_TEXTURE  );
	m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_CURRENT  );

	m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_ALPHAARG1, D3DTA_TEXTURE  );
	m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_ALPHAARG2, D3DTA_CURRENT  );

	m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1 );

	//blend output is source * sourcemode + dest * destmode
	//where source is this texture stage
	//and   dest   is input from last stage

	//default setup is arg1 - stage texture
	//                 arg2 - input from last stage

	switch( uiSrcMode )
	{	
		case BlendMode::SRC_ONE:
		{
			switch( uiDstMode )
			{	
				case BlendMode::DEST_ZERO:
				{
					// RGBf = 1 * RGBs + 0 * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_SELECTARG1 );
					break;
				}

				case BlendMode::DEST_ONE:
				{
					// RGBf = 1 * RGBs + 1 * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_ADD );
					break;
				}

				case BlendMode::DEST_SRCCOLOR:
				{
					//RGBf = 1 * RGBs + RGBs * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG0, D3DTA_TEXTURE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MULTIPLYADD );
					break;
				}

				case BlendMode::DEST_ONEMINUSSRCCOLOR:
				{
					// RGBf = 1 * RGBs + ( 1 - RGBs ) * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_ADDSMOOTH );
					break;
				}

				case BlendMode::DEST_SRCALPHA:
				{
					// RGBf = 1 * RGBs + As * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MODULATEALPHA_ADDCOLOR );
					break;
				}

				case BlendMode::DEST_ONEMINUSSRCALPHA:
				{
					// RGBf = 1 * RGBs + ( 1 - As ) * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MODULATEINVALPHA_ADDCOLOR );
					break;
				}

				case BlendMode::DEST_DESTALPHA:
				{
					// RGBf = 1 * RGBs + Ad * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_ALPHAARG1, D3DTA_CURRENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MODULATEALPHA_ADDCOLOR );
					break;
				}

				case BlendMode::DEST_ONEMINUSDESTALPHA:
				{
					// RGBf = 1 * RGBs + ( 1 - Ad ) * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_ALPHAARG1, D3DTA_CURRENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MODULATEINVALPHA_ADDCOLOR );
					break;
				}

				default:
					assert( false );
			}

			break;
		}

		case BlendMode::SRC_ZERO:
		{
			switch( uiDstMode )
			{
				case BlendMode::DEST_ZERO:
				{
					// RGBf = 0 * RGBs + 0 * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_CURRENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_SUBTRACT );
					break;
				}

				case BlendMode::DEST_ONE:
				{
					// RGBf = 0 * RGBs + 1 * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_SELECTARG2 );
					break;
				}

				case BlendMode::DEST_SRCCOLOR:
				{
					//RGBf = 0 * RGBs + RGBs * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MODULATE );
					break;
				}

				case BlendMode::DEST_ONEMINUSSRCCOLOR:
				{
					// RGBf = 0 * RGBs + ( 1 - RGBs ) * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_TEXTURE | D3DTA_COMPLEMENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MODULATE );
					break;
				}

				case BlendMode::DEST_SRCALPHA:
				{
					// RGBf = 0 * RGBs + As * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_TEXTURE | D3DTA_ALPHAREPLICATE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MODULATE );
					break;
				}

				case BlendMode::DEST_ONEMINUSSRCALPHA:
				{
					// RGBf = 0 * RGBs + ( 1 - As ) * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_TEXTURE | D3DTA_COMPLEMENT | D3DTA_ALPHAREPLICATE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MODULATE );
					break;
				}

				case BlendMode::DEST_DESTALPHA:
				{
					// RGBf = 0 * RGBs + Ad * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_CURRENT | D3DTA_ALPHAREPLICATE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MODULATE );
					break;
				}

				case BlendMode::DEST_ONEMINUSDESTALPHA:
				{
					// RGBf = 0 * RGBs + ( 1 - Ad ) * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_CURRENT | D3DTA_COMPLEMENT | D3DTA_ALPHAREPLICATE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MODULATE );
					break;
				}

				default:
					assert( false );
			}
		
			break;
		}

		case BlendMode::SRC_DESTCOLOR:
		{
			switch( uiDstMode )
			{
				case BlendMode::DEST_ZERO:
				{
					// RGBf = RGBd * RGBs + 0 * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MODULATE );
					break;
				}

				case BlendMode::DEST_ONE:
				{
					// RGBf = RGBd * RGBs + 1 * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_CURRENT );
					//m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_CURRENT ); already set as default
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG0, D3DTA_TEXTURE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MULTIPLYADD );
					break;
				}

				case BlendMode::DEST_SRCCOLOR:
				{
					//RGBf = RGBd * RGBs + RGBs * RGBd = ( RGBs * RGBd ) << 1
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MODULATE2X );
					break;
				}

				case BlendMode::DEST_ONEMINUSSRCCOLOR:
				{
					// RGBf = RGBd * RGBs + ( 1 - RGBs ) * RGBd = RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_SELECTARG2 );
					break;
				}

				case BlendMode::DEST_SRCALPHA:
				{
					// RGBf = RGBd * RGBs + As * RGBd
					return false;
				}

				case BlendMode::DEST_ONEMINUSSRCALPHA:
				{
					// RGBf = RGBd * RGBs + ( 1 - As ) * RGBd
					return false;
				}

				case BlendMode::DEST_DESTALPHA:
				{
					// RGBf = RGBd * RGBs + Ad * RGBd
					return false;
				}

				case BlendMode::DEST_ONEMINUSDESTALPHA:
				{
					// RGBf = RGBd * RGBs + ( 1 - Ad ) * RGBd
					return false;
				}

				default:
					assert( false );
			}
		
			break;
		}


		case BlendMode::SRC_ONEMINUSDESTCOLOR:
		{
			switch( uiDstMode )
			{
				case BlendMode::DEST_ZERO:
				{
					// RGBf = ( 1 - RGBd ) * RGBs + 0 * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_CURRENT | D3DTA_COMPLEMENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MODULATE );
					break;
				}

				case BlendMode::DEST_ONE:
				{
					// RGBf = ( 1 - RGBd ) * RGBs + 1 * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_CURRENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_TEXTURE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG0, D3DTA_CURRENT | D3DTA_COMPLEMENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MULTIPLYADD );
					break;
				}

				case BlendMode::DEST_SRCCOLOR:
				{
					//RGBf = ( 1 - RGBd ) * RGBs + RGBs * RGBd = RGBs
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_SELECTARG1 );
					break;
				}

				case BlendMode::DEST_ONEMINUSSRCCOLOR:
				{
					// RGBf = ( 1 - RGBd ) * RGBs + ( 1 - RGBs ) * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_CURRENT | D3DTA_COMPLEMENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG0, D3DTA_CURRENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_LERP );
					break;
				}

				case BlendMode::DEST_SRCALPHA:
				{
					// RGBf = ( 1 - RGBd ) * RGBs + As * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_CURRENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_TEXTURE | D3DTA_ALPHAREPLICATE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG0, D3DTA_TEXTURE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_LERP );
					break;
				}

				case BlendMode::DEST_ONEMINUSSRCALPHA:
				{
					// RGBf = ( 1 - RGBd ) * RGBs + ( 1 - As ) * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_CURRENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_TEXTURE | D3DTA_COMPLEMENT | D3DTA_ALPHAREPLICATE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG0, D3DTA_TEXTURE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_LERP );
					break;
				}

				case BlendMode::DEST_DESTALPHA:
				{
					// RGBf = ( 1 - RGBd ) * RGBs + Ad * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_CURRENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_CURRENT | D3DTA_ALPHAREPLICATE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG0, D3DTA_TEXTURE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_LERP );
					break;
				}

				case BlendMode::DEST_ONEMINUSDESTALPHA:
				{
					// RGBf = ( 1 - RGBd ) * RGBs + ( 1 - Ad ) * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_CURRENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_CURRENT | D3DTA_COMPLEMENT | D3DTA_ALPHAREPLICATE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG0, D3DTA_TEXTURE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_LERP );
					break;
				}

				default:
					assert( false );
			}
		
			break;
		}

		case BlendMode::SRC_SRCCOLOR:
		{
			switch( uiDstMode )
			{
				case BlendMode::DEST_ZERO:
				{
					// RGBf = RGBs * RGBs + 0 * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_TEXTURE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MODULATE );
					break;
				}

				case BlendMode::DEST_ONE:
				{
					// RGBf = RGBs * RGBs + 1 * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_CURRENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_TEXTURE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG0, D3DTA_TEXTURE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MULTIPLYADD );
					break;
				}

				case BlendMode::DEST_SRCCOLOR:
				{
					//RGBf = RGBs * RGBs + RGBs * RGBd = RGBs * ( 1 + RGBd )
					return false;
				}

				case BlendMode::DEST_ONEMINUSSRCCOLOR:
				{
					// RGBf = RGBs * RGBs + ( 1 - RGBs ) * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_TEXTURE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG0, D3DTA_CURRENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_LERP );
					break;
				}

				case BlendMode::DEST_SRCALPHA:
				{
					// RGBf = RGBs * RGBs + As * RGBd
					return false;
				}

				case BlendMode::DEST_ONEMINUSSRCALPHA:
				{
					// RGBf = RGBs * RGBs + ( 1 - As ) * RGBd
					return false;
				}

				case BlendMode::DEST_DESTALPHA:
				{
					// RGBf = RGBs * RGBs + Ad * RGBd
					return false;
				}

				case BlendMode::DEST_ONEMINUSDESTALPHA:
				{
					// RGBf = RGBs * RGBs + ( 1 - Ad ) * RGBd
					return false;
				}

				default:
					assert( false );
			}
		
			break;
		}

		case BlendMode::SRC_FACTOR:
		{
			switch( uiDstMode )
			{
			case BlendMode::DEST_ONEMINUSFACTOR:
				{
					// RGBf = RGBs * f + ( 1 - f ) * RGBd
					m_pkDevice->m_pkDevice->SetRenderState( D3DRS_TEXTUREFACTOR, D3DCOLOR_COLORVALUE( fFactor, fFactor, fFactor, fFactor ) );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_BLENDFACTORALPHA );
					break;
				}

				default:
					assert( false );
			}

			break;
		}


		case BlendMode::SRC_SRCALPHA:
		{
			switch( uiDstMode )
			{
				case BlendMode::DEST_ZERO:
				{
					// RGBf = As * RGBs + 0 * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_TEXTURE | D3DTA_ALPHAREPLICATE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MODULATE );
					break;
				}

				case BlendMode::DEST_ONE:
				{
					// RGBf = As * RGBs + 1 * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_CURRENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_TEXTURE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG0, D3DTA_TEXTURE | D3DTA_ALPHAREPLICATE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MULTIPLYADD );
					break;
				}

				case BlendMode::DEST_SRCCOLOR:
				{
					//RGBf = As * RGBs + RGBs * RGBd = RGBs * ( As + RGBd )
					return false;
				}

				case BlendMode::DEST_ONEMINUSSRCCOLOR:
				{
					// RGBf = As * RGBs + ( 1 - RGBs ) * RGBd
					return false;
				}

				case BlendMode::DEST_SRCALPHA:
				{
					// RGBf = As * RGBs + As * RGBd = As * ( RGBs + RGBd )
					return false;
				}

				case BlendMode::DEST_ONEMINUSSRCALPHA:
				{
					// RGBf = As * RGBs + ( 1 - As ) * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_BLENDTEXTUREALPHA );

					// Af   = As + Ad
					// This allows two layers with srcalpha|oneminussrcalpha to be rendered
					// correctly, since the final framebuffer blend of srcalpha|oneminussrcalpha
					// should not use only base layer texture alpha, since that would also mask out
					// parts of second layer
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_ALPHAOP,   D3DTOP_ADD );
					break;
				}

				case BlendMode::DEST_DESTALPHA:
				{
					// RGBf = As * RGBs + Ad * RGBd
					return false;
				}

				case BlendMode::DEST_ONEMINUSDESTALPHA:
				{
					// RGBf = As * RGBs + ( 1 - Ad ) * RGBd
					return false;
				}

				default:
					assert( false );
			}
		
			break;
		}

		case BlendMode::SRC_ONEMINUSSRCALPHA:
		{
			switch( uiDstMode )
			{
				case BlendMode::DEST_ZERO:
				{
					// RGBf = ( 1 - As ) * RGBs + 0 * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_TEXTURE | D3DTA_ALPHAREPLICATE | D3DTA_COMPLEMENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MODULATE );
					break;
				}

				case BlendMode::DEST_ONE:
				{
					// RGBf = ( 1 - As ) * RGBs + 1 * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_CURRENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_TEXTURE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG0, D3DTA_TEXTURE | D3DTA_ALPHAREPLICATE | D3DTA_COMPLEMENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MULTIPLYADD );
					break;
				}

				case BlendMode::DEST_SRCCOLOR:
				{
					//RGBf = ( 1 - As ) * RGBs + RGBs * RGBd = RGBs * ( ( 1 - As ) + RGBd )
					return false;
				}

				case BlendMode::DEST_ONEMINUSSRCCOLOR:
				{
					// RGBf = ( 1 - As ) * RGBs + ( 1 - RGBs ) * RGBd
					return false;
				}

				case BlendMode::DEST_SRCALPHA:
				{
					// RGBf = ( 1 - As ) * RGBs + As * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_CURRENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_TEXTURE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_BLENDTEXTUREALPHA );
					break;
				}

				case BlendMode::DEST_ONEMINUSSRCALPHA:
				{
					// RGBf = ( 1 - As ) * RGBs + ( 1 - As ) * RGBd = ( 1 - As ) * ( RGBs + RGBd )
					return false;
				}

				case BlendMode::DEST_DESTALPHA:
				{
					// RGBf = ( 1 - As ) * RGBs + Ad * RGBd
					return false;
				}

				case BlendMode::DEST_ONEMINUSDESTALPHA:
				{
					// RGBf = ( 1 - As ) * RGBs + ( 1 - Ad ) * RGBd
					return false;
				}

				default:
					assert( false );
			}
		
			break;
		}

		case BlendMode::SRC_DESTALPHA:
		{
			switch( uiDstMode )
			{
				case BlendMode::DEST_ZERO:
				{
					// RGBf = Ad * RGBs + 0 * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_CURRENT | D3DTA_ALPHAREPLICATE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MODULATE );
					break;
				}

				case BlendMode::DEST_ONE:
				{
					// RGBf = Ad * RGBs + 1 * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_CURRENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_TEXTURE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG0, D3DTA_CURRENT | D3DTA_ALPHAREPLICATE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MULTIPLYADD );
					break;
				}

				case BlendMode::DEST_SRCCOLOR:
				{
					//RGBf = Ad * RGBs + RGBs * RGBd = RGBs * ( Ad + RGBd )
					return false;
				}

				case BlendMode::DEST_ONEMINUSSRCCOLOR:
				{
					// RGBf = Ad * RGBs + ( 1 - RGBs ) * RGBd
					return false;
				}

				case BlendMode::DEST_SRCALPHA:
				{
					// RGBf = Ad * RGBs + As * RGBd
					return false;
				}

				case BlendMode::DEST_ONEMINUSSRCALPHA:
				{
					// RGBf = Ad * RGBs + ( 1 - As ) * RGBd
					return false;
				}

				case BlendMode::DEST_DESTALPHA:
				{
					// RGBf = Ad * RGBs + Ad * RGBd = Ad * ( RGBs + RGBd )
					return false;
				}

				case BlendMode::DEST_ONEMINUSDESTALPHA:
				{
					// RGBf = Ad * RGBs + ( 1 - Ad ) * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_BLENDCURRENTALPHA );
					break;
				}

				default:
					assert( false );
			}
		
			break;
		}

		case BlendMode::SRC_ONEMINUSDESTALPHA:
		{
			switch( uiDstMode )
			{
				case BlendMode::DEST_ZERO:
				{
					// RGBf = ( 1 - Ad ) * RGBs + 0 * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_CURRENT | D3DTA_ALPHAREPLICATE | D3DTA_COMPLEMENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MODULATE );
					break;
				}

				case BlendMode::DEST_ONE:
				{
					// RGBf = ( 1 - Ad ) * RGBs + 1 * RGBd
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_CURRENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_TEXTURE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG0, D3DTA_CURRENT | D3DTA_ALPHAREPLICATE | D3DTA_COMPLEMENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_MULTIPLYADD );
					break;
				}

				case BlendMode::DEST_SRCCOLOR:
				{
					//RGBf = ( 1 - Ad ) * RGBs + RGBs * RGBd = RGBs * ( ( 1 - Ad ) + RGBd )
					return false;
				}

				case BlendMode::DEST_ONEMINUSSRCCOLOR:
				{
					// RGBf = ( 1 - Ad ) * RGBs + ( 1 - RGBs ) * RGBd
					return false;
				}

				case BlendMode::DEST_SRCALPHA:
				{
					// RGBf = ( 1 - Ad ) * RGBs + As * RGBd
					return false;
				}

				case BlendMode::DEST_ONEMINUSSRCALPHA:
				{
					// RGBf = ( 1 - Ad ) * RGBs + ( 1 - As ) * RGBd
					return false;
				}

				case BlendMode::DEST_DESTALPHA:
				{
					// RGBf = ( 1 - Ad ) * RGBs + Ad * RGBd = Ad * ( RGBs + RGBd )
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG1, D3DTA_CURRENT );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLORARG2, D3DTA_TEXTURE );
					m_pkDevice->m_pkDevice->SetTextureStageState( m_uiID, D3DTSS_COLOROP, D3DTOP_BLENDCURRENTALPHA );
					break;
				}

				case BlendMode::DEST_ONEMINUSDESTALPHA:
				{
					// RGBf = ( 1 - Ad ) * RGBs + ( 1 - Ad ) * RGBd = ( 1 - Ad ) * ( RGBs + RGBd )
					break;
				}

				default:
					assert( false );
			}
		
			break;
		}

		
		case BlendMode::SRC_ALPHASATURATE:
		{
			return false;
		}

		default:
		{
			assert( false );
			break;
		}
	}

	return true;
}


}; // namespace NeoD3D9
