/***************************************************************************
                      vertexbuffer.cpp  -  Vertex buffer
                             -------------------
    begin                : Sun Mar 9 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoDevD3D9, vertexbuffer.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/


#include "device.h"
#include "vertexbuffer.h"
#include "buffermanager.h"
#include "bufferregion.h"

#include <neoengine/logstream.h>

using namespace std;
using namespace NeoEngine;



namespace NeoD3D9
{


VertexBufferPtr Device::CreateVertexBuffer( unsigned int uiType, unsigned int uiNumVertices, const VertexDeclaration *pkFormat, const void *pData )
{
	if( uiType & Buffer::NORENDER )
		return new NeoEngine::VertexBuffer( uiType, uiNumVertices, pkFormat, pData );

	//We don't care about NOREADWRITE buffers, infact, it will make us crash since we always need to be able to lock buffer
	uiType &= ~Buffer::NOREADWRITE;

	return new VertexBufferD3D( this, m_pkVertexBufferManager, uiType, uiNumVertices, pkFormat, pData );
}





VertexBufferD3D::VertexBufferD3D( Device *pkDevice, BufferManager *pkBufferManager, unsigned int uiType, unsigned int uiNumVertices, const VertexDeclaration *pkFormat, const void *pData ) :
	VertexBuffer( uiType, 0, 0 )
{
	m_pkD3DBuffer         = 0;
	m_pkDevice            = pkDevice;
	m_pkManager           = pkBufferManager;
	m_pkRegion            = 0;
	m_pkVertexDeclaration = 0;

	AllocateVertices( uiNumVertices, pkFormat, pData );
}


VertexBufferD3D::~VertexBufferD3D()
{
	m_pkManager->Free( m_pkRegion );

	//We don't want base buffer class to deallocate the pointer
	m_pucBuffer = 0;

	if( m_pkVertexDeclaration )
		m_pkVertexDeclaration->Release();
}


void VertexBufferD3D::AllocateVertices( unsigned int uiNumVertices, const VertexDeclaration *pkFormat, const void *pData )
{
	if( ( uiNumVertices == m_uiNumAllocated ) && pkFormat && m_pkVertexFormat && ( *m_pkVertexFormat == *pkFormat ) )
	{
		m_uiNumCurrent = m_uiNumAllocated;

		if( !pData )
			return;

		Lock( Buffer::WRITE );

		LoadVertexData( pData );

		Unlock();

		return;
	}

	m_pkManager->Free( m_pkRegion );

	m_pucBuffer           = 0;

	if( pkFormat )
	{
		delete m_pkVertexFormat;
		m_pkVertexFormat  = pkFormat->Duplicate();
	}

	m_uiVertexFormatSize  = m_pkVertexFormat ? m_pkVertexFormat->GetVertexSize() : 0;
	m_uiNumAllocated      = m_pkVertexFormat ? uiNumVertices : 0;
	m_uiNumCurrent        = m_uiNumAllocated;

	m_pkRegion            = m_pkManager->Alloc( m_uiNumAllocated * m_uiVertexFormatSize, m_uiVertexFormatSize );
	m_pkD3DBuffer         = m_pkDevice->GetVertexBufferStore();
	m_uiD3DFVF            = 0;

	static int aiD3DFVFTEX[9] = { D3DFVF_TEX0, D3DFVF_TEX1, D3DFVF_TEX2, D3DFVF_TEX3, D3DFVF_TEX4, D3DFVF_TEX5, D3DFVF_TEX6, D3DFVF_TEX7, D3DFVF_TEX8 };

	D3DVERTEXELEMENT9 aDeclaration[ MAXD3DDECLUSAGEINDEX + 1 ];

	int          iDeclaration   = 0;
	unsigned int uiLastTexLayer = 0;

	if( m_pkVertexFormat )
	{
		assert( m_pkVertexFormat->m_uiNumElements <= MAXD3DDECLUSAGEINDEX );

		for( unsigned int uiElement = 0; uiElement < m_pkVertexFormat->m_uiNumElements; ++uiElement )
		{
			VertexElement *pkElement = &m_pkVertexFormat->m_pkElements[ uiElement ];

			if( pkElement->m_uiUsage == VertexElement::POSITION )
			{
				m_uiD3DFVF |= D3DFVF_XYZ;

				aDeclaration[ iDeclaration   ].Stream     = 0;
				aDeclaration[ iDeclaration   ].Offset     = pkElement->m_uiOffset;
				aDeclaration[ iDeclaration   ].Type       = D3DDECLTYPE_FLOAT1 + pkElement->m_uiType;
				aDeclaration[ iDeclaration   ].Method     = D3DDECLMETHOD_DEFAULT;
				aDeclaration[ iDeclaration   ].Usage      = D3DDECLUSAGE_POSITION;
				aDeclaration[ iDeclaration++ ].UsageIndex = 0;
			}
			else if( pkElement->m_uiUsage == VertexElement::NORMAL )
			{
				m_uiD3DFVF |= D3DFVF_NORMAL;

				aDeclaration[ iDeclaration   ].Stream     = 0;
				aDeclaration[ iDeclaration   ].Offset     = pkElement->m_uiOffset;
				aDeclaration[ iDeclaration   ].Type       = D3DDECLTYPE_FLOAT1 + pkElement->m_uiType;
				aDeclaration[ iDeclaration   ].Method     = D3DDECLMETHOD_DEFAULT;
				aDeclaration[ iDeclaration   ].Usage      = D3DDECLUSAGE_NORMAL;
				aDeclaration[ iDeclaration++ ].UsageIndex = 0;
			}
			else if( pkElement->m_uiUsage == VertexElement::DIFFUSECOLOR )
			{
				m_uiD3DFVF |= D3DFVF_DIFFUSE;

				aDeclaration[ iDeclaration   ].Stream     = 0;
				aDeclaration[ iDeclaration   ].Offset     = pkElement->m_uiOffset;
				aDeclaration[ iDeclaration   ].Type       = ( pkElement->m_uiType == VertexElement::COLOR32 ) ? D3DDECLTYPE_D3DCOLOR : ( D3DDECLTYPE_FLOAT1 + pkElement->m_uiType );
				aDeclaration[ iDeclaration   ].Method     = D3DDECLMETHOD_DEFAULT;
				aDeclaration[ iDeclaration   ].Usage      = D3DDECLUSAGE_COLOR;
				aDeclaration[ iDeclaration++ ].UsageIndex = 0;
			}
			else if( pkElement->m_uiUsage == VertexElement::TEXCOORD )
			{
				if( pkElement->m_uiType == VertexElement::FLOAT )
				{
					m_uiD3DFVF |= D3DFVF_TEXCOORDSIZE1( pkElement->m_uiIndex );
					m_auiTexLayerDims[ pkElement->m_uiIndex ] = 1;
				}
				else if( pkElement->m_uiType == VertexElement::FLOAT2 )
				{
					m_uiD3DFVF |= D3DFVF_TEXCOORDSIZE2( pkElement->m_uiIndex );
					m_auiTexLayerDims[ pkElement->m_uiIndex ] = 2;
				}
				else if( pkElement->m_uiType == VertexElement::FLOAT3 )
				{
					m_uiD3DFVF |= D3DFVF_TEXCOORDSIZE3( pkElement->m_uiIndex );
					m_auiTexLayerDims[ pkElement->m_uiIndex ] = 3;
				}

				aDeclaration[ iDeclaration   ].Stream     = 0;
				aDeclaration[ iDeclaration   ].Offset     = pkElement->m_uiOffset;
				aDeclaration[ iDeclaration   ].Type       = D3DDECLTYPE_FLOAT1 + pkElement->m_uiType;
				aDeclaration[ iDeclaration   ].Method     = D3DDECLMETHOD_DEFAULT;
				aDeclaration[ iDeclaration   ].Usage      = D3DDECLUSAGE_TEXCOORD;
				aDeclaration[ iDeclaration++ ].UsageIndex = pkElement->m_uiIndex;

				if( pkElement->m_uiIndex >= uiLastTexLayer )
					uiLastTexLayer = pkElement->m_uiIndex + 1;
			}
			else if( pkElement->m_uiUsage == VertexElement::TANGENT )
			{
				aDeclaration[ iDeclaration   ].Stream     = 0;
				aDeclaration[ iDeclaration   ].Offset     = pkElement->m_uiOffset;
				aDeclaration[ iDeclaration   ].Type       = ( pkElement->m_uiType == VertexElement::COLOR32 ) ? D3DDECLTYPE_D3DCOLOR : ( D3DDECLTYPE_FLOAT1 + pkElement->m_uiType );
				aDeclaration[ iDeclaration   ].Method     = D3DDECLMETHOD_DEFAULT;
				aDeclaration[ iDeclaration   ].Usage      = D3DDECLUSAGE_TANGENT;
				aDeclaration[ iDeclaration++ ].UsageIndex = 0;
			}
			else if( pkElement->m_uiUsage == VertexElement::BINORMAL )
			{
				aDeclaration[ iDeclaration   ].Stream     = 0;
				aDeclaration[ iDeclaration   ].Offset     = pkElement->m_uiOffset;
				aDeclaration[ iDeclaration   ].Type       = ( pkElement->m_uiType == VertexElement::COLOR32 ) ? D3DDECLTYPE_D3DCOLOR : ( D3DDECLTYPE_FLOAT1 + pkElement->m_uiType );
				aDeclaration[ iDeclaration   ].Method     = D3DDECLMETHOD_DEFAULT;
				aDeclaration[ iDeclaration   ].Usage      = D3DDECLUSAGE_BINORMAL;
				aDeclaration[ iDeclaration++ ].UsageIndex = 0;
			}
		}
	}

	m_uiD3DFVF |= aiD3DFVFTEX[ uiLastTexLayer ];

	aDeclaration[ iDeclaration   ].Stream     = 0xFF;
	aDeclaration[ iDeclaration   ].Offset     = 0;
	aDeclaration[ iDeclaration   ].Type       = D3DDECLTYPE_UNUSED;
	aDeclaration[ iDeclaration   ].Method     = 0;
	aDeclaration[ iDeclaration   ].Usage      = 0;
	aDeclaration[ iDeclaration++ ].UsageIndex = 0;

	m_pkDevice->m_pkDevice->CreateVertexDeclaration( aDeclaration, &m_pkVertexDeclaration );

	if( !m_uiNumAllocated )
		return;

	if( pData )
	{
		Lock( Buffer::WRITE );

		LoadVertexData( pData );

		Unlock();
	}
}


void VertexBufferD3D::ReleaseLock()
{
#ifdef _DEBUG
	HRESULT hResult;
	if( D3D_OK != ( hResult = m_pkD3DBuffer->Unlock() ) )
		neolog << LogLevel( WARNING ) << "[VertexBufferD3D::ReleaseLock] unable to unlock D3D backing store buffer [" << (unsigned int)hResult << "]" << endl;
#else
	m_pkD3DBuffer->Unlock();
#endif

	m_pucBuffer = 0;

	VertexBuffer::ReleaseLock();
}


bool VertexBufferD3D::AcquireLock()
{
	DWORD dwFlags = 0;

	if( m_uiLock == READ )
	{
		dwFlags |= D3DLOCK_NOOVERWRITE;
		dwFlags |= D3DLOCK_READONLY;
	}

#ifdef _DEBUG
	HRESULT hResult;
	if( D3D_OK != ( hResult = m_pkD3DBuffer->Lock( m_pkRegion->m_uiRegionOffset, m_pkRegion->m_uiRegionSize, (void**)&m_pucBuffer, dwFlags ) ) )
		neolog << LogLevel( WARNING ) << "[VertexBufferD3D::AcquireLock] unable to lock D3D backing store buffer [" << (unsigned int)hResult << "]" << endl;
#else
	m_pkD3DBuffer->Lock( m_pkRegion->m_uiRegionOffset, m_pkRegion->m_uiRegionSize, (void**)&m_pucBuffer, dwFlags );
#endif

	return VertexBuffer::AcquireLock();
}


}; // namespace NeoD3D9




