/***************************************************************************
                    device.cpp  -  Export symbols for engine
                             -------------------
    begin                : Wed Sep 24 2003
    copyright            : (C) 2003 by emedia-solutions wolf
    email                : markus@emedia-solutions-wolf.de
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoDevDInput8, device.cpp

 The Initial Developer of the Original Code is Markus Wolf.
 Portions created by Markus Wold are Copyright (C) 2003
 emedia-solutions wolf. All Rights Reserved.

 ***************************************************************************/
#define INITGUID
#include "device.h"

// Helper macro
#define KEYDOWN(key) (m_aKeyData[key] & 0x80)
#define BUTTONDOWN(x) (x.dwData & 0x80)

namespace NeoDI {

HINSTANCE Device::s_hInstance = 0;

/*---------------------------------------------------------------------------*/
Device::Device( HWND hWnd, NeoEngine::InputManager *pkInputManager ) :
        NeoEngine::InputDevice( pkInputManager ),
        m_pkDI( 0 ),
        m_pkKeyboard( 0 ),
        m_hWnd( hWnd ),
        m_iKeyIndex( -1 ),
        m_eMouseCollectState(COLLECT),
        m_bKeyboardInitialized( false ),
		m_bKeyboardBuffered( false ),
        m_bMouseInitialized( false ),
		m_bMouseBuffered( false ),
        m_iZeroX( 0 ),
        m_iZeroY( 0 ),
		m_iZeroZ( 0 ),
		m_iX( 0 ),
		m_iY( 0 ),
		m_iZ( 0 ),
		m_iButtonState( 0 )
{
	// Get DirectInput
	HRESULT hResult;
	hResult = DirectInput8Create( s_hInstance, DIRECTINPUT_VERSION, IID_IDirectInput8, (void **)&m_pkDI, NULL );
	if( FAILED( hResult ) )
	{
		NeoEngine::neolog << NeoEngine::LogLevel( NeoEngine::ERROR ) << "**** Could not get direct input [" << hResult << "]" << std::endl;
		return;
	}
	InitializeMouse();
	InitializeKeyboard();
};

/*---------------------------------------------------------------------------*/
Device::~Device()
{
	if ( m_pkDI )
	{
		if( m_bMouseInitialized )
			ReleaseMouse();
		if( m_bKeyboardInitialized )
			ReleaseKeyboard();
		m_pkDI->Release();
		m_pkDI = 0;
	}
};

/*---------------------------------------------------------------------------*/
void Device::InitializeMouse() 
{
    HRESULT hResult;

    ZeroMemory( &m_kMouseData, sizeof( m_kMouseData ) );

    hResult = m_pkDI->CreateDevice( GUID_SysMouse, &m_pkMouse, NULL );
    if( FAILED( hResult ) ) 
	{
        NeoEngine::neolog << NeoEngine::LogLevel( NeoEngine::ERROR ) << "**** Failed to create mouse device [" << hResult << "]" << std::endl;
        return;
    }
    
    hResult = m_pkMouse->SetDataFormat( &c_dfDIMouse2 );
    if( FAILED( hResult ) )
	{
        NeoEngine::neolog << NeoEngine::LogLevel( NeoEngine::ERROR ) << "**** Failed to set data format [" << hResult << "]" << std::endl;
        return;
    }
    
    hResult = m_pkMouse->SetCooperativeLevel( m_hWnd, DISCL_FOREGROUND | DISCL_NONEXCLUSIVE );
    if( FAILED( hResult ) ) 
	{
        NeoEngine::neolog << NeoEngine::LogLevel( NeoEngine::ERROR ) << "**** Failed to set cooperative level [" << hResult << "]" << std::endl;
        return;
    }
    
    // Set absolute axis (instead of relative)
    {
        DIPROPDWORD kProperty;
        kProperty.diph.dwSize = sizeof( DIPROPDWORD );
        kProperty.diph.dwHeaderSize = sizeof( DIPROPHEADER );
        kProperty.diph.dwHow = DIPH_DEVICE;
        kProperty.diph.dwObj = 0;
        kProperty.dwData = DIPROPAXISMODE_ABS;
        hResult = m_pkMouse->SetProperty(DIPROP_AXISMODE, &kProperty.diph);
        if ( FAILED( hResult ) )
		{
            NeoEngine::neolog << NeoEngine::LogLevel( NeoEngine::ERROR ) << "**** Failed to set absolute mouse coords [" << hResult << "]" << std::endl;
        }
    }
    
    RECT r;
    if( GetClientRect( m_hWnd, &r ) )
	{
        SetMouseRange( r.left, r.right, r.top, r.bottom );
    }
    
	if( !SetBufferSize( m_pkMouse, 10 ) )
	{
		return;
	}

    hResult = m_pkMouse->Acquire();
    m_bMouseInitialized = true;
}

/*---------------------------------------------------------------------------*/
void Device::InitializeKeyboard()
{
    HRESULT hResult;

    hResult = m_pkDI->CreateDevice( GUID_SysKeyboard, &m_pkKeyboard, NULL );
    if( FAILED( hResult ) )
	{
        NeoEngine::neolog << NeoEngine::LogLevel( NeoEngine::ERROR ) << "**** Could not create keyboard device [" << hResult << "]" << std::endl;
        return;
    }

    hResult = m_pkKeyboard->SetDataFormat( &c_dfDIKeyboard );
    if( FAILED( hResult ) )
	{
        NeoEngine::neolog << NeoEngine::LogLevel( NeoEngine::ERROR ) << "**** Could not set keyboard layout [" << hResult << "]" << std::endl;
        return;
    }

    hResult = m_pkKeyboard->SetCooperativeLevel( m_hWnd, DISCL_FOREGROUND | DISCL_NONEXCLUSIVE );
    if( FAILED( hResult ) )
	{
        NeoEngine::neolog << NeoEngine::LogLevel( NeoEngine::ERROR ) << "**** Failed to set cooperative level [" << hResult << "]" << std::endl;
        return;
    }

	if( !SetBufferSize( m_pkKeyboard, 10 ) )
	{
		return;
	}

    hResult = m_pkKeyboard->Acquire();
    m_bKeyboardInitialized = true;
}

/*---------------------------------------------------------------------------*/
void Device::ReleaseMouse()
{
	if( m_bMouseInitialized )
	{
		m_pkMouse->Unacquire();
		m_pkMouse->Release();
		m_pkMouse = 0;
	}
	m_bMouseInitialized = false;
}

/*---------------------------------------------------------------------------*/
void Device::ReleaseKeyboard()
{
	if( m_bKeyboardInitialized )
	{
		m_pkKeyboard->Unacquire();
		m_pkKeyboard->Release();
		m_pkKeyboard = 0;
	}
	m_bKeyboardInitialized = false;
}

/*---------------------------------------------------------------------------*/
bool Device::Collect( NeoEngine::InputEvent *pkEvent )
{
    bool bEvent = false;
	if( HasInputEventGroup( MOUSEINPUT ) && m_bMouseInitialized )
	{
		if( m_bMouseBuffered )
		{
			bEvent = CollectBufferedMouse( pkEvent );
		} else {
			bEvent = CollectMouse( pkEvent );
		}
	}
	if( HasInputEventGroup( KEYBOARDINPUT ) && m_bKeyboardInitialized && !bEvent )
	{
		if( m_bKeyboardBuffered )
		{
			bEvent = CollectBufferedKeyboard( pkEvent );
		} else
		{
			bEvent = CollectKeyboard( pkEvent );
		}
	}
    return bEvent;
};

/*---------------------------------------------------------------------------*/
bool Device::CollectBufferedMouse( NeoEngine::InputEvent *pkEvent )
{
    HRESULT hResult;
    NeoEngine::InputEvent *pkNewEvent = 0;
	unsigned long ulSize = 1;
	DIDEVICEOBJECTDATA aData[ 1 ];

	hResult = m_pkMouse->GetDeviceData( sizeof( DIDEVICEOBJECTDATA ), aData, &ulSize, 0 );
	if( ( hResult == DIERR_NOTACQUIRED ) || ( hResult == DIERR_INPUTLOST ) )
	{
		NeoEngine::neolog << NeoEngine::LogLevel( NeoEngine::INFO ) << "**** Attempt to reaquire device" << std::endl;
		hResult = m_pkMouse->Acquire();
		return false;
	}

	if( ulSize == 0 )
	{
		return false;
	}

	switch( aData[ 0 ].dwOfs )
	{
		case DIMOFS_X:
		{
			// Mouse x move
            if( m_iZeroX == 0 )
			{
                m_iZeroX = -(int)aData[ 0 ].dwData;
            }
			m_iX = m_iZeroX + aData[ 0 ].dwData;
			pkNewEvent = new NeoEngine::InputEvent( NeoEngine::IE_MOUSEMOVE, m_iX, m_iY, m_iZ, m_iButtonState );
			break;
		}
		case DIMOFS_Y:
		{
			// Mouse y move
            if( m_iZeroY == 0 ) {
                m_iZeroY = -(int)aData[ 0 ].dwData;
            }
			m_iY = m_iZeroY + aData[ 0 ].dwData;
			pkNewEvent = new NeoEngine::InputEvent( NeoEngine::IE_MOUSEMOVE, m_iX, m_iY, m_iZ, m_iButtonState );
			break;
		}
		case DIMOFS_Z:
		{
			// Mouse z move
            if( m_iZeroZ == 0 )
			{
                m_iZeroZ = -(int)aData[ 0 ].dwData;
            }
			m_iZ = m_iZeroZ + aData[ 0 ].dwData;
			pkNewEvent = new NeoEngine::InputEvent( NeoEngine::IE_MOUSEMOVE, m_iX, m_iY, m_iZ, m_iButtonState );
			break;
		}
		case DIMOFS_BUTTON0:
		case DIMOFS_BUTTON1:
		case DIMOFS_BUTTON2:
		case DIMOFS_BUTTON3:
		case DIMOFS_BUTTON4:
		case DIMOFS_BUTTON5:
		case DIMOFS_BUTTON6:
		case DIMOFS_BUTTON7:
		{
			int iButton = aData[ 0 ].dwOfs - DIMOFS_BUTTON0;
			if( BUTTONDOWN( aData[ 0 ] ) )
			{
				m_iButtonState |= (1 << iButton);
				pkNewEvent = new NeoEngine::InputEvent( NeoEngine::IE_MOUSEDOWN, iButton, m_iX, m_iY, m_iZ );
			} else
			{
				m_iButtonState &= ~(1 << iButton);
				pkNewEvent = new NeoEngine::InputEvent( NeoEngine::IE_MOUSEUP, iButton, m_iX, m_iY, m_iZ );
			}
			break;
		}
	}

	*pkEvent = *pkNewEvent;
	delete pkNewEvent, pkNewEvent = 0;
	return true;
}

/*---------------------------------------------------------------------------*/
bool Device::CollectMouse( NeoEngine::InputEvent *pkEvent )
{
    HRESULT hResult;
    NeoEngine::InputEvent *pkNewEvent = 0;

	switch ( m_eMouseCollectState )
	{
		case COLLECT:
		{
			// Need to get new mouse data
	        hResult = m_pkMouse->GetDeviceState( sizeof( m_kMouseData ), (LPVOID)&m_kMouseData );
		    if( ( hResult == DIERR_NOTACQUIRED ) || ( hResult == DIERR_INPUTLOST ) )
			{
				ZeroMemory( &m_kMouseData, sizeof( m_kMouseData ) );
	            NeoEngine::neolog << NeoEngine::LogLevel( NeoEngine::INFO ) << "**** Attempt to reaquire device" << std::endl;
		        hResult = m_pkMouse->Acquire();
			    return false;
	        }
			break;
		}
		case CHECKAXIS:
		{
			if( ( m_iZeroX == 0 ) && ( m_iZeroY == 0 ) && ( m_iZeroZ == 0 ) )
			{
                m_iZeroX = -m_kMouseData.lX;
                m_iZeroY = -m_kMouseData.lY;
				m_iZeroZ = -m_kMouseData.lZ;
            }
			m_iX = m_iZeroX + m_kMouseData.lX;
			m_iY = m_iZeroY + m_kMouseData.lY;
			m_iZ = m_iZeroZ + m_kMouseData.lZ;
			if ( ( m_iX != 0 ) || ( m_iY != 0 ) || ( m_iZ != 0 ) )
			{
				pkNewEvent = new NeoEngine::InputEvent( NeoEngine::IE_MOUSEMOVE, m_iX, m_iY, m_iZ, m_iButtonState );
			}
			break;
		}
		case CHECKBUTTON1:
		case CHECKBUTTON2:
		case CHECKBUTTON3:
		case CHECKBUTTON4:
		case CHECKBUTTON5:
		case CHECKBUTTON6:
		case CHECKBUTTON7:
		case CHECKBUTTON8:
		{
			int iButton = m_eMouseCollectState - CHECKBUTTON1;
			if( m_kMouseData.rgbButtons[ iButton ] != 0 )
			{
				pkNewEvent = new NeoEngine::InputEvent( NeoEngine::IE_MOUSEDOWN, iButton, m_iX, m_iY, m_iZ );
				m_iButtonState |= (1 << iButton);
			} else if( m_iButtonState & (1 << iButton) )
			{
					pkNewEvent = new NeoEngine::InputEvent( NeoEngine::IE_MOUSEUP, iButton, m_iX, m_iY, m_iZ );
					m_iButtonState &= ~(1 << iButton);
			}
			break;
		}
	}

	// send event if we have one
	if( pkNewEvent != 0 )
	{
		*pkEvent = *pkNewEvent;
	}

	if ( m_eMouseCollectState == CHECKBUTTON8 )
	{
		m_eMouseCollectState = COLLECT;
		return false;
	} else
	{
		m_eMouseCollectState++;
		return true;
	}
}

/*---------------------------------------------------------------------------*/
bool Device::CollectBufferedKeyboard( NeoEngine::InputEvent *pkEvent )
{
    HRESULT hResult;
    NeoEngine::InputEvent *pkNewEvent = 0;
	unsigned long ulSize = 1;
	DIDEVICEOBJECTDATA aData[ 1 ];

	hResult = m_pkKeyboard->GetDeviceData( sizeof( DIDEVICEOBJECTDATA ), aData, &ulSize, 0 );
	if( ( hResult == DIERR_NOTACQUIRED ) || ( hResult == DIERR_INPUTLOST ) )
	{
		NeoEngine::neolog << NeoEngine::LogLevel( NeoEngine::INFO ) << "**** Attempt to reaquire device" << std::endl;
		hResult = m_pkKeyboard->Acquire();
		return false;
	}

	if( ulSize == 0 )
	{
		return false;
	}

	pkNewEvent = CreateEventByKeyCode( aData[ 0 ].dwOfs, ( aData[ 0 ].dwData > 0 ? NeoEngine::IE_KEYDOWN : NeoEngine::IE_KEYUP ) );
	*pkEvent = *pkNewEvent;
	return true;
}

/*---------------------------------------------------------------------------*/
bool Device::CollectKeyboard( NeoEngine::InputEvent *pkEvent )
{
	HRESULT hResult;
	NeoEngine::InputEvent *pkNewEvent = 0;

	// Need to get new keyboard data
	if( m_iKeyIndex < 0 )
	{
		hResult = m_pkKeyboard->GetDeviceState( sizeof( m_aKeyData ), (LPVOID)&m_aKeyData );
		if( ( hResult == DIERR_NOTACQUIRED ) || ( hResult == DIERR_INPUTLOST ) )
		{
			m_iKeyIndex = -1;
			NeoEngine::neolog << NeoEngine::LogLevel( NeoEngine::INFO ) << "**** Attempt to reaquire device" << std::endl;
			hResult = m_pkKeyboard->Acquire();
			return false;
		}
	}

    while( !pkNewEvent )
	{
        // Select key index
        m_iKeyIndex++;
       
        // Keys to query left?
        if( m_iKeyIndex > sizeof( m_aKeyData ) )
		{
            m_iKeyIndex = -1;
            return false;
        }

        // Is current key down?
        if( m_aKeyData[ m_iKeyIndex ] & 0x80 )
		{
            // Create event
			pkNewEvent = CreateEventByKeyCode( m_iKeyIndex, NeoEngine::IE_KEYDOWN );
        }
    }
    
    *pkEvent = *pkNewEvent;
    return true;
}

/*---------------------------------------------------------------------------*/
NeoEngine::InputEvent* Device::CreateEventByKeyCode( int iCode, int iType )
{
	NeoEngine::InputEvent *pkEvent = 0;

	switch ( iCode )
	{
		case DIK_0:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_0, NeoEngine::KC_0 );
			break;
		case DIK_1:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_1, NeoEngine::KC_1 );
			break;
		case DIK_2:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_2, NeoEngine::KC_2 );
			break;
		case DIK_3:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_3, NeoEngine::KC_3 );
			break;
		case DIK_4:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_4, NeoEngine::KC_4 );
			break;
		case DIK_5:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_5, NeoEngine::KC_5 );
			break;
		case DIK_6:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_6, NeoEngine::KC_6 );
			break;
		case DIK_7:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_7, NeoEngine::KC_7 );
			break;
		case DIK_8:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_8, NeoEngine::KC_8 );
			break;
		case DIK_9:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_9, NeoEngine::KC_9 );
			break;
		case DIK_A:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_A, NeoEngine::KC_A );
			break;
		case DIK_ABNT_C1:
			break;
		case DIK_ABNT_C2:
			break;
		case DIK_ADD:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_KP_PLUS );
			break;
		case DIK_APOSTROPHE:
			break;
		case DIK_APPS:
			break;
		case DIK_AT:
			// Japanese
			break;
		case DIK_AX:
			break;
		case DIK_B:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_B, NeoEngine::KC_B );
			break;
		case DIK_BACK:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_BACKSPACE, 8 );
			break;
		case DIK_BACKSLASH:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_BACKSLASH, '\\' );
			break;
		case DIK_C:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_C, NeoEngine::KC_C );
			break;
		case DIK_CALCULATOR:
			break;
		case DIK_CAPITAL:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_CAPSLOCK );
			break;
		case DIK_COLON:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_COLON, ':' );
			break;
		case DIK_COMMA:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_COMMA, ',' );
			break;
		case DIK_CONVERT:
			// Japanese
			break;
		case DIK_D:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_D, NeoEngine::KC_D );
			break;
		case DIK_DECIMAL:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_KP_DECIMAL );
			break;
		case DIK_DELETE:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_DELETE );
			break;
		case DIK_DIVIDE:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_KP_DIVIDE );
			break;
		case DIK_DOWN:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_DOWN );
			break;
		case DIK_E:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_E, NeoEngine::KC_E );
			break;
		case DIK_END:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_END );
			break;
		case DIK_EQUALS:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_EQUAL, '=' );
			break;
		case DIK_ESCAPE:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_ESCAPE, 27 );
			break;
		case DIK_F:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_F, NeoEngine::KC_F );
			break;
		case DIK_F1:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_F1 );
			break;
		case DIK_F2:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_F2 );
			break;
		case DIK_F3:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_F3 );
			break;
		case DIK_F4:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_F4 );
			break;
		case DIK_F5:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_F5 );
			break;
		case DIK_F6:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_F6 );
			break;
		case DIK_F7:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_F7 );
			break;
		case DIK_F8:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_F8 );
			break;
		case DIK_F9:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_F9 );
			break;
		case DIK_F10:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_F10 );
			break;
		case DIK_F11:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_F11 );
			break;
		case DIK_F12:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_F12 );
			break;
		case DIK_F13:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_F13 );
			break;
		case DIK_F14:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_F14 );
			break;
		case DIK_F15:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_F15 );
			break;
		case DIK_G:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_G, NeoEngine::KC_G );
			break;
		case DIK_GRAVE:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_GRAVE, '^' );
			break;
		case DIK_H:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_H, NeoEngine::KC_H );
			break;
		case DIK_HOME:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_HOME );
			break;
		case DIK_I:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_I, NeoEngine::KC_I );
			break;
		case DIK_INSERT:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_INSERT );
			break;
		case DIK_J:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_J, NeoEngine::KC_J );
			break;
		case DIK_K:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_K, NeoEngine::KC_K );
			break;
		case DIK_KANA:
			// Japanese
			break;
		case DIK_KANJI:
			// Japanese
			break;
		case DIK_L:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_L, NeoEngine::KC_L );
			break;
		case DIK_LBRACKET:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_LEFTBRACKET, '[' );
			break;
		case DIK_LCONTROL:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_LCTRL );
			break;
		case DIK_LEFT:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_LEFT );
			break;
		case DIK_LMENU:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_LALT );
			break;
		case DIK_LSHIFT:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_LSHIFT );
			break;
		case DIK_LWIN:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_LWIN );
			break;
		case DIK_M:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_M, NeoEngine::KC_M );
			break;
		case DIK_MAIL:
			break;
		case DIK_MEDIASELECT:
			break;
		case DIK_MEDIASTOP:
			break;
		case DIK_MINUS:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_MINUS, '-' );
			break;
		case DIK_MULTIPLY:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_KP_ASTERISK );
			break;
		case DIK_MUTE:
			break;
		case DIK_MYCOMPUTER:
			break;
		case DIK_N:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_N, NeoEngine::KC_N );
			break;
		case DIK_NEXT:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_PAGEDOWN );
			break;
		case DIK_NEXTTRACK:
			break;
		case DIK_NOCONVERT:
			// Japanese
			break;
		case DIK_NUMLOCK:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_KP_NUMLOCK );
		break;
		case DIK_NUMPAD0:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_KP_0 );
			break;
		case DIK_NUMPAD1:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_KP_1 );
			break;
		case DIK_NUMPAD2:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_KP_2 );
			break;
		case DIK_NUMPAD3:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_KP_3 );
			break;
		case DIK_NUMPAD4:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_KP_4 );
			break;
		case DIK_NUMPAD5:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_KP_5 );
			break;
		case DIK_NUMPAD6:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_KP_6 );
			break;
		case DIK_NUMPAD7:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_KP_7 );
			break;
		case DIK_NUMPAD8:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_KP_8 );
			break;
		case DIK_NUMPAD9:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_KP_9 );
			break;
		case DIK_NUMPADCOMMA:
			// Japanese
			break;
		case DIK_NUMPADENTER:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_KP_ENTER );
			break;
		case DIK_NUMPADEQUALS:
			// Japanese
			break;
		case DIK_O:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_O, NeoEngine::KC_O );
			break;
		case DIK_OEM_102:
			// British and German
			break;
		case DIK_P:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_P, NeoEngine::KC_P );
			break;
		case DIK_PAUSE:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_PAUSE );
			break;
		case DIK_PERIOD:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_DOT, '.' );
			break;
		case DIK_PLAYPAUSE:
			break;
		case DIK_POWER:
			break;
		case DIK_PREVTRACK:
			break;
		case DIK_PRIOR:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_PAGEUP );
			break;
		case DIK_Q:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_Q, NeoEngine::KC_Q );
			break;
		case DIK_R:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_R, NeoEngine::KC_R );
			break;
		case DIK_RBRACKET:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_RIGHTBRACKET, ']' );
			break;
		case DIK_RCONTROL:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_RCTRL );
			break;
		case DIK_RETURN:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_RETURN, 13 );
			break;
		case DIK_RIGHT:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_RIGHT );
			break;
		case DIK_RMENU:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_RALT );
			break;
		case DIK_RSHIFT:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_RSHIFT );
			break;
		case DIK_RWIN:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_RWIN );
			break;
		case DIK_S:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_S, NeoEngine::KC_S );
			break;
		case DIK_SCROLL:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_SCROLLLOCK );
			break;
		case DIK_SEMICOLON:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_SEMICOLON, ';' );
			break;
		case DIK_SLASH:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_SLASH, '/' );
			break;
		case DIK_SLEEP:
			break;
		case DIK_SPACE:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_SPACE, NeoEngine::KC_SPACE );
			break;
		case DIK_STOP:
			// Japanese
			break;
		case DIK_SUBTRACT:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_KP_MINUS );
			break;
		case DIK_SYSRQ:
			break;
		case DIK_T:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_T, NeoEngine::KC_T );
			break;
		case DIK_TAB:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_TAB );
			break;
		case DIK_U:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_U, NeoEngine::KC_U );
			break;
		case DIK_UNDERLINE:
			// Japanese
			break;
		case DIK_UNLABELED:
			// Japanese
			break;
		case DIK_UP:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_UP );
			break;
		case DIK_V:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_V, NeoEngine::KC_V );
			break;
		case DIK_VOLUMEDOWN:
			break;
		case DIK_VOLUMEUP:
			break;
		case DIK_W:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_W, NeoEngine::KC_W );
			break;
		case DIK_WAKE:
			break;
		case DIK_WEBBACK:
			break;
		case DIK_WEBFAVORITES:
			break;
		case DIK_WEBFORWARD:
			break;
		case DIK_WEBHOME:
			break;
		case DIK_WEBREFRESH:
			break;
		case DIK_WEBSEARCH:
			break;
		case DIK_WEBSTOP:
			break;
		case DIK_X:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_X, NeoEngine::KC_X );
			break;
		case DIK_Y:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_Y, NeoEngine::KC_Y );
			break;
		case DIK_YEN:
			// Japanese
			break;
		case DIK_Z:
			pkEvent = new NeoEngine::InputEvent( iType, NeoEngine::KC_Z, NeoEngine::KC_Z );
			break;
	}

	return pkEvent;
}

/*---------------------------------------------------------------------------*/
void Device::SetMouseRange( int left, int right, int top, int bottom )
{
    HRESULT hResult;

    DIPROPRANGE kProperty;
    kProperty.diph.dwSize = sizeof( DIPROP_RANGE );
    kProperty.diph.dwHeaderSize = sizeof( DIPROPHEADER );
    kProperty.diph.dwObj = DIMOFS_X;
    kProperty.diph.dwHow = DIPH_BYOFFSET;
    kProperty.lMin = left;
    kProperty.lMax = right;
    hResult = m_pkMouse->SetProperty( DIPROP_RANGE, &kProperty.diph );
    if( FAILED( hResult ) )
	{
        NeoEngine::neolog << NeoEngine::LogLevel( NeoEngine::ERROR ) << "**** Failed to set mouse x range [" << hResult << "]" << std::endl;
    }
    
    kProperty.diph.dwSize = sizeof( DIPROP_RANGE );
    kProperty.diph.dwHeaderSize = sizeof( DIPROPHEADER );
    kProperty.diph.dwObj = DIMOFS_Y;
    kProperty.diph.dwHow = DIPH_BYOFFSET;
    kProperty.lMin = top;
    kProperty.lMax = bottom;
    hResult = m_pkMouse->SetProperty( DIPROP_RANGE, &kProperty.diph );
    if( FAILED( hResult ) )
	{
        NeoEngine::neolog << NeoEngine::LogLevel( NeoEngine::ERROR ) << "**** Failed to set mouse y range [" << hResult << "]" << std::endl;
    }    
}

/*---------------------------------------------------------------------------*/
bool Device::SetBufferSize( IDirectInputDevice8 *pkDevice, int iSize )
{
	HRESULT hResult;
	DIPROPDWORD kProperty;

	kProperty.diph.dwSize = sizeof( DIPROPDWORD );
	kProperty.diph.dwHeaderSize = sizeof( DIPROPHEADER );
	kProperty.diph.dwHow = DIPH_DEVICE;
	kProperty.diph.dwObj = 0;
	kProperty.dwData = iSize;
	hResult = pkDevice->SetProperty( DIPROP_BUFFERSIZE, &kProperty.diph );
    if( FAILED( hResult ) )
	{
        NeoEngine::neolog << NeoEngine::LogLevel( NeoEngine::ERROR ) << "**** Failed to set buffersize [" << hResult << "]" << std::endl;
		return false;
    }    

	return true;
}

/*---------------------------------------------------------------------------*/
void Device::ToggleKeyboardInputMethod()
{
	m_bKeyboardBuffered = !m_bKeyboardBuffered;
	m_iKeyIndex = -1;
}

/*---------------------------------------------------------------------------*/
void Device::ToggleMouseInputMethod()
{
	m_bMouseBuffered = !m_bMouseBuffered;
}

}; /* namespace */
