/***************************************************************************
                     sound.cpp  -  Sound classes and codecs
                             -------------------
    begin                : Wed Mar 17 2004
    copyright            : (C) 2004 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, sound.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "sound.h"

#include <neoengine/file.h>
#include <neoengine/filemanager.h>
#include <neoengine/logstream.h>

using namespace NeoEngine;
using namespace std;


namespace NeoDSound
{


NeoEngine::Sound *AudioDevice::LoadSound( const std::string &rstrFilename )
{
	vector< SoundCodec* >::iterator ppkCodec    = m_vpkCodecs.begin();
	vector< SoundCodec* >::iterator ppkCodecEnd = m_vpkCodecs.end();

	for( ; ppkCodec != ppkCodecEnd; ++ppkCodec )
	{
		string strFilename = rstrFilename + ".";

		const vector<string> &rvstrExtensions = (*ppkCodec)->GetExtensions();

		vector<string>::const_iterator pstrExt    = rvstrExtensions.begin();
		vector<string>::const_iterator pstrExtEnd = rvstrExtensions.end();

		for( ; pstrExt != pstrExtEnd; ++pstrExt )
		{
			File        *pkFile   = m_pkFileManager->GetByName( strFilename + (*pstrExt) );
			SoundStream *pkStream = 0;

			if( pkFile && pkFile->Open() && (*ppkCodec)->IsType( pkFile ) && ( pkStream = (*ppkCodec)->GetStream( pkFile ) ) )
				return new Sound( this, rstrFilename, pkStream );
			else
				delete pkFile;
		}
	}

	neolog << LogLevel( WARNING ) << "*** unable to load sound [" << rstrFilename << "]" << endl;

	return 0;
}



Sound::Sound( AudioDevice *pkDevice, const std::string &rstrName, SoundStream *pkStream ) :
	NeoEngine::Sound( rstrName, pkStream ),
	m_pkDevice( pkDevice ),
	m_pkBuffer( 0 ),
	m_bStreaming( false ),
	m_hNotifyEvent( 0 ),
	m_pkThread( 0 ),
	m_uiBufferSize( 0 ),
	m_uiLastPlayPos( 0 ),
	m_uiPlayProgress( 0 ),
	m_uiNotifySize( 0 ),
	m_uiNextWriteOffset( 0 ),
	m_bRefill( false )
{
	//Stream sounds > 256k
	if( m_pkStream->GetSize() > ( 256 * 1024 ) )
	{
		m_bStreaming = true;
		m_hNotifyEvent = CreateEvent( 0, FALSE, FALSE, 0 );
		m_pkThread = new Thread( this, 0 );
	}
}


Sound::~Sound()
{
	if( m_pkThread )
	{
		m_pkThread->m_bTerminate = true;

		SetEvent( m_hNotifyEvent );

		delete m_pkThread;

		CloseHandle( m_hNotifyEvent );
	}

	if( m_pkBuffer )
	{
		m_pkBuffer->Stop();
		m_pkBuffer->Release();
	}
}

#define NUM_NOTIFICATIONS 16

void Sound::Play( bool bLoop )
{
	HRESULT res;

	m_bLoop = bLoop;

	if( IsPlaying() )
		return;

	if( !m_pkBuffer )
	{
		m_uiBufferSize = m_pkStream->GetSize();

		WAVEFORMATEX WaveFormat;
		DSBUFFERDESC desc;

		WaveFormat.wFormatTag      = 1;
		WaveFormat.nChannels       = m_pkStream->GetNumChannels();
		WaveFormat.nSamplesPerSec  = m_pkStream->GetSamplesPerSecond();
		WaveFormat.wBitsPerSample  = m_pkStream->GetBitsPerSample();
		WaveFormat.nBlockAlign     = WaveFormat.nChannels * ( WaveFormat.wBitsPerSample / 8 );
		WaveFormat.nAvgBytesPerSec = WaveFormat.nSamplesPerSec * WaveFormat.nBlockAlign;
		WaveFormat.cbSize          = 0;
	
		//Setup a three-second circular buffer for streaming
		if( m_bStreaming )
		{
			m_uiBufferSize  = m_pkStream->GetNumChannels() * m_pkStream->GetSamplesPerSecond() * ( m_pkStream->GetBitsPerSample() / 8 ) * 3;
			m_uiNotifySize  = m_uiBufferSize / NUM_NOTIFICATIONS;
			m_uiNotifySize -= m_uiNotifySize % WaveFormat.nBlockAlign;
			m_uiBufferSize  = m_uiNotifySize * NUM_NOTIFICATIONS;
		}

		memset( &desc, 0, sizeof( DSBUFFERDESC ) );
		desc.dwSize        = sizeof( DSBUFFERDESC );
		desc.dwFlags       = DSBCAPS_GETCURRENTPOSITION2 | DSBCAPS_CTRLPOSITIONNOTIFY /*| DSBCAPS_CTRLPAN | DSBCAPS_CTRLFREQUENCY | DSBCAPS_CTRLVOLUME*/ | DSBCAPS_GLOBALFOCUS;
		desc.lpwfxFormat   = &WaveFormat;
		desc.dwBufferBytes = m_uiBufferSize;

		IDirectSoundBuffer *pkBuffer = 0;

		if( FAILED( ( res = m_pkDevice->m_pkDS->CreateSoundBuffer( &desc, &pkBuffer, 0 ) ) ) )
		{
			neolog << LogLevel( WARNING ) << "*** unable to play sound: failed to create buffer" << endl;
			return;
		}

		if( FAILED( ( res = pkBuffer->QueryInterface( IID_IDirectSoundBuffer8, (void**)&m_pkBuffer ) ) ) )
		{
			neolog << LogLevel( WARNING ) << "*** unable to play sound: failed to query buffer interface" << endl;
			return;
		}

		void  *pData   = 0;
		DWORD  dwBytes = m_uiBufferSize;
		bool bSilence = false;

		if( FAILED( ( res = m_pkBuffer->Lock( 0, 0, &pData, &dwBytes, 0, 0, DSBLOCK_ENTIREBUFFER ) ) ) )
		{
			neolog << LogLevel( WARNING ) << "*** unable to play sound: failed to lock buffer" << endl;
			return;
		}

		FillBuffer( (unsigned char*)pData, dwBytes, &bSilence );

		if( FAILED( ( res = m_pkBuffer->Unlock( pData, dwBytes, 0, 0 ) ) ) )
		{
			neolog << LogLevel( WARNING ) << "*** unable to play sound: failed to unlock buffer" << endl;
			return;
		}

		if( m_bStreaming )
		{
			IDirectSoundNotify *pkNotify   = 0;
			DSBPOSITIONNOTIFY*  aPosNotify = 0; 

			if( FAILED( res = pkBuffer->QueryInterface( IID_IDirectSoundNotify, (void**)&pkNotify ) ) )
			{
				neolog << LogLevel( WARNING ) << "*** unable to play sound: failed to unlock buffer" << endl;
				return;
			}

			aPosNotify = new DSBPOSITIONNOTIFY[ NUM_NOTIFICATIONS ];
			
			for( unsigned int i = 0; i < NUM_NOTIFICATIONS; ++i )
			{
				aPosNotify[i].dwOffset     = ( m_uiNotifySize * i ) + m_uiNotifySize - 1;
				aPosNotify[i].hEventNotify = m_hNotifyEvent;
			}
		    
			if( FAILED( res = pkNotify->SetNotificationPositions( NUM_NOTIFICATIONS, aPosNotify ) ) )
			{
				neolog << LogLevel( WARNING ) << "*** unable to play sound: failed to set stream buffer notification" << endl;
				return;
			}

			delete [] aPosNotify;
			pkNotify->Release();
		}
	}

	m_pkBuffer->Play( 0, 0, m_bLoop ? DSBPLAY_LOOPING : 0 );

	//If not streaming, we don't keep track of the playing status
	if( m_bStreaming )
		m_bIsPlaying = true;
}


void Sound::Pause()
{
	if( m_pkBuffer )
		m_pkBuffer->Stop();

	m_bIsPlaying = false;
}


void Sound::Stop()
{
	if( m_pkBuffer )
	{
		m_pkBuffer->Stop();
		m_pkBuffer->SetCurrentPosition( 0 );

		if( m_bStreaming )
		{
			m_pkStream->Reset();

			m_bRefill = true;
			SetEvent( m_hNotifyEvent );
		}
	}

	m_bIsPlaying = false;
}


bool Sound::Restore( bool *pbRestored )
{
	HRESULT res;
	DWORD dwStatus = 0;

	*pbRestored = false;

    if( FAILED( res = m_pkBuffer->GetStatus( &dwStatus ) ) )
		return false;

	if( dwStatus & DSBSTATUS_BUFFERLOST )
	{
		// Since the app could have just been activated, then
		// DirectSound may not be giving us control yet, so 
		// the restoring the buffer may fail.  
		// If it does, sleep until DirectSound gives us control.
		do 
		{
			res = m_pkBuffer->Restore();
			if( res == DSERR_BUFFERLOST )
				Sleep( 10 );
		}
		while( ( res = m_pkBuffer->Restore() ) == DSERR_BUFFERLOST );

		*pbRestored = true;
	}

	return true;
}


void *Sound::ThreadEntry( NeoEngine::Thread *pkThread, void *pData )
{
	pkThread->m_bStarted = true;

	HRESULT res;
	bool bSilence = false;

    while( !pkThread->m_bTerminate ) 
    { 
        unsigned int uiResult = MsgWaitForMultipleObjects( 1, &m_hNotifyEvent, FALSE, INFINITE, QS_ALLEVENTS );
        
		switch( uiResult )
        {
            case WAIT_OBJECT_0 + 0:
			{
				if( pkThread->m_bTerminate )
					break;

				bool bRestored = false;

				if( !Restore( &bRestored ) )
				{
					neolog << LogLevel( WARNING ) << "*** unable to play streaming sound: buffer lost and not restorable" << endl;
					pkThread->m_bTerminate = true;
					break;
				}

				if( bRestored || m_bRefill )
				{
					if( bRestored )
						neolog << LogLevel( INFO ) << "streaming sound buffer lost and restored" << endl;

					//Stream in the full buffer
					void *pData = 0;
					DWORD dwBytes = m_uiBufferSize;

					if( FAILED( ( res = m_pkBuffer->Lock( 0, 0, &pData, &dwBytes, 0, 0, DSBLOCK_ENTIREBUFFER ) ) ) )
					{
						neolog << LogLevel( WARNING ) << "*** unable to play streaming sound: failed to lock buffer after restore" << endl;
						pkThread->m_bTerminate = true;
						break;
					}

					FillBuffer( (unsigned char*)pData, dwBytes, &bSilence );

					if( FAILED( ( res = m_pkBuffer->Unlock( pData, dwBytes, 0, 0 ) ) ) )
					{
						neolog << LogLevel( WARNING ) << "*** unable to play streaming sound: failed to unlock buffer after restore" << endl;
						pkThread->m_bTerminate = true;
						break;
					}

					m_bRefill = false;
					m_uiNextWriteOffset = 0;

					break;
				}

				void* pBuffer  = 0;
				void* pBuffer2 = 0;
				DWORD dwSize   = 0;
				DWORD dwSize2  = 0;
				DWORD dwThisSize = 0;

				if( FAILED( res = m_pkBuffer->Lock( m_uiNextWriteOffset, m_uiNotifySize, &pBuffer, &dwSize, &pBuffer2, &dwSize2, 0L ) ) )
				{
					neolog << LogLevel( WARNING ) << "*** unable to play streaming sound: failed to lock buffer segment [" << m_uiNotifySize << " @ " << m_uiNextWriteOffset << " of " << m_uiBufferSize << "]" << endl;
					pkThread->m_bTerminate = true;
					break;
				}

				dwThisSize = dwSize + dwSize2;

				while( pBuffer )
				{
					FillBuffer( (unsigned char*)pBuffer, dwSize, &bSilence );

					if( FAILED( res = m_pkBuffer->Unlock( pBuffer, dwSize, 0, 0 ) ) )
					{
						neolog << LogLevel( WARNING ) << "*** unable to play streaming sound: failed to unlock buffer segment" << endl;
						pkThread->m_bTerminate = true;
						break;
					}

					pBuffer  = pBuffer2;
					dwSize   = dwSize2;
					pBuffer2 = 0;
				}

				DWORD uiCurrentPlayPos, uiPlayDelta;

				if( FAILED( res = m_pkBuffer->GetCurrentPosition( &uiCurrentPlayPos, 0 ) ) )
				{
					neolog << LogLevel( WARNING ) << "*** unable to play streaming sound: failed to get current play position" << endl;
					pkThread->m_bTerminate = true;
					break;
				}

				// Check to see if the position counter looped
				if( uiCurrentPlayPos < m_uiLastPlayPos )
					uiPlayDelta = ( m_uiBufferSize - m_uiLastPlayPos ) + uiCurrentPlayPos;
				else
					uiPlayDelta = uiCurrentPlayPos - m_uiLastPlayPos;

				m_uiPlayProgress += uiPlayDelta;
				m_uiLastPlayPos   = uiCurrentPlayPos;

				if( !m_bLoop && ( m_uiPlayProgress >= m_pkStream->GetSize() ) )
				{
					pkThread->m_bTerminate = true;
					m_pkBuffer->Stop();
					m_bIsPlaying = false;
				}

				m_uiNextWriteOffset += dwThisSize; 
				m_uiNextWriteOffset %= m_uiBufferSize;

                break;
			}
        }
    }

	pkThread->m_bTerminated = true;

	return 0;
}


void Sound::FillBuffer( unsigned char *pucData, unsigned int uiBytes, bool *pbSilence )
{
	if( *pbSilence )
		memset( pucData, m_pkStream->GetBitsPerSample() == 8 ? 128 : 0, uiBytes );
	else
	{
		unsigned int uiDecoded = m_pkStream->Decode( pucData, uiBytes );

		while( uiDecoded < uiBytes )
		{
			if( !m_bLoop )
			{
				memset( pucData, m_pkStream->GetBitsPerSample() == 8 ? 128 : 0, uiBytes - uiDecoded );
				*pbSilence = true;
				uiDecoded = uiBytes;
			}
			else
			{
				uiBytes -= uiDecoded;

				m_pkStream->Reset();

				uiDecoded = m_pkStream->Decode( pucData + uiDecoded, uiBytes );
			}
		}
	}
}


};


