/***************************************************************************
                     intialize.cpp  -  Initialize OpenGL
                             -------------------
    begin                : Tue Apr 1 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoDevOpenGL, initialize.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#define NEEDGDI

#include "device.h"
#include "pixelbuffer.h"
#include "pixelbuffertarget.h"
#include "framebuffertarget.h"
#include "textureunit.h"
#include "extensions.h"
#include "vertexbuffermanager-nobs.h"
#include "vertexbuffermanager-ati.h"
#include "vertexbuffermanager-arb.h"
#include "shader.h"

#include <neoengine/nemath.h>
#include <neoengine/config.h>
#include <neoengine/logstream.h>

#ifdef __APPLE__
#  include <mach-o/dyld.h>
#endif

using namespace std;
using namespace NeoEngine;


namespace NeoOGL
{


bool Device::Initialize( const RenderCaps *pkCaps )
{
	bool bDisableVBO = false;

	//Set active render target
	SetRenderTarget( FRAMEBUFFER );

	if( !m_pkCurTarget || !m_pkCurTarget->IsActive() )
	{
		neolog << LogLevel( ERROR ) << "*** unable to initialize render device: unable to activate frame buffer render target" << endl;
		return false;
	}

	m_kExtensions.ResetAll();

	string strVendor = (const char*)glGetString( GL_VENDOR );
	string strVersion = (const char*)glGetString( GL_VERSION );
	string strExt = (const char*)glGetString( GL_EXTENSIONS );

	neolog << LogLevel( INFO ) << "Opened device @ <" << GetWidth() << "x" << GetHeight() << "x" << GetBPP() << "> "
	       << ( m_kWindow.m_kCaps.IsSet( RenderCaps::FULLSCREEN ) ? "fullscreen" : "windowed mode" ) << endl
	       << "Device Information:" << endl
	       << "--------------------------" << endl
	       << "Vendor: " << strVendor << endl
	       << "Renderer: " << glGetString( GL_RENDERER ) << endl
	       << "Version: " << strVersion << endl
	       << "Extensions: " << strExt << endl;

	//Get data
	glGetIntegerv( GL_MAX_TEXTURE_SIZE, (GLint*)&m_uiMaxTextureSize );
	neolog << "Max texture size: " << m_uiMaxTextureSize << endl;

#if defined( LINUX )

	// Selectively disable ARB vertex buffer objects on broken ATI systems
	if( strVendor == "ATI Technologies Inc." && strVersion == "1.3 (X4.3.0-3.7.6)" )
		bDisableVBO = true;

#endif

#if defined( WIN32 )

	if( strExt.find( "WGL_EXT_extensions_string" ) < string::npos )
	{
		neolog << LogLevel( INFO ) << "WGL_EXT_extensions_string supported" << endl;

		//Get WGL extension string proc
		if( !( fpneowglGetExtensionsStringEXT          = (fpwglGetExtensionsStringEXTProc)GetProcAddress( "wglGetExtensionsStringEXT" ) ) )
			neolog << LogLevel( ERROR ) << "WGL_EXT_extensions_string supported, but unable to get \"wglGetExtensionsStringEXT\" proc address" << endl;
	}
	else if( ( fpneowglGetExtensionsStringEXT          = (fpwglGetExtensionsStringEXTProc)GetProcAddress( "wglGetExtensionsStringEXT" ) ) )
		neolog << LogLevel( INFO ) << "WGL_EXT_extensions_string NOT supported, but able to get \"wglGetExtensionsStringEXT\" proc address anyway" << endl;

	if( fpneowglGetExtensionsStringEXT )
	{
		string strWGLExt = neowglGetExtensionsStringEXT();

		neolog << LogLevel( INFO ) << "WGL Extensions: " << strWGLExt << endl;

		strExt += strWGLExt;
	}

/*
	if( strExt.find( "WGL_ARB_pbuffer" ) < strExt.length() )
	{
		fpneowglCreatePbufferARB     = (fpwglCreatePbufferARBProc   )GetProcAddress( "wglCreatePbufferARB" );
		fpneowglGetPbufferDCARB      = (fpwglGetPbufferDCARBProc    )GetProcAddress( "wglGetPbufferDCARB" );
		fpneowglReleasePbufferDCARB  = (fpwglReleasePbufferDCARBProc)GetProcAddress( "wglReleasePbufferDCARB" );
		fpneowglDestroyPbufferARB    = (fpwglDestroyPbufferARBProc  )GetProcAddress( "wglDestroyPbufferARB" );
		fpneowglQueryPbufferARB      = (fpwglQueryPbufferARBProc    )GetProcAddress( "wglQueryPbufferARB" );

		m_kWindow.m_kCaps.Set( RenderCaps::PIXELBUFFER );

		m_pkPixelBufferTarget = new PixelBufferRenderTarget( this );
	}

	if( strExt.find( "WGL_ARB_render_texture" ) < strExt.length() )
	{
		fpneowglBindTexImageARB       = (fpwglBindTexImageARBProc    )GetProcAddress( "wglBindTexImageARB" );
		fpneowglReleaseTexImageARB    = (fpwglReleaseTexImageARBProc )GetProcAddress( "wglReleaseTexImageARB" );
		fpneowglSetPbufferAttribARB   = (fpwglSetPbufferAttribARBProc)GetProcAddress( "wglSetPbufferAttribARB" );

		m_kWindow.m_kCaps.Set( RenderCaps::RENDERTEXTURE );
	}
*/
#endif

	m_kWindow.m_kCaps.Set( RenderCaps::PIXELBUFFER );
	m_kWindow.m_kCaps.Set( RenderCaps::RENDERTEXTURE );

	m_pkPixelBufferTarget = new PixelBufferRenderTarget( this );


	//Look for fog coordinate support
	if( strExt.find( "GL_EXT_fog_coord" ) < std::string::npos )
	{
		neolog << LogLevel( INFO ) << "GL_EXT_fog_coord supported" << endl;

		m_kExtensions.Set( Extensions::EXT_FOG_COORD );

		fpneoglFogCoordfEXT       = (fpglFogCoordfEXTProc)      GetProcAddress( "glFogCoordfEXT" );
		fpneoglFogCoorddEXT       = (fpglFogCoorddEXTProc)      GetProcAddress( "glFogCoorddEXT" );
		fpneoglFogCoordfvEXT      = (fpglFogCoordfvEXTProc)     GetProcAddress( "glFogCoordfvEXT" );
		fpneoglFogCoorddvEXT      = (fpglFogCoorddvEXTProc)     GetProcAddress( "glFogCoorddvEXT" );
		fpneoglFogCoordPointerEXT = (fpglFogCoordPointerEXTProc)GetProcAddress( "glFogCoordPointerEXT" );

		m_kWindow.m_kCaps.Set( RenderCaps::FOGCOORD );

	}

	//Look for transpose matrix
	if( strExt.find( "GL_ARB_transpose_matrix" ) < std::string::npos )
	{
		neolog << LogLevel( INFO ) << "GL_ARB_transpose_matrix supported" << endl;

		m_kExtensions.Set( Extensions::ARB_TRANSPOSE_MATRIX );

		fpneoglLoadMatrixfARB = (fpglLoadMatrixfARBProc)GetProcAddress( "glLoadTransposeMatrixfARB" );
		fpneoglMultMatrixfARB = (fpglMultMatrixfARBProc)GetProcAddress( "glMultTransposeMatrixfARB" );
	}

	//Look for cubemap support
	if( ( strExt.find( "GL_ARB_texture_cube_map" ) < std::string::npos ) || ( strExt.find( "GL_EXT_texture_cube_map" ) < std::string::npos ) )
	{
		//Extensions ARB and EXT are identical
		neolog << LogLevel( INFO ) << "GL_ARB_texture_cube_map supported" << endl;

		m_kExtensions.Set( Extensions::ARB_TEXTURE_CUBE_MAP );

		m_kWindow.m_kCaps.Set( RenderCaps::CUBEMAP );
	}

	//Look for multitexturing
	if( strExt.find( "GL_ARB_multitexture" ) < strExt.length() )
	{
		//multitexture is supported
		glGetIntegerv( GL_MAX_TEXTURE_UNITS_ARB, (GLint*)&m_uiNumTMUs );

#if !defined(__APPLE__)

		fpneoglActiveTextureARB       = (fpglActiveTextureARBProc      )GetProcAddress( "glActiveTextureARB" );
		fpneoglClientActiveTextureARB = (fpglClientActiveTextureARBProc)GetProcAddress( "glClientActiveTextureARB" );

#else
		//Symbols defined as default on OSX
#endif
		
		neolog << LogLevel( INFO ) << "Texture units: " << m_uiNumTMUs << " [arb]" << endl;
		m_kWindow.m_kCaps.Set( RenderCaps::MULTITEXTURE );
	}
	else
	{
		neolog << LogLevel( INFO ) << "No ARB multitexturing" << endl;
		m_uiNumTMUs = 1;
	}

	m_kWindow.m_kCaps.Set( RenderCaps::MULTITEXTURE_GROUP | ( ( m_uiNumTMUs << RenderCaps::MULTITEXTURE_UNITS_SHIFT ) & RenderCaps::MULTITEXTURE_UNITS_MASK ) );

	//Look for ADD texenv support
	if( strExt.find( "GL_ARB_texture_env_add" ) < std::string::npos )
	{
		neolog << LogLevel( INFO ) << "GL_ARB_texture_env_add supported" << endl;
		m_kExtensions.Set( Extensions::ARB_TEXENV_ADD );
	}
	else if( strExt.find( "GL_EXT_texture_env_add" ) < std::string::npos )
	{
		neolog << LogLevel( INFO ) << "GL_EXT_texture_env_add supported" << endl;
		m_kExtensions.Set( Extensions::ARB_TEXENV_ADD );
	}

	//Look for COMBINE texenv support
	if( strExt.find( "GL_ARB_texture_env_combine" ) < std::string::npos )
	{
		neolog << LogLevel( INFO ) << "GL_ARB_texture_env_combine supported" << endl;
		m_kExtensions.Set( Extensions::ARB_TEXENV_COMBINE );
		m_kWindow.m_kCaps.Set( RenderCaps::BLENDMODE_FACTOR );
		m_kWindow.m_kCaps.Set( RenderCaps::BLENDMODE_INTERPOLATE );
	}

	m_ppkTMU = new TextureUnit*[ m_uiNumTMUs ];

	for( unsigned int uiUnit = 0; uiUnit < m_uiNumTMUs; ++uiUnit )
	{
		if( m_uiNumTMUs > 1 )
			TextureUnit::SetActiveTextureUnit( uiUnit );

		glDisableClientState( GL_TEXTURE_COORD_ARRAY );

		if( m_kExtensions.IsSupported( Extensions::ARB_TEXENV_COMBINE ) )
			m_ppkTMU[ uiUnit ] = new TextureUnitCombine( this, uiUnit );
		else
			m_ppkTMU[ uiUnit ] = new TextureUnit( this, uiUnit );
	}

	if( m_uiNumTMUs > 1 )
		TextureUnit::SetActiveTextureUnit( 0 );

	neolog << LogLevel( INFO ) << "  texture units setup" << endl;

	neoglCheckError( "*** Error setting up texture units: " );

	//Check for automatic mipmap generation
	if( strExt.find( "GL_SGIS_generate_mipmap" ) < std::string::npos )
	{
		neolog << LogLevel( INFO ) << "GL_SGIS_generate_mipmap supported" << endl;
		m_kExtensions.Set( Extensions::SGIS_GENERATE_MIPMAP );
		glHint( GL_GENERATE_MIPMAP_HINT, GL_NICEST );
	}

	// Check for anisotropic filtering
	if( strExt.find( "GL_EXT_texture_filter_anisotropic" ) < std::string::npos )
	{
		GLint iMax;

		glGetIntegerv( GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &iMax );

		neolog << LogLevel( INFO ) << "GL_EXT_texture_filter_anisotropic supported (max: " << iMax << ")" << endl;
		m_kExtensions.Set( Extensions::EXT_TEXTURE_FILTER_ANISOTROPIC );
	}

	if( strExt.find( "GL_ARB_vertex_buffer_object" ) < string::npos && !bDisableVBO )
	{
		neolog << LogLevel( INFO ) << "GL_ARB_vertex_buffer_object supported" << endl;

		fpneoglBindBufferARB           = (fpglBindBufferARB          )GetProcAddress( "glBindBufferARB"           );
		fpneoglDeleteBuffersARB        = (fpglDeleteBuffersARB       )GetProcAddress( "glDeleteBuffersARB"        );
		fpneoglGenBuffersARB           = (fpglGenBuffersARB          )GetProcAddress( "glGenBuffersARB"           );
		fpneoglIsBufferARB             = (fpglIsBufferARB            )GetProcAddress( "glIsBufferARB"             );
		fpneoglBufferDataARB           = (fpglBufferDataARB          )GetProcAddress( "glBufferDataARB"           );
		fpneoglBufferSubDataARB        = (fpglBufferSubDataARB       )GetProcAddress( "glBufferSubDataARB"        );
		fpneoglGetBufferSubDataARB     = (fpglGetBufferSubDataARB    )GetProcAddress( "glGetBufferSubDataARB"     );
		fpneoglMapBufferARB            = (fpglMapBufferARB           )GetProcAddress( "glMapBufferARB"            );
		fpneoglUnmapBufferARB          = (fpglUnmapBufferARB         )GetProcAddress( "glUnmapBufferARB"          );
		fpneoglGetBufferParameterivARB = (fpglGetBufferParameterivARB)GetProcAddress( "glGetBufferParameterivARB" );
		fpneoglGetBufferPointervARB    = (fpglGetBufferPointervARB   )GetProcAddress( "glGetBufferPointervARB"    );

		m_kExtensions.Set( Extensions::ARB_VERTEX_BUFFER_OBJECT );

		m_pkVertexBufferManager = new VertexBufferManagerARB( this );

		if( !m_pkVertexBufferManager->SetSize( m_uiVertexStorageSize ) )
			delete m_pkVertexBufferManager, m_pkVertexBufferManager = 0;

		neoglCheckError( "*** Error setting up ARB vertex buffer extension: " );
	}
	else if( strExt.find( "GL_ATI_vertex_array_object" ) < string::npos )
	{
		neolog << LogLevel( INFO ) << "GL_ATI_vertex_array_object supported" << endl;

		if( strExt.find( "GL_ATI_vertex_attrib_array_object" ) == string::npos )
			neolog << LogLevel( INFO ) << "  disabled, GL_ATI_vertex_attrib_array_object unsupported" << endl;
		else
		{
			fpneoglNewObjectBufferATI          = (fpglNewObjectBufferATI         )GetProcAddress( "glNewObjectBufferATI"          );
			fpneoglUpdateObjectBufferATI       = (fpglUpdateObjectBufferATI      )GetProcAddress( "glUpdateObjectBufferATI"       );
			fpneoglFreeObjectBufferATI         = (fpglFreeObjectBufferATI        )GetProcAddress( "glFreeObjectBufferATI"         );
			fpneoglArrayObjectATI              = (fpglArrayObjectATI             )GetProcAddress( "glArrayObjectATI"              );
			fpneoglVertexAttribArrayObjectATI  = (fpglVertexAttribArrayObjectATI )GetProcAddress( "glVertexAttribArrayObjectATI"  );

			fpneoglEnableVertexAttribArrayARB  = (fpglEnableVertexAttribArrayARB )GetProcAddress( "glEnableVertexAttribArrayARB"  );
			fpneoglDisableVertexAttribArrayARB = (fpglDisableVertexAttribArrayARB)GetProcAddress( "glDisableVertexAttribArrayARB" );

			m_kExtensions.Set( Extensions::ATI_VERTEX_ARRAY_OBJECT );

			m_pkVertexBufferManager = new VertexBufferManagerATI( this );

			if( !m_pkVertexBufferManager->SetSize( m_uiVertexStorageSize ) )
				delete m_pkVertexBufferManager, m_pkVertexBufferManager = 0;

			neoglCheckError( "*** Error setting up ATI vertex buffer extension: " );
		}
	}

	if( !m_pkVertexBufferManager )
		m_pkVertexBufferManager = new VertexBufferManagerNOBS( this );

	if( !Shader::s_veVSPriority.size() )
	{
		//Setup default priority
		Shader::s_veVSPriority.push_back( Shader::NEPT_GL_ARBVP_1 );
	}

	vector< Shader::SHADERTARGET >::iterator peTarget    = Shader::s_veVSPriority.begin();
	vector< Shader::SHADERTARGET >::iterator peTargetEnd = Shader::s_veVSPriority.end();

	for( ; peTarget != peTargetEnd; ++peTarget )
	{
		if( ( (*peTarget) == Shader::NEPT_GL_ARBVP_1 ) && ( strExt.find( "GL_ARB_vertex_program" ) < string::npos ) )
		{
			neolog << LogLevel( INFO ) << "GL_ARB_vertex_program supported" << endl;

			fpneoglVertexAttribPointerARB      = (fpglVertexAttribPointerARB     )GetProcAddress( "glVertexAttribPointerARB"      );
			fpneoglEnableVertexAttribArrayARB  = (fpglEnableVertexAttribArrayARB )GetProcAddress( "glEnableVertexAttribArrayARB"  );
			fpneoglDisableVertexAttribArrayARB = (fpglDisableVertexAttribArrayARB)GetProcAddress( "glDisableVertexAttribArrayARB" );

			fpneoglProgramStringARB            = (fpglProgramStringARB           )GetProcAddress( "glProgramStringARB"            );
			fpneoglBindProgramARB              = (fpglBindProgramARB             )GetProcAddress( "glBindProgramARB"              );
			fpneoglDeleteProgramsARB           = (fpglDeleteProgramsARB          )GetProcAddress( "glDeleteProgramsARB"           );
			fpneoglGenProgramsARB              = (fpglGenProgramsARB             )GetProcAddress( "glGenProgramsARB"              );
			fpneoglProgramEnvParameter4fARB    = (fpglProgramEnvParameter4fARB   )GetProcAddress( "glProgramEnvParameter4fARB"    );
			fpneoglProgramEnvParameter4fvARB   = (fpglProgramEnvParameter4fvARB  )GetProcAddress( "glProgramEnvParameter4fvARB"   );
			fpneoglProgramLocalParameter4fARB  = (fpglProgramLocalParameter4fARB )GetProcAddress( "glProgramLocalParameter4fARB"  );
			fpneoglProgramLocalParameter4fvARB = (fpglProgramLocalParameter4fvARB)GetProcAddress( "glProgramLocalParameter4fvARB" );
			fpneoglGetProgramivARB             = (fpglGetProgramivARB            )GetProcAddress( "glGetProgramivARB"             );

			m_kExtensions.Set( Extensions::ARB_VERTEX_PROGRAM );

			Shader::s_eVSTarget = Shader::NEPT_GL_ARBVP_1;

			break;
		}
	}

	if( Shader::s_eVSTarget != Shader::NOTSUPPORTED )
		m_kWindow.m_kCaps.Set( RenderCaps::VERTEXSHADER );

	if( !Shader::s_veFSPriority.size() )
	{
		//Setup default priority
		Shader::s_veFSPriority.push_back( Shader::NEPT_GL_ARBFP_1 );
		Shader::s_veFSPriority.push_back( Shader::NEPT_GL_NVRC_2 );
		Shader::s_veFSPriority.push_back( Shader::NEPT_GL_NVRC );
		Shader::s_veFSPriority.push_back( Shader::NEPT_GL_ATIFS );
	}

	peTarget    = Shader::s_veFSPriority.begin();
	peTargetEnd = Shader::s_veFSPriority.end();

	for( ; peTarget != peTargetEnd; ++peTarget )
	{
		if( ( (*peTarget) == Shader::NEPT_GL_ARBFP_1 ) && ( strExt.find( "GL_ARB_fragment_program" ) < string::npos ) )
		{
			neolog << LogLevel( INFO ) << "GL_ARB_fragment_program supported" << endl;

			fpneoglProgramStringARB            = (fpglProgramStringARB           )GetProcAddress( "glProgramStringARB"            );
			fpneoglBindProgramARB              = (fpglBindProgramARB             )GetProcAddress( "glBindProgramARB"              );
			fpneoglDeleteProgramsARB           = (fpglDeleteProgramsARB          )GetProcAddress( "glDeleteProgramsARB"           );
			fpneoglGenProgramsARB              = (fpglGenProgramsARB             )GetProcAddress( "glGenProgramsARB"              );
			fpneoglProgramEnvParameter4fARB    = (fpglProgramEnvParameter4fARB   )GetProcAddress( "glProgramEnvParameter4fARB"    );
			fpneoglProgramEnvParameter4fvARB   = (fpglProgramEnvParameter4fvARB  )GetProcAddress( "glProgramEnvParameter4fvARB"   );
			fpneoglProgramLocalParameter4fARB  = (fpglProgramLocalParameter4fARB )GetProcAddress( "glProgramLocalParameter4fARB"  );
			fpneoglProgramLocalParameter4fvARB = (fpglProgramLocalParameter4fvARB)GetProcAddress( "glProgramLocalParameter4fvARB" );
			fpneoglGetProgramivARB             = (fpglGetProgramivARB            )GetProcAddress( "glGetProgramivARB"             );

			m_kExtensions.Set( Extensions::ARB_FRAGMENT_PROGRAM );

			Shader::s_eFSTarget = Shader::NEPT_GL_ARBFP_1;

			break;
		}
		else if( ( ( (*peTarget) == Shader::NEPT_GL_NVRC ) || ( (*peTarget) == Shader::NEPT_GL_NVRC_2 ) ) && ( ( strExt.find( "GL_NV_register_combiners" ) < string::npos ) || ( strExt.find( "GL_NV_register_combiners2" ) < string::npos ) ) )
		{
			neolog << LogLevel( INFO ) << "GL_NV_register_combiners supported" << endl;

			fpneoglCombinerParameterfvNV = (fpglCombinerParameterfvNV)GetProcAddress( "glCombinerParameterfvNV" );
			fpneoglCombinerParameterfNV  = (fpglCombinerParameterfNV )GetProcAddress( "glCombinerParameterfNV"  );
			fpneoglCombinerParameterivNV = (fpglCombinerParameterivNV)GetProcAddress( "glCombinerParameterivNV" );
			fpneoglCombinerParameteriNV  = (fpglCombinerParameteriNV )GetProcAddress( "glCombinerParameteriNV"  );
			fpneoglCombinerInputNV       = (fpglCombinerInputNV      )GetProcAddress( "glCombinerInputNV"       );
			fpneoglCombinerOutputNV      = (fpglCombinerOutputNV     )GetProcAddress( "glCombinerOutputNV"      );
			fpneoglFinalCombinerInputNV  = (fpglFinalCombinerInputNV )GetProcAddress( "glFinalCombinerInputNV"  );

			if( strExt.find( "GL_NV_register_combiners2" ) < string::npos )
			{
				neolog << LogLevel( INFO ) << "GL_NV_register_combiners2 supported" << endl;
			
				fpneoglCombinerStageParameterfvNV = (fpglCombinerStageParameterfvNV)GetProcAddress( "glCombinerStageParameterfvNV" );

				m_kExtensions.Set( Extensions::NV_REGISTER_COMBINERS2 );

				Shader::s_eFSTarget = Shader::NEPT_GL_NVRC_2;
			}
			else
			{
				m_kExtensions.Set( Extensions::NV_REGISTER_COMBINERS );

				Shader::s_eFSTarget = Shader::NEPT_GL_NVRC;
			}

			break;
		}
		else if( ( (*peTarget) == Shader::NEPT_GL_ATIFS ) && ( ( strExt.find( "GL_ATI_fragment_shader" ) < string::npos ) || ( strExt.find( "GL_ATI_text_fragment_shader" ) < string::npos ) ) )
		{
			neolog << LogLevel( INFO ) << "GL_ATI_fragment_shader supported" << endl;

			fpneoglGenFragmentShadersATI        = (fpglGenFragmentShadersATI       )GetProcAddress( "glGenFragmentShadersATI"        );
			fpneoglBindFragmentShaderATI        = (fpglBindFragmentShaderATI       )GetProcAddress( "glBindFragmentShaderATI"        );
			fpneoglDeleteFragmentShaderATI      = (fpglDeleteFragmentShaderATI     )GetProcAddress( "glDeleteFragmentShaderATI"      );
			fpneoglBeginFragmentShaderATI       = (fpglBeginFragmentShaderATI      )GetProcAddress( "glBeginFragmentShaderATI"       );
			fpneoglEndFragmentShaderATI         = (fpglEndFragmentShaderATI        )GetProcAddress( "glEndFragmentShaderATI"         );
			fpneoglPassTexCoordATI              = (fpglPassTexCoordATI             )GetProcAddress( "glPassTexCoordATI"              );
			fpneoglSampleMapATI                 = (fpglSampleMapATI                )GetProcAddress( "glSampleMapATI"                 );
			fpneoglColorFragmentOp1ATI          = (fpglColorFragmentOp1ATI         )GetProcAddress( "glColorFragmentOp1ATI"          );
			fpneoglColorFragmentOp2ATI          = (fpglColorFragmentOp2ATI         )GetProcAddress( "glColorFragmentOp2ATI"          );
			fpneoglColorFragmentOp3ATI          = (fpglColorFragmentOp3ATI         )GetProcAddress( "glColorFragmentOp3ATI"          );
			fpneoglAlphaFragmentOp1ATI          = (fpglAlphaFragmentOp1ATI         )GetProcAddress( "glAlphaFragmentOp1ATI"          );
			fpneoglAlphaFragmentOp2ATI          = (fpglAlphaFragmentOp2ATI         )GetProcAddress( "glAlphaFragmentOp2ATI"          );
			fpneoglAlphaFragmentOp3ATI          = (fpglAlphaFragmentOp3ATI         )GetProcAddress( "glAlphaFragmentOp3ATI"          );
			fpneoglSetFragmentShaderConstantATI = (fpglSetFragmentShaderConstantATI)GetProcAddress( "glSetFragmentShaderConstantATI" );

			m_kExtensions.Set( Extensions::ATI_FRAGMENT_SHADER );

			Shader::s_eFSTarget = Shader::NEPT_GL_ATIFS;
			
			break;
		}
	}

	if( Shader::s_eFSTarget != Shader::NOTSUPPORTED )
		m_kWindow.m_kCaps.Set( RenderCaps::FRAGMENTSHADER );

	neoglCheckError( "*** Error in extension initialization: " );

	glGetIntegerv( GL_MAX_LIGHTS, (GLint*)&m_uiMaxLights );

	neolog << LogLevel( INFO ) << "[" << m_uiMaxLights << "] max lights" << endl;

	if( m_kWindow.m_kCaps.IsSet( RenderCaps::DOUBLEBUFFER ) )
		glDrawBuffer( GL_BACK );

	glShadeModel( GL_SMOOTH );
	glLightModeli( GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE );
	glLightModeli( GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE );
	glLightModeli( GL_LIGHT_MODEL_COLOR_CONTROL, GL_SINGLE_COLOR );

	glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );

	glClearStencil( 0 );
	glClearDepth( 1.0f );

	glDepthMask( GL_TRUE );

	glEnable( GL_CULL_FACE );
	glFrontFace( GL_CCW );
	glCullFace( GL_BACK );

	glBlendFunc( GL_ONE, GL_ZERO );
	glEnable( GL_BLEND );

	glDisable( GL_LIGHT0 );
	glDisable( GL_LIGHTING );

	glEnable( GL_DEPTH_TEST );
	glDepthFunc( GL_LEQUAL );
	glDepthMask( GL_TRUE );

	if( m_kWindow.m_kCaps.IsSet( RenderCaps::STENCILBUFFER ) )
		glEnable( GL_STENCIL_TEST );
	else
		glDisable( GL_STENCIL_TEST );
	glStencilOp( GL_KEEP, GL_KEEP, GL_KEEP );

	//glColorMaterial( GL_FRONT_AND_BACK, GL_DIFFUSE );
	glDisable( GL_COLOR_MATERIAL );
	glColor3f( 1.0f, 1.0f, 1.0f );

	glDisable( GL_TEXTURE_2D );

	glEnableClientState( GL_VERTEX_ARRAY );

	if( pkCaps && pkCaps->IsSet( RenderCaps::NORMALIZATION ) )
	{
		glEnable( GL_NORMALIZE );
		m_kWindow.m_kCaps.Set( RenderCaps::NORMALIZATION );
	}
	else
		glDisable( GL_NORMALIZE );

	//Get default light from config repository
	float fAmb = 0.3f;

	if( Core::Get()->GetConfig() )
		Core::Get()->GetConfig()->GetValue( "ambientlight", &fAmb );

	float fDefLight[4] = { fAmb, fAmb, fAmb, 1.0f };

	glLightModelfv( GL_LIGHT_MODEL_AMBIENT, fDefLight );

	SetViewport( 0, 0, GetWidth(), GetHeight() );

	m_kPerspectiveProjection.BuildProjection( m_kViewport );
	m_kOrthographicProjection.BuildProjection( m_kViewport );

	SetPerspectiveProjection();


	m_pkDefaultMaterial = new Material( "__defmat", Core::Get()->GetMaterialManager() );


	//Setup modelview transformation matrix
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	neoglCheckError( "*** Error in context state setup: " );

	//Clear buffers
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT );

	neoglCheckError( "*** Error in clearing buffers: " );

	return true;
}


void *Device::GetProcAddress( const char* pszName )
{
	if( !pszName )
		return 0;

#if defined (WIN32)

	return (void*)wglGetProcAddress( pszName );
    
#elif defined (POSIX)

	// Prepend a '_' for the Unix C symbol mangling convention
	string strSymbol = "_"; strSymbol += pszName;

	return (void*)neoglXGetProcAddressARB( (GLubyte*)pszName );
    
#elif defined (__APPLE__)

	// Prepend a '_' for the Unix C symbol mangling convention
	NSSymbol symbol;
	string strSymbolName = "_"; strSymbolName += pszName;
        
	symbol = 0;
        
	if( NSIsSymbolNameDefined( strSymbolName.c_str() ) )
		symbol = NSLookupAndBindSymbol( strSymbolName.c_str() );
        
	return( symbol ? NSAddressOfSymbol( symbol ) : 0 );
        
#else
#  error "Platform unimplemented"
#endif
}


}; // namespace NeoOGL
