/***************************************************************************
              pixelbuffer.cpp  -  OpenGL pixel buffer implementation
                             -------------------
    begin                : Thu April 17, 2003
    copyright            : (C) 2003 by Cody Russell
    email                : cody [at] jhu.edu
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoDevOpenGL, pixelbuffer.cpp

 The Initial Developer of the Original Code is Cody Russell.
 Portions created by Cody Russell are Copyright (C) 2003
 Cody Russell. All Rights Reserved.

 Contributors: Mattias Jansson (mattias@realityrift.com)

 ***************************************************************************/

#define __NEED_VECTOR_INT
#define NEEDGDI

#include "device.h"
#include "pixelbuffer.h"
#include "pixelbuffertarget.h"
#include "textureunit.h"

#include <neoengine/logstream.h>

using namespace std;
using namespace NeoEngine;


namespace NeoOGL
{


/*! Free buffer IDs */
vector< int > g_vuiFreePBufferID;

/*! Last buffer ID */
unsigned int  g_uiNextPBufferID = 0;


#ifdef WIN32

//Helper method
string GetLastErrorString( DWORD dwError )
{
	LPVOID lpMsgBuf;

    FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, dwError, MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ), (LPTSTR)&lpMsgBuf, 0, NULL );

	string strRet = (char*)lpMsgBuf;

	LocalFree( lpMsgBuf );

	return strRet;
}


#endif



NeoEngine::PixelBuffer *Device::CreatePixelBuffer( unsigned int uiWidth, unsigned int uiHeight, unsigned int uiBPP, Texture::TEXTURETYPE eTextureType )
{
	if( m_kWindow.m_kCaps.IsSet( RenderCaps::RENDERTEXTURE ) )
	{
		PixelBuffer *pkPBuffer = new PixelBuffer( this );

		if( pkPBuffer->Open( uiWidth, uiHeight, uiBPP, eTextureType ) )
			return pkPBuffer;

		neolog << LogLevel( ERROR ) << "*** Unable to create pixel buffer: Unable to open buffer [" << uiWidth << "x" << uiHeight << "x" << uiBPP << "]" << endl;

		delete pkPBuffer;
	}

	return 0;
}





PixelBuffer::PixelBuffer( Device *pkDevice ) :
	Activator(),
	NeoEngine::PixelBuffer(),
	m_pkDevice( pkDevice ),
	m_bIsBound( false ),
	m_kViewport( m_pkDevice->GetViewport() ),
	m_eCurCubemapFace( Texture::CUBEFACE_FRONT )
{
	Deactivate();

	if( g_vuiFreePBufferID.size() )
	{
		m_uiID = *( g_vuiFreePBufferID.end() - 1 );
		g_vuiFreePBufferID.pop_back();
	}
	else
		m_uiID = g_uiNextPBufferID++;

	m_pkDevice->m_vpkPixelBuffers.push_back( this );
}


PixelBuffer::~PixelBuffer()
{
	Close();

	g_vuiFreePBufferID.push_back( m_uiID );

	vector< PixelBuffer* >::iterator ppkBuffer    = m_pkDevice->m_vpkPixelBuffers.begin();
	vector< PixelBuffer* >::iterator ppkBufferEnd = m_pkDevice->m_vpkPixelBuffers.end();

	for( ; ppkBuffer != ppkBufferEnd; ++ppkBuffer )
		if( *ppkBuffer == this )
		{
			m_pkDevice->m_vpkPixelBuffers.erase( ppkBuffer );
			break;
		}

	if( m_pkDevice->m_pkPixelBufferTarget->m_pkCurBuffer == this )
		m_pkDevice->m_pkPixelBufferTarget->m_pkCurBuffer = 0;
}


bool PixelBuffer::Open( unsigned int uiWidth, unsigned int uiHeight, unsigned int uiBPP, Texture::TEXTURETYPE eTextureType )
{
	m_uiWidth   = uiWidth;
	m_uiHeight  = uiHeight;
	m_uiBPP     = uiBPP;

	float fExp    = float( log( float( m_uiWidth ) ) ) / float( log( 2.0f ) );
	float fFrac   = fExp - float( int( fExp ) );

	if( ( fFrac > 0.0001f ) && ( fFrac < 0.9999f ) )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to create pixelbuffer: width [" << m_uiWidth << "] not a power of 2, exponent [" << fExp << "]" << endl;
		return false;
	}

	fExp  = float( log( float( m_uiHeight ) ) ) / float( log( 2.0f ) );
	fFrac = fExp - float( int( fExp ) );	

	if( ( fFrac > 0.0001f ) && ( fFrac < 0.9999f ) )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to create pixelbuffer: height [" << m_uiHeight << "] not a power of 2, exponent [" << fExp << "]" << endl;
		return false;
	}


	while( m_uiWidth > m_pkDevice->GetWidth() )
		m_uiWidth >>= 2;

	while( m_uiHeight > m_pkDevice->GetHeight() )
		m_uiHeight >>= 2;

	// We shouldn't ruin the squareness when resizing
	if( m_uiWidth != m_uiHeight )
	{
		m_uiWidth  = MIN( m_uiWidth, m_uiHeight );
		m_uiHeight = m_uiWidth;
	}

	m_pkTexture = new NeoOGL::Texture( m_pkDevice, "__pbuffer", eTextureType, Texture::RGB, Texture::LINEAR | ( Texture::LINEAR << 8 ), 1, 0, 0, m_uiWidth, m_uiHeight );

	m_pkDevice->m_ppkTMU[0]->SetTexture( (NeoOGL::Texture*)(NeoEngine::Texture*)m_pkTexture );

	if( eTextureType == Texture::TEX2D )
	{
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

		glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB, m_uiWidth, m_uiHeight, 0, GL_RGB, GL_UNSIGNED_BYTE, 0 );
	}
	else if( eTextureType == Texture::CUBEMAP )
	{
		if( !m_pkDevice->m_kExtensions.IsSupported( Extensions::ARB_TEXTURE_CUBE_MAP ) )
		{
			neolog << LogLevel( ERROR ) << "*** Unable to create pixelbuffer: no cubemap support" << endl;
			return false;
		}

		glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
		glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
		glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE );

		glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
		glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

		for( int iFace = 0; iFace < 6; ++iFace )
			glTexImage2D( Texture::s_aeOGLCubeFaces[iFace], 0, GL_RGB, m_uiWidth, m_uiHeight, 0, GL_RGB, GL_UNSIGNED_BYTE, 0 );
	}
	else
	{
		neolog << LogLevel( ERROR ) << "*** Unable to create pixelbuffer: unsupported texture type [" << (int)eTextureType << "]" << endl;
		return false;
	}

	neoglCheckError();

	GLint iRealWidth = 0;
	GLint iRealHeight = 0;
	GLint iA = 0;
	GLint iR = 0;
	GLint iG = 0;
	GLint iB = 0;

	GLenum eTextureID = ( eTextureType == Texture::TEX2D ) ? GL_TEXTURE_2D : GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB;

	glGetTexLevelParameteriv( eTextureID, 0, GL_TEXTURE_WIDTH,      &iRealWidth );
	glGetTexLevelParameteriv( eTextureID, 0, GL_TEXTURE_HEIGHT,     &iRealHeight );
	glGetTexLevelParameteriv( eTextureID, 0, GL_TEXTURE_ALPHA_SIZE, &iA );
	glGetTexLevelParameteriv( eTextureID, 0, GL_TEXTURE_RED_SIZE,   &iR );
	glGetTexLevelParameteriv( eTextureID, 0, GL_TEXTURE_GREEN_SIZE, &iG );
	glGetTexLevelParameteriv( eTextureID, 0, GL_TEXTURE_BLUE_SIZE,  &iB );

	neolog << LogLevel( DEBUG ) << "Created pixelbuffer texture " << uiWidth << "x" << uiHeight << " ID " << m_pkTexture->GetID() << " -> [" << iRealWidth << "x" << iRealHeight << "] ARGB" << iA << iR << iG << iB << ( ( eTextureType == Texture::CUBEMAP ) ? " x 6" : "" ) << endl;

	m_uiWidth  = (unsigned int)iRealWidth;
	m_uiHeight = (unsigned int)iRealHeight;

	return true;
}


void PixelBuffer::Close()
{
	m_pkTexture = 0;
}


void PixelBuffer::Activate()
{
	if( IsActive() )
		return;

	Activator::Activate();

	m_kViewport = m_pkDevice->GetViewport();

	m_pkDevice->SetViewport( 0, 0, m_uiWidth, m_uiHeight );

	m_pkDevice->m_bIsRenderingTexture = true;

	m_pkDevice->SetCullMode( m_pkDevice->m_eCullMode );
}


void PixelBuffer::Deactivate()
{
	if( !IsActive() )
		return;

	Activator::Deactivate();

	GrabTexture();

	m_pkDevice->SetViewport( m_kViewport.m_iX, m_kViewport.m_iY, m_kViewport.m_iWidth, m_kViewport.m_iHeight );

	m_pkDevice->m_bIsRenderingTexture = false;

	m_pkDevice->SetCullMode( m_pkDevice->m_eCullMode );
}


void PixelBuffer::SetCubeMapFace( Texture::TEXTURECUBEMAPFACE eCubeMapFace )
{
	if( m_eCurCubemapFace == eCubeMapFace )
		return;

	if( IsActive() && m_pkTexture && ( m_pkTexture->GetType() == Texture::CUBEMAP ) )
		GrabTexture();

	m_eCurCubemapFace = eCubeMapFace;
}


void PixelBuffer::GrabTexture()
{
	if( m_pkTexture )
	{
		//Copy texture data
		m_pkDevice->m_ppkTMU[0]->SetTexture( (NeoOGL::Texture*)(NeoEngine::Texture*)m_pkTexture );

		GLenum eTarget = GL_TEXTURE_2D;

		if( m_pkTexture->GetType() == Texture::CUBEMAP )
			eTarget = Texture::s_aeOGLCubeFaces[ m_eCurCubemapFace ];

		glCopyTexSubImage2D( eTarget, 0, 0, 0, 0, m_pkDevice->GetHeight() - m_uiHeight, m_uiWidth, m_uiHeight );
	}
}



# if 0

#if defined( WIN32 )

bool PixelBuffer::Open( unsigned int uiWidth, unsigned int uiHeight, unsigned int uiBPP, bool bShareContext )
{
	int iFormat;

	m_uiWidth   = uiWidth;
	m_uiHeight  = uiHeight;
	m_uiBPP     = uiBPP;

	HDC   hDC   = wglGetCurrentDC();
	HGLRC hGLRC = wglGetCurrentContext();

	m_bShareContext = bShareContext;

	if( m_bShareContext )
	{
		iFormat = GetPixelFormat( hDC );

		if( !iFormat )
		{
			neolog << LogLevel( ERROR ) << "*** Unable to open pixel buffer: GetPixelFormat() failed" << endl;
			return false;
		}
	}
	else
	{
		// ### TODO: Make attributes fully configurable in the class definition
		// ###       Color channels especially should be configurable; e.g., R, RG, RGB, RGBA,
		// ###       and the bits per channel.
		vector< int > viAttribs;

		//viAttribs.push_back( WGL_SUPPORT_OPENGL_ARB );
		//viAttribs.push_back( TRUE );

		viAttribs.push_back( WGL_DRAW_TO_PBUFFER_ARB );
		viAttribs.push_back( GL_TRUE );

		//viAttribs.push_back( WGL_DOUBLE_BUFFER_ARB );
		//viAttribs.push_back( 0 );

		//if( m_pkDevice->GetCaps().IsSet( RenderCaps::RENDERTEXTURE ) )
		{
			viAttribs.push_back( WGL_BIND_TO_TEXTURE_RGB_ARB );
			viAttribs.push_back( GL_TRUE );
		}

		viAttribs.push_back( 0 );

		unsigned int uiNumFormats;

		neowglChoosePixelFormatARB( hDC, (const int*)&viAttribs.front(), NULL, 1, &iFormat, &uiNumFormats );

		if( !uiNumFormats )
		{
			neolog << LogLevel( ERROR ) << "*** Unable to open pixel buffer: Failed to locate compatible pixel format" << endl;
			return false;
		}
	}

	// ### TODO: Make attributes fully configurable in the class definition
	vector< int > viBufferAttribs;

	viBufferAttribs.push_back( WGL_TEXTURE_FORMAT_ARB );
	viBufferAttribs.push_back( WGL_TEXTURE_RGB_ARB );

	viBufferAttribs.push_back( WGL_TEXTURE_TARGET_ARB );
	viBufferAttribs.push_back( WGL_TEXTURE_2D_ARB );

	viBufferAttribs.push_back( WGL_MIPMAP_TEXTURE_ARB );
	viBufferAttribs.push_back( 0 );

	viBufferAttribs.push_back( WGL_PBUFFER_LARGEST_ARB );
	viBufferAttribs.push_back( FALSE );

	viBufferAttribs.push_back( 0 );

	if( !( m_hPbuffer = neowglCreatePbufferARB( hDC, iFormat, m_uiWidth, m_uiHeight, (const int*)&viBufferAttribs.front() ) ) )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to open pixel buffer: Failed to create pbuffer" << endl;
		return false;
	}

	if( !( m_hDC = neowglGetPbufferDCARB( m_hPbuffer ) ) )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to open pixel buffer: Failed to get pbuffer DC" << endl;
		return false;
	}

	if( m_bShareContext )
	{
		m_hGLRC = hGLRC;
	}
	else
	{
		m_hGLRC = wglCreateContext( m_hDC );

		if( !m_hGLRC )
		{
			neolog << LogLevel( ERROR ) << "*** Unable to open pixel buffer: Failed to create context" << endl;
			return false;
		}

		if( m_bShareLists )
		{
			if( !wglShareLists( hGLRC, m_hGLRC ) )
			{
				neolog << LogLevel( ERROR ) << "*** Unable to open pixel buffer: Failed to share lists" << endl;
				return false;
			}
		}
	}

	neowglQueryPbufferARB( m_hPbuffer, WGL_PBUFFER_WIDTH_ARB,  (int*)&m_uiWidth  );
	neowglQueryPbufferARB( m_hPbuffer, WGL_PBUFFER_HEIGHT_ARB, (int*)&m_uiHeight );

	neolog << LogLevel( INFO ) << "Created pixel buffer [" << m_uiWidth << "x" << m_uiHeight << "]" << endl;

/*
	wglMakeCurrent( m_hDC, m_hGLRC );

	//glDrawBuffer( GL_FRONT );

	glEnable( GL_DEPTH_TEST );
	glDepthFunc( GL_LEQUAL );
	glDepthMask( GL_TRUE );

	glStencilOp( GL_KEEP, GL_KEEP, GL_KEEP );
	glEnable( GL_STENCIL_TEST );

	//m_pkDevice->SetViewport( 0, 0, m_uiWidth, m_uiHeight );

	//m_pkDevice->SetPerspectiveProjection();

	//Setup modelview transformation matrix
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	//Clear buffers
	glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );
	glClearDepth( 1.0f );
	glClearStencil( 0 );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT );

*/
	neoglCheckError( "*** Error in creating pixel buffer: OpenGL error occurred: " );

	wglMakeCurrent( m_pkDevice->m_pkFrameBufferTarget->m_hDC, m_pkDevice->m_pkFrameBufferTarget->m_hRC );


	Bind();

	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

	//glTexParameteri( GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE ); //Generate mipmaps automatically


	GLint iRealWidth = 0;
	GLint iRealHeight = 0;
	GLint iA = 0;
	GLint iR = 0;
	GLint iG = 0;
	GLint iB = 0;

	GLenum eTextureID = GL_TEXTURE_2D;

	glGetTexLevelParameteriv( eTextureID, 0, GL_TEXTURE_WIDTH,      &iRealWidth );
	glGetTexLevelParameteriv( eTextureID, 0, GL_TEXTURE_HEIGHT,     &iRealHeight );
	glGetTexLevelParameteriv( eTextureID, 0, GL_TEXTURE_ALPHA_SIZE, &iA );
	glGetTexLevelParameteriv( eTextureID, 0, GL_TEXTURE_RED_SIZE,   &iR );
	glGetTexLevelParameteriv( eTextureID, 0, GL_TEXTURE_GREEN_SIZE, &iG );
	glGetTexLevelParameteriv( eTextureID, 0, GL_TEXTURE_BLUE_SIZE,  &iB );

	neolog << LogLevel( DEBUG ) << "Pixelbuffer texture [" << m_pkTexture->GetName() << "] " << m_uiWidth << "x" << m_uiHeight << " ID " << m_pkTexture->GetID() << " -> [" << iRealWidth << "x" << iRealHeight << "] ARGB" << iA << iR << iG << iB << endl;

	((NeoOGL::PixelBufferTexture*)(NeoEngine::Texture*)m_pkTexture)->SetDimensions( iRealWidth, iRealHeight );

	return true;
}


void PixelBuffer::Close()
{
	if( m_hPbuffer )
	{
		if( !m_bShareContext )
			wglDeleteContext( m_hGLRC );

		neowglReleasePbufferDCARB( m_hPbuffer, m_hDC );
		neowglDestroyPbufferARB( m_hPbuffer );
	}
}


void PixelBuffer::ModeSwitch()
{
	int iLost = 0;

	if( iLost )
	{
		Close();
		Open( m_uiWidth, m_uiHeight, m_uiBPP, m_bShareContext );
	}
}


int PixelBuffer::Bind()
{
	if( !m_bIsTexture )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to bind pixel buffer: No support for render to texture" << endl;
		return 0;
	}

	if( m_bIsBound )
		return 0;

	if( IsActive() )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to bind pixel buffer: Pixel buffer is current render target" << endl;
		return 0;
	}

	assert( !neoglCheckError() );

	m_pkDevice->m_ppkTMU[0]->SetTexture( (NeoOGL::Texture*)(NeoEngine::Texture*)m_pkTexture );

	assert( !neoglCheckError() );

	int iRet = neowglBindTexImageARB( m_hPbuffer, WGL_FRONT_LEFT_ARB );

	assert( !neoglCheckError() );

	if( !iRet )
	{
		DWORD dwError = GetLastError();
		neolog << LogLevel( ERROR ) << "*** Unable to bind pixel buffer: Failed to bind texture image" << endl <<
		                               "    " << GetLastErrorString( dwError ) << endl;
		neoglCheckError( "    " );
		return 0;
	}

	m_bIsBound = true;

	return iRet;
}


int PixelBuffer::Release()
{
	if( !m_bIsTexture )
		return 0;

	if( !m_bIsBound )
		return 0;

	if( IsActive() )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to release pixel buffer: Pixel buffer is current render target" << endl;
		return 0;
	}

	assert( !neoglCheckError() );

	m_pkDevice->m_ppkTMU[0]->SetTexture( (NeoOGL::Texture*)(NeoEngine::Texture*)m_pkTexture );

	assert( !neoglCheckError() );

	int iRet = neowglReleaseTexImageARB( m_hPbuffer, WGL_FRONT_LEFT_ARB );

	assert( !neoglCheckError() );

	if( !iRet )
	{
		DWORD dwError = GetLastError();
		neolog << LogLevel( ERROR ) << "*** Unable to release pixel buffer: Failed to release texture image" << endl <<
		                               "    " << GetLastErrorString( dwError ) << endl;
		neoglCheckError( "    " );
	}

	m_bIsBound = false;

	return iRet;
}


void PixelBuffer::Activate()
{
	if( IsActive() )
		return;

	if( m_bIsBound )
		Release();

	if( !wglMakeCurrent( m_hDC, m_hGLRC ) )
	{
		DWORD dwError = GetLastError();
		neolog << LogLevel( ERROR ) << "*** Unable to activate pixel buffer: Failed to make context current" << endl <<
		                               "    " << GetLastErrorString( dwError ) << endl;
	}
	else
		Activator::Activate();
}


void PixelBuffer::Deactivate()
{
	if( !IsActive() )
		return;

	Activator::Deactivate();

	if( !m_bIsBound && m_hPbuffer )
		Bind();
}


#elif defined( POSIX )


bool PixelBuffer::Open( unsigned int uiWidth, unsigned int uiHeight, unsigned int uiBPP, bool bShareContext )
{
	Display *pDisplay     = glXGetCurrentDisplay();
	int iScreen           = DefaultScreen( pDisplay );
	GLXContext glxContext = glXGetCurrentContext();

	GLXFBConfig *glxConfig;
	int          iConfigCount;

	m_bShareContext = bShareContext;
	m_bShareObjects = true;

	m_uiWidth = uiWidth;
	m_uiHeight = uiHeight;

	int aiFBAttribs[] =
	{
		GLX_DRAWABLE_TYPE,
		GLX_PBUFFER_BIT,
		GLX_RENDER_TYPE,
		GLX_RGBA_BIT
	};

	int aiCreateAttribs[] =
	{
		GLX_LARGEST_PBUFFER, true,
		GLX_PRESERVED_CONTENTS, true,
		0
	};

	if( m_bShareContext )
	{
		glxConfig = glXGetFBConfigs( pDisplay, iScreen, &iConfigCount );
		if( !glxConfig )
		{
			neolog << LogLevel( ERROR ) << "*** Unable to open pixel buffer: Failed to get FB configs" << endl;
			return false;
		}
	}
	else
	{
		glxConfig = glXChooseFBConfigSGIX( pDisplay, iScreen, aiFBAttribs, &iConfigCount );
		if( !glxConfig )
		{
			neolog << LogLevel( ERROR ) << "*** Unable to open pixel buffer: Failed to choose FB config" << endl;
			return false;
		}
	}

	m_glxPbuffer = glXCreateGLXPbufferSGIX( pDisplay, glxConfig[0], m_uiWidth, m_uiHeight, aiCreateAttribs );

	if( !m_glxPbuffer )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to open pixel buffer: Failed to create pbuffer" << endl;
		return false;
	}

	if( m_bShareContext )
	{
		m_glxContext = glxContext;
	}
	else
	{
		if( m_bShareObjects )
		{
			m_glxContext = glXCreateContextWithConfigSGIX( pDisplay, glxConfig[0], GLX_RGBA_TYPE, glxContext, true );
		}
		else
		{
			m_glxContext = glXCreateContextWithConfigSGIX( pDisplay, glxConfig[0], GLX_RGBA_TYPE, NULL, true );
		}

		if( !glxConfig )
		{
			neolog << LogLevel( ERROR ) << "*** Unable to open pixel buffer: Failed to create new context" << endl;
			return false;
		}
	}

	m_pDisplay = pDisplay;

	glXQueryGLXPbufferSGIX( m_pDisplay, m_glxPbuffer, GLX_WIDTH, &m_uiWidth );
	glXQueryGLXPbufferSGIX( m_pDisplay, m_glxPbuffer, GLX_HEIGHT, &m_uiHeight );

	neolog << LogLevel( INFO ) << "Create pixel buffer [" << m_uiWidth << "x" << m_uiHeight << "]" << endl;

	return true;
}


void PixelBuffer::Close()
{
	if( m_glxContext )
	{
		if( !m_bShareContext )
		{
			glXDestroyContext( m_pDisplay, m_glxContext );
		}
	}

	if( m_glxPbuffer )
	{
		glXDestroyGLXPbufferSGIX( m_pDisplay, m_glxPbuffer );
	}
}


int PixelBuffer::Bind()
{
	return 0;
}


int PixelBuffer::Release()
{
	if( !IsActive() )
		return 0;

	return 0;
}


void PixelBuffer::Activate()
{
	if( !glXMakeCurrent( m_pDisplay, m_glxPbuffer, m_glxContext ) )
		neolog << LogLevel( ERROR ) << "*** Unable to activate pixel buffer: Failed to make context current" << endl;

	Activator::Activate();
}


void PixelBuffer::Deactivate()
{
	Activator::Deactivate();
}


#elif defined( __APPLE__ )


bool PixelBuffer::Open( unsigned int uiWidth, unsigned int uiHeight, unsigned int uiBPP, bool bShareContext )
{
	return false;
}


void PixelBuffer::Close()
{
}


int PixelBuffer::Bind( int iBuffer )
{
}


int PixelBuffer::Release( int iBuffer )
{
}


void PixelBuffer::Activate()
{
	Activator::Activate();
}


void PixelBuffer::Deactivate()
{
	Activator::Deactivate() ;
}


#endif


#endif


}; // namespace NeoOGL

