/***************************************************************************
               general.cpp  -  nVidia register combiners parser
                             -------------------
    begin       : Wed Jan 7 2004
    copyright   : (C) 2004 by Reality Rift Studios
    email       : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoDevOpenGL, rcparse/general.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2004
 Reality Rift Studios. All Rights Reserved.
 ***************************************************************************/

#include "general.h"
#include "combiners.h"

#include <neoengine/shader.h>
#include <neoengine/logstream.h>

using namespace std;
using namespace NeoEngine;


namespace NeoOGL
{


bool NVGeneralCombiners::Validate( int numConsts, NVConstColor *pcc )
{
	GLint maxGCs;
	
	glGetIntegerv( GL_MAX_GENERAL_COMBINERS_NV, &maxGCs );

	if( num > maxGCs )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to compile fragment shader: " << num << " general combiners specified, only " << maxGCs << " supported" << endl;
		return false;
	}

	if( 0 == num )
	{
		//Setup a "fake" general combiner 0
		general[0].ZeroOut();
		num = 1;
	}

	localConsts = 0;
	int i;

	for( i = 0; i < num; ++i )
		localConsts += general[i].numConsts;

	if( localConsts > 0 )
		if( !fpneoglCombinerStageParameterfvNV )
		{
			neolog << LogLevel( WARNING ) << "*** Register combiners setup: Local constant(s) specified, but not supported -- ignored" << endl;
			localConsts = 0;
		}
		else
			for( i = 0; i < num; ++i )
				general[i].SetUnusedLocalConsts( numConsts, pcc );

	for( i = 0; i < num; ++i)
		if( !general[i].Validate(i) )
			return false;

	for( ; i < maxGCs; ++i )
		general[i].ZeroOut();

	return true;
}


void NVGeneralCombiners::Invoke( Device *pkDevice, NeoEngine::RenderPrimitive *pkPrimitive )
{
	for( int i = 0; i < num; ++i )
		general[i].Invoke( pkDevice, i, pkPrimitive );
	
	if( localConsts > 0 )
		glEnable( GL_PER_STAGE_CONSTANTS_NV );
	else
		glDisable( GL_PER_STAGE_CONSTANTS_NV );
}


void NVGeneralCombiners::BindLightParams( Device *pkDevice, NeoEngine::RenderPrimitive *pkPrimitive )
{
	for( int i = 0; i < num; ++i )
		general[i].BindLightParams( pkDevice, i, pkPrimitive );
}




void NVGeneralCombiner::ZeroOut()
{
	numPortions = 2;
	numConsts   = 0;

	portion[0].ZeroOut();
	portion[0].designator = RCP_RGB;
	portion[1].ZeroOut();
	portion[1].designator = RCP_ALPHA;
}


void NVGeneralCombiner::SetUnusedLocalConsts( int numGlobalConsts, NVConstColor *globalCCs )
{
	int i;

	for( i = 0; i < numGlobalConsts; ++i )
	{
		bool constUsed = false;
		int j;

		for( j = 0; j < numConsts; ++j )
			constUsed |= ( cc[j].reg.bits.name == globalCCs[i].reg.bits.name );
		
		if( !constUsed )
			cc[numConsts++] = globalCCs[i];
	}
}


bool NVGeneralCombiner::Validate( int stage )
{
	if( ( 2 == numConsts ) && ( cc[0].reg.bits.name == cc[1].reg.bits.name ) )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to compile fragment shader: Local constant set twice" << endl;
		return false;
	}

	switch( numPortions )
	{
		case 0:
			portion[0].designator = RCP_RGB;
			// Fallthru

		case 1:
			portion[1].designator = ( ( RCP_RGB == portion[0].designator ) ? RCP_ALPHA : RCP_RGB );
			// Fallthru

		case 2:
			if( portion[0].designator == portion[1].designator )
				neolog << LogLevel( ERROR ) << "*** Unable to compile fragment shader: Portion declared twice" << endl;
			break;
	}
	
	int i;
	
	for( i = 0; i < numPortions; ++i )
		if( !portion[i].Validate( stage ) )
			return false;

	for( ; i < 2; ++i )
		portion[i].ZeroOut();

	if( !fpneoglCombinerStageParameterfvNV )
		numConsts = 0;

	return true;
}


void NVGeneralCombiner::Invoke( Device *pkDevice, int stage, NeoEngine::RenderPrimitive *pkPrimitive )
{
	int i;

	for( i = 0; i < numConsts; ++i )
		NVRegisterCombiners::BindParam( pkDevice, stage, cc[i].reg.bits.name, m_pkParams[i], pkPrimitive );

	for( i = 0; i < 2; ++i )
		portion[i].Invoke( stage );
}


void NVGeneralCombiner::BindLightParams( Device *pkDevice, int stage, NeoEngine::RenderPrimitive *pkPrimitive )
{
	for( int i = 0; i < numConsts; ++i )
	{
		if( ( m_pkParams[i]->m_eType == ShaderParam::PPT_SYSTEM ) && ( m_pkParams[i]->m_Value.m_eSystemID & ShaderParam::LIGHTPARAM ) )
			NVRegisterCombiners::BindParam( pkDevice, stage, cc[i].reg.bits.name, m_pkParams[i], pkPrimitive );
	}
}




bool NVGeneralPortion::Validate( int stage )
{
	return gf.Validate( stage, designator );
}


void NVGeneralPortion::Invoke( int stage )
{
	gf.Invoke( stage, designator, bs );
}


void NVGeneralPortion::ZeroOut()
{
	gf.ZeroOut();
	bs.word = RCP_SCALE_BY_ONE;
}




void NVGeneralFunction::ZeroOut()
{
	// Create mapped registers for zero and discard
	NVMappedRegister unsignedZero;
	NVRegister zero;
	zero.word = RCP_ZERO;
	unsignedZero.Init( zero );

	NVMappedRegister unsignedDiscard;
	NVRegister discard;
	discard.word = RCP_DISCARD;
	unsignedDiscard.Init( discard );

	numOps = 3;

	op[0].op = RCP_MUL;
	op[0].reg[0] = unsignedDiscard;
	op[0].reg[1] = unsignedZero;
	op[0].reg[2] = unsignedZero;

	op[1].op = RCP_MUL;
	op[1].reg[0] = unsignedDiscard;
	op[1].reg[1] = unsignedZero;
	op[1].reg[2] = unsignedZero;

	op[2].op = RCP_SUM;
	op[2].reg[0] = unsignedDiscard;
}


bool NVGeneralFunction::Validate( int stage, int portion )
{
	int i;

	for( i = 0; i < numOps; ++i )
		if( !op[i].Validate(stage, portion) )
			return false;

	// Check if multiple ops are writing to same register (and it's not DISCARD)
	if( ( ( numOps > 1 ) &&
	      ( op[0].reg[0].reg.bits.name == op[1].reg[0].reg.bits.name ) &&
	      ( GL_DISCARD_NV != op[0].reg[0].reg.bits.name ) ) ||
	    ( ( ( numOps > 2 ) &&
	      ( ( op[0].reg[0].reg.bits.name == op[2].reg[0].reg.bits.name ) ||
	        ( op[1].reg[0].reg.bits.name == op[2].reg[0].reg.bits.name ) ) &&
	      ( GL_DISCARD_NV != op[2].reg[0].reg.bits.name ) ) ) )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to compile fragment shader: Writing to same register twice" << endl;
		return false;
	}

	// Set unused outputs to discard, unused inputs to zero/unsigned_identity
	if( numOps < 2 )
	{
		// Set C input to zero
		op[1].reg[1].reg.bits.name = GL_ZERO;
		op[1].reg[1].map = GL_UNSIGNED_IDENTITY_NV;
		op[1].reg[1].reg.bits.channel = portion;

		// Set D input to zero
		op[1].reg[2].reg.bits.name = GL_ZERO;
		op[1].reg[2].map = GL_UNSIGNED_IDENTITY_NV;
		op[1].reg[2].reg.bits.channel = portion;

		// Discard CD output
		op[1].op = false;
		op[1].reg[0].reg.bits.name = GL_DISCARD_NV;
	}

	if( numOps < 3 )
	{
		// Discard muxSum output
		op[2].reg[0].reg.bits.name = GL_DISCARD_NV;
		op[2].op = RCP_SUM;
	}

	return true;
}


void NVGeneralFunction::Invoke( int stage, int portion, NVBiasScale bs )
{
	GLenum portionEnum = ( RCP_RGB == portion ) ? GL_RGB : GL_ALPHA;

	neoglCombinerInputNV( GL_COMBINER0_NV + stage,
		portionEnum,
		GL_VARIABLE_A_NV,
		op[0].reg[1].reg.bits.name,
		op[0].reg[1].map,
		MAP_CHANNEL( op[0].reg[1].reg.bits.channel ) );

	neoglCombinerInputNV( GL_COMBINER0_NV + stage,
		portionEnum,
		GL_VARIABLE_B_NV,
		op[0].reg[2].reg.bits.name,
		op[0].reg[2].map,
		MAP_CHANNEL( op[0].reg[2].reg.bits.channel ) );

	neoglCombinerInputNV(GL_COMBINER0_NV + stage,
		portionEnum,
		GL_VARIABLE_C_NV,
		op[1].reg[1].reg.bits.name,
		op[1].reg[1].map,
		MAP_CHANNEL( op[1].reg[1].reg.bits.channel ) );

	neoglCombinerInputNV( GL_COMBINER0_NV + stage,
		portionEnum,
		GL_VARIABLE_D_NV,
		op[1].reg[2].reg.bits.name,
		op[1].reg[2].map,
		MAP_CHANNEL( op[1].reg[2].reg.bits.channel ) );

	neoglCombinerOutputNV( GL_COMBINER0_NV + stage,
		portionEnum,
		op[0].reg[0].reg.bits.name,
		op[1].reg[0].reg.bits.name,
		op[2].reg[0].reg.bits.name,
		bs.bits.scale,
		bs.bits.bias,
		op[0].op,
		op[1].op,
		( op[2].op == RCP_MUX ) ? true : false );
}




// This helper function assigns a channel to an undesignated input register
static void ConvertRegister( NVRegister &reg, int portion )
{
	if( RCP_NONE == reg.bits.channel ) {
		reg.bits.channel = portion;
		if( ( GL_FOG == reg.bits.name ) && ( RCP_ALPHA == portion ) )
			// Special case where fog alpha is final only, but RGB is not
			reg.bits.finalOnly = true;
	}
}


bool NVOp::Validate( int stage, int portion )
{
	int args = 1;

	if( ( RCP_DOT == op ) || ( RCP_MUL == op ) )
		args = 3;
	else
		args = 1;

	string strError;

	do
	{
		if( reg[0].reg.bits.readOnly )
		{
			strError = "Writing to a read-only register";
			break;
		}

		if( ( RCP_ALPHA == portion ) && ( RCP_DOT == op ) )
		{
			strError = "dot used in alpha portion";
			break;
		}

		int i;

		for( i = 0; i < args; ++i )
		{
			ConvertRegister( reg[i].reg, portion );
		
			if( reg[i].reg.bits.finalOnly )
			{
				strError = "final register used in general combiner";
				break;
			}

			if( ( RCP_RGB == portion ) && ( RCP_BLUE == reg[i].reg.bits.channel ) )
			{
				strError = "blue register used in rgb portion";
				break;
			}

			if( ( RCP_ALPHA == portion ) && ( RCP_RGB == reg[i].reg.bits.channel ) )
			{
				strError = "rgb register used in alpha portion";
				break;
			}
		
			if( ( i > 0 ) && ( GL_DISCARD_NV == reg[i].reg.bits.name ) )
			{
				strError = "reading from discard";
				break;
			}
		}
	} while( false );

	if( strError.length() )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to compile fragment shader: " << strError << endl;
		return false;
	}

	return true;
}


};

