/***************************************************************************
                    aabb.cpp  -  Axis-aligned bounding box
                             -------------------
    begin                : Wed Sep 18 2002
    copyright            : (C) 2002 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, aabb.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "aabb.h"
#include "obb.h"
#include "sphere.h"
#include "capsule.h"
#include "line.h"
#include "plane.h"
#include "contact.h"
#include "vertexbuffer.h"

#include <assert.h>

using namespace std;


namespace NeoEngine
{


AABB::AABB( const Vector3d &rkDim ) :
	BoundingVolume()
{
	m_kRealDim = rkDim;
	m_kDim     = rkDim;
	m_fRadius  = rkDim.Len();
}


void AABB::SetRotation( const Quaternion &rkRotation, bool bNotifyUpdate )
{
	BoundingVolume::SetRotation( rkRotation, bNotifyUpdate );

	RecalcDim();
}


void AABB::SetDim( const Vector3d &rkDim )
{
	m_kRealDim = rkDim;

	RecalcDim();
}


void AABB::RecalcDim()
{
	Matrix kMat( GetRotation() );

	Vector3d kX = kMat.GetColumn( 0 ) * m_kRealDim.x;
	Vector3d kY = kMat.GetColumn( 1 ) * m_kRealDim.y;
	Vector3d kZ = kMat.GetColumn( 2 ) * m_kRealDim.z;

	m_fRadius = m_kRealDim.Len();

	float fMaxX = 0.0f;
	float fMaxY = 0.0f;
	float fMaxZ = 0.0f;

	Vector3d kTest = kX + kY + kZ;

	if( fabsf( kTest.x ) > fMaxX )
		fMaxX = fabsf( kTest.x );
	if( fabsf( kTest.y ) > fMaxY )
		fMaxY = fabsf( kTest.y );
	if( fabsf( kTest.z ) > fMaxZ )
		fMaxZ = fabsf( kTest.z );

	kTest = kX + kY - kZ;

	if( fabsf( kTest.x ) > fMaxX )
		fMaxX = fabsf( kTest.x );
	if( fabsf( kTest.y ) > fMaxY )
		fMaxY = fabsf( kTest.y );
	if( fabsf( kTest.z ) > fMaxZ )
		fMaxZ = fabsf( kTest.z );

	kTest = kX - kY - kZ;

	if( fabsf( kTest.x ) > fMaxX )
		fMaxX = fabsf( kTest.x );
	if( fabsf( kTest.y ) > fMaxY )
		fMaxY = fabsf( kTest.y );
	if( fabsf( kTest.z ) > fMaxZ )
		fMaxZ = fabsf( kTest.z );

	kTest = kX - kY + kZ;

	if( fabsf( kTest.x ) > fMaxX )
		fMaxX = fabsf( kTest.x );
	if( fabsf( kTest.y ) > fMaxY )
		fMaxY = fabsf( kTest.y );
	if( fabsf( kTest.z ) > fMaxZ )
		fMaxZ = fabsf( kTest.z );

	m_kDim.x = fMaxX;
	m_kDim.y = fMaxY;
	m_kDim.z = fMaxZ;
}


bool AABB::Intersection( const Vector3d &rkPoint )
{
	return( ( m_kTranslation.x + m_kDim.x >= rkPoint.x ) &&
	        ( m_kTranslation.x - m_kDim.x <= rkPoint.x ) &&
	        ( m_kTranslation.y + m_kDim.y >= rkPoint.y ) &&
	        ( m_kTranslation.y - m_kDim.y <= rkPoint.y ) &&
	        ( m_kTranslation.z + m_kDim.z >= rkPoint.z ) &&
	        ( m_kTranslation.z - m_kDim.z <= rkPoint.z ) );
}


bool AABB::Intersection( const Line &rkLine )
{
	Vector3d kAbsDir;
	float fAbsWCrossDdU, fRhs;

	Vector3d kDiff = rkLine.m_kOrigin - m_kTranslation;
	Vector3d kWCrossD = rkLine.m_kDirection % kDiff;

	kAbsDir.y     = fabsf( rkLine.m_kDirection.y );
	kAbsDir.z     = fabsf( rkLine.m_kDirection.z );
	fAbsWCrossDdU = fabsf( kWCrossD.x );
	fRhs          = m_kDim.y * kAbsDir.z + m_kDim.z * kAbsDir.y;
	if( fAbsWCrossDdU > fRhs )
		return false;

	kAbsDir.x     = fabsf( rkLine.m_kDirection.x );
	fAbsWCrossDdU = fabsf( kWCrossD.y );
	fRhs          = m_kDim.x * kAbsDir.z + m_kDim.z * kAbsDir.x;
	if( fAbsWCrossDdU > fRhs )
		return false;

	fAbsWCrossDdU = fabsf( kWCrossD.z );
	fRhs          = m_kDim.x * kAbsDir.y + m_kDim.y * kAbsDir.x;
	if( fAbsWCrossDdU > fRhs )
		return false;

	return true;
}


bool AABB::Intersection( const Plane &rkPlane )
{
	Vector3d kTest = m_kTranslation;

	//Test most negative point
	kTest.x -= ( rkPlane.m_kNormal.x < 0.0f ) ? -m_kDim.x : m_kDim.x;
	kTest.y -= ( rkPlane.m_kNormal.y < 0.0f ) ? -m_kDim.y : m_kDim.y;
	kTest.z -= ( rkPlane.m_kNormal.z < 0.0f ) ? -m_kDim.z : m_kDim.z;

	float fDistance = rkPlane.Distance( kTest );

	if( fDistance > 0.0f )
		return false;

	//Test most positive point
	kTest = m_kTranslation;

	kTest.x += ( rkPlane.m_kNormal.x < 0.0f ) ? -m_kDim.x : m_kDim.x;
	kTest.y += ( rkPlane.m_kNormal.y < 0.0f ) ? -m_kDim.y : m_kDim.y;
	kTest.z += ( rkPlane.m_kNormal.z < 0.0f ) ? -m_kDim.z : m_kDim.z;

	fDistance = rkPlane.Distance( kTest );

	if( fDistance < 0.0f )
		return false;

	return true;
}


bool AABB::Intersection( const Vector3d &rkV0, const Vector3d &rkV1, const Vector3d &rkV2, ContactSet *pkContactSet, bool bForceTriNormal, const Vector3d &rkNormal )
{
	OBB kOBB( m_kDim );

	kOBB.SetTranslation( GetTranslation() );

	return kOBB.Intersection( rkV0, rkV1, rkV2, pkContactSet, bForceTriNormal, rkNormal );
}


//Cohen-Sutherland clipping outcodes
#define CSCLIPRIGHT  (1<<0)
#define CSCLIPLEFT   (1<<1)
#define CSCLIPTOP    (1<<2)
#define CSCLIPBOTTOM (1<<3)
#define CSCLIPFRONT  (1<<4)
#define CSCLIPBACK   (1<<5)
#define COHENSUTHERLAND_OUTCODE( code, pt, min, max ) { code = 0; if( pt.x > max.x ) code |= CSCLIPRIGHT; else if( pt.x < min.x ) code |= CSCLIPLEFT; if( pt.y > max.y ) code |= CSCLIPTOP; else if( pt.y < min.y ) code |= CSCLIPBOTTOM; if( pt.z > max.z ) code |= CSCLIPBACK; else if( pt.z < min.z ) code |= CSCLIPFRONT; }

bool AABB::Intersection( const Ray &rkRay, ContactSet *pkContactSet )
{
	//Based on the Cohen-Sutherland line-clipping algorithm
	int iOutcode;

	Vector3d kMin( m_kTranslation - m_kDim );
	Vector3d kMax( m_kTranslation + m_kDim );

	Vector3d kOrg( rkRay.m_kOrigin );

	COHENSUTHERLAND_OUTCODE( iOutcode, kOrg, kMin, kMax );
	if( !iOutcode )
	{
		//Inside bounding box
		if( pkContactSet )
		{
			Contact *pkData = new Contact;

			pkData->m_pkSceneNode = 0;
			
			pkData->m_pfDepths[0] = 0.0f;
			pkData->m_pkPoints[0] = kOrg;
			pkData->m_uiNumPoints = 1;

			pkContactSet->Add( pkData );
		}

		return true;
	}
/*
	Vector3d kEnd( rkPoint + ( rkPointVel * fMaxT ) );

	COHENSUTHERLAND_OUTCODE( iOutcode2, kEnd, kMin, kMax );
	if( !iOutcode2 )
	{
		//Inside bounding box
		//FIXME: If kOrg is outside and kEnd is inside, the collision point
		//       is somewhere on the line segment, NOT inside the box
		if( pkContactSet )
		{
			Contact *pkData = new Contact;

			pkData->m_fT = fMaxT;
			pkData->m_vkPoints.push_back( kEnd );

			pkContactSet->Add( pkData );
		}

		return true;
	}

	if( ( iOutcode & iOutcode2 ) )
	{
		//Both points on same side of box
		return false;
	}
*/
	Vector3d kColl;

	//Check intersections
	if( iOutcode & ( CSCLIPRIGHT | CSCLIPLEFT ) )
	{
		if( iOutcode & CSCLIPRIGHT )
			kColl.x = kMax.x;
		else
			kColl.x = kMin.x;

		float x1 = rkRay.m_kDirection.x;
		float x2 = kColl.x - kOrg.x;
		float x2divx1 = x2 / x1;

		kColl.y = kOrg.y + rkRay.m_kDirection.y * x2divx1;
		kColl.z = kOrg.z + rkRay.m_kDirection.z * x2divx1;

		if( ( kColl.y <= kMax.y ) && ( kColl.y >= kMin.y ) && ( kColl.z <= kMax.z ) && ( kColl.z >= kMin.z ) )
		{
			if( pkContactSet )
			{
				Contact *pkData = new Contact;

				pkData->m_pkPoints[0] = kColl;
				pkData->m_pfDepths[0] = x2divx1;
				pkData->m_uiNumPoints = 1;

				pkContactSet->Add( pkData );
			}

			return true;
		}
	}

	if( iOutcode & ( CSCLIPTOP | CSCLIPBOTTOM ) )
	{
		if( iOutcode & CSCLIPTOP )
			kColl.y = kMax.y;
		else
			kColl.y = kMin.y;

		float y1 = rkRay.m_kDirection.y;
		float y2 = kColl.y - kOrg.y;
		float y2divy1 = y2 / y1;

		kColl.x = kOrg.x + rkRay.m_kDirection.x * y2divy1;
		kColl.z = kOrg.z + rkRay.m_kDirection.z * y2divy1;

		if( ( kColl.x <= kMax.x ) && ( kColl.x >= kMin.x ) && ( kColl.z <= kMax.z ) && ( kColl.z >= kMin.z ) )
		{
			if( pkContactSet )
			{
				Contact *pkData = new Contact;

				pkData->m_pfDepths[0] = y2divy1;
				pkData->m_pkPoints[0] = kColl;
				pkData->m_uiNumPoints = 1;

				pkContactSet->Add( pkData );
			}

			return true;
		}
	}

	if( iOutcode & ( CSCLIPFRONT | CSCLIPBACK ) )
	{
		if( iOutcode & CSCLIPBACK )
			kColl.z = kMax.z;
		else
			kColl.z = kMin.z;

		float z1 = rkRay.m_kDirection.z;
		float z2 = kColl.z - kOrg.z;
		float z2divz1 = z2 / z1;

		kColl.x = kOrg.x + rkRay.m_kDirection.x * z2divz1;
		kColl.y = kOrg.y + rkRay.m_kDirection.y * z2divz1;

		if( ( kColl.x <= kMax.x ) && ( kColl.x >= kMin.x ) && ( kColl.y <= kMax.y ) && ( kColl.y >= kMin.y ) )
		{
			if( pkContactSet )
			{
				Contact *pkData = new Contact;

				pkData->m_pfDepths[0] = z2divz1;
				pkData->m_pkPoints[0] = kColl;
				pkData->m_uiNumPoints = 1;

				pkContactSet->Add( pkData );
			}

			return true;
		}
	}

	return false;
}



void AABB::Generate( AABB *pkAABB )
{
	SetTranslation( pkAABB->GetTranslation() );

	m_kRotation = pkAABB->m_kRotation;
	m_kRealDim  = pkAABB->m_kRealDim;
	m_kDim      = pkAABB->m_kDim;
	m_fRadius   = pkAABB->m_fRadius;
}


void AABB::Generate( OBB *pkOBB )
{
	m_kRealDim = pkOBB->GetDim();

	SetTranslation( pkOBB->GetTranslation() );
	SetRotation( pkOBB->GetRotation() );
}


void AABB::Generate( Sphere *pkSphere )
{
	SetTranslation( pkSphere->GetTranslation() );

	m_kRotation.Reset();

	m_kDim.Set( pkSphere->m_fRadius, pkSphere->m_fRadius, pkSphere->m_fRadius );

	m_kRealDim = m_kDim;
	m_fRadius  = pkSphere->m_fRadius;
}


void AABB::Generate( Capsule *pkCapsule )
{
	m_kRealDim.Set( pkCapsule->m_fLength + pkCapsule->m_fRadius, pkCapsule->m_fRadius, pkCapsule->m_fRadius );

	SetTranslation( pkCapsule->GetTranslation() );
	SetRotation( pkCapsule->GetRotation() );
}


void AABB::Generate( VertexBufferPtr &pkVertexBuffer )
{
	Vector3d kDim;
	bool bLocked = pkVertexBuffer->IsLocked();

	if( !bLocked )
		pkVertexBuffer->Lock( Buffer::READ );

	unsigned char *pucVertex     = (unsigned char*)pkVertexBuffer->GetVertex();
	unsigned int   uiNumVertices = pkVertexBuffer->GetNumElements();
	unsigned int   uiVertexSize  = pkVertexBuffer->GetVertexSize();

	Vector3d *pkVec;

	for( unsigned int i = 0; i < uiNumVertices; ++i, pucVertex += uiVertexSize )
	{
		pkVec = (Vector3d*)pucVertex;

		if( fabs( pkVec->x ) > kDim.x )
			kDim.x = fabsf( pkVec->x );
		if( fabs( pkVec->y ) > kDim.y )
			kDim.y = fabsf( pkVec->y );
		if( fabs( pkVec->z ) > kDim.z )
			kDim.z = fabsf( pkVec->z );
	}

	if( !bLocked )
		pkVertexBuffer->Unlock();

	m_kRealDim = kDim;
	m_kDim     = kDim;

	m_kTranslation.Reset();
	m_kRotation.Reset();
}


void AABB::Merge( AABB *pkAABB )
{
	Vector3d kMinOne = pkAABB->GetTranslation() - pkAABB->m_kDim;
	Vector3d kMinTwo = m_kTranslation           - m_kDim;

	Vector3d kMaxOne = pkAABB->GetTranslation() + pkAABB->m_kDim;
	Vector3d kMaxTwo = m_kTranslation           + m_kDim;

	if( kMinTwo.x < kMinOne.x )
		kMinOne.x = kMinTwo.x;
	if( kMinTwo.y < kMinOne.y )
		kMinOne.y = kMinTwo.y;
	if( kMinTwo.z < kMinOne.z )
		kMinOne.z = kMinTwo.z;

	if( kMaxTwo.x > kMaxOne.x )
		kMaxOne.x = kMaxTwo.x;
	if( kMaxTwo.y > kMaxOne.y )
		kMaxOne.y = kMaxTwo.y;
	if( kMaxTwo.z > kMaxOne.z )
		kMaxOne.z = kMaxTwo.z;

	m_kRealDim = ( kMaxOne - kMinOne ) * 0.5f;
	m_kDim     = m_kRealDim;
	m_fRadius  = m_kDim.Len();

	SetTranslation( kMinOne + m_kDim );

	m_kRotation.Reset();
}


void AABB::Merge( OBB *pkOBB )
{
	Vector3d kCenter = pkOBB->GetTranslation();
	Vector3d kDim    = pkOBB->GetRotation() * pkOBB->GetDim();

	kDim.x = fabsf( kDim.x );
	kDim.y = fabsf( kDim.y );
	kDim.z = fabsf( kDim.z );

	Vector3d kMinOne = kCenter - kDim;
	Vector3d kMaxOne = kCenter + kDim;

	Vector3d kMinTwo = m_kTranslation - m_kDim;
	Vector3d kMaxTwo = m_kTranslation + m_kDim;

	if( kMinTwo.x < kMinOne.x )
		kMinOne.x = kMinTwo.x;
	if( kMinTwo.y < kMinOne.y )
		kMinOne.y = kMinTwo.y;
	if( kMinTwo.z < kMinOne.z )
		kMinOne.z = kMinTwo.z;

	if( kMaxTwo.x > kMaxOne.x )
		kMaxOne.x = kMaxTwo.x;
	if( kMaxTwo.y > kMaxOne.y )
		kMaxOne.y = kMaxTwo.y;
	if( kMaxTwo.z > kMaxOne.z )
		kMaxOne.z = kMaxTwo.z;

	m_kDim     = ( kMaxOne - kMinOne ) * 0.5f;
	m_kRealDim = m_kDim;
	m_fRadius  = m_kDim.Len();

	SetTranslation( kMinOne + m_kDim );

	m_kRotation.Reset();
}


void AABB::Merge( Sphere *pkSphere )
{
	Vector3d kMinOne = pkSphere->GetTranslation() - Vector3d( pkSphere->m_fRadius, pkSphere->m_fRadius, pkSphere->m_fRadius );
	Vector3d kMinTwo = m_kTranslation             - m_kDim;

	Vector3d kMaxOne = pkSphere->GetTranslation() + Vector3d( pkSphere->m_fRadius, pkSphere->m_fRadius, pkSphere->m_fRadius );;
	Vector3d kMaxTwo = m_kTranslation             + m_kDim;

	if( kMinTwo.x < kMinOne.x )
		kMinOne.x = kMinTwo.x;
	if( kMinTwo.y < kMinOne.y )
		kMinOne.y = kMinTwo.y;
	if( kMinTwo.z < kMinOne.z )
		kMinOne.z = kMinTwo.z;

	if( kMaxTwo.x > kMaxOne.x )
		kMaxOne.x = kMaxTwo.x;
	if( kMaxTwo.y > kMaxOne.y )
		kMaxOne.y = kMaxTwo.y;
	if( kMaxTwo.z > kMaxOne.z )
		kMaxOne.z = kMaxTwo.z;

	m_kDim     = ( kMaxOne - kMinOne ) * 0.5f;
	m_kRealDim = m_kDim;
	m_fRadius  = m_kDim.Len();

	SetTranslation( kMinOne + m_kDim );

	m_kRotation.Reset();
}


void AABB::Merge( Capsule *pkCapsule )
{
	Vector3d kCenter = pkCapsule->GetTranslation();
	Vector3d kDim    = pkCapsule->GetRotation() * Vector3d( pkCapsule->m_fLength + pkCapsule->m_fRadius, pkCapsule->m_fRadius, pkCapsule->m_fRadius );

	kDim.x = fabsf( kDim.x );
	kDim.y = fabsf( kDim.y );
	kDim.z = fabsf( kDim.z );

	Vector3d kMinOne = kCenter - kDim;
	Vector3d kMaxOne = kCenter + kDim;

	Vector3d kMinTwo = m_kTranslation - m_kDim;
	Vector3d kMaxTwo = m_kTranslation + m_kDim;

	if( kMinTwo.x < kMinOne.x )
		kMinOne.x = kMinTwo.x;
	if( kMinTwo.y < kMinOne.y )
		kMinOne.y = kMinTwo.y;
	if( kMinTwo.z < kMinOne.z )
		kMinOne.z = kMinTwo.z;

	if( kMaxTwo.x > kMaxOne.x )
		kMaxOne.x = kMaxTwo.x;
	if( kMaxTwo.y > kMaxOne.y )
		kMaxOne.y = kMaxTwo.y;
	if( kMaxTwo.z > kMaxOne.z )
		kMaxOne.z = kMaxTwo.z;

	m_kDim     = ( kMaxOne - kMinOne ) * 0.5f;
	m_kRealDim = m_kDim;
	m_fRadius  = m_kDim.Len();

	SetTranslation( kMinOne + m_kDim );

	m_kRotation.Reset();
}


void AABB::RenderOutlines( const Color &rkColor ) const
{
	OBB kOBB( m_kDim );

	kOBB.SetTranslation( GetTranslation() );

	kOBB.RenderOutlines( rkColor );
}


BoundingVolume *AABB::Duplicate() const
{
	return new AABB( *this );
}


}; // namespace NeoEngine
