/***************************************************************************
                 astar.cpp  -  A* Pathfinding implementation
                             -------------------
    begin                : Fri Jan 30 2004
    copyright            : (C) 2004 by Reality Rift Studios
    email                : kmullis@zianet.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, astar.cpp

 The Initial Developer of the Original Code is Keaton Mullis.
 Portions created by Mattias Jansson are Copyright (C) 2004
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "astar.h"
#include "timer.h"

using namespace std;

namespace NeoEngine
{

AStarNodeLink::AStarNodeLink( AStarNode *pkTarget, float fLength ) : m_pkTarget( pkTarget ), m_fLength( fLength )
{

}

AStarNodeLink::AStarNodeLink( const AStarNodeLink &rkLink ) : m_pkTarget( rkLink.m_pkTarget ), m_fLength( rkLink.m_fLength )
{

}

AStarNodeLink::~AStarNodeLink()
{

}

AStarNode::AStarNode() : SRTNode(), m_pkParent( 0 ), m_fCost( numeric_limits< float >::max() ), m_fEstimate( numeric_limits< float >::max() ), m_bOpen( false )
{

}

AStarNode::AStarNode( const NeoEngine::AStarNode &rkNode ) : SRTNode( rkNode ), m_pkParent( 0 ), m_fCost( numeric_limits< float >::max() ), m_fEstimate( numeric_limits< float >::max() ), m_bOpen( false )
{

}

AStarNode::~AStarNode()
{
	unsigned int i;

	for( i = 0; i < m_vpkLinks.size(); i++ )
	{
		if( m_vpkLinks[i] )
		{
			delete m_vpkLinks[i];
			m_vpkLinks[i] = NULL;
		}
	}
	m_vpkLinks.clear();
}

void AStarNode::AddLink( const AStarNodeLink &rkLink )
{
	AStarNodeLink *pkLink = new AStarNodeLink( rkLink );

	m_vpkLinks.push_back( pkLink );
}

AStar::AStar()
{

}

AStar::~AStar()
{
	unsigned int i;

	for( i = 0; i < m_vpkNodes.size(); i++ )
	{
		if( m_vpkNodes[i] )
		{
			delete m_vpkNodes[i];
			m_vpkNodes[i] = NULL;
		}
	}
	m_vpkNodes.clear();
}

AStar::Status AStar::GeneratePath( const Vector3d &rkFrom, const Vector3d &rkTo, float fThreshold, vector< SRTNode > &rvpkPath, float fMaxTime )
{
	AStarNode *pkCurrent = NULL, *pkStart = NULL;
	Timer kTimer;
	unsigned int i;
	bool bSuccess = false;

	if( !( pkCurrent = m_vpkNodes[0] ) )
		return NOPATH;
	if( rkFrom != m_kFrom || rkTo != m_kTo )
	{
		float fClosest = ( pkCurrent->GetTranslation() - m_kFrom ).Len2();

		m_kFrom = rkFrom;
		m_kTo = rkTo;
		m_vpkOpenNodes.clear();
		for( i = 0; i < m_vpkNodes.size(); i++ )
		{
			float fDist2 = ( m_vpkNodes[i]->GetTranslation() - m_kFrom ).Len2();

			m_vpkNodes[i]->m_pkParent = NULL;
			m_vpkNodes[i]->m_fCost = numeric_limits< float >::max();
			m_vpkNodes[i]->m_fEstimate = numeric_limits< float >::max();
			m_vpkNodes[i]->m_bOpen = false;
			if( fDist2 < fClosest )
			{
				fClosest = fDist2;
				pkCurrent = m_vpkNodes[i];
			}
		}
		pkStart = pkCurrent;
		m_vpkOpenNodes.push_back( pkCurrent );
	}
	while( !m_vpkOpenNodes.empty() )
	{
		unsigned int uiCurrent = 0;

		if( kTimer.GetDeltaTime( false ) >= fMaxTime )
			return NOTFINISHED;
		pkCurrent = m_vpkOpenNodes[uiCurrent];
		for( i = 1; i < m_vpkOpenNodes.size(); i++ )
		{
			if( m_vpkOpenNodes[i]->m_fEstimate < pkCurrent->m_fEstimate )
			{
				uiCurrent = i;
				pkCurrent = m_vpkOpenNodes[i];
			}
		}
		m_vpkOpenNodes.erase( m_vpkOpenNodes.begin() + uiCurrent );
		pkCurrent->m_bOpen = false;
		if( ( m_kTo - pkCurrent->GetTranslation() ).Len2() < ( fThreshold * fThreshold ) )
		{
			bSuccess = true;
			break;
		}
		for( i = 0; i < pkCurrent->m_vpkLinks.size(); i++ )
		{
			AStarNode *pkTarget = pkCurrent->m_vpkLinks[i]->m_pkTarget;
			float fCost = pkCurrent->m_fCost + pkCurrent->m_vpkLinks[i]->m_fLength;
			float fEstimate = ( m_kTo - pkTarget->GetTranslation() ).Len();

			if( fCost > pkTarget->m_fCost )
				continue;
			pkTarget->m_pkParent = pkCurrent;
			pkTarget->m_fCost = fCost;
			pkTarget->m_fEstimate = fCost + fEstimate;
			if( !pkTarget->m_bOpen )
			{
				m_vpkOpenNodes.push_back( pkTarget );
				pkTarget->m_bOpen = true;
			}
		}
	}
	if( !bSuccess )
		return NOPATH;
	for( ; pkCurrent != NULL; pkCurrent = pkCurrent->m_pkParent )
	{
		rvpkPath.insert( rvpkPath.begin(), *pkCurrent );
		if( pkCurrent == pkStart )
			break;
	}
	return SUCCESS;
}

void AStar::AddNode( const AStarNode &rkNode )
{
	AStarNode *pkNode = new AStarNode( rkNode );

	m_vpkNodes.push_back( pkNode );
}

};
