/***************************************************************************
                          audio.cpp  -  Audio device
                             -------------------
    begin                : Wed Mar 17 2004
    copyright            : (C) 2004 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, audio.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2004
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "audio.h"
#include "core.h"
#include "logstream.h"

using namespace std;


namespace NeoEngine
{


AudioDevice::AudioDevice( FileManager *pkFileManager ) :
	m_pkFileManager( pkFileManager ? pkFileManager : Core::Get()->GetFileManager() )
{
}


AudioDevice::~AudioDevice()
{
	vector< SoundCodec* >::iterator ppkCodec    = m_vpkCodecs.begin();
	vector< SoundCodec* >::iterator ppkCodecEnd = m_vpkCodecs.end();

	for( ; ppkCodec != ppkCodecEnd; ++ppkCodec )
	{
		//By creating new reference to module, it won't get deleted inside codec dtor,
		//since that is a Bad Thing<tm> - the dtor is inside module space
		ModulePtr pkModule = (*ppkCodec)->m_pkModule;

		delete( *ppkCodec );
	}
}


bool AudioDevice::LoadSoundCodec( const std::string &rstrName )
{
	ModSymbol kSymLoad;

#if defined(WIN32) && defined(_DEBUG)
	ModulePtr pkModule = Core::Get()->GetModuleManager()->LoadModule( "neoac" + rstrName + "_debug" );
#else
	ModulePtr pkModule = Core::Get()->GetModuleManager()->LoadModule( "neoac" + rstrName );
#endif

	if( !pkModule || !pkModule->IsValid() )
		return false;

	kSymLoad = pkModule->LookupSymbol( "LoadSoundCodec" );

	if( !kSymLoad )
	{
		neolog << LogLevel( ERROR ) << "ERROR: Sound codec does not export LoadSoundCodec() method" << endl;
		return false;
	}

	SoundCodec *pkCodec = ( ( SoundCodec* ( NE_CDECL* )( const ModulePtr& ) )kSymLoad )( pkModule );

	if( !pkCodec )
	{
		neolog << LogLevel( ERROR ) << "ERROR: Unable to load sound codec" << endl;
		return false;
	}

	m_vpkCodecs.push_back( pkCodec );

	pkCodec->m_pkModule = pkModule;

	return true;
}



AudioDevice *Core::CreateAudioDevice( const string &rstrName, FileManager *pkFileManager )
{
#if defined(WIN32) && defined(_DEBUG)
	ModulePtr pkModule = s_pkModuleManager->LoadModule( "neodev" + rstrName + "_debug" );
#else
	ModulePtr pkModule = s_pkModuleManager->LoadModule( "neodev" + rstrName );
#endif

	if( !pkModule || !pkModule->IsValid() )
		return false;

	ModSymbol kSymInfo;
	ModSymbol kSymLoad;

	kSymInfo = pkModule->LookupSymbol( "GetVersion" );

	if( !kSymInfo )
		neolog << LogLevel( WARNING ) << "*** audio device module [" << rstrName << "] does not export GetVersion symbol" << endl;
	else
	{
		string strName;
		int iMajor;
		int iMinor;
		int iRevision;

		(( void ( NE_CDECL* )( std::string*, int*, int*, int* ))kSymInfo)( &strName, &iMajor, &iMinor, &iRevision );

		neolog << LogLevel( INFO ) << "Creating audio device " << rstrName << " version " << iMajor << "." << iMinor << "." << iRevision << endl;
	}

	kSymLoad = pkModule->LookupSymbol( "CreateDevice" );

	if( !kSymLoad )
	{
		neolog << LogLevel( ERROR ) << "*** render device module [" << rstrName << "] does not export CreateDevice method" << endl;
		return 0;
	}

	AudioDevice *pkDevice = (( AudioDevice* ( NE_CDECL* )( FileManager* ))kSymLoad)( pkFileManager );

	if( pkDevice )
	{
		m_pkAudioDevice = pkDevice;
		m_pkAudioDevice->m_pkModule = pkModule;
	}

	return pkDevice;
}


void Core::DeleteAudioDevice( AudioDevice *pkDevice )
{
	if( pkDevice != m_pkAudioDevice )
	{
		neolog << LogLevel( ERROR ) <<  "*** trying to delete non-current audio device" << endl;
		return;
	}

	if( m_pkAudioDevice )
	{
		//By creating new reference to module, it won't get deleted inside render device dtor,
		//since that is a Bad Thing<tm> - the dtor is inside module space
		ModulePtr pkModule = m_pkAudioDevice->m_pkModule;

		m_pkAudioDevice = 0;
	
		delete pkDevice;
	}
}


AudioDevice *Core::SetAudioDevice( AudioDevice *pkDevice )
{
	AudioDevice *pkOld = m_pkAudioDevice;

	m_pkAudioDevice = pkDevice;

	return( pkOld );
}


};
