/***************************************************************************
                       camera.cpp  -  Base camera class
                             -------------------
    begin                : Sat Nov 3 2001
    copyright            : (C) 2001 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, camera.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2001
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "camera.h"
#include "room.h"
#include "render.h"
#include "frustum.h"
#include "logstream.h"
#include "profile.h"

using namespace std;


namespace NeoEngine
{


Camera *Camera::s_pkActiveCamera = 0;


Camera::Camera( const string &rstrName, SceneNode *pkParent ) :
	SceneNode( rstrName, pkParent )
{
	m_pkRoom = 0;

	if( !s_pkActiveCamera )
		SetActive();
}


bool Camera::Render( Frustum *pkIgnoredFrustum, bool bForce )
{
	BEGIN_PROFILE( "Camera render" );
	
	assert( Core::Get()->GetRenderDevice() );

	pkIgnoredFrustum = 0;
	
	if( !m_pkRoom )
	{
		neolog << LogLevel( ERROR ) << "[Camera::Render]: No room" << endl;
		return false;
	}

	SetActive();

	Frustum kFrustum;

	const Matrix &rkProjection = Core::Get()->GetRenderDevice()->GetProjectionMatrix();

	float fWidth  = 2.0f * Core::Get()->GetRenderDevice()->GetZNear() / rkProjection[0][0];
	float fHeight = 2.0f * Core::Get()->GetRenderDevice()->GetZNear() / rkProjection[1][1];

	Quaternion kWorldRot = GetWorldRotation();
	Vector3d   kWorldPos = GetWorldTranslation();

	Vector3d kVecOne;
	Vector3d kVecTwo;
	Vector3d kNormal;
	float fDist;

	Matrix kViewProj( kWorldRot );

	//Left clipping plane
	kVecOne.x = -fWidth * 0.5f;
	kVecOne.y = fHeight * 0.5f;
	kVecOne.z = -Core::Get()->GetRenderDevice()->GetZNear();

	kVecTwo.x = -fWidth * 0.5f;
	kVecTwo.y = -fHeight * 0.5f;
	kVecTwo.z = -Core::Get()->GetRenderDevice()->GetZNear();

	kNormal   = kVecOne % kVecTwo;
	kNormal.Normalize();

	kNormal   = kViewProj * kNormal;

	fDist = -( kNormal * kWorldPos );
	kFrustum.m_akPlanes[ Frustum::LEFT ].Set( kNormal, fDist );


	//Right clipping plane
	kVecOne.x = fWidth * 0.5f;
	kVecOne.y = -fHeight * 0.5f;
	kVecOne.z = -Core::Get()->GetRenderDevice()->GetZNear();

	kVecTwo.x = fWidth * 0.5f;
	kVecTwo.y = fHeight * 0.5f;
	kVecTwo.z = -Core::Get()->GetRenderDevice()->GetZNear();

	kNormal   = kVecOne % kVecTwo;
	kNormal.Normalize();

	kNormal   = kViewProj * kNormal;

	fDist = -( kNormal * kWorldPos );
	kFrustum.m_akPlanes[ Frustum::RIGHT ].Set( kNormal, fDist );


	//Top clipping plane
	kVecOne.x = fWidth * 0.5f;
	kVecOne.y = fHeight * 0.5f;
	kVecOne.z = -Core::Get()->GetRenderDevice()->GetZNear();

	kVecTwo.x = -fWidth * 0.5f;
	kVecTwo.y = fHeight * 0.5f;
	kVecTwo.z = -Core::Get()->GetRenderDevice()->GetZNear();

	kNormal   = kVecOne % kVecTwo;
	kNormal.Normalize();

	kNormal   = kViewProj * kNormal;

	fDist = -( kNormal * kWorldPos );
	kFrustum.m_akPlanes[ Frustum::TOP ].Set( kNormal, fDist );


	//Bottom clipping plane
	kVecOne.x = -fWidth * 0.5f;
	kVecOne.y = -fHeight * 0.5f;
	kVecOne.z = -Core::Get()->GetRenderDevice()->GetZNear();

	kVecTwo.x = fWidth * 0.5f;
	kVecTwo.y = -fHeight * 0.5f;
	kVecTwo.z = -Core::Get()->GetRenderDevice()->GetZNear();

	kNormal   = kVecOne % kVecTwo;
	kNormal.Normalize();

	kNormal   = kViewProj * kNormal;

	fDist = -( kNormal * kWorldPos );
	kFrustum.m_akPlanes[ Frustum::BOTTOM ].Set( kNormal, fDist );


	// Near clipping plane
	kNormal   = kViewProj.GetColumn( 2 );

	kVecOne   = kWorldPos;
	kVecOne  += kNormal * -Core::Get()->GetRenderDevice()->GetZNear();

	fDist    = -( kNormal * kVecOne );
	kFrustum.m_akPlanes[ Frustum::FRONT ].Set( kNormal, fDist );


	// Far clipping plane
	kNormal   = -kViewProj.GetColumn( 2 );

	kVecOne   = kWorldPos;
	kVecOne  += kNormal * Core::Get()->GetRenderDevice()->GetZFar();

	fDist    = -( kNormal * kVecOne );
	kFrustum.m_akPlanes[ Frustum::BACK ].Set( kNormal, fDist );


	for( int i = 0; i < 6; i++ )
	{
		kFrustum.m_aiPVertexSign[i][0] = ( kFrustum.m_akPlanes[i].m_kNormal.x < 0.0f ) ? 1 : 0;  //1 means negative sign
		kFrustum.m_aiPVertexSign[i][1] = ( kFrustum.m_akPlanes[i].m_kNormal.y < 0.0f ) ? 1 : 0;
		kFrustum.m_aiPVertexSign[i][2] = ( kFrustum.m_akPlanes[i].m_kNormal.z < 0.0f ) ? 1 : 0;

		kFrustum.m_aiNVertexSign[i][0] = ( kFrustum.m_akPlanes[i].m_kNormal.x < 0.0f ) ? 0 : 1;
		kFrustum.m_aiNVertexSign[i][1] = ( kFrustum.m_akPlanes[i].m_kNormal.y < 0.0f ) ? 0 : 1;
		kFrustum.m_aiNVertexSign[i][2] = ( kFrustum.m_akPlanes[i].m_kNormal.z < 0.0f ) ? 0 : 1;
	}


	//Store data
	kFrustum.SetTranslation( GetWorldTranslation() );
	kFrustum.SetRotation( GetWorldRotation() );
	kFrustum.m_kDirection = kFrustum.m_akPlanes[ Frustum::BACK ].m_kNormal;
	kFrustum.m_fZNear     = Core::Get()->GetRenderDevice()->GetZNear();
	kFrustum.m_fZFar      = Core::Get()->GetRenderDevice()->GetZFar();


	m_pkRoom->Render( &kFrustum, bForce );

	END_PROFILE();

	return true;
}


void Camera::SetRoom( Room *pkRoom )
{
	m_pkRoom = pkRoom;
}


void Camera::LookAt( const Vector3d &rkAt, const Vector3d &rkUp )
{
	Vector3d kZAxis = GetWorldTranslation() - rkAt; kZAxis.Normalize(); //Looking down negative z axis

	//This is unstable if new kZAxis equals kUp vector
	Vector3d kXAxis = rkUp % kZAxis;   kXAxis.Normalize();
	Vector3d kYAxis = kZAxis % kXAxis; kYAxis.Normalize();

	Matrix kMat;

	kMat[0][0] = kXAxis[0];
	kMat[1][0] = kXAxis[1];
	kMat[2][0] = kXAxis[2];
	kMat[3][0] = 0.0f;

	kMat[0][1] = kYAxis[0];
	kMat[1][1] = kYAxis[1];
	kMat[2][1] = kYAxis[2];
	kMat[3][1] = 0.0f;

	kMat[0][2] = kZAxis[0];
	kMat[1][2] = kZAxis[1];
	kMat[2][2] = kZAxis[2];
	kMat[3][2] = 0.0f;

	kMat[0][3] = 0.0f;
	kMat[1][3] = 0.0f;
	kMat[2][3] = 0.0f;
	kMat[3][3] = 1.0f;

	SetRotation( Quaternion( kMat ) );
}


}; // namespace NeoEngine
