/***************************************************************************
                        collision.cpp  -  description
                             -------------------
    begin                : Thu May 16 2002
    copyright            : (C) 2002 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, collision.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "collision.h"
#include "aabb.h"
#include "obb.h"
#include "sphere.h"
#include "frustum.h"
#include "line.h"
#include "plane.h"
#include "contact.h"

#include <limits>

#include <assert.h>

using namespace std;


namespace NeoEngine
{


ContactSet::~ContactSet()
{
	vector< Contact* >::iterator ppkData    = m_vpkContacts.begin();
	vector< Contact* >::iterator ppkDataEnd = m_vpkContacts.end();

	for( ; ppkData != ppkDataEnd; ++ppkData )
		delete( *ppkData );
}


void ContactSet::Add( Contact *pkData )
{
	vector< Contact* >::iterator ppkData    = m_vpkContacts.begin();
	vector< Contact* >::iterator ppkDataEnd = m_vpkContacts.end();

	for( ; ppkData != ppkDataEnd; ++ppkData )
	{
		if( !(*ppkData)->m_uiNumPoints || ( pkData->m_uiNumPoints && ( pkData->m_pfDepths[0] >= (*ppkData)->m_pfDepths[0] ) ) )
		{
			m_vpkContacts.insert( ppkData, pkData );
			return;
		}
	}

	m_vpkContacts.push_back( pkData );
}


void ContactSet::Clear()
{
	vector< Contact* >::iterator ppkData    = m_vpkContacts.begin();
	vector< Contact* >::iterator ppkDataEnd = m_vpkContacts.end();

	for( ; ppkData != ppkDataEnd; ++ppkData )
		delete( *ppkData );

	m_vpkContacts.clear();
}



NEOENGINE_API bool Intersection( AABB *pkAABB0, AABB *pkAABB1, ContactSet *pkContactSet, bool bInvertNormal )
{
	if( !pkContactSet )
	{
		const Vector3d &rkC0   = pkAABB0->GetTranslation();
		const Vector3d &rkC1   = pkAABB1->GetTranslation();

		const Vector3d &rkDim0 = pkAABB0->GetDim();
		const Vector3d &rkDim1 = pkAABB1->GetDim();

		return( ( ( rkC0.x + rkDim0.x ) > ( rkC1.x - rkDim1.x ) ) &&
		        ( ( rkC0.x - rkDim0.x ) < ( rkC1.x + rkDim1.x ) ) &&
		        ( ( rkC0.y + rkDim0.y ) > ( rkC1.y - rkDim1.y ) ) &&
		        ( ( rkC0.y - rkDim0.y ) < ( rkC1.y + rkDim1.y ) ) &&
		        ( ( rkC0.z + rkDim0.z ) > ( rkC1.z - rkDim1.z ) ) &&
		        ( ( rkC0.z - rkDim0.z ) < ( rkC1.z + rkDim1.z ) ) );
	}
	else
	{
		OBB kOBB0( pkAABB0->GetDim() );
		OBB kOBB1( pkAABB1->GetDim() );

		kOBB0.SetTranslation( pkAABB0->GetTranslation() );
		kOBB1.SetTranslation( pkAABB1->GetTranslation() );

		return Intersection( &kOBB0, &kOBB1, pkContactSet, bInvertNormal );
	}
}


NEOENGINE_API bool Intersection( AABB *pkAABB, OBB *pkOBB, ContactSet *pkContactSet, bool bInvertNormal )
{
	static OBB kOBBTmp;

	kOBBTmp.Generate( pkAABB );

	return Intersection( &kOBBTmp, pkOBB, pkContactSet, bInvertNormal );
}


NEOENGINE_API bool Intersection( AABB *pkAABB, Sphere *pkSphere, ContactSet *pkContactSet, bool bInvertNormal )
{
	const Vector3d &rkAABBCenter   = pkAABB->GetTranslation();
	const Vector3d &rkSphereCenter = pkSphere->GetTranslation();
	const float    &rfRadius       = pkSphere->m_fRadius;

	const Vector3d &rkAABBDim      = pkAABB->GetDim();

	float fSide[3] = { 1.0f, 1.0f, 1.0f };
	float fS[3]    = { 0.0f, 0.0f, 0.0f };
	float fD       = 0.0f;

	if( rkSphereCenter.x < rkAABBCenter.x )
	{
		fS[0] = rkSphereCenter.x - ( rkAABBCenter.x - rkAABBDim.x );

		if( fS[0] <= -rfRadius )
			return false;
		else if( fS[0] < 0.0f )
			fD += fS[0] * fS[0];

		fS[0] += rfRadius;
	}
	else
	{
		fS[0] = ( rkAABBCenter.x + rkAABBDim.x ) - rkSphereCenter.x;

		if( fS[0] <= -rfRadius )
			return false;
		else if( fS[0] < 0.0f )
			fD += fS[0] * fS[0];

		fS[0] += rfRadius;
		fSide[0] = -1.0f;
	}

	if( rkSphereCenter.y < rkAABBCenter.y )
	{
		fS[1] = rkSphereCenter.y - ( rkAABBCenter.y - rkAABBDim.y );

		if( fS[1] <= -rfRadius )
			return false;
		else if( fS[1] < 0.0f )
			fD += fS[1] * fS[1];

		fS[1] += rfRadius;
	}
	else
	{
		fS[1] = ( rkAABBCenter.y + rkAABBDim.y ) - rkSphereCenter.y;

		if( fS[1] <= -rfRadius )
			return false;
		else if( fS[1] < 0.0f )
			fD += fS[1] * fS[1];

		fS[1] += rfRadius;
		fSide[1] = -1.0f;
	}

	if( rkSphereCenter.z < rkAABBCenter.z )
	{
		fS[2] = rkSphereCenter.z - ( rkAABBCenter.z - rkAABBDim.z );

		if( fS[2] <= -rfRadius )
			return false;
		else if( fS[2] < 0.0f )
			fD += fS[2] * fS[2];

		fS[2] += rfRadius;
	}
	else
	{
		fS[2] = ( rkAABBCenter.z + rkAABBDim.z ) - rkSphereCenter.z;

		if( fS[2] <= -rfRadius )
			return false;
		else if( fS[2] < 0.0f )
			fD += fS[2] * fS[2];

		fS[2] += rfRadius;
		fSide[2] = -1.0f;
	}

	if( fD <= ( rfRadius * rfRadius ) )
	{
		//We've got intersection
		if( !pkContactSet )
			return true;

		int iMinAxis = 0;
		int iSecAxis = 1;
		int iMajAxis = 2;

		if( fS[1] < fS[0] )
		{
			if( fS[2] < fS[1] )
			{
				iMinAxis = 2;
				iSecAxis = 1;
				iMajAxis = 0;
			}
			else
			{
				iMinAxis = 1;
				if( fS[2] < fS[0] )
				{
					iSecAxis = 2;
					iMajAxis = 0;
				}
				else
				{
					iSecAxis = 0;
					iMajAxis = 2;
				}
			}
		}
		else
		{
			if( fS[2] < fS[0] )
			{
				iMinAxis = 2;
				iSecAxis = 0;
				iMajAxis = 1;
			}
			else
			{
				iMinAxis = 0;
				if( fS[2] < fS[1] )
				{
					iSecAxis = 2;
					iMajAxis = 1;
				}
				else
				{
					iSecAxis = 1;
					iMajAxis = 2;
				}
			}
		}

		Contact *pkContact = new Contact;

		pkContact->m_kNormal = Vector3d::AXES[iMinAxis] * fSide[iMinAxis];

		Vector3d kPoint = rkSphereCenter + ( pkContact->m_kNormal * pkSphere->m_fRadius );
		Vector3d kPointDiff = kPoint - rkAABBCenter;
		
		if( ( fabsf( kPointDiff[iSecAxis] ) < rkAABBDim[iSecAxis] ) && ( fabsf( kPointDiff[iMajAxis] ) < rkAABBDim[iMajAxis] ) )
		{
			//Inside cube
			kPoint -= ( pkContact->m_kNormal * ( fS[iMinAxis] * 0.5f ) );

			pkContact->m_pkPoints[0] = kPoint;
			pkContact->m_pfDepths[0] = fS[iMinAxis];
			pkContact->m_uiNumPoints = 1;
		}
		else
		{
			//Outside cube
			kPoint = rkAABBCenter;
			
			kPoint[iMinAxis] -= rkAABBDim[iMinAxis] * fSide[iMinAxis];
			kPoint[iSecAxis] -= rkAABBDim[iSecAxis] * fSide[iSecAxis];

			Plane kPlane( Vector3d::AXES[iMajAxis], rkSphereCenter );

			float fDist = kPlane.Distance( kPoint );

			if( fDist < -rkAABBDim[iMajAxis] )
				fDist = -rkAABBDim[iMajAxis];
			else if( fDist > rkAABBDim[iMajAxis] )
				fDist = rkAABBDim[iMajAxis];

			kPoint -= Vector3d::AXES[iMajAxis] * fDist;

			Vector3d kNormal = kPoint - rkSphereCenter;

			float fLen = kNormal.Len();

			float fDepth = pkSphere->m_fRadius - fLen;

			kNormal *= 1.0f / fLen;

			kPoint += kNormal * ( fDepth * 0.5f );

			pkContact->m_pkPoints[0] = kPoint;
			pkContact->m_pfDepths[0] = fDepth;
			pkContact->m_uiNumPoints = 1;
			pkContact->m_kNormal     = kNormal;
		}

		if( bInvertNormal )
			pkContact->m_kNormal = -pkContact->m_kNormal;

		pkContactSet->Add( pkContact );

		return true;
	}

	return false;
}


NEOENGINE_API bool Intersection( AABB *pkAABB, Capsule *pkCapsule, ContactSet *pkContactSet, bool bInvertNormal )
{
	assert( !"Not implemented" );
	return false;
}


bool Intersection( AABB *pkAABB, Frustum *pkFrustum )
{
	const Vector3d &rkAABBPos = pkAABB->GetTranslation();
	const Vector3d &rkAABBDim = pkAABB->GetDim();

	Vector3d kTest;

	for( int iPlane = 0; iPlane < 6; ++iPlane )
	{
		kTest = rkAABBPos;

		//Test most negative point
		kTest.x += ( pkFrustum->m_akPlanes[iPlane].m_kNormal.x > 0.0f ) ? (-rkAABBDim.x) : rkAABBDim.x;
		kTest.y += ( pkFrustum->m_akPlanes[iPlane].m_kNormal.y > 0.0f ) ? (-rkAABBDim.y) : rkAABBDim.y;
		kTest.z += ( pkFrustum->m_akPlanes[iPlane].m_kNormal.z > 0.0f ) ? (-rkAABBDim.z) : rkAABBDim.z;

		if( pkFrustum->m_akPlanes[iPlane].Distance( kTest ) > 0.0f )
			return false;
	}

	return true;
}


NEOENGINE_API bool Intersection( OBB *pkOBB0, OBB *pkOBB1, ContactSet *pkContactSet, bool bInvertNormal )
{
	//Early out test
	if( ( ( pkOBB0->GetRadius() + pkOBB1->GetRadius() ) * ( pkOBB0->GetRadius() + pkOBB1->GetRadius() ) ) < ( pkOBB0->GetTranslation() - pkOBB1->GetTranslation() ).Len2() )
		return false;

	const float fParallellCutoff = 0.99999f;
	bool bParallellAxes = false;

	Matrix kRotMatA( pkOBB0->GetRotation() );
	Matrix kRotMatB( pkOBB1->GetRotation() );

	Vector3d akAxesA[3] = { kRotMatA.GetColumn(0), kRotMatA.GetColumn(1), kRotMatA.GetColumn(2) };
	Vector3d akAxesB[3] = { kRotMatB.GetColumn(0), kRotMatB.GetColumn(1), kRotMatB.GetColumn(2) };

	float afDimA[3] = { pkOBB0->GetDim().x, pkOBB0->GetDim().y, pkOBB0->GetDim().z };
	float afDimB[3] = { pkOBB1->GetDim().x, pkOBB1->GetDim().y, pkOBB1->GetDim().z };

    //Diffrence of box positions
    Vector3d kDiff = pkOBB1->GetTranslation() - pkOBB0->GetTranslation();

	float afAxisDot[3][3];    // afAxisDot[i][j]    = akAxesA[i] * akAxesB[j];
	float afAbsAxisDot[3][3]; // afAbsAxisDot[i][j] = |afAxisDot[i][j]|

	float afAxesADotDiff[3];  // afAxesADotDiff[i]  = akAxesA[i] * kDiff

	//Test all 15 possible separating axes

	//Axis Ax
	//First calculate Ax * Bi axis dot products (length of each B axis along Ax)
	afAxisDot[0][0] = akAxesA[0] * akAxesB[0];
	afAxisDot[0][1] = akAxesA[0] * akAxesB[1];
	afAxisDot[0][2] = akAxesA[0] * akAxesB[2];

	//Get absolute value of dot products
	afAbsAxisDot[0][0] = fabsf( afAxisDot[0][0] ); if( afAbsAxisDot[0][0] > fParallellCutoff ) bParallellAxes = true;
	afAbsAxisDot[0][1] = fabsf( afAxisDot[0][1] ); if( afAbsAxisDot[0][1] > fParallellCutoff ) bParallellAxes = true;
	afAbsAxisDot[0][2] = fabsf( afAxisDot[0][2] ); if( afAbsAxisDot[0][2] > fParallellCutoff ) bParallellAxes = true;

	//Calculate Ax * D dot product (length of center difference along Ax axis)
	afAxesADotDiff[0]  = akAxesA[0] * kDiff;

	int      iDeepAxis = 0;
	Vector3d kNormal;

	float fOverlapMax  = -numeric_limits< float >::max();
	float fOverlap;

	//Check if distance between centers along axis is greater than sum of boxes dimensions along axis
	if( ( fOverlap = fabsf( afAxesADotDiff[0] ) - ( ( afDimA[0] ) + ( afDimB[0] * afAbsAxisDot[0][0] + afDimB[1] * afAbsAxisDot[0][1] + afDimB[2] * afAbsAxisDot[0][2] ) ) ) > 0.0f )
		return false;
	if( pkContactSet && ( fOverlap > fOverlapMax ) )
	{
		fOverlapMax = fOverlap;
		kNormal     = ( afAxesADotDiff[0] < 0.0f ) ? -akAxesA[0] : akAxesA[0];
		iDeepAxis   = 0;
	}

	//Axis Ay
	afAxisDot[1][0] = akAxesA[1] * akAxesB[0];
	afAxisDot[1][1] = akAxesA[1] * akAxesB[1];
	afAxisDot[1][2] = akAxesA[1] * akAxesB[2];

	afAbsAxisDot[1][0] = fabsf( afAxisDot[1][0] ); if( afAbsAxisDot[1][0] > fParallellCutoff ) bParallellAxes = true;
	afAbsAxisDot[1][1] = fabsf( afAxisDot[1][1] ); if( afAbsAxisDot[1][1] > fParallellCutoff ) bParallellAxes = true;
	afAbsAxisDot[1][2] = fabsf( afAxisDot[1][2] ); if( afAbsAxisDot[1][2] > fParallellCutoff ) bParallellAxes = true;

	afAxesADotDiff[1]  = akAxesA[1] * kDiff;

	if( ( fOverlap = fabsf( afAxesADotDiff[1] ) - ( ( afDimA[1] ) + ( afDimB[0] * afAbsAxisDot[1][0] + afDimB[1] * afAbsAxisDot[1][1] + afDimB[2] * afAbsAxisDot[1][2] ) ) ) > 0.0f )
		return false;
	if( pkContactSet && ( fOverlap > fOverlapMax ) )
	{
		fOverlapMax = fOverlap;
		kNormal     = ( afAxesADotDiff[1] < 0.0f ) ? -akAxesA[1] : akAxesA[1];
		iDeepAxis   = 1;
	}


	//Axis Az
	afAxisDot[2][0] = akAxesA[2] * akAxesB[0];
	afAxisDot[2][1] = akAxesA[2] * akAxesB[1];
	afAxisDot[2][2] = akAxesA[2] * akAxesB[2];

	afAbsAxisDot[2][0] = fabsf( afAxisDot[2][0] ); if( afAbsAxisDot[2][0] > fParallellCutoff ) bParallellAxes = true;
	afAbsAxisDot[2][1] = fabsf( afAxisDot[2][1] ); if( afAbsAxisDot[2][1] > fParallellCutoff ) bParallellAxes = true;
	afAbsAxisDot[2][2] = fabsf( afAxisDot[2][2] ); if( afAbsAxisDot[2][2] > fParallellCutoff ) bParallellAxes = true;

	afAxesADotDiff[2]  = akAxesA[2] * kDiff;

	if( ( fOverlap = fabsf( afAxesADotDiff[2] ) - ( ( afDimA[2] ) + ( afDimB[0] * afAbsAxisDot[2][0] + afDimB[1] * afAbsAxisDot[2][1] + afDimB[2] * afAbsAxisDot[2][2] ) ) ) > 0.0f )
		return false;
	if( pkContactSet && ( fOverlap > fOverlapMax ) )
	{
		fOverlapMax = fOverlap;
		kNormal     = ( afAxesADotDiff[2] < 0.0f ) ? -akAxesA[2] : akAxesA[2];
		iDeepAxis   = 2;
	}


	float fProj;
	float fScale;

	//Axis Bx
	//We already have all axis*axis dot products, only calculate center difference along Bx axis and compare
	if( ( fOverlap = fabsf( ( fProj = akAxesB[0] * kDiff ) ) - ( ( afDimA[0] * afAbsAxisDot[0][0] + afDimA[1] * afAbsAxisDot[1][0] + afDimA[2] * afAbsAxisDot[2][0] ) + ( afDimB[0] ) ) ) > 0.0f )
		return false;
	if( pkContactSet && ( fOverlap > fOverlapMax ) )
	{
		fOverlapMax = fOverlap;
		kNormal     = ( fProj < 0.0f ) ? -akAxesB[0] : akAxesB[0];
		iDeepAxis   = 3;
	}


	//Axis By
	if( ( fOverlap = fabsf( ( fProj = akAxesB[1] * kDiff ) ) - ( ( afDimA[0] * afAbsAxisDot[0][1] + afDimA[1] * afAbsAxisDot[1][1] + afDimA[2] * afAbsAxisDot[2][1] ) + ( afDimB[1] ) ) ) > 0.0f )
		return false;
	if( pkContactSet && ( fOverlap > fOverlapMax ) )
	{
		fOverlapMax = fOverlap;
		kNormal     = ( fProj < 0.0f ) ? -akAxesB[1] : akAxesB[1];
		iDeepAxis   = 4;
	}


	//Axis Bz
	if( ( fOverlap = fabsf( ( fProj = akAxesB[2] * kDiff ) ) - ( ( afDimA[0] * afAbsAxisDot[0][2] + afDimA[1] * afAbsAxisDot[1][2] + afDimA[2] * afAbsAxisDot[2][2] ) + ( afDimB[2] ) ) ) > 0.0f )
		return false;
	if( pkContactSet && ( fOverlap > fOverlapMax ) )
	{
		fOverlapMax = fOverlap;
		kNormal     = ( fProj < 0.0f ) ? -akAxesB[2] : akAxesB[2];
		iDeepAxis   = 5;
	}

	//if not contact set, do extra tests to avoid reporting false collisions in parallell axis threshold zone
	if( !bParallellAxes || !pkContactSet )
	{

	//Axis Ax X Bx
	if( ( fOverlap = fabsf( ( fProj = afAxisDot[1][0] * afAxesADotDiff[2] - afAxisDot[2][0] * afAxesADotDiff[1] ) ) -
	                 ( ( afDimA[1] * afAbsAxisDot[2][0] + afDimA[2] * afAbsAxisDot[1][0] ) + ( afDimB[1] * afAbsAxisDot[0][2] + afDimB[2] * afAbsAxisDot[0][1] ) ) ) > 0.0f )
		return false;
	if( pkContactSet )
	{
		fScale = sqrtf( afAxisDot[2][0] * afAxisDot[2][0] + afAxisDot[1][0] * afAxisDot[1][0] );

		if( fScale > 0.0f )
		{
			fScale    = 1.0f / fScale;
			fOverlap *= fScale;

			if( fOverlap > fOverlapMax )
			{
				fOverlapMax = fOverlap;
				kNormal.Set( 0.0f, -afAxisDot[2][0] * fScale, afAxisDot[1][0] * fScale );

				if( fProj < 0.0f )
					kNormal = -kNormal;

				iDeepAxis   = 6;
			}
		}
	}

	//Axis Ax X By
	if( ( fOverlap = fabsf( ( fProj = afAxisDot[1][1] * afAxesADotDiff[2] - afAxisDot[2][1] * afAxesADotDiff[1] ) ) -
	                 ( ( afDimA[1] * afAbsAxisDot[2][1] + afDimA[2] * afAbsAxisDot[1][1] ) + ( afDimB[0] * afAbsAxisDot[0][2] + afDimB[2] * afAbsAxisDot[0][0] ) ) ) > 0.0f )
		return false;
	if( pkContactSet )
	{
		fScale = sqrtf( afAxisDot[2][1] * afAxisDot[2][1] + afAxisDot[1][1] * afAxisDot[1][1] );

		if( fScale > 0.0f )
		{
			fScale    = 1.0f / fScale;
			fOverlap *= fScale;

			if( fOverlap > fOverlapMax )
			{
				fOverlapMax = fOverlap;
				kNormal.Set( 0.0f, -afAxisDot[2][1] * fScale, afAxisDot[1][1] * fScale );

				if( fProj < 0.0f )
					kNormal = -kNormal;

				iDeepAxis   = 7;
			}
		}
	}

	//Axis Ax X Bz
	if( ( fOverlap = fabsf( ( fProj = afAxisDot[1][2] * afAxesADotDiff[2] - afAxisDot[2][2] * afAxesADotDiff[1] ) ) -
	                 ( ( afDimA[1] * afAbsAxisDot[2][2] + afDimA[2] * afAbsAxisDot[1][2] ) + ( afDimB[0] * afAbsAxisDot[0][1] + afDimB[1] * afAbsAxisDot[0][0] ) ) ) > 0.0f )
		return false;
	if( pkContactSet )
	{
		fScale = sqrtf( afAxisDot[2][2] * afAxisDot[2][2] + afAxisDot[1][2] * afAxisDot[1][2] );

		if( fScale > 0.0f )
		{
			fScale    = 1.0f / fScale;
			fOverlap *= fScale;

			if( fOverlap > fOverlapMax )
			{
				fOverlapMax = fOverlap;
				kNormal.Set( 0.0f, -afAxisDot[2][2] * fScale, afAxisDot[1][2] * fScale );

				if( fProj < 0.0f )
					kNormal = -kNormal;

				iDeepAxis   = 8;
			}
		}
	}

	//Axis Ay X Bx
	if( ( fOverlap = fabsf( ( fProj = afAxisDot[2][0] * afAxesADotDiff[0] - afAxisDot[0][0] * afAxesADotDiff[2] ) ) -
	                 ( ( afDimA[0] * afAbsAxisDot[2][0] + afDimA[2] * afAbsAxisDot[0][0] ) + ( afDimB[1] * afAbsAxisDot[1][2] + afDimB[2] * afAbsAxisDot[1][1] ) ) ) > 0.0f )
		return false;
	if( pkContactSet )
	{
		fScale = sqrtf( afAxisDot[2][0] * afAxisDot[2][0] + afAxisDot[0][0] * afAxisDot[0][0] );

		if( fScale > 0.0f )
		{
			fScale    = 1.0f / fScale;
			fOverlap *= fScale;

			if( fOverlap > fOverlapMax )
			{
				fOverlapMax = fOverlap;
				kNormal.Set( afAxisDot[2][0] * fScale, 0.0f, -afAxisDot[0][0] * fScale );

				if( fProj < 0.0f )
					kNormal = -kNormal;

				iDeepAxis   = 9;
			}
		}
	}

	//Axis Ay X By
	if( ( fOverlap = fabsf( ( fProj = afAxisDot[2][1] * afAxesADotDiff[0] - afAxisDot[0][1] * afAxesADotDiff[2] ) ) -
	                 ( ( afDimA[0] * afAbsAxisDot[2][1] + afDimA[2] * afAbsAxisDot[0][1] ) + ( afDimB[0] * afAbsAxisDot[1][2] + afDimB[2] * afAbsAxisDot[1][0] ) ) ) > 0.0f )
		return false;
	if( pkContactSet )
	{
		fScale = sqrtf( afAxisDot[2][1] * afAxisDot[2][1] + afAxisDot[0][1] * afAxisDot[0][1] );

		if( fScale > 0.0f )
		{
			fScale    = 1.0f / fScale;
			fOverlap *= fScale;

			if( fOverlap > fOverlapMax )
			{
				fOverlapMax = fOverlap;
				kNormal.Set( afAxisDot[2][1] * fScale, 0.0f, -afAxisDot[0][1] * fScale );

				if( fProj < 0.0f )
					kNormal = -kNormal;

				iDeepAxis   = 10;
			}
		}
	}

	//Axis Ay X Bz
	if( ( fOverlap = fabsf( ( fProj = afAxisDot[2][2] * afAxesADotDiff[0] - afAxisDot[0][2] * afAxesADotDiff[2] ) ) -
	                 ( ( afDimA[0] * afAbsAxisDot[2][2] + afDimA[2] * afAbsAxisDot[0][2] ) + ( afDimB[0] * afAbsAxisDot[1][1] + afDimB[1] * afAbsAxisDot[1][0] ) ) ) > 0.0f )
		return false;
	if( pkContactSet )
	{
		fScale = sqrtf( afAxisDot[2][2] * afAxisDot[2][2] + afAxisDot[0][2] * afAxisDot[0][2] );

		if( fScale > 0.0f )
		{
			fScale    = 1.0f / fScale;
			fOverlap *= fScale;

			if( fOverlap > fOverlapMax )
			{
				fOverlapMax = fOverlap;
				kNormal.Set( afAxisDot[2][2] * fScale, 0.0f, -afAxisDot[0][2] * fScale );

				if( fProj < 0.0f )
					kNormal = -kNormal;

				iDeepAxis   = 11;
			}
		}
	}

	//Axis Az X Bx
	if( ( fOverlap = fabsf( ( fProj = afAxisDot[0][0] * afAxesADotDiff[1] - afAxisDot[1][0] * afAxesADotDiff[0] ) ) -
	                 ( ( afDimA[0] * afAbsAxisDot[1][0] + afDimA[1] * afAbsAxisDot[0][0] ) + ( afDimB[1] * afAbsAxisDot[2][2] + afDimB[2] * afAbsAxisDot[2][1] ) ) ) > 0.0f )
		return false;
	if( pkContactSet )
	{
		fScale = sqrtf( afAxisDot[1][0] * afAxisDot[1][0] + afAxisDot[0][0] * afAxisDot[0][0] );

		if( fScale > 0.0f )
		{
			fScale    = 1.0f / fScale;
			fOverlap *= fScale;

			if( fOverlap > fOverlapMax )
			{
				fOverlapMax = fOverlap;
				kNormal.Set( -afAxisDot[1][0] * fScale, afAxisDot[0][0] * fScale, 0.0f );

				if( fProj < 0.0f )
					kNormal = -kNormal;

				iDeepAxis   = 12;
			}
		}
	}

	//Axis Az X By
	if( ( fOverlap = fabsf( ( fProj = afAxisDot[0][1] * afAxesADotDiff[1] - afAxisDot[1][1] * afAxesADotDiff[0] ) ) -
	                 ( ( afDimA[0] * afAbsAxisDot[1][1] + afDimA[1] * afAbsAxisDot[0][1] ) + ( afDimB[0] * afAbsAxisDot[2][2] + afDimB[2] * afAbsAxisDot[2][0] ) ) ) > 0.0f )
		return false;
	if( pkContactSet )
	{
		fScale = sqrtf( afAxisDot[1][1] * afAxisDot[1][1] + afAxisDot[0][1] * afAxisDot[0][1] );

		if( fScale > 0.0f )
		{
			fScale    = 1.0f / fScale;
			fOverlap *= fScale;

			if( fOverlap > fOverlapMax )
			{
				fOverlapMax = fOverlap;
				kNormal.Set( -afAxisDot[1][1] * fScale, afAxisDot[0][1] * fScale, 0.0f );

				if( fProj < 0.0f )
					kNormal = -kNormal;

				iDeepAxis   = 13;
			}
		}
	}

	//Axis Az X Bz
	if( ( fOverlap = fabsf( ( fProj = afAxisDot[0][2] * afAxesADotDiff[1] - afAxisDot[1][2] * afAxesADotDiff[0] ) ) -
	                 ( ( afDimA[0] * afAbsAxisDot[1][2] + afDimA[1] * afAbsAxisDot[0][2] ) + ( afDimB[0] * afAbsAxisDot[2][1] + afDimB[1] * afAbsAxisDot[2][0] ) ) ) > 0.0f )
		return false;
	if( pkContactSet )
	{
		fScale = sqrtf( afAxisDot[1][2] * afAxisDot[1][2] + afAxisDot[0][2] * afAxisDot[0][2] );

		if( fScale > 0.0f )
		{
			fScale    = 1.0f / fScale;
			fOverlap *= fScale;

			if( fOverlap > fOverlapMax )
			{
				fOverlapMax = fOverlap;
				kNormal.Set( -afAxisDot[1][2] * fScale, afAxisDot[0][2] * fScale, 0.0f );

				if( fProj < 0.0f )
					kNormal = -kNormal;

				iDeepAxis   = 14;
			}
		}
	}

	} // if( bParallellAxes )

	if( !pkContactSet )
		return true;

	Contact *pkContact = new Contact;

	if( iDeepAxis >= 6 )
	{
		//Edge from box zero touches and edge from box one

		//Rotate normal into world space
		kNormal = kRotMatA * kNormal;

		float fF[3];
		
		fF[0] = fabsf( kNormal * akAxesA[0] );
		fF[1] = fabsf( kNormal * akAxesA[1] );
		fF[2] = fabsf( kNormal * akAxesA[2] );

		if( fF[1] > fF[0] )
		{
			if( fF[2] > fF[1] )
				iDeepAxis = 2;
			else
				iDeepAxis = 1;
		}
		else
		{
			if( fF[2] > fF[0] )
				iDeepAxis = 2;
			else
				iDeepAxis = 0;
		}

		kNormal = akAxesA[ iDeepAxis ];

		if( kNormal * kDiff < 0.0f )
			kNormal = -kNormal;

/*
		//Calculate point on edge of boxes
		Vector3d kPoint0    = pkOBB0->GetTranslation();
		Vector3d kPoint1    = pkOBB1->GetTranslation();

		kPoint0 += ( akAxesA[0] * ( afDimA[0] * ( ( ( kNormal * akAxesA[0] ) > 0.0f ) ?  1.0f : -1.0f ) ) ) +
		           ( akAxesA[1] * ( afDimA[1] * ( ( ( kNormal * akAxesA[1] ) > 0.0f ) ?  1.0f : -1.0f ) ) ) +
		           ( akAxesA[2] * ( afDimA[2] * ( ( ( kNormal * akAxesA[2] ) > 0.0f ) ?  1.0f : -1.0f ) ) );

		kPoint1 += ( akAxesB[0] * ( afDimB[0] * ( ( ( kNormal * akAxesB[0] ) > 0.0f ) ? -1.0f :  1.0f ) ) ) +
		           ( akAxesB[1] * ( afDimB[1] * ( ( ( kNormal * akAxesB[1] ) > 0.0f ) ? -1.0f :  1.0f ) ) ) +
		           ( akAxesB[2] * ( afDimB[2] * ( ( ( kNormal * akAxesB[2] ) > 0.0f ) ? -1.0f :  1.0f ) ) );


		float fT0, fT1;

		int iAxis0 = ( iDeepAxis - 6 ) / 3;
		int iAxis1 = ( iDeepAxis - 6 ) % 3;

		Vector3d kDir0 = kRotMatA.GetColumn( iAxis0 );
		Vector3d kDir1 = kRotMatB.GetColumn( iAxis1 );

		//Get closest point t values between the two lines defined by
		//  l0 = kPoint0 + fT0 * kDir0
		//  l1 = kPoint1 + fT1 * kDir1

		ClosestPointLineLine( kPoint0, kDir0, kPoint1, kDir1, &fT0, &fT1 );

		kPoint0 += kDir0 * fT0;
		kPoint1 += kDir1 * fT1;

		pkContact->m_pkPoints[0] = ( kPoint0 + kPoint1 ) * 0.5f;
		pkContact->m_pfDepths[0] = -fOverlapMax;
		pkContact->m_uiNumPoints = 1;

		float fDir = kNormal * kDiff;

		if( ( !bInvertNormal && ( fDir > 0.0f ) ) || ( bInvertNormal && ( fDir < 0.0f ) ) )
			kNormal = -kNormal;

		pkContact->m_kNormal  = kNormal;

		pkContactSet->Add( pkContact );

		return true;
*/
	}

	float fDir = kNormal * kDiff;

	if( ( !bInvertNormal && ( fDir > 0.0f ) ) || ( bInvertNormal && ( fDir < 0.0f ) ) )
		pkContact->m_kNormal = -kNormal;
	else
		pkContact->m_kNormal = kNormal;

	if( iDeepAxis >= 3 )
		kNormal = -kNormal;


	// index 0 is box which reference face belongs to
	Matrix         *pkRotMat0;
	Matrix         *pkRotMat1;
	const Vector3d *pkPos0;
	const Vector3d *pkPos1;
	const Vector3d *pkDim0;
	const Vector3d *pkDim1;


	//Get "local" (in reference face coordinate space) x and y coordinate axes
	int iLocalX, iLocalY;
	int iNormal = ( iDeepAxis < 3 ) ? iDeepAxis : ( iDeepAxis - 3 );

	if( iNormal == 0 )
	{
		iLocalX = 1;
		iLocalY = 2;
	}
	else if( iNormal == 1 )
	{
		iLocalX = 0;
		iLocalY = 2;
	}
	else
	{
		iLocalX = 0;
		iLocalY = 1;
	}

	int iMajor, iOther[2];

	if( iDeepAxis < 3 )
	{
		//Face from box zero touches vertex/edge/face from box one
		pkRotMat0 = &kRotMatA;
		pkRotMat1 = &kRotMatB;
		pkPos0    = &pkOBB0->GetTranslation();
		pkPos1    = &pkOBB1->GetTranslation();
		pkDim0    = &pkOBB0->GetDim();
		pkDim1    = &pkOBB1->GetDim();

		// Get the normal axis for the interecting face (or one of them, in edge or vertex case)
		if( afAbsAxisDot[ iNormal ][1] > afAbsAxisDot[ iNormal ][0] )
		{
			if( afAbsAxisDot[ iNormal ][2] > afAbsAxisDot[ iNormal ][1] )
			{
				iMajor    = 2;
				iOther[0] = 0;
				iOther[1] = 1;
			}
			else
			{
				iMajor    = 1;
				iOther[0] = 0;
				iOther[1] = 2;
			}
		}
		else
		{
			if( afAbsAxisDot[ iNormal ][2] > afAbsAxisDot[ iNormal ][0] )
			{
				iMajor    = 2;
				iOther[0] = 0;
				iOther[1] = 1;
			}
			else
			{
				iMajor    = 0;
				iOther[0] = 1;
				iOther[1] = 2;
			}
		}
	}
	else
	{
		//Face from box one touches vertex/edge/face from box zero
		pkRotMat0 = &kRotMatB;
		pkRotMat1 = &kRotMatA;
		pkPos0    = &pkOBB1->GetTranslation();
		pkPos1    = &pkOBB0->GetTranslation();
		pkDim0    = &pkOBB1->GetDim();
		pkDim1    = &pkOBB0->GetDim();

		// Get the normal axis for the interecting face (or one of them, in edge or vertex case)
		if( afAbsAxisDot[1][ iNormal ] > afAbsAxisDot[0][ iNormal ] )
		{
			if( afAbsAxisDot[2][ iNormal ] > afAbsAxisDot[1][ iNormal ] )
			{
				iMajor    = 2;
				iOther[0] = 0;
				iOther[1] = 1;
			}
			else
			{
				iMajor    = 1;
				iOther[0] = 0;
				iOther[1] = 2;
			}
		}
		else
		{
			if( afAbsAxisDot[2][ iNormal ] > afAbsAxisDot[0][ iNormal ] )
			{
				iMajor    = 2;
				iOther[0] = 0;
				iOther[1] = 1;
			}
			else
			{
				iMajor    = 0;
				iOther[0] = 1;
				iOther[1] = 2;
			}
		}
	}


	//Calculate intersecting face center relative reference face
	Vector3d kCenter;
	Vector3d kOtherNormal = pkRotMat1->GetColumn( iMajor );

	if( ( kNormal * kOtherNormal ) > 0.0f )
		kCenter = ( (*pkPos1) - (*pkPos0) ) - ( kOtherNormal * (*pkDim1)[ iMajor ] );
	else
		kCenter = ( (*pkPos1) - (*pkPos0) ) + ( kOtherNormal * (*pkDim1)[ iMajor ] );



	//Project intersecting face onto reference face coordinate space (i.e a 2D coordinate quad)
	float aafQuad[4][2];
	float afCenter[2];

	afCenter[0] = kCenter * pkRotMat0->GetColumn( iLocalX );
	afCenter[1] = kCenter * pkRotMat0->GetColumn( iLocalY );

	float fXOffset[2], fYOffset[2];


	if( iDeepAxis >= 3 )
	{
		//Swizzle dot product data
		float fTemp     = afAxisDot[1][0];
		afAxisDot[1][0] = afAxisDot[0][1];
		afAxisDot[0][1] = fTemp;

		fTemp           = afAxisDot[2][0];
		afAxisDot[2][0] = afAxisDot[0][2];
		afAxisDot[0][2] = fTemp;

		fTemp           = afAxisDot[2][1];
		afAxisDot[2][1] = afAxisDot[1][2];
		afAxisDot[1][2] = fTemp;
	}

	fXOffset[0] = (*pkDim1)[ iOther[0] ] * afAxisDot[ iLocalX ][ iOther[0] ];
	fYOffset[0] = (*pkDim1)[ iOther[0] ] * afAxisDot[ iLocalY ][ iOther[0] ];
	fXOffset[1] = (*pkDim1)[ iOther[1] ] * afAxisDot[ iLocalX ][ iOther[1] ];
	fYOffset[1] = (*pkDim1)[ iOther[1] ] * afAxisDot[ iLocalY ][ iOther[1] ];

	aafQuad[0][0] = afCenter[0] - fXOffset[0] - fXOffset[1];
	aafQuad[0][1] = afCenter[1] - fYOffset[0] - fYOffset[1];
	aafQuad[1][0] = afCenter[0] - fXOffset[0] + fXOffset[1];
	aafQuad[1][1] = afCenter[1] - fYOffset[0] + fYOffset[1];
	aafQuad[2][0] = afCenter[0] + fXOffset[0] + fXOffset[1];
	aafQuad[2][1] = afCenter[1] + fYOffset[0] + fYOffset[1];
	aafQuad[3][0] = afCenter[0] + fXOffset[0] - fXOffset[1];
	aafQuad[3][1] = afCenter[1] + fYOffset[0] - fYOffset[1];


	//Get the dimensions of the reference face
	float afDim[2] = { (*pkDim0)[ iLocalX ], (*pkDim0)[ iLocalY ] };

	float aafClipped[8][2];

	int iNum = ClipRectPoly2D( afDim, (float*)aafQuad, 4, (float*)aafClipped );

	if( !iNum )
	{
		// ### FIXME: Wtf?
        //assert( iNum > 0 );

		delete pkContact;
		
		return false;
	}

	//Un-project points into reference face coordinate 3D space and calculate depths
	//(ignore points with positive depth)

	float fDet = 1.0f / ( afAxisDot[ iLocalX ][ iOther[0] ] * afAxisDot[ iLocalY ][ iOther[1] ] - afAxisDot[ iLocalX ][ iOther[1] ] * afAxisDot[ iLocalY ][ iOther[0] ] );

	afAxisDot[ iLocalX ][ iOther[0] ] *= fDet;
	afAxisDot[ iLocalX ][ iOther[1] ] *= fDet;
	afAxisDot[ iLocalY ][ iOther[0] ] *= fDet;
	afAxisDot[ iLocalY ][ iOther[1] ] *= fDet;

	for( int iPoint = 0; iPoint < iNum; ++iPoint )
	{
		float f0 =  afAxisDot[ iLocalY ][ iOther[1] ] * ( aafClipped[ iPoint ][0] - afCenter[0] ) - afAxisDot[ iLocalX ][ iOther[1] ] * ( aafClipped[ iPoint ][1] - afCenter[1] );
		float f1 = -afAxisDot[ iLocalY ][ iOther[0] ] * ( aafClipped[ iPoint ][0] - afCenter[0] ) + afAxisDot[ iLocalX ][ iOther[0] ] * ( aafClipped[ iPoint ][1] - afCenter[1] );

		Vector3d kPoint = kCenter + ( pkRotMat1->GetColumn( iOther[0] ) * f0 ) + ( pkRotMat1->GetColumn( iOther[1] ) * f1 );

		float fDot   = kNormal * kPoint;
		float fDepth = (*pkDim0)[ iNormal ] - fDot;

		// ### TEST: clamp very small negative depths to zero
		if( fDepth > -0.005f )
		{
			if( fDepth < 0.0f )
				fDepth = 0.0f;

			pkContact->m_pkPoints[ pkContact->m_uiNumPoints   ] = kPoint + *pkPos0;
			pkContact->m_pfDepths[ pkContact->m_uiNumPoints++ ] = fDepth;
		}
	}

	if( !pkContact->m_uiNumPoints )
	{
		//Possible due to parallell axes threshold and floating point inaccuracy causing non-collisions to slip through
		delete pkContact;
		return false;
	}

	pkContact->Sort();

	pkContactSet->Add( pkContact );

    return true;
}


NEOENGINE_API bool Intersection( OBB *pkOBB, Sphere *pkSphere, ContactSet *pkContactSet, bool bInvertNormal )
{
	Vector3d kDiff = pkSphere->GetTranslation() - pkOBB->GetTranslation();

	Matrix kRotMat( pkOBB->GetRotation() );

	const Vector3d &rkOBBCenter    = pkOBB->GetTranslation();
	const Vector3d &rkSphereCenter = pkSphere->GetTranslation();
	const float    &rfRadius       = pkSphere->m_fRadius;
	const Vector3d &rkOBBDim       = pkOBB->GetDim();

	Vector3d akAxes[] = { kRotMat.GetColumn(0), kRotMat.GetColumn(1), kRotMat.GetColumn(2) };

	float fSide[3] = { 1.0f, 1.0f, 1.0f };
	float fS[3]    = { 0.0f, 0.0f, 0.0f };
	float fD       = 0.0f;

	if( kDiff * akAxes[0] < 0.0f )
	{
		fS[0] = ( rkSphereCenter - ( rkOBBCenter - ( akAxes[0] * rkOBBDim.x ) ) ) * akAxes[0];

		if( fS[0] <= -rfRadius )
			return false;
		else if( fS[0] < 0.0f )
			fD += fS[0] * fS[0];

		fS[0] += rfRadius;
	}
	else
	{
		fS[0] = ( ( rkOBBCenter + ( akAxes[0] * rkOBBDim.x ) ) - rkSphereCenter ) * akAxes[0];

		if( fS[0] <= -rfRadius )
			return false;
		else if( fS[0] < 0.0f )
			fD += fS[0] * fS[0];

		fS[0] += rfRadius;
		fSide[0] = -1.0f;
	}

	if( kDiff * akAxes[1] < 0.0f )
	{
		fS[1] = ( rkSphereCenter - ( rkOBBCenter - ( akAxes[1] * rkOBBDim.y ) ) ) * akAxes[1];

		if( fS[1] <= -rfRadius )
			return false;
		else if( fS[1] < 0.0f )
			fD += fS[1] * fS[1];

		fS[1] += rfRadius;
	}
	else
	{
		fS[1] = ( ( rkOBBCenter + ( akAxes[1] * rkOBBDim.y ) ) - rkSphereCenter ) * akAxes[1];

		if( fS[1] <= -rfRadius )
			return false;
		else if( fS[1] < 0.0f )
			fD += fS[1] * fS[1];

		fS[1] += rfRadius;
		fSide[1] = -1.0f;
	}

	if( kDiff * akAxes[2] < 0.0f )
	{
		fS[2] = ( rkSphereCenter - ( rkOBBCenter - ( akAxes[2] * rkOBBDim.z ) ) ) * akAxes[2];

		if( fS[2] <= -rfRadius )
			return false;
		else if( fS[2] < 0.0f )
			fD += fS[2] * fS[2];

		fS[2] += rfRadius;
	}
	else
	{
		fS[2] = ( ( rkOBBCenter + ( akAxes[2] * rkOBBDim.z ) ) - rkSphereCenter ) * akAxes[2];

		if( fS[2] <= -rfRadius )
			return false;
		else if( fS[2] < 0.0f )
			fD += fS[2] * fS[2];

		fS[2] += rfRadius;
		fSide[2] = -1.0f;
	}

	if( fD <= ( rfRadius * rfRadius ) )
	{
		//We've got intersection
		if( !pkContactSet )
			return true;

		int iMinAxis = 0;
		int iSecAxis = 1;
		int iMajAxis = 2;

		if( fS[1] < fS[0] )
		{
			if( fS[2] < fS[1] )
			{
				iMinAxis = 2;
				iSecAxis = 1;
				iMajAxis = 0;
			}
			else
			{
				iMinAxis = 1;
				if( fS[2] < fS[0] )
				{
					iSecAxis = 2;
					iMajAxis = 0;
				}
				else
				{
					iSecAxis = 0;
					iMajAxis = 2;
				}
			}
		}
		else
		{
			if( fS[2] < fS[0] )
			{
				iMinAxis = 2;
				iSecAxis = 0;
				iMajAxis = 1;
			}
			else
			{
				iMinAxis = 0;
				if( fS[2] < fS[1] )
				{
					iSecAxis = 2;
					iMajAxis = 1;
				}
				else
				{
					iSecAxis = 1;
					iMajAxis = 2;
				}
			}
		}

		Contact *pkContact = new Contact;

		pkContact->m_kNormal = akAxes[iMinAxis] * fSide[iMinAxis];

		Vector3d kPoint = rkSphereCenter + ( pkContact->m_kNormal * pkSphere->m_fRadius );
		Vector3d kPointDiff = kPoint - rkOBBCenter;
		
		if( ( fabsf( kPointDiff * akAxes[iSecAxis] ) < rkOBBDim[iSecAxis] ) && ( fabsf( kPointDiff * akAxes[iMajAxis] ) < rkOBBDim[iMajAxis] ) )
		{
			//Inside cube
			kPoint -= ( pkContact->m_kNormal * ( fS[iMinAxis] * 0.5f ) );

			pkContact->m_pkPoints[0] = kPoint;
			pkContact->m_pfDepths[0] = fS[iMinAxis];
			pkContact->m_uiNumPoints = 1;
		}
		else
		{
			//Outside cube
			kPoint = rkOBBCenter + ( akAxes[iMinAxis] * ( -fSide[iMinAxis] * rkOBBDim[iMinAxis] ) ) + ( akAxes[iSecAxis] * ( -fSide[iSecAxis] * rkOBBDim[iSecAxis] ) );

			Plane kPlane( akAxes[iMajAxis], rkSphereCenter );

			float fDist = kPlane.Distance( kPoint );

			if( fDist < -rkOBBDim[iMajAxis] )
				fDist = -rkOBBDim[iMajAxis];
			else if( fDist > rkOBBDim[iMajAxis] )
				fDist = rkOBBDim[iMajAxis];

			kPoint -= akAxes[iMajAxis] * fDist;

			Vector3d kNormal = kPoint - rkSphereCenter;

			float fLen = kNormal.Len();

			float fDepth = pkSphere->m_fRadius - fLen;

			kNormal *= 1.0f / fLen;

			kPoint += kNormal * ( fDepth * 0.5f );

			pkContact->m_pkPoints[0] = kPoint;
			pkContact->m_pfDepths[0] = fDepth;
			pkContact->m_uiNumPoints = 1;
			pkContact->m_kNormal = kNormal;
		}

		if( bInvertNormal )
			pkContact->m_kNormal = -pkContact->m_kNormal;

		pkContactSet->Add( pkContact );

		return true;
	}

	return false;

}


NEOENGINE_API bool Intersection( OBB *pkOBB, Capsule *pkCapsule, ContactSet *pkContactSet, bool bInvertNormal )
{
	assert( !"Not implemented" );
	return false;
}


NEOENGINE_API bool Intersection( OBB *pkOBB, Frustum *pkFrustum )
{
	const Vector3d &rkOBBPos = pkOBB->GetTranslation();
	const Vector3d &rkOBBDim = pkOBB->GetDim();
	Matrix kRotMat( pkOBB->GetRotation() );

	Vector3d kOBBDimRot = kRotMat * rkOBBDim;

	kOBBDimRot.x = fabsf( kOBBDimRot.x );
	kOBBDimRot.y = fabsf( kOBBDimRot.y );
	kOBBDimRot.z = fabsf( kOBBDimRot.z );

	Vector3d kTest;

	for( int iPlane = 0; iPlane < 6; ++iPlane )
	{
		kTest = rkOBBPos;

		//Test most negative point
		kTest.x += ( pkFrustum->m_akPlanes[iPlane].m_kNormal.x > 0.0f ) ? (-kOBBDimRot.x) : kOBBDimRot.x;
		kTest.y += ( pkFrustum->m_akPlanes[iPlane].m_kNormal.y > 0.0f ) ? (-kOBBDimRot.y) : kOBBDimRot.y;
		kTest.z += ( pkFrustum->m_akPlanes[iPlane].m_kNormal.z > 0.0f ) ? (-kOBBDimRot.z) : kOBBDimRot.z;

		if( pkFrustum->m_akPlanes[iPlane].Distance( kTest ) > 0.0f )
			return false;
	}

	return true;
}


NEOENGINE_API bool Intersection( Sphere *pkSphere0, Sphere *pkSphere1, ContactSet *pkContactSet, bool bInvertNormal )
{
	float fMaxDistance = pkSphere0->m_fRadius + pkSphere1->m_fRadius;
	Vector3d kDiff = pkSphere0->GetTranslation() - pkSphere1->GetTranslation();
	float fLen = kDiff.Len2();

	if( fLen <= ( fMaxDistance * fMaxDistance ) )
	{
		if( !pkContactSet )
			return true;

		fLen = sqrtf( fLen );

		float fOverlap = fMaxDistance - fLen;

		kDiff *= 1.0f / fLen;

		Vector3d kPoint = pkSphere1->GetTranslation() + kDiff * ( pkSphere1->m_fRadius - fOverlap * 0.5f );

		Contact *pkColl = new Contact;

		pkColl->m_kNormal = !bInvertNormal ? kDiff : -kDiff;
		pkColl->m_pfDepths[0] = fOverlap;
		pkColl->m_pkPoints[0] = kPoint;
		pkColl->m_uiNumPoints = 1;

		pkContactSet->Add( pkColl );

		return true;
	}

	return false;
}


NEOENGINE_API bool Intersection( Sphere *pkSphere, Capsule *pkCapsule, ContactSet *pkContactSet, bool bInvertNormal )
{
	assert( !"Not implemented" );
	return false;
}


NEOENGINE_API bool Intersection( Sphere *pkSphere, Frustum *pkFrustum )
{
	//Check distance to all frustum planes (this simple algorithm will generate
	// false Intersections at corners of frustum, but it's acceptable for now)

	Vector3d rkWorldPos = pkSphere->GetTranslation();
	float fRadii          = pkSphere->m_fRadius;

	for( int iPlane = 0; iPlane < 6; ++iPlane )
		if( pkFrustum->m_akPlanes[iPlane].Distance( rkWorldPos ) > fRadii )
			return false;

	return true;
}


NEOENGINE_API bool Intersection( Capsule *pkCapsule0, Capsule *pkCapsule1, ContactSet *pkContactSet, bool bInvertNormal )
{
	assert( !"Not implemented" );
	return false;
}


NEOENGINE_API bool Intersection( Capsule *pkCapsule, Frustum *pkFrustum )
{
	assert( !"Not implemented" );
	return false;
}

/*
*  Algorithm based on MllerTrumbore97
*  http://www.acm.org/jgt/papers/MollerTrumbore97/index.html
*  Original algorithm is implemented in C code.
*/
NEOENGINE_API bool Intersection( const Ray &rkRay, const Vector3d &rkV0, const Vector3d &rkV1, const Vector3d &rkV2, ContactSet *pkContactSet, const Vector3d &rkNormal )
{
	Vector3d kTVec, kQVec, kPoint;
	float fU, fV, fT, fDet;
	
	Vector3d kEdge1 = rkV1 - rkV0;
	Vector3d kEdge2 = rkV2 - rkV0;
	
	Vector3d kPVec = rkRay.m_kDirection % kEdge2;
	
	fDet = kEdge1 * kPVec;

	if( fDet < EPSILON )
		return false;

	kTVec = rkRay.m_kOrigin - rkV0;
	fU    = kTVec * kPVec;

	if( ( fU < 0.0f ) || ( fU > fDet ) )
		return false;

	kQVec = kTVec % kEdge1;
	fV  = rkRay.m_kDirection * kQVec;
	
	if( ( fV < 0.0f ) || ( ( fU + fV ) > fDet ) )
		return false;

	fT = ( kEdge2 * kQVec ) / fDet;

	if( fT < 0.0f )
		return false;

	kPoint = rkRay.m_kOrigin + ( rkRay.m_kDirection * fT );

/* #ifndef CULLING
	if (fDet > -EPSILON && fDet < EPSILON)
		return false;
	fInvDet = 1.0f / fDet;
	kTVec = rkRay.m_kOrigin - rkV0;
	kPoint.y = (kTVec * kPVec) * fInvDet;
	if (kPoint.y < 0.0f || kPoint.y > 1.0f)
		return false;
	kQVec = kTVec % kEdge1;
	kPoint.z = (rkRay.m_kDirection * kQVec) * fInvDet;
	if (kPoint.z < 0.0f || (kPoint.y + kPoint.z) > 1.0f)
		return false;
	kPoint.x = (kEdge2 * kQVec) * fInvDet;
#endif */

	if( !pkContactSet )
		return true;

	Contact *pkContact = new Contact;

	pkContact->m_pkPoints[0] = kPoint;
	pkContact->m_pfDepths[0] = fT;
	pkContact->m_uiNumPoints = 1;
	pkContact->m_kNormal = kEdge1 % kEdge2;

	if( ( pkContact->m_kNormal * rkRay.m_kDirection ) >= 0.0f)
		pkContact->m_kNormal *= -1.0f;

	pkContact->m_kNormal.Normalize();

	pkContactSet->Add( pkContact );

	return true;
}


inline void SliceIntersect(
	const Vector3d &rkV0, const Vector3d &rkV1, const Vector3d &rkV2, 
	float fVP0, float fVP1, float fVP2,
	float fdV0, float fdV1, float fdV2, 
	float *pfISect0, float *pfISect1, Vector3d *pkI0, Vector3d *pkI1 )
{
	float fTmp;
	Vector3d kDiff;

	fTmp      = fdV0 / ( fdV0 - fdV1 );

	*pfISect0 = fVP0 + ( fVP1 - fVP0 ) * fTmp;
	kDiff     = ( rkV1 - rkV0 ) * fTmp;

	*pkI0     = rkV0 + kDiff;


	fTmp      = fdV0 / ( fdV0 - fdV2 );

	*pfISect1 = fVP0 + ( fVP2 - fVP0 ) * fTmp;
	kDiff     = ( rkV2 - rkV0 ) * fTmp;

	*pkI1     = rkV0 + kDiff;
}


inline int SliceComputeIntervals( 
	const Vector3d &rkV0, const Vector3d &rkV1, const Vector3d &rkV2, 
	float fVP0, float fVP1, float fVP2, float fdV0, float fdV1, float fdV2,
	float fdV0dV1, float fdV0dV2, float *pfISect0, float *pfISect1, Vector3d *pkI0, Vector3d *pkI1 )
{
	if( fdV0dV1 > 0.0f)
	{
		// Now fdV0dV2 <= 0.0f ( fdV0, fdV1 are on the same side, fdV2 on the other or on the plane 
		SliceIntersect( rkV2, rkV0, rkV1, fVP2, fVP0, fVP1, fdV2, fdV0, fdV1, pfISect0, pfISect1, pkI0, pkI1 );
	}
	else if( fdV0dV2 > 0.0f )
	{
		// Now fdV0dV1 <= 0.0f
		SliceIntersect( rkV1, rkV0, rkV2, fVP1, fVP0, fVP2, fdV1, fdV0, fdV2, pfISect0, pfISect1, pkI0, pkI1 );
	}
	else if( ( fdV1 * fdV2 > 0.0f ) || ( fdV0 != 0.0f ) )
	{
		// Now either fdV0dV1 <= 0.0f or fdV0 != 0.0f
		SliceIntersect( rkV0, rkV1, rkV2, fVP0, fVP1, fVP2, fdV0, fdV1, fdV2, pfISect0, pfISect1, pkI0, pkI1 );
	}
	else if( fdV1 != 0.0f )
	{
		SliceIntersect( rkV1, rkV0, rkV2, fVP1, fVP0, fVP2, fdV1, fdV0, fdV2, pfISect0, pfISect1, pkI0, pkI1 );
	}
	else if( fdV2 != 0.0f )
	{
		SliceIntersect( rkV2, rkV0, rkV1, fVP2, fVP0, fVP1, fdV2, fdV0, fdV1, pfISect0, pfISect1, pkI0, pkI1 );
	}
	else
		return 1;

	return 0;
}


NEOENGINE_API bool Slice( const Vector3d &rkV0, const Vector3d &rkV1, const Vector3d &rkV2, const Vector3d &rkU0, const Vector3d &rkU1, const Vector3d &rkU2, Vector3d *pkI0, Vector3d *pkI1, int *piCoplanar )
{
	Vector3d kEdge0, kEdge1;
	Plane    kPlane0, kPlane1;

	float fdU0, fdU1, fdU2;
	float fdV0, fdV1, fdV2;

	float fdU0dU1, fdU0dU2, fdV0dV1, fdV0dV2;

	float fVP0, fVP1, fVP2, fUP0, fUP1, fUP2;

	Vector3d kDir;

	float fMax, fTemp;

	int iIndex;
	
	float fISect0[2], fISect1[2];

	Vector3d kISectPointV0, kISectPointV1;
	Vector3d kISectPointU0, kISectPointU1;

	int iCoplanar;

	int iSmallest0, iSmallest1;



	// Calculate edges and plane equation of triangle V
	kEdge0 = rkV1 - rkV0;
	kEdge1 = rkV2 - rkV0;

	kPlane0.m_kNormal   = kEdge0 % kEdge1;
	kPlane0.m_fDistance = -( kPlane0.m_kNormal * rkV0 );

	// Calculate distances of U vertices to V plane
	fdU0 = kPlane0.Distance( rkU0 );
	fdU1 = kPlane0.Distance( rkU1 );
	fdU2 = kPlane0.Distance( rkU2 );

	/*
	if( fabs( fdU0 ) < EPSILON ) fdU0 = 0.0f;
	if( fabs( fdU1 ) < EPSILON ) fdU1 = 0.0f;
	if( fabs( fdU2 ) < EPSILON ) fdU2 = 0.0f;
	*/

	fdU0dU1 = fdU0 * fdU1;
	fdU0dU2 = fdU0 * fdU2;

	if( ( fdU0dU1 > 0.0f ) && ( fdU0dU2 > 0.0f ) )
		return false; //No intersection


	// Calculate edges and plane equation of triangle V
	kEdge0 = rkU1 - rkU0;
	kEdge1 = rkU2 - rkU0;

	kPlane1.m_kNormal   = kEdge0 % kEdge1;
	kPlane1.m_fDistance = -( kPlane1.m_kNormal * rkU0 );

	// Calculate distances of V vertices to U plane
	fdV0 = kPlane1.Distance( rkV0 );
	fdV1 = kPlane1.Distance( rkV1 );
	fdV2 = kPlane1.Distance( rkV2 );

	/*
	if( fabs( fdV0 ) < EPSILON ) fdV0 = 0.0f;
	if( fabs( fdV1 ) < EPSILON ) fdV1 = 0.0f;
	if( fabs( fdV2 ) < EPSILON ) fdV2 = 0.0f;
	*/

	fdV0dV1 = fdV0 * fdV1;
	fdV0dV2 = fdV0 * fdV2;
        
	if( ( fdV0dV1 > 0.0f ) && ( fdV0dV2 > 0.0f ) )
		return false; //No intersection


	// Calculate direction of intersection line
	kDir = kPlane0.m_kNormal % kPlane1.m_kNormal;

	// Calculate index of largest component of intersection line direction
	fMax   = fabsf( kDir.x );
	iIndex = 0;

	if( ( fTemp = fabsf( kDir.y ) ) > fMax ) { fMax = fTemp; iIndex = 1; }
	if( ( fTemp = fabsf( kDir.z ) ) > fMax ) { fMax = fTemp; iIndex = 2; }


	// Calculate simplified projection of V and U vertices on intersection line
	fVP0 = rkV0[ iIndex ];
	fVP1 = rkV1[ iIndex ];
	fVP2 = rkV2[ iIndex ];

	fUP0 = rkU0[ iIndex ];
	fUP1 = rkU1[ iIndex ];
	fUP2 = rkU2[ iIndex ];


	// Calculate interval for triangle V
	iCoplanar = SliceComputeIntervals( rkV0, rkV1, rkV2, fVP0, fVP1, fVP2, fdV0, fdV1, fdV2, fdV0dV1, fdV0dV2, &fISect0[0], &fISect0[1], &kISectPointV0, &kISectPointV1 );

	if( iCoplanar )
	{
		if( piCoplanar )
			*piCoplanar = iCoplanar;

		return false;
	}
	
	// Calculate interval for triangle U
	SliceComputeIntervals( rkU0, rkU1, rkU2, fUP0, fUP1, fUP2, fdU0, fdU1, fdU2, fdU0dU1, fdU0dU2, &fISect1[0], &fISect1[1], &kISectPointU0, &kISectPointU1 );

	if( fISect0[0] > fISect0[1] )
	{
		fTemp      = fISect0[0];
		fISect0[0] = fISect0[1];
		fISect0[1] = fTemp;

		iSmallest0 = 1;
	}
	else
		iSmallest0 = 0;

	if( fISect1[0] > fISect1[1] )
	{
		fTemp      = fISect1[0];
		fISect1[0] = fISect1[1];
		fISect1[1] = fTemp;

		iSmallest1 = 1;
	}
	else
		iSmallest1 = 0;


	if( ( fISect0[1] < fISect1[0] ) || ( fISect1[1] < fISect0[0] ) )
		return false;


	// Now we know triangles intersect

	if( fISect1[0] < fISect0[0] )
	{
		if( iSmallest0 == 0 )
			*pkI0 = kISectPointV0;
		else
			*pkI0 = kISectPointV1;
		
		if( fISect1[1] < fISect0[1] )
		{
			if( iSmallest1 == 0 )
				*pkI1 = kISectPointU1;
			else
				*pkI1 = kISectPointU0;
		}
		else
		{
			if( iSmallest0 == 0 )
				*pkI1 = kISectPointV1;
			else
				*pkI1 = kISectPointV0;
		}
	}
	else
	{
		if( iSmallest1 == 0 )
			*pkI0 = kISectPointU0;
		else
			*pkI0 = kISectPointU1;

		if( fISect1[1] > fISect0[1] )
		{
			if( iSmallest0 == 0 )
				*pkI1 = kISectPointV1;
			else
				*pkI1 = kISectPointV0;
		}
		else
		{
			if( iSmallest1 == 0 )
				*pkI1 = kISectPointU1;
			else
				*pkI1 = kISectPointU0;
		}
	}

	return true;
}




NEOENGINE_API int ClipRectPoly2D( float *pfRect, float *pfPoly, int iNum, float *pfClipped )
{
	assert( ( iNum == 3 ) || ( iNum == 4 ) );

	int   iNumQ = iNum;
	int   iNumR;
	float afBuffer[16];
	int   iMax = ( iNum == 3 ) ? 5 : 8;

	float *pfQ = pfPoly;
	float *pfR = pfClipped;

	//Axis denoted as 0 == x and 1 == y
	for( int iDir = 0; iDir < 2; ++iDir )
	{
		for( int iSign = -1; iSign <= 1; iSign +=2 )
		{
			//Clip pfQ along the axis iDir = ( iSign * pfRect[ iDir ] )
			float *pfCurQ = pfQ;
			float *pfCurR = pfR;

			iNumR = 0;

			for( int i = iNumQ; i > 0; --i )
			{
				//Loop all points in pfQ and all edges in pfQ
				if( iSign * pfCurQ[ iDir ] < pfRect[ iDir ] )
				{
					//Point inside clip line
					*pfCurR++ = pfCurQ[0];
					*pfCurR++ = pfCurQ[1];

					++iNumR;

					if( iNumR == iMax )
					{
						pfQ = pfR;
						goto clipdone;
					}
				}

				float *pfNextQ = ( i > 1 ) ? ( pfCurQ + 2 ) : pfQ;

				if( ( iSign * pfCurQ[ iDir ] < pfRect[ iDir ] ) ^ ( iSign * pfNextQ[ iDir ] < pfRect[ iDir ] ) )
				{
					//This line should be clipped
					pfCurR[ 1 - iDir ] = pfCurQ[ 1 - iDir ] + ( pfNextQ[ 1 - iDir ] - pfCurQ[ 1 - iDir ] ) / ( pfNextQ[ iDir ] - pfCurQ[ iDir ] ) * ( iSign * pfRect[ iDir ] - pfCurQ[ iDir ] );
					pfCurR[ iDir     ] = iSign * pfRect[ iDir ];

					pfCurR += 2;
					++iNumR;

					if( iNumR == iMax )
					{
						pfQ = pfR;
						goto clipdone;
					}
				}

				pfCurQ += 2;
			}

			pfQ = pfR;
			pfR = ( pfQ == pfClipped ) ? afBuffer : pfClipped;

			iNumQ = iNumR;
		}
	}

	clipdone:

	if( pfQ != pfClipped )
		fmemcpy( pfClipped, pfQ, iNumR * 2 * sizeof( float ) );

	return iNumR;
}


NEOENGINE_API int ClipPolyPoly2D( float *pfRefPoly, float *pfRefPolyNormals, int iNumRef, float *pfPoly, int iNum, float *pfClipped )
{
	assert( ( iNumRef == 3 ) || ( iNumRef == 4 ) );
	assert( ( iNum == 3 ) || ( iNum == 4 ) );

	int   iNumQ = iNum;
	int   iNumR = 0;
	float afBuffer[16];

	float *pfQ = pfPoly;
	float *pfR = pfClipped;
	float *pfE = pfRefPoly;
	float *pfN = pfRefPolyNormals;

	for( int iEdge = iNumRef; iEdge > 0; --iEdge )
	{
		//Clip pfQ along the edge
		float *pfCurQ = pfQ;
		float *pfCurR = pfR;

		iNumR = 0;

		for( int i = iNumQ; i > 0; --i )
		{
			//Loop all points in pfQ and all edges in pfQ
			float fRel[2] = { pfCurQ[0] - pfE[0], pfCurQ[1] - pfE[1] };
			float fCurDot = fRel[0] * pfN[0] + fRel[1] * pfN[1];

			if( fCurDot <= 0.0f )
			{
				//Point inside clip line
				*pfCurR++ = pfCurQ[0];
				*pfCurR++ = pfCurQ[1];

				++iNumR;
			}

			float *pfNextQ = ( i > 1 ) ? ( pfCurQ + 2 ) : pfQ;

			float fNextRel[2] = { pfNextQ[0] - pfE[0], pfNextQ[1] - pfE[1] };

			float fNextDot = fNextRel[0] * pfN[0] + fNextRel[1] * pfN[1];

			if( ( fCurDot > 0.0f ) ^ ( fNextDot > 0.0f ) )
			{
				float *pfNextE = ( iEdge > 1 ) ? ( pfE + 2 ) : pfRefPoly;

				float fDir0[2] = { pfNextQ[0] - pfCurQ[0], pfNextQ[1] - pfCurQ[1] };
				float fDir1[2] = { pfNextE[0] - pfE[0], pfNextE[1] - pfE[1] };

				//This line should be clipped
				float fT = ( ( fDir1[0] * -fRel[1] ) - ( fDir1[1] * -fRel[0] ) ) / ( fDir1[0] * fDir0[1] - fDir1[1] * fDir0[0] );

				*pfCurR++ = pfCurQ[0] + fDir0[0] * fT;
				*pfCurR++ = pfCurQ[1] + fDir0[1] * fT;

				++iNumR;
			}

			pfCurQ += 2;
		}

		pfQ = pfR;
		pfR = ( pfQ == pfClipped ) ? afBuffer : pfClipped;

		iNumQ = iNumR;

		pfE += 2;
		pfN += 2;
	}

	// pfQ is now the last used destination buffer!
	if( pfQ != pfClipped )
		fmemcpy( pfClipped, pfQ, iNumR * 2 * sizeof( float ) );

	return iNumR;
}


void NEOENGINE_API ClosestPointLineLine( const Vector3d &rkP0, const Vector3d &rkD0, const Vector3d &rkP1, const Vector3d &rkD1, float *pfT0, float *pfT1 )
{
	Vector3d kDiff = rkP0 - rkP1;

	float fDir00    = rkD0 * rkD0;
	float fDir01    = -( rkD0 * rkD1 );
	float fDir11    = rkD1 * rkD1;
	float fDiffDir0 = kDiff * rkD0;

	float fDet      = fabsf( fDir00 * fDir11 - fDir01 * fDir01 );

	if( fDet > EPSILON )
	{
		float fDiffDir1 = -( kDiff * rkD1 );

		fDet = 1.0f / fDet;

		*pfT0 = ( fDir01 * fDiffDir1 - fDir11 * fDiffDir0 ) * fDet;
		*pfT1 = ( fDir01 * fDiffDir0 - fDir00 * fDiffDir1 ) * fDet;
	}
	else
	{
		//Parallell, any closest pair of points will do
		*pfT0 = -( fDiffDir0 / fDir00 );
		*pfT1 = 0.0f;
	}
}



};


