/***************************************************************************
                    config.cpp  -  Configuration repository
                             -------------------
    begin                : Wed Jul 03 2002
    copyright            : (C) 2002 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, config.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "config.h"
#include "console.h"
#include "file.h"
#include "logstream.h"
#include "strutil.h"

#include <algorithm>

using namespace std;


namespace NeoEngine
{


#ifndef HAVE_STRTOF
#  define strtof(a,b) ((float)atof((a)))
#endif


//sorting comparison method
inline bool CompareConfigValue( HashTableNode<ConfigValue> *pkNode0, HashTableNode<ConfigValue> *pkNode1 )
{
	return( pkNode0->m_strKey.m_strData < pkNode1->m_strKey.m_strData );
}


Config::Config( Console *pkConsole ) :
	LoadableEntity(),
	ConsoleCmdCallback()
{
	m_pkHashTable = new HashTable<ConfigValue>;

	if( !pkConsole )
		pkConsole = Core::Get()->GetConsole();

	if( pkConsole )
	{
		pkConsole->RegisterCommand( "set", this );
		pkConsole->RegisterCommand( "showconfig", this );
	}
}


Config::~Config()
{
	if( m_pkHashTable )
	{
		vector< HashTableNode<ConfigValue>* > vpkNodes;

		m_pkHashTable->GetAllNodes( &vpkNodes );

		for( unsigned int i = 0; i < vpkNodes.size(); ++i )
			delete vpkNodes[i]->m_pkData;

		delete m_pkHashTable, m_pkHashTable = 0;
	}
}


void Config::ConsoleCmd( const HashString &rstrCmd, const HashString &rstrArgs )
{
	if( rstrCmd == "set" )
	{
		int iSpace;

		string strVar = rstrArgs.m_strData.substr( 0, ( iSpace = rstrArgs.m_strData.find_first_of( ' ' ) ) );
		string strVal = rstrArgs.m_strData.substr( iSpace + 1, rstrArgs.m_strData.length() );

		SetValue( strVar, strVal );
	}
	else if( rstrCmd == "showconfig" )
		Print();
}


void Config::Print()
{
	vector< HashTableNode<ConfigValue>* > vpkNodes;

	if( m_pkHashTable )
		m_pkHashTable->GetAllNodes( &vpkNodes );

	std::sort( vpkNodes.begin(), vpkNodes.end(), CompareConfigValue );

	for( unsigned int i = 0; i < vpkNodes.size(); ++i )
		neolog << LogLevel( INFO ) << vpkNodes[i]->m_strKey.m_strData << ": " << vpkNodes[i]->m_pkData->m_strValue << endl;
}


bool Config::LoadNode( unsigned int uiFlags )
{
	if( !m_pkFile || !m_pkFile->IsValid() )
		return false;

	m_pkFile->SetBinary( false );

	string strToken;
	string strValue;

	//Get name
	while( true )
	{
		strToken = strValue = "";

		//Read token
		if( !( *m_pkFile >> strToken ) )
			break;

		if( ( strToken[0] == '#' ) || ( strToken[0] == ';' ) )
		{
			//Comment line, skip
			m_pkFile->GetLine( 0, 0 );
			continue;
		}

		*m_pkFile >> strValue;

		if( strValue[0] == '"' )
		{
			if( strValue.length() == 1 )
			{
				*m_pkFile >> strValue;
				strValue = "\"" + strValue;
			}

			while( strValue[ strValue.length() - 1 ] != '"' )
			{
				string strVal;

				*m_pkFile >> strVal;

				strValue += " " + strVal;
			}
				
			SetValue( strToken, Strip( strValue, "\"" ) );
		}
		else if( strValue == "true" )
			SetValue( strToken, true );
		else if( strValue == "false" )
			SetValue( strToken, false );
		else if( strValue.find( "." ) < strValue.length() )
			SetValue( strToken, strtof( strValue.c_str(), 0 ) );
		else
			SetValue( strToken, int( strtol( strValue.c_str(), 0, 10 ) ) );

		//Skip rest of line
		m_pkFile->GetLine( 0, 0, '\n' );
	}

	return true;
}


void Config::SetValue( const HashString &rstrKey, const HashString &rstrValue )
{
	if( !m_pkHashTable )
		return;

	ConfigValue *pVal = m_pkHashTable->Find( rstrKey );

	if( !pVal )
	{
		pVal = new ConfigValue();

		m_pkHashTable->Insert( rstrKey, pVal );
	}

	pVal->m_strValue = rstrValue;
	pVal->m_iValue   = int( strtol( rstrValue.c_str(), 0, 10 ) );
	pVal->m_fValue   = strtof( rstrValue.c_str(), 0 );
	pVal->m_bValue   = ( ( rstrValue == "true" ) || ( rstrValue == "on" ) ) ? true : ( pVal->m_iValue ? true : false );

	for( unsigned int i = 0; i < m_vpkCallbacks.size(); ++i )
		m_vpkCallbacks[i]->ConfigValueChange( rstrKey );
}


void Config::SetValue( const HashString &rstrKey, float fValue )
{
	if( !m_pkHashTable )
		return;

	ConfigValue *pVal = m_pkHashTable->Find( rstrKey );

	if( !pVal )
	{
		pVal = new ConfigValue();

		m_pkHashTable->Insert( rstrKey, pVal );
	}

	char szVal[128];
	snprintf( szVal, 128, "%.4f", fValue );

	pVal->m_strValue = szVal;
	pVal->m_iValue   = int( fValue );
	pVal->m_fValue   = fValue;
	pVal->m_bValue   = pVal->m_iValue ? true : false;

	for( unsigned int i = 0; i < m_vpkCallbacks.size(); ++i )
		m_vpkCallbacks[i]->ConfigValueChange( rstrKey );
}


void Config::SetValue( const HashString &rstrKey, int iValue )
{
	if( !m_pkHashTable )
		return;

	ConfigValue *pVal = m_pkHashTable->Find( rstrKey );

	if( !pVal )
	{
		pVal = new ConfigValue();

		m_pkHashTable->Insert( rstrKey, pVal );
	}

	char szVal[128];
	snprintf( szVal, 128, "%d", iValue );

	pVal->m_strValue = szVal;
	pVal->m_iValue   = iValue;
	pVal->m_fValue   = float( iValue );
	pVal->m_bValue   = pVal->m_iValue ? true : false;

	for( unsigned int i = 0; i < m_vpkCallbacks.size(); ++i )
		m_vpkCallbacks[i]->ConfigValueChange( rstrKey );
}


void Config::SetValue( const HashString &rstrKey, bool bValue )
{
	if( !m_pkHashTable )
		return;

	ConfigValue *pVal = m_pkHashTable->Find( rstrKey );

	if( !pVal )
	{
		pVal = new ConfigValue();

		m_pkHashTable->Insert( rstrKey, pVal );
	}

	pVal->m_strValue = bValue ? "true" : "false";
	pVal->m_iValue   = bValue ? 1 : 0;
	pVal->m_fValue   = bValue ? 1.0f : 0.0f;
	pVal->m_bValue   = bValue;

	for( unsigned int i = 0; i < m_vpkCallbacks.size(); ++i )
		m_vpkCallbacks[i]->ConfigValueChange( rstrKey );
}


int Config::GetValue( const HashString &rstrKey, std::string *pstrValue )
{
	if( !m_pkHashTable )
		return 0;

	ConfigValue *pVal = m_pkHashTable->Find( rstrKey );

	if( pVal )
	{
		*pstrValue = (std::string&)pVal->m_strValue;
		return 1;
	}

	return 0;
}


int Config::GetValue( const HashString &rstrKey, float *pfValue )
{
	if( !m_pkHashTable )
		return 0;

	ConfigValue *pVal = m_pkHashTable->Find( rstrKey );

	if( pVal )
	{
		*pfValue = pVal->m_fValue;
		return 1;
	}

	return 0;
}


int Config::GetValue( const HashString &rstrKey, int *piValue )
{
	if( !m_pkHashTable )
		return 0;

	ConfigValue *pVal = m_pkHashTable->Find( rstrKey );

	if( pVal )
	{
		*piValue = pVal->m_iValue;
		return 1;
	}

	return 0;
}


int Config::GetValue( const HashString &rstrKey, bool *pbValue )
{
	if( !m_pkHashTable )
		return 0;

	ConfigValue *pVal = m_pkHashTable->Find( rstrKey );

	if( pVal )
	{
		*pbValue = pVal->m_bValue;
		return 1;
	}

	return 0;
}


void Config::RegisterCallback( ConfigCallback *pkCallback )
{
	vector< ConfigCallback* >::iterator ppkCallback    = m_vpkCallbacks.begin();
	vector< ConfigCallback* >::iterator ppkCallbackEnd = m_vpkCallbacks.end();

	for( ; ppkCallback != ppkCallbackEnd; ++ppkCallback )
		if( *ppkCallback == pkCallback )
			return;

	m_vpkCallbacks.push_back( pkCallback );
}


void Config::UnregisterCallback( ConfigCallback *pkCallback )
{
	vector< ConfigCallback* >::iterator ppkCallback    = m_vpkCallbacks.begin();
	vector< ConfigCallback* >::iterator ppkCallbackEnd = m_vpkCallbacks.end();

	for( ; ppkCallback != ppkCallbackEnd; ++ppkCallback )
		if( *ppkCallback == pkCallback )
		{
			m_vpkCallbacks.erase( ppkCallback );
			return;
		}
}


}; // namespace NeoEngine

