/***************************************************************************
                       core.cpp  -  Core engine object
                             -------------------
    begin                : Mon Jul 28 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, core.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "core.h"
#include "module.h"
#include "material.h"
#include "mesh.h"
#include "room.h"
#include "terrain.h"
#include "font.h"
#include "filemanager.h"
#include "audio.h"
#include "input.h"
#include "profile.h"
#include "console.h"
#include "config.h"
#include "vertex.h"
#include "nemath.h"
#include "timer.h"
#include "logstream.h"
#include "util.h"

#include <math.h>
#include <time.h>

#ifdef WIN32
#  ifdef _MSC_VER
#    define strcasecmp(a,b) stricmp((a),(b))
#  endif
#  ifndef WIN32_LEAN_AND_MEAN
#    define WIN32_LEAN_AND_MEAN
#  endif
#  include <windows.h>
#  include <winsock2.h>
#  undef ERROR
#elif defined(POSIX) || defined(__APPLE__)
#  include <sys/time.h>
#endif

#ifdef __APPLE__
#  include <sys/sysctl.h>
#endif

using namespace std;


namespace NeoEngine
{


struct TerrainPrivate
{
        ModulePtr m_pkTerrainModule;
 
        ~TerrainPrivate() { m_pkTerrainModule = 0; }
};


#ifdef ARCH_X86
#  if !defined( __MINGW32__ ) || defined( BUILD_STATIC )
const Core::BYTEORDER Core::BYTEORDER_SYSTEM = BYTEORDER_LITTLEENDIAN;
#  else
      Core::BYTEORDER Core::BYTEORDER_SYSTEM = BYTEORDER_LITTLEENDIAN;
#  endif
#elif defined(ARCH_PPC)
const Core::BYTEORDER Core::BYTEORDER_SYSTEM = BYTEORDER_BIGENDIAN;
#else
const Core::BYTEORDER Core::BYTEORDER_SYSTEM = BYTEORDER_LITTLEENDIAN;
#endif


Buffer::BUFFERUPLOADPOLICY Buffer::s_eUploadPolicy = Buffer::ONRENDER;



NE_STATIC Core *Core::s_pkSingleton = 0;


ModuleManager *Core::s_pkModuleManager = 0;


Core::Core() :
#ifdef __APPLE__
	m_iHasAltiVec( 0 ),
#endif

#ifdef ARCH_X86
	m_iHasMMX( 0 ),
	m_iHasFXSR( 0 ),
	m_iHasSSE( 0 ),
	m_iHasSSE2( 0 ),
	m_iHas3DNow( 0 ),
	m_iHas3DNowEx( 0 ),
#endif

#ifdef ARCH_X86_64
	m_iHasMMX( 1 ),
	m_iHasSSE( 1 ),
	m_iHasSSE2( 1 ),
	m_iHas3DNow( 1 ),
	m_iHas3DNowEx( 1 ),
#endif

	m_pkFileManager( 0 ),
	m_pkInputManager( 0 ),
	m_pkMaterialManager( 0 ),
	m_pkMeshManager( 0 ),
	m_pkRoomManager( 0 ),
	m_pkTerrainManager( 0 ),
	m_pkTerrainPrivate( 0 ),
	m_pkFontManager( 0 ),
	m_pkProfileManager( 0 ),
	m_pkConsole( 0 ),
	m_pkConfig( 0 ),
	m_pkRenderDevice( 0 ),
	m_pkStdoutSink( new LogStdoutSink ),
	m_pkAudioDevice( 0 )
{
	srand( time(0) );

#ifdef WIN32
	QueryPerformanceCounter( (LARGE_INTEGER*)&Timer::s_uliOldHeartbeatClock );
	QueryPerformanceFrequency( (LARGE_INTEGER*)&Timer::s_uliDivisor );
#elif defined(POSIX) || defined(__APPLE__)
	timeval tv;

	gettimeofday( &tv, 0 );

	Timer::s_uliOldHeartbeatClock = uint64_t( tv.tv_sec ) * 1000000LL + (uint64_t)tv.tv_usec;
	Timer::s_uliDivisor           = 1000000LL;
#else
#  error Timer heartbeat reset not implemented
#endif
}


Core::~Core()
{
	delete m_pkStdoutSink;
}


int Core::Initialize( int iArgs, char **ppszArgs )
{
	neolog << LogLevelPush( INFO )
	       << "NeoEngine v" << NEOENGINEVERSIONSTRING << " initializing" << endl
	       << "  Copyright 2001-2004 Reality Rift Studios" << endl
	       << "  Released under MPL license" << endl
	       << "  http://neoengine.sourceforge.net" << endl
	       << "  #neoengine @ irc.freenode.net" << endl
	       << "  Mattias Jansson (mattias@realityrift.com)" << endl
	       << "  Cody Russell, Keaton Mullis, Markus Wolf" << endl << LogLevelPop();

#ifdef WIN32
	//On Win32 platforms, do not set working directory by default
	bool bChdir = false;
#else
	//Set working directory by default
	bool bChdir = true;
#endif

	//Parse arguments
	for( int iArg = 0; iArg < iArgs; ++iArg )
	{
		if( !strcmp( ppszArgs[iArg], "--loglevel" ) )
		{
			++iArg;

			if( iArg < iArgs )
			{
				LOGLEVEL eLevel = INFO;

				if( !strcasecmp( ppszArgs[iArg], "debug" ) )
					eLevel = DEBUG;
				else if( !strcasecmp( ppszArgs[iArg], "info" ) )
					eLevel = INFO;
				else if( !strcasecmp( ppszArgs[iArg], "warning" ) )
					eLevel = WARNING;
				else if( !strcasecmp( ppszArgs[iArg], "error" ) )
					eLevel = ERROR;
				else if( !strcasecmp( ppszArgs[iArg], "panic" ) )
					eLevel = PANIC;
				else
					eLevel = (LOGLEVEL)atoi( ppszArgs[iArg] );

				neolog.SetLogLevel( eLevel );
			}
		}
		else if( !strcmp( ppszArgs[iArg], "--chdir" ) )
			bChdir = true;
		else if( !strcmp( ppszArgs[iArg], "--no-chdir" ) )
			bChdir = false;
	}

#ifdef ARCH_X86
	unsigned int regs[4];
	/* Check for CPUID support */
#  ifdef __GNUC__
	asm volatile(
		"pushf\n\t"
		"popl %0\n\t"
		"movl %0, %1\n\t"
		"xorl $0x200000, %0\n\t"
		"push %0\n\t"
		"popf\n\t"
		"pushf\n\t"
		"popl %0\n\t"
		: "=a" (regs[0]), "=c" (regs[1])
		:
		: "cc");
#  elif WIN32
	__asm
	{
		pushfd
		pop eax
		mov ecx, eax
		xor eax, 200000h
		push eax
		popfd
		pushfd
		pop eax
		mov regs[0], eax
		mov regs[1], ecx
	}
#  endif
	if (regs[0] != regs[1]) {
#  ifdef __GNUC__
		asm volatile(
			"movl %%ebx, %%esi\n\t"
			"cpuid\n\t"
			"xchgl %%ebx, %%esi\n\t"
			: "=a" (regs[0]), "=S" (regs[1]),
			  "=c" (regs[2]), "=d" (regs[3])
			: "0" (0x00000000));
#  elif WIN32
		__asm
		{
			mov esi, ebx
			mov eax, 0
			cpuid
			xchg esi, ebx
			mov regs[0], eax
			mov regs[1], esi
			mov regs[2], ecx
			mov regs[3], edx
		}
#  endif

#  ifdef __GNUC__
		/* Check for Intel extensions */
		asm volatile(
			"movl %%ebx, %%esi\n\t"
			"cpuid\n\t"
			"xchgl %%ebx, %%esi\n\t"
			: "=a" (regs[0]), "=S" (regs[1]),
			  "=c" (regs[2]), "=d" (regs[3])
			: "0" (0x00000001));
#  elif WIN32
		__asm
		{
			mov esi, ebx
			mov eax, 1
			cpuid
			xchg esi, ebx
			mov regs[0], eax
			mov regs[1], esi
			mov regs[2], ecx
			mov regs[3], edx
		}
#  endif
		m_iHasMMX = (regs[3] & (1 << 23)) >> 23;
		m_iHasFXSR = (regs[3] & (1 << 24)) >> 24;
		m_iHasSSE = (regs[3] & (1 << 25)) >> 25;
		m_iHasSSE2 = (regs[3] & (1 << 26)) >> 26;
		m_iHasFXSR = m_iHasSSE;

		/* Check for AMD extensions */
#  ifdef __GNUC__
		asm volatile(
			"movl %%ebx, %%esi\n\t"
			"cpuid\n\t"
			"xchgl %%ebx, %%esi\n\t"
			: "=a" (regs[0]), "=S" (regs[1]),
			  "=c" (regs[2]), "=d" (regs[3])
			: "0" (0x80000000));
#  elif WIN32
		__asm
		{
			mov esi, ebx
			mov eax, 80000000h
			cpuid
			xchg esi, ebx
			mov regs[0], eax
			mov regs[1], esi
			mov regs[2], ecx
			mov regs[3], edx
		}
#  endif
		if (regs[0] >= 0x8000001) {
#  ifdef __GNUC__
			asm volatile(
				"movl %%ebx, %%esi\n\t"
				"cpuid\n\t"
				"xchgl %%ebx, %%esi\n\t"
				: "=a" (regs[0]), "=S" (regs[1]),
				  "=c" (regs[2]), "=d" (regs[3])
				: "0" (0x80000001));
#  elif WIN32
			__asm
			{
				mov esi, ebx
				mov eax, 80000001h
				cpuid
				xchg esi, ebx
				mov regs[0], eax
				mov regs[1], esi
				mov regs[2], ecx
				mov regs[3], edx
			}
#  endif
			m_iHasMMX |= (regs[3] & (1 << 23)) >> 23;
			m_iHasFXSR |= (regs[3] & (1 << 22)) >> 22;
			m_iHas3DNow = (regs[3] & (1 << 31)) >> 31;
			m_iHas3DNowEx = (regs[3] & (1 << 30)) >> 30;
		}
	} else {
		neolog << LogLevel( INFO ) << "  ... CPUID not supported by processor" << endl;
	}

	neolog << LogLevel( INFO ) << "  ... CPU Flags: "
	       << ( m_iHasMMX ? "MMX " : "" )
	       << ( m_iHasFXSR ? "FXSR " : "" )
	       << ( m_iHasSSE ? "SSE " : "" )
	       << ( m_iHasSSE2 ? "SSE2 " : "" )
	       << ( m_iHas3DNow ? "3DNow " : "" )
	       << ( m_iHas3DNowEx ? "3DNowEx " : "" ) << endl;
#endif

#ifdef __APPLE__

	int aiSel[2]    = { CTL_HW, HW_VECTORUNIT };
	int iHasVecUnit = 0;
	size_t iLen     = sizeof (iHasVecUnit);
	int error       = sysctl (aiSel, 2, &iHasVecUnit, &iLen, NULL, 0);

	if( ( error == 0 ) && ( m_iHasAltiVec = ( iHasVecUnit != 0 ) ? 1 : 0 ) )
		neolog << LogLevel( INFO ) << "  ... AltiVec present" << endl;

#endif

#if defined(ARCH_X86) && defined(INTEL_INTRINSICS)

	if( m_iHasSSE )
	{
		/*Timer *kTimer = new Timer();
		char *tmp1 = new char[1024 * 1024 * 10];
		char *tmp2 = new char[1024 * 1024 * 10];
		float fSSE = 0, fStandard = 0;

	  	neolog << LogLevel( INFO ) << "[Benchmarking memcpys]" << endl;

		kTimer->Reset( );
		memcpy_sse( tmp1, tmp2, 1024 * 1024 * 10 );
		fSSE = kTimer->GetDeltaTime( true );
		memcpy( tmp1, tmp2, 1024 * 1024 * 10 );
		fStandard = kTimer->GetDeltaTime( false );

		delete tmp1, tmp1 = NULL;
		delete tmp2, tmp2 = NULL;
		delete kTimer, kTimer = NULL;

	        neolog << LogLevel( INFO ) << "[SSE: " << fSSE << "\tStandard: " << fStandard << "]" << endl;

		if( fSSE < fStandard )
		{
		        neolog << LogLevel( INFO ) << "[SSE optimized core]" << endl;
			fmemcpy = memcpy_sse;
		}
		else
		{
		        neolog << LogLevel( INFO ) << "[Using kernel memcpy]" << endl;
			fmemcpy = memcpy;
		}
		fmemcpy = memcpy_sse;*/
	}
	else
#endif /* ARCH_X86 && INTEL_INTRINSICS */
		fmemcpy = memcpy;

	//Optionally set working directory
	if( bChdir && ( iArgs > 0 ) )
	{
		//extract path from program argument
		std::string strPath( ppszArgs[0] );

		std::string::size_type uiPathDelimiter = strPath.find_last_of( '/' );

		if( uiPathDelimiter == std::string::npos )
			uiPathDelimiter   = strPath.find_last_of( '\\' );

		if( uiPathDelimiter != std::string::npos )
		{
			strPath = strPath.substr( 0, uiPathDelimiter );

			neolog << LogLevel( DEBUG ) << "  ... setting working directory to [" << strPath << "]" << endl;
#ifdef WIN32
			SetCurrentDirectory( strPath.c_str() );
#elif defined(POSIX) || defined(__APPLE__)
			chdir( strPath.c_str() );
#endif
		}
	}

	//Create configuration repository
	neolog << LogLevel( INFO ) << "  ... creating config repository" << endl;
	m_pkConfig  = new Config( 0 );


	//Create module manager
	neolog << LogLevel( INFO ) << "  ... creating module manager" << endl;
	if( !s_pkModuleManager )
		s_pkModuleManager = new ModuleManager;


	//Create file manager
	neolog << LogLevel( INFO ) << "  ... creating file manager" << endl;
	m_pkFileManager = new FileManager( "core" );


	//Create input manager
	neolog << LogLevel( INFO ) << "  ... creating input manager" << endl;
	m_pkInputManager = new InputManager;


	//Create console
	neolog << LogLevel( INFO ) << "  ... creating console" << endl;
	m_pkConsole = new Console;

	//Register basic configure commands for the console
	neolog << LogLevel( INFO ) << "    ... registering console configure callbacks" << endl;
	m_pkConsole->RegisterCommand( "set",        m_pkConfig );
	m_pkConsole->RegisterCommand( "showconfig", m_pkConfig );


	//Create function lookup tables
	neolog << LogLevel( INFO ) << "  ... creating lookup tables" << endl;

	for( int i = 0; i < TextureMatrixGen::TEXMODFUNCTABLE_SIZE; ++i )
	{
		TextureMatrixGen::s_afSinTable[i]             = sinf( ( i * 360.0f / float( TextureMatrixGen::TEXMODFUNCTABLE_SIZE - 1 ) ) * PI / 180.0f );
		TextureMatrixGen::s_afSquareTable[i]          = ( i < ( TextureMatrixGen::TEXMODFUNCTABLE_SIZE / 2 ) ) ? 1.0f : -1.0f;
		TextureMatrixGen::s_afSawToothTable[i]        = (float)i / TextureMatrixGen::TEXMODFUNCTABLE_SIZE;
		TextureMatrixGen::s_afInverseSawToothTable[i] = 1.0f - TextureMatrixGen::s_afSawToothTable[i];

		if( i < ( TextureMatrixGen::TEXMODFUNCTABLE_SIZE / 2 ) )
		{
			if( i < ( TextureMatrixGen::TEXMODFUNCTABLE_SIZE / 4 ) )
				TextureMatrixGen::s_afTriangleTable[i] = (float)i / float( TextureMatrixGen::TEXMODFUNCTABLE_SIZE / 4 );
			else
				TextureMatrixGen::s_afTriangleTable[i] = 1.0f - TextureMatrixGen::s_afTriangleTable[ i - ( TextureMatrixGen::TEXMODFUNCTABLE_SIZE / 4 ) ];
		}
		else
			TextureMatrixGen::s_afTriangleTable[i] = -TextureMatrixGen::s_afTriangleTable[ i - ( TextureMatrixGen::TEXMODFUNCTABLE_SIZE / 2 ) ];
	}


	//Create material table
	neolog << "  ... creating material manager" << endl;
	m_pkMaterialManager = new MaterialTable;


	//Create mesh blueprint pool
	neolog << "  ... creating mesh manager" << endl;
	m_pkMeshManager     = new MeshPool;


	//Create room manager
	neolog << "  ... creating room manager" << endl;
	m_pkRoomManager     = new RoomManager;


	//Create font manager
	neolog << "  ... creating font manager" << endl;
	m_pkFontManager     = new FontManager;

	//Create profile manager
        neolog << "  ... creating profile manager" << endl;
	m_pkProfileManager  = new ProfileManager;

	//Create silly private terrain junk
	m_pkTerrainPrivate  = new TerrainPrivate;


#ifdef WIN32
	//Initialize winsock
	WSADATA kWSAData;
	if( WSAStartup( MAKEWORD(2,1), &kWSAData ) == SOCKET_ERROR )
		neolog << LogLevel( ERROR ) << "FATAL ERROR: Unable to initialize WinSOCK, network subsystem unavailable" << endl;
#endif


	//Initialize basic vertex format declarations
	Vertex::s_kDecl.m_pkElements[0].m_uiType   = VertexElement::FLOAT3;
	Vertex::s_kDecl.m_pkElements[0].m_uiUsage  = VertexElement::POSITION;
	Vertex::s_kDecl.m_pkElements[0].m_uiOffset = 0;


	NormalVertex::s_kDecl.m_pkElements[0].m_uiType   = VertexElement::FLOAT3;
	NormalVertex::s_kDecl.m_pkElements[0].m_uiUsage  = VertexElement::POSITION;
	NormalVertex::s_kDecl.m_pkElements[0].m_uiOffset = 0;

	NormalVertex::s_kDecl.m_pkElements[1].m_uiType   = VertexElement::FLOAT3;
	NormalVertex::s_kDecl.m_pkElements[1].m_uiUsage  = VertexElement::NORMAL;
	NormalVertex::s_kDecl.m_pkElements[1].m_uiOffset = 12;


	DiffuseVertex::s_kDecl.m_pkElements[0].m_uiType   = VertexElement::FLOAT3;
	DiffuseVertex::s_kDecl.m_pkElements[0].m_uiUsage  = VertexElement::POSITION;
	DiffuseVertex::s_kDecl.m_pkElements[0].m_uiOffset = 0;

	DiffuseVertex::s_kDecl.m_pkElements[1].m_uiType   = VertexElement::COLOR32;
	DiffuseVertex::s_kDecl.m_pkElements[1].m_uiUsage  = VertexElement::DIFFUSECOLOR;
	DiffuseVertex::s_kDecl.m_pkElements[1].m_uiOffset = 12;


	TexVertex::s_kDecl.m_pkElements[0].m_uiType   = VertexElement::FLOAT3;
	TexVertex::s_kDecl.m_pkElements[0].m_uiUsage  = VertexElement::POSITION;
	TexVertex::s_kDecl.m_pkElements[0].m_uiOffset = 0;

	TexVertex::s_kDecl.m_pkElements[1].m_uiType   = VertexElement::FLOAT2;
	TexVertex::s_kDecl.m_pkElements[1].m_uiUsage  = VertexElement::TEXCOORD;
	TexVertex::s_kDecl.m_pkElements[1].m_uiIndex  = 0;
	TexVertex::s_kDecl.m_pkElements[1].m_uiOffset = 12;


	NormalDiffuseVertex::s_kDecl.m_pkElements[0].m_uiType   = VertexElement::FLOAT3;
	NormalDiffuseVertex::s_kDecl.m_pkElements[0].m_uiUsage  = VertexElement::POSITION;
	NormalDiffuseVertex::s_kDecl.m_pkElements[0].m_uiOffset = 0;

	NormalDiffuseVertex::s_kDecl.m_pkElements[1].m_uiType   = VertexElement::FLOAT3;
	NormalDiffuseVertex::s_kDecl.m_pkElements[1].m_uiUsage  = VertexElement::NORMAL;
	NormalDiffuseVertex::s_kDecl.m_pkElements[1].m_uiOffset = 12;

	NormalDiffuseVertex::s_kDecl.m_pkElements[2].m_uiType   = VertexElement::COLOR32;
	NormalDiffuseVertex::s_kDecl.m_pkElements[2].m_uiUsage  = VertexElement::DIFFUSECOLOR;
	NormalDiffuseVertex::s_kDecl.m_pkElements[2].m_uiOffset = 24;


	NormalTexVertex::s_kDecl.m_pkElements[0].m_uiType   = VertexElement::FLOAT3;
	NormalTexVertex::s_kDecl.m_pkElements[0].m_uiUsage  = VertexElement::POSITION;
	NormalTexVertex::s_kDecl.m_pkElements[0].m_uiOffset = 0;

	NormalTexVertex::s_kDecl.m_pkElements[1].m_uiType   = VertexElement::FLOAT3;
	NormalTexVertex::s_kDecl.m_pkElements[1].m_uiUsage  = VertexElement::NORMAL;
	NormalTexVertex::s_kDecl.m_pkElements[1].m_uiOffset = 12;

	NormalTexVertex::s_kDecl.m_pkElements[2].m_uiType   = VertexElement::FLOAT2;
	NormalTexVertex::s_kDecl.m_pkElements[2].m_uiUsage  = VertexElement::TEXCOORD;
	NormalTexVertex::s_kDecl.m_pkElements[2].m_uiIndex  = 0;
	NormalTexVertex::s_kDecl.m_pkElements[2].m_uiOffset = 24;


	DiffuseTexVertex::s_kDecl.m_pkElements[0].m_uiType   = VertexElement::FLOAT3;
	DiffuseTexVertex::s_kDecl.m_pkElements[0].m_uiUsage  = VertexElement::POSITION;
	DiffuseTexVertex::s_kDecl.m_pkElements[0].m_uiOffset = 0;

	DiffuseTexVertex::s_kDecl.m_pkElements[1].m_uiType   = VertexElement::COLOR32;
	DiffuseTexVertex::s_kDecl.m_pkElements[1].m_uiUsage  = VertexElement::DIFFUSECOLOR;
	DiffuseTexVertex::s_kDecl.m_pkElements[1].m_uiOffset = 12;

	DiffuseTexVertex::s_kDecl.m_pkElements[2].m_uiType   = VertexElement::FLOAT2;
	DiffuseTexVertex::s_kDecl.m_pkElements[2].m_uiUsage  = VertexElement::TEXCOORD;
	DiffuseTexVertex::s_kDecl.m_pkElements[2].m_uiIndex  = 0;
	DiffuseTexVertex::s_kDecl.m_pkElements[2].m_uiOffset = 16;


	NormalDiffuseTexVertex::s_kDecl.m_pkElements[0].m_uiType   = VertexElement::FLOAT3;
	NormalDiffuseTexVertex::s_kDecl.m_pkElements[0].m_uiUsage  = VertexElement::POSITION;
	NormalDiffuseTexVertex::s_kDecl.m_pkElements[0].m_uiOffset = 0;

	NormalDiffuseTexVertex::s_kDecl.m_pkElements[1].m_uiType   = VertexElement::FLOAT3;
	NormalDiffuseTexVertex::s_kDecl.m_pkElements[1].m_uiUsage  = VertexElement::NORMAL;
	NormalDiffuseTexVertex::s_kDecl.m_pkElements[1].m_uiOffset = 12;

	NormalDiffuseTexVertex::s_kDecl.m_pkElements[2].m_uiType   = VertexElement::COLOR32;
	NormalDiffuseTexVertex::s_kDecl.m_pkElements[2].m_uiUsage  = VertexElement::DIFFUSECOLOR;
	NormalDiffuseTexVertex::s_kDecl.m_pkElements[2].m_uiOffset = 24;

	NormalDiffuseTexVertex::s_kDecl.m_pkElements[3].m_uiType   = VertexElement::FLOAT2;
	NormalDiffuseTexVertex::s_kDecl.m_pkElements[3].m_uiUsage  = VertexElement::TEXCOORD;
	NormalDiffuseTexVertex::s_kDecl.m_pkElements[3].m_uiIndex  = 0;
	NormalDiffuseTexVertex::s_kDecl.m_pkElements[3].m_uiOffset = 28;

	return 0;
}


int Core::Shutdown()
{
	neolog << LogLevel( INFO ) << "NeoEngine shutting down" << endl;

#ifdef WIN32
	//Shutdown winsock
	WSACleanup();
#endif

	delete m_pkConsole,         m_pkConsole           = 0;
	delete m_pkConfig,          m_pkConfig            = 0;
	delete m_pkProfileManager,  m_pkProfileManager    = 0;
	delete m_pkFontManager,     m_pkFontManager       = 0;
	delete m_pkRoomManager,     m_pkRoomManager       = 0;
	delete m_pkTerrainManager,  m_pkTerrainManager    = 0;
	delete m_pkMeshManager,     m_pkMeshManager       = 0;
	delete m_pkMaterialManager, m_pkMaterialManager   = 0;
	delete m_pkTerrainPrivate,  m_pkTerrainPrivate    = 0;

	if( m_pkRenderDevice )
	{
		neolog << LogLevel( WARNING ) <<  "*** render device not deleted before shutdown" << endl;
		DeleteRenderDevice( m_pkRenderDevice );
	}

	delete m_pkInputManager,    m_pkInputManager      = 0;
	delete m_pkFileManager,     m_pkFileManager       = 0;
	delete s_pkModuleManager,   s_pkModuleManager     = 0;

	s_pkSingleton = 0;

	delete this;

	return 0;
}


ModuleManager *Core::GetModuleManager()
{
	if( !s_pkModuleManager )
		s_pkModuleManager = new ModuleManager;

	return s_pkModuleManager;
}


void Core::InitializeTerrain( const std::string &rstrName )
{
#if defined( WIN32 ) && defined( _DEBUG )
        HashString strModuleName = "neo" + rstrName + "_debug";
#else
        HashString strModuleName = "neo" + rstrName;
#endif
 
        m_pkTerrainPrivate->m_pkTerrainModule = GetModuleManager()->LoadModule( strModuleName );
 
        if( m_pkTerrainPrivate->m_pkTerrainModule )
        {
                ModSymbol kSymInitialize;
 
                kSymInitialize = m_pkTerrainPrivate->m_pkTerrainModule->LookupSymbol( "Initialize" );
 
                if( !kSymInitialize )
                {
                        neolog << LogLevel( WARNING ) << "[Core::InitializeTerrain] Terrain module ["
                               << rstrName << "] does not export Initialize method" << endl;
                }
                else
                {
                        (( void ( NE_CDECL* )() )kSymInitialize)();
                }
        }
        else
        {
                neolog << LogLevel( ERROR ) << "*** Unable to load terrain module [" << rstrName << "]" << endl;
        }
 
        if( !m_pkTerrainPrivate->m_pkTerrainModule )
        {
                neolog << LogLevel( ERROR ) << "[Core::InitializeTerrain] No module available for terrain type [" << rstrName << "]" << endl;
                return;
        }
                                                                                                                                              
        ModSymbol kSymCreate = m_pkTerrainPrivate->m_pkTerrainModule->LookupSymbol( "CreateTerrainManager" );
                                                                                                                                              
        if( !kSymCreate )
        {
                neolog << LogLevel( ERROR ) << "[Core::InitializeTerrain] Terrain module [" << m_pkTerrainPrivate->m_pkTerrainModule->GetName() << "] does not export CreateTerrainManager method" << endl;
                return;
        }
                                                                                                                                              
        m_pkTerrainManager = (( TerrainManager* ( NE_CDECL* )() ) kSymCreate)();
                                                                                                                                              
        if( !m_pkTerrainManager )
        {
                neolog << "[Core::InitializeTerrain]: Could not create TerrainManager object." << endl;
        }
}


void Core::SetConsole( Console *pkConsole )
{
	delete m_pkConsole;

	m_pkConsole = pkConsole;
}


};

