/***************************************************************************
                   directory.cpp  -  Directory abstraction
                             -------------------
    begin                : Wed Sep 18 2002
    copyright            : (C) 2002 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, directory.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "directory.h"
#include "file.h"
#include "logstream.h"

#ifdef WIN32
 //Windows-specific includes
#  include <windows.h>
#elif defined( POSIX ) || defined( __APPLE__ )
//Posix-specific includes
#  include <dirent.h>
#  if !defined( _DIRENT_HAVE_D_TYPE ) && ( defined( __APPLE__ ) || defined( FREEBSD ) )
#    define _DIRENT_HAVE_D_TYPE
#  endif
#endif

using namespace std;


namespace NeoEngine
{


Directory::Directory( const std::string &rstrName, Directory *pkParent, bool bParse, bool bRecurse ) :
	m_pkFiles( new HashTable< FileTemplate > ),
	m_pkParent( 0 ),
	m_strName( rstrName ),
	m_bIsParsed( false ),
	m_bIsPackage( false )
{
	if( pkParent )
		pkParent->AttachNode( this );

	if( bParse )
		Parse( bRecurse );
}


Directory::~Directory()
{
	if( m_pkParent )
		m_pkParent->DetachNode( this ), m_pkParent = 0;

	vector<FileTemplate*> vpkFiles;

	m_pkFiles->GetAllNodeData( &vpkFiles );

	vector<FileTemplate*>::iterator ppkNode = vpkFiles.begin();
	vector<FileTemplate*>::iterator ppkEnd  = vpkFiles.end();

	for( ; ppkNode != ppkEnd; ++ppkNode )
		delete (*ppkNode);

	delete m_pkFiles;

	while( m_vpkDirectories.size() )
		delete m_vpkDirectories[0];
}


void Directory::AttachNode( Directory *pkDirectory )
{
	if( !pkDirectory )
		return;

	vector< Directory* >::iterator ppkDir    = m_vpkDirectories.begin();
	vector< Directory* >::iterator ppkDirEnd = m_vpkDirectories.end();

	for( ; ppkDir != ppkDirEnd; ++ppkDir )
		if( (*ppkDir) == pkDirectory )
		{
			neolog << LogLevel( WARNING ) << "[Directory::AttachNode]: Directory already attached, ignoring request" << endl;
			return;
		}

	bool bAdded = false;

	ppkDir    = m_vpkDirectories.begin();
	ppkDirEnd = m_vpkDirectories.end();

	//Insert in correct priority (name sort, if same name exists, real directories above packages)
	for( ; ppkDir != ppkDirEnd; ++ppkDir )
	{
		if( ( pkDirectory->GetName() < (*ppkDir)->GetName() ) ||
		    ( ( pkDirectory->GetName() == (*ppkDir)->GetName() ) &&
		      ( (*ppkDir)->IsPackage() && !pkDirectory->IsPackage() ) ) )
		{
		 	m_vpkDirectories.insert( ppkDir, pkDirectory );
			bAdded = true;
			break;
		}
	}

	if( !bAdded )
		m_vpkDirectories.push_back( pkDirectory );

	pkDirectory->m_pkParent    = this;
	pkDirectory->m_strFullPath = "";
}


void Directory::DetachNode( Directory *pkDirectory )
{
	if( !pkDirectory )
		return;

	for( vector<Directory*>::iterator ppkNode = m_vpkDirectories.begin(); ppkNode != m_vpkDirectories.end(); ++ppkNode )
		if( (*ppkNode) == pkDirectory )
		{
			m_vpkDirectories.erase( ppkNode );
			return;
		}

	pkDirectory->m_pkParent    = 0;
	pkDirectory->m_strFullPath = "";
}


File *Directory::GetByName( const std::string &rstrName, bool bRecurse )
{
	string strDir = File::ExtractPathName( rstrName );

	if( strDir.length() )
	{
		Directory *pkDirectory = GetDirectoryByName( strDir );

		return( pkDirectory ? pkDirectory->GetByName( File::ExtractFileName( rstrName ), false ) : 0 );
	}

	FileTemplate *pkFileTemplate = m_pkFiles->Find( rstrName );

	if( pkFileTemplate )
		return pkFileTemplate->AllocateFile( rstrName, this );

	if( bRecurse )
	{
		vector<Directory*>::iterator ppkNode = m_vpkDirectories.begin();
		vector<Directory*>::iterator ppkEnd  = m_vpkDirectories.end();

		File *pkFile = 0;

		for( ; ppkNode != ppkEnd; ++ppkNode )
			if( ( pkFile = (*ppkNode)->GetByName( rstrName, true ) ) )
				return pkFile;
	}

	return 0;
}


Directory *Directory::GetDirectoryByName( const std::string &rstrName )
{
	if( !rstrName.length() || ( rstrName == "." ) )
		return this;

	//Splice path
	std::string::size_type iDelim = rstrName.find_first_of( "\\" );

	if( iDelim == string::npos )
		iDelim = rstrName.find_first_of( "/", 1 );

	string strDir = ( ( iDelim != string::npos ) ? rstrName.substr( 0, iDelim ) : rstrName );
	string strSec = ( ( iDelim != string::npos ) ? rstrName.substr( iDelim + 1, string::npos ) : "" );

	Directory *pkDirectory = 0;

	if( !strDir.length() )
		pkDirectory = this;
	else if( strDir == "." )
		return GetDirectoryByName( strSec );
	else
	{
		vector< Directory* >::iterator ppkNode = m_vpkDirectories.begin();
		vector< Directory* >::iterator ppkEnd  = m_vpkDirectories.end();

		for( ; ppkNode != ppkEnd; ++ppkNode )
			if( (*ppkNode)->GetName() == strDir )
			{
				pkDirectory = *ppkNode;
				break;
			}
	}

	if( pkDirectory )
		return pkDirectory->GetDirectoryByName( strSec );
	else if( ( strDir == ".." ) && GetParentDirectory() )
		return GetParentDirectory()->GetDirectoryByName( strSec );

	return 0;
}


const string &Directory::GetFullPath()
{
	if( m_strFullPath.length() )
		return m_strFullPath;

	if( m_pkParent )
	{
		m_strFullPath = m_pkParent->GetFullPath();

		if( m_strFullPath.length() && ( m_strFullPath[ m_strFullPath.length() - 1 ] != '/' ) )
			m_strFullPath += '/';

		m_strFullPath += m_strName;
	}
	else
		m_strFullPath = m_strName;

	return m_strFullPath;
}


bool Directory::Parse( bool bRecurse )
{
	if( m_bIsParsed )
	{
		if( bRecurse )
		{
			vector< Directory* >::iterator ppkNode = m_vpkDirectories.begin();
			vector< Directory* >::iterator ppkEnd  = m_vpkDirectories.end();

			for( ; ppkNode != ppkEnd; ++ppkNode )
				(*ppkNode)->Parse( true );
		}

		return true;
	}

	//Check if it is a directory in normal file system
#ifdef WIN32

	//Windows specific implementation of directory check
	WIN32_FIND_DATA sData;

	char szPattern[1024];

	sprintf( szPattern, "%s/*", GetFullPath().length() ? GetFullPath().c_str() : "." );

	HANDLE hFind = FindFirstFile( szPattern, &sData );

	if( hFind != INVALID_HANDLE_VALUE )
	{
		do
		{
			if( sData.cFileName[0] != '.' )
			{
				if( sData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
					new Directory( sData.cFileName, this, bRecurse, bRecurse );
				else
					m_pkFiles->Insert( sData.cFileName, new FileTemplate );
			}

		} while( FindNextFile( hFind, &sData ) );

		m_bIsParsed = true;
	}
	else
	{
		char szWD[512];

		memset( szWD, 0, 512 );
		GetCurrentDirectory( 512, szWD );

		neolog << LogLevel( WARNING ) << "[Directory::ParseDirectory] Unable to parse directory [" << GetFullPath() << "], wd [" << szWD << "]" << endl;
	}

	FindClose( hFind );

#elif defined(POSIX) || defined(__APPLE__)

	//Linux specific implementation of directory check

	DIR *pDir = 0;

	if( ( pDir = opendir( GetFullPath().length() ? GetFullPath().c_str() : "." ) ) )
	{
		//We have a directory, parse and create virtual file system
		struct dirent *pEntry = 0;

		while( ( pEntry = readdir( pDir ) ) )
		{
			if( strcmp( pEntry->d_name, "." ) && strcmp( pEntry->d_name, ".." ) )
			{
				//Create new directory or file
#ifdef _DIRENT_HAVE_D_TYPE
				if( pEntry->d_type == DT_DIR )
				{
#else
				DIR *pNextDir = opendir( ( GetFullPath() + "/" + pEntry->d_name ).c_str() );
				if( pNextDir )
				{
					closedir( pNextDir );
#endif
					new Directory( pEntry->d_name, this, bRecurse, bRecurse );
				}
				else //if( pEntry->d_type == DT_BLK )
					m_pkFiles->Insert( pEntry->d_name, new FileTemplate );
			}
		}

		closedir( pDir ), pDir = 0;

		m_bIsParsed = true;
	}
	else
		neolog << LogLevel( WARNING ) << "[Directory::ParseDirectory]: Unable to parse directory [" << GetFullPath() << "]" << endl;
#else
#  error Directory::ParseDirectory not implemented
#endif

	return m_bIsParsed;
}


Directory *Directory::GetParentDirectory()
{
	return m_pkParent;
}


void Directory::PrintHierarchy( int iDepth )
{
	unsigned int i;

	if( iDepth < 0 )
		iDepth = 0;

	for( i = 0; i < (unsigned int)iDepth; ++i )
		neolog << LogLevel( DEBUG ) << "  ";

	vector< HashTableNode< FileTemplate >* > vpkFileNodes;

	m_pkFiles->GetAllNodes( &vpkFileNodes );

	neolog << LogLevel( DEBUG ) << m_strName << ( IsPackage() ? " [P] [" : " [" ) << m_vpkDirectories.size() << " directories] [" << vpkFileNodes.size() << " files]" << endl;

	for( i = 0; i < m_vpkDirectories.size(); ++i )
		m_vpkDirectories[i]->PrintHierarchy( iDepth + 1 );

	vector< HashTableNode< FileTemplate >* >::iterator ppkFileNode    = vpkFileNodes.begin();
	vector< HashTableNode< FileTemplate >* >::iterator ppkFileNodeEnd = vpkFileNodes.end();

	for( ; ppkFileNode != ppkFileNodeEnd; ++ppkFileNode )
	{
		for( int j = 0; j < iDepth + 1; ++j )
			neolog << LogLevel( DEBUG ) << "  ";

		neolog << (*ppkFileNode)->m_strKey << endl;
	}
}


void Directory::GetFiles( std::vector<File*> *pvpkFiles ) const
{
	vector< HashTableNode< FileTemplate >* > vpkFileNodes;

	m_pkFiles->GetAllNodes( &vpkFileNodes );

	vector< HashTableNode< FileTemplate >* >::iterator ppkNode    = vpkFileNodes.begin();
	vector< HashTableNode< FileTemplate >* >::iterator ppkNodeEnd = vpkFileNodes.end();

	for( ; ppkNode != ppkNodeEnd; ++ppkNode )
		pvpkFiles->push_back( (*ppkNode)->m_pkData->AllocateFile( (*ppkNode)->m_strKey, (Directory*)this ) );
}





File *FileTemplate::AllocateFile( const std::string &rstrName, Directory *pkDirectory )
{
	return new File( pkDirectory ? pkDirectory->GetFullPath() : "", rstrName, ios_base::in | ios_base::binary, false, pkDirectory );
}


}; // namespace NeoEngine

