/***************************************************************************
      file.cpp  -  File abstraction. Normal files, memory buffer files
                       and virtual files in packages
                             -------------------
    begin                : Thu Nov 1 2001
    copyright            : (C) 2001 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, file.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2001
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "file.h"
#include "filecodec.h"
#include "package.h"
#include "filemanager.h"
#include "logstream.h"
#include "util.h"

#include <limits>

using namespace std;


namespace NeoEngine
{


File::File( const string &rstrPath, const string &rstrName, ios_base::openmode uiOpenMode, bool bOpen, Directory *pkDirectory ) :
	m_pkStream( 0 ),
	m_uiOpenMode( uiOpenMode ),
	m_bBinary( true ),
	m_eByteOrder( Core::BYTEORDER_LITTLEENDIAN ),
	m_strPath( ( !rstrPath.length() || ( rstrPath[ rstrPath.length() - 1 ] != '/' ) ) ? rstrPath : rstrPath.substr( 0, rstrPath.length() - 1 ) ),
	m_strName( rstrName ),
	m_pkDirectory( pkDirectory )
{
	if( bOpen )
		Open();
}


File::~File()
{
	Close();
}


bool File::Open( const string &rstrPath, const string &rstrFilename, ios_base::openmode uiOpenMode )
{
	Close();

	if( rstrFilename.length() )
		m_strName = rstrFilename;

	if( rstrPath.length() )
		m_strPath = ( !rstrPath.length() || ( rstrPath[ rstrPath.length() - 1 ] != '/' ) ) ? rstrPath : rstrPath.substr( 0, rstrPath.length() - 1 );

	if( uiOpenMode != 0 )
		m_uiOpenMode = uiOpenMode;

	if( !m_uiOpenMode )
		m_uiOpenMode = ios_base::in | ios_base::binary;

	m_bBinary = ( ( m_uiOpenMode & ios_base::binary ) != 0 );

	string strFullPath;

	if( m_strPath.length() )
	{
		strFullPath = m_strPath;

		if( strFullPath[ strFullPath.length() - 1 ] != '/' )
			strFullPath += '/';

		strFullPath += m_strName;
	}
	else
		strFullPath = m_strName;

	AllocStream( strFullPath );

	if( !m_pkStream || !( *m_pkStream ) )
	{
		Close();
		return false;
	}

	return true;
}


void File::Close()
{
	delete m_pkStream, m_pkStream = 0;
}


void File::AllocStream( const string &rstrFullPath )
{
	m_pkStream = new fstream( rstrFullPath.c_str(), m_uiOpenMode | ios_base::binary );
}


int File::GetSize()
{
	bool bClose = false;

	if( !IsValid() )
	{
		if( !Open() )
			return -1;
		
		bClose = true;
	}
	
	streampos uiCur = Tellg();

	Seekg( 0, ios_base::end );

	int iSize = (int)Tellg();

	Seekg( uiCur, ios_base::beg );

	if( bClose )
		Close();

	return iSize;
}


File &File::GetLine( char *pcDest, unsigned int uiCount, char cDelimiter )
{
	if( !uiCount )
		uiCount = numeric_limits< int >::max();

	if( pcDest )
		m_pkStream->getline( pcDest, uiCount, cDelimiter );
	else
		m_pkStream->ignore( uiCount, cDelimiter );

	return( *this );
}


File &File::operator << ( bool bData )
{
	if( m_bBinary )
	{
		unsigned char ucVal = ( bData ? 1 : 0 );
		m_pkStream->write( (char*)&ucVal, 1 );
	}
	else
		*m_pkStream << ( bData ? "true" : "false" );
	
	return( *this );
}


File &File::operator << ( int8_t cData )
{
	if( m_bBinary )
		m_pkStream->write( (char*)&cData, 1 );
	else
		*m_pkStream << cData;
	
	return( *this );
}


File &File::operator << ( uint8_t ucData )
{
	if( m_bBinary )
		m_pkStream->write( (char*)&ucData, 1 );
	else
		*m_pkStream << ucData;

	return( *this );
}


File &File::operator << ( int16_t sData )
{
	if( m_bBinary )
	{
		if( m_eByteOrder == Core::BYTEORDER_SYSTEM )
			m_pkStream->write( (char*)&sData, 2 );
		else
		{
			char *pData = ( (char*)&sData ) + 1;
			m_pkStream->write( pData--, 1 );
			m_pkStream->write( pData,   1 );
		}
	}
	else
		*m_pkStream << sData;
	
	return( *this );
}


File &File::operator << ( uint16_t usData )
{
	if( m_bBinary )
	{
		if( m_eByteOrder == Core::BYTEORDER_SYSTEM )
			m_pkStream->write( (char*)&usData, 2 );
		else
		{
			char *pData = ( (char*)&usData ) + 1;
			m_pkStream->write( pData--, 1 );
			m_pkStream->write( pData,   1 );
		}
	}
	else
		*m_pkStream << usData;
	
	return( *this );
}


File &File::operator << ( int32_t iData )
{
	if( m_bBinary )
	{
		if( m_eByteOrder == Core::BYTEORDER_SYSTEM )
			m_pkStream->write( (char*)&iData, 4 );
		else
		{
			char *pData = ( (char*)&iData ) + 3;
			m_pkStream->write( pData--, 1 );
			m_pkStream->write( pData--, 1 );
			m_pkStream->write( pData--, 1 );
			m_pkStream->write( pData,   1 );
		}
	}
	else
		*m_pkStream << iData;
	
	return( *this );
}


File &File::operator << ( uint32_t uiData )
{
	if( m_bBinary )
	{
		if( m_eByteOrder == Core::BYTEORDER_SYSTEM )
			m_pkStream->write( (char*)&uiData, 4 );
		else
		{
			char *pData = ( (char*)&uiData ) + 3;
			m_pkStream->write( pData--, 1 );
			m_pkStream->write( pData--, 1 );
			m_pkStream->write( pData--, 1 );
			m_pkStream->write( pData,   1 );
		}
	}
	else
		*m_pkStream << uiData;
	
	return( *this );
}


File &File::operator << ( float fData )
{
	if( m_bBinary )
	{
		if( m_eByteOrder == Core::BYTEORDER_SYSTEM )
			m_pkStream->write( (char*)&fData, 4 );
		else
		{
			char *pData = ( (char*)&fData ) + 3;
			m_pkStream->write( pData--, 1 );
			m_pkStream->write( pData--, 1 );
			m_pkStream->write( pData--, 1 );
			m_pkStream->write( pData,   1 );
		}
	}
	else
		*m_pkStream << fData;
	
	return( *this );
}


File &File::operator << ( const char *pszData )
{
	if( m_bBinary )
		do m_pkStream->write( pszData, 1 ); while( *pszData++ );
	else
		*m_pkStream << pszData;

	return( *this );
}


File &File::operator << ( const string &rstrData )
{
	if( m_bBinary )
		*this << rstrData.c_str();
	else
		*m_pkStream << rstrData;

	return( *this );
}


File &File::operator << ( const HashString &rstrData )
{
	return( *this << rstrData.m_strData );
}


File &File::operator >> ( int8_t &rcData )
{
	if( m_bBinary )
	{
		char c;
		m_pkStream->read( &c, 1 );
		rcData = c;
	}
	else
		*m_pkStream >> rcData;
	
	return( *this );
}


File &File::operator >> ( bool &rbData )
{
	if( m_bBinary )
	{
		unsigned char ucVal = 0;
		m_pkStream->read( (char*)&ucVal, 1 );
		rbData = ( ucVal ? true : false );
	}
	else
	{
		string strVal;
		*m_pkStream >> strVal;
		rbData = ( ( strVal == "false" ) ? false : true );
	}
	
	return( *this );
}


File &File::operator >> ( uint8_t &rucData )
{
	if( m_bBinary )
		m_pkStream->read( (char*)&rucData, 1 );
	else
		*m_pkStream >> rucData;

	return( *this );
}


File &File::operator >> ( int16_t &rsData )
{
	if( m_bBinary )
	{
		if( m_eByteOrder == Core::BYTEORDER_SYSTEM )
			m_pkStream->read( (char*)&rsData, 2 );
		else
		{
			char *pData = ( (char*)&rsData ) + 1;
			m_pkStream->read( pData--, 1 );
			m_pkStream->read( pData,   1 );
		}
	}
	else
		*m_pkStream >> rsData;
	
	return( *this );
}


File &File::operator >> ( uint16_t &rusData )
{
	if( m_bBinary )
	{
		if( m_eByteOrder == Core::BYTEORDER_SYSTEM )
			m_pkStream->read( (char*)&rusData, 2 );
		else
		{
			char *pData = ( (char*)&rusData ) + 1;
			m_pkStream->read( pData--, 1 );
			m_pkStream->read( pData,   1 );
		}
	}
	else
		*m_pkStream >> rusData;
	
	return( *this );
}


File &File::operator >> ( int32_t &riData )
{
	if( m_bBinary )
	{
		if( m_eByteOrder == Core::BYTEORDER_SYSTEM )
			m_pkStream->read( (char*)&riData, 4 );
		else
		{
			char *pData = ( (char*)&riData ) + 3;
			m_pkStream->read( pData--, 1 );
			m_pkStream->read( pData--, 1 );
			m_pkStream->read( pData--, 1 );
			m_pkStream->read( pData,   1 );
		}
	}
	else
		*m_pkStream >> riData;
	
	return( *this );
}


File &File::operator >> ( uint32_t &ruiData )
{
	if( m_bBinary )
	{
		if( m_eByteOrder == Core::BYTEORDER_SYSTEM )
			m_pkStream->read( (char*)&ruiData, 4 );
		else
		{
			char *pData = ( (char*)&ruiData ) + 3;
			m_pkStream->read( pData--, 1 );
			m_pkStream->read( pData--, 1 );
			m_pkStream->read( pData--, 1 );
			m_pkStream->read( pData,   1 );
		}
	}
	else
		*m_pkStream >> ruiData;
	
	return( *this );
}


File &File::operator >> ( float &rfData )
{
	if( m_bBinary )
	{
		if( m_eByteOrder == Core::BYTEORDER_SYSTEM )
			m_pkStream->read( (char*)&rfData, 4 );
		else
		{
			char *pData = ( (char*)&rfData ) + 3;
			m_pkStream->read( pData--, 1 );
			m_pkStream->read( pData--, 1 );
			m_pkStream->read( pData--, 1 );
			m_pkStream->read( pData,   1 );
		}
	}
	else
		*m_pkStream >> rfData;
	
	return( *this );
}


File &File::operator >> ( string &rstrData )
{
	if( m_bBinary )
	{
		static const int uiSize = 128;

		char  acData[uiSize];
		char *pcData = acData;
		char *pcEnd  = acData + uiSize;

		while( true )
		{
			if( !m_pkStream->read( pcData, 1 ) )
				break;

			if( !*pcData )
				break;

			if( ++pcData == pcEnd )
			{
				rstrData.append( acData, uiSize );
				pcData = acData;
			}
		}

		if( pcData != acData )
			rstrData.append( acData, pcData - acData );
	}
	else
		*m_pkStream >> rstrData;

	return( *this );
}


File &File::operator >> ( HashString &rstrData )
{
	*this >> rstrData.m_strData;

	rstrData.HashMe();

	return( *this );
}


File &File::Read( void *pBuffer, int iNumBytes )
{
	m_pkStream->read( (char*)pBuffer, iNumBytes );

	return( *this );
}


File &File::Write( const void *pBuffer, int iNumBytes )
{
	m_pkStream->write( (char*)pBuffer, iNumBytes );

	return( *this );
}


bool File::DetermineByteOrder( int iReferenceValue )
{
	bool bBinary   = IsBinary();
	int  iCurValue = 0;

	int iCurPos = Tellg();

	SetBinary( true );

	*this >> iCurValue;

	SetBinary( bBinary );

	Seekg( iCurPos, ios_base::beg );

	if( iCurValue == iReferenceValue )
		return true; //we have correct byte ordering

	//try swapping bytes
	unsigned char cTemp;
	unsigned char *pucCur = (unsigned char*)&iCurValue;
	unsigned char *pucSwap = pucCur + ( sizeof( int ) - 1 );

	while( pucCur < pucSwap )
	{
		cTemp = *pucCur;

		*pucCur  = *pucSwap;
		*pucSwap = cTemp;

		++pucCur;
		--pucSwap;
	}

	if( iCurValue == iReferenceValue )
	{
		//we should be other byte order
		SetByteOrder( ( GetByteOrder() == Core::BYTEORDER_LITTLEENDIAN ) ? Core::BYTEORDER_BIGENDIAN : Core::BYTEORDER_LITTLEENDIAN );
		return true;
	}

	//Unable to determine byte order, possible bad reference value
	neolog << LogLevel( WARNING ) << "*** unable to determine file [" << GetName() << "] byte order: Bad values, ref[ " << hex << iReferenceValue << " ], swapped read [ " << iCurValue << " ]" << dec << endl;

	return false;
}


bool File::SetBinary( bool bBinary )
{
	bool bRet = IsBinary();

	m_bBinary = bBinary;

	return bRet;
}


bool File::IsBinary() const
{
	return m_bBinary;
}


bool File::DetermineBinaryMode( const unsigned char *pucReferenceValues )
{
	unsigned char aucRead[4];

	streampos kCurPos = m_pkStream->tellg();

	m_pkStream->read( (char*)aucRead, 4 );

	m_pkStream->seekg( kCurPos );

	return( m_bBinary = ( ( aucRead[0] == pucReferenceValues[0] ) && ( aucRead[1] == pucReferenceValues[1] ) && ( aucRead[2] == pucReferenceValues[2] ) && ( aucRead[3] == pucReferenceValues[3] ) ) );
}


std::string File::ExtractBaseFileName( const std::string &rstrPath )
{
	int    iDelim;
	string strBase = ExtractFileName( rstrPath );

	return( ( (std::string::size_type)( iDelim = strBase.find_last_of( "." ) ) != string::npos ) ? strBase.substr( 0, iDelim ) : strBase );
}


std::string File::ExtractFileExtension( const std::string &rstrPath )
{
	int    iDelim;
	string strBase = ExtractFileName( rstrPath );

	return( ( (std::string::size_type)( iDelim = strBase.find_last_of( "." ) ) != string::npos ) ? strBase.substr( iDelim + 1 ) : "" );
}


std::string File::ExtractFileName( const std::string &rstrPath )
{
	std::string::size_type iDelim = rstrPath.find_last_of( "\\" );

	if( iDelim == string::npos )
		iDelim = rstrPath.find_last_of( "/" );

	return( ( iDelim != string::npos ) ? rstrPath.substr( iDelim + 1, string::npos ) : rstrPath );
}


std::string File::ExtractPathName( const std::string &rstrPath )
{
	std::string::size_type iDelim = rstrPath.find_last_of( "\\" );

	if( iDelim == string::npos )
		iDelim = rstrPath.find_last_of( "/" );

	return( ( iDelim != string::npos ) ? rstrPath.substr( 0, iDelim ) : "" );
}





BufferFileStreamBuf::BufferFileStreamBuf( unsigned char *pucBuffer, unsigned int uiSize ) :
	basic_streambuf< char >(),
	m_pucBuffer( 0 ),
	m_uiSize( 0 )
{
	pubsetbuf( (char*)pucBuffer, uiSize );
}


BufferFileStreamBuf::~BufferFileStreamBuf()
{
	pubsetbuf( 0, 0 );
}


basic_streambuf< char > *BufferFileStreamBuf::setbuf( char *pcBuffer, streamsize uiSize )
{
	delete [] m_pucBuffer;

	m_pucBuffer = (unsigned char*)pcBuffer;
	m_uiSize    = uiSize;

	setg( (char*)m_pucBuffer, (char*)m_pucBuffer, (char*)( m_pucBuffer + m_uiSize ) );

	return this;
}


BufferFileStreamBuf::pos_type BufferFileStreamBuf::seekoff( BufferFileStreamBuf::off_type iOffset, ios_base::seekdir uiDir, ios_base::openmode uiMode )
{
	if( uiMode & ios_base::in )
	{
		int iCur = gptr() - eback();

		if( uiDir == ios_base::beg )
		{
			if( ( iOffset < (int)m_uiSize ) && ( iOffset >= 0 ) )
				setg( (char*)m_pucBuffer, (char*)( m_pucBuffer + iOffset ), (char*)( m_pucBuffer + m_uiSize ) );
			else
				return( basic_streambuf<char>::seekoff( iOffset, uiDir, uiMode ) );

			return( streampos( iOffset ) );
		}
		else if( uiDir == ios_base::cur )
		{
			if( iOffset == 0 )
				return( streampos( iCur ) );
			else if( ( iOffset > 0 ) && ( ( iCur + iOffset ) < (int)m_uiSize ) )
				setg( (char*)m_pucBuffer, (char*)( m_pucBuffer + iOffset + iCur ), (char*)( m_pucBuffer + m_uiSize ) );
			else if( ( iOffset < 0 ) && ( ( iCur + iOffset ) >= 0 ) )
				setg( (char*)m_pucBuffer, (char*)( m_pucBuffer + iOffset + iCur ), (char*)( m_pucBuffer + m_uiSize ) );
			else
				return( basic_streambuf<char>::seekoff( iOffset, uiDir, uiMode ) );

			return( streampos( iOffset + iCur ) );
		}
		else // end
		{
			if( ( iOffset < (int)m_uiSize ) && ( iOffset >= 0 ) )
				setg( (char*)m_pucBuffer, (char*)( m_pucBuffer + m_uiSize - iOffset ), (char*)( m_pucBuffer + m_uiSize ) );
			else
				return( basic_streambuf<char>::seekoff( iOffset, uiDir, uiMode ) );

			return( streampos( m_uiSize - iOffset ) );
		}
	}

	if( uiMode & ios_base::out )
	{

	}

	return( basic_streambuf<char>::seekoff( iOffset, uiDir, uiMode ) );
}


BufferFileStreamBuf::pos_type BufferFileStreamBuf::seekpos( BufferFileStreamBuf::pos_type uiPosition, ios_base::openmode uiMode )
{
	return seekoff( uiPosition, ios_base::beg, uiMode );
}


BufferFileStreamBuf::int_type BufferFileStreamBuf::underflow()
{
	return basic_streambuf< char >::underflow();
}





BufferFile::BufferFile( unsigned char *pucBuffer, int iSize, const std::string &rstrPath, const std::string &rstrFilename, std::ios_base::openmode iOpenFlags, bool bOpen, Directory *pkDirectory ) :
	File( rstrPath, rstrFilename, iOpenFlags, false, pkDirectory ),
	m_pkBuffer( new BufferFileStreamBuf( pucBuffer, iSize ) )
{
	if( bOpen )
		Open();
}


BufferFile::~BufferFile()
{
	delete m_pkBuffer;
}


void BufferFile::AllocStream( const std::string &rstrFullPath )
{
	m_pkStream = new iostream( m_pkBuffer );
}


void BufferFile::Close()
{
	if( IsValid() )
		SetBuffer( 0, 0 );

	File::Close();
}


void BufferFile::SetBuffer( unsigned char *pucBuffer, int iSize )
{
	m_pkBuffer->pubsetbuf( (char*)pucBuffer, iSize );
}








VirtualFile::VirtualFile( const std::string &rstrPath, const std::string &rstrFilename, Directory *pkDirectory ) :
	BufferFile( 0, 0, rstrPath, rstrFilename, ios_base::in | ios_base::binary, false, pkDirectory )
{
	m_iFileOffset     = 0;
	m_iFileFlags      = 0;
	m_pkPackage       = 0;
	m_iFileData[0]    = 0;
	m_iFileData[1]    = 0;
	m_iFileData[2]    = 0;
	m_iFileData[3]    = 0;
}


VirtualFile::~VirtualFile()
{
}


void VirtualFile::AllocStream( const std::string &rstrFullPath )
{
	if( ( m_iFileFlags == Package::FILEFLAG_BZIP2_COMPRESSED )
		|| ( m_iFileFlags == Package::FILEFLAG_ZIP_COMPRESSED ))
	{
		FileCodec *pkCodec = pkCodec = FileManager::FindCodec( m_iFileFlags );

		if( !pkCodec )
		{
			neolog << LogLevel( ERROR ) << "*** invalid virtual file: unable to locate codec [" << m_iFileFlags << "]" << endl;
			return;
		}

		if( m_iFileData[0] <= 0 )
		{
			neolog << LogLevel( ERROR ) << "*** invalid virtual file: invalid real size of file" << endl;
			return;
		}

		unsigned char *pucBuffer = new unsigned char[ m_iFileData[0] + 1 ];

		pucBuffer[ m_iFileData[0] ] = 0;

		int iBufSize = m_iFileData[0];

		int iCurOffset = m_pkPackage->GetBaseFile()->Tellg();

		m_pkPackage->GetBaseFile()->Seekg( m_iFileOffset, ios_base::beg );

		if( pkCodec->LoadFileData( m_pkPackage->GetBaseFile(), m_iStoredFileSize, (void*)pucBuffer, &iBufSize ) )
		{
			neolog << LogLevel( ERROR ) << "*** invalid virtual file: Error in uncompress" << endl;

			delete [] pucBuffer, pucBuffer = 0;

			return;
		}

		SetBuffer( pucBuffer, m_iFileData[0] );

		m_pkPackage->GetBaseFile()->Seekg( iCurOffset, ios_base::beg );
	}
	else if( m_iFileFlags )
	{
		neolog << LogLevel( ERROR ) << "*** invalid virtual file: File [" << m_strName << "] has invalid file flags [" << m_iFileFlags << "]" << endl;
		return;
	}
	else
	{
		unsigned char *pucBuffer = new unsigned char[ m_iStoredFileSize + 1 ];

		pucBuffer[ m_iStoredFileSize ] = 0;

		int iCurOffset = m_pkPackage->GetBaseFile()->Tellg();

		m_pkPackage->GetBaseFile()->Seekg( m_iFileOffset, ios_base::beg );

		m_pkPackage->GetBaseFile()->Read( pucBuffer, m_iStoredFileSize );

		SetBuffer( pucBuffer, m_iStoredFileSize );

		m_pkPackage->GetBaseFile()->Seekg( iCurOffset, ios_base::beg );

	}

	BufferFile::AllocStream( rstrFullPath );
}


}; // namespace NeoEngine
