/***************************************************************************
                 filemanager.cpp  -  File and package management
                             -------------------
    begin                : Fri Nov 2 2001
    copyright            : (C) 2001 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, filemanager.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2001
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "filemanager.h"
#include "filecodec.h"
#include "module.h"
#include "file.h"
#include "directory.h"
#include "package.h"
#include "npacpackage.h"
#include "zippackage.h"
#include "strutil.h"
#include "logstream.h"

#include <stdio.h>

using namespace std;


namespace NeoEngine
{

bool                 g_bLoadedBZIP2 = false;
bool                 g_bLoadedZIP = false;
vector< FileCodec* > g_vpkFileCodecs;
int                  FileManager::s_iFileManagers = 0;


FileManager::FileManager( const std::string &rstrName ) :
	m_strName( rstrName ),
	m_pkRoot( 0 ),
	m_bChanged( false )
{
	Clear();

	s_iFileManagers++;

	m_bChanged = false;
}


FileManager::~FileManager()
{
	if( !--s_iFileManagers )
	{
		//Clean up codecs
		vector< FileCodec* >::iterator ppkCodec    = g_vpkFileCodecs.begin();
		vector< FileCodec* >::iterator ppkCodecEnd = g_vpkFileCodecs.end();

		for( ; ppkCodec != ppkCodecEnd; ++ppkCodec )
		{
			//By creating new reference to module, it won't get deleted inside codec dtor,
			//since that is a Bad Thing<tm> - the dtor is inside module space
			ModulePtr pkModule = (*ppkCodec)->m_pkModule;

			delete( *ppkCodec );
		}

		g_vpkFileCodecs.clear();
	}

	delete m_pkRoot;
}


bool FileManager::AddPackage( const std::string &rstrPackage )
{
	string strPackage = rstrPackage.substr( 0, rstrPackage.find_last_not_of( "/\\" ) + 1 );

	string strPath = File::ExtractPathName( strPackage );
	string strFile = File::ExtractFileName( strPackage );

	Directory *pkDirectory     = 0;
	Directory *pkLastDirectory = 0;
	Directory *pkNewDirectory  = 0;
	Directory *pkPackage       = 0;

	//Make sure path exists
	if( !( pkDirectory = strPath.length() ? m_pkRoot->GetDirectoryByName( strPath ) : m_pkRoot ) )
	{
		pkDirectory = pkLastDirectory = m_pkRoot;

		//Create placeholder directory structure
		vector< string > vstrPaths;

		if( ( strPath[0] == '/' ) || ( strPath[0] == '\\' ) )
			vstrPaths.push_back( "/" );

		Explode( "/\\", strPath, &vstrPaths );

		vector< string >::iterator pstrCurPath = vstrPaths.begin();
		vector< string >::iterator pstrPathEnd = vstrPaths.end();

		for( ; pstrCurPath != pstrPathEnd; ++pstrCurPath, pkLastDirectory = pkDirectory )
			if( !( pkDirectory = pkDirectory->GetDirectoryByName( *pstrCurPath ) ) )
				break;

		assert( pstrCurPath != pstrPathEnd ); //Or dir would have been located in first test above

		pkDirectory = pkNewDirectory = new Directory( *pstrCurPath++, pkLastDirectory, false );

		for( ; pstrCurPath != pstrPathEnd; ++pstrCurPath )
			pkDirectory = new Directory( *pstrCurPath, pkDirectory, false );

		m_bChanged = true;
	}

	if( !pkDirectory->IsParsed() )
	{
		pkDirectory->Parse( false );
		m_bChanged = true;
	}

	//Locate already loaded directory/package with same path and name
	if( ( pkPackage = pkDirectory->GetDirectoryByName( strFile ) ) )
	{
		if( !pkPackage->IsParsed() )
			pkPackage->Parse( true );
		else
			return true;
	}
	else
	{
		if( ( pkPackage = pkDirectory->GetDirectoryByName( File::ExtractBaseFileName( strFile ) ) ) && pkPackage->IsPackage() )
			return true; //Already loaded

		File *pkFile = pkDirectory->GetByName( strFile, false );

		if( pkFile || pkDirectory->GetDirectoryByName( strFile ) )
		{
			delete pkFile;

			pkPackage = DeterminePackageType( pkDirectory, strFile, this );
		}
		else if( !strPath.length() )
		{
			//Try to locate file
			File *pkFile = m_pkRoot->GetByName( strFile, true );

			if( pkFile )
			{
				pkDirectory = pkFile->GetDirectory();

				delete pkFile;

				if( ( pkPackage = pkDirectory->GetDirectoryByName( File::ExtractBaseFileName( strFile ) ) ) && pkPackage->IsPackage() )
					return true; //Already loaded

				pkPackage = DeterminePackageType( pkDirectory, strFile, this );
			}
		}
		else
			pkPackage = 0;

		if( !pkPackage || !pkPackage->IsParsed() )
		{
			delete pkPackage;
			return false;
		}
	}

	neolog << LogLevel( DEBUG ) << "package [" << pkPackage->GetFullPath() << "] loaded" << endl;

	m_bChanged = true;

	return true;
}

Package* FileManager::DeterminePackageType( Directory *pkParent, const string &rstrFile, FileManager *pkManager )
{
	// prepare pathnames
	string strPackage = rstrFile.substr( 0, rstrFile.find_last_not_of( "/\\" ) + 1 );
	string strPath = File::ExtractPathName( strPackage );
	string strFile = File::ExtractFileName( strPackage );

	// open file
	File *pkFile = pkParent->GetByName( rstrFile );
	if ( !pkFile )
	{
		neolog << LogLevel( ERROR ) << "[FileManager::DeterminePackageType]: File not found [" << rstrFile << "]" << endl;
		return 0;
	}
	if( !pkFile->Open() )
	{
		delete pkFile;
		neolog << LogLevel( WARNING ) << "[FileManager::DeterminePackageType]: Unable to Open() file [" << ( pkFile->GetPath().length() ? pkFile->GetPath() + "/" : "" ) << pkFile->GetName() << "]" << endl;
		return 0;
	}

	vector< FileCodec* >::iterator ppkCodec    = g_vpkFileCodecs.begin();
	vector< FileCodec* >::iterator ppkCodecEnd = g_vpkFileCodecs.end();

	// search for filetype through loaded codecs
	for( ; ppkCodec != ppkCodecEnd; ++ppkCodec )
	{
		if ( (**ppkCodec).IsType( pkFile ) )
		{
			if ( (**ppkCodec).GetTypeName() == "bzip2 compressed file" )
			{
				return new NpacPackage( pkParent, rstrFile, pkManager );
			} 
			else if ( (**ppkCodec).GetTypeName() == "zip compressed file" )
			{
				return new ZipPackage( pkParent, rstrFile, pkManager );
			}
		}
	}

	pkFile->Close();
	delete pkFile;

	neolog << LogLevel( WARNING ) << "[FileManager::DeterminePackageType]: No matching file codec found File[" << strFile << "]" << endl;
	return 0;
}

void FileManager::PrintHierarchy( int iDepth )
{
    unsigned int i;

	if( iDepth < 0 )
		iDepth = 0;
	
	for( i = 0; i < (unsigned int)iDepth; ++i )
		neolog << LogLevel( DEBUG ) << "  ";

	neolog << m_strName << endl;

	m_pkRoot->PrintHierarchy( iDepth + 1 );
}


File *FileManager::GetByName( const std::string &rstrName )
{
	return m_pkRoot->GetByName( rstrName );
}


FileCodec *FileManager::FindCodec( int iFileFlag )
{
	/* FileCodecs have to be loaded before AddPackage is called
	if( !g_bLoadedBZIP2 && ( iFileFlag == Package::FILEFLAG_BZIP2_COMPRESSED ) )
	{
		LoadCodec( "bzip2" );
		g_bLoadedBZIP2 = true;
	}
	if( !g_bLoadedZIP && ( iFileFlag == Package::FILEFLAG_ZIP_COMPRESSED ) )
	{
		LoadCodec( "zip" );
		g_bLoadedZIP = true;
	}
	*/

	vector< FileCodec* >::iterator ppkCodec    = g_vpkFileCodecs.begin();
	vector< FileCodec* >::iterator ppkCodecEnd = g_vpkFileCodecs.end();

	for( ; ppkCodec != ppkCodecEnd; ++ppkCodec )
		if( (*ppkCodec)->m_iFileFlag == iFileFlag )
			return( *ppkCodec );

	return 0;
}


void FileManager::Clear()
{
	delete m_pkRoot;

	m_pkRoot = new Directory( "", 0, false );

	m_bChanged = true;
}


bool FileManager::HasChanged()
{
	bool bRet = m_bChanged;

	m_bChanged = false;

	return bRet;
}


bool FileManager::LoadCodec( const std::string &rstrName )
{
	ModSymbol kSymLoad;

#if defined(WIN32) && defined(_DEBUG)
	ModulePtr pkModule = Core::Get()->GetModuleManager()->LoadModule( "neofc" + rstrName + "_debug" );
#else
	ModulePtr pkModule = Core::Get()->GetModuleManager()->LoadModule( "neofc" + rstrName );
#endif

	if( !pkModule || !pkModule->IsValid() )
		return false;

	kSymLoad = pkModule->LookupSymbol( "LoadFileCodec" );

	if( !kSymLoad )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to load [" << rstrName << "] file codec: Codec does not export LoadFileCodec() method" << endl;
		return false;        
	}

	FileCodec *pkCodec = ( ( FileCodec* ( NE_CDECL* )() )kSymLoad )();

	if( !pkCodec )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to load [" << rstrName << "] file codec" << endl;
		return false;
	}

	g_vpkFileCodecs.insert( g_vpkFileCodecs.begin(), pkCodec );

	pkCodec->m_pkModule = pkModule;

	return true;
}


};
