/***************************************************************************
                  font.cpp  -  Font class for text rendering
                             -------------------
    begin                : Fri Nov 16 2001
    copyright            : (C) 2001 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, font.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2001
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "font.h"
#include "file.h"
#include "nemath.h"
#include "render.h"
#include "renderprimitive.h"
#include "vertex.h"
#include "polygon.h"
#include "logstream.h"

#include <stdarg.h>
#include <stdio.h>

#include <string>

#if defined( WIN32 ) && defined( _MSC_VER )
#define vsnprintf _vsnprintf
#endif

using namespace std;

//#define NEFONT_DEBUGLOADER

namespace NeoEngine
{


float NE_STATIC Font::s_fSubPixelOffset = 0.0f;


Font::Font( FontManager *pkFontManager ) :
	LoadableEntity()
{
	m_kColor.Set( 1.0f, 1.0f, 1.0f, 1.0f );

	m_iSpacing          = 0;
	m_iSpaceWidth       = 6;
	m_iLineHeight       = 12;
	m_eLineWrap         = NOWRAP;
	m_pkFontManager     = pkFontManager;
	m_strWordSeparators = " \n\r";
	m_eHorizontalAlign  = LEFT;

	ResetClipBox();
}



Font::~Font()
{
	if( m_pkFontManager )
		m_pkFontManager->DeregisterFont( this );

	m_pkFontManager = 0;
}


bool Font::LoadNode( unsigned int uiFlags )
{
	if( !m_pkFile->IsValid() )
		return false;

	for( int i = 0; i < 256; ++i )
		m_akCharacters[i].m_iWidth = 0;

	m_strName = m_pkFile->GetName().substr( 0, m_pkFile->GetName().find_last_of( '.' ) );

	neolog << LogLevel( DEBUG ) << "parsing font file [" << m_pkFile->GetName() << "] - [" << m_strName << "]" << endl;

	//Load font file
	m_pkFile->SetByteOrder( Core::BYTEORDER_LITTLEENDIAN );
	m_pkFile->SetBinary( false );

	char szID[5]; szID[4] = 0;

	m_pkFile->Read( szID, 4 );

	if( string( szID ) != "NFON" )
	{
		neolog << LogLevel( ERROR ) << "[Font::LoadEntity] invalid ID" << endl;

		return false;
	}

	//File ID and version ok, load
	string strTexture;

	*m_pkFile >> strTexture;

	TexturePtr pkTexture = Core::Get()->GetRenderDevice()->LoadTexture( strTexture, Texture::TEX2D, Texture::DEFAULT, Texture::NOMIPMAPS );

	if( !pkTexture )
	{
		neolog << LogLevel( ERROR ) << "[Font::LoadEntity] invalid texture" << endl;
		return false;
	}

	if( !m_pkMaterial )
		m_pkMaterial = new Material( "font", 0 );

	m_pkMaterial->m_pkTexture = pkTexture;
	m_pkMaterial->m_kBlendMode.Set( BlendMode::DECAL );
	m_pkMaterial->m_kZBufferMode.Set( ZBufferMode::ALWAYSNOWRITE );

	*m_pkFile >> m_iLineHeight >> m_iSpaceWidth;

	#ifdef NEFONT_DEBUGLOADER
	neolog << LogLevel( DEBUG ) << "line height [" << m_iLineHeight << "]" << endl
	       << "space width [" << m_iSpaceWidth << "]" << endl;
	#endif

	while( true )
	{
		string strChar;
		int iX      = 0;
		int iY      = 0;
		int iWidth  = 0;
		int iHeight = 0;

		if( !( *m_pkFile >> strChar >> iX >> iY >> iWidth >> iHeight ) )
			break;

		unsigned char cChar = strChar[0];

		//Store character
		m_akCharacters[cChar].m_iX       = iX;
		m_akCharacters[cChar].m_iY       = iY;
		m_akCharacters[cChar].m_iWidth   = iWidth;
		m_akCharacters[cChar].m_iHeight  = iHeight;

		float fOOWidth                   = 1.0f / float( pkTexture->GetWidth()  );
		float fOOHeight                  = 1.0f / float( pkTexture->GetHeight() );

		m_akCharacters[cChar].m_fU       = ( float( iX ) + s_fSubPixelOffset ) * fOOWidth;
		m_akCharacters[cChar].m_fV       = ( float( iY ) + s_fSubPixelOffset ) * fOOHeight;

		float fUOffset                   = 2.0f * ( iWidth  > 0 ) ? -s_fSubPixelOffset : ( ( iWidth  < 0 ) ? s_fSubPixelOffset : 0.0f );
		float fVOffset                   = 2.0f * ( iHeight > 0 ) ? -s_fSubPixelOffset : ( ( iHeight < 0 ) ? s_fSubPixelOffset : 0.0f );

		m_akCharacters[cChar].m_fUWidth  = ( float( iWidth  ) + fUOffset ) * fOOWidth;
		m_akCharacters[cChar].m_fVHeight = ( float( iHeight ) + fVOffset ) * fOOHeight;

		#ifdef NEFONT_DEBUGLOADER
		        neolog << LogLevel( DEBUG )
			       << "[DEBUG] char '" << cChar << "' read" << endl
			       << "  x [ " << m_akCharacters[cChar].m_iX << " ]" << endl
			       << "  y [ " << m_akCharacters[cChar].m_iY << " ]" << endl
			       << "  u [ " << setprecision( 3 ) << m_akCharacters[cChar].m_fU << " ]" << endl
			       << "  v [ " << m_akCharacters[cChar].m_fU << " ]" << endl
			       << "  width   [ " << m_akCharacters[cChar].m_iWidth << " ]" << endl
			       << "  height  [ " << m_akCharacters[cChar].m_iHeight << " ]" << endl
			       << "  uwidth  [ " << m_akCharacters[cChar].m_fUWidth << " ]" << endl
			       << "  vheight [ " << m_akCharacters[cChar].m_fVHeight << " ]" << endl;
		#endif
	}

	m_akCharacters[ (unsigned char)' ' ].m_iWidth = m_iSpaceWidth;

	return true;
}


int Font::Printf( float fX, float fY, const char *pszText, ... )
{
#ifdef NEOENGINE_THREAD_SAFE
#  ifdef WIN32
	__declspec( thread ) static char szStr[256];
#  else
#    error "Thread-safe storage not implemented on this platform"
#  endif
#else
	static char szStr[256];
#endif

	va_list arg;

	va_start( arg, pszText );

	char *pszMem = szStr;

	int iWritten;
	int iResize = sizeof( szStr );

	while( ( ( iWritten = vsnprintf( pszMem, iResize, pszText, arg ) ) == -1 ) || ( iWritten > iResize ) )
	{
		iResize = ( iWritten == -1 ) ? ( iResize * 4 ) : ( iWritten + 1 );

		if( pszMem != szStr )
			delete [] pszMem;

		pszMem = new char[ iResize ];
	}

	va_end( arg );

	int iX = int( fX * (float)Core::Get()->GetRenderDevice()->GetWidth() );
	int iY = int( fY * (float)Core::Get()->GetRenderDevice()->GetHeight() );

	int iRet = Printf( iX, iY, pszMem );

	if( pszMem != szStr )
		delete [] pszMem;

	return iRet;
}


int Font::Printf( int iX, int iY, const char *pszText, ... )
{
#ifdef NEOENGINE_THREAD_SAFE
#  ifdef WIN32
	__declspec( thread ) static char szStr[256];
#  else
#    error "Thread-safe storage not implemented on this platform"
#  endif
#else
	static char szStr[256];
#endif

	va_list arg;

	va_start( arg, pszText );

	char *pszMem = szStr;

	int iWritten;
	int iResize = sizeof( szStr );

	while( ( ( iWritten = vsnprintf( pszMem, iResize, pszText, arg ) ) == -1 ) || ( iWritten > iResize ) )
	{
		iResize = ( iWritten == -1 ) ? ( iResize * 4 ) : ( iWritten + 1 );

		if( pszMem != szStr )
			delete [] pszMem;

		pszMem = new char[ iResize ];
	}

	va_end( arg );

	//Clip line to clip box
	int iLines;
	string strClipped( ClipLine( pszMem, iX, iY, &iLines ) );

	if( pszMem != szStr )
		delete [] pszMem;

	if( !strClipped.length() )
		return 0;

	int iStartX = iX;

	//Build vertex/poly buffers and send to device

	//Four vertices per character
	VertexBufferPtr pkBuffer = Core::Get()->GetRenderDevice()->CreateVertexBuffer( Buffer::DYNAMIC, strClipped.length() * 4, &DiffuseTexVertex::s_kDecl );
	PolygonBufferPtr pkPolys = Core::Get()->GetRenderDevice()->CreatePolygonBuffer( Buffer::DYNAMIC, strClipped.length() * 2 );  //Two polys per character

	unsigned short usVert = 0;
	unsigned int   uiPoly = 0;

	const Viewport &rkViewport = Core::Get()->GetRenderDevice()->GetViewport();

	assert( rkViewport.m_iWidth > 0 && rkViewport.m_iHeight > 0 );
	assert( m_pkMaterial->m_pkTexture && m_pkMaterial->m_pkTexture->GetWidth() > 0 && m_pkMaterial->m_pkTexture->GetHeight() > 0 );

	float fOOWidth      = 1.0f / float( rkViewport.m_iWidth );
	float fOOHeight     = 1.0f / float( rkViewport.m_iHeight );

	float fOOTexWidth   = 1.0f / float( m_pkMaterial->m_pkTexture->GetWidth() );
	float fOOTexHeight  = 1.0f / float( m_pkMaterial->m_pkTexture->GetHeight() );

	int  iDirection     = ( ( m_eHorizontalAlign == LEFT ) ? 1 : -1 );
	bool bFirst         = true;

	pkBuffer->Lock( Buffer::WRITE );
	pkPolys->Lock( Buffer::WRITE );

	Polygon *pkPoly = pkPolys->GetPolygon( 0 );

	DiffuseTexVertex *pkVertex = (DiffuseTexVertex*)pkBuffer->GetVertex();

	//ClipLine will reverse string for right-to-left printing modes, so just loop normally

	for( unsigned int i = 0; i < strClipped.length(); ++i )
	{
		int iChar = (unsigned char)strClipped[i];

		if( iChar == ' ' )
		{
			iX += ( m_iSpaceWidth + m_iSpacing ) * iDirection;

			if( bFirst && ( m_eHorizontalAlign == RIGHT ) )
				iX += m_iSpacing; //Spacing not for first character when right-to-left printing

			bFirst = false;
		}
		else if( iChar == '\n' )
		{	
			iX  = iStartX;
			iY += m_iLineHeight;

			bFirst = true;

			continue;
		}
		else
		{
			//We know character is valid or ClipLine would have removed it
			assert( m_akCharacters[ iChar ].m_iWidth );

			if( m_eHorizontalAlign == RIGHT )
			{
				iX -= m_akCharacters[ iChar ].m_iWidth + m_iSpacing;

				if( bFirst )
					iX += m_iSpacing; //No spacing for first character
			}

			bFirst = false;

			if( ( ( iX + m_akCharacters[ iChar ].m_iWidth  ) >= m_iClipX ) && ( iX <= ( m_iClipX + m_iClipWidth  ) ) &&
			    ( ( iY + m_akCharacters[ iChar ].m_iHeight ) >= m_iClipY ) && ( iY <= ( m_iClipY + m_iClipHeight ) ) )
			{
				int iClipBeginX = 0;
				int iClipEndX   = 0;
				int iClipBeginY = 0;
				int iClipEndY   = 0;

				float fClipBeginU = 0.0f;
				float fClipEndU   = 0.0f;
				float fClipBeginV = 0.0f;
				float fClipEndV   = 0.0f;

				//Check for sub-clipping
				if( iX < m_iClipX )
				{
					iClipBeginX = m_iClipX - iX;
					fClipBeginU = float( iClipBeginX ) * fOOTexWidth;
				}

				if( ( iX + m_akCharacters[ iChar ].m_iWidth ) > ( m_iClipX + m_iClipWidth ) )
				{
					iClipEndX = ( iX + m_akCharacters[ iChar ].m_iWidth ) - ( m_iClipX + m_iClipWidth );
					fClipEndU = float( iClipEndX ) * fOOTexWidth;
				}

				if( iY < m_iClipY )
				{
					iClipBeginY = m_iClipY - iY;
					fClipBeginV = float( iClipBeginY ) * fOOTexHeight;
				}

				if( ( iY + m_akCharacters[ iChar ].m_iHeight ) > ( m_iClipY + m_iClipHeight ) )
				{
					iClipEndY = ( iY + m_akCharacters[ iChar ].m_iHeight ) - ( m_iClipY + m_iClipHeight );
					fClipEndV = float( iClipEndY ) * fOOTexHeight;
				}


				//Add two new start left-edge vertices to the buffer
				pkVertex->m_kPosition.Set( (float)( iX + iClipBeginX ) * fOOWidth, (float)( iY + iClipBeginY ) * fOOHeight, 0.0f );
				pkVertex->m_kColor = m_kColor;
				pkVertex->m_afTexCoord[0] = m_akCharacters[ iChar ].m_fU + fClipBeginU;
				pkVertex->m_afTexCoord[1] = m_akCharacters[ iChar ].m_fV + fClipBeginV;

				++pkVertex;
				++usVert;

				pkVertex->m_kPosition.Set( (float)( iX + iClipBeginX ) * fOOWidth, (float)( iY + m_akCharacters[ iChar ].m_iHeight - iClipEndY ) * fOOHeight, 0.0f );
				pkVertex->m_kColor = m_kColor;
				pkVertex->m_afTexCoord[0] = m_akCharacters[ iChar ].m_fU + fClipBeginU;
				pkVertex->m_afTexCoord[1] = m_akCharacters[ iChar ].m_fV + m_akCharacters[ iChar ].m_fVHeight - fClipEndV;

				++pkVertex;
				++usVert;

				//Add two right-edge vertices to the buffer
				pkVertex->m_kPosition.Set( (float)( iX + m_akCharacters[ iChar ].m_iWidth - iClipEndX ) * fOOWidth, (float)( iY + iClipBeginY ) * fOOHeight, 0.0f );
				pkVertex->m_kColor = m_kColor;
				pkVertex->m_afTexCoord[0] = m_akCharacters[ iChar ].m_fU + m_akCharacters[ iChar ].m_fUWidth - fClipEndU;
				pkVertex->m_afTexCoord[1] = m_akCharacters[ iChar ].m_fV + fClipBeginV;

				++pkVertex;
				++usVert;

				pkVertex->m_kPosition.Set( (float)( iX + m_akCharacters[ iChar ].m_iWidth - iClipEndX ) * fOOWidth, (float)( iY + m_akCharacters[ iChar ].m_iHeight -iClipEndY ) * fOOHeight, 0.0f );
				pkVertex->m_kColor = m_kColor;
				pkVertex->m_afTexCoord[0] = m_akCharacters[ iChar ].m_fU + m_akCharacters[ iChar ].m_fUWidth - fClipEndU;
				pkVertex->m_afTexCoord[1] = m_akCharacters[ iChar ].m_fV + m_akCharacters[ iChar ].m_fVHeight - fClipEndV;

				++pkVertex;
				++usVert;

				//Ok, make polys
				pkPoly->v[ 0 ] = usVert - 4;
				pkPoly->v[ 1 ] = usVert - 3;
				pkPoly->v[ 2 ] = usVert - 2;

				++pkPoly;
				++uiPoly;

				pkPoly->v[ 0 ] = usVert - 2;
				pkPoly->v[ 1 ] = usVert - 3;
				pkPoly->v[ 2 ] = usVert - 1;

				++uiPoly;
				++pkPoly;
			}

			//Move offset
			if( m_eHorizontalAlign == LEFT )
				iX += ( m_akCharacters[ iChar ].m_iWidth + m_iSpacing ) * iDirection;
			else
				iX -= m_iSpacing;
		}
	}

	pkPolys->Unlock();
	pkBuffer->Unlock();

	if( uiPoly )
	{
		assert( usVert == uiPoly * 2 );

		//Setup primitive batch data
		RenderPrimitive kData;

		kData.m_ePrimitive        = RenderPrimitive::TRIANGLES;
		kData.m_pkMaterial        = m_pkMaterial;
		kData.m_pkVertexBuffer    = pkBuffer;
		kData.m_pkPolygonBuffer   = pkPolys;
		kData.m_uiNumPrimitives   = uiPoly;

		Core::Get()->GetRenderDevice()->Render( kData, 0 );
	}
	
	return iLines;
}


string Font::ClipLine( const string &rstrLine, int iX, int iY, int *piLines )
{
	//Check if start is outside clip box on "far" border
	if( ( iX > ( m_iClipX + m_iClipWidth ) ) || ( iY > ( m_iClipY + m_iClipWidth ) ) )
	{
		if( piLines )
			*piLines = 0;

		return "";
	}

	int  iLines        = 0;
	int  iStartX       = iX;
	bool bLinefeed     = false;
	int  iDirection    = ( m_eHorizontalAlign == LEFT ) ? 1 : -1;
	bool bWordClipped  = false;

	string strResult;
	string strClip;

	int iStart         = ( m_eHorizontalAlign == LEFT ) ? 0 : ( rstrLine.length() - 1 );
	int iStop          = -1; if( m_eHorizontalAlign == LEFT ) iStop = rstrLine.length();

	bool bFirst        = true;

	//Loop through whole string
	for( int i = iStart; i != iStop; i += iDirection )
	{
		int iChar = (unsigned char)rstrLine[i];

		//If character is a word separator, check for word clipping
		if( m_strWordSeparators.find_first_of( rstrLine[i] ) != string::npos )
			bWordClipped = false;

		if( iChar == ' ' )
		{
			//Don't add spaces right after linefeed
			if( !bLinefeed )
			{
				iX += ( m_iSpaceWidth + m_iSpacing ) * iDirection;

				if( bFirst && ( m_eHorizontalAlign == RIGHT ) )
					iX += m_iSpacing; //Spacing not for first character when right-to-left printing

				strResult += ' ';

				bFirst = false;
			}
		}
		else if( iChar == '\n' )
		{	
			//Linefeed
			iX         = iStartX;
			iY        += m_iLineHeight;
			bLinefeed  = true;
			bFirst     = true; //xt is first character on new line

			strResult += '\n';

			++iLines;

			continue;
		}
		else if( m_akCharacters[ iChar ].m_iWidth > 0 )
		{
			if( m_eHorizontalAlign == RIGHT )
				iX -= m_akCharacters[ iChar ].m_iWidth + ( bFirst ? 0 : m_iSpacing ); //Back up

			bFirst = false; //Reset flag, we don't need it more

			if( ( m_eLineWrap == WORDWRAP ) && !bWordClipped )
			{
				if( m_eHorizontalAlign == LEFT )
				{
					strClip = rstrLine.substr( i, rstrLine.length() - i );

					strClip = strClip.substr( 0, strClip.find_first_of( m_strWordSeparators ) );
				
					int iWidth = 0;

					for( unsigned int iLoop = 0; iLoop < strClip.length(); ++iLoop )
						iWidth += m_akCharacters[ (unsigned char)strClip[ iLoop ] ].m_iWidth + m_iSpacing;

					iWidth -= m_iSpacing; //Remove spacing for last character

					if( ( iX + iWidth ) > ( m_iClipX + m_iClipWidth ) )
					{
						iX  = iStartX;
						iY += m_iLineHeight;

						strResult += '\n';

						bFirst = true;

						iLines++;
					}

					bWordClipped = true;
				}
				else // RIGHT
				{
					strClip = rstrLine.substr( 0, i + 1 );

					//Find separator
					int iSep = i - 1;

					while( ( iSep > 0 ) && ( m_strWordSeparators.find_first_of( strClip[ iSep ] ) != string::npos ) )
						--iSep;

					int iWidth = 0;

					for( int iLoop = i; iLoop >= iSep; --iLoop )
						iWidth += m_akCharacters[ (unsigned char)strClip[ iLoop ] ].m_iWidth + m_iSpacing;

					iWidth -= m_iSpacing; //Remove first characters spacing

					if( ( iX - iWidth ) < m_iClipX )
					{
						iX  = iStartX;
						iY += m_iLineHeight;

						strResult += '\n';

						bFirst = true;

						iLines++;
					}

					bWordClipped = true;
				}
			}

			if( ( m_eHorizontalAlign == LEFT ) && ( iX > ( m_iClipX + m_iClipWidth ) ) )
				break; //nothing more to do

			if( ( m_eHorizontalAlign == RIGHT ) && ( iX < m_iClipX ) )
				break; //nothing more to do

			if( iY > ( m_iClipY + m_iClipHeight ) )
				break; //nothing more to do

			//Move offset
			if( m_eHorizontalAlign == LEFT )
				iX += m_akCharacters[ iChar ].m_iWidth + m_iSpacing;
			else
				iX -= m_iSpacing;

			strResult += (char)iChar;
		}

		bLinefeed = false;
	}

	if( !bLinefeed )
		++iLines;

	if( piLines )
		*piLines = iLines;

	return strResult;
}


int Font::GetStringLength( const string &rstrString )
{
	int iLength = 0;

	for( unsigned int i = 0; i < rstrString.length(); ++i )
		iLength += m_akCharacters[ (unsigned char)rstrString[i] ].m_iWidth + m_iSpacing;

	return iLength;
}


int Font::GetStringHeight( const std::string &rstrString, int iX )
{
	int iLines;

	ClipLine( rstrString, iX, -1000, &iLines );

	return( iLines * GetLineHeight() );
}


void Font::SetClipBox( int iX, int iY, int iWidth, int iHeight )
{
	m_iClipX        = iX;
	m_iClipY        = iY;
	m_iClipWidth    = iWidth - 1;
	m_iClipHeight   = iHeight - 1;
}


void Font::GetClipBox( int *piX, int *piY, int *piWidth, int *piHeight )
{
	assert( piX && piY && piWidth && piHeight );

	*piX      = m_iClipX;
	*piY      = m_iClipY;
	*piWidth  = m_iClipWidth + 1;
	*piHeight = m_iClipHeight + 1;
}


void Font::ResetClipBox()
{
	SetClipBox( 0, 0, Core::Get()->GetRenderDevice()->GetWidth() - 1, Core::Get()->GetRenderDevice()->GetHeight() - 1 );
}


const Color &Font::GetColor() const
{
	return m_kColor;
}


Color Font::SetColor( const Color &rkColor )
{
	Color kRet( m_kColor );

	m_kColor = rkColor;

	return kRet;
}


void Font::SetWordSeparators( const std::string &rstrSeparators )
{
	m_strWordSeparators = rstrSeparators;
}


Font::FONTALIGN Font::SetHorizontalAlign( FONTALIGN eAlign )
{
	FONTALIGN eOld = m_eHorizontalAlign;

	m_eHorizontalAlign = eAlign;

	return eOld;
}


Font::FONTALIGN Font::GetHorizontalAlign()
{
	return m_eHorizontalAlign;
}









FontManager::FontManager()
{
}


FontManager::~FontManager()
{
	if( m_vpkFonts.size() )
	{
		neolog << LogLevel( WARNING ) << "[FontManager::~FontManager] font vector not empty [" << m_vpkFonts.size() << "]" << endl;

		for( unsigned int uiFont = 0; uiFont < m_vpkFonts.size(); ++uiFont )
		{
			neolog << LogLevel( DEBUG ) << "  [" << m_vpkFonts[ uiFont ] << "] ref count [" << m_vpkFonts[ uiFont ]->GetRefCount() << "]" << endl;

			m_vpkFonts[ uiFont ]->m_pkMaterial->m_pkTexture  = 0;
			m_vpkFonts[ uiFont ]->m_pkFontManager            = 0;
		}
	}
}


void FontManager::DeregisterFont( Font *pkFont )
{
	for( unsigned int uiFont = 0; uiFont < m_vpkFonts.size(); ++uiFont )
	{
		if( m_vpkFonts[ uiFont ] == pkFont )
		{
			m_vpkFonts.erase( m_vpkFonts.begin() + uiFont );
			break;
		}
	}
}


FontPtr FontManager::CreateFont( const std::string &rstrName, bool bLoad )
{
	FontPtr pkFont = 0;

	if( rstrName.size() )
		pkFont = FindFont( rstrName );

	if( !pkFont )
	{
		pkFont = new Font( this );

		m_vpkFonts.push_back( pkFont );

		if( rstrName.size() && bLoad )
			pkFont->Load( rstrName + ".nfon" );
	}

	return pkFont;
}


FontPtr FontManager::FindFont( const string &rstrName )
{
	vector< Font* >::iterator ppkFont    = m_vpkFonts.begin();
	vector< Font* >::iterator ppkFontEnd = m_vpkFonts.end();

	for( ; ppkFont != ppkFontEnd; ++ppkFont )
	{
		if( (*ppkFont)->m_strName == rstrName )
			return( *ppkFont );
	}

	return 0;
}


}; // namespace NeoEngine
