/***************************************************************************
                          input.cpp  -  Input subsystem
                             -------------------
    begin                : Thu Nov 8 2001
    copyright            : (C) 2001 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, input.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2001
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "input.h"
#include "inputentity.h"
#include "core.h"
#include "render.h"
#include "renderwindow.h"

#ifdef _WIN32
#   include <windows.h>
#endif

using namespace std;


namespace NeoEngine
{


InputEntity::InputEntity( InputGroup *pkGroup ) :
	Activator()
{
	m_pkGroup = 0;

	AttachToGroup( pkGroup );	
}


InputEntity::~InputEntity()
{
	AttachToGroup( 0 );
}


void InputEntity::AttachToGroup( InputGroup *pkGroup )
{
	if( m_pkGroup == pkGroup )
		return;
	
	if( m_pkGroup )
	{
		InputGroup *pkOldGroup = m_pkGroup;

		m_pkGroup = 0;
		
		pkOldGroup->DetachEntity( this );
	}

	if( ( m_pkGroup = pkGroup ) )
		m_pkGroup->AttachEntity( this );
}



InputGroup::InputGroup( InputManager *pkManager ) :
	Activator(),
	m_pkManager( pkManager ? pkManager : Core::Get()->GetInputManager() )
{
	if( m_pkManager )
		m_pkManager->AttachGroup( this );
}


InputGroup::~InputGroup()
{
	if( m_pkManager )
		m_pkManager->DetachGroup( this );

	while( m_vpkInputObjects.size() )
		m_vpkInputObjects[0]->AttachToGroup( 0 );
}


void InputGroup::AttachEntity( InputEntity *pkObject )
{
	if( !pkObject )
		return;

	vector< InputEntity* >::iterator ppkEntity    = m_vpkInputObjects.begin();
	vector< InputEntity* >::iterator ppkEntityEnd = m_vpkInputObjects.end();

	for( ; ppkEntity != ppkEntityEnd; ++ppkEntity )
		if(  pkObject == (*ppkEntity) )
			return;

	m_vpkInputObjects.push_back( pkObject );

	pkObject->AttachToGroup( this );
}


void InputGroup::DetachEntity( InputEntity *pkObject )
{
	if( !pkObject )
		return;

	vector< InputEntity* >::iterator ppkEntity    = m_vpkInputObjects.begin();
	vector< InputEntity* >::iterator ppkEntityEnd = m_vpkInputObjects.end();

	for( ; ppkEntity != ppkEntityEnd; ++ppkEntity )
		if( (*ppkEntity) == pkObject )
		{
			m_vpkInputObjects.erase( ppkEntity );
			break;
		}

	pkObject->AttachToGroup( 0 );
}


void InputGroup::Distribute( const vector< InputEvent* > &rvpkEvents )
{
	vector< InputEvent* >::const_iterator ppkEvent = rvpkEvents.begin();
	vector< InputEvent* >::const_iterator ppkEnd   = rvpkEvents.end();

	for( ; ppkEvent != ppkEnd; ++ppkEvent )
		Distribute( *ppkEvent );
}


void InputGroup::Distribute( const InputEvent *pkEvent )
{
	vector< InputEntity* >::iterator ppkNode = m_vpkInputObjects.begin();
	vector< InputEntity* >::iterator ppkEnd  = m_vpkInputObjects.end();

	for( ; ppkNode != ppkEnd; ++ppkNode )
		if( (*ppkNode)->IsActive() )
			(*ppkNode)->Input( pkEvent );
}





InputManager::InputManager()
{
}


InputManager::~InputManager()
{
	vector< InputDevice* >::iterator ppkDevice    = m_vpkInputDevices.begin();
	vector< InputDevice* >::iterator ppkDeviceEnd = m_vpkInputDevices.end();

	for( ; ppkDevice != ppkDeviceEnd; ++ppkDevice )
		(*ppkDevice)->m_pkManager = 0;

	vector< InputGroup* >::iterator ppkGroup    = m_vpkInputGroups.begin();
	vector< InputGroup* >::iterator ppkGroupEnd = m_vpkInputGroups.end();

	for( ; ppkGroup != ppkGroupEnd; ++ppkGroup )
		(*ppkGroup)->m_pkManager = 0;
}


void InputManager::AttachGroup( InputGroup *pkGroup )
{
	if( !pkGroup )
		return;

	vector< InputGroup* >::iterator ppkGroup    = m_vpkInputGroups.begin();
	vector< InputGroup* >::iterator ppkGroupEnd = m_vpkInputGroups.end();

	for( ; ppkGroup != ppkGroupEnd; ++ppkGroup )
		if( (*ppkGroup) == pkGroup )
			return;

	m_vpkInputGroups.push_back( pkGroup );
}


void InputManager::AttachDevice( InputDevice *pkDevice )
{
	if( !pkDevice )
		return;

	vector< InputDevice* >::iterator ppkDevice    = m_vpkInputDevices.begin();
	vector< InputDevice* >::iterator ppkDeviceEnd = m_vpkInputDevices.end();

	for( ; ppkDevice != ppkDeviceEnd; ++ppkDevice )
		if( (*ppkDevice) == pkDevice )
			return;

	m_vpkInputDevices.push_back( pkDevice );
}


void InputManager::DetachGroup( InputGroup *pkObject )
{
	if( !pkObject )
		return;

	vector< InputGroup* >::iterator ppkGroup    = m_vpkInputGroups.begin();
	vector< InputGroup* >::iterator ppkGroupEnd = m_vpkInputGroups.end();

	for( ; ppkGroup != ppkGroupEnd; ++ppkGroup )
		if( (*ppkGroup) == pkObject )
		{
			m_vpkInputGroups.erase( ppkGroup );
			return;
		}
}


void InputManager::DetachDevice( InputDevice *pkDevice )
{
	if( !pkDevice )
		return;

	vector< InputDevice* >::iterator ppkDevice    = m_vpkInputDevices.begin();
	vector< InputDevice* >::iterator ppkDeviceEnd = m_vpkInputDevices.end();

	for( ; ppkDevice != ppkDeviceEnd; ++ppkDevice )
		if( (*ppkDevice) == pkDevice )
		{
			m_vpkInputDevices.erase( ppkDevice );
			return;
		}
}


void InputManager::Process()
{
	if( !m_vpkInputDevices.size() )
		return;

	vector< InputEvent* > vpkEvents;

	int iNumEvents = 0;
	bool bAllocate = true;

	vector< InputDevice* >::iterator ppkDevice    = m_vpkInputDevices.begin();
	vector< InputDevice* >::iterator ppkDeviceEnd = m_vpkInputDevices.end();

	for( ; ppkDevice != ppkDeviceEnd; ++ppkDevice )
	{
		if( (*ppkDevice)->IsActive() )
		{
			do
			{
				if( bAllocate )
					vpkEvents.push_back( new InputEvent() );
				else
					bAllocate = true;

			} while( (*ppkDevice)->Collect( vpkEvents[ iNumEvents++ ] ) );

			iNumEvents--;

			bAllocate = false;
		}
	}

	delete vpkEvents[ iNumEvents ];
	vpkEvents.pop_back();

	vector< InputGroup* >::iterator ppkGroup    = m_vpkInputGroups.begin();
	vector< InputGroup* >::iterator ppkGroupEnd = m_vpkInputGroups.end();

	for( ; ppkGroup != ppkGroupEnd; ++ppkGroup )
		if( (*ppkGroup)->IsActive() )
			(*ppkGroup)->Distribute( vpkEvents );

	vector< InputEvent* >::iterator ppkEvent    = vpkEvents.begin();
	vector< InputEvent* >::iterator ppkEventEnd = vpkEvents.end();

	for( ; ppkEvent != ppkEventEnd; ++ppkEvent )
		delete( *ppkEvent );
}





InputDevice::InputDevice( InputManager *pkManager ) :
	Activator(),
	m_pkManager( pkManager ? pkManager : Core::Get()->GetInputManager() ),
	m_eInputEventGroups( SYSTEMINPUT | KEYBOARDINPUT | MOUSEINPUT | JOYSTICKINPUT )
{
	if( m_pkManager )
		m_pkManager->AttachDevice( this );
}


InputDevice::~InputDevice()
{
	if( m_pkManager )
		m_pkManager->DetachDevice( this );
}


InputDevice* Core::CreateInputDevice( const std::string &rstrName, RenderDevice *pkRenderDevice, InputManager *pkInputManager )
{
	// Check if render window is open
	if( !pkRenderDevice )
		pkRenderDevice = Core::Get()->GetRenderDevice();

	RenderWindow kRenderWindow = pkRenderDevice->GetWindow();

#ifdef WIN32
	if( !kRenderWindow.m_hWnd )
#elif __APPLE__
	if( !kRenderWindow.m_pWindow )
#else
	if( !kRenderWindow.m_Window )
#endif 
	{
		neolog << LogLevel( ERROR ) << "*** unable to create input device: no open render device" << endl;
		return 0;
	}

#if defined(WIN32) && defined(_DEBUG)
	ModulePtr pkModule = s_pkModuleManager->LoadModule( "neodev" + rstrName + "_debug" );
#else
	ModulePtr pkModule = s_pkModuleManager->LoadModule( "neodev" + rstrName );
#endif

	if( !pkModule || !pkModule->IsValid() )
		return 0;

	ModSymbol kSymInfo;
	ModSymbol kSymLoad;

	kSymInfo = pkModule->LookupSymbol( "GetVersion" );

	if( !kSymInfo )
		neolog << LogLevel( WARNING ) << "*** input device module [" << rstrName << "] does not export GetVersion symbol" << endl;
	else
	{
		string strName;
		int iMajor;
		int iMinor;
		int iRevision;

		(( void ( NE_CDECL* )( std::string*, int*, int*, int* ))kSymInfo)( &strName, &iMajor, &iMinor, &iRevision );

		neolog << LogLevel( INFO ) << "Creating device " << rstrName << " version " << iMajor << "." << iMinor << "." << iRevision << endl;
	}

	kSymLoad = pkModule->LookupSymbol( "CreateDevice" );

	if( !kSymLoad )
	{
		neolog << LogLevel( ERROR ) << "*** input device module [" << rstrName << "] does not export CreateDevice method" << endl;
		return 0;
	}

	InputDevice *pkInputDevice = (( InputDevice* ( NE_CDECL* )( RenderWindow*, InputManager* ))kSymLoad)( &kRenderWindow, pkInputManager );

	if( pkInputDevice )
		pkInputDevice->m_pkModule = pkModule;

	return pkInputDevice;
}


void Core::DeleteInputDevice( InputDevice *pkDevice )
{
	if( !pkDevice )
	{
		neolog << LogLevel( ERROR ) <<  "*** trying to delete invalid input device" << endl;
		return;
	}

    //By creating new reference to module, it won't get deleted inside render device dtor,
	//since that is a Bad Thing<tm> - the dtor is inside module space
	ModulePtr pkModule = pkDevice->m_pkModule;

	delete pkDevice;
}


}; // namespace NeoEngine

