/***************************************************************************
                        light.h  -  Dynamic lighting
                             -------------------
    begin                : Mon Nov 5 2001
    copyright            : (C) 2001 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, light.h

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2001
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#ifndef __NELIGHT_H
#define __NELIGHT_H


#include "base.h"
#include "sceneentity.h"
#include "color.h"
#include "nemath.h"


/**
  * \file neoengine/light.h
  * Dynamic light
  */


namespace NeoEngine
{


/**
  * \brief Dynamic light object
  * Dynamic lights are scene entities. The Render method
  * will register the light as active with the render device
  * for this frame, allowing bounding volume culling of light
  * entities.
  * \author Mattias Jansson (mattias@realityrift.com)
  * \todo Support all light types
  * \todo Change direction to be world rotation z-axis to enable light to be positioned in hierarchies
  */
class NEOENGINE_API Light : public SceneEntity
{
	public:

		DefineVisitable();


	public:

		/**
		* \brief Light type identifiers
		*/
		enum LIGHTTYPE
		{
		  /*! Directional light. The direction of the light is the negative z axis of the parent node in world space */
		  DIRECTIONAL                   = 0,

		  /*! Point light. Cast light in all directions from a point that is the translation of the parent node in world space */
		  POINT                         = 1,

		  /*! Ambient light. Only contributes to the total ambient light in the scene, does not cast any shadows (naturally) */
		  AMBIENT                       = 2
		};


		/**
		* \brief Light flags
		*/
		enum LIGHTFLAG
		{
		  /*! No light flags */
		  NOFLAGS                        = 0,

		  /*! Light should cause dynamic shadows to be generated by geometry with materials that have shadow flag set */
		  CASTSHADOWS                    = 1
		};


	protected:

		/*! Type */
		LIGHTTYPE                       m_eType;

		/*! Flags */
		int                             m_iFlags;


	public:

		/*! Ambient color */
		Color                           m_kAmbient;

		/*! Diffuse color */
		Color                           m_kDiffuse;

		/*! Specular color */
		Color                           m_kSpecular;

		/*! Constant attenuation, default 1.0f */
		float                           m_fConstantAttenuation;

		/*! Linear attenuation, default 0.0f */
		float                           m_fLinearAttenuation;

		/*! Quadratic attenuation, default 0.0f */
		float                           m_fQuadraticAttenuation;


		/**
		* \param eType                  Light type
		* \param iFlags                 Flags
		* \param rkDirection            Light direction (for DIRECTIONAL and SPOT)
		*/
		                                Light( LIGHTTYPE eType, int iFlags = NOFLAGS );

		/**
		* \return                       Light type
		*/
		inline LIGHTTYPE                GetType() const { return m_eType; }

		/**
		* \return                       Light flags
		*/
		inline int                      GetFlags() const { return m_iFlags; }

		/**
		* Add light to this frame rendering queue
		* \param pkFrustum              Ignored
		* \param bForce                 Ignored
		* \return                       true
		*/
		virtual bool                    Render( Frustum *pkFrustum = 0, bool bForce = false );
};


};


#endif
