/***************************************************************************
         logstream.h  -  Filtered logging to multiple output targets
                             -------------------
    begin                : Tue Aug 26 2003
    copyright            : (C) 2003 by Cody Russell
    email                : cody `at' jhu.edu
***************************************************************************
                                                                                
 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, logstream.cpp

 The Initial Developer of the Original Code is Cody Russell.
 Portions created by Cody Russell are Copyright (C) 2003
 Cody Russell. All Rights Reserved.

 Contributors: Mattias Jansson (mattias@realityrift.com)
               Bert van der Weerd
                                                                                
***************************************************************************/

#include "logstream.h"
#include "file.h"

#include <iostream>

#ifdef WIN32
#  ifndef WIN32_LEAN_AND_MEAN
#    define WIN32_LEAN_AND_MEAN
#  endif
#  include <windows.h>
#endif

using namespace std;


namespace NeoEngine
{


/* Global engine log object */
LogSource neolog;



LogSink::LogSink() :
	m_uiThreshold( DEBUG )
{
}


LogSink::~LogSink()
{
}


void LogSink::SetLogThreshold( unsigned int uiLevel )
{
	m_uiThreshold = uiLevel;
}




LogStdoutSink::LogStdoutSink() :
	LogSink()
{
}


LogStdoutSink::~LogStdoutSink()
{
}


void LogStdoutSink::Write( const string &rstrMsg )
{
	cout << rstrMsg.c_str() << flush;
}




LogFileSink::LogFileSink( const std::string &rstrFileName ) :
	LogSink(),
	m_pkFile( new File( "", rstrFileName, ios_base::out, true, 0 ) )
{
}


LogFileSink::LogFileSink( File *pkFile ) :
	LogSink(),
	m_pkFile( pkFile )
{
}


LogFileSink::~LogFileSink()
{
	delete m_pkFile;
}


void LogFileSink::Write( const string &rstrMsg )
{
	if( m_pkFile )
		*m_pkFile << rstrMsg << flush;
}




LogMSVCDebugSink::LogMSVCDebugSink() :
	LogSink()
{
}


LogMSVCDebugSink::~LogMSVCDebugSink()
{
}


void LogMSVCDebugSink::Write( const std::string &rstrMsg )
{
#ifdef WIN32
	OutputDebugString( rstrMsg.c_str() );
#endif
}




LogSourceStreamBuf::LogSourceStreamBuf() :
	basic_streambuf< char >()
{
}


LogSourceStreamBuf::~LogSourceStreamBuf()
{
}


void LogSourceStreamBuf::SetLogSource( LogSource *pkSource )
{
	m_pkLogSource = pkSource;
}


basic_ios< char >::int_type LogSourceStreamBuf::overflow( int_type c )
{
	if( !char_traits< char >::eq_int_type( c, char_traits< char >::eof() ) )
	{
		m_strMsg += c;

		if( c == '\n' )
		{
			m_pkLogSource->Write( m_strMsg );
			m_strMsg = "";
		}
	}

	return char_traits< char >::not_eof( c );
}




LogSource::LogSource() :
	basic_ostream< char >( &m_kStreamBuf ),
	m_uiLogLevel( DEBUG ),
	m_uiThreshold( DEBUG ),
	m_pkErrorCallback( 0 )
{
	m_kStreamBuf.SetLogSource( this );
}


void LogSource::SetLogLevel( unsigned int uiLevel )
{
	m_uiLogLevel = uiLevel;
}


void LogSource::PushLogLevel( unsigned int uiLevel )
{
	m_vuiLogLevelStack.push_back( m_uiLogLevel );

	m_uiLogLevel = uiLevel;
}


void LogSource::PopLogLevel()
{
	if( m_vuiLogLevelStack.size() )
	{
		m_uiLogLevel = m_vuiLogLevelStack.back();
		m_vuiLogLevelStack.pop_back();
	}
}


void LogSource::SetLogThreshold( unsigned int uiLevel )
{
	m_uiThreshold = uiLevel;
}


void LogSource::SetErrorCallback( LogErrorCallback *pkCallback )
{
	m_pkErrorCallback = pkCallback;
}


void LogSource::AttachSink( LogSink *pkSink )
{
	if( !pkSink )
		return;

	std::vector< LogSink* >::iterator ppkSink = m_vpkSinks.begin();
	std::vector< LogSink* >::iterator ppkEnd  = m_vpkSinks.end();

	for( ; ppkSink != ppkEnd; ++ppkSink )
		if( *ppkSink == pkSink )
			return;

	m_vpkSinks.push_back( pkSink );
}


void LogSource::DetachSink( LogSink *pkSink )
{
	std::vector< LogSink* >::iterator ppkSink = m_vpkSinks.begin();
	std::vector< LogSink* >::iterator ppkEnd  = m_vpkSinks.end();

	for( ; ppkSink != ppkEnd; ++ppkSink )
		if( *ppkSink == pkSink )
		{
			m_vpkSinks.erase( ppkSink );
			return;
		}
}


void LogSource::Write( const string &rstrMsg )
{
	if( m_pkErrorCallback && ( m_uiLogLevel <= ERROR ) )
		m_pkErrorCallback->ErrorCallback( m_uiLogLevel, rstrMsg );

	if( m_uiThreshold < m_uiLogLevel )
		return;

	std::vector< LogSink* >::const_iterator ppkIter = m_vpkSinks.begin();
	std::vector< LogSink* >::const_iterator ppkEnd  = m_vpkSinks.end();

	for( ; ppkIter != ppkEnd; ++ppkIter )
	{
		LogSink *pkSink = (*ppkIter);

		if( pkSink->m_uiThreshold >= m_uiLogLevel )
			pkSink->Write( rstrMsg );
	}
}


ostream &operator << ( ostream &rkOut, const LogManipulator &rkManip )
{
	return rkManip( rkOut );
}


ostream &LogLevel::operator()( ostream &rkOut ) const
{
	LogSource *pkLog = dynamic_cast< LogSource* >( &rkOut );
	
	if( pkLog )
		pkLog->SetLogLevel( m_uiLevel );

	return rkOut;
}


ostream &LogLevelPush::operator()( ostream &rkOut ) const
{
	LogSource *pkLog = dynamic_cast< LogSource* >( &rkOut );
	
	if( pkLog )
		pkLog->PushLogLevel( m_uiLevel );

	return rkOut;
}


ostream &LogLevelPop::operator()( ostream &rkOut ) const
{
	LogSource *pkLog = dynamic_cast< LogSource* >( &rkOut );
	
	if( pkLog )
		pkLog->PopLogLevel();

	return rkOut;
}


};
