/***************************************************************************
                       mesh.cpp  -  Mesh highlevel class
                             -------------------
    begin                : Thu Nov 1 2001
    copyright            : (C) 2001 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, mesh.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2001
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "mesh.h"
#include "animatedsubmesh.h"
#include "core.h"
#include "polygon.h"
#include "submesh.h"
#include "bone.h"
#include "skeleton.h"
#include "skin.h"
#include "scenenode.h"
#include "renderprimitive.h"
#include "aabb.h"
#include "line.h"
#include "collision.h"
#include "contact.h"

using namespace std;


namespace NeoEngine
{


Mesh::Mesh( const HashString &rstrName, MeshPool *pkPool ) :
	m_pkPool( pkPool ? pkPool : Core::Get()->GetMeshManager() ),
	m_strName( rstrName ),
	m_pkSkeleton( 0 )
{
	m_pkPool->Insert( m_strName, this );
}


Mesh::~Mesh()
{
	m_pkPool->Delete( m_strName, this );

	vector< SubMesh* >::iterator ppkSubMesh    = m_vpkSubMeshes.begin();
	vector< SubMesh* >::iterator ppkSubMeshEnd = m_vpkSubMeshes.end();

	for( ; ppkSubMesh != ppkSubMeshEnd; ++ppkSubMesh )
		delete( *ppkSubMesh );

	delete m_pkSkeleton;
}


void Mesh::SetName( const HashString &rstrName )
{
	m_pkPool->Delete( m_strName, this );

	m_strName = rstrName;

	m_pkPool->Insert( m_strName, this );
}






MeshEntity::MeshEntity( const MeshPtr &pkMesh ) :
	SceneEntity(),
	m_pkMesh( pkMesh ),
	m_pkSkeleton( pkMesh->m_pkSkeleton ? static_cast< Skeleton* >( pkMesh->m_pkSkeleton->Duplicate() ) : 0 ),
	m_uiFlags( 0 )
{
	vector< SubMesh* >::iterator ppkSubMesh = m_pkMesh->m_vpkSubMeshes.begin();
	vector< SubMesh* >::iterator ppkEnd     = m_pkMesh->m_vpkSubMeshes.end();

	m_vpkSubMeshes.resize( m_pkMesh->m_vpkSubMeshes.size() );

	vector< SubMesh* >::iterator ppkCurrent = m_vpkSubMeshes.begin();

	for( ; ppkSubMesh != ppkEnd; ++ppkSubMesh, ++ppkCurrent )
	{
		*ppkCurrent = (*ppkSubMesh)->Duplicate();

		if( (*ppkCurrent)->GetType() == SubMesh::SKELETALSUBMESH )
		{
			if( !m_pkSkeleton )
				neolog << LogLevel( WARNING ) << "*** skin submeshes but no skeleton" << endl;

			dynamic_cast< SkeletalSubMesh* >( *ppkCurrent )->m_pkSkeleton = m_pkSkeleton;
		}
	}
}


MeshEntity::MeshEntity( MeshEntity &rkMesh ) :
	SceneEntity(),
	m_pkMesh( rkMesh.GetMesh() ),
	m_pkSkeleton( rkMesh.m_pkSkeleton ? static_cast< Skeleton* >( rkMesh.m_pkSkeleton->Duplicate() ) : 0 ),
	m_uiFlags( rkMesh.m_uiFlags )
{
	vector< SubMesh* >::iterator ppkSubMesh = rkMesh.m_vpkSubMeshes.begin();
	vector< SubMesh* >::iterator ppkEnd     = rkMesh.m_vpkSubMeshes.end();

	m_vpkSubMeshes.resize( rkMesh.m_vpkSubMeshes.size() );

	vector< SubMesh* >::iterator ppkCurrent = m_vpkSubMeshes.begin();

	for( ; ppkSubMesh != ppkEnd; ++ppkSubMesh, ++ppkCurrent )
	{
		*ppkCurrent = (*ppkSubMesh)->Duplicate();

		if( (*ppkCurrent)->GetType() == SubMesh::SKELETALSUBMESH )
		{
			if( !m_pkSkeleton )
				neolog << LogLevel( WARNING ) << "*** skin submeshes but no skeleton" << endl;

			dynamic_cast< SkeletalSubMesh* >( *ppkCurrent )->m_pkSkeleton = m_pkSkeleton;
		}
	}
}


MeshEntity::~MeshEntity()
{
	vector< SubMesh* >::iterator ppkSubMesh = m_vpkSubMeshes.begin();
	vector< SubMesh* >::iterator ppkEnd     = m_vpkSubMeshes.end();

	for( ; ppkSubMesh != ppkEnd; ++ppkSubMesh )
		delete( *ppkSubMesh );

	delete m_pkSkeleton;
}


void MeshEntity::SetFlags( unsigned int uiFlags )
{
	m_uiFlags |= uiFlags;
}


void MeshEntity::ResetFlags( unsigned int uiFlags )
{
	m_uiFlags &= (~uiFlags);
}


void MeshEntity::SetNode()
{
	if( m_pkSkeleton )
		m_pkNode->AttachNode( m_pkSkeleton, false );
}


bool MeshEntity::Render( Frustum *pkFrustum, bool bForce )
{
	if( !SceneEntity::Render( pkFrustum, bForce ) )
		return false;

#ifdef NEOENGINE_THREAD_SAFE
#  ifdef WIN32
	//Ugly as hell, but you can't declare object with constructor as thread-local
	static __declspec( thread ) unsigned char saucPrimBuffer[ sizeof( RenderPrimitive ) ];
	static __declspec( thread ) RenderPrimitive *spkPrimitive = 0;
	spkPrimitive = (RenderPrimitive*)saucPrimBuffer;
#  else
#    error Thread local storage not implemented on platform
#  endif
#else
	static RenderPrimitive  skPrimitive;
	static RenderPrimitive *spkPrimitive = &skPrimitive;
#endif

	if( m_pkNode )
	{
		spkPrimitive->m_kModelMatrix    = m_pkNode->GetWorldTransform();
		spkPrimitive->m_kInvModelMatrix = m_pkNode->GetInverseWorldTransform();
	}

	vector< SubMesh* >::iterator ppkSubMesh = m_vpkSubMeshes.begin();
	vector< SubMesh* >::iterator ppkEnd     = m_vpkSubMeshes.end();

	for( ; ppkSubMesh != ppkEnd; ++ppkSubMesh )
		(*ppkSubMesh)->Render( spkPrimitive, pkFrustum );

	//Reset pointers to allow reference counter to clean up
	spkPrimitive->m_pkVertexBuffer       = 0;
	spkPrimitive->m_pkPolygonBuffer      = 0;
	spkPrimitive->m_pkPolygonStripBuffer = 0;
	spkPrimitive->m_pkMaterial           = 0;

	return true;
}


void MeshEntity::Update( float fDeltaTime )
{
	SceneEntity::Update( fDeltaTime );

	if( m_pkSkeleton )
		m_pkSkeleton->Update( fDeltaTime );

	bool bChanged = false;

	vector< SubMesh* >::iterator ppkSubMesh = m_vpkSubMeshes.begin();
	vector< SubMesh* >::iterator ppkEnd     = m_vpkSubMeshes.end();

	for( ; ppkSubMesh != ppkEnd; ++ppkSubMesh )
	{
		(*ppkSubMesh)->Update( fDeltaTime );

		if( (*ppkSubMesh)->m_bChanged )
		{
			(*ppkSubMesh)->m_bChanged = false;
			bChanged = true;
		}
	}

	if( bChanged && ( m_uiFlags & AUTOGENVOLUME ) )
		GenerateBoundingVolume();
}


SceneEntity *MeshEntity::Duplicate()
{
	return new MeshEntity( *this );
}


bool MeshEntity::Intersection( BoundingVolume *pkObj, ContactSet *pkContactSet, bool bInvertNormal )
{
	if( m_uiFlags & VOLUMEINTERSECTION )
		return( GetBoundingVolume() ? GetBoundingVolume()->Intersection( pkObj, pkContactSet, bInvertNormal ) : false );

	ContactSet  kSet;
	ContactSet *pkSet      = 0;
	bool        bIntersect = false;

	if( pkContactSet )
		pkSet = &kSet;

	// Transform bounding volume into local space
	Vector3d   kCurTranslation = pkObj->GetTranslation();
	Quaternion kCurRotation    = pkObj->GetRotation();

	if( m_pkNode )
	{
		pkObj->TranslateWorld( ~m_pkNode->GetWorldRotation() * -m_pkNode->GetWorldTranslation() );
		pkObj->RotateWorld( ~m_pkNode->GetWorldRotation() );
	}

	vector< SubMesh* >::iterator ppkSubMesh = m_vpkSubMeshes.begin();
	vector< SubMesh* >::iterator ppkEnd     = m_vpkSubMeshes.end();

	for( ; ppkSubMesh != ppkEnd; ++ppkSubMesh )
	{
		VertexBuffer  *pkVBuffer = (*ppkSubMesh)->GetVertexBuffer();
		PolygonBuffer *pkPBuffer = (*ppkSubMesh)->GetPolygonBuffer();

		pkVBuffer->Lock( Buffer::READ );
		pkPBuffer->Lock( Buffer::READ );

		//Intersection test with all polygons
		NeoEngine::Polygon *pkPolygon    = pkPBuffer->GetPolygon();
		int                 iNumPolygons = pkPBuffer->GetNumElements();

		for( int i = 0; i < iNumPolygons; ++i, ++pkPolygon )
		{
			if( pkObj->Intersection( *(Vector3d*)pkVBuffer->GetVertex( pkPolygon->v[0] ), *(Vector3d*)pkVBuffer->GetVertex( pkPolygon->v[1] ), *(Vector3d*)pkVBuffer->GetVertex( pkPolygon->v[2] ), pkSet ) )
			{
				bIntersect = true;

				if( !pkContactSet )
				{
					ppkSubMesh = ppkEnd - 1;
					break;
				}
			}
		}

		if( pkSet && pkSet->m_vpkContacts.size() )
		{
			vector< Contact* >::iterator ppkContact    = pkSet->m_vpkContacts.end() - 1;
			vector< Contact* >::iterator ppkContactEnd = pkSet->m_vpkContacts.begin();

			for( ; ppkContact >= ppkContactEnd; --ppkContact )
			{
				(*ppkContact)->m_pkSceneNode     = m_pkNode;
				(*ppkContact)->m_apkMaterials[0] = (*ppkSubMesh)->m_pkMaterial;

				pkContactSet->Add( *ppkContact );
			}

			pkSet->m_vpkContacts.clear();
		}

		pkVBuffer->Unlock();
		pkPBuffer->Unlock();
	}

	//Restore bounding volume
	if( m_pkNode )
	{
		pkObj->SetTranslation( kCurTranslation );
		pkObj->SetRotation( kCurRotation );
	}

	return bIntersect;
}


bool MeshEntity::Intersection( const Ray &rkRay, ContactSet *pkContactSet )
{
	if( m_uiFlags & VOLUMEINTERSECTION )
		return( GetBoundingVolume() ? GetBoundingVolume()->Intersection( rkRay, pkContactSet ) : false );

	ContactSet  kSet;
	ContactSet *pkSet      = 0;
	bool        bIntersect = false;

	if( pkContactSet )
		pkSet = &kSet;

	// Transform ray into local space
	Ray kRay( rkRay.m_kOrigin, rkRay.m_kDirection );

	if( m_pkNode )
	{
		Matrix kInvRot( ~m_pkNode->GetWorldRotation() );
		kRay.m_kOrigin    += kInvRot * -m_pkNode->GetWorldTranslation();
		kRay.m_kDirection  = kInvRot * rkRay.m_kDirection;
	}

	vector< SubMesh* >::iterator ppkSubMesh = m_vpkSubMeshes.begin();
	vector< SubMesh* >::iterator ppkEnd     = m_vpkSubMeshes.end();

	for( ; ppkSubMesh != ppkEnd; ++ppkSubMesh )
	{
		VertexBuffer  *pkVBuffer = (*ppkSubMesh)->GetVertexBuffer();
		PolygonBuffer *pkPBuffer = (*ppkSubMesh)->GetPolygonBuffer();

		pkVBuffer->Lock( Buffer::READ );
		pkPBuffer->Lock( Buffer::READ );

		//Intersection test with all polygons
		NeoEngine::Polygon *pkPolygon    = pkPBuffer->GetPolygon();
		int                 iNumPolygons = pkPBuffer->GetNumElements();

		for( int i = 0; i < iNumPolygons; ++i, ++pkPolygon )
		{
			if( NeoEngine::Intersection( kRay, *(Vector3d*)pkVBuffer->GetVertex( pkPolygon->v[0] ), *(Vector3d*)pkVBuffer->GetVertex( pkPolygon->v[1] ), *(Vector3d*)pkVBuffer->GetVertex( pkPolygon->v[2] ), pkSet ) )
			{
				bIntersect = true;

				if( !pkContactSet )
				{
					ppkSubMesh = ppkEnd - 1;
					break;
				}
			}
		}

		if( pkSet && pkSet->m_vpkContacts.size() )
		{
			vector< Contact* >::iterator ppkContact    = pkSet->m_vpkContacts.end() - 1;
			vector< Contact* >::iterator ppkContactEnd = pkSet->m_vpkContacts.begin();

			for( ; ppkContact >= ppkContactEnd; --ppkContact )
			{
				(*ppkContact)->m_pkSceneNode     = m_pkNode;
				(*ppkContact)->m_apkMaterials[0] = (*ppkSubMesh)->m_pkMaterial;

				pkContactSet->Add( *ppkContact );
			}

			pkSet->m_vpkContacts.clear();
		}

		pkVBuffer->Unlock();
		pkPBuffer->Unlock();
	}

	return bIntersect;
}


void MeshEntity::GenerateBoundingVolume()
{
	if( !GetBoundingVolume() )
		SetBoundingVolume( new AABB );

	vector< SubMesh* >::iterator ppkSubMesh = m_vpkSubMeshes.begin();
	vector< SubMesh* >::iterator ppkEnd     = m_vpkSubMeshes.end();

	if( ppkSubMesh == ppkEnd )
		return;

	BoundingVolume *pkVolume = GetBoundingVolume();
	BoundingVolume *pkTemp   = pkVolume->Duplicate();

	pkVolume->Generate( (*ppkSubMesh)->GetVertexBuffer() );

	++ppkSubMesh;

	for( ; ppkSubMesh != ppkEnd; ++ppkSubMesh )
	{
		pkTemp->Generate( (*ppkSubMesh)->GetVertexBuffer() );
		pkVolume->Merge( pkTemp );
	}

	delete pkTemp;
}


void MeshEntity::CalculateTangents()
{
	vector< SubMesh* >::iterator ppkSubMesh = m_vpkSubMeshes.begin();
	vector< SubMesh* >::iterator ppkEnd     = m_vpkSubMeshes.end();

	for( ; ppkSubMesh != ppkEnd; ++ppkSubMesh )
		(*ppkSubMesh)->CalculateTangents();
}


};
