/***************************************************************************
           module.cpp  -  Wrapper for loading dynamic library modules
                             -------------------
    begin                : Mon Oct 14 2002
    copyright            : (C) 2002 by Cody Russell
    email                : cody [at] jhu.edu
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, module.cpp

 The Initial Developer of the Original Code is Cody Russell.
 Portions created by Cody Russell are Copyright (C) 2002
 Cody Russell. All Rights Reserved.

 Contributors: Mattias Jansson (mattias@realityrift.com)

 ***************************************************************************/

#include "module.h"
#include "logstream.h"

#ifdef WIN32
#  ifndef WIN32_LEAN_AND_MEAN
#    define WIN32_LEAN_AND_MEAN
#  endif
#  include <windows.h>
#endif

#ifdef __APPLE__
#  include <Carbon/Carbon.h>
#  include <CoreFoundation/CFBundle.h>
#  include <mach-o/dyld.h>
#endif

#ifdef POSIX
#  include <dlfcn.h>
#endif

#if defined(POSIX) || defined(__APPLE__)
#  include <unistd.h>
#endif

#include <stdio.h>

using namespace std;


namespace NeoEngine
{


ModuleManager::ModuleManager()
{
}


ModuleManager::~ModuleManager()
{
	vector< Module* > vpkModules;

	m_hashModules.GetAllNodeData( &vpkModules );

	vector< Module* >::iterator ppkModule    = vpkModules.begin();
	vector< Module* >::iterator ppkModuleEnd = vpkModules.end();

	for( ; ppkModule != ppkModuleEnd; ++ppkModule )
	{
		if( ( (*ppkModule)->GetRefCount() > 0 ) && !(*ppkModule)->IsStatic() )
		{
			//If you hit this assert, module was not unloaded/deregistered correctly
			assert( false );

			neolog << LogLevel( WARNING ) << "[ModuleManager::~ModuleManager]: Module [" << (*ppkModule)->GetName() << "] not unloaded correctly" << endl;

			delete( *ppkModule );
		}
		else
			RemoveModule( *ppkModule );
	}
}


bool ModuleManager::AddSearchPath( const HashString &rstrPath, bool bForce )
{
	unsigned int uiPos;
	HashString strNewPath;

	for( uiPos = 0; uiPos < rstrPath.length(); ++uiPos )
	{
		if( rstrPath.m_strData[ uiPos ] == '\\' )
			strNewPath.m_strData += '/';
		else
			strNewPath.m_strData += rstrPath.m_strData[ uiPos ];
	}

	if( !uiPos || ( strNewPath.m_strData[ uiPos - 1 ] != '/' ) )
		strNewPath.m_strData += '/';

	strNewPath.HashMe();

	vector< HashString >::iterator pstrPath    = m_vstrSearchPaths.begin();
	vector< HashString >::iterator pstrPathEnd = m_vstrSearchPaths.end();

	for( ; pstrPath != pstrPathEnd; ++pstrPath )
	{
		if( *pstrPath == strNewPath )
		{
			// The directory is already in our search path
			return false;
		}
	}

	if( !PathExists( strNewPath ) && ( bForce == false ) )
	{
		neolog << LogLevel( INFO ) << "[ModuleManager::AddSearchPath]: The path [" << strNewPath << "] does not exist" << endl;

		return false;
	}
	else
	{
		m_vstrSearchPaths.push_back( strNewPath );
		neolog << LogLevel( INFO ) << "[ModuleManager::AddSearchPath]: Added [" << strNewPath << "] to search path" << endl;

		return true;
	}
}


void ModuleManager::RemoveSearchPath( const HashString &rstrPath )
{
	unsigned int uiPos;
	HashString strNewPath;

	for( uiPos = 0; uiPos < rstrPath.length(); ++uiPos )
	{
		if( rstrPath.m_strData[ uiPos ] == '\\' )
			strNewPath.m_strData += '/';
		else
			strNewPath.m_strData += rstrPath.m_strData[ uiPos ];
	}

	if( !uiPos || ( strNewPath.m_strData[ uiPos - 1 ] != '/' ) )
		strNewPath.m_strData += '/';

	strNewPath.HashMe();

	vector< HashString >::iterator pstrPath    = m_vstrSearchPaths.begin();
	vector< HashString >::iterator pstrPathEnd = m_vstrSearchPaths.end();

	for( ; pstrPath != pstrPathEnd; ++pstrPath )
	{
		if( *pstrPath == strNewPath )
		{
			// The directory is already in our search path
			m_vstrSearchPaths.erase( pstrPath );
			return;
		}
	}
}


bool ModuleManager::PathExists( const HashString &rstrPath )
{
#ifdef WIN32

	WIN32_FIND_DATA sData;

	char szPattern[1024];

	sprintf( szPattern, "%s/*", rstrPath.c_str() );

	HANDLE hFind = FindFirstFile( szPattern, &sData );

	if( hFind != INVALID_HANDLE_VALUE )
	{
		if( sData.cFileName[0] != '.' )
		{
			if( sData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
			{
				return true;
			}
			else
			{
				return false;
			}
		}
		else
		{
			return true;
		}
	}
	else
	{
		return false;
	}

#else

	if( !access( rstrPath.c_str(), F_OK ) )
	{
		return true;
	}
	else
	{
		return false;
	}

#endif
}


ModulePtr ModuleManager::LoadModule( const HashString &rstrName )
{
	ModulePtr  pkMod;
	FILE      *pFile;

	//Try locating module in hash table
	if( ( pkMod = m_hashModules.Find( rstrName ) ) )
		return pkMod;

#if defined(WIN32)
	string strFilename = rstrName.m_strData + ".dll";
#elif defined(POSIX) || defined(__APPLE__)
	string strFilename = "lib" + rstrName.m_strData + ".so";
#endif

	string strModulePath;

	vector< HashString >::iterator pstrPath    = m_vstrSearchPaths.begin();
	vector< HashString >::iterator pstrPathEnd = m_vstrSearchPaths.end();

	for( ; pstrPath != pstrPathEnd; ++pstrPath )
	{
		strModulePath = pstrPath->m_strData + strFilename;

		/* Check to see if the file exists at this path */
		pFile = fopen( strModulePath.c_str(), "r" );

		if( pFile )
		{
			fclose( pFile );

			if( ( pkMod = new ModuleDynamic( rstrName, strModulePath, this ) )->IsValid() )
			{
				InsertModule( pkMod );
				return pkMod;
			}

			pkMod = 0;
		}
	}

	// On MacOS X, consider the module to be part of the NeoEngine core bundle, so pass true.
	// This doesn't mean anything on any other platform, so they'll just disregard it.
	if( ( pkMod = new ModuleDynamic( rstrName, strFilename, this, true ) )->IsValid() )
		InsertModule( pkMod );
	else
		pkMod = 0;

	if( !pkMod )
	{
		neolog << LogLevel( WARNING ) << "[ModuleManager::GetModule]: Unable to load module [" << rstrName << "]"
#ifdef POSIX
		       << endl <<                "                            dlerror [" << dlerror() << "]"
#endif
		       << endl;
	}

	return pkMod;
}


void ModuleManager::InsertModule( Module *pkModule )
{
	if( !pkModule )
		return;

	ModSymbol pfVersion = pkModule->LookupSymbol( "GetVersion" );

	if( !pfVersion )
	{
#ifndef BUILD_STATIC
		neolog << LogLevel( WARNING ) << "*** Module [" << pkModule->GetName() << "] does not export GetVersion symbol" << endl;
#endif
	}
	else
	{
		string strName;
		int iMajor, iMinor, iRevision;

		(( void ( NE_CDECL* )( std::string*, int*, int*, int* ))pfVersion)( &strName, &iMajor, &iMinor, &iRevision );

#ifndef BUILD_STATIC
		neolog << LogLevel( INFO ) << "Loading module " << strName << " v" << iMajor << "." << iMinor << "." << iRevision << endl;
#endif
	}

	ModSymbol pfInit = pkModule->LookupSymbol( "Initialize" );

	if( !pfInit )
	{
#ifndef BUILD_STATIC
		neolog << LogLevel( WARNING ) << "*** Module [" << pkModule->GetName() << "] does not export Initialize symbol" << endl;
#endif
	}
	else
		(( int ( NE_CDECL* )())pfInit)();

	pkModule->m_pkManager = this;

	m_hashModules.Insert( pkModule->GetName(), pkModule );
}


void ModuleManager::RemoveModule( Module *pkModule, bool bDeleted )
{
	if( !pkModule )
		return;

	if( !m_hashModules.Delete( pkModule->GetName(), pkModule ) || bDeleted )
		return;

	ModSymbol pfShut = pkModule->LookupSymbol( "Shutdown" );

	if( !pfShut )
		neolog << LogLevel( WARNING ) << "*** Module [" << pkModule->GetName() << "] does not export Shutdown symbol" << endl;
	else
		(( int ( NE_CDECL* )())pfShut)();
}




Module::Module( const HashString &rstrName, ModuleManager *pkManager ) :
	RefCounter(),
	m_pkManager( pkManager ),
	m_strName( rstrName )
{
}


Module::~Module()
{
	if( m_pkManager )
		m_pkManager->RemoveModule( this, true );
}




ModuleDynamic::ModuleDynamic( const HashString &rstrName, const HashString &rstrFullPath, ModuleManager *pkManager, bool bMacBundleLib ) :
	Module( rstrName, pkManager )
{
	m_Lib         = 0;
	m_strPathName = rstrFullPath;

#ifdef WIN32

	m_Lib         = (void*)LoadLibrary( m_strPathName.c_str() );

	if( !m_Lib )
	{
		LPVOID lpMsgBuf;

		FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, GetLastError(), MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ), (LPTSTR)&lpMsgBuf, 0, NULL );

		string strErr = (char*)lpMsgBuf;

		LocalFree( lpMsgBuf );

		neolog << LogLevel( WARNING ) << "*** Unable to load dynamic module [" << m_strName << "] : " << strErr << endl;	

		return;
	}

#elif __APPLE__

	m_bIsMacBundleLib = bMacBundleLib;

	// This is not a NeoEngine bundle lib, so we should have gotten a full path
	if( bMacBundleLib == false )
	{
		NSObjectFileImage objImage;
		NSObjectFileImageReturnCode ret;

		neolog << LogLevel( DEBUG ) << "[ModuleDynamic::ModuleDynamic] loading [" << m_strPathName << "]" << endl;

		ret = NSCreateObjectFileImageFromFile( m_strPathName.c_str(), &objImage );
		switch( ret )
		{
			case NSObjectFileImageSuccess:
				neolog << "NSCreateObjectFileImageFromFile successfully loaded file [" << m_strPathName << "]" << endl;
				break;

			case NSObjectFileImageFailure:
				neolog << "NSCreateObjectFileImageFromFile failed to load file [" << m_strPathName << "]" << endl;
				break;

			case NSObjectFileImageInappropriateFile:
				neolog << "NSCreateObjectFileImageFromFile [" << m_strPathName << "]: Inappropriate File type" << endl;
				break;

			case NSObjectFileImageArch:
				neolog << "NSCreateObjectFileImageFromFile [" << m_strPathName << "]: Different CPU architecture" << endl;
				break;

			case NSObjectFileImageFormat:
				neolog << "NSCreateObjectFileImageFromFile [" << m_strPathName << "]: This isn't a Mach-O file" << endl;
				break;

			case NSObjectFileImageAccess:
				neolog << "NSCreateObjectFileImageFromFile [" << m_strPathName << "]: Incorrect privileges" << endl;
				break;
		}

		unsigned long options = NSLINKMODULE_OPTION_RETURN_ON_ERROR;

		m_Lib = NSLinkModule(objImage, m_strPathName.c_str(), options );

		if( m_Lib )
		{
			neolog << "[ModuleDynamic::ModuleDynamic]: Successfully created module " << m_strPathName << endl;
		}
	}
	else  // It's a bundle inside the NeoEngine framework, and hence does not need a full path
	{
		// First search by bundle identifier to see if we've already loaded this bundle
		m_strPathName = "com.RealityRift.NeoEngine." + rstrName.m_strData;

		CFStringRef bundleName = CFStringCreateWithCString( NULL, m_strPathName.c_str(), kCFStringEncodingASCII );
		CFBundleRef pkBundle = CFBundleGetBundleWithIdentifier( bundleName );

		if( pkBundle != NULL )
		{
			// The bundle is already loaded
			m_Lib = pkBundle;
			return;
		}

		// We have to create the bundle.  Start by getting a
		// reference to the neoengine.framework bundle.
		CFStringRef frameworkName = CFStringCreateWithCString( NULL,
		                                                       "com.RealityRift.NeoEngine",
		                                                       kCFStringEncodingASCII );
		CFBundleRef frameworkBundle = CFBundleGetBundleWithIdentifier( frameworkName );

		if( !frameworkBundle )
		{
			neolog << LogLevel( ERROR ) << "[ModuleDynamic::ModuleDynamic]: Couldn't open framework handle" << endl;
		}

		// Now get the resource URL to our bundle that is contained within the framework.
		std::string s = rstrName.m_strData + ".bundle";

		CFStringRef cfBundleName = CFStringCreateWithCString( NULL,
		                                                      s.c_str(),
		                                                      kCFStringEncodingASCII );

		CFURLRef bundleURL = CFBundleCopyResourceURL( frameworkBundle, cfBundleName, NULL, NULL );
		if( !bundleURL )
		{
			neolog << LogLevel( ERROR ) << "[ModuleDynamic::ModuleDynamic]: Failed to get the bundle URL for " << rstrName << endl;
		}

		// Create a reference to the bundle
		m_Lib = CFBundleCreate( kCFAllocatorDefault, bundleURL );

		// And now load the code in the bundle
		if ( !CFBundleLoadExecutable( (CFBundleRef) m_Lib ))
		{
			neolog << LogLevel( ERROR ) << "[ModuleDynamic::ModuleDynamic]: Failed while loading " << rstrName << endl;
		}
	}

#elif POSIX

	m_Lib = dlopen( m_strPathName.c_str(), RTLD_NOW | RTLD_GLOBAL );

	if( !m_Lib )
		return;

#else
	#error "[ModuleDynamic::ModuleDynamic] ERROR Unsupported platform"
#endif
}


ModuleDynamic::~ModuleDynamic()
{
	if( IsValid() )
	{
		ModSymbol pfShut = LookupSymbol( "Shutdown" );

		if( !pfShut )
			neolog << LogLevel( WARNING ) << "*** Module [" << GetName() << "] does not export Shutdown symbol" << endl;
		else
			(( int ( NE_CDECL* )())pfShut)();
	}

#ifdef WIN32
	if( m_Lib )
		FreeLibrary( (HINSTANCE)m_Lib );
#elif __APPLE__
	if( m_bIsMacBundleLib )
	{
		CFBundleUnloadExecutable( (CFBundleRef) m_Lib );
	}
	else
	{
		NSUnLinkModule( m_Lib, NSUNLINKMODULE_OPTION_NONE );
	}
#elif POSIX
	if( m_Lib )
		dlclose( m_Lib );
#else
    #error "[ModuleDynamic::~ModuleDynamic] ERROR: Unsupported platform"
#endif
}


ModSymbol ModuleDynamic::LookupSymbol( const HashString &rstrSymbol )
{
	ModSymbol ret = 0;

#ifdef WIN32
	ret = (void*)GetProcAddress( (HINSTANCE)m_Lib, rstrSymbol.c_str() );
#elif __APPLE__

	if( m_bIsMacBundleLib )
	{
		CFStringRef cfstrSymbolName = CFStringCreateWithCString( NULL,
		                                                         rstrSymbol.c_str(),
		                                                         kCFStringEncodingASCII );

		ret = (ModSymbol) CFBundleGetFunctionPointerForName( (CFBundleRef) m_Lib, cfstrSymbolName);
	}
	else
	{
		string s = "_" + rstrSymbol.m_strData;
		NSSymbol symbol = NSLookupSymbolInModule( m_Lib, s.c_str() );

		if( symbol )
			ret = NSAddressOfSymbol (symbol);
	}
        
#elif POSIX
	ret = dlsym( m_Lib, rstrSymbol.c_str() );
#else
	#error "[ModuleDynamic::LookupSymbol] ERROR: Unsupported platform"
#endif

	if( !ret )
	{
		neolog << LogLevel( WARNING ) << "[ModuleDynamic::LookupSymbol]: Unable to find symbol " << rstrSymbol << endl;
	}
#ifdef __APPLE__
	else
	{
		neolog << LogLevel( DEBUG ) << "[ModuleDynamic::LookupSymbol]: Found symbol " << rstrSymbol << endl;
	}
#endif

	return ret;
}




ModuleStatic::ModuleStatic( const HashString &rstrName ) :
	Module( rstrName, 0 )
{
	//Make sure we don't get deallocated
	IncRef();
}


ModuleStatic::~ModuleStatic()
{
	//We might get deleted after the manager
	m_pkManager = 0;
}


ModSymbol ModuleStatic::LookupSymbol( const HashString &rstrSymbol )
{
	return m_Symbols.Find( rstrSymbol );
}


}; // namespace NeoEngine
