/***************************************************************************
                 mutex.cpp  -  Mutex wrapper and automatic lock
                             -------------------
    begin                : Tue Feb 12 2002
    copyright            : (C) 2002 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, mutex.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "mutex.h"
#include "core.h"
#include "logstream.h"

#include <stdlib.h>
#include <string.h>

using namespace std;


namespace NeoEngine
{


#ifndef NEOENGINE_DEBUGMUTEXLOCK
//Uncomment line below to get lock/unlock debug messages for mutexes
//#define NEOENGINE_DEBUGMUTEXLOCK
#endif


MutexObject::MutexObject( const string &rstrMutexName )
{
	m_strMutexName = rstrMutexName;

#ifdef NEOENGINE_DEBUGMUTEXLOCK
	neolog << LogLevel( DEBUG ) << "[mutex] created [" << m_strMutexName << "]" << endl;
#endif
	
#ifdef WIN32
	m_kMutex = CreateMutex( 0, FALSE, 0 );
	m_kEvent = CreateEvent( 0, FALSE, FALSE, 0 );
#elif defined(POSIX) || defined(__APPLE__)
	pthread_cond_init( &m_kCond, 0 );
	pthread_mutex_init( &m_kMutex, 0 );
#else
#  error MutexObject::MutexObject not implemented
#endif
}

MutexObject::~MutexObject()
{
	Unlock();

#ifdef WIN32
	ReleaseMutex( m_kMutex );
	CloseHandle( m_kMutex );
	CloseHandle( m_kEvent );
#elif defined(POSIX) || defined(__APPLE__)
	pthread_mutex_destroy( &m_kMutex );
#else
#  error MutexObject::~MutexObject not implemented
#endif

#ifdef NEOENGINE_DEBUGMUTEXLOCK
	neolog << LogLevel( DEBUG ) << "[mutex] destroyed [" << m_strMutexName << "]" << endl;
#endif
}


void MutexObject::Lock()
{
#ifdef WIN32
	if( WaitForSingleObject( m_kMutex, INFINITE ) == WAIT_FAILED )
#elif defined(POSIX) || defined(__APPLE__)
	if( pthread_mutex_lock( &m_kMutex ) < 0 )
#else
#  error MutexObject::Lock not implemented
#endif
	{
		neolog << LogLevel( ERROR ) << "unable to lock mutex [" << m_strMutexName << "]" << endl;
		return;
	}

#ifdef NEOENGINE_DEBUGMUTEXLOCK
	neolog << LogLevel( DEBUG ) << "[mutex] locked [" << m_strMutexName << "]" << endl;
#endif
}

void MutexObject::Unlock()
{
#ifdef WIN32
	ReleaseMutex( m_kMutex );
#elif defined(POSIX) || defined(__APPLE__)
	if( pthread_mutex_unlock( &m_kMutex ) < 0 )
		neolog << LogLevel( ERROR ) << "unable to unlock mutex [" << m_strMutexName << "]" << endl;
#else
#  error MutexObject::Unlock not implemented
#endif

#ifdef NEOENGINE_DEBUGMUTEXLOCK
	neolog << LogLevel( DEBUG ) << "[mutex] unlocked [" << m_strMutexName << "]" << endl;
#endif
}


void MutexObject::WaitForEvent()
{
#ifdef NEOENGINE_DEBUGMUTEXLOCK
	neolog << LogLevel( DEBUG ) << "[mutex] wait for event on [" << m_strMutexName << "]" << endl;
#endif

#ifdef WIN32
	WaitForSingleObject( m_kEvent, INFINITE );
#elif defined(POSIX) || defined(__APPLE__)
	Lock();

	pthread_cond_wait( &m_kCond, &m_kMutex );

	Unlock();
#else
#  error MutexObject::WaitForEvent not implemented
#endif

#ifdef NEOENGINE_DEBUGMUTEXLOCK
	neolog << LogLevel( DEBUG ) << "[mutex] event triggered on [" << m_strMutexName << "]" << endl;
#endif
}


void MutexObject::BroadcastEvent()
{
#ifdef NEOENGINE_DEBUGMUTEXLOCK
	neolog << LogLevel( DEBUG ) << "[mutex] broadcasting event for [" << m_strMutexName << "]" << endl;
#endif

#ifdef WIN32	
	PulseEvent( m_kEvent );
#elif defined(POSIX) || defined(__APPLE__)
	Lock();

	pthread_cond_broadcast( &m_kCond );

	Unlock();
#else
#  error MutexObject::BroadcastEvent not implemented
#endif

#ifdef NEOENGINE_DEBUGMUTEXLOCK
	neolog << LogLevel( DEBUG ) << "[mutex] broadcasted event to [" << m_strMutexName << "]" << endl;
#endif
}




MutexLock::MutexLock( MutexObject *pkObject )
{
	m_pkMutexObject = pkObject;
	m_pkMutexObject->Lock();
}


MutexLock::~MutexLock()
{
	m_pkMutexObject->Unlock();
}


}; // namespace NeoEngine
