/***************************************************************************
                         nemath.cpp  -  description
                             -------------------
    begin                : Tue Oct 30 2001
    copyright            : (C) 2001 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, nemath.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2001
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/


#include "nemath.h"
#include "util.h"

#include <assert.h>
#include <string.h>


namespace NeoEngine
{


const float PI            = 3.1415926535897932384626433832795f;
const float TWO_PI        = 2.0f * PI;
const float HALF_PI       = 0.5f * PI;
const float SQRT_3        = sqrtf( 3.0f );
const float HALF_SQRT_3   = 0.5f * sqrtf( 3.0f );
const float EPSILON       = 0.0001f;
const float SQREPSILON    = 0.000001f;


const Vector3d NE_STATIC Vector3d::ZERO( 0.0f, 0.0f, 0.0f );
const Vector3d NE_STATIC Vector3d::ORIGO( 0.0f, 0.0f, 0.0f );
const Vector3d NE_STATIC Vector3d::AXES[3] = { Vector3d( 1.0f, 0.0f, 0.0f ), Vector3d( 0.0f, 1.0f, 0.0f ), Vector3d( 0.0f, 0.0f, 1.0f ) };

const float NE_STATIC Matrix::g_afIdentityMatrix[16] = { 1.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f };
const Matrix NE_STATIC Matrix::IDENTITY( (const float*)g_afIdentityMatrix );

const Quaternion NE_STATIC Quaternion::IDENTITY( 0.0f, 0.0f, 0.0f, 1.0f );


std::ostream &operator << ( std::ostream &rkStream, const Vector3d &rkVector )
{
	rkStream << "( " << rkVector.x << ", " << rkVector.y << ", " << rkVector.z << " )";

	return rkStream;
}

std::ostream &operator << ( std::ostream &rkStream, const Quaternion &rkQuat )
{
	rkStream << "( " << rkQuat.qx << ", " << rkQuat.qy << ", " << rkQuat.qz << ", " << rkQuat.qw << " )";

	return rkStream;
}

std::ostream &operator << ( std::ostream &rkStream, const Matrix &rkMatrix )
{
	rkStream << "( ( " << rkMatrix.m_aafMatrix[0][0] << ", " << rkMatrix.m_aafMatrix[0][1] << ", " << rkMatrix.m_aafMatrix[0][2] << ", " << rkMatrix.m_aafMatrix[0][3] << " ),"
	         <<  " ( " << rkMatrix.m_aafMatrix[1][0] << ", " << rkMatrix.m_aafMatrix[1][1] << ", " << rkMatrix.m_aafMatrix[1][2] << ", " << rkMatrix.m_aafMatrix[1][3] << " ),"
	         <<  " ( " << rkMatrix.m_aafMatrix[2][0] << ", " << rkMatrix.m_aafMatrix[2][1] << ", " << rkMatrix.m_aafMatrix[2][2] << ", " << rkMatrix.m_aafMatrix[2][3] << " ),"
	         <<  " ( " << rkMatrix.m_aafMatrix[3][0] << ", " << rkMatrix.m_aafMatrix[3][1] << ", " << rkMatrix.m_aafMatrix[3][2] << ", " << rkMatrix.m_aafMatrix[3][3] << " ) )";

	return rkStream;
}


static inline unsigned int GetSafe( unsigned int uiOrder ) { static unsigned int aiSafe[] = { 0, 1, 2, 0 }; return( aiSafe[ uiOrder & 0x3 ] ); }
static inline unsigned int GetNext( unsigned int uiOrder ) { static unsigned int aiNext[] = { 1, 2, 0, 1 }; return( aiNext[ uiOrder & 0x3 ] ); }
static inline void         GetOrderData( unsigned int uiOrder, int &i, int &j, int &k, int &h, int &n, int &s, int &f )
{
	f = uiOrder & 0x1; uiOrder >>= 1;
	s = uiOrder & 0x1; uiOrder >>= 1;
	n = uiOrder & 0x1; uiOrder >>= 1;
	i = GetSafe( uiOrder & 0x3 );
	j = GetNext( i + n );
	k = GetNext( i + 1 - n );
	h = s ? k : i;
}


Quaternion &Quaternion::operator = ( const EulerAngles &rkEulerAngles )
{
	//From http://vered.rose.utoronto.ca/people/david_dir/GEMS/GEMS.html
	float fAngle[3] = { rkEulerAngles.m_kAngle.x, rkEulerAngles.m_kAngle.y, rkEulerAngles.m_kAngle.z };
	float a[3], ti, tj, th, ci, cj, ch, si, sj, sh, cc, cs, sc, ss;
	int i,j,k,h,n,s,f;

	GetOrderData( rkEulerAngles.m_uiOrder, i, j, k, h, n, s, f );

	if( f == EulerAngles::ROTATEFRAME ) { float t = fAngle[0]; fAngle[0] = fAngle[2]; fAngle[2] = t; }
	if( n == EulerAngles::ODD         ) { fAngle[1] = -fAngle[1]; }

	ti = fAngle[0] * 0.5f;
	tj = fAngle[1] * 0.5f;
	th = fAngle[2] * 0.5f;

	ci = cosf( ti );  cj = cosf( tj );  ch = cosf( th );
	si = sinf( ti );  sj = sinf( tj );  sh = sinf( th );

	cc = ci*ch; cs = ci*sh; sc = si*ch; ss = si*sh;

	if( s == EulerAngles::REPEAT )
	{
		a[i] = cj*(cs + sc);
		a[j] = sj*(cc + ss);
		a[k] = sj*(cs - sc);
		qw   = cj*(cc - ss);
	}
	else
	{
		a[i] = cj*sc - sj*cs;
		a[j] = cj*ss + sj*cc;
		a[k] = cj*cs - sj*sc;
		qw   = cj*cc + sj*ss;
	}
	
	if( n == EulerAngles::ODD ) a[j] = -a[j];
	
	qx = a[EulerAngles::X];
	qy = a[EulerAngles::Y];
	qz = a[EulerAngles::Z];
	
	return Normalize();
}


EulerAngles Quaternion::ToEulerAngles( unsigned int uiOrder ) const
{
	Matrix kM( *this );

	//From http://vered.rose.utoronto.ca/people/david_dir/GEMS/GEMS.html
	int i,j,k,h,n,s,f;
	Vector3d kAngles;

	GetOrderData( uiOrder, i, j, k, h, n, s, f );

	if( s == EulerAngles::REPEAT )
	{
		float sy = sqrtf( kM[i][j] * kM[i][j] + kM[i][k] * kM[i][k] );
	
		if( sy > EPSILON )
		{
			kAngles.x = atan2f( kM[i][j],  kM[i][k] );
			kAngles.y = atan2f( sy,        kM[i][i] );
			kAngles.z = atan2f( kM[j][i], -kM[k][i] );
		}
		else
		{
			kAngles.x = atan2f( -kM[j][k], kM[j][j] );
			kAngles.y = atan2f( sy,        kM[i][i] );
			kAngles.z = 0.0f;
		}
	}
	else
	{
		float cy = sqrtf( kM[i][i] * kM[i][i] + kM[j][i] * kM[j][i] );
	
		if( cy > EPSILON )
		{
			kAngles.x = atan2f(  kM[k][j], kM[k][k] );
			kAngles.y = atan2f( -kM[k][i], cy       );
			kAngles.z = atan2f(  kM[j][i], kM[i][i] );
		}
		else
		{
			kAngles.x = atan2f( -kM[j][k], kM[j][j] );
			kAngles.y = atan2f( -kM[k][i], cy       );
			kAngles.z = 0.0f;
		}
	}

	if( n == EulerAngles::ODD         ) { kAngles.x = -kAngles.x; kAngles.y = -kAngles.y; kAngles.z = -kAngles.z; }
	if( f == EulerAngles::ROTATEFRAME ) { float t = kAngles.x; kAngles.x = kAngles.z; kAngles.z = t; }
	
	return EulerAngles( kAngles, uiOrder );
}





ODE::ODE()
{
	m_iStateSize    = 0;
	m_ppfState      = 0;
	m_pfDerive      = 0;
}


ODE::~ODE()
{
	if( m_ppfState )
		delete [] m_ppfState, m_ppfState = 0;

	if( m_pfDerive )
		delete [] m_pfDerive, m_pfDerive = 0;
}


void ODE::SetStateSize( int iStateSize )
{
	m_iStateSize = iStateSize;

	if( m_pfDerive )
		delete [] m_pfDerive;

	if( m_ppfState )
		delete [] m_ppfState;

	m_ppfState  = new float*[ m_iStateSize ];
	m_pfDerive  = new float [ m_iStateSize ];
}


void ODE::EulerODE( float fDeltaTime, int iSteps )
{
	if( m_iStateSize == 0 )
		return;

	if( iSteps < 1 )
		iSteps = 1;

	float fStepTime = fDeltaTime / float( iSteps );

	while( iSteps )
	{
		Derive();

		for( int i = 0; i < m_iStateSize; i++ )
			*(m_ppfState[i]) += fStepTime * m_pfDerive[i];

		--iSteps;

		CalcAuxiliary( fStepTime );
	}

	PostODE();
}


}; // namespace NeoEngine
