/***************************************************************************
                          obb.cpp  -  description
                             -------------------
    begin                : Wed Sep 18 2002
    copyright            : (C) 2002 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, obb.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "obb.h"
#include "aabb.h"
#include "sphere.h"
#include "capsule.h"
#include "line.h"
#include "plane.h"
#include "contact.h"
#include "render.h"
#include "renderprimitive.h"
#include "vertex.h"

#include <limits>

#include <assert.h>

using namespace std;


namespace NeoEngine
{


OBB::OBB( const Vector3d &rkDim ) :
	BoundingVolume()
{
	m_kDim    = rkDim;
	m_fRadius = rkDim.Len();
}


void OBB::SetDim( const Vector3d &rkDim )
{
	m_kDim    = rkDim;
	m_fRadius = rkDim.Len();
}


bool OBB::Intersection( const Vector3d &rkPoint )
{
	Vector3d kDiff = rkPoint - m_kTranslation;

	Matrix kRotMat( m_kRotation );

	if( fabsf( kDiff * kRotMat.GetColumn(0) ) > m_kDim.x )
		return false;

	if( fabsf( kDiff * kRotMat.GetColumn(1) ) > m_kDim.y )
		return false;

	if( fabsf( kDiff * kRotMat.GetColumn(2) ) > m_kDim.z )
		return false;

	return true;
}


bool OBB::Intersection( const Line &rkLine )
{
	float fAWdU[3], fAWxDdU[3], fRhs;

	Vector3d kDiff = rkLine.m_kOrigin - m_kTranslation;
	Vector3d kWxD  = rkLine.m_kDirection % kDiff;

	Matrix kRotMat = m_kRotation;	

	Vector3d kAxis[3];

	kAxis[0] = kRotMat.GetColumn(0);
	kAxis[1] = kRotMat.GetColumn(1);
	kAxis[2] = kRotMat.GetColumn(2);

	fAWdU[1] = fabsf( rkLine.m_kDirection * kAxis[1] );
	fAWdU[2] = fabsf( rkLine.m_kDirection * kAxis[2] );
	fAWxDdU[0] = fabsf( kWxD * kAxis[0] );
	fRhs = m_kDim.y * fAWdU[2] + m_kDim.z * fAWdU[1];
	if( fAWxDdU[0] > fRhs )
		return false;

	fAWdU[0] = fabsf( rkLine.m_kDirection * kAxis[0] );
	fAWxDdU[1] = fabsf( kWxD * kAxis[1] );
	fRhs = m_kDim.x * fAWdU[2] + m_kDim.z * fAWdU[0];
	if( fAWxDdU[1] > fRhs )
		return false;

	fAWxDdU[2] = fabsf( kWxD * kAxis[2] );
	fRhs = m_kDim.x * fAWdU[1] + m_kDim.y * fAWdU[0];
	if( fAWxDdU[2] > fRhs )
		return false;

	return true;
}


bool OBB::Intersection( const Plane &rkPlane )
{
	Matrix kRotMat( m_kRotation );

	Vector3d kDist( m_kDim.x * ( rkPlane.m_kNormal * kRotMat.GetColumn(0) ),
	                  m_kDim.y * ( rkPlane.m_kNormal * kRotMat.GetColumn(1) ),
	                  m_kDim.z * ( rkPlane.m_kNormal * kRotMat.GetColumn(2) ) );

	float fRadius = fabsf( kDist.x ) + fabsf( kDist.y ) + fabsf( kDist.z );

	float fCenterDistToPlane = rkPlane.Distance( m_kTranslation );

	return( fabsf( fCenterDistToPlane ) <= fRadius );
}


bool OBB::Intersection( const Vector3d &rkV0, const Vector3d &rkV1, const Vector3d &rkV2, ContactSet *pkContactSet, bool bForceTriNormal, const Vector3d &rkNormal )
{
	assert( bForceTriNormal );

	Vector3d kDiff     = rkV0 - m_kTranslation;
	Vector3d akEdge[3] = { rkV1 - rkV0, rkV2 - rkV0, rkV2 - rkV1 };
	Vector3d kNormal   = ( akEdge[0] % akEdge[1] );//( rkNormal == Vector3d::ORIGO ) ? ( akEdge[0] % akEdge[1] ) : rkNormal;
	Vector3d kNormalized = kNormal;
	float fNormLen = kNormal.Len();
	kNormalized *= 1.0f / fNormLen;

	//Early out
	if( fabsf( kDiff * kNormalized ) > m_fRadius )
		return false;
	
	Matrix kRotMat( m_kRotation );

	Vector3d akAxes[3] = { kRotMat.GetColumn(0), kRotMat.GetColumn(1), kRotMat.GetColumn(2) };

	float fNormDotDiff;           // fNormDotDiff           = kNormal * kDiff;
	float fAxesDotNormalSum;      // Sum of all projected box axes along polygon normal
	float afNormDotAxis[3];       // afNormDotAxis[i]       = kNormal * akAxes[i]
	float afAxisDotEdge[3][3];    // afAxisDotEdge[i][j]    = akAxes[i] * akEdge[j]
	float afAbsAxisDotEdge[3][3]; // afAbsAxisDotEdge[i][j] = |afAxisDotEdge[i][j]|
	float fParallellThreshold = 0.999f * fNormLen;
	bool  bParallellAxes[3] = { false, false, false };

	float f0, f1, f2, fR;


	//Axis N
	afNormDotAxis[0] = kNormal * akAxes[0]; if( fabs( afNormDotAxis[0] ) > fParallellThreshold ) bParallellAxes[0] = true;
	afNormDotAxis[1] = kNormal * akAxes[1]; if( fabs( afNormDotAxis[1] ) > fParallellThreshold ) bParallellAxes[1] = true;
	afNormDotAxis[2] = kNormal * akAxes[2]; if( fabs( afNormDotAxis[2] ) > fParallellThreshold ) bParallellAxes[2] = true;

	assert( !( bParallellAxes[0] && bParallellAxes[1] ) && !( bParallellAxes[0] && bParallellAxes[2] ) && !( bParallellAxes[1] && bParallellAxes[2] ) );

	float fOverlapMax = -numeric_limits< float >::max();
	float fOverlap;

	int iDeepAxis = 0;
	Vector3d kCollNormal, kPlaneNormal;

	float fMin, fMax;

	fNormDotDiff      = kNormal * kDiff;
	fAxesDotNormalSum = m_kDim.x * fabsf( afNormDotAxis[0] ) + m_kDim.y * fabsf( afNormDotAxis[1] ) + m_kDim.z * fabsf( afNormDotAxis[2] );

	if( ( fOverlap = fabsf( fNormDotDiff ) - fAxesDotNormalSum ) > 0.0f )
		return false;

	if( pkContactSet )
	{
		fOverlapMax = fOverlap;
		kCollNormal = kNormal;
		iDeepAxis   = 0;
	}


	//Axis Ax
	afAxisDotEdge[0][0] = akAxes[0] * akEdge[0];
	afAxisDotEdge[0][1] = akAxes[0] * akEdge[1];

	f0 = akAxes[0] * kDiff;
	f1 = f0 + afAxisDotEdge[0][0];
	f2 = f0 + afAxisDotEdge[0][1];

	fR = m_kDim.x;

	MINMAX3( f0, f1, f2, fMin, fMax );

	if( ( fMin > fR ) || ( fMax < -fR ) )
		return false;
/*
	if( !bForceTriNormal && pkContactSet )
	{
		float fLeftOverlap  =  fMin - fR;
		float fRightOverlap = -fR   - fMax;

		if( ( fLeftOverlap > fRightOverlap ) && ( fLeftOverlap > fOverlapMax ) )
		{
			fOverlapMax = fLeftOverlap;
			kCollNormal = akAxes[0];
			iDeepAxis   = 1;
		}
		else if( fRightOverlap > fOverlapMax )
		{
			fOverlapMax = fRightOverlap;
			kCollNormal = -akAxes[0];
			iDeepAxis   = 1;
		}
	}
*/

	//Axis Ay
	afAxisDotEdge[1][0] = akAxes[1] * akEdge[0];
	afAxisDotEdge[1][1] = akAxes[1] * akEdge[1];

	f0 = akAxes[1] * kDiff;
	f1 = f0 + afAxisDotEdge[1][0];
	f2 = f0 + afAxisDotEdge[1][1];

	fR = m_kDim.y;

	MINMAX3( f0, f1, f2, fMin, fMax );

	if( ( fMin > fR ) || ( fMax < -fR ) )
		return false;
/*
	if( !bForceTriNormal && pkContactSet )
	{
		float fLeftOverlap  =  fMin - fR;
		float fRightOverlap = -fR   - fMax;

		if( ( fLeftOverlap > fRightOverlap ) && ( fLeftOverlap > fOverlapMax ) )
		{
			fOverlapMax = fLeftOverlap;
			kCollNormal = akAxes[1];
			iDeepAxis   = 2;
		}
		else if( fRightOverlap > fOverlapMax )
		{
			fOverlapMax = fRightOverlap;
			kCollNormal = -akAxes[1];
			iDeepAxis   = 2;
		}
	}
*/

	//Axis Az
	afAxisDotEdge[2][0] = akAxes[2] * akEdge[0];
	afAxisDotEdge[2][1] = akAxes[2] * akEdge[1];

	f0 = akAxes[2] * kDiff;
	f1 = f0 + afAxisDotEdge[2][0];
	f2 = f0 + afAxisDotEdge[2][1];

	fR = m_kDim.z;

	MINMAX3( f0, f1, f2, fMin, fMax );

	if( ( fMin > fR ) || ( fMax < -fR ) )
		return false;
/*
	if( !bForceTriNormal && pkContactSet )
	{
		float fLeftOverlap  =  fMin - fR;
		float fRightOverlap = -fR   - fMax;

		if( ( fLeftOverlap > fRightOverlap ) && ( fLeftOverlap > fOverlapMax ) )
		{
			fOverlapMax = fLeftOverlap;
			kCollNormal = akAxes[2];
			iDeepAxis   = 3;
		}
		else if( fRightOverlap > fOverlapMax )
		{
			fOverlapMax = fRightOverlap;
			kCollNormal = -akAxes[2];
			iDeepAxis   = 3;
		}
	}
*/

	if( bParallellAxes[0] || !bForceTriNormal )
	{

	//Axis Ax X E0

	kPlaneNormal = akAxes[0] % akEdge[0];
	f0 = kPlaneNormal * kDiff;
	f1 = f0 + afNormDotAxis[0];

	afAbsAxisDotEdge[2][0] = fabsf( afAxisDotEdge[2][0] );
	afAbsAxisDotEdge[1][0] = fabsf( afAxisDotEdge[1][0] );

	fR = m_kDim.y * afAbsAxisDotEdge[2][0] + m_kDim.z * afAbsAxisDotEdge[1][0];

	MINMAX( f0, f1, fMin, fMax );

	if( ( fMin > fR ) || ( fMax < -fR ) )
		return false;
/*
	if( !bForceTriNormal && pkContactSet )
	{
		float fLeftOverlap  =  fMin - fR;
		float fRightOverlap = -fR   - fMax;

		float fScale = kPlaneNormal.Len();

		if( fScale > 0.0f )
		{
			if( fLeftOverlap > fRightOverlap )
			{
				fLeftOverlap /= fScale;
				
				if( fLeftOverlap > fOverlapMax )
				{
					fOverlapMax = fLeftOverlap;
					kCollNormal = kPlaneNormal;
					iDeepAxis   = 4;
				}
			}
			else
			{
				fRightOverlap /= fScale;

				if( fRightOverlap > fOverlapMax )
				{
					fOverlapMax = fRightOverlap;
					kCollNormal = -kPlaneNormal;
					iDeepAxis   = 4;
				}
			}
		}
	}
*/

	//Axis Ax X E1
	kPlaneNormal = akAxes[0] % akEdge[1];
	f0 = kPlaneNormal * kDiff;
	f1 = f0 - afNormDotAxis[0];

	afAbsAxisDotEdge[2][1] = fabsf( afAxisDotEdge[2][1] );
	afAbsAxisDotEdge[1][1] = fabsf( afAxisDotEdge[1][1] );

	fR = m_kDim.y * afAbsAxisDotEdge[2][1] + m_kDim.z * afAbsAxisDotEdge[1][1];

	MINMAX( f0, f1, fMin, fMax );

	if( ( fMin > fR ) || ( fMax < -fR ) )
		return false;
/*
	if( !bForceTriNormal && pkContactSet )
	{
		float fLeftOverlap  =  fMin - fR;
		float fRightOverlap = -fR   - fMax;

		float fScale = kPlaneNormal.Len();

		if( fScale > 0.0f )
		{
			if( fLeftOverlap > fRightOverlap )
			{
				fLeftOverlap /= fScale;

				if( fLeftOverlap > fOverlapMax )
				{
					fOverlapMax = fLeftOverlap;
					kCollNormal = kPlaneNormal;
					iDeepAxis   = 5;
				}
			}
			else
			{
				fRightOverlap /= fScale;

				if( fRightOverlap > fOverlapMax )
				{
					fOverlapMax = fRightOverlap;
					kCollNormal = -kPlaneNormal;
					iDeepAxis   = 5;
				}
			}
		}
	}
*/

	//Axis Ax X E2
	kPlaneNormal = akAxes[0] % akEdge[2];
	f0 = kPlaneNormal * kDiff;
	f1 = f0 - afNormDotAxis[0];

	afAxisDotEdge[2][2]    = akAxes[2] * akEdge[2];
	afAxisDotEdge[1][2]    = akAxes[1] * akEdge[2];
	afAbsAxisDotEdge[2][2] = fabsf( afAxisDotEdge[2][2] );
	afAbsAxisDotEdge[1][2] = fabsf( afAxisDotEdge[1][2] );

	fR = m_kDim.y * afAbsAxisDotEdge[2][2] + m_kDim.z * afAbsAxisDotEdge[1][2];

	MINMAX( f0, f1, fMin, fMax );

	if( ( fMin > fR ) || ( fMax < -fR ) )
		return false;
/*
	if( !bForceTriNormal && pkContactSet )
	{
		float fLeftOverlap  =  fMin - fR;
		float fRightOverlap = -fR   - fMax;

		float fScale = kPlaneNormal.Len();

		if( fScale > 0.0f )
		{
            if( fLeftOverlap > fRightOverlap )
			{
				fLeftOverlap /= fScale;

				if( fLeftOverlap > fOverlapMax )
				{
					fOverlapMax = fLeftOverlap;
					kCollNormal = kPlaneNormal;
					iDeepAxis   = 6;
				}
			}
			else
			{
				fRightOverlap /= fScale;

				if( fRightOverlap > fOverlapMax )
				{
					fOverlapMax = fRightOverlap;
					kCollNormal = -kPlaneNormal;
					iDeepAxis   = 6;
				}
			}
		}
	}
*/
	}
	if( bParallellAxes[1] || !bForceTriNormal )
	{

	//Axis Ay X E0
	kPlaneNormal = akAxes[1] % akEdge[0];
	f0 = kPlaneNormal * kDiff;
	f1 = f0 + afNormDotAxis[1];

	afAbsAxisDotEdge[0][0] = fabsf( afAxisDotEdge[0][0] );
	afAbsAxisDotEdge[2][0] = fabsf( afAxisDotEdge[2][0] );

	fR = m_kDim.x * afAbsAxisDotEdge[2][0] + m_kDim.z * afAbsAxisDotEdge[0][0];

	MINMAX( f0, f1, fMin, fMax );

	if( ( fMin > fR ) || ( fMax < -fR ) )
		return false;
/*
	if( !bForceTriNormal && pkContactSet )
	{
		float fLeftOverlap  =  fMin - fR;
		float fRightOverlap = -fR   - fMax;

		float fScale = kPlaneNormal.Len();

		if( fScale > 0.0f )
		{
			if( fLeftOverlap > fRightOverlap )
			{
				fLeftOverlap /= fScale;

				if( fLeftOverlap > fOverlapMax )
				{
					fOverlapMax = fLeftOverlap;
					kCollNormal = kPlaneNormal;
					iDeepAxis   = 7;
				}
			}
            else
			{
				fRightOverlap /= fScale;

				if( fRightOverlap > fOverlapMax )
				{
					fOverlapMax = fRightOverlap;
					kCollNormal = -kPlaneNormal;
					iDeepAxis   = 7;
				}
			}
		}
	}
*/

	//Axis Ay X E1
	kPlaneNormal = akAxes[1] % akEdge[1];
	f0 = kPlaneNormal * kDiff;
	f1 = f0 - afNormDotAxis[1];

	afAbsAxisDotEdge[0][1] = fabsf( afAxisDotEdge[0][1] );
	afAbsAxisDotEdge[2][1] = fabsf( afAxisDotEdge[2][1] );

	fR = m_kDim.x * afAbsAxisDotEdge[2][1] + m_kDim.z * afAbsAxisDotEdge[0][1];

	MINMAX( f0, f1, fMin, fMax );

	if( ( fMin > fR ) || ( fMax < -fR ) )
		return false;
/*
	if( !bForceTriNormal && pkContactSet )
	{
		float fLeftOverlap  =  fMin - fR;
		float fRightOverlap = -fR   - fMax;

		float fScale = kPlaneNormal.Len();

		if( fScale > 0.0f )
		{
			if( fLeftOverlap > fRightOverlap )
			{
				fLeftOverlap /= fScale;

				if( fLeftOverlap > fOverlapMax )
				{
					fOverlapMax = fLeftOverlap;
					kCollNormal = kPlaneNormal;
					iDeepAxis   = 8;
				}
			}
            else
			{
				fRightOverlap /= fScale;

				if( fRightOverlap > fOverlapMax )
				{
					fOverlapMax = fRightOverlap;
					kCollNormal = -kPlaneNormal;
					iDeepAxis   = 8;
				}
			}
		}
	}
*/

	//Axis Ay X E2
	kPlaneNormal = akAxes[1] % akEdge[2];
	f0 = kPlaneNormal * kDiff;
	f1 = f0 - afNormDotAxis[1];

	afAxisDotEdge[0][2]    = akAxes[0] * akEdge[2];
	afAxisDotEdge[2][2]    = akAxes[2] * akEdge[2];
	afAbsAxisDotEdge[0][2] = fabsf( afAxisDotEdge[0][2] );
	afAbsAxisDotEdge[2][2] = fabsf( afAxisDotEdge[2][2] );

	fR = m_kDim.x * afAbsAxisDotEdge[2][2] + m_kDim.z * afAbsAxisDotEdge[0][2];

	MINMAX( f0, f1, fMin, fMax );

	if( ( fMin > fR ) || ( fMax < -fR ) )
		return false;
/*
	if( !bForceTriNormal && pkContactSet )
	{
		float fLeftOverlap  =  fMin - fR;
		float fRightOverlap = -fR   - fMax;

		float fScale = kPlaneNormal.Len();

		if( fScale > 0.0f )
		{
			if( fLeftOverlap > fRightOverlap )
			{
				fLeftOverlap /= fScale;

				if( fLeftOverlap > fOverlapMax )
				{
					fOverlapMax = fLeftOverlap;
					kCollNormal = kPlaneNormal;
					iDeepAxis   = 9;
				}
			}
            else
			{
				fRightOverlap /= fScale;

				if( fRightOverlap > fOverlapMax )
				{
					fOverlapMax = fRightOverlap;
					kCollNormal = -kPlaneNormal;
					iDeepAxis   = 9;
				}
			}
		}
	}
*/
	}
	if( bParallellAxes[2] || !bForceTriNormal )
	{

	//Axis Az X E0
	kPlaneNormal = akAxes[2] % akEdge[0];
	f0 = kPlaneNormal * kDiff;
	f1 = f0 + afNormDotAxis[2];

	afAbsAxisDotEdge[1][0] = fabsf( afAxisDotEdge[1][0] );
	afAbsAxisDotEdge[0][0] = fabsf( afAxisDotEdge[0][0] );

	fR = m_kDim.x * afAbsAxisDotEdge[1][0] + m_kDim.y * afAbsAxisDotEdge[0][0];

	MINMAX( f0, f1, fMin, fMax );

	if( ( fMin > fR ) || ( fMax < -fR ) )
		return false;
/*
	if( !bForceTriNormal && pkContactSet )
	{
		float fLeftOverlap  =  fMin - fR;
		float fRightOverlap = -fR   - fMax;

		float fScale = kPlaneNormal.Len();

		if( fScale > 0.0f )
		{
			if( fLeftOverlap > fRightOverlap )
			{
				fLeftOverlap /= fScale;

				if( fLeftOverlap > fOverlapMax )
				{
					fOverlapMax = fLeftOverlap;
					kCollNormal = kPlaneNormal;
					iDeepAxis   = 10;
				}
			}
            else
			{
				fRightOverlap /= fScale;

				if( fRightOverlap > fOverlapMax )
				{
					fOverlapMax = fRightOverlap;
					kCollNormal = -kPlaneNormal;
					iDeepAxis   = 10;
				}
			}
		}
	}
*/

	//Axis Az X E1
	kPlaneNormal = akAxes[2] % akEdge[1];
	f0 = kPlaneNormal * kDiff;
	f1 = f0 - afNormDotAxis[2];

	afAbsAxisDotEdge[1][1] = fabsf( afAxisDotEdge[1][1] );
	afAbsAxisDotEdge[0][1] = fabsf( afAxisDotEdge[0][1] );

	fR = m_kDim.x * afAbsAxisDotEdge[1][1] + m_kDim.y * afAbsAxisDotEdge[0][1];

	MINMAX( f0, f1, fMin, fMax );

	if( ( fMin > fR ) || ( fMax < -fR ) )
		return false;
/*
	if( !bForceTriNormal && pkContactSet )
	{
		float fLeftOverlap  =  fMin - fR;
		float fRightOverlap = -fR   - fMax;

		float fScale = kPlaneNormal.Len();

		if( fScale > 0.0f )
		{
			if( fLeftOverlap > fRightOverlap )
			{
				fLeftOverlap /= fScale;

				if( fLeftOverlap > fOverlapMax )
				{
					fOverlapMax = fLeftOverlap;
					kCollNormal = kPlaneNormal;
					iDeepAxis   = 11;
				}
			}
            else
			{
				fRightOverlap /= fScale;

				if( fRightOverlap > fOverlapMax )
				{
					fOverlapMax = fRightOverlap;
					kCollNormal = -kPlaneNormal;
					iDeepAxis   = 11;
				}
			}
		}
	}
*/

	//Axis Az X E2
	kPlaneNormal = akAxes[2] % akEdge[2];
	f0 = kPlaneNormal * kDiff;
	f1 = f0 - afNormDotAxis[2];

	afAxisDotEdge[0][2]    = akAxes[0] * akEdge[2];
	afAxisDotEdge[1][2]    = akAxes[1] * akEdge[2];
	afAbsAxisDotEdge[0][2] = fabsf( afAxisDotEdge[0][2] );
	afAbsAxisDotEdge[1][2] = fabsf( afAxisDotEdge[1][2] );

	fR = m_kDim.x * afAbsAxisDotEdge[1][2] + m_kDim.y * afAbsAxisDotEdge[0][2];

	MINMAX( f0, f1, fMin, fMax );

	if( ( fMin > fR ) || ( fMax < -fR ) )
		return false;
/*
	if( !bForceTriNormal && pkContactSet )
	{
		float fLeftOverlap  =  fMin - fR;
		float fRightOverlap = -fR   - fMax;

		float fScale = kPlaneNormal.Len();

		if( fScale > 0.0f )
		{
			if( fLeftOverlap > fRightOverlap )
			{
				fLeftOverlap /= fScale;

				if( fLeftOverlap > fOverlapMax )
				{
					fOverlapMax = fLeftOverlap;
					kCollNormal = kPlaneNormal;
					iDeepAxis   = 12;
				}
			}
            else
			{
				fRightOverlap /= fScale;

				if( fRightOverlap > fOverlapMax )
				{
					fOverlapMax = fRightOverlap;
					kCollNormal = -kPlaneNormal;
					iDeepAxis   = 12;
				}
			}
		}
	}
*/
	}

	if( !pkContactSet )
		return true;

	if( iDeepAxis == 0 )
	{
		//Box intersecting triangle face
		Contact *pkContact = new Contact;

		kNormal = kNormalized;

		int aiAxes[3][3] = { { 0, 1, 2 }, { 1, 0, 2 }, { 2, 0, 1 } };

		for( int iAxis = 0; iAxis < 3; ++iAxis )
		{
			int iMajor = aiAxes[iAxis][0];
			int iOther[2] = { aiAxes[iAxis][1], aiAxes[iAxis][2] };

			//Calculate intersecting face center relative reference face
			Vector3d kCenter;

			if( afNormDotAxis[ iMajor ] > 0.0f )
				kCenter = ( akAxes[ iMajor ] * -m_kDim[ iMajor ] ) - kDiff;
			else
				kCenter = ( akAxes[ iMajor ] *  m_kDim[ iMajor ] ) - kDiff;

			Plane kBoxPlane( ( afNormDotAxis[ iMajor ] > 0.0f ) ? -akAxes[ iMajor ] : akAxes[ iMajor ], kCenter + rkV0 );

			float fDenom   = kNormal * kBoxPlane.m_kNormal;

			if( fabsf( fDenom ) < EPSILON )
				continue;

			//Project intersecting face onto reference face coordinate space (i.e a 2D coordinate quad)

			//Construct coordinate system
			Vector3d kLocalX = akEdge[0];
			Vector3d kLocalY = kNormal % kLocalX;

			float afCenter[2];
			float aafTri[3][2];
			float aafNormal[3][2];
			float aafQuad[4][2];
			float aafClipped[7][2];

			afCenter[0] = kCenter * kLocalX;
			afCenter[1] = kCenter * kLocalY;

			float fXOffset[2], fYOffset[2];

			float afLocalXDotAxis[2] = { afAxisDotEdge[ iOther[0] ][0], afAxisDotEdge[ iOther[1] ][0] };
			float afLocalYDotAxis[2] = { akAxes[ iOther[0] ] * kLocalY, akAxes[ iOther[1] ] * kLocalY };

			fXOffset[0] = m_kDim[ iOther[0] ] * afLocalXDotAxis[0];
			fYOffset[0] = m_kDim[ iOther[0] ] * afLocalYDotAxis[0];
			fXOffset[1] = m_kDim[ iOther[1] ] * afLocalXDotAxis[1];
			fYOffset[1] = m_kDim[ iOther[1] ] * afLocalYDotAxis[1];

			aafQuad[0][0] = afCenter[0] - fXOffset[0] - fXOffset[1];
			aafQuad[0][1] = afCenter[1] - fYOffset[0] - fYOffset[1];
			aafQuad[1][0] = afCenter[0] - fXOffset[0] + fXOffset[1];
			aafQuad[1][1] = afCenter[1] - fYOffset[0] + fYOffset[1];
			aafQuad[2][0] = afCenter[0] + fXOffset[0] + fXOffset[1];
			aafQuad[2][1] = afCenter[1] + fYOffset[0] + fYOffset[1];
			aafQuad[3][0] = afCenter[0] + fXOffset[0] - fXOffset[1];
			aafQuad[3][1] = afCenter[1] + fYOffset[0] - fYOffset[1];

			aafTri[0][0] = 0.0f;
			aafTri[0][1] = 0.0f;
			aafTri[1][0] = akEdge[0] * kLocalX;
			aafTri[1][1] = 0.0f;
			aafTri[2][0] = akEdge[1] * kLocalX;
			aafTri[2][1] = akEdge[1] * kLocalY;

			//aafTri[1][0] > 0

			if( aafTri[2][1] < 0.0f )
			{
				aafNormal[0][0] =  0.0f;
				aafNormal[0][1] =  aafTri[1][0];
				aafNormal[1][0] = -aafTri[2][1]; //aafTri[1][1] == 0
				aafNormal[1][1] =  ( aafTri[2][0] - aafTri[1][0] );
				aafNormal[2][0] =  aafTri[2][1]; //aafTri[0][1] == 0
				aafNormal[2][1] = -aafTri[2][0]; //aafTri[0][0] == 0
			}
			else
			{
				aafNormal[0][0] =  0.0f;
				aafNormal[0][1] = -aafTri[1][0];
				aafNormal[1][0] =  aafTri[2][1]; //aafTri[1][1] == 0
				aafNormal[1][1] = -( aafTri[2][0] - aafTri[1][0] );
				aafNormal[2][0] = -aafTri[2][1]; //aafTri[0][1] == 0
				aafNormal[2][1] =  aafTri[2][0]; //aafTri[0][0] == 0
			}

			int iNum = ClipPolyPoly2D( (float*)aafTri, (float*)aafNormal, 3, (float*)aafQuad, 4, (float*)aafClipped );

			if( !iNum )
				continue;

			//Un-project points into reference face coordinate 3D space and calculate depths
			//(ignore points with positive depth)
			float fNormDot = 1.0f / fDenom;

			float f0Factor = 1.0f / aafTri[1][0];
			float f1Factor = 1.0f / ( kLocalY * kLocalY );

			for( int iPoint = 0; iPoint < iNum; ++iPoint )
			{
				float f0 = aafClipped[ iPoint ][0] * f0Factor;
				float f1 = aafClipped[ iPoint ][1] * f1Factor;

				Vector3d kPoint = rkV0 + ( kLocalX * f0 ) + ( kLocalY * f1 );

				float fDepth = kBoxPlane.Distance( kPoint ) * fNormDot;

				// ### TEST: clamp very small negative depths to zero
				if( fDepth > -EPSILON )
				{
					if( fDepth < EPSILON )
						fDepth = 0.0f;

					assert( pkContact->m_uiNumPoints < 8 );

					unsigned int iTest;

					for( iTest = 0; iTest < pkContact->m_uiNumPoints; ++iTest )
						if( pkContact->m_pkPoints[ iTest ] == kPoint )
							break;

					if( ( iTest >= pkContact->m_uiNumPoints ) && ( pkContact->m_uiNumPoints < 8 ) )
					{
						pkContact->m_pkPoints[ pkContact->m_uiNumPoints   ] = kPoint + ( kNormal * -fDepth  );
						pkContact->m_pfDepths[ pkContact->m_uiNumPoints++ ] = fDepth;
					}
				}
			}
		}

		pkContact->m_kNormal = kNormal;

		if( !pkContact->m_uiNumPoints )
		{
			//FIXME: This should not happen
			delete pkContact;
			return false;
		}

		pkContact->Sort();

		pkContactSet->Add( pkContact );
	}
/*	else if( iDeepAxis < 4 )
	{
		//Triangle is intersecting box face

		//Rotate normal into world space
		kCollNormal = kRotMat * kCollNormal;

		//Get "local" (in reference face coordinate space) x and y coordinate axes
		int iLocalX, iLocalY;
		int iNormal = iDeepAxis - 1;

		if( iNormal == 0 )
		{
			iLocalX = 1;
			iLocalY = 2;
		}
		else if( iNormal == 1 )
		{
			iLocalX = 0;
			iLocalY = 2;
		}
		else
		{
			iLocalX = 0;
			iLocalY = 1;
		}

		//Project intersecting face onto reference face coordinate space (i.e a 2D coordinate triangle)
		float aafTriangle[3][2];

		Vector3d &kLocalX = akAxes[ iLocalX ];
		Vector3d &kLocalY = akAxes[ iLocalY ];

		aafTriangle[0][0] = ( rkV0 - m_kTranslation ) * kLocalX;
		aafTriangle[0][1] = ( rkV0 - m_kTranslation ) * kLocalY;

		aafTriangle[1][0] = ( rkV1 - m_kTranslation ) * kLocalX;
		aafTriangle[1][1] = ( rkV1 - m_kTranslation ) * kLocalY;

		aafTriangle[2][0] = ( rkV2 - m_kTranslation ) * kLocalX;
		aafTriangle[2][1] = ( rkV2 - m_kTranslation ) * kLocalY;

		float afDim[2] = { m_kDim[ iLocalX ], m_kDim[ iLocalY ] };

		float aafClipped[6][2];

		int iNum = ClipRectPoly2D( afDim, (float*)aafTriangle, 3, (float*)aafClipped );

		if( !iNum )
		{
			// ### FIXME: Wtf?
			//assert( iNum > 0 );
			return false;
		}

		Contact *pkContact = new Contact;

		Plane kTriPlane( kNormal, rkV0 );

		//Un-project points into reference face coordinate 3D space and calculate depths
		//(ignore points with positive depth)

		for( int iPoint = 0; iPoint < iNum; ++iPoint )
		{
			Vector3d kPoint = m_kTranslation + ( kCollNormal * m_kDim[ iNormal ] ) + ( akAxes[ iLocalX ] * aafClipped[ iPoint ][0] ) + ( akAxes[ iLocalY ] * aafClipped[ iPoint ][1] );

			float fDepth = kTriPlane.Distance( kPoint ) / ( kCollNormal * kTriPlane.m_kNormal );

			// ### TEST: clamp very small negative depths to zero
			if( fDepth > -EPSILON )
			{
				if( fDepth < EPSILON )
					fDepth = 0.0f;

				pkContact->m_pkPoints[ pkContact->m_uiNumPoints   ] = kPoint + ( kCollNormal * -fDepth );
				pkContact->m_pfDepths[ pkContact->m_uiNumPoints++ ] = fDepth;
			}
		}

		if( !pkContact->m_uiNumPoints )
		{
			delete pkContact;
			return false;
		}

		pkContact->Sort();

		pkContact->m_kNormal = -kCollNormal;

		pkContactSet->Add( pkContact );
	}
	else
	{
		// edge-edge contact
		Contact *pkContact = new Contact;

		kCollNormal.Normalize();

		//Rotate normal into world space
		kCollNormal = kRotMat * kCollNormal;

		pkContact->m_kNormal = kCollNormal;

		//Calculate point on edges
		Vector3d kPoint0 = GetTranslation();
		Vector3d kPoint1;

		kPoint0 += ( akAxes[0] * ( m_kDim[0] * ( ( ( kCollNormal * akAxes[0] ) > 0.0f ) ?  1.0f : -1.0f ) ) ) +
		           ( akAxes[1] * ( m_kDim[1] * ( ( ( kCollNormal * akAxes[1] ) > 0.0f ) ?  1.0f : -1.0f ) ) ) +
		           ( akAxes[2] * ( m_kDim[2] * ( ( ( kCollNormal * akAxes[2] ) > 0.0f ) ?  1.0f : -1.0f ) ) );

		float fT0, fT1;

		int iAxisBox = ( iDeepAxis - 4 ) / 3;
		int iEdgeTri = ( iDeepAxis - 4 ) % 3;

		if( iEdgeTri == 0 )
			kPoint1 = rkV0;
		else if( iEdgeTri == 0 )
			kPoint1 = rkV1;
		else
			kPoint1 = rkV2;

		Vector3d kDir0 = akAxes[ iAxisBox ];
		Vector3d kDir1 = akEdge[ iEdgeTri ];

		//Get closest point t values between the two lines defined by
		//  l0 = kPoint0 + fT0 * kDir0
		//  l1 = kPoint1 + fT1 * kDir1

		ClosestPointLineLine( kPoint0, kDir0, kPoint1, kDir1, &fT0, &fT1 );

		kPoint0 += kDir0 * fT0;
		kPoint1 += kDir1 * fT1;

		pkContact->m_pfDepths[0] = -fOverlapMax;
		pkContact->m_pkPoints[0] = ( kPoint0 + kPoint1 ) * 0.5f;
		pkContact->m_uiNumPoints = 1;

		pkContactSet->Add( pkContact );
	}
*/
	return true;
}


//Helper method
inline bool Clip( float fDenom, float fNumer, float *pfT0/*, float *pfT1*/ )
{
	if ( fDenom > 0.0f )
	{
		/*if( fNumer > ( fDenom * (*pfT1) ) )
			return false;*/
		if( fNumer > ( fDenom * (*pfT0) ) )
			*pfT0 = fNumer / fDenom;
		return true;
	}
	else if( fDenom < 0.0f )
	{
		if( fNumer > ( fDenom * (*pfT0) ) )
			return false;
		/*if( fNumer > ( fDenom * (*pfT1) ) )
			*pfT1 = fNumer / fDenom;*/
		return true;
	}
	else
		return( fNumer <= 0.0f );
}


bool OBB::Intersection( const Ray &rkRay, ContactSet *pkContactSet )
{
	// convert ray to box coordinates
	Vector3d kOrig = rkRay.m_kOrigin;
	Vector3d kDiff = kOrig - m_kTranslation;

	Matrix kRot( m_kRotation );

	Vector3d kOrigin( kDiff * kRot.GetColumn(0), kDiff * kRot.GetColumn(1), kDiff * kRot.GetColumn(2) );
	Vector3d kDirection( rkRay.m_kDirection * kRot.GetColumn(0), rkRay.m_kDirection * kRot.GetColumn(1), rkRay.m_kDirection * kRot.GetColumn(2) );

	float fT0 = 0.0f; //fT1 = 1.0f;

	bool bNotEntirelyClipped =
	    Clip( +kDirection.x, -kOrigin.x - m_kDim.x, &fT0 ) &&
	    Clip( -kDirection.x, +kOrigin.x - m_kDim.x, &fT0 ) &&
	    Clip( +kDirection.y, -kOrigin.y - m_kDim.y, &fT0 ) &&
	    Clip( -kDirection.y, +kOrigin.y - m_kDim.y, &fT0 ) &&
	    Clip( +kDirection.z, -kOrigin.z - m_kDim.z, &fT0 ) &&
	    Clip( -kDirection.z, +kOrigin.z - m_kDim.z, &fT0 );

	if( bNotEntirelyClipped && ( fT0 != 0.0f ) ) //( ( fT0 != 0.0f ) || ( fT1 != 1.0f ) ) )
	{
		if( pkContactSet )
		{
			Contact *pkData = new Contact;

			pkData->m_pkPoints[0] = kOrig + ( rkRay.m_kDirection * fT0 );
			pkData->m_pfDepths[0] = fT0;
			pkData->m_uiNumPoints = 1;

			pkContactSet->Add( pkData );
        }

		return true;
    }

	return false;
}


void OBB::Generate( AABB *pkAABB )
{
	m_kRotation    = pkAABB->GetRotation();
	m_kTranslation = pkAABB->GetTranslation();
	m_kDim         = pkAABB->GetRealDim();
	m_fRadius      = pkAABB->GetRadius();
}


void OBB::Generate( OBB *pkOBB )
{
	m_kRotation    = pkOBB->GetRotation();
	m_kTranslation = pkOBB->GetTranslation();
	m_kDim         = pkOBB->m_kDim;
	m_fRadius      = pkOBB->m_fRadius;
}


void OBB::Generate( Sphere *pkSphere )
{
	m_kRotation.Reset();
	m_kTranslation = pkSphere->GetTranslation();
	m_kDim.Set( pkSphere->m_fRadius, pkSphere->m_fRadius, pkSphere->m_fRadius );
	m_fRadius      = pkSphere->m_fRadius;
}


void OBB::Generate( Capsule *pkCapsule )
{
	m_kRotation     = pkCapsule->GetRotation();
	m_kTranslation = pkCapsule->GetTranslation();
	m_kDim.Set( pkCapsule->m_fLength + pkCapsule->m_fRadius, pkCapsule->m_fRadius, pkCapsule->m_fRadius );
	m_fRadius = pkCapsule->m_fLength + pkCapsule->m_fRadius;
}


void OBB::Generate( VertexBufferPtr &pkVertexBuffer )
{
	AABB kAABB;

	kAABB.Generate( pkVertexBuffer );

	Generate( &kAABB );
}


void OBB::Merge( AABB *pkAABB )
{
	OBB kOBB;

	kOBB.SetRotation( pkAABB->GetRotation() );
	kOBB.SetTranslation( pkAABB->GetTranslation() );
	kOBB.m_kDim = pkAABB->GetRealDim();

	Merge( &kOBB );
}


void OBB::Merge( OBB *pkOBB )
{
	//Calculate new translation
	Vector3d kNewTranslation( ( m_kTranslation + pkOBB->GetTranslation() ) * 0.5f );

	//Calculate new rotation
	const Quaternion &rkOBBQuat = pkOBB->GetRotation();

	Quaternion kNewQuat;

	kNewQuat.qx = m_kRotation.qx + rkOBBQuat.qx;
	kNewQuat.qy = m_kRotation.qy + rkOBBQuat.qy;
	kNewQuat.qz = m_kRotation.qz + rkOBBQuat.qz;
	kNewQuat.qw = m_kRotation.qw + rkOBBQuat.qw;

	kNewQuat.Normalize();

	//Calculate new dimension
	Vector3d kNewDim;

	Matrix kRotMat[2] = { m_kRotation, rkOBBQuat };

	Vector3d kBoxAxes[2][3] = { { kRotMat[0].GetColumn(0), kRotMat[0].GetColumn(1), kRotMat[0].GetColumn(2) },
	                              { kRotMat[1].GetColumn(0), kRotMat[1].GetColumn(1), kRotMat[1].GetColumn(2) } };

	Vector3d kBoxCenter[2]  = { m_kTranslation, pkOBB->GetTranslation() };

	Vector3d kBoxDim[2]     = { kRotMat[0] * m_kDim, kRotMat[1] * pkOBB->m_kDim };

	//Loop all corners of both boxes
	Vector3d kCorner[2][8];
	Vector3d kDiff[2];
	float fDot;

	int i, iAxis;

	for( i = 0; i < 2; ++i )
	{
		kCorner[i][0]  = kBoxCenter[i] + kBoxDim[i];

		kCorner[i][1]    = kBoxCenter[i];
		kCorner[i][1].x += kBoxDim[i].x;
		kCorner[i][1].y += kBoxDim[i].y;
		kCorner[i][1].z -= kBoxDim[i].z;

		kCorner[i][2]    = kBoxCenter[i];
		kCorner[i][2].x -= kBoxDim[i].x;
		kCorner[i][2].y += kBoxDim[i].y;
		kCorner[i][2].z += kBoxDim[i].z;

		kCorner[i][3]    = kBoxCenter[i];
		kCorner[i][3].x -= kBoxDim[i].x;
		kCorner[i][3].y += kBoxDim[i].y;
		kCorner[i][3].z -= kBoxDim[i].z;

		kCorner[i][4]    = kBoxCenter[i];
		kCorner[i][4].x += kBoxDim[i].x;
		kCorner[i][4].y -= kBoxDim[i].y;
		kCorner[i][4].z += kBoxDim[i].z;

		kCorner[i][5]    = kBoxCenter[i];
		kCorner[i][5].x += kBoxDim[i].x;
		kCorner[i][5].y -= kBoxDim[i].y;
		kCorner[i][5].z -= kBoxDim[i].z;

		kCorner[i][6]    = kBoxCenter[i];
		kCorner[i][6].x -= kBoxDim[i].x;
		kCorner[i][6].y += kBoxDim[i].y;
		kCorner[i][6].z += kBoxDim[i].z;

		kCorner[i][7]    = kBoxCenter[i] - kBoxDim[i];
	}

	for( i = 0; i < 8; ++i )
	{
		kDiff[0] = kCorner[0][i] - kNewTranslation;
		kDiff[1] = kCorner[1][i] - kNewTranslation;

		for( iAxis = 0; iAxis < 3; ++iAxis )
		{
			fDot = MAX( kBoxAxes[0][ iAxis ] * kDiff[0], kBoxAxes[1][ iAxis ] * kDiff[1] );

			if( fDot > kNewDim[ iAxis ] )
				kNewDim[ iAxis ] = fDot;
		}
	}

	SetDim( kNewDim );
	SetTranslation( kNewTranslation );
	SetRotation( kNewQuat );
}


void OBB::Merge( Sphere *pkSphere )
{
	//Sphere is merged as an OBB with same rotation
	OBB kOBB;

	kOBB.SetTranslation( pkSphere->GetTranslation() );
	kOBB.SetRotation( pkSphere->GetRotation() );

	kOBB.m_kDim.Set( pkSphere->m_fRadius, pkSphere->m_fRadius, pkSphere->m_fRadius );

	Merge( &kOBB );
}


void OBB::Merge( Capsule *pkCapsule )
{
	//Capsule is merged as an OBB with same rotation
	OBB kOBB;

	kOBB.SetTranslation( pkCapsule->GetTranslation() );
	kOBB.SetRotation( pkCapsule->GetRotation() );

	kOBB.m_kDim.Set( pkCapsule->m_fLength + pkCapsule->m_fRadius, pkCapsule->m_fRadius, pkCapsule->m_fRadius );

	Merge( &kOBB );
}


void OBB::RenderOutlines( const Color &rkColor ) const
{
	VertexBufferPtr pkBuffer = Core::Get()->GetRenderDevice()->CreateVertexBuffer( Buffer::DYNAMIC, 24, &DiffuseVertex::s_kDecl );

	pkBuffer->Lock( Buffer::WRITE );
	{
		DiffuseVertex *pkVertex = (DiffuseVertex*)pkBuffer->GetVertex();

		pkVertex->m_kPosition.Set(  m_kDim.x,  m_kDim.y,  m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;
		pkVertex->m_kPosition.Set(  m_kDim.x, -m_kDim.y,  m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;

		pkVertex->m_kPosition.Set(  m_kDim.x, -m_kDim.y,  m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;
		pkVertex->m_kPosition.Set( -m_kDim.x, -m_kDim.y,  m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;

		pkVertex->m_kPosition.Set( -m_kDim.x, -m_kDim.y,  m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;
		pkVertex->m_kPosition.Set( -m_kDim.x,  m_kDim.y,  m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;

		pkVertex->m_kPosition.Set( -m_kDim.x,  m_kDim.y,  m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;
		pkVertex->m_kPosition.Set(  m_kDim.x,  m_kDim.y,  m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;

		pkVertex->m_kPosition.Set(  m_kDim.x,  m_kDim.y,  m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;
		pkVertex->m_kPosition.Set(  m_kDim.x,  m_kDim.y, -m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;

		pkVertex->m_kPosition.Set(  m_kDim.x, -m_kDim.y,  m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;
		pkVertex->m_kPosition.Set(  m_kDim.x, -m_kDim.y, -m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;

		pkVertex->m_kPosition.Set( -m_kDim.x, -m_kDim.y,  m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;
		pkVertex->m_kPosition.Set( -m_kDim.x, -m_kDim.y, -m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;

		pkVertex->m_kPosition.Set( -m_kDim.x,  m_kDim.y,  m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;
		pkVertex->m_kPosition.Set( -m_kDim.x,  m_kDim.y, -m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;

		pkVertex->m_kPosition.Set(  m_kDim.x,  m_kDim.y, -m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;
		pkVertex->m_kPosition.Set(  m_kDim.x, -m_kDim.y, -m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;

		pkVertex->m_kPosition.Set(  m_kDim.x, -m_kDim.y, -m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;
		pkVertex->m_kPosition.Set( -m_kDim.x, -m_kDim.y, -m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;

		pkVertex->m_kPosition.Set( -m_kDim.x, -m_kDim.y, -m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;
		pkVertex->m_kPosition.Set( -m_kDim.x,  m_kDim.y, -m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;

		pkVertex->m_kPosition.Set( -m_kDim.x,  m_kDim.y, -m_kDim.z ); pkVertex->m_kColor = rkColor; ++pkVertex;
		pkVertex->m_kPosition.Set(  m_kDim.x,  m_kDim.y, -m_kDim.z ); pkVertex->m_kColor = rkColor;
	}
	pkBuffer->Unlock();

	RenderPrimitive kPrimitive;

	kPrimitive.m_ePrimitive      = RenderPrimitive::LINES;
	kPrimitive.m_uiNumPrimitives = 12;
	kPrimitive.m_pkVertexBuffer  = pkBuffer;
	kPrimitive.m_fSize           = 1.0f;

	kPrimitive.m_kModelMatrix.Set( GetRotation(), GetTranslation() );

	Core::Get()->GetRenderDevice()->Render( kPrimitive );

	DiffuseVertex akAxes[6];
	
	akAxes[0].m_kPosition = Vector3d::ORIGO;   akAxes[0].m_kColor = Color32( 0, (unsigned char)0xFF, 0 );
	akAxes[1].m_kPosition = Vector3d::AXES[0]; akAxes[1].m_kColor = Color32( 0, (unsigned char)0xFF, 0 );
	akAxes[2].m_kPosition = Vector3d::ORIGO;   akAxes[2].m_kColor = Color32( (unsigned char)0xFF, 0, 0 );
	akAxes[3].m_kPosition = Vector3d::AXES[1]; akAxes[3].m_kColor = Color32( (unsigned char)0xFF, 0, 0 );
	akAxes[4].m_kPosition = Vector3d::ORIGO;   akAxes[4].m_kColor = Color32( 0, 0, (unsigned char)0xFF );
	akAxes[5].m_kPosition = Vector3d::AXES[2]; akAxes[5].m_kColor = Color32( 0, 0, (unsigned char)0xFF );

	kPrimitive.m_uiNumPrimitives = 3;
	kPrimitive.m_pkVertexBuffer = Core::Get()->GetRenderDevice()->CreateVertexBuffer( Buffer::NORMAL, 6, &DiffuseVertex::s_kDecl, akAxes );

	Core::Get()->GetRenderDevice()->Render( kPrimitive );
}


BoundingVolume *OBB::Duplicate() const
{
	return new OBB( *this );
}

}; // namespace NeoEngine
