/***************************************************************************
                    particle.cpp  -  Particle system classes
                             -------------------
    begin                : Thu May 9 2002
    copyright            : (C) 2002 by Mattias Jansson
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, particle.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "particle.h"
#include "room.h"
#include "timer.h"
#include "logstream.h"

using namespace std;


namespace NeoEngine
{


//#define NEOENGINE_PROFILE_PARTICLESYSTEM


ParticleSystem::ParticleSubSystem::~ParticleSubSystem()
{
	for( int i = 0; i < NUMPARTICLES; ++i )
		delete m_apkParticles[i];
}




ParticleSystem::ParticleSystem( ParticleGenerator *pkGenerator, unsigned int uiMaxParticles, float fTTL, float fTTLAccuracy ) :
	SceneNode()
{
	m_pkGenerator            = pkGenerator;
	m_uiMaxParticles         = uiMaxParticles;
	m_fNormalTTL             = fTTL;
	m_fTTLAccuracy           = fTTLAccuracy;
	m_uiNumParticles         = 0;
	m_ppkSubSystems          = 0;
	m_uiNumSubSystems        = 1;
	m_uiGeneratePoolSize     = int( pkGenerator->m_fDensity / 3.0f );
	m_ppkGeneratePool        = m_uiGeneratePoolSize ? new Particle*[ m_uiGeneratePoolSize ] : 0;
	m_uiNumFreeParticles     = 0;
	m_pkRoom                 = 0;
	m_bGenerate              = true;

	if( m_uiMaxParticles )
		m_uiNumSubSystems    = ( m_uiMaxParticles / ParticleSubSystem::NUMPARTICLES ) + 1;

	m_uiNumSubSystemPointers = ( ( m_uiNumSubSystems * 2 ) < 10 ) ? 10 : ( m_uiNumSubSystems * 2 );

	m_ppkSubSystems          = new ParticleSubSystem*[ m_uiNumSubSystemPointers ];

	for( unsigned int uiSubSystem = 0; uiSubSystem < m_uiNumSubSystems; ++uiSubSystem )
		m_ppkSubSystems[ uiSubSystem ] = new ParticleSubSystem;
}


ParticleSystem::~ParticleSystem()
{
	delete m_pkGenerator;

	//Delete freed particles
	for( unsigned int uiFree = 0; uiFree < m_uiNumFreeParticles; ++uiFree )
		delete m_ppkGeneratePool[ uiFree ];

	for( unsigned int i = 0; i < m_uiNumSubSystems; ++i )
	{
		ParticleSubSystem *pkSubSystem = m_ppkSubSystems[ i ];

		if( m_pkRoom )
		{
			for( unsigned int uiSlot = 0; uiSlot < pkSubSystem->m_uiNumUsedSlots; ++uiSlot )
				m_pkRoom->DetachNode( pkSubSystem->m_apkParticles[ uiSlot ] );
		}

		delete pkSubSystem;
	}

	delete [] m_ppkSubSystems;

	delete [] m_ppkGeneratePool;
}


SceneNode *ParticleSystem::Duplicate()
{
	assert( !"[ParticleSystem::Duplicate] Not implemented!" );
	return 0;
}


void ParticleSystem::Update( float fDeltaTime )
{
	float fTimeMultiplier = 0.01f * fDeltaTime;

#ifdef NEOENGINE_PROFILE_PARTICLESYSTEM
	Timer kTimer;
#endif

    //Update existing particles
	for( unsigned int uiSubSystem = 0; uiSubSystem < m_uiNumSubSystems; ++uiSubSystem )
	{
		ParticleSubSystem *pkSubSystem = m_ppkSubSystems[ uiSubSystem ];

		for( unsigned int uiSlot = 0; uiSlot < pkSubSystem->m_uiNumUsedSlots; )
		{
			if( ( pkSubSystem->m_afLifetime[ uiSlot ] + fDeltaTime ) > pkSubSystem->m_afTTL[ uiSlot ] )
			{
				Particle *pkParticle = pkSubSystem->m_apkParticles[ uiSlot ];

				pkSubSystem->FreeSlot( uiSlot );

				--m_uiNumParticles;

				if( m_uiGeneratePoolSize == m_uiNumFreeParticles )
					delete pkParticle; //No free space in generator pool, delete particle
				else
				{
					if( m_pkRoom )
						m_pkRoom->DetachNode( pkParticle );

					m_ppkGeneratePool[ m_uiNumFreeParticles++ ] = pkParticle;
				}

				assert( m_uiNumFreeParticles <= m_uiGeneratePoolSize );
			}
			else
			{
				pkSubSystem->m_apkParticles[ uiSlot ]->Update( fDeltaTime );
				pkSubSystem->m_afLifetime[ uiSlot ] += fDeltaTime;

				++uiSlot;
			}
		}
	}

#ifdef NEOENGINE_PROFILE_PARTICLESYSTEM
	float fElapsedTime = kTimer.GetDeltaTime( false );

	neolog << LogLevel( DEBUG ) << "[Particle system] update loop: " << setprecision( 3 ) << fElapsedTime << endl;
#endif

	//Generate new particles if needed
	if( m_bGenerate && ( !m_uiMaxParticles || ( m_uiNumParticles < m_uiMaxParticles ) ) )
	{
#ifdef NEOENGINE_PROFILE_PARTICLESYSTEM
		kTimer.Reset();
#endif

		unsigned int uiMaxGenerate = m_uiMaxParticles - m_uiNumParticles;

		if( m_uiGeneratePoolSize < uiMaxGenerate )
		{
			Particle **ppkPool = new Particle*[ ( m_uiGeneratePoolSize = uiMaxGenerate ) ];

			//Copy pointers to freed particles
			if( m_ppkGeneratePool )
			{
				memcpy( ppkPool, m_ppkGeneratePool, sizeof( Particle* ) * m_uiNumFreeParticles );
				memset( ppkPool + m_uiNumFreeParticles, 0, sizeof( Particle* ) * ( m_uiGeneratePoolSize - m_uiNumFreeParticles ) );

				delete [] m_ppkGeneratePool;
			}
			else
				memset( ppkPool, 0, sizeof( Particle* ) * m_uiGeneratePoolSize );

			m_ppkGeneratePool = ppkPool;
		}

		//Populate generate pool with pointers

		unsigned int uiGenerated = m_pkGenerator->Generate( fDeltaTime, uiMaxGenerate, m_ppkGeneratePool, GetWorldTranslation() );

#ifdef NEOENGINE_PROFILE_PARTICLESYSTEM
		fElapsedTime = kTimer.GetDeltaTime( false );

		neolog << LogLevel( DEBUG ) << "[Particle system] Generate loop: " << setprecision( 3 ) << fElapsedTime << endl;

		kTimer.Reset();
#endif

		unsigned int uiSubSystem = 0;
		ParticleSubSystem *ppkSubSystem = m_ppkSubSystems[ uiSubSystem ];

		for( unsigned int uiParticle = 0; uiParticle < uiGenerated; ++uiParticle )
		{
			int iSlot = 0;

			while( ( iSlot = ppkSubSystem->GetFreeSlot() ) < 0 )
			{
				if( ++uiSubSystem >= m_uiNumSubSystems )
				{
					if( uiSubSystem >= m_uiNumSubSystemPointers )
					{
						ParticleSubSystem **ppkSubSystems = new ParticleSubSystem*[ m_uiNumSubSystemPointers + 5 ];

						fmemcpy( ppkSubSystems, m_ppkSubSystems, sizeof( ParticleSubSystem* ) * m_uiNumSubSystemPointers );

						delete [] m_ppkSubSystems;

						m_ppkSubSystems = ppkSubSystems;

						m_uiNumSubSystemPointers += 5;
					}

					m_ppkSubSystems[ uiSubSystem ] = new ParticleSubSystem;

					++m_uiNumSubSystems;
				}

				ppkSubSystem = m_ppkSubSystems[ uiSubSystem ];
			}

			Particle *pkParticle = m_ppkGeneratePool[ uiParticle ];

			//Update new particle
			float fTime = float( rand() % 100 ) * fTimeMultiplier;
			
			pkParticle->Update( fTime );
			
			if( m_pkRoom )
				m_pkRoom->AttachNode( pkParticle );

			ppkSubSystem->m_apkParticles[ iSlot ] = pkParticle;
			ppkSubSystem->m_afLifetime[ iSlot ]   = fTime;

			if( m_fNormalTTL != 0.0f )
				ppkSubSystem->m_afTTL[ iSlot ]    = m_fNormalTTL + m_fNormalTTL * ( ( rand() % 2 ) ? 1.0f : -1.0f ) * ( 1.0f - m_fTTLAccuracy );
			else
				ppkSubSystem->m_afTTL[ iSlot ]    = 0.0f;

			++ppkSubSystem->m_uiNumUsedSlots;
		}


		m_uiNumParticles += uiGenerated;


		if( uiGenerated >= m_uiNumFreeParticles )
		{
			m_uiNumFreeParticles = 0;

			memset( m_ppkGeneratePool, 0, sizeof( Particle* ) * m_uiGeneratePoolSize );
		}
		else
		{
			m_uiNumFreeParticles -= uiGenerated;

			memmove( m_ppkGeneratePool, m_ppkGeneratePool + uiGenerated, sizeof( Particle* ) * m_uiNumFreeParticles );
			memset( m_ppkGeneratePool + m_uiNumFreeParticles, 0, sizeof( Particle* ) * uiGenerated );
		}

#ifdef NEOENGINE_PROFILE_PARTICLESYSTEM
		fElapsedTime = kTimer.GetDeltaTime( false );

		neolog << LogLevel( DEBUG ) << "[Particle system] Insert loop: " << setprecision( 3 ) << fElapsedTime << endl;
#endif
	}
}


void ParticleSystem::SetRoom( Room *pkRoom )
{
	m_pkRoom = pkRoom;
}







void ParticleGenerator::Allocate( unsigned int uiNumParticles, Particle **ppkParticles )
{
	if( m_pkAllocatorCallback )
		m_pkAllocatorCallback->AllocateParticles( uiNumParticles, ppkParticles );
	else
	{
		for( unsigned int uiCurrent = 0; uiCurrent < uiNumParticles; ++uiCurrent )
			if( !ppkParticles[ uiCurrent ] )
				ppkParticles[ uiCurrent ] = new Particle;
	}
}






SphereParticleGenerator::SphereParticleGenerator( float fVelocity, float fVelocityAccuracy, float fDensity, ParticleAllocatorCallback *pkAllocatorCallback ) :
	ParticleGenerator( fVelocity, fVelocityAccuracy, fDensity, pkAllocatorCallback )
{
}


SphereParticleGenerator::~SphereParticleGenerator()
{
}


unsigned int SphereParticleGenerator::Generate( float fDeltaTime, unsigned int uiMaxParticles, Particle **ppkParticles, const Vector3d &kPos )
{
	m_fFraction += fDeltaTime * m_fDensity;

	unsigned int uiToGenerate = (int)floor( m_fFraction );

	if( !uiToGenerate )
		return 0;

	m_fFraction -= float( uiToGenerate );

	if( uiToGenerate > uiMaxParticles )
		uiToGenerate = uiMaxParticles;

	Allocate( uiToGenerate, ppkParticles );

	Vector3d kVel;

	for( unsigned int uiCurrent = 0; uiCurrent < uiToGenerate; ++uiCurrent )
	{
		kVel.Set( float( double( rand() ) / double( RAND_MAX - 1 ) ) * 2.0f - 1.0f,
		          float( double( rand() ) / double( RAND_MAX - 1 ) ) * 2.0f - 1.0f,
		          float( double( rand() ) / double( RAND_MAX - 1 ) ) * 2.0f - 1.0f );

		kVel.ScaleTo( m_fVelocity * ( 1.0f + ( 2.0f * float( double( rand() ) / double( RAND_MAX - 1 ) ) - 1.0f ) * ( 1.0f - m_fVelocityAccuracy ) ) );

		ppkParticles[ uiCurrent ]->SetVelocity( kVel );
		ppkParticles[ uiCurrent ]->SetTranslation( kPos );
	}

	return uiToGenerate;
}


}; // namespace NeoEngine
