/***************************************************************************
                          patch.cpp  -  Patch surfaces
                             -------------------
    begin                : Thu May 2 2002
    copyright            : (C) 2002 by Mattias Jansson
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, patch.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "patch.h"
#include "vertex.h"
#include "polygon.h"
#include "render.h"
#include "logstream.h"

#include <assert.h>

using namespace std;


namespace NeoEngine
{


#ifndef NEOENGINE_DEBUGPATCHTESSELATOR
//define to get patch tesselator debug messages
//#define NEOENGINE_DEBUGPATCHTESSELATOR
#endif


//FIXME: Use lookup table instead?
static inline int Comb( int a, int b )
{
	uint64_t i;
	uint64_t j;
	uint64_t k = 1;
	int u;

	if( ( a - b ) < b )
	{
		for( u = 2; u <= ( a - b ); ++u ) k *= u;

		j = k;

		for( ; u <= b; ++u ) j *= u;
	}
	else
	{
		for( u = 2; u <= b; ++u ) k *= u;
		
		j = k;

		for( ; u <= ( a - b ); ++u ) j *= u;
	}

	i = j;

	for( ; u <= a; ++u ) i *= u;

	return int( i / (j*k) );
}



PatchSurfaceData::PatchSurfaceData()
{
	m_pkControlPoints = 0;
	m_ePatchType      = BEZIER;
	m_eVisibleSide    = VISIBLE_FRONT;
	m_iControlWidth   = 0;
	m_iControlHeight  = 0;
	m_iULevel         = -1;
	m_iVLevel         = -1;
	m_iTessWidth      = -1;
	m_iTessHeight     = -1;
}


PatchSurfaceData::~PatchSurfaceData()
{
	m_pkControlPoints = 0;
}


const PatchSurfaceData &PatchSurfaceData::operator =( const PatchSurfaceData &rkPatchData )
{
	m_ePatchType      = rkPatchData.m_ePatchType;
	m_eVisibleSide    = rkPatchData.m_eVisibleSide;
	m_iControlWidth   = rkPatchData.m_iControlWidth;
	m_iControlHeight  = rkPatchData.m_iControlHeight;
	m_iULevel         = rkPatchData.m_iULevel;
	m_iVLevel         = rkPatchData.m_iVLevel;
	m_iTessWidth      = rkPatchData.m_iTessWidth;
	m_iTessHeight     = rkPatchData.m_iTessHeight;

	if( rkPatchData.m_pkControlPoints )
	{
		//Ugly const cast
		VertexBuffer *pkVertexBuffer = (VertexBuffer*)( (const VertexBuffer*)rkPatchData.m_pkControlPoints );

		pkVertexBuffer->Lock( Buffer::READ );

		m_pkControlPoints = Core::Get()->GetRenderDevice()->CreateVertexBuffer( Buffer::DYNAMIC | Buffer::NORENDER, pkVertexBuffer->GetNumElements(), pkVertexBuffer->GetVertexFormat(), pkVertexBuffer->GetVertex() );

		pkVertexBuffer->Unlock();
	}
	else
		m_pkControlPoints = 0;

	return( *this );
}





PatchSurfaceSubMesh::PatchSurfaceSubMesh( const PatchSurfaceData &rkPatchData ) :
	SubMesh()
{
	m_bNeedUpdate     = true;
	m_kPatchData      = rkPatchData;

//	if( ( m_kPatchData.m_ePatchType == PatchSurfaceData::BEZIER ) &&
//	    ( m_kPatchData.m_iControlWidth  == 3 ) &&
//	    ( m_kPatchData.m_iControlHeight == 3 ) )
//		m_pkTesselator  = new QuadraticBezierPatchTesselator;
//	else
		m_pkTesselator  = new PatchSurfaceTesselator;
}


PatchSurfaceSubMesh::PatchSurfaceSubMesh( const PatchSurfaceSubMesh &rkSubMesh ) :
	SubMesh( rkSubMesh )
{
	m_bNeedUpdate     = true;
	m_pkVertices      = 0;
	m_pkPolygons      = 0;
	m_kPatchData      = rkSubMesh.m_kPatchData;

//	if( ( m_kPatchData.m_ePatchType == PatchSurfaceData::BEZIER ) &&
//	    ( m_kPatchData.m_iControlWidth  == 3 ) &&
//	    ( m_kPatchData.m_iControlHeight == 3 ) )
//		m_pkTesselator  = new QuadraticBezierPatchTesselator;
//	else
		m_pkTesselator  = new PatchSurfaceTesselator;
}


PatchSurfaceSubMesh::~PatchSurfaceSubMesh()
{
	delete m_pkTesselator;
}


void PatchSurfaceSubMesh::UpdateData()
{
	m_bNeedUpdate = false;

	m_pkTesselator->Tesselate( &m_kPatchData, this );
}


SubMesh *PatchSurfaceSubMesh::Duplicate() const
{
	return( new PatchSurfaceSubMesh( *this ) );
}


		




PatchSurfaceTesselator::PatchSurfaceTesselator()
{
}


PatchSurfaceTesselator::~PatchSurfaceTesselator()
{
}


void PatchSurfaceTesselator::SetSubdivisionLevel( PatchSurfaceData *pkPatchData, int iULevel, int iVLevel )
{
	if( !pkPatchData )
		return;

	assert( (int)pkPatchData->m_pkControlPoints->GetNumElements() >= ( pkPatchData->m_iControlHeight * pkPatchData->m_iControlWidth ) );

	if( iULevel < 0 )
		iULevel = 2;

	if( iVLevel < 0 )
		iVLevel = 2;

	if( pkPatchData->m_iULevel != iULevel )
	{
#ifdef NEOENGINE_DEBUGPATCHTESSELATOR
		neolog << LogLevel( DEBUG ) << "[PatchSurfaceTesselator::SetSubdivisionLevel] set u level to " << iULevel << endl;
#endif
		pkPatchData->m_iULevel = iULevel;
	}

	if( pkPatchData->m_iVLevel != iVLevel )
	{
#ifdef NEOENGINE_DEBUGPATCHTESSELATOR
		neolog << LogLevel( DEBUG ) << "[PatchSurfaceTesselator::SetSubdivisionLevel] set v level to " << iVLevel << endl;
#endif
		pkPatchData->m_iVLevel = iVLevel;	
	}

	pkPatchData->m_iTessWidth  = pkPatchData->m_iControlWidth  + ( ( pkPatchData->m_iControlWidth  - 1 ) * pkPatchData->m_iULevel );
	pkPatchData->m_iTessHeight = pkPatchData->m_iControlHeight + ( ( pkPatchData->m_iControlHeight - 1 ) * pkPatchData->m_iVLevel );
}


void PatchSurfaceTesselator::Tesselate( PatchSurfaceData *pkPatchData, PatchSurfaceSubMesh *pkSubMesh )
{
	assert( pkPatchData->m_ePatchType == PatchSurfaceData::BEZIER );

	VertexBufferPtr  pkVertexBuffer  = pkSubMesh->m_pkVertices;
	PolygonBufferPtr pkPolygonBuffer = pkSubMesh->m_pkPolygons;

#ifdef NEOENGINE_DEBUGPATCHTESSELATOR
	neolog << LogLevel( DEBUG ) << "[PatchSurfaceTesselator::Tesselate]: starting tesselation of bezier patch" << endl;
#endif
	
	if( !pkPatchData->m_pkControlPoints->GetNumElements() )
	{
		neolog << LogLevel( WARNING ) << "[PatchSurfaceTesselator::Tesselate] tesselation aborted, no control pointers" << endl;
		return;
	}

	if( ( pkPatchData->m_iTessWidth <= 0 ) || ( pkPatchData->m_iTessHeight <= 0 ) )
		SetSubdivisionLevel( pkPatchData, pkPatchData->m_iULevel, pkPatchData->m_iVLevel );

	if( ( pkPatchData->m_iTessWidth <= 0 ) || ( pkPatchData->m_iTessHeight <= 0 ) )
	{
		neolog << LogLevel( WARNING ) << "[PatchSurfaceTesselator::Tesselate] tesselation aborted, zero tesselation dimensions" << endl;
		return;
	}

#ifdef NEOENGINE_DEBUGPATCHTESSELATOR
	neolog << LogLevel( DEBUG ) << "[PatchSurfaceTesselator::Tesselate] allocating buffer for [" << pkPatchData->m_iTessWidth << "x" << pkPatchData->m_iTessHeight << "] vertices" << endl;
#endif
	
	if( !pkVertexBuffer )
		pkSubMesh->m_pkVertices = ( pkVertexBuffer = Core::Get()->GetRenderDevice()->CreateVertexBuffer( Buffer::WRITEPRIORITIZED | Buffer::DYNAMIC ) );

	pkVertexBuffer->AllocateVertices( pkPatchData->m_iTessWidth * pkPatchData->m_iTessHeight, pkPatchData->m_pkControlPoints->GetVertexFormat() );

	pkVertexBuffer->Lock( Buffer::WRITE );

	//Interpolate vertices
	InterpolateVertices( pkPatchData, pkVertexBuffer );

	//Create polygon buffer
	{
		int iVinc, iUinc, iV, iU, iIterations;
		int iVcount, iUcount;

		if( pkPatchData->m_eVisibleSide == PatchSurfaceData::VISIBLE_BOTH )
		{
			iIterations = 2;
			iVinc       = 1;
			iV          = 0;
		}
		else
		{
			iIterations = 1;

			if( pkPatchData->m_eVisibleSide == PatchSurfaceData::VISIBLE_BACK )
			{
				iVinc = -1;
				iV    = pkPatchData->m_iTessHeight - 1;
			}
			else
			{
				iVinc = 1;
				iV    = 0;
			}
		}

		//Allocate buffer
#ifdef NEOENGINE_DEBUGPATCHTESSELATOR
		neolog << LogLevel( DEBUG ) << "[PatchSurfaceTesselator::Tesselate] allocating buffer for ["
		       << ( pkPatchData->m_iTessWidth - 1 ) * ( pkPatchData->m_iTessHeight - 1 ) * 2 * iIterations << " polygons" << endl;
#endif

		if( !pkPolygonBuffer )
			pkSubMesh->m_pkPolygons = ( pkPolygonBuffer = Core::Get()->GetRenderDevice()->CreatePolygonBuffer( Buffer::DYNAMIC | Buffer::READPRIORITIZED | Buffer::WRITEPRIORITIZED, 0 ) );

		pkPolygonBuffer->AllocatePolygons( ( pkPatchData->m_iTessWidth - 1 ) * ( pkPatchData->m_iTessHeight - 1 ) * 2 * iIterations );

		pkPolygonBuffer->Lock( Buffer::WRITE );

		Polygon *pkPoly = pkPolygonBuffer->GetPolygon( 0 );

		while( iIterations-- )
		{
			//Make strip-compatible order
			iU      = 0;
			iUinc   = 1;
			iVcount = pkPatchData->m_iTessHeight - 1;

			while( iVcount-- )
			{
				iUcount = pkPatchData->m_iTessWidth - 1;

				while( iUcount-- )
				{
					pkPoly->v[ 0 ] = (   iV           * pkPatchData->m_iTessWidth ) + ( iU + iUinc );
					pkPoly->v[ 1 ] = ( ( iV + iVinc ) * pkPatchData->m_iTessWidth ) + ( iU + iUinc );
					pkPoly->v[ 2 ] = (   iV           * pkPatchData->m_iTessWidth ) +   iU;

					++pkPoly;

					pkPoly->v[ 0 ] = (   iV           * pkPatchData->m_iTessWidth ) +   iU;
					pkPoly->v[ 1 ] = ( ( iV + iVinc ) * pkPatchData->m_iTessWidth ) + ( iU + iUinc );
					pkPoly->v[ 2 ] = ( ( iV + iVinc ) * pkPatchData->m_iTessWidth ) +   iU;

					++pkPoly;

					iU += iUinc;
				}

				iV += iVinc;
				iU  = 0;
			}

			iV    = pkPatchData->m_iTessHeight - 1;
			iVinc = -iVinc;
		}

		pkPolygonBuffer->Unlock();

		pkPolygonBuffer->Stripify();
	}


	const VertexDeclaration *pkVertexFormat  = pkVertexBuffer->GetVertexFormat();
	const VertexElement     *pkNormalElement = pkVertexFormat->GetElement( VertexElement::FLOAT3, VertexElement::NORMAL );
	unsigned int               uiVertexSize    = pkVertexBuffer->GetVertexSize();

	if( pkNormalElement )
	{
#ifdef NEOENGINE_DEBUGPATCHTESSELATOR
		neolog << LogLevel( DEBUG ) << "[PatchSurfaceTesselator::Tesselate] calculating polygon normals" << endl;
#endif

		pkPolygonBuffer->CalculateNormals( pkVertexBuffer, true );

		Vector3d *pkPolyNormalBuffer = pkPolygonBuffer->GetNormals();

#ifdef NEOENGINE_DEBUGPATCHTESSELATOR
		neolog << LogLevel( DEBUG ) << "[PatchSurfaceTesselator::Tesselate] calculating vertex normals" << endl;
#endif

		unsigned char *pucVertex = (unsigned char*)pkVertexBuffer->GetVertex() + pkNormalElement->m_uiOffset;
		unsigned int   uiVert;
		
		for( uiVert = 0; uiVert < pkVertexBuffer->GetNumElements(); ++uiVert, pucVertex += uiVertexSize )
			((Vector3d*)pucVertex)->Reset();

		pkPolygonBuffer->Lock( Buffer::READ );

		Polygon  *pkPoly       = pkPolygonBuffer->GetPolygon( 0 );
		Vector3d *pkPolyNormal = pkPolyNormalBuffer;

		unsigned int uiNumPolygons = pkPolygonBuffer->GetNumElements();

		pucVertex = (unsigned char*)pkVertexBuffer->GetVertex() + pkNormalElement->m_uiOffset;

		for( unsigned int uiPoly = 0; uiPoly < uiNumPolygons; ++uiPoly, ++pkPoly, ++pkPolyNormal )
		{
			*((Vector3d*)( pucVertex + pkPoly->v[0] * uiVertexSize ) ) += *pkPolyNormal;
			*((Vector3d*)( pucVertex + pkPoly->v[1] * uiVertexSize ) ) += *pkPolyNormal;
			*((Vector3d*)( pucVertex + pkPoly->v[2] * uiVertexSize ) ) += *pkPolyNormal;
		}

		pkPolygonBuffer->Unlock();

		pucVertex = (unsigned char*)pkVertexBuffer->GetVertex() + pkNormalElement->m_uiOffset;
		
		for( uiVert = 0; uiVert < pkVertexBuffer->GetNumElements(); ++uiVert, pucVertex += uiVertexSize )
			((Vector3d*)pucVertex)->Normalize();
	}

	//FIXME: Generate bounding box

	pkVertexBuffer->Unlock();

#ifdef NEOENGINE_DEBUGPATCHTESSELATOR
	neolog << LogLevel( DEBUG ) << "[PatchSurfaceTesselator::Tesselate] tesselation done" << endl;
#endif
}


void PatchSurfaceTesselator::InterpolateVertices( PatchSurfaceData *pkPatchData, VertexBufferPtr pkVertexBuffer )
{
	int iULoop = pkPatchData->m_iControlWidth - 1;
	int iVLoop = pkPatchData->m_iControlHeight - 1;

	//Loop and evaluate vertices
	pkPatchData->m_pkControlPoints->Lock( Buffer::READ );

	const VertexDeclaration *pkVertexFormat = pkVertexBuffer->GetVertexFormat();
	unsigned int               uiVertexSize   = pkVertexBuffer->GetVertexSize();

	unsigned char             *pucSrcVertex   = (unsigned char*)pkPatchData->m_pkControlPoints->GetVertex();
	unsigned char             *pucVertex      = (unsigned char*)pkVertexBuffer->GetVertex();

	float fOOWidth  = 1.0f / float( pkPatchData->m_iTessWidth  - 1 );
	float fOOHeight = 1.0f / float( pkPatchData->m_iTessHeight - 1 );

	float afTmp[32];

	int iTmpUsed = 0;

	for( unsigned int uiElement = 0; uiElement < pkVertexFormat->m_uiNumElements; ++uiElement )
		iTmpUsed += ( pkVertexFormat->m_pkElements[ uiElement ].m_uiType < VertexElement::COLOR32 ) ? ( pkVertexFormat->m_pkElements[ uiElement ].m_uiType + 1 ) : 4;

	int iSizeTmp = iTmpUsed * sizeof( float );

	//Reset everything to zero
	memset( pucVertex, 0, uiVertexSize * pkPatchData->m_iTessWidth * pkPatchData->m_iTessHeight );

	for( int iY = 0; iY < pkPatchData->m_iTessHeight; ++iY )
	{
		float fV = float( iY ) * fOOHeight;

		for( int iX = 0; iX < pkPatchData->m_iTessWidth; ++iX, pucVertex += uiVertexSize )
		{
			float fU = float( iX ) * fOOWidth;

			pucSrcVertex = (unsigned char*)pkPatchData->m_pkControlPoints->GetVertex();

			for( int iVCP = 0; iVCP <= iVLoop; ++iVCP )
			{
				float fBPv = float( Comb( iVLoop, iVCP ) ) * powf( fV, (float)iVCP ) * powf( 1.0f - fV, float( iVLoop - iVCP ) );

				memset( afTmp, 0, iSizeTmp );

				float *pfTmp;
				
				for( int iUCP = 0; iUCP <= iULoop; ++iUCP )
				{
					float fBPu =  float( Comb( iULoop, iUCP ) ) * powf( fU, (float)iUCP ) * powf( 1.0f - fU, float( iULoop - iUCP ) );

					pfTmp = afTmp;

					for( unsigned int uiElement = 0; uiElement < pkVertexFormat->m_uiNumElements; ++uiElement )
					{
						if( pkVertexFormat->m_pkElements[ uiElement ].m_uiType < VertexElement::COLOR32 )
						{
							for( unsigned int uiFloat = 0; uiFloat < ( pkVertexFormat->m_pkElements[ uiElement ].m_uiType + 1 ); ++uiFloat, pucSrcVertex += 4 )
								*pfTmp++ += *((float*)pucSrcVertex) * fBPu;
						}
						else // COLOR32
						{
							unsigned int uiColor = *(unsigned int*)pucSrcVertex; pucSrcVertex += 4;
							*pfTmp++ += float(   uiColor         & 0xFF ) * fBPu;
							*pfTmp++ += float( ( uiColor >> 8  ) & 0xFF ) * fBPu;
							*pfTmp++ += float( ( uiColor >> 16 ) & 0xFF ) * fBPu;
							*pfTmp++ += float( ( uiColor >> 24 ) & 0xFF ) * fBPu;
						}
					}
				}

				pfTmp = afTmp;

				unsigned char *pucDst = pucVertex;

				for( unsigned int uiElement = 0; uiElement < pkVertexFormat->m_uiNumElements; ++uiElement )
				{
					if( pkVertexFormat->m_pkElements[ uiElement ].m_uiType < VertexElement::COLOR32 )
					{
						for( unsigned int uiFloat = 0; uiFloat < ( pkVertexFormat->m_pkElements[ uiElement ].m_uiType + 1 ); ++uiFloat, pucDst += 4 )
							*(float*)pucDst += *pfTmp++ * fBPv;
					}
					else // COLOR32
					{
						unsigned char aucDst[4] = { *pucDst, *( pucDst + 1 ), *( pucDst + 2 ), *( pucDst + 3 ) };

						aucDst[0] += static_cast<unsigned char>( *( pfTmp     ) * fBPv );
						aucDst[1] += static_cast<unsigned char>( *( pfTmp + 1 ) * fBPv );
						aucDst[2] += static_cast<unsigned char>( *( pfTmp + 2 ) * fBPv );
						aucDst[3] += static_cast<unsigned char>( *( pfTmp + 3 ) * fBPv );

						*(unsigned int*)pucDst = aucDst[0] | ( aucDst[1] << 8 ) | ( aucDst[2] << 16 ) | ( aucDst[3] << 24 );

						pucDst += 4;
						pfTmp  += 4;
					}
				}
			}
		}
	}

	pkPatchData->m_pkControlPoints->Unlock();
}





QuadraticBezierPatchTesselator::QuadraticBezierPatchTesselator()
{
}


QuadraticBezierPatchTesselator::~QuadraticBezierPatchTesselator()
{
}


void QuadraticBezierPatchTesselator::InterpolateVertices( PatchSurfaceData *pkPatchData, VertexBufferPtr pkVertexBuffer )
{
	// ### FIXME: Fill in corner vertices (just copy from control points)
	//            for x == 0, x == ( tesswidth - 1 ), y == 0, y == ( tessheight - 1 )

	// ### FIXME: Simple evaluation of border vertices

#if 0

	//Loop and evaluate vertices
	pkPatchData->m_pkControlPoints->Lock( Buffer::READ );

	const VertexDeclaration *pkVertexFormat = pkVertexBuffer->GetVertexFormat();
	unsigned int               uiVertexSize   = pkVertexBuffer->GetVertexSize();

	unsigned char             *pucVertex      = (unsigned char*)pkVertexBuffer->GetVertex();

	float fOOWidth  = 1.0f / float( pkPatchData->m_iTessWidth  - 1 );
	float fOOHeight = 1.0f / float( pkPatchData->m_iTessHeight - 1 );

	//Reset everything to zero
	memset( pucVertex, 0, uiVertexSize * pkPatchData->m_iTessWidth * pkPatchData->m_iTessHeight );

	for( int iY = 0; iY < pkPatchData->m_iTessHeight; ++iY )
	{
		float fV = float( iY ) * fOOHeight;

		float fFactorV0 = ( 1.0f - fV ) * ( 1.0f - fV );
		float fFactorV1 = 2.0f * fV * ( 1.0f - fV );
		float fFactorV2 = fV * fV;

		for( int iX = 0; iX < pkPatchData->m_iTessWidth; ++iX, pucVertex += uiVertexSize )
		{
			float fU = float( iX ) * fOOWidth;

			float fFactorU0 = ( 1.0f - fU ) * ( 1.0f - fU );

			unsigned char *pucSrcVertex0 = (unsigned char*)pkPatchData->m_pkControlPoints->GetVertex();
			unsigned char *pucSrcVertex1 = pucSrcVertex0 + uiVertexSize *pkPatchData->m_iControlWidth;
			unsigned char *pucSrcVertex2 = pucSrcVertex1 + uiVertexSize *pkPatchData->m_iControlWidth;
			unsigned int   uiElement;
			unsigned char *pucDst        = pucVertex;

			for( uiElement = 0; uiElement < pkVertexFormat->m_uiNumElements; ++uiElement )
			{
				if( pkVertexFormat->m_pkElements[ uiElement ].m_uiType < VertexElement::COLOR32 )
				{
					for( unsigned int uiFloat = 0; uiFloat < ( pkVertexFormat->m_pkElements[ uiElement ].m_uiType + 1 ); ++uiFloat, pucDst += 4 )
						*(float*)pucDst += fFactorU * ( ( *((float*)pucSrcVertex0) * fFactorV0 ) + ( *((float*)pucSrcVertex1) * fFactorV1 ) + ( *((float*)pucSrcVertex2) * fFactorV2 ) );
				}
				else // COLOR32
				{
					unsigned int uiColor0 = *(unsigned int*)pucSrcVertex0;
					unsigned int uiColor1 = *(unsigned int*)pucSrcVertex1;
					unsigned int uiColor2 = *(unsigned int*)pucSrcVertex2;

					*(unsigned int*)pucDst += static_cast<unsigned char>( fFactorU * ( ( float(   uiColor0         & 0xFF ) * fFactorV0 ) + ( float(   uiColor1         & 0xFF ) * fFactorV1 ) + ( float(   uiColor2         & 0xFF ) * fFactorV2 ) ) )       |
					                          static_cast<unsigned char>( fFactorU * ( ( float( ( uiColor0 >> 8  ) & 0xFF ) * fFactorV0 ) + ( float( ( uiColor1 >> 8  ) & 0xFF ) * fFactorV1 ) + ( float( ( uiColor2 >> 8  ) & 0xFF ) * fFactorV2 ) ) ) << 8  |
					                          static_cast<unsigned char>( fFactorU * ( ( float( ( uiColor0 >> 16 ) & 0xFF ) * fFactorV0 ) + ( float( ( uiColor1 >> 16 ) & 0xFF ) * fFactorV1 ) + ( float( ( uiColor2 >> 16 ) & 0xFF ) * fFactorV2 ) ) ) << 16 |
					                          static_cast<unsigned char>( fFactorU * ( ( float( ( uiColor0 >> 24 ) & 0xFF ) * fFactorV0 ) + ( float( ( uiColor1 >> 24 ) & 0xFF ) * fFactorV1 ) + ( float( ( uiColor2 >> 24 ) & 0xFF ) * fFactorV2 ) ) ) << 24;

					pucDst += 4;
				}
			}

			pucSrcVertex0 += uiVertexSize;
			pucSrcVertex1 += uiVertexSize;
			pucSrcVertex2 += uiVertexSize;

			pucDst         = pucVertex;

			fFactorU = 2.0f * fU * ( 1.0f - fU );

			for( uiElement = 0; uiElement < pkVertexFormat->m_uiNumElements; ++uiElement )
			{
				if( pkVertexFormat->m_pkElements[ uiElement ].m_uiType < VertexElement::COLOR32 )
				{
					for( unsigned int uiFloat = 0; uiFloat < ( pkVertexFormat->m_pkElements[ uiElement ].m_uiType + 1 ); ++uiFloat, pucDst += 4 )
						*(float*)pucDst += fFactorU * ( ( *((float*)pucSrcVertex0) * fFactorV0 ) + ( *((float*)pucSrcVertex1 ) * fFactorV1 ) + ( *((float*)pucSrcVertex2) * fFactorV2 ) );
				}
				else // COLOR32
				{
					unsigned int uiColor0 = *(unsigned int*)pucSrcVertex0;
					unsigned int uiColor1 = *(unsigned int*)pucSrcVertex1;
					unsigned int uiColor2 = *(unsigned int*)pucSrcVertex2;

					*(unsigned int*)pucDst += static_cast<unsigned char>( fFactorU * ( ( float(   uiColor0         & 0xFF ) * fFactorV0 ) + ( float(   uiColor1         & 0xFF ) * fFactorV1 ) + ( float(   uiColor2         & 0xFF ) * fFactorV2 ) ) )       |
					                          static_cast<unsigned char>( fFactorU * ( ( float( ( uiColor0 >> 8  ) & 0xFF ) * fFactorV0 ) + ( float( ( uiColor1 >> 8  ) & 0xFF ) * fFactorV1 ) + ( float( ( uiColor2 >> 8  ) & 0xFF ) * fFactorV2 ) ) ) << 8  |
					                          static_cast<unsigned char>( fFactorU * ( ( float( ( uiColor0 >> 16 ) & 0xFF ) * fFactorV0 ) + ( float( ( uiColor1 >> 16 ) & 0xFF ) * fFactorV1 ) + ( float( ( uiColor2 >> 16 ) & 0xFF ) * fFactorV2 ) ) ) << 16 |
					                          static_cast<unsigned char>( fFactorU * ( ( float( ( uiColor0 >> 24 ) & 0xFF ) * fFactorV0 ) + ( float( ( uiColor1 >> 24 ) & 0xFF ) * fFactorV1 ) + ( float( ( uiColor2 >> 24 ) & 0xFF ) * fFactorV2 ) ) ) << 24;

					pucDst += 4;
				}
			}

			pucSrcVertex0 += uiVertexSize;
			pucSrcVertex1 += uiVertexSize;
			pucSrcVertex2 += uiVertexSize;

			pucDst         = pucVertex;

			fFactorU = fU * fU;

			for( uiElement = 0; uiElement < pkVertexFormat->m_uiNumElements; ++uiElement )
			{
				if( pkVertexFormat->m_pkElements[ uiElement ].m_uiType < VertexElement::COLOR32 )
				{
					for( unsigned int uiFloat = 0; uiFloat < ( pkVertexFormat->m_pkElements[ uiElement ].m_uiType + 1 ); ++uiFloat, pucDst += 4 )
						*(float*)pucDst += fFactorU * ( ( *((float*)pucSrcVertex0) * fFactorV0 ) + ( *((float*)pucSrcVertex1 ) * fFactorV1 ) + ( *((float*)pucSrcVertex2) * fFactorV2 ) );
				}
				else // COLOR32
				{
					unsigned int uiColor0 = *(unsigned int*)pucSrcVertex0;
					unsigned int uiColor1 = *(unsigned int*)pucSrcVertex1;
					unsigned int uiColor2 = *(unsigned int*)pucSrcVertex2;

					*(unsigned int*)pucDst += static_cast<unsigned char>( fFactorU * ( ( float(   uiColor0         & 0xFF ) * fFactorV0 ) + ( float(   uiColor1         & 0xFF ) * fFactorV1 ) + ( float(   uiColor2         & 0xFF ) * fFactorV2 ) ) )       |
					                          static_cast<unsigned char>( fFactorU * ( ( float( ( uiColor0 >> 8  ) & 0xFF ) * fFactorV0 ) + ( float( ( uiColor1 >> 8  ) & 0xFF ) * fFactorV1 ) + ( float( ( uiColor2 >> 8  ) & 0xFF ) * fFactorV2 ) ) ) << 8  |
					                          static_cast<unsigned char>( fFactorU * ( ( float( ( uiColor0 >> 16 ) & 0xFF ) * fFactorV0 ) + ( float( ( uiColor1 >> 16 ) & 0xFF ) * fFactorV1 ) + ( float( ( uiColor2 >> 16 ) & 0xFF ) * fFactorV2 ) ) ) << 16 |
					                          static_cast<unsigned char>( fFactorU * ( ( float( ( uiColor0 >> 24 ) & 0xFF ) * fFactorV0 ) + ( float( ( uiColor1 >> 24 ) & 0xFF ) * fFactorV1 ) + ( float( ( uiColor2 >> 24 ) & 0xFF ) * fFactorV2 ) ) ) << 24;

					pucDst += 4;
				}
			}
		}
	}

	pkPatchData->m_pkControlPoints->Unlock();

#endif

}



}; // namespace NeoEngine
