/***************************************************************************
                     perlin.cpp  -  Perlin noise generator
                             -------------------
    begin                : Mon Oct 14 2002
    copyright            : (C) 2002 by Rob Wanders
    email                : forteq@softhome.net
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, perlin.cpp

 The Initial Developer of the Original Code is Rob Wanders.
 Portions created by Rob Wanders are Copyright (C) 2002
 Rob Wanders. All Rights Reserved.

 ***************************************************************************/

#include "perlin.h"
#include "nemath.h"

#include <math.h>
#include <stdlib.h>


namespace NeoEngine
{


PerlinNoise::PerlinNoise()
{
	m_uiPrimeOne   = 15731;
	m_uiPrimeTwo   = 789221;
	m_uiPrimeThree = 1376312589;
	m_afLookUp     = 0;
	m_uiMaxX       = 0;
	m_uiMaxY       = 0;
	m_uiMaxZ       = 0;
}


PerlinNoise::~PerlinNoise()
{
	delete [] m_afLookUp;
}


void PerlinNoise::NewPrimeNumbers()
{
	m_uiPrimeOne   = RandomPrime();
	m_uiPrimeTwo   = RandomPrime();
	m_uiPrimeThree = RandomPrime();
}


bool PerlinNoise::IsPrime( unsigned int iA )
{
	int iSA;
	int iJ;

	iSA = (int)sqrtf( (float)iA );

	for( iJ=2; iJ <= iSA; iJ++ )
	{
		if ( iA % iJ == 0 )
			return false;
	}

	return true;
}


unsigned int PerlinNoise::RandomPrime()
{
	unsigned int iPrime = (unsigned int)rand();

	while( !IsPrime( iPrime ) )
		iPrime++;

	return iPrime;
}


float PerlinNoise::Noise( int iX )
{
	if( m_afLookUp )
		return m_afLookUp[ iX ];

	iX = ( iX << 13 ) ^ iX;
	return ( 1.0f - float( double( ( iX * ( iX * iX * m_uiPrimeOne + m_uiPrimeTwo ) + m_uiPrimeThree ) & 0x7fffffff ) / 1073741824.0 ) );
}


float PerlinNoise::Noise( int iX,int iY )
{
	if( m_afLookUp )
		return m_afLookUp[ iX + iY * m_uiMaxX ];

	unsigned int iN = iX + iY * 57;
	iN              = ( iN << 13 ) ^ iN;
	return( 1.0f - float( double( ( iN * ( iN * iN * m_uiPrimeOne + m_uiPrimeTwo ) + m_uiPrimeThree ) & 0x7fffffff ) / 1073741824.0 ) );
}


float PerlinNoise::Noise( int iX, int iY, int iZ )
{
	if( m_afLookUp )
		return m_afLookUp[ iX + ( iZ * m_uiMaxY + iY ) * m_uiMaxX ];

	int iL = iX + iY * 57;
	int iM = iY + iZ * 57;
	unsigned int iN = iL + iM * 57;
	iN              = ( iN << 13 ) ^ iN;
	return( 1.0f - float( double( ( iN * ( iN * iN * m_uiPrimeOne + m_uiPrimeTwo ) + m_uiPrimeThree ) & 0x7fffffff ) / 1073741824.0 ) );
}


float PerlinNoise::Interpolate( float fA, float fB, float fX )
{
	float fFt = fX * PI;
	float fF  = ( 1.0f - cosf( fFt ) ) * 0.5f;

	return( fA * ( 1.0f - fF ) + fB * fF );
}


float PerlinNoise::SmoothNoise(int iX)
{
	return( Noise( iX ) / 2.0f + Noise( iX - 1 ) / 4.0f + Noise( iX + 1 ) / 4.0f );
}


float PerlinNoise::SmoothNoise( int iX,int iY )
{
	float fCorners = ( Noise( iX - 1, iY - 1 ) + Noise( iX + 1, iY - 1 ) + Noise( iX - 1, iY + 1 ) + Noise( iX + 1, iY + 1 ) ) / 16.0f;
	float fSides   = ( Noise( iX - 1, iY     ) + Noise( iX + 1, iY     ) + Noise( iX    , iY - 1 ) + Noise( iX    , iY + 1 ) ) / 8.0f;
	float fCenter  =   Noise( iX, iY ) / 4.0f;
    return( fCorners + fSides + fCenter );
}


float PerlinNoise::SmoothNoise( int iX, int iY, int iZ )
{
	float fCorners, fSides, fCenter;
	float fZm1, fZ, fZp1;

	// average of neighbours in z-1
	fCorners = ( Noise( iX - 1, iY - 1, iZ - 1 ) + Noise( iX + 1, iY - 1, iZ - 1 ) + Noise( iX - 1, iY + 1, iZ - 1 ) + Noise( iX + 1, iY + 1, iZ - 1 ) ) / 16.0f;
	fSides   = ( Noise( iX - 1, iY    , iZ - 1 ) + Noise( iX + 1, iY    , iZ - 1 ) + Noise( iX    , iY - 1, iZ - 1 ) + Noise( iX    , iY + 1, iZ - 1 ) ) / 8.0f;
	fCenter  =   Noise( iX, iY, iZ - 1 ) / 4.0f;
	fZm1     = fCorners + fSides + fCenter;

	// average of neighbours in z
	fCorners = ( Noise( iX - 1, iY - 1, iZ ) + Noise( iX + 1, iY - 1, iZ ) + Noise( iX - 1, iY + 1, iZ ) + Noise( iX + 1, iY + 1, iZ ) ) / 16.0f;
	fSides   = ( Noise( iX - 1, iY    , iZ ) + Noise( iX + 1, iY    , iZ ) + Noise( iX    , iY - 1, iZ ) + Noise( iX    , iY + 1, iZ ) ) / 8.0f;
	fCenter  =   Noise( iX, iY, iZ ) / 4.0f;
	fZ       = fCorners + fSides + fCenter;

	// average of neighbours in z+1
	fCorners = ( Noise( iX - 1, iY - 1, iZ + 1 ) + Noise( iX + 1, iY - 1, iZ + 1 ) + Noise( iX - 1, iY + 1, iZ + 1 ) + Noise( iX + 1, iY + 1, iZ + 1 ) ) / 16.0f;
	fSides   = ( Noise( iX - 1, iY    , iZ + 1 ) + Noise( iX + 1, iY    , iZ + 1 ) + Noise( iX    , iY - 1, iZ + 1 ) + Noise( iX    , iY + 1, iZ + 1 ) ) / 8.0f;
	fCenter  =   Noise( iX, iY, iZ + 1 ) / 4.0f;
	fZp1     = fCorners + fSides + fCenter;

	return( ( fZm1 / 4.0f ) + ( fZ / 2.0f ) + ( fZp1 / 4.0f ) );
}


float PerlinNoise::InterpolateNoise( float fX )
{
	int iX = int(fX);
	float fFractX = fX - float( iX );
	
	float fV1 = SmoothNoise( iX );
	float fV2 = SmoothNoise( iX + 1 );
	
	return Interpolate( fV1, fV2, fFractX );
}


float PerlinNoise::InterpolateNoise( float fX, float fY )
{
	int iX = int( fX );
	float fFractX = fX - float( iX );

	int iY = int( fY );
	float fFractY = fY - float( iY );

	float fV1 = SmoothNoise( iX    , iY );
	float fV2 = SmoothNoise( iX + 1, iY );
	float fV3 = SmoothNoise( iX    , iY + 1 );
	float fV4 = SmoothNoise( iX + 1, iY + 1 );

	float fI1 = Interpolate( fV1, fV2, fFractX );
	float fI2 = Interpolate( fV3, fV4, fFractX );

	return Interpolate( fI1, fI2, fFractY );
}


float PerlinNoise::InterpolateNoise( float fX, float fY, float fZ )
{
	int iX = int( fX );
	float fFractX = fX - float( iX );

	int iY = int( fY );
	float fFractY = fY - float( iY );

	int iZ = int( fZ );
	float fFractZ = fZ - float( iZ );

	float fV1 = SmoothNoise( iX    , iY    , iZ );
	float fV2 = SmoothNoise( iX + 1, iY    , iZ );
	float fV3 = SmoothNoise( iX    , iY + 1, iZ );
	float fV4 = SmoothNoise( iX + 1, iY + 1, iZ );
	float fV5 = SmoothNoise( iX    , iY    , iZ + 1 );
	float fV6 = SmoothNoise( iX + 1, iY    , iZ + 1 );
	float fV7 = SmoothNoise( iX    , iY + 1, iZ + 1 );
	float fV8 = SmoothNoise( iX + 1, iY + 1, iZ + 1 );

	float fI1 = Interpolate( fV1, fV2, fFractX );
	float fI2 = Interpolate( fV3, fV4, fFractX );
	float fI3 = Interpolate( fV5, fV6, fFractX );
	float fI4 = Interpolate( fV7, fV8, fFractX );

	float fI5 = Interpolate( fI1, fI2, fFractY );
	float fI6 = Interpolate( fI3, fI4, fFractY );

	return Interpolate( fI5, fI6, fFractZ );
}


float PerlinNoise::PerlinNoise1d( float fX, float fAmplitude, float fXfrequency)
{
	return InterpolateNoise( fX * fXfrequency ) * fAmplitude;
}


float PerlinNoise::PerlinNoise2d( float fX, float fY, float fAmplitude, float fXfrequency, float fYfrequency )
{
	return InterpolateNoise( fX * fXfrequency, fY * fYfrequency ) * fAmplitude;
}


float PerlinNoise::PerlinNoise3d( float fX, float fY, float fZ, float fAmplitude, float fXfrequency, float fYfrequency, float fZfrequency )
{
	return InterpolateNoise( fX * fXfrequency, fY * fYfrequency, fZ * fZfrequency ) * fAmplitude;
}


void PerlinNoise::GenerateLookup( unsigned int uiMaxX )
{
	m_uiMaxX   = uiMaxX;
	m_afLookUp = new float[ uiMaxX ];

	for( unsigned int iX = 0; iX < uiMaxX; ++iX )
		m_afLookUp[iX] = Noise( iX );
}


void PerlinNoise::GenerateLookup( unsigned int uiMaxX, unsigned int uiMaxY )
{
	m_uiMaxX   = uiMaxX;
	m_uiMaxY   = uiMaxY;
	m_afLookUp = new float[ uiMaxX * uiMaxY ];

	int iYOff  = 0;

	for( unsigned int iY = 0; iY < uiMaxY; ++iY )
	{
		for( unsigned int iX = 0; iX < uiMaxX; ++iX )
			m_afLookUp[ iX + iYOff ] = Noise( iX,iY );

		iYOff += uiMaxX;
	}
}


void PerlinNoise::GenerateLookup( unsigned int uiMaxX, unsigned int uiMaxY, unsigned int uiMaxZ )
{
	m_uiMaxX   = uiMaxX;
	m_uiMaxY   = uiMaxY;
	m_uiMaxZ   = uiMaxZ;
	m_afLookUp = new float[ uiMaxX * uiMaxY * uiMaxZ ];

	int iZOff = 0;
	int iYOff = 0;
	int iZStep = uiMaxY * uiMaxX;

	for( unsigned int iZ = 0; iZ < uiMaxZ; ++iZ )
	{
		for( unsigned int iY = 0; iY < uiMaxY; ++iY )
		{
			for( unsigned int iX = 0; iX < uiMaxX; ++iX )
				m_afLookUp[ iX + iYOff + iZOff ] = Noise( iX,iY,iZ );

			iYOff += uiMaxX;
		}

		iYOff  = 0;
		iZOff += iZStep;
	}
}


}; // namespace NeoEngine
