/***************************************************************************
                   physics.cpp  -  Physics simulation classes
                             -------------------
    begin                : Tue Nov 6 2001
    copyright            : (C) 2001 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, physics.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2001
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "physics.h"
#include "joint.h"
#include "radix.h"
#include "logstream.h"

#ifndef DEBUG_PHYSICS_MANAGER
//#  define DEBUG_PHYSICS_MANAGER
#endif

#ifdef DEBUG_PHYSICS_MANAGER
#  include "render.h"
#  include "renderprimitive.h"
#  include "vertex.h"
#  include "camera.h"
#endif

#include <algorithm>

using namespace std;


namespace NeoEngine
{


namespace __internal
{


inline Vector3d operator * ( const Vector3d &rkVector, const Matrix &rkMatrix )
{
	return Vector3d(

		rkVector.x * rkMatrix[0][0] + rkVector.y * rkMatrix[1][0] + rkVector.z * rkMatrix[2][0] + rkMatrix[3][0],
		rkVector.x * rkMatrix[0][1] + rkVector.y * rkMatrix[1][1] + rkVector.z * rkMatrix[2][1] + rkMatrix[3][1],
		rkVector.x * rkMatrix[0][2] + rkVector.y * rkMatrix[1][2] + rkVector.z * rkMatrix[2][2] + rkMatrix[3][2]

	);
}


inline Matrix Mat3x3Inv( const Matrix &rkMatrix )
{
	Matrix kRet;

	kRet[0][0] = rkMatrix.m_aafMatrix[1][1] * rkMatrix.m_aafMatrix[2][2] - rkMatrix.m_aafMatrix[1][2] * rkMatrix.m_aafMatrix[2][1];
	kRet[0][1] = rkMatrix.m_aafMatrix[0][2] * rkMatrix.m_aafMatrix[2][1] - rkMatrix.m_aafMatrix[0][1] * rkMatrix.m_aafMatrix[2][2];
	kRet[0][2] = rkMatrix.m_aafMatrix[0][1] * rkMatrix.m_aafMatrix[1][2] - rkMatrix.m_aafMatrix[0][2] * rkMatrix.m_aafMatrix[1][1];
	kRet[1][0] = rkMatrix.m_aafMatrix[1][2] * rkMatrix.m_aafMatrix[2][0] - rkMatrix.m_aafMatrix[1][0] * rkMatrix.m_aafMatrix[2][2];
	kRet[1][1] = rkMatrix.m_aafMatrix[0][0] * rkMatrix.m_aafMatrix[2][2] - rkMatrix.m_aafMatrix[0][2] * rkMatrix.m_aafMatrix[2][0];
	kRet[1][2] = rkMatrix.m_aafMatrix[0][2] * rkMatrix.m_aafMatrix[1][0] - rkMatrix.m_aafMatrix[0][0] * rkMatrix.m_aafMatrix[1][2];
	kRet[2][0] = rkMatrix.m_aafMatrix[1][0] * rkMatrix.m_aafMatrix[2][1] - rkMatrix.m_aafMatrix[1][1] * rkMatrix.m_aafMatrix[2][0];
	kRet[2][1] = rkMatrix.m_aafMatrix[0][1] * rkMatrix.m_aafMatrix[2][0] - rkMatrix.m_aafMatrix[0][0] * rkMatrix.m_aafMatrix[2][1];
	kRet[2][2] = rkMatrix.m_aafMatrix[0][0] * rkMatrix.m_aafMatrix[1][1] - rkMatrix.m_aafMatrix[0][1] * rkMatrix.m_aafMatrix[1][0];

	float fDet = rkMatrix.m_aafMatrix[0][0] * kRet[0][0] + rkMatrix.m_aafMatrix[0][1] * kRet[1][0]+ rkMatrix.m_aafMatrix[0][2] * kRet[2][0];

	float fInvDet = 1.0f / fDet;

	for( int iRow = 0; iRow < 3; ++iRow )
		for( int iCol = 0; iCol < 3; ++iCol )
			kRet[iRow][iCol] *= fInvDet;

	return kRet;
}


void SetContactNodePartition( ContactNode **ppkNodes, unsigned int *puiPartitions, unsigned int uiCurNode, unsigned int uiPartition, unsigned int uiPass )
{
	ContactNode *pkNode = ppkNodes[ uiCurNode ];

	puiPartitions[ uiCurNode ] = uiPartition++;
	pkNode->m_bVisited         = true;
	pkNode->m_uiPartitionPass  = uiPass;

	vector< ContactNode* >::iterator ppkChild    = pkNode->m_vpkChildren.begin();
	vector< ContactNode* >::iterator ppkChildEnd = pkNode->m_vpkChildren.end();

	ContactNode *pkChild;

	for( ; ppkChild != ppkChildEnd; ++ppkChild )
	{
		pkChild = *ppkChild;

		if( pkChild->m_bVisited )
		{
			if( ( pkChild->m_uiPartitionPass == uiPass ) || ( puiPartitions[ pkChild->m_uiID ] >= uiPartition ) )
				continue;
			else
				pkChild->ResetVisited();
		}

		SetContactNodePartition( ppkNodes, puiPartitions, pkChild->m_uiID, uiPartition, uiPass );
	}
}


//Scratchpad memory
unsigned int  g_uiPadSize        = 0;
unsigned int *g_puiState         = 0;
Matrix       *g_pkCollMatrix     = 0;
Matrix       *g_pkInvCollMatrix  = 0;

class __physDataManager { public: ~__physDataManager() { delete [] g_pkCollMatrix, g_pkCollMatrix = 0; delete [] g_pkInvCollMatrix, g_pkInvCollMatrix = 0; delete [] g_puiState, g_puiState = 0; g_uiPadSize = 0; } } __physDataManagerObj;

};


using namespace __internal;


bool Joint::Solve( ContactNode::CONTACTMODE eMode )
{
	Vector3d kRelPoint = m_apkBody[0]->GetWorldRotation() * m_akOffset[0];
	Vector3d kRelPointOther;

	Vector3d kPointVel = m_apkBody[0]->GetPointVelocity( kRelPoint + m_apkBody[0]->GetWorldTranslation(), false );

	if( m_apkBody[1] )
	{
		kRelPointOther = m_apkBody[1]->GetWorldRotation() * m_akOffset[1];

		kPointVel -= m_apkBody[1]->GetPointVelocity( kRelPointOther + m_apkBody[1]->GetWorldTranslation(), false );
	}

	if( kPointVel.Len2() > SQREPSILON )
	{
		Matrix kCollisionMatrix = ( Matrix::IDENTITY * m_apkBody[0]->GetInvMass() ) + ( ( Matrix( kRelPoint ).Transpose() * m_apkBody[0]->GetInvInertiaTensor() ) * Matrix( kRelPoint ) );

		if( m_apkBody[1] )
			kCollisionMatrix += ( Matrix::IDENTITY * m_apkBody[1]->GetInvMass() ) + ( ( Matrix( kRelPointOther ).Transpose() * m_apkBody[1]->GetInvInertiaTensor() ) * Matrix( kRelPointOther ) );

		Matrix kInvCollisionMatrix = Mat3x3Inv( kCollisionMatrix );

		Vector3d kImpulse = kInvCollisionMatrix * -kPointVel;

		m_apkBody[0]->ApplyImpulse( kImpulse, kRelPoint + m_apkBody[0]->GetWorldTranslation(), false );

		if( m_apkBody[1] )
			m_apkBody[1]->ApplyImpulse( -kImpulse, kRelPointOther + m_apkBody[1]->GetWorldTranslation(), false );

		return true;
	}

	return false;
}



ContactNode::ContactNode( RigidBody *pkBody ) :
	m_pkBody( pkBody ),
	m_bVisited( false ),
	m_bFree( false ),
	m_uiPartitionPass( 0 )
{
}


ContactNode::~ContactNode()
{
	vector< ContactSet* >::iterator ppkSet    = m_vpkContactSets.begin();
	vector< ContactSet* >::iterator ppkSetEnd = m_vpkContactSets.end();

	for( ; ppkSet != ppkSetEnd; ++ppkSet )
		delete( *ppkSet );
}


bool ContactNode::Solve( CONTACTMODE eMode, unsigned int uiPass, unsigned int uiNumPasses )
{
	if( !m_vpkContactSets.size() )
		return false;

	bool bHaveColliding  = false;
	bool bHaveThisColl   = false;
	bool bLoop           = false;
	unsigned int uiCount = 0;

	static unsigned int MAXSHOCKPASSES = 25;

	if( g_uiPadSize < m_uiTotalPoints )
	{
		g_uiPadSize = m_uiTotalPoints * 2;

		delete [] g_puiState;
		delete [] g_pkCollMatrix;
		delete [] g_pkInvCollMatrix;

		g_puiState        = new unsigned int[ g_uiPadSize ];
		g_pkCollMatrix    = new Matrix[ g_uiPadSize ];
		g_pkInvCollMatrix = new Matrix[ g_uiPadSize ];
	}

	memset( g_puiState, 0, sizeof( unsigned int ) * m_uiTotalPoints );

	do
	{
		bLoop         = false;
		bHaveThisColl = false;

		vector< ContactSet*  >::iterator ppkSet          = m_vpkContactSets.begin();
		vector< ContactSet*  >::iterator ppkSetEnd       = m_vpkContactSets.end();
		vector< ContactNode* >::iterator ppkNode         = m_vpkParents.begin();

		unsigned int *puiState        = g_puiState;
		Matrix       *pkCollMatrix    = g_pkCollMatrix;
		Matrix       *pkInvCollMatrix = g_pkInvCollMatrix;

		for( ; ppkSet != ppkSetEnd; ++ppkSet, ++ppkNode )
		{
			ContactNode  *pkParent  = *ppkNode;

			Contact *pkContact = (*ppkSet)->m_vpkContacts[0];

			Vector3d *pkPoint = pkContact->m_pkPoints;
			Vector3d  kNormal = pkContact->m_kNormal;

			for( unsigned int uiPoint = 0; uiPoint < pkContact->m_uiNumPoints; ++uiPoint, ++pkPoint, ++puiState, ++pkCollMatrix, ++pkInvCollMatrix )
			{
				if( *puiState & 0x1 )
					continue;

				Vector3d kRelPoint   = (*pkPoint) - m_pkBody->GetWorldTranslation();

				Vector3d kPointVel = m_pkBody->GetPointVelocity( *pkPoint, false );

				RigidBody *pkOtherBody = pkParent ? pkParent->m_pkBody : 0;

				if( pkOtherBody )
					kPointVel -= pkOtherBody->GetPointVelocity( *pkPoint, false );

				float fVelN = kPointVel * kNormal;

				if( fVelN < 0.0f )
				{
					bHaveThisColl        = true;

					if( !( *puiState & 0x2 ) )
					{
						*pkCollMatrix = ( Matrix::IDENTITY * m_pkBody->GetInvMass() ) + ( ( Matrix( kRelPoint ).Transpose() * m_pkBody->GetInvInertiaTensor() ) * Matrix( kRelPoint ) );

						if( pkOtherBody && ( eMode != SHOCKPROPAGATION ) )
						{
							Vector3d kRelPointOther = (*pkPoint) - pkOtherBody->GetWorldTranslation();

							*pkCollMatrix += ( Matrix::IDENTITY * pkOtherBody->GetInvMass() ) + ( ( Matrix( kRelPointOther ).Transpose() * pkOtherBody->GetInvInertiaTensor() ) * Matrix( kRelPointOther ) );
						}

						*pkInvCollMatrix = Mat3x3Inv( *pkCollMatrix );

						*puiState |= 0x2;
					}

					float fElasticity    = MIN< float >( m_pkBody->GetRestitution(), pkOtherBody ? pkOtherBody->GetRestitution() : 1.0f );
					float fFriction      = MAX< float >( m_pkBody->GetFriction(),    pkOtherBody ? pkOtherBody->GetFriction()    : 0.0f );

					if( eMode == CONTACT )
						fElasticity = MIN< float >( -1.0f + ( float( uiPass + 1 ) / float( uiNumPasses ) ), 0.0f );
					else if( eMode == SHOCKPROPAGATION )
						fElasticity = 0.0f;

					Vector3d kImpulse = (*pkInvCollMatrix) * ( ( kNormal * ( fElasticity * -fVelN ) ) - kPointVel );

					float fImpulseN = kImpulse * kNormal;

					// Check if impulse is within friction cone, |Jt| <= u*Jn
					if( ( kImpulse - ( kNormal * fImpulseN ) ).Len() > ( fFriction * fImpulseN ) )
					{
						// vrel,t = vrel - vrel,n * n
						// T = vrel,t / |vrel,t|
						Vector3d kTangent = ( kPointVel - ( kNormal * fVelN ) ).Normalize();

						// Jn = -( e + 1 )*vrel,n / ( nT * Kt * ( N - u*T ) )
						fImpulseN = ( fVelN * -( 1.0f + fElasticity ) ) / ( ( kNormal * (*pkCollMatrix) ) * ( kNormal - ( kTangent * fFriction ) ) );
					
						// J = Jn*N - u*Jn*T
						kImpulse = ( kNormal * fImpulseN ) - ( kTangent * ( fFriction * fImpulseN ) );
					}

					m_pkBody->ApplyImpulse( kImpulse, *pkPoint, false );

					if( pkOtherBody && ( eMode != SHOCKPROPAGATION ) )
						pkOtherBody->ApplyImpulse( -kImpulse, *pkPoint, false );

					if( eMode == COLLISION )
						break; //Only consider the deepest point each local iteration
				}
			}
		}

		bHaveColliding |= bHaveThisColl;

		if( ( eMode == SHOCKPROPAGATION ) && ( uiCount == MAXSHOCKPASSES ) )
			break;

		++uiCount;

		if( bHaveThisColl )
		{
			if( eMode == COLLISION )
			{
				unsigned int *puiState = g_puiState;
				ppkSet  = m_vpkContactSets.begin();
				ppkNode = m_vpkParents.begin();

				//Remove points that are non-intersecting and redo collision process
				for( ; ppkSet != ppkSetEnd; ++ppkSet, ++ppkNode )
				{
					ContactNode *pkParent  = *ppkNode;

					Contact *pkContact = (*ppkSet)->m_vpkContacts[0];

					Vector3d *pkPoint  = pkContact->m_pkPoints;

					for( unsigned int uiPoint = 0; uiPoint < pkContact->m_uiNumPoints; ++uiPoint, ++pkPoint, ++puiState )
					{
						if( *puiState & 0x1 )
							continue;

						Vector3d kPointVel = m_pkBody->GetPointVelocity( *pkPoint, false );

						RigidBody *pkOtherBody = pkParent ? pkParent->m_pkBody : 0;

						if( pkOtherBody )
							kPointVel -= pkOtherBody->GetPointVelocity( *pkPoint, false );

						if( kPointVel * pkContact->m_kNormal > -EPSILON )
							*puiState |= 0x1;
						else
							bLoop = true;
					}
				}
			}
			else if( eMode == SHOCKPROPAGATION )
				bLoop = true;
		}

	} while( bLoop );

	return bHaveColliding;
}


void ContactNode::ResetVisited()
{
	if( !m_bVisited )
		return;

	m_bVisited = false;
	m_uiPartitionPass = 0;

	vector< ContactNode* >::iterator ppkNode    = m_vpkChildren.begin();
	vector< ContactNode* >::iterator ppkNodeEnd = m_vpkChildren.end();

	for( ; ppkNode != ppkNodeEnd; ++ppkNode )
		(*ppkNode)->ResetVisited();
}


void ContactNode::AddContact( ContactNode *pkParent, ContactSet *pkSet )
{
	vector< ContactNode* >::iterator ppkNode    = m_vpkParents.begin();
	vector< ContactNode* >::iterator ppkNodeEnd = m_vpkParents.end();
	vector< ContactSet*  >::iterator ppkSet     = m_vpkContactSets.begin();

	for( ; ppkNode != ppkNodeEnd; ++ppkNode, ++ppkSet )
	{
		if( *ppkNode == pkParent )
		{
			assert( !"Contact already registered for parent" );
			delete pkSet;
			return;
		}
	}

	if( pkParent )
		pkParent->m_vpkChildren.push_back( this );
	
	m_vpkParents.push_back( pkParent );
	m_vpkContactSets.push_back( pkSet );

	vector< Contact* >::iterator ppkContact    = pkSet->m_vpkContacts.begin();
	vector< Contact* >::iterator ppkContactEnd = pkSet->m_vpkContacts.end();

	for( ; ppkContact != ppkContactEnd; ++ppkContact )
		m_uiTotalPoints += (*ppkContact)->m_uiNumPoints;
}


void ContactNode::Clear()
{
	m_bVisited      = false;
	m_bFree         = false;
	m_uiTotalPoints = 0;
	m_uiPartitionPass = 0;

	m_vpkParents.clear();
	m_vpkChildren.clear();

	vector< ContactSet* >::iterator ppkSet       = m_vpkContactSets.begin();
	vector< ContactSet* >::iterator ppkSetEnd    = m_vpkContactSets.end();

	for( ; ppkSet != ppkSetEnd; ++ppkSet )
		delete( *ppkSet );

	m_vpkContactSets.clear();
}



#ifdef DEBUG_PHYSICS_MANAGER

VertexBufferPtr *g_pkDebugLines[3];
VertexBufferPtr *g_pkDebugPoints[3];

#endif




PhysicsManager::PhysicsManager( Room *pkRoom, unsigned int uiCollisionPasses, unsigned int uiContactPasses ) :
	UpdateEntity(),
	m_uiCollisionPasses( uiCollisionPasses ),
	m_uiContactPasses( uiContactPasses ),
	m_pkRoom( pkRoom )
{
#ifdef DEBUG_PHYSICS_MANAGER
	g_pkDebugLines[0]  = new VertexBufferPtr[ uiCollisionPasses ];
	g_pkDebugPoints[0] = new VertexBufferPtr[ uiCollisionPasses ];
	g_pkDebugLines[1]  = new VertexBufferPtr[ uiContactPasses ];
	g_pkDebugPoints[1] = new VertexBufferPtr[ uiContactPasses ];
	g_pkDebugLines[2]  = new VertexBufferPtr[ 1 ];
	g_pkDebugPoints[2] = new VertexBufferPtr[ 1 ];
#endif

	if( g_uiPadSize < 16 )
	{
		g_uiPadSize = 16;

		delete [] g_puiState;
		delete [] g_pkCollMatrix;
		delete [] g_pkInvCollMatrix;

		g_puiState        = new unsigned int[ g_uiPadSize ];
		g_pkCollMatrix    = new Matrix[ g_uiPadSize ];
		g_pkInvCollMatrix = new Matrix[ g_uiPadSize ];
	}

	m_uiStoreSize      = 100;
	m_pkContactSet     = new ContactSet;
	m_pbIsActive       = new bool[ m_uiStoreSize ];
	m_ppkContacts      = new ContactNode*[ m_uiStoreSize ];
	m_ppkResting       = new ContactNode*[ m_uiStoreSize ];
	m_puiNodePartition = new unsigned int[ m_uiStoreSize ];
}


PhysicsManager::~PhysicsManager()
{
	vector< Joint* >::iterator ppkJoint    = m_vpkJoints.begin();
	vector< Joint* >::iterator ppkJointEnd = m_vpkJoints.end();

	for( ; ppkJoint != ppkJointEnd; ++ppkJoint )
		delete( *ppkJoint );

	vector< RigidBody* >::iterator ppkBody    = m_vpkBodies.begin();
	vector< RigidBody* >::iterator ppkBodyEnd = m_vpkBodies.end();

	for( ; ppkBody != ppkBodyEnd; ++ppkBody )
		delete *ppkBody;

	vector< ContactNode* >::iterator ppkBodyNode    = m_vpkBodyNodes.begin();
	vector< ContactNode* >::iterator ppkBodyNodeEnd = m_vpkBodyNodes.end();

	for( ; ppkBodyNode != ppkBodyNodeEnd; ++ppkBodyNode )
		delete *ppkBodyNode;

	delete m_pkContactSet;
	delete [] m_pbIsActive;
	delete [] m_ppkContacts;
	delete [] m_ppkResting;
	delete [] m_puiNodePartition;

#ifdef DEBUG_PHYSICS_MANAGER
	for( int i = 0; i < 3; ++i )
	{
		delete [] g_pkDebugLines[i];
		delete [] g_pkDebugPoints[i];
	}
#endif
}


void PhysicsManager::SetGravity( const Vector3d &rkGravity )
{
	m_kGravity = rkGravity;
}


void PhysicsManager::Update( float fDeltaTime )
{
	vector< ContactNode* >::iterator ppkBodyNode    = m_vpkBodyNodes.begin();
	vector< ContactNode* >::iterator ppkBodyNodeEnd = m_vpkBodyNodes.end();

	unsigned int  uiNumActive  = 0;
	unsigned int  uiNumResting = 0;
	unsigned int  uiCurNode    = 0;
	unsigned int  uiPass       = 0;
	bool         *pbActive     = m_pbIsActive;

	//Initial setup phase
	for( uiCurNode = 0; ppkBodyNode != ppkBodyNodeEnd; ++ppkBodyNode, ++pbActive, ++uiCurNode )
	{
		RigidBody *pkBody = (*ppkBodyNode)->m_pkBody;

		pkBody->m_uiID = uiCurNode;

		if( !( *pbActive = pkBody->IsActive() ) )
			continue;

		if( pkBody->m_bResting )
			++uiNumResting;

		pkBody->m_bPenetrating = false;

		pkBody->PushState();
		pkBody->Deactivate();

		++uiNumActive;
	}

	if( !uiNumActive )
		return;

	static RadixSort kCollisionSort( RadixSort::UINT32 );

	//Collision phase
	for( uiPass = 0; uiPass < m_uiCollisionPasses; ++uiPass )
	{
		ppkBodyNode  = m_vpkBodyNodes.begin();
		pbActive     = m_pbIsActive;
		uiNumActive  = 0;
		uiNumResting = 0;

		// Predict new position and keep velocities
		for( ; ppkBodyNode != ppkBodyNodeEnd; ++ppkBodyNode, ++pbActive )
		{
			ContactNode *pkBodyNode = *ppkBodyNode;
			RigidBody   *pkBody     = pkBodyNode->m_pkBody;

			if( !*pbActive )
				continue;

			if( pkBody->m_bResting )
			{
				//Check if still resting
				vector< RigidBody* >::iterator ppkRestee     = pkBody->m_vpkRestOn.begin();
				vector< RigidBody* >::iterator ppkResteeEnd  = pkBody->m_vpkRestOn.end();
				
				for( ; ppkRestee != ppkResteeEnd; ++ppkRestee )
					if( (*ppkRestee) && ( !(*ppkRestee)->m_bResting || !m_pbIsActive[ (*ppkRestee)->m_uiID ] ) )
					{
						pkBody->m_bResting = false;
						break;
					}
			}

			if( !pkBody->m_bResting )
			{
				if( uiPass )
				{
					Vector3d kLinearVelocity  = pkBody->m_kVelocity;
					Vector3d kAngularVelocity = pkBody->m_kAngularVelocity;

					pkBody->PopState();

					pkBody->m_kVelocity        = kLinearVelocity;
					pkBody->m_kAngularVelocity = kAngularVelocity;

					pkBody->PushState();
				}

				pkBody->m_kTranslation += ( ( ( ( pkBody->m_kForce * pkBody->m_fInvMass ) + m_kGravity ) * fDeltaTime ) + pkBody->m_kVelocity ) * fDeltaTime;
				pkBody->m_kRotation     = ( pkBody->m_kRotation + ( ( Quaternion( pkBody->m_kAngularVelocity.x, pkBody->m_kAngularVelocity.y, pkBody->m_kAngularVelocity.z, 0.0f ) * pkBody->m_kRotation ) * ( 0.5f * fDeltaTime ) ) ).Normalize();

				pkBody->m_kRotation.ToMatrix( &pkBody->m_kRotationMatrix );
				pkBody->m_kInvInertiaTensor = ( pkBody->m_kRotationMatrix * pkBody->m_kInvBodyInertiaTensor );

				Matrix kRotTranspose( pkBody->m_kRotationMatrix ); kRotTranspose.Transpose();
				pkBody->m_kInvInertiaTensor = ( pkBody->m_kInvInertiaTensor * kRotTranspose );

				pkBody->NotifyUpdate( false );

				m_ppkContacts[ uiNumActive++ ] = pkBodyNode;
			}
			else
				m_ppkResting[ uiNumResting++ ] = pkBodyNode;

			pkBodyNode->m_bVisited = false;
			pkBodyNode->Clear();
		}

		//Either we are in first pass or we must have active
		assert( !uiPass || uiNumActive );

		if( !uiNumActive )
		{
			ppkBodyNode  = m_vpkBodyNodes.begin();
			pbActive     = m_pbIsActive;

			for( ; ppkBodyNode != ppkBodyNodeEnd; ++ppkBodyNode, ++pbActive )
			{
				if( *pbActive )
				{
					(*ppkBodyNode)->m_pkBody->PopState( false );
					(*ppkBodyNode)->m_pkBody->Activate();
				}
			}

			return;
		}

		unsigned int uiAdded = 0;

		//Collision tests
		for( uiCurNode = 0 ; uiCurNode != uiNumActive; ++uiCurNode )
		{
			ContactNode *pkBodyNode = m_ppkContacts[ uiCurNode ];
			RigidBody   *pkBody     = pkBodyNode->m_pkBody;

			pkBodyNode->m_uiID      = uiCurNode;

			if( m_pkRoom && m_pkRoom->Intersection( pkBody->GetBoundingVolume(), m_pkContactSet ) )
			{
				pkBodyNode->AddContact( 0, m_pkContactSet );
				m_pkContactSet = new ContactSet;
			}

			for( unsigned int uiNextNode = uiCurNode + 1; uiNextNode < uiNumActive; ++uiNextNode )
			{
				ContactNode *pkOtherNode = m_ppkContacts[ uiNextNode ];
				RigidBody   *pkOtherBody = pkOtherNode->m_pkBody;

				//Check if joints connect bodies
				vector< Joint* >::iterator ppkJoint    = pkBody->m_vpkJoints.begin();
				vector< Joint* >::iterator ppkJointEnd = pkBody->m_vpkJoints.end();

				for( ; ppkJoint != ppkJointEnd; ++ppkJoint )
				{
					Joint *pkJoint = *ppkJoint;

					if( ( ( pkJoint->m_apkBody[0] == pkBody ) && ( pkJoint->m_apkBody[1] == pkOtherBody ) ) ||
					    ( ( pkJoint->m_apkBody[1] == pkBody ) && ( pkJoint->m_apkBody[0] == pkOtherBody ) ) )
						break;
				}

				if( ppkJoint < ppkJointEnd )
					continue;

				ContactNode *pkFirstNode  = pkBodyNode;
				ContactNode *pkSecondNode = pkOtherNode;

				if( ( ( pkSecondNode->m_pkBody->GetWorldTranslation() - pkFirstNode->m_pkBody->GetWorldTranslation() ) * m_kGravity ) < 0.0f )
				{
					pkFirstNode  = pkOtherNode;
					pkSecondNode = pkBodyNode;
				}

				if( pkFirstNode->m_pkBody->GetBoundingVolume()->Intersection( pkSecondNode->m_pkBody->GetBoundingVolume(), m_pkContactSet ) )
				{
					if( m_pkContactSet->m_vpkContacts[0]->m_kNormal * m_kGravity < -EPSILON )
						pkFirstNode->AddContact( pkSecondNode, m_pkContactSet );
					else
					{
						vector< Contact* >::iterator ppkContact    = m_pkContactSet->m_vpkContacts.begin();
						vector< Contact* >::iterator ppkContactEnd = m_pkContactSet->m_vpkContacts.end();

						for( ; ppkContact != ppkContactEnd; ++ppkContact )
							(*ppkContact)->m_kNormal = -(*ppkContact)->m_kNormal;

						pkSecondNode->AddContact( pkFirstNode, m_pkContactSet );
					}

					m_pkContactSet = new ContactSet;
				}
			}

			//Loop and test with resting bodies
			for( unsigned int uiRestNode = 0; uiRestNode < uiNumResting; ++uiRestNode )
			{
				ContactNode *pkOtherNode = m_ppkResting[ uiRestNode ];
				RigidBody   *pkOtherBody = pkOtherNode->m_pkBody;

				//Check if joints connect bodies
				vector< Joint* >::iterator ppkJoint    = pkBody->m_vpkJoints.begin();
				vector< Joint* >::iterator ppkJointEnd = pkBody->m_vpkJoints.end();

				for( ; ppkJoint != ppkJointEnd; ++ppkJoint )
				{
					Joint *pkJoint = *ppkJoint;

					if( ( ( pkJoint->m_apkBody[0] == pkBody ) && ( pkJoint->m_apkBody[1] == pkOtherBody ) ) ||
					    ( ( pkJoint->m_apkBody[1] == pkBody ) && ( pkJoint->m_apkBody[0] == pkOtherBody ) ) )
						break;
				}

				if( ppkJoint < ppkJointEnd )
					continue;

				ContactNode *pkFirstNode  = pkBodyNode;
				ContactNode *pkSecondNode = pkOtherNode;

				if( ( ( pkSecondNode->m_pkBody->GetWorldTranslation() - pkFirstNode->m_pkBody->GetWorldTranslation() ) * m_kGravity ) < 0.0f )
				{
					pkFirstNode  = pkOtherNode;
					pkSecondNode = pkBodyNode;
				}

				if( pkFirstNode->m_pkBody->GetBoundingVolume()->Intersection( pkSecondNode->m_pkBody->GetBoundingVolume(), m_pkContactSet ) )
				{
					if( m_pkContactSet->m_vpkContacts[0]->m_kNormal * m_kGravity < -EPSILON )
						pkFirstNode->AddContact( pkSecondNode, m_pkContactSet );
					else
					{
						vector< Contact* >::iterator ppkContact    = m_pkContactSet->m_vpkContacts.begin();
						vector< Contact* >::iterator ppkContactEnd = m_pkContactSet->m_vpkContacts.end();

						for( ; ppkContact != ppkContactEnd; ++ppkContact )
							(*ppkContact)->m_kNormal = -(*ppkContact)->m_kNormal;

						pkSecondNode->AddContact( pkFirstNode, m_pkContactSet );
					}

					m_pkContactSet = new ContactSet;

					if( pkOtherBody->m_bResting )
					{
						m_ppkContacts[ uiNumActive + uiAdded ] = pkOtherNode;
						pkOtherNode->m_uiID = uiNumActive + uiAdded;
						pkOtherBody->m_bResting = false;
						++uiAdded;
					}
				}
			}
		}

		uiNumActive += uiAdded;

		//Partition contact nodes
		for( uiCurNode = 0; uiCurNode < uiNumActive; ++uiCurNode )
		{
			if( m_ppkContacts[ uiCurNode ]->m_bVisited )
				continue;

			SetContactNodePartition( m_ppkContacts, m_puiNodePartition, uiCurNode, 0, uiCurNode + 1 );
		}

		//Sort according to partition
		kCollisionSort.Sort( m_puiNodePartition, uiNumActive );

		bool bCollision = false;

		// FIXME : Does propagating collisions top-down --> bottom-up --> top-down --> ... increase stability? Or does it make everything worse?
		unsigned int *puiIndex = kCollisionSort.GetIndices();
		
		for( uiCurNode = 0; uiCurNode < uiNumActive; ++uiCurNode, ++puiIndex )
		{
			bCollision |= m_ppkContacts[ *puiIndex ]->Solve( ContactNode::COLLISION, uiPass, m_uiCollisionPasses );
			
			RigidBody *pkBody = m_vpkBodies[ *puiIndex ];

			//Solve joints
			vector< Joint* >::iterator ppkJoint    = pkBody->m_vpkJoints.begin();
			vector< Joint* >::iterator ppkJointEnd = pkBody->m_vpkJoints.end();

			for( ; ppkJoint != ppkJointEnd; ++ppkJoint )
				(*ppkJoint)->Solve( ContactNode::COLLISION );
		}


#ifdef DEBUG_PHYSICS_MANAGER
	
		unsigned int uiNumVerts = 0;
		unsigned int uiNumPoints = 0;

		for( uiCurNode = 0; uiCurNode < uiNumActive; ++uiCurNode )
		{
			for( unsigned int i = 0; i < ppkContacts[ uiCurNode ]->m_vpkParents.size(); ++i )
				if( !ppkContacts[ uiCurNode ]->m_vpkParents[i] )
					uiNumVerts += 2;

			for( unsigned int j = 0; j < ppkContacts[ uiCurNode ]->m_vpkContactSets.size(); ++j )
				for( unsigned int c = 0; c < ppkContacts[ uiCurNode ]->m_vpkContactSets[j]->m_vpkContacts.size(); ++c )
					uiNumPoints += ppkContacts[ uiCurNode ]->m_vpkContactSets[j]->m_vpkContacts[c]->m_uiNumPoints;

			uiNumVerts += ppkContacts[ uiCurNode ]->m_vpkChildren.size() * 2;
		}

		g_pkDebugLines[0][uiPass] = Core::Get()->GetRenderDevice()->CreateVertexBuffer( Buffer::NORMAL, uiNumVerts, &DiffuseVertex::s_kDecl );

		g_pkDebugLines[0][uiPass]->Lock( Buffer::WRITE );

		DiffuseVertex *pkVert = (DiffuseVertex*)g_pkDebugLines[0][uiPass]->GetVertex();

		for( uiCurNode = 0; uiCurNode < uiNumActive; ++uiCurNode )
		{
			unsigned int i;

			for( i = 0; i < ppkContacts[ uiCurNode ]->m_vpkParents.size(); ++i )
			{
				if( !ppkContacts[ uiCurNode ]->m_vpkParents[i] )
				{
					pkVert->m_kPosition = Vector3d( 0.0f, -9.0f, 0.0f );
					pkVert->m_kColor.Set( 0.0f, 1.0f, 0.0f ); ++pkVert;

					pkVert->m_kPosition = ppkContacts[ uiCurNode ]->m_pkBody->GetWorldTranslation();
					pkVert->m_kColor.Set( 1.0f, 0.0f, 0.0f ); ++pkVert;
				}
			}

			for( i = 0; i < ppkContacts[ uiCurNode ]->m_vpkChildren.size(); ++i )
			{
				pkVert->m_kPosition = ppkContacts[ uiCurNode ]->m_pkBody->GetWorldTranslation();
				pkVert->m_kColor.Set( 0.0f, 1.0f, 0.0f ); ++pkVert;

				pkVert->m_kPosition = ppkContacts[ uiCurNode ]->m_vpkChildren[i]->m_pkBody->GetWorldTranslation();
				pkVert->m_kColor.Set( 1.0f, 0.0f, 0.0f ); ++pkVert;
			}
		}

		g_pkDebugLines[0][uiPass]->Unlock();

		g_pkDebugPoints[0][uiPass] = Core::Get()->GetRenderDevice()->CreateVertexBuffer( Buffer::NORMAL, uiNumPoints, &DiffuseVertex::s_kDecl );

		g_pkDebugPoints[0][uiPass]->Lock( Buffer::WRITE );

		pkVert = (DiffuseVertex*)g_pkDebugPoints[0][uiPass]->GetVertex();

		for( uiCurNode = 0; uiCurNode < uiNumActive; ++uiCurNode )
		{
			for( unsigned int j = 0; j < ppkContacts[ uiCurNode ]->m_vpkContactSets.size(); ++j )
				for( unsigned int c = 0; c < ppkContacts[ uiCurNode ]->m_vpkContactSets[j]->m_vpkContacts.size(); ++c )
					for( unsigned int p = 0; p < ppkContacts[ uiCurNode ]->m_vpkContactSets[j]->m_vpkContacts[c]->m_uiNumPoints; ++p )
					{
						pkVert->m_kPosition = ppkContacts[ uiCurNode ]->m_vpkContactSets[j]->m_vpkContacts[c]->m_pkPoints[p];
						pkVert->m_kColor.Set( 0.0f, 0.0f, 1.0f );
						++pkVert;
					}
		}

		g_pkDebugPoints[0][uiPass]->Unlock();

#endif

		if( !bCollision )
			break;
	}

#ifdef DEBUG_PHYSICS_MANAGER
	for( ++uiPass; uiPass < m_uiCollisionPasses; ++uiPass )
	{
		g_pkDebugLines[0][uiPass]  = 0;
		g_pkDebugPoints[0][uiPass] = 0;
	}

	g_pkDebugLines[2][0]  = 0;
	g_pkDebugPoints[2][0] = 0;
#endif

	bool bCollision = true;

	//Contact phase
	for( uiPass = 0; uiPass < m_uiContactPasses + 1; ++uiPass )
	{
		ppkBodyNode  = m_vpkBodyNodes.begin();
		pbActive     = m_pbIsActive;
		uiNumActive  = 0;
		uiNumResting = 0;

		// Predict new position and new velocities
		for( ; ppkBodyNode != ppkBodyNodeEnd; ++ppkBodyNode, ++pbActive )
		{
			ContactNode *pkBodyNode = *ppkBodyNode;
			RigidBody   *pkBody     = pkBodyNode->m_pkBody;

			if( !*pbActive )
				continue;

			if( pkBody->m_bResting )
			{
				//Check if still resting
				vector< RigidBody* >::iterator ppkRestee     = pkBody->m_vpkRestOn.begin();
				vector< RigidBody* >::iterator ppkResteeEnd  = pkBody->m_vpkRestOn.end();
				
				for( ; ppkRestee != ppkResteeEnd; ++ppkRestee )
					if( (*ppkRestee) && ( !(*ppkRestee)->m_bResting || !m_pbIsActive[ (*ppkRestee)->m_uiID ] ) )
					{
						pkBody->m_bResting = false;
						break;
					}
			}

			if( !pkBody->m_bResting )
			{
				Vector3d kLinearVelocity  = pkBody->m_kVelocity;
				Vector3d kAngularVelocity = pkBody->m_kAngularVelocity;

				pkBody->PopState();

				pkBody->m_kVelocity        = kLinearVelocity;
				pkBody->m_kAngularVelocity = kAngularVelocity;

				pkBody->PushState();

				if( !uiPass )
				{
					pkBody->m_kVelocity        += ( ( pkBody->m_kForce * pkBody->m_fInvMass ) + m_kGravity ) * fDeltaTime;
					pkBody->m_kAngularVelocity += pkBody->m_kInvInertiaTensor * ( pkBody->m_kTorque * fDeltaTime );
				}

				pkBody->m_kTranslation += pkBody->m_kVelocity * fDeltaTime;
				pkBody->m_kRotation     = ( pkBody->m_kRotation + ( ( Quaternion( pkBody->m_kAngularVelocity.x, pkBody->m_kAngularVelocity.y, pkBody->m_kAngularVelocity.z, 0.0f ) * pkBody->m_kRotation ) * ( 0.5f * fDeltaTime ) ) ).Normalize();

				pkBody->m_kRotation.ToMatrix( &pkBody->m_kRotationMatrix );
				pkBody->m_kInvInertiaTensor = ( pkBody->m_kRotationMatrix * pkBody->m_kInvBodyInertiaTensor );

				Matrix kRotTranspose( pkBody->m_kRotationMatrix ); kRotTranspose.Transpose();
				pkBody->m_kInvInertiaTensor = ( pkBody->m_kInvInertiaTensor * kRotTranspose );

				pkBody->NotifyUpdate( false );

				m_ppkContacts[ uiNumActive++ ] = pkBodyNode;
			}
			else
				m_ppkResting[ uiNumResting++ ] = pkBodyNode;

			pkBodyNode->m_bVisited = false;
			pkBodyNode->Clear();
		}

		unsigned int uiAdded = 0;

		//Collision tests
		for( uiCurNode = 0 ; uiCurNode != uiNumActive; ++uiCurNode )
		{
			ContactNode *pkBodyNode = m_ppkContacts[ uiCurNode ];
			RigidBody   *pkBody     = pkBodyNode->m_pkBody;

			pkBodyNode->m_uiID      = uiCurNode;

			if( m_pkRoom && m_pkRoom->Intersection( pkBody->GetBoundingVolume(), m_pkContactSet ) )
			{
				pkBodyNode->AddContact( 0, m_pkContactSet );
				m_pkContactSet = new ContactSet;
			}

			for( unsigned int uiNextNode = uiCurNode + 1; uiNextNode < uiNumActive; ++uiNextNode )
			{
				ContactNode *pkOtherNode = m_ppkContacts[ uiNextNode ];
				RigidBody   *pkOtherBody = pkOtherNode->m_pkBody;

				//Check if joints connect bodies
				vector< Joint* >::iterator ppkJoint    = pkBody->m_vpkJoints.begin();
				vector< Joint* >::iterator ppkJointEnd = pkBody->m_vpkJoints.end();

				for( ; ppkJoint != ppkJointEnd; ++ppkJoint )
				{
					Joint *pkJoint = *ppkJoint;

					if( ( ( pkJoint->m_apkBody[0] == pkBody ) && ( pkJoint->m_apkBody[1] == pkOtherBody ) ) ||
					    ( ( pkJoint->m_apkBody[1] == pkBody ) && ( pkJoint->m_apkBody[0] == pkOtherBody ) ) )
						break;
				}

				if( ppkJoint < ppkJointEnd )
					continue;

				ContactNode *pkFirstNode  = pkBodyNode;
				ContactNode *pkSecondNode = pkOtherNode;

				if( ( ( pkSecondNode->m_pkBody->GetWorldTranslation() - pkFirstNode->m_pkBody->GetWorldTranslation() ) * m_kGravity ) < 0.0f )
				{
					pkFirstNode  = pkOtherNode;
					pkSecondNode = pkBodyNode;
				}

				if( pkFirstNode->m_pkBody->GetBoundingVolume()->Intersection( pkSecondNode->m_pkBody->GetBoundingVolume(), m_pkContactSet ) )
				{
					if( m_pkContactSet->m_vpkContacts[0]->m_kNormal * m_kGravity < -EPSILON )
						pkFirstNode->AddContact( pkSecondNode, m_pkContactSet );
					else
					{
						vector< Contact* >::iterator ppkContact    = m_pkContactSet->m_vpkContacts.begin();
						vector< Contact* >::iterator ppkContactEnd = m_pkContactSet->m_vpkContacts.end();

						for( ; ppkContact != ppkContactEnd; ++ppkContact )
							(*ppkContact)->m_kNormal = -(*ppkContact)->m_kNormal;

						pkSecondNode->AddContact( pkFirstNode, m_pkContactSet );
					}

					m_pkContactSet = new ContactSet;
				}
			}

			//Loop and test with resting bodies
			for( unsigned int uiRestNode = 0; uiRestNode < uiNumResting; ++uiRestNode )
			{
				ContactNode *pkOtherNode = m_ppkResting[ uiRestNode ];
				RigidBody   *pkOtherBody = pkOtherNode->m_pkBody;

				//Check if joints connect bodies
				vector< Joint* >::iterator ppkJoint    = pkBody->m_vpkJoints.begin();
				vector< Joint* >::iterator ppkJointEnd = pkBody->m_vpkJoints.end();

				for( ; ppkJoint != ppkJointEnd; ++ppkJoint )
				{
					Joint *pkJoint = *ppkJoint;

					if( ( ( pkJoint->m_apkBody[0] == pkBody ) && ( pkJoint->m_apkBody[1] == pkOtherBody ) ) ||
					    ( ( pkJoint->m_apkBody[1] == pkBody ) && ( pkJoint->m_apkBody[0] == pkOtherBody ) ) )
						break;
				}

				if( ppkJoint < ppkJointEnd )
					continue;

				ContactNode *pkFirstNode  = pkBodyNode;
				ContactNode *pkSecondNode = pkOtherNode;

				if( ( ( pkSecondNode->m_pkBody->GetWorldTranslation() - pkFirstNode->m_pkBody->GetWorldTranslation() ) * m_kGravity ) < 0.0f )
				{
					pkFirstNode  = pkOtherNode;
					pkSecondNode = pkBodyNode;
				}

				if( pkFirstNode->m_pkBody->GetBoundingVolume()->Intersection( pkSecondNode->m_pkBody->GetBoundingVolume(), m_pkContactSet ) )
				{
					if( m_pkContactSet->m_vpkContacts[0]->m_kNormal * m_kGravity < -EPSILON )
						pkFirstNode->AddContact( pkSecondNode, m_pkContactSet );
					else
					{
						vector< Contact* >::iterator ppkContact    = m_pkContactSet->m_vpkContacts.begin();
						vector< Contact* >::iterator ppkContactEnd = m_pkContactSet->m_vpkContacts.end();

						for( ; ppkContact != ppkContactEnd; ++ppkContact )
							(*ppkContact)->m_kNormal = -(*ppkContact)->m_kNormal;

						pkSecondNode->AddContact( pkFirstNode, m_pkContactSet );
					}

					m_pkContactSet = new ContactSet;

					if( pkOtherBody->m_bResting )
					{
						m_ppkContacts[ uiNumActive + uiAdded ] = pkOtherNode;
						pkOtherNode->m_uiID = uiNumActive + uiAdded;
						pkOtherBody->m_bResting = false;
						++uiAdded;
					}
				}
			}
		}

		uiNumActive += uiAdded;

		//Partition contact nodes
		for( uiCurNode = 0; uiCurNode < uiNumActive; ++uiCurNode )
		{
			if( m_ppkContacts[ uiCurNode ]->m_bVisited )
				continue;

			SetContactNodePartition( m_ppkContacts, m_puiNodePartition, uiCurNode, 0, uiCurNode + 1 );
		}

		//Sort according to partition
		kCollisionSort.Sort( m_puiNodePartition, uiNumActive );

		bCollision = false;

		unsigned int *puiIndex = kCollisionSort.GetIndices();

		if( uiPass < m_uiContactPasses )
		{
			for( uiCurNode = 0; uiCurNode < uiNumActive; ++uiCurNode, ++puiIndex )
			{
				bCollision |= m_ppkContacts[ *puiIndex ]->Solve( ContactNode::CONTACT, uiPass, m_uiContactPasses );
			
				RigidBody *pkBody = m_vpkBodies[ *puiIndex ];

				//Solve joints
				vector< Joint* >::iterator ppkJoint    = pkBody->m_vpkJoints.begin();
				vector< Joint* >::iterator ppkJointEnd = pkBody->m_vpkJoints.end();

				for( ; ppkJoint != ppkJointEnd; ++ppkJoint )
					bCollision |= (*ppkJoint)->Solve( ContactNode::CONTACT );
			}
		}
		else if( uiPass == m_uiContactPasses )
		{
			// Shock propagation
			for( uiCurNode = 0; uiCurNode < uiNumActive; ++uiCurNode, ++puiIndex )
				m_ppkContacts[ *puiIndex ]->Solve( ContactNode::SHOCKPROPAGATION, 0, 1 );

			puiIndex = kCollisionSort.GetIndices();

			//Apply small separation impulses to resolve final penetrating points with non-separating velocity
			for( uiCurNode = 0; uiCurNode < uiNumActive; ++uiCurNode, ++puiIndex )
			{
				ContactNode *pkBodyNode = m_ppkContacts[ *puiIndex ];

				vector< ContactSet*  >::iterator ppkSet    = pkBodyNode->m_vpkContactSets.begin();
				vector< ContactSet*  >::iterator ppkEnd    = pkBodyNode->m_vpkContactSets.end();
				vector< ContactNode* >::iterator ppkParent = pkBodyNode->m_vpkParents.begin();

				for( ; ppkSet != ppkEnd; ++ppkSet, ++ppkParent )
				{
					Contact *pkContact = (*ppkSet)->m_vpkContacts[0];

					Vector3d *pkPoint = pkContact->m_pkPoints;
					float    *pfDepth = pkContact->m_pfDepths;

					Vector3d kPoint;
					float fMaxVel  = 0.0f;
					float fDeepest = 0.0f;

					for( unsigned int uiPoint = 0; uiPoint < pkContact->m_uiNumPoints; ++uiPoint, ++pkPoint, ++pfDepth )
					{
						Vector3d kPointVel = pkBodyNode->m_pkBody->GetPointVelocity( *pkPoint, false );

						RigidBody *pkOtherBody = (*ppkParent) ? (*ppkParent)->m_pkBody : 0;

						if( pkOtherBody )
							kPointVel -= pkOtherBody->GetPointVelocity( *pkPoint, false );

						float fVelN = kPointVel * pkContact->m_kNormal;

						if( fVelN <= fMaxVel )
						{
							fMaxVel = fVelN;
							kPoint = *pkPoint;
						}

						if( *pfDepth > fDeepest )
							fDeepest = *pfDepth;
					}

					if( fDeepest > 0.01f )
					{
						pkBodyNode->m_pkBody->ApplyImpulse( pkContact->m_kNormal * ( ( -fMaxVel + ( fDeepest * 0.5f ) ) * pkBodyNode->m_pkBody->GetMass() ), pkBodyNode->m_pkBody->GetWorldTranslation(), false );

						pkBodyNode->m_pkBody->m_bPenetrating = true;
						pkBodyNode->m_pkBody->m_bResting     = false;
					}
				}
			}

			bCollision = true;
		}

#ifdef DEBUG_PHYSICS_MANAGER
	
		int iMode = ( uiPass < m_uiContactPasses ) ? ContactNode::CONTACT : ContactNode::SHOCKPROPAGATION;

		if( iMode == ContactNode::SHOCKPROPAGATION )
			uiPass = 0;

		unsigned int uiNumVerts = 0;
		unsigned int uiNumPoints = 0;

		for( uiCurNode = 0; uiCurNode < uiNumActive; ++uiCurNode )
		{
			for( unsigned int i = 0; i < ppkContacts[ uiCurNode ]->m_vpkParents.size(); ++i )
				if( !ppkContacts[ uiCurNode ]->m_vpkParents[i] )
					uiNumVerts += 2;

			for( unsigned int j = 0; j < ppkContacts[ uiCurNode ]->m_vpkContactSets.size(); ++j )
				for( unsigned int c = 0; c < ppkContacts[ uiCurNode ]->m_vpkContactSets[j]->m_vpkContacts.size(); ++c )
					uiNumPoints += ppkContacts[ uiCurNode ]->m_vpkContactSets[j]->m_vpkContacts[c]->m_uiNumPoints;

			uiNumVerts += ppkContacts[ uiCurNode ]->m_vpkChildren.size() * 2;
		}

		g_pkDebugLines[iMode][uiPass] = Core::Get()->GetRenderDevice()->CreateVertexBuffer( Buffer::NORMAL, uiNumVerts, &DiffuseVertex::s_kDecl );

		g_pkDebugLines[iMode][uiPass]->Lock( Buffer::WRITE );

		DiffuseVertex *pkVert = (DiffuseVertex*)g_pkDebugLines[iMode][uiPass]->GetVertex();

		for( uiCurNode = 0; uiCurNode < uiNumActive; ++uiCurNode )
		{
			unsigned int i;

			for( i = 0; i < ppkContacts[ uiCurNode ]->m_vpkParents.size(); ++i )
			{
				if( !ppkContacts[ uiCurNode ]->m_vpkParents[i] )
				{
					pkVert->m_kPosition = Vector3d( 0.0f, -9.0f, 0.0f );
					pkVert->m_kColor.Set( 0.0f, 1.0f, 0.0f ); ++pkVert;

					pkVert->m_kPosition = ppkContacts[ uiCurNode ]->m_pkBody->GetWorldTranslation();
					pkVert->m_kColor.Set( 1.0f, 0.0f, 0.0f ); ++pkVert;
				}
			}

			for( i = 0; i < ppkContacts[ uiCurNode ]->m_vpkChildren.size(); ++i )
			{
				pkVert->m_kPosition = ppkContacts[ uiCurNode ]->m_pkBody->GetWorldTranslation();
				pkVert->m_kColor.Set( 0.0f, 1.0f, 0.0f ); ++pkVert;

				pkVert->m_kPosition = ppkContacts[ uiCurNode ]->m_vpkChildren[i]->m_pkBody->GetWorldTranslation();
				pkVert->m_kColor.Set( 1.0f, 0.0f, 0.0f ); ++pkVert;
			}
		}

		g_pkDebugLines[iMode][uiPass]->Unlock();

		g_pkDebugPoints[iMode][uiPass] = Core::Get()->GetRenderDevice()->CreateVertexBuffer( Buffer::NORMAL, uiNumPoints, &DiffuseVertex::s_kDecl );

		g_pkDebugPoints[iMode][uiPass]->Lock( Buffer::WRITE );

		pkVert = (DiffuseVertex*)g_pkDebugPoints[iMode][uiPass]->GetVertex();

		for( uiCurNode = 0; uiCurNode < uiNumActive; ++uiCurNode )
		{
			for( unsigned int j = 0; j < ppkContacts[ uiCurNode ]->m_vpkContactSets.size(); ++j )
				for( unsigned int c = 0; c < ppkContacts[ uiCurNode ]->m_vpkContactSets[j]->m_vpkContacts.size(); ++c )
					for( unsigned int p = 0; p < ppkContacts[ uiCurNode ]->m_vpkContactSets[j]->m_vpkContacts[c]->m_uiNumPoints; ++p )
					{
						pkVert->m_kPosition = ppkContacts[ uiCurNode ]->m_vpkContactSets[j]->m_vpkContacts[c]->m_pkPoints[p];
						pkVert->m_kColor.Set( 0.0f, 0.0f, 1.0f );
						++pkVert;
					}
		}

		g_pkDebugPoints[iMode][uiPass]->Unlock();

		if( iMode == ContactNode::SHOCKPROPAGATION )
			uiPass = m_uiContactPasses;

#endif

		if( !bCollision )
		{
#ifdef DEBUG_PHYSICS_MANAGER
			for( ++uiPass; uiPass < m_uiContactPasses; ++uiPass )
			{
				g_pkDebugLines[1][uiPass] = 0;
				g_pkDebugPoints[1][uiPass] = 0;
			}
#endif

			uiPass = m_uiContactPasses - 1; //Skip to shock propagation
		}
	}

	//Calculate final positions and velocities
	ppkBodyNode = m_vpkBodyNodes.begin();
	pbActive = m_pbIsActive;

	for( ; ppkBodyNode != ppkBodyNodeEnd; ++ppkBodyNode, ++pbActive )
	{
		ContactNode *pkBodyNode = *ppkBodyNode;
		RigidBody *pkBody = pkBodyNode->m_pkBody;

		if( !*pbActive )
			continue;

		if( !pkBody->m_bResting )
		{
			if( pkBodyNode->m_bFree && !pkBodyNode->m_vpkChildren.size() )
				pkBody->PopState( false );
			else
			{
				Vector3d kLinVel = pkBody->m_kVelocity;
				Vector3d kAngVel = pkBody->m_kAngularVelocity;

				pkBody->PopState();

				bool bUpdateAng = true;

				if( !pkBody->m_bPenetrating && ( kLinVel.Len2() < 0.0009f ) )
				{
					pkBody->m_kVelocity.Reset();
				
					if( kAngVel.Len2() < 0.001f )
					{
						pkBody->m_kAngularVelocity.Reset();
						bUpdateAng = false;

						if( pkBodyNode->m_vpkParents.size() )
							pkBody->m_bResting = true;
					}
				}
				else
				{
					pkBody->m_kVelocity     = kLinVel;
					pkBody->m_kTranslation += kLinVel * fDeltaTime;
				}

				if( bUpdateAng )
				{
					pkBody->m_kAngularVelocity = kAngVel;
					pkBody->m_kRotation        = ( ( ( Quaternion( kAngVel.x, kAngVel.y, kAngVel.z, 0.0f ) * pkBody->m_kRotation ) * ( 0.5f * fDeltaTime ) ) + pkBody->m_kRotation ).Normalize();

					pkBody->m_kRotation.ToMatrix( &pkBody->m_kRotationMatrix );
					pkBody->m_kInvInertiaTensor = ( pkBody->m_kRotationMatrix * pkBody->m_kInvBodyInertiaTensor );

					Matrix kRotTranspose( pkBody->m_kRotationMatrix ); kRotTranspose.Transpose();
					pkBody->m_kInvInertiaTensor = ( pkBody->m_kInvInertiaTensor * kRotTranspose );
				}
			}
		}
		else
			pkBody->PopState( false );

		pkBody->ResetForceAndTorque();

		pkBody->Activate();
	}

	//Sweep contact graph and reset invalid rest states
	{
		unsigned int *puiIndex = kCollisionSort.GetIndices();

		for( uiCurNode = 0; uiCurNode < uiNumActive; ++uiCurNode, ++puiIndex )
		{
			ContactNode *pkContactNode = m_ppkContacts[ *puiIndex ];
			RigidBody   *pkBody        = pkContactNode->m_pkBody;

			if( !pkBody->m_bResting )
				continue;

			if( pkContactNode->m_vpkParents.size() )
			{
				vector< ContactNode* >::iterator ppkParent    = pkContactNode->m_vpkParents.begin();
				vector< ContactNode* >::iterator ppkParentEnd = pkContactNode->m_vpkParents.end();

				for( ; ppkParent != ppkParentEnd; ++ppkParent )
				{
					if( *ppkParent && (*ppkParent)->m_pkBody && ( !(*ppkParent)->m_pkBody->m_bResting || !(*ppkParent)->m_pkBody->IsActive() ) )
					{
						pkBody->m_bResting = false;
						break;
					}
				}

				if( pkBody->m_bResting )
				{
					//Body is actually resting, store
					pkBody->m_vpkRestOn.clear();

					ppkParent = pkContactNode->m_vpkParents.begin();

					for( ; ppkParent != ppkParentEnd; ++ppkParent )
					{
						if( *ppkParent && (*ppkParent)->m_pkBody )
							pkBody->m_vpkRestOn.push_back( (*ppkParent)->m_pkBody );
						else
							pkBody->m_vpkRestOn.push_back( 0 );
					}
				}
			}
		}
	}
}


void PhysicsManager::AddRigidBody( RigidBody *pkBody )
{
	vector< RigidBody* >::iterator ppkBody    = m_vpkBodies.begin();
	vector< RigidBody* >::iterator ppkBodyEnd = m_vpkBodies.end();

	for( ; ppkBody != ppkBodyEnd; ++ppkBody )
		if( (*ppkBody) == pkBody )
			return;

	m_vpkBodies.push_back( pkBody );
	m_vpkBodyNodes.push_back( new ContactNode( pkBody ) );

	pkBody->m_bManaged = true;

	if( m_vpkBodyNodes.size() >= m_uiStoreSize )
	{
		delete [] m_pbIsActive;
		delete [] m_ppkContacts;
		delete [] m_ppkResting;
		delete [] m_puiNodePartition;

		m_uiStoreSize      += 100;
		m_pbIsActive       = new bool[ m_uiStoreSize ];
		m_ppkContacts      = new ContactNode*[ m_uiStoreSize ];
		m_ppkResting       = new ContactNode*[ m_uiStoreSize ];
		m_puiNodePartition = new unsigned int[ m_uiStoreSize ];
	}
}


void PhysicsManager::RemoveRigidBody( RigidBody *pkBody )
{
	vector< RigidBody* >::iterator   ppkBody    = m_vpkBodies.begin();
	vector< RigidBody* >::iterator   ppkBodyEnd = m_vpkBodies.end();
	vector< ContactNode* >::iterator ppkNode    = m_vpkBodyNodes.begin();

	for( ; ppkBody != ppkBodyEnd; ++ppkBody, ++ppkNode )
		if( (*ppkBody) == pkBody )
		{
			delete( *ppkNode );
			m_vpkBodies.erase( ppkBody );
			m_vpkBodyNodes.erase( ppkNode );

			//Check if any other node is resting on this
			ppkBody    = m_vpkBodies.begin();
			ppkBodyEnd = m_vpkBodies.end();

			for( ; ppkBody != ppkBodyEnd; ++ppkBody )
				if( (*ppkBody)->m_bResting )
				{
					vector< RigidBody* >::iterator ppkRestee    = (*ppkBody)->m_vpkRestOn.begin();
					vector< RigidBody* >::iterator ppkResteeEnd = (*ppkBody)->m_vpkRestOn.end();

					for( ; ppkRestee != ppkResteeEnd; ++ppkRestee )
						if( (*ppkRestee) == pkBody )
						{
							(*ppkBody)->m_bResting = false;
							break;
						}
				}

			return;
		}

	pkBody->m_bManaged = false;
}


void PhysicsManager::AddJoint( Joint *pkJoint )
{
	vector< Joint* >::iterator ppkJoint    = m_vpkJoints.begin();
	vector< Joint* >::iterator ppkJointEnd = m_vpkJoints.end();

	for( ; ppkJoint != ppkJointEnd; ++ppkJoint )
	{
		if( (*ppkJoint) == pkJoint )
			return;
	}

	m_vpkJoints.push_back( pkJoint );

	pkJoint->m_apkBody[0]->m_vpkJoints.push_back( pkJoint );

	if( pkJoint->m_apkBody[1] )
		pkJoint->m_apkBody[1]->m_vpkJoints.push_back( pkJoint );
}


void PhysicsManager::RemoveJoint( Joint *pkJoint )
{
	vector< Joint* >::iterator ppkJoint    = find( m_vpkJoints.begin(), m_vpkJoints.end(), pkJoint );

	if( ppkJoint != m_vpkJoints.end() )
		m_vpkJoints.erase( ppkJoint );

	ppkJoint = find( pkJoint->m_apkBody[0]->m_vpkJoints.begin(), pkJoint->m_apkBody[0]->m_vpkJoints.end(), pkJoint );

	if( ppkJoint != pkJoint->m_apkBody[0]->m_vpkJoints.end() )
		pkJoint->m_apkBody[0]->m_vpkJoints.erase( ppkJoint );

	if( pkJoint->m_apkBody[1] )
	{
		ppkJoint = find( pkJoint->m_apkBody[1]->m_vpkJoints.begin(), pkJoint->m_apkBody[1]->m_vpkJoints.end(), pkJoint );

		if( ppkJoint != pkJoint->m_apkBody[1]->m_vpkJoints.end() )
			pkJoint->m_apkBody[1]->m_vpkJoints.erase( ppkJoint );
	}
}


void PhysicsManager::RenderDebug( ContactNode::CONTACTMODE eMode, unsigned int uiPass )
{
#ifdef DEBUG_PHYSICS_MANAGER
	Core::Get()->GetRenderDevice()->Begin( Camera::GetActive()->GetViewMatrix() );

	if( ( eMode == ContactNode::COLLISION ) && ( uiPass >= m_uiCollisionPasses ) )
		return;

	if( ( eMode == ContactNode::CONTACT ) && ( uiPass >= m_uiContactPasses ) )
		return;

	if( ( eMode == ContactNode::SHOCKPROPAGATION ) && uiPass )
		return;

	if( ( g_pkDebugLines[eMode][uiPass] != (VertexBuffer*)0 ) && ( g_pkDebugLines[eMode][uiPass]->GetNumElements() > 0 ) )
	{
		RenderPrimitive kPrim;

		kPrim.m_ePrimitive = RenderPrimitive::LINES;
		kPrim.m_uiNumPrimitives = g_pkDebugLines[eMode][uiPass]->GetNumElements() / 2;
		kPrim.m_pkVertexBuffer = g_pkDebugLines[eMode][uiPass];
		kPrim.m_fSize = 1.0f;

		kPrim.m_pkMaterial = new Material( "lines", 0 );

		kPrim.m_pkMaterial->m_kZBufferMode.Set( ZBufferMode::ALWAYSNOWRITE );

		Core::Get()->GetRenderDevice()->Render( kPrim );
	}

	if( ( g_pkDebugPoints[eMode][uiPass] != (VertexBuffer*)0 ) && ( g_pkDebugPoints[eMode][uiPass]->GetNumElements() > 0 ) )
	{
		RenderPrimitive kPrim;

		kPrim.m_ePrimitive = RenderPrimitive::POINTS;
		kPrim.m_uiNumPrimitives = g_pkDebugPoints[eMode][uiPass]->GetNumElements();
		kPrim.m_pkVertexBuffer = g_pkDebugPoints[eMode][uiPass];
		kPrim.m_fSize = 5.0f;

		kPrim.m_pkMaterial = new Material( "points", 0 );

		kPrim.m_pkMaterial->m_kZBufferMode.Set( ZBufferMode::ALWAYSNOWRITE );

		Core::Get()->GetRenderDevice()->Render( kPrim );
	}

	Core::Get()->GetRenderDevice()->End();
#endif
}


};
