/***************************************************************************
                planarreflection.cpp  -  Planar reflection class
                             -------------------
    begin                : Sun Apr 27 2003
    copyright            : (C) 2003 by Cody Russell
    email                : cody [at] jhu.edu
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, planarreflection.cpp

 The Initial Developer of the Original Code is Cody Russell.
 Portions created by Cody Russell are Copyright (C) 2002
 Cody Russell. All Rights Reserved.

 ***************************************************************************/

#include "camera.h"
#include "logstream.h"
#include "planarreflection.h"
#include "polygon.h"
#include "render.h"
#include "renderprimitive.h"
#include "vertex.h"


namespace NeoEngine
{


PlanarReflection::PlanarReflection( const MaterialPtr &rkMaterial, VertexBufferPtr &pkVertexBuffer, PolygonBufferPtr &pkPolygonBuffer )
{
	m_pkVertexBuffer  = pkVertexBuffer;
	m_pkPolygonBuffer = pkPolygonBuffer;
	m_pkMaterial      = rkMaterial;

	pkVertexBuffer->Lock( Buffer::READ );

	unsigned char *pucVertex    = (unsigned char*)pkVertexBuffer->GetVertex();
	unsigned int   uiVertexSize = pkVertexBuffer->GetVertexSize();

	Vector3d &v0 = *((Vector3d*)( pucVertex                    ) );
	Vector3d &v1 = *((Vector3d*)( pucVertex + uiVertexSize     ) );
	Vector3d &v2 = *((Vector3d*)( pucVertex + uiVertexSize * 2 ) );

	m_akPoints[0] = v0;
	m_akPoints[1] = v1;
	m_akPoints[2] = v2;

	Set( v0, v1, v2 );

	pkVertexBuffer->Unlock();

	m_bIsRendering = false;
	Core::Get()->GetRenderDevice()->RegisterFrameCallback( FrameCallback::FRAMEEND, this, true );
}

PlanarReflection::~PlanarReflection()
{
	if( m_pkTexture )
	{
		delete m_pkTexture;
	}
}

void PlanarReflection::Draw()
{
	/* Render the geometry */
	static RenderPrimitive  skPrimitive;
	static RenderPrimitive *spkPrimitive = &skPrimitive;

	RenderDevice *pkRenderDevice = Core::Get()->GetRenderDevice();

        pkRenderDevice->Begin( pkRenderDevice->GetViewMatrix(), 0 ); // RenderDevice::NOSORT );
 
        if( !m_pkPolygonBuffer->IsStripped() )
        {
                spkPrimitive->m_ePrimitive           = RenderPrimitive::TRIANGLES;
                spkPrimitive->m_pkPolygonBuffer      = m_pkPolygonBuffer;
                spkPrimitive->m_uiNumPrimitives      = spkPrimitive->m_pkPolygonBuffer->GetNumElements();
        }
        else
        {
                spkPrimitive->m_ePrimitive           = RenderPrimitive::TRIANGLESTRIP;
                spkPrimitive->m_pkPolygonStripBuffer = m_pkPolygonBuffer->GetStrip();
                spkPrimitive->m_uiNumPrimitives      = spkPrimitive->m_pkPolygonStripBuffer->GetNumElements();
        }
                                                                                                                       
        spkPrimitive->m_pkMaterial           = m_pkMaterial;
        spkPrimitive->m_pkVertexBuffer       = m_pkVertexBuffer;
                                                                                                                       
        pkRenderDevice->Render( *spkPrimitive, 0 );
                                                                                                                       
        pkRenderDevice->End();
}

bool PlanarReflection::Render( Frustum *pkFrustum, bool bForce )
{
	return true;
}

VertexBufferPtr &PlanarReflection::GetVertexBuffer()
{
        return m_pkVertexBuffer;
}

PolygonBufferPtr &PlanarReflection::GetPolygonBuffer()
{
        return m_pkPolygonBuffer;
}

void PlanarReflection::SetVertexBuffer( VertexBufferPtr pkVertexBuffer )
{
	m_pkVertexBuffer = pkVertexBuffer;

	m_pkVertexBuffer->Lock( Buffer::READ );

	unsigned char *pucVertex    = (unsigned char*)pkVertexBuffer->GetVertex();
	unsigned int   uiVertexSize = pkVertexBuffer->GetVertexSize();

	m_akPoints[0] = *((Vector3d*)( pucVertex                    ) );
	m_akPoints[1] = *((Vector3d*)( pucVertex + uiVertexSize     ) );
	m_akPoints[2] = *((Vector3d*)( pucVertex + uiVertexSize * 2 ) );

	m_pkVertexBuffer->Unlock();

	m_kNormal = ((m_akPoints[2] - m_akPoints[1]) % (m_akPoints[0] - m_akPoints[1]));
	m_kNormal.Normalize();
}

void PlanarReflection::SetPolygonBuffer( PolygonBufferPtr pkPolygonBuffer )
{
	m_pkPolygonBuffer = pkPolygonBuffer;
}

void PlanarReflection::FrameEvent( FrameCallback::FRAMECALLBACKTYPE eType, void *pData )
{
	RenderDevice *pkRenderDevice = Core::Get()->GetRenderDevice();

        // If we're already in a callback, don't start another one
        if( m_bIsRendering )
                return;

        m_bIsRendering = true;
 
        pkRenderDevice->Clear( RenderDevice::STENCILBUFFER, Color::BLACK, 0.0f, 0 );
        int ZSave = m_pkMaterial->m_kZBufferMode.Get();
        m_pkMaterial->m_kZBufferMode.Set( ZBufferMode::ALWAYSNOWRITE );
        pkRenderDevice->SetTargetMask( false );
 
        // Draw the shape of the reflector to the stencil buffer
        pkRenderDevice->SetStencilOp( RenderDevice::STENCILOP_REPLACE, RenderDevice::STENCILOP_REPLACE, RenderDevice::STENCILOP_REPLACE );
        pkRenderDevice->SetStencilFunc( RenderDevice::STENCILFUNC_ALWAYS, 1, 0xffffffff );
 
        Draw();
 
        // Re-enable the color and depth buffers
        pkRenderDevice->SetTargetMask( true );
        m_pkMaterial->m_kZBufferMode.Set( ZSave );
 
        // Now only render where the stencil buffer is 1
        pkRenderDevice->SetStencilOp( RenderDevice::STENCILOP_KEEP, RenderDevice::STENCILOP_KEEP, RenderDevice::STENCILOP_KEEP );
        pkRenderDevice->SetStencilFunc( RenderDevice::STENCILFUNC_EQUAL, 1, 0xffffffff );
 
	Draw();
 
        // Save the current camera position and create a temporary camera for rendering the reflection
        Camera *pkCamera;
        Camera *pkSavedCamera = Camera::GetActive();
 
        pkCamera = new Camera( "reflection camera" );
        pkCamera->SetRoom( pkSavedCamera->GetRoom() );

        pkCamera->Translate( m_kNormal * ( -2.0f * Distance( pkSavedCamera->GetTranslation() ) ) );

        Vector3d kOrigin = pkSavedCamera->GetWorldTranslation();
        Vector3d kVector = -( pkSavedCamera->GetWorldTransform().GetColumn( 2 ) );
        Vector3d kTarget = kOrigin + ( kVector * ( Distance( kOrigin ) / -( m_kNormal * kVector ) ) );
        Vector3d kOldUp  = pkSavedCamera->GetWorldTransform().GetColumn( 1 );

        pkCamera->LookAt( kTarget, kOldUp - m_kNormal * 2.0f * ( m_kNormal * kOldUp ) );

	pkRenderDevice->Begin( pkCamera->GetViewMatrix() );
	pkCamera->Render();
	pkRenderDevice->End();

        // Draw the blended floor here
        m_pkMaterial->m_kBlendMode.Set( BlendMode::SRC_SRCALPHA, BlendMode::SRC_ONEMINUSSRCALPHA );
        Draw();

        m_bIsRendering = false;

        pkRenderDevice->SetStencilFunc( RenderDevice::STENCILFUNC_ALWAYS, 1, 0xffffffff );

        pkSavedCamera->SetRoom( pkCamera->GetRoom() );
}


}; // namespace NeoEngine

