/***************************************************************************
           polygonbuffer.cpp  -  Polygon and polygon strip buffers
                             -------------------
    begin                : Fri Nov 2 2001
    copyright            : (C) 2001 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, polygonbuffer.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2001
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "polygon.h"
#include "render.h"
#include "material.h"
#include "nemath.h"
#include "radix.h"
#include "logstream.h"

using namespace std;


namespace NeoEngine
{


#ifndef POLYGONBUFFER_DEBUGSTRIPIFY
//Define this to get debug messages in stripifier method
//#  define POLYGONBUFFER_DEBUGSTRIPIFY
#endif

#ifndef POLYGONBUFFER_REQUIRESTRIPEFF
//Define required efficiency of strips
#  define POLYGONBUFFER_REQUIRESTRIPEFF   0.90f
#endif


// 0 - merge with last+first index repeat
// 1 - switch strips
// 2 - reverse last strip
// 3 - strip padding from last strip

const int POLYGONSTRIPMERGEOP_MERGE        = 0;
const int POLYGONSTRIPMERGEOP_SWITCH       = 1;
const int POLYGONSTRIPMERGEOP_REVERSELAST  = 2;
const int POLYGONSTRIPMERGEOP_STRIPPADLAST = 3;


int g_aaiPolygonStripMergeOp[9][4] = { { POLYGONSTRIPMERGEOP_MERGE,        -1, -1, -1 },
                                       { POLYGONSTRIPMERGEOP_MERGE,        -1, -1, -1 },
                                       { POLYGONSTRIPMERGEOP_MERGE,        -1, -1, -1 },
                                       { POLYGONSTRIPMERGEOP_SWITCH,       POLYGONSTRIPMERGEOP_MERGE,        -1, -1 },
                                       { POLYGONSTRIPMERGEOP_REVERSELAST,  POLYGONSTRIPMERGEOP_MERGE,        -1, -1 },
                                       { POLYGONSTRIPMERGEOP_STRIPPADLAST, POLYGONSTRIPMERGEOP_MERGE,        -1, -1 },
                                       { POLYGONSTRIPMERGEOP_SWITCH,       POLYGONSTRIPMERGEOP_MERGE,        -1, -1 },
                                       { POLYGONSTRIPMERGEOP_SWITCH,       POLYGONSTRIPMERGEOP_STRIPPADLAST, POLYGONSTRIPMERGEOP_MERGE, -1 },
                                       { POLYGONSTRIPMERGEOP_STRIPPADLAST, POLYGONSTRIPMERGEOP_MERGE,        -1, -1 } };











PolygonStripBuffer::~PolygonStripBuffer()
{
	delete [] m_pusBuffer;
}


void PolygonStripBuffer::AllocateStrip( unsigned int uiNumPolygons, const unsigned short *pusData )
{
	assert( !m_uiLock );

	if( uiNumPolygons )
	{
		//If we reach this point, we should not care about NOREADWRITE type, but allow Lock:s below to copy data
		m_uiType &= ~Buffer::NOREADWRITE;
	}

	if( m_uiNumAllocated == uiNumPolygons )
	{
		m_uiNumCurrent = m_uiNumAllocated;

		if( !pusData )
			return;

		Lock( Buffer::WRITE | Buffer::NOUPLOAD );

		LoadStripData( pusData );

		Unlock();

		return;
	}


	delete [] m_pusBuffer;

	m_pusBuffer      = 0;
	m_uiNumAllocated = uiNumPolygons;
	m_uiNumCurrent   = m_uiNumAllocated;

	if( !uiNumPolygons )
		return;

	m_pusBuffer      = new unsigned short[ uiNumPolygons + 2 ];

	if( pusData )
	{
		Lock( Buffer::WRITE | Buffer::NOUPLOAD );

		LoadStripData( pusData );

		Unlock();
	}
}








PolygonBuffer::~PolygonBuffer()
{
	delete [] m_pkBuffer;
	delete [] m_pkNormals;
	delete [] m_pkEdges;
}


void PolygonBuffer::AllocatePolygons( unsigned int uiNumPolygons, const Polygon *pkData, bool bStripify )
{
	assert( !m_uiLock );

	if( uiNumPolygons )
	{
		//If we reach this point, we should not care about NOREADWRITE type, but allow Lock:s below to copy data
		m_uiType &= ~Buffer::NOREADWRITE;
	}

	if( m_uiNumAllocated == uiNumPolygons )
	{
		m_uiNumCurrent = m_uiNumAllocated;

		if( !pkData )
			return;

		Lock( Buffer::WRITE | Buffer::NOUPLOAD );

		LoadPolygonData( pkData, bStripify );

		Unlock();

		return;
	}

	delete [] m_pkBuffer;
	delete [] m_pkNormals, m_pkNormals = 0;
	delete [] m_pkEdges, m_pkEdges = 0;
	
	m_pkBuffer       = 0;
	m_uiNumAllocated = uiNumPolygons;
	m_uiNumCurrent   = m_uiNumAllocated;

	if( !uiNumPolygons )
		return;

	m_pkBuffer       = new Polygon[ uiNumPolygons ];

	if( pkData )
	{
		Lock( Buffer::WRITE | Buffer::NOUPLOAD );

		LoadPolygonData( pkData, bStripify );

		Unlock();
	}
}


void PolygonBuffer::Stripify()
{
	unsigned int uiStrip;

	if( !m_uiNumCurrent )
	{
		neolog << LogLevel( WARNING ) << "*** Unable to stripify polygon buffer: no polygons" << endl;
		return;
	}

	//Lock buffer for reading
	bool bHadLock = IsLocked();

	if( !bHadLock )
		Lock( Buffer::READ );

	//Polygon-added-to-strip flag buffer
	unsigned char *pucAdded         = new unsigned char[ m_uiNumCurrent];
	unsigned char *pucLocalAdded[3];
	
	pucLocalAdded[0] = new unsigned char[ m_uiNumCurrent ];
	pucLocalAdded[1] = new unsigned char[ m_uiNumCurrent ];
	pucLocalAdded[2] = new unsigned char[ m_uiNumCurrent ];

	memset( pucAdded, 0, m_uiNumCurrent );

	//Build connectivity database
	AdjacencyData *pkAdj  = new AdjacencyData[ m_uiNumCurrent ];
	unsigned int    *puiCon = new unsigned int[ m_uiNumCurrent ];

	unsigned int uiPoly;

	memset( puiCon, 0, sizeof( unsigned int ) * m_uiNumCurrent );

	AdjacencyData *pkCurAdj  = pkAdj;
	Polygon       *pkCurPoly = m_pkPolygons;

	for( uiPoly = 0; uiPoly < m_uiNumCurrent; ++uiPoly, ++pkCurAdj, ++pkCurPoly )
	{
		pkCurAdj->m_uiEdges[0][0] = pkCurPoly->v[0];
		pkCurAdj->m_uiEdges[0][1] = pkCurPoly->v[1];

		pkCurAdj->m_uiEdges[1][0] = pkCurPoly->v[1];
		pkCurAdj->m_uiEdges[1][1] = pkCurPoly->v[2];

		pkCurAdj->m_uiEdges[2][0] = pkCurPoly->v[2];
		pkCurAdj->m_uiEdges[2][1] = pkCurPoly->v[0];

		pkCurAdj->m_uiAdjTri[0]   = 0xFFFFFFFF;
		pkCurAdj->m_uiAdjTri[1]   = 0xFFFFFFFF;
		pkCurAdj->m_uiAdjTri[2]   = 0xFFFFFFFF;

		pkCurAdj->m_uiOtherVertex[0] = pkCurPoly->v[2];
		pkCurAdj->m_uiOtherVertex[1] = pkCurPoly->v[0];
		pkCurAdj->m_uiOtherVertex[2] = pkCurPoly->v[1];

		puiCon[ uiPoly ] = uiPoly & 0x3FFFFFFF;
	}

	//Fill neighbour data
	pkCurAdj  = pkAdj;

	AdjacencyData *pkOtherAdj;

	for( uiPoly = 0; uiPoly < m_uiNumCurrent; ++uiPoly, ++pkCurAdj )
	{
		for( int iAdj = 0; iAdj < 3; ++iAdj )
		{
			if( pkCurAdj->m_uiAdjTri[ iAdj ] == 0xFFFFFFFF )
			{
				pkOtherAdj  = pkCurAdj + 1;

				for( unsigned int uiOtherPoly = uiPoly + 1; uiOtherPoly < m_uiNumCurrent; ++uiOtherPoly, ++pkOtherAdj )
				{
					bool bFound = false;

					for( int iOtherAdj = 0; iOtherAdj < 3; ++iOtherAdj )
					{
						if( pkOtherAdj->m_uiAdjTri[ iOtherAdj ] == 0xFFFFFFFF )
						{
							if( ( pkCurAdj->m_uiEdges[iAdj][0] == pkOtherAdj->m_uiEdges[iOtherAdj][1] ) &&
							    ( pkCurAdj->m_uiEdges[iAdj][1] == pkOtherAdj->m_uiEdges[iOtherAdj][0] ) )
							{
								pkCurAdj->m_uiAdjTri[ iAdj ]        = uiOtherPoly;
								pkOtherAdj->m_uiAdjTri[ iOtherAdj ] = uiPoly;

								unsigned char ucLastCon = ( ( puiCon[ uiPoly ] & 0xB0000000 ) >> 30 ) & 0x3;

								assert( ucLastCon < 3 );

								++ucLastCon;

								puiCon[ uiPoly ] = ( puiCon[ uiPoly ] & 0x3FFFFFFF ) | ( ucLastCon << 30 );


								ucLastCon = ( ( puiCon[ uiOtherPoly ] & 0xB0000000 ) >> 30 ) & 0x3;

								assert( ucLastCon < 3 );

								++ucLastCon;

								puiCon[ uiOtherPoly ] = ( puiCon[ uiOtherPoly ] & 0x3FFFFFFF ) | ( ucLastCon << 30 );


								bFound = true;

								break;
							}
						}
					}

					if( bFound )
						break;
				}
			}
		}
	}

	//Sort on connectivity
	RadixSort kSort( RadixSort::UINT32 );

	kSort.Sort( puiCon, m_uiNumCurrent );

	unsigned int *puiPolyIndex = kSort.GetIndices();


	//Create strips
	unsigned int uiCurStart  = 0;
	unsigned int uiNumStrips = 0;
	unsigned int uiPolysLeft = m_uiNumCurrent;

	vector< unsigned short* > vpusStrips;
	vector< unsigned int    > vuiStripLen;
	vector< bool            > vbPadded;

	while( true )
	{
		for( ; uiCurStart < m_uiNumCurrent; ++uiCurStart )
		{		
			uiPoly = puiPolyIndex[ uiCurStart ] & 0x3FFFFFFF;

			if( pucAdded[ uiPoly ] )
				continue;

			//Ok, we got a start poly
			//Now throw out strips in all directions
			unsigned int *puiStrips[3];
			
			puiStrips[0] = new unsigned int[ uiPolysLeft ];
			puiStrips[1] = new unsigned int[ uiPolysLeft ];
			puiStrips[2] = new unsigned int[ uiPolysLeft ];

			unsigned int auiStripLen[3] = { 0, 0, 0 };
			bool abPad[3]               = { false, false, false };

			for( unsigned int uiAdj = 0; uiAdj < 3; ++uiAdj )
			{
				//Copy added flag to local array to let us modify it
				//during loop to avoid iterating back through this local
				//strip's polygons more than once
				fmemcpy( pucLocalAdded[ uiAdj ], pucAdded, m_uiNumCurrent );

				ExtendStrip( uiPoly, uiAdj, pkAdj, puiStrips[ uiAdj ], &auiStripLen[ uiAdj ], pucLocalAdded[ uiAdj ] );
			}


			//Try to extend strips by reversing direction
			for( int iStrip = 0; iStrip < 3; ++iStrip )
			{
				//We must have at least one extra poly in (iStrip+2)%3 strip for iStrip strip to be extendable in reverse directon
				if( ( auiStripLen[ iStrip ] > 1 ) && ( auiStripLen[ ( iStrip + 2 ) % 3 ] > 1 ) )
				{
					unsigned int *puiReverseStrip = new unsigned int[ uiPolysLeft ];

					unsigned int uiStartPoly  = puiStrips[ iStrip ][1];
					unsigned int uiReverseLen = 0;
					unsigned int uiEdge       = 0;

					if( pkAdj[ uiStartPoly ].m_uiAdjTri[0] == uiPoly )
						uiEdge = 0;
					else if( pkAdj[ uiStartPoly ].m_uiAdjTri[1] == uiPoly )
						uiEdge = 1;
					else if( pkAdj[ uiStartPoly ].m_uiAdjTri[2] == uiPoly )
						uiEdge = 2;
					else
						assert( false );

					pucLocalAdded[ iStrip ][ uiStartPoly ] = 0;
					pucLocalAdded[ iStrip ][ uiPoly ]      = 0;

					ExtendStrip( uiStartPoly, uiEdge, pkAdj, puiReverseStrip, &uiReverseLen, pucLocalAdded[ iStrip ] );

					assert( uiReverseLen >= 2 );

					if( uiReverseLen > 2 )
					{
						unsigned int *puiNewStrip = new unsigned int[ uiPolysLeft ];
						unsigned int  uiNewStrip  = 0;
						unsigned int  i;

						//Note: These tests might look strange but are correct if
						//      given some thought about start polygon (reverse
						//      strip starts one polygon _before_ iPoly where normal
						//      strip starts)
						if( ( uiReverseLen % 2 ) && ( auiStripLen[ iStrip ] % 2 ) )
						{
							//Pad with extra void polygon at beginning, merge with reverse part first
							abPad[ iStrip ] = true;

							for( i = uiReverseLen - 1; i > 1; --i )
								puiNewStrip[ uiNewStrip++ ] = puiReverseStrip[i];
						
							for( i = 0; i < auiStripLen[ iStrip ]; ++i )
								puiNewStrip[ uiNewStrip++ ] = puiStrips[ iStrip ][i];

						}
						else if( !( uiReverseLen % 2 ) )
						{
							//Merge with reverse part first, no need to pad with void polygons
							for( i = uiReverseLen - 1; i > 1; --i )
								puiNewStrip[ uiNewStrip++ ] = puiReverseStrip[i];
						
							for( i = 0; i < auiStripLen[ iStrip ]; ++i )
								puiNewStrip[ uiNewStrip++ ] = puiStrips[ iStrip ][i];
						}
						else
						{
							//Invert normal strip and merge with tail first, no need to pad
							for( i = auiStripLen[ iStrip ]; i > 0; --i )
								puiNewStrip[ uiNewStrip++ ] = puiStrips[iStrip][ i - 1 ];
						
							for( i = 2; i < uiReverseLen; ++i )
								puiNewStrip[ uiNewStrip++ ] = puiReverseStrip[i];
						}

						delete [] puiStrips[ iStrip ];

						puiStrips[ iStrip ]   = puiNewStrip;
						auiStripLen[ iStrip ] = uiNewStrip;
					}

					delete [] puiReverseStrip;
				}
			}


			int iLongest = 0;

			if( auiStripLen[1] > auiStripLen[iLongest] )
				iLongest = 1;

			if( auiStripLen[2] > auiStripLen[iLongest] )
				iLongest = 2;

			//Create strip
			unsigned short *pusBuffer = new unsigned short[ auiStripLen[ iLongest ] + 2 + ( abPad[ iLongest ] ? 1 : 0 ) ];
			unsigned short *pusCur    = pusBuffer;

			//Add first polygon, get glue edge at correct position (as edge 1, rotate vertices)
			{
				int iVertices[3] = { 2, 0, 1 };

				if( pkAdj[ puiStrips[ iLongest ][0] ].m_uiAdjTri[1] == puiStrips[ iLongest ][1] )
				{
					iVertices[0] = 0;
					iVertices[1] = 1;
					iVertices[2] = 2;
				}
				else if( pkAdj[ puiStrips[ iLongest ][0] ].m_uiAdjTri[2] == puiStrips[ iLongest ][1] )
				{
					iVertices[0] = 1;
					iVertices[1] = 2;
					iVertices[2] = 0;
				}

				if( abPad[ iLongest ] )
				{
					*pusCur++ = m_pkPolygons[ puiStrips[ iLongest ][0] ].v[ iVertices[0] ];
					*pusCur++ = m_pkPolygons[ puiStrips[ iLongest ][0] ].v[ iVertices[0] ];
					*pusCur++ = m_pkPolygons[ puiStrips[ iLongest ][0] ].v[ iVertices[2] ];
					*pusCur++ = m_pkPolygons[ puiStrips[ iLongest ][0] ].v[ iVertices[1] ];
				}
				else
				{
					*pusCur++ = m_pkPolygons[ puiStrips[ iLongest ][0] ].v[ iVertices[0] ];
					*pusCur++ = m_pkPolygons[ puiStrips[ iLongest ][0] ].v[ iVertices[1] ];
					*pusCur++ = m_pkPolygons[ puiStrips[ iLongest ][0] ].v[ iVertices[2] ];
				}

				++pucAdded[ puiStrips[ iLongest ][0] ];

				assert( pucAdded[ puiStrips[ iLongest ][0] ] == 1 );
			}

			for( unsigned int uiLoopPoly = 1; uiLoopPoly < auiStripLen[ iLongest ]; ++uiLoopPoly )
			{
				//Get which edge is the glue to our current poly
				int iEdge = 0;

				if( pkAdj[ puiStrips[ iLongest ][ uiLoopPoly ] ].m_uiAdjTri[1] == puiStrips[ iLongest ][ uiLoopPoly - 1 ] )
					iEdge = 1;
				else if( pkAdj[ puiStrips[ iLongest ][ uiLoopPoly ] ].m_uiAdjTri[2] == puiStrips[ iLongest ][ uiLoopPoly - 1 ] )
					iEdge = 2;

				*pusCur++ = pkAdj[ puiStrips[ iLongest ][ uiLoopPoly ] ].m_uiOtherVertex[ iEdge ];

				++pucAdded[ puiStrips[ iLongest ][ uiLoopPoly ] ];

				assert( pucAdded[ puiStrips[ iLongest ][ uiLoopPoly ] ] == 1 );
			}

			vpusStrips.push_back( pusBuffer );
			vuiStripLen.push_back( auiStripLen[ iLongest ] + ( abPad[ iLongest ] ? 1 : 0 ) );
			vbPadded.push_back( abPad[ iLongest ] );

			++uiNumStrips;

			delete [] puiStrips[0];
			delete [] puiStrips[1];
			delete [] puiStrips[2];

			uiPolysLeft -= auiStripLen[ iLongest ];
		}

		if( uiCurStart >= m_uiNumCurrent )
			break;
	}

	assert( !uiPolysLeft );

	if( uiNumStrips == 1 )
	{
		//Just create a new accelerated storage strip
		m_pkStrip = Core::Get()->GetRenderDevice()->CreatePolygonStripBuffer( Buffer::STATIC, vuiStripLen[0], vpusStrips[0] );

		delete [] vpusStrips[0];
	}
	else
	{
		//Strip concatenation, like [1,2,3,4] + [5,6,7,8] = [1,2,3,4,4,5,5,6,7,8]
		unsigned int uiNumIndices = 0;

		//Add extra indices for concatenation void indices
		for( uiStrip = 0; uiStrip < uiNumStrips; ++uiStrip )
			uiNumIndices += vuiStripLen[ uiStrip ] + 2 + 4;

		unsigned short *pusFinal = new unsigned short[ uiNumIndices ];
		unsigned short *pusTmp   = new unsigned short[ uiNumIndices ];

		while( vpusStrips.size() > 1 )
		{
			int aiStrip[2]  = { 0, 1 };
			int iOp         = 0;

			if( vuiStripLen[ aiStrip[0] ] & 1 )
			{
				if( vbPadded[ aiStrip[0] ] )
					iOp = 6;
				else
					iOp = 3;
			}
			else
			{
				if( vbPadded[ aiStrip[0] ] )
				{
					assert( !"Padded even strip (first) is not possible!" );
				}
				else
					iOp = 0;
			}

			if( vuiStripLen[ aiStrip[1] ] & 1 )
			{
				if( vbPadded[ aiStrip[1] ] )
					iOp += 2;
				else
					iOp += 1;
			}
			else
			{
				if( vbPadded[ aiStrip[1] ] )
				{
					assert( !"Padded even strip (second) is not possible!" );
				}
				else
					iOp += 0;
			}

			int  iLoop   = 0;
			bool bMerged = false;

			while( g_aaiPolygonStripMergeOp[iOp][iLoop] != -1 )
			{
				switch( g_aaiPolygonStripMergeOp[iOp][iLoop] )
				{
					case POLYGONSTRIPMERGEOP_MERGE:
					{
						unsigned int    uiTotalIndices = vuiStripLen[ aiStrip[0] ] + vuiStripLen[ aiStrip[1] ] + 2 + 2 + 2;
						unsigned short *pusDst         = pusTmp;
						unsigned short *pusSrc         = vpusStrips[ aiStrip[0] ];
						unsigned int    uiSize         = ( vuiStripLen[ aiStrip[0] ] + 2 ) * sizeof( unsigned short );

						//Copy first polygon strip data
						fmemcpy( pusDst, pusSrc, uiSize );

						//Move pointers
						pusDst += vuiStripLen[ aiStrip[0] ] + 2;
						pusSrc += vuiStripLen[ aiStrip[0] ] + 2 - 1;

						//Duplicate last index of first strip
						*pusDst++ = *pusSrc;

						//Duplicate first index of last strip
						pusSrc  = vpusStrips[ aiStrip[1] ];

						*pusDst++ = *pusSrc;

						//Copy last polygons strip data
						uiSize  = ( vuiStripLen[ aiStrip[1] ] + 2 ) * sizeof( unsigned short );

						fmemcpy( pusDst, pusSrc, uiSize );

						//Keep first strip's padded flag
						bool bPadded = vbPadded[ aiStrip[0] ];

						//Copy tmp to final
						fmemcpy( pusFinal, pusTmp, uiTotalIndices * sizeof( unsigned short ) );

						//Delete old strips if not final
						if( vpusStrips[ aiStrip[1] ] != pusFinal )
							delete [] vpusStrips[ aiStrip[1] ];

						if( vpusStrips[ aiStrip[0] ] != pusFinal )
							delete [] vpusStrips[ aiStrip[0] ];

						//Erase second strip and set new ptr
						vpusStrips.erase( vpusStrips.begin() + aiStrip[1] );
						vuiStripLen.erase( vuiStripLen.begin() + aiStrip[1] );
						vbPadded.erase( vbPadded.begin() + aiStrip[1] );

						if( aiStrip[1] < aiStrip[0] )
							--aiStrip[0];

						vpusStrips[  aiStrip[0] ] = pusFinal;
						vuiStripLen[ aiStrip[0] ] = uiTotalIndices - 2;
						vbPadded[    aiStrip[0] ] = bPadded;

						bMerged = true;

						break;
					}

					case POLYGONSTRIPMERGEOP_SWITCH:
					{
						int iTemp = aiStrip[0];

						aiStrip[0] = aiStrip[1];
						aiStrip[1] = iTemp;

						break;
					}

					case POLYGONSTRIPMERGEOP_REVERSELAST:
					{
						unsigned short *pusFirstPtr =  vpusStrips[ aiStrip[1] ];
						unsigned short *pusLastPtr  = &vpusStrips[ aiStrip[1] ][ vuiStripLen[ aiStrip[1] ] + 2 - 1 ];
						unsigned short  usTemp;

						while( pusLastPtr > pusFirstPtr )
						{
							usTemp = *pusFirstPtr;

							*pusFirstPtr = *pusLastPtr;
							*pusLastPtr  = usTemp;

							++pusFirstPtr;
							--pusLastPtr;
						}

						break;
					}

					case POLYGONSTRIPMERGEOP_STRIPPADLAST:
					{
						assert( vbPadded[ aiStrip[1] ] );

						unsigned short *pusFirstPtr =  vpusStrips[ aiStrip[1] ];
						
						unsigned int uiLoopLen = ( vuiStripLen[ aiStrip[1] ] + 1 );

						for( unsigned int uiLoop = 0; uiLoop < uiLoopLen; ++uiLoop )
							*pusFirstPtr++ = *( pusFirstPtr + 1 );

						--vuiStripLen[ aiStrip[1] ];

						vbPadded[ aiStrip[1] ] = false;

						break;
					}

					default:
						assert( !"Unknown op in strip merge" );
				}
		
				++iLoop;
			}

			assert( bMerged );
		}

		m_pkStrip = Core::Get()->GetRenderDevice()->CreatePolygonStripBuffer( Buffer::STATIC, vuiStripLen[0], pusFinal );

		delete [] pusTmp;
		delete [] pusFinal;
	}

#ifdef POLYGONBUFFER_DEBUGSTRIPIFY
	{
		unsigned int uiNumIndices = m_pkStrip->GetNumPolygons() + 2;

		neolog << LogLevel( DEBUG ) << "[PolygonBuffer::Stripify] RESULT: Stripification of [ " << m_uiNumCurrent << " ] polygons, efficiency [ " << int( float( uiNumIndices * 100 ) / float( m_uiNumPolygons * 3 ) ) ) << "% ]" << endl;
	}
#endif

	//Unlock buffer
	if( !bHadLock )
		Unlock();

	//Free memory
	delete [] pucLocalAdded[0];
	delete [] pucLocalAdded[1];
	delete [] pucLocalAdded[2];
	delete [] pucAdded;
	delete [] pkAdj;
	delete [] puiCon;

#ifdef POLYGONBUFFER_REQUIRESTRIPEFF
	{
		//Calculate efficiency
		unsigned int uiTotalPolygons = m_pkStrip->GetNumElements();
		unsigned int uiNumIndices    = uiTotalPolygons + 2;

		float fEff = float( uiNumIndices ) / float( m_uiNumCurrent * 3 );

		if( fEff > POLYGONBUFFER_REQUIRESTRIPEFF )
		{
#ifdef POLYGONBUFFER_DEBUGSTRIPIFY
			neolog << LogLevel( DEBUG ) << "[PolygonBuffer::Stripify] Required rate is " << int( POLYGONBUFFER_REQUIRESTRIPEFF * 100.0f ) << "%, deleting strip" << endl;
#endif

			m_pkStrip = 0;
		}
	}
#endif
}


void PolygonBuffer::ExtendStrip( unsigned int uiPolygon, unsigned int uiEdge, AdjacencyData *pkAdj, unsigned int *puiStrip, unsigned int *puiStripLen, unsigned char *pucAdded )
{
	//Get strip with current triangle rotation
	unsigned int uiLoopPoly    = uiPolygon;
	bool         bSecondEdge   = false;
	unsigned int uiGlueEdge    = uiEdge;
	unsigned int uiThisLen     = 0;

	while( true )
	{
		puiStrip[ uiThisLen++ ] = uiLoopPoly;

		++pucAdded[ uiLoopPoly ]; 

		assert( pucAdded[ uiLoopPoly ] == 1 );

		unsigned int uiNextPoly = pkAdj[ uiLoopPoly ].m_uiAdjTri[ uiGlueEdge ];

		bSecondEdge = !bSecondEdge;

		if( ( uiNextPoly == 0xFFFFFFFF ) || pucAdded[ uiNextPoly ] )
			break;

		//Calculate glue edge for next polygon
		unsigned int uiNextGlueEdge = 0;

		if( pkAdj[ uiNextPoly ].m_uiAdjTri[1] == uiLoopPoly )
			uiNextGlueEdge = 1;
		else if( pkAdj[ uiNextPoly ].m_uiAdjTri[2] == uiLoopPoly )
			uiNextGlueEdge = 2;

		//Next poly is connected along iNextGlueEdge with our polygon, find correct strip-compatible edge for next-next polygon
		uiNextGlueEdge = ( uiNextGlueEdge + ( bSecondEdge ? 2 : 1 ) ) % 3;

		uiLoopPoly     = uiNextPoly;
		uiGlueEdge     = uiNextGlueEdge;
	}

	*puiStripLen += uiThisLen;
}


void PolygonBuffer::CalculateNormals( VertexBufferPtr pkVertexBuffer, bool bNormalize )
{
	if( !pkVertexBuffer )
	{
		pkVertexBuffer = m_pkVertexBuffer;

		if( !pkVertexBuffer )
			return;
	}

	if( !m_pkNormals )
		m_pkNormals = new Vector3d[ m_uiNumCurrent ];

	bool bLock = IsLocked();

	if( !bLock )
		Lock( Buffer::READ );

	Polygon  *pkPoly   = m_pkPolygons;
	Vector3d *pkNormal = m_pkNormals;

	bool bVBLock = pkVertexBuffer->IsLocked();

	if( !bVBLock )
		pkVertexBuffer->Lock( Buffer::READ );

	// ### FIXME: We assume postition is at offset 0
	unsigned char *pucVertex    = (unsigned char*)pkVertexBuffer->GetVertex();
	unsigned int   uiVertexSize = pkVertexBuffer->GetVertexSize();

	for( unsigned int uiPoly = 0; uiPoly < m_uiNumCurrent; ++uiPoly, ++pkPoly, ++pkNormal )
	{
		Vector3d *pkVec0 = (Vector3d*)( pucVertex + uiVertexSize * pkPoly->v[0] );
		Vector3d *pkVec1 = (Vector3d*)( pucVertex + uiVertexSize * pkPoly->v[1] );
		Vector3d *pkVec2 = (Vector3d*)( pucVertex + uiVertexSize * pkPoly->v[2] );

		Vector3d kV0     = ( *pkVec1 - *pkVec0 ) * 10.0f;
		Vector3d kV1     = ( *pkVec2 - *pkVec0 ) * 10.0f;

		Vector3d kNormal = kV0 % kV1;

		if( bNormalize )
			kNormal.Normalize();

		*pkNormal = kNormal;
	}

	if( !bVBLock )
		pkVertexBuffer->Unlock();

	if( !bLock )
		Unlock();
}


void PolygonBuffer::CalculateEdges( VertexBufferPtr pkVertexBuffer )
{
	if( !m_pkVertexBuffer && !pkVertexBuffer )
		return;

	if( !pkVertexBuffer )
		pkVertexBuffer = m_pkVertexBuffer;

	if( !m_pkEdges )
		m_pkEdges = new Edge[ m_uiNumCurrent * 3 ];

	bool bLock = IsLocked();

	if( !bLock )
		Lock( Buffer::READ );

	Polygon  *pkPoly   = m_pkPolygons;
	Edge     *pkEdge   = m_pkEdges;

	unsigned int uiPoly = 0;

	//Add edges for this polygon
	for( ; uiPoly < m_uiNumCurrent; ++uiPoly, ++pkPoly )
	{
		pkEdge->m_usVertices[ 0 ] = pkPoly->v[0];
		pkEdge->m_usVertices[ 1 ] = pkPoly->v[1];

		pkEdge->m_uiPolygon       = uiPoly;

		++pkEdge;

		pkEdge->m_usVertices[ 0 ] = pkPoly->v[1];
		pkEdge->m_usVertices[ 1 ] = pkPoly->v[2];

		pkEdge->m_uiPolygon       = uiPoly;

		++pkEdge;

		pkEdge->m_usVertices[ 0 ] = pkPoly->v[2];
		pkEdge->m_usVertices[ 1 ] = pkPoly->v[0];

		pkEdge->m_uiPolygon       = uiPoly;

		++pkEdge;
	}

	if( !bLock )
		Unlock();

	bool bVBLock = pkVertexBuffer->IsLocked();

	if( !bVBLock )
		pkVertexBuffer->Lock( Buffer::READ );

	// ### FIXME: We assume postition is at offset 0
	unsigned char *pucVertex    = (unsigned char*)pkVertexBuffer->GetVertex();
	unsigned int   uiVertexSize = pkVertexBuffer->GetVertexSize();

	//All edges stored, check neighbours
	for( uiPoly = 0; uiPoly < m_uiNumCurrent; ++uiPoly )
	{
		pkEdge = &m_pkEdges[ uiPoly * 3 ];

		Edge *pkEdge0 = pkEdge++;
		Edge *pkEdge1 = pkEdge++;
		Edge *pkEdge2 = pkEdge++;

		Vector3d &rkPos0Vert0 = *((Vector3d*)( pucVertex + uiVertexSize * pkEdge0->m_usVertices[0] ));
		Vector3d &rkPos0Vert1 = *((Vector3d*)( pucVertex + uiVertexSize * pkEdge0->m_usVertices[1] ));

		Vector3d &rkPos1Vert0 = *((Vector3d*)( pucVertex + uiVertexSize * pkEdge1->m_usVertices[0] ));
		Vector3d &rkPos1Vert1 = *((Vector3d*)( pucVertex + uiVertexSize * pkEdge1->m_usVertices[1] ));

		Vector3d &rkPos2Vert0 = *((Vector3d*)( pucVertex + uiVertexSize * pkEdge2->m_usVertices[0] ));
		Vector3d &rkPos2Vert1 = *((Vector3d*)( pucVertex + uiVertexSize * pkEdge2->m_usVertices[1] ));

		//Check for neighbour edges
		for( unsigned int uiEdge = ( uiPoly + 1 ) * 3; uiEdge < m_uiNumCurrent * 3; ++uiEdge, ++pkEdge )
		{
			Vector3d &rkVert0 = *((Vector3d*)( pucVertex + uiVertexSize * pkEdge->m_usVertices[0] ));
			Vector3d &rkVert1 = *((Vector3d*)( pucVertex + uiVertexSize * pkEdge->m_usVertices[1] ));

			//WE SHOULD NOT ALLOW 0==0 && 1==1 case since that means the polys are winded in the same direction
			//and that means edge will never cast shadow (if normals are similar, both polys will (almost) always
			//be visible at the same time, and no shadow cast. if not similar, the normals are (almost) parallell
			//and diffrent direction, and we can just as well split them)

			if( ( rkVert0 == rkPos0Vert1 ) && ( rkVert1 == rkPos0Vert0 ) )
			{
				pkEdge0->m_pkNeighbour = pkEdge;
				pkEdge->m_pkNeighbour  = pkEdge0;
				break;
			}
			else if( ( rkVert0 == rkPos1Vert1 ) && ( rkVert1 == rkPos1Vert0 ) )
			{
				pkEdge1->m_pkNeighbour = pkEdge;
				pkEdge->m_pkNeighbour  = pkEdge1;
				break;
			}
			else if( ( rkVert0 == rkPos2Vert1 ) && ( rkVert1 == rkPos2Vert0 ) )
			{
				pkEdge2->m_pkNeighbour = pkEdge;
				pkEdge->m_pkNeighbour  = pkEdge2;
				break;
			}
		}
	}

	if( !bVBLock )
		pkVertexBuffer->Unlock();
}


void PolygonBuffer::CalculateTangents( VertexBufferPtr pkVertices, const VertexElement *pkNormalElem, const VertexElement *pkTexCoordElem, const VertexElement *pkTangentElem, const VertexElement *pkBinormalElem )
{
	bool bUnlock = false;
	bool bUnlockVB = false;

	if( !IsLocked() )
	{
		Lock( Buffer::READ );
		bUnlock = true;
	}

	if( !pkVertices->IsLocked() )
	{
		pkVertices->Lock( Buffer::WRITE );
		bUnlockVB = true;
	}

	int iEnd = pkVertices->GetNumElements();
	int iPolyEnd = GetNumElements();
	int i;

	//Calculate tangent vectors
	Vector3d *pkTangent      = new Vector3d[ iEnd ];
	int      *piTangentCount = new int[ iEnd ];

	memset( pkTangent,      0, sizeof( Vector3d ) * iEnd );
	memset( piTangentCount, 0, sizeof( int      ) * iEnd );

	Polygon  *pkPolygon = GetPolygon();
	Vector3d *pkNormal  = GetNormals();

	if( !pkNormal )
	{
		CalculateNormals( pkVertices, true );
		pkNormal = GetNormals();
	}

	unsigned char *pucVertex    = (unsigned char*)pkVertices->GetVertex();
	unsigned int   uiVertexSize = pkVertices->GetVertexSize();

	for( i = 0; i < iPolyEnd; ++i, ++pkPolygon, ++pkNormal )
	{
		unsigned char *apucVertices[3] =  { pucVertex + uiVertexSize * pkPolygon->v[0], pucVertex + uiVertexSize * pkPolygon->v[1], pucVertex + uiVertexSize * pkPolygon->v[2] };

		Vector3d *apkVertices[3]  = { (Vector3d*)apucVertices[0], (Vector3d*)apucVertices[1], (Vector3d*)apucVertices[2] };
		float    *apfTexCoords[3] = { (float*)( apucVertices[0] + pkTexCoordElem->m_uiOffset ), (float*)( apucVertices[1] + pkTexCoordElem->m_uiOffset ), (float*)( apucVertices[2] + pkTexCoordElem->m_uiOffset ) };

		Vector3d kEdge0 = *apkVertices[0] - *apkVertices[1];
		Vector3d kEdge1 = *apkVertices[2] - *apkVertices[0];

		float dv0 = apfTexCoords[0][1] - apfTexCoords[1][1];
		float dv1 = apfTexCoords[2][1] - apfTexCoords[0][1];

		Vector3d kTangent = ( kEdge0 * dv1 ) - ( kEdge1 * dv0 );

		kTangent.Normalize();

		float du0 = apfTexCoords[0][0] - apfTexCoords[1][0];
		float du1 = apfTexCoords[2][0] - apfTexCoords[0][0];

		Vector3d kBinorm = ( kEdge0 * du1 ) - ( kEdge1 * du0 );

		if( ( kTangent % kBinorm ) * (*pkNormal) < 0.0f )
			kTangent = -kTangent;

		pkTangent[ pkPolygon->v[0] ] += kTangent;
		pkTangent[ pkPolygon->v[1] ] += kTangent;
		pkTangent[ pkPolygon->v[2] ] += kTangent;

		piTangentCount[ pkPolygon->v[0] ]++;
		piTangentCount[ pkPolygon->v[1] ]++;
		piTangentCount[ pkPolygon->v[2] ]++;
	}

	pucVertex = (unsigned char*)pkVertices->GetVertex();

	//Renormalize
	if( pkTangentElem->m_uiType == VertexElement::COLOR32 )
	{
		for( i = 0; i < iEnd; ++i, pucVertex += uiVertexSize )
		{
			Vector3d *pkNormal = (Vector3d*)( pucVertex + pkNormalElem->m_uiOffset );

			pkTangent[i] *= 1.0f / float( piTangentCount[i] );

			pkTangent[i]  = ( *pkNormal % pkTangent[i] ) % *pkNormal;

			Vector3d kTangent = pkTangent[i].Normalize();

			//Store as ranged-compressed
			kTangent *= 0.5f;
			kTangent += Vector3d( 0.5f, 0.5f, 0.5f );
			kTangent *= 255.0f;
			
			((Color32*)( pucVertex + pkTangentElem->m_uiOffset ) )->Set( (unsigned char)kTangent.x, (unsigned char)kTangent.y, (unsigned char)kTangent.z );
		}
	}
	else // Assume FLOAT3
	{
		for( i = 0; i < iEnd; ++i, pucVertex += uiVertexSize )
		{
			Vector3d *pkNormal = (Vector3d*)( pucVertex + pkNormalElem->m_uiOffset );

			pkTangent[i] *= 1.0f / float( piTangentCount[i] );

			pkTangent[i]  = ( *pkNormal % pkTangent[i] ) % *pkNormal;

			pkTangent[i].Normalize();

			*((Vector3d*)( pucVertex + pkTangentElem->m_uiOffset ) ) = pkTangent[i];
		}
	}

	if( pkBinormalElem )
	{
		pucVertex = (unsigned char*)pkVertices->GetVertex();

		//Renormalize
		if( pkBinormalElem->m_uiType == VertexElement::COLOR32 )
		{
			for( i = 0; i < iEnd; ++i, pucVertex += uiVertexSize )
			{
				Vector3d *pkNormal = (Vector3d*)( pucVertex + pkNormalElem->m_uiOffset );

				//Store as ranged-compressed
				Vector3d kBinormal = *pkNormal % pkTangent[i];
				kBinormal *= 0.5f;
				kBinormal += Vector3d( 0.5f, 0.5f, 0.5f );
				kBinormal *= 255.0f;
				
				((Color32*)( pucVertex + pkBinormalElem->m_uiOffset ) )->Set( (unsigned char)kBinormal.x, (unsigned char)kBinormal.y, (unsigned char)kBinormal.z );
			}
		}
		else // Assume FLOAT3
		{
			for( i = 0; i < iEnd; ++i, pucVertex += uiVertexSize )
			{
				Vector3d *pkNormal = (Vector3d*)( pucVertex + pkNormalElem->m_uiOffset );

				*((Vector3d*)( pucVertex + pkBinormalElem->m_uiOffset ) ) = *pkNormal % pkTangent[i];
			}
		}
	}

	delete [] pkTangent;
	delete [] piTangentCount;

	if( bUnlock )
		Unlock();

	if( bUnlockVB )
		pkVertices->Unlock();
}


};

