/***************************************************************************
                          profile.cpp  -  Profiler
                             -------------------
    begin                : Sat Aug 31 2002
    copyright            : (C) 2002 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, profile.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "profile.h"
#include "timer.h"
#include "logstream.h"
#include "console.h"

#include <stdio.h>
#include <assert.h>

using namespace std;


namespace NeoEngine
{


ProfileData::ProfileData( const string &rstrName, ProfileData *pkParent ) :
	m_strName( rstrName ),
	m_pkParent( pkParent ),
	m_vpkChildren(),
	m_ulActivated( Timer::GetHeartbeat() ),
	m_ulAccTime(0),
	m_iCount(1),
	m_iLevel( pkParent ? pkParent->m_iLevel + 1 : 0 )
{
}


ProfileData::~ProfileData()
{
	for( unsigned int i = 0; i < m_vpkChildren.size(); ++i )
		delete m_vpkChildren[i];
}


void ProfileData::Dump()
{
	std::string strDst;

	DumpToString( &strDst );

	neolog << LogLevel( DEBUG ) << strDst;
}


void ProfileData::DumpToString( std::string *pstrDst )
{
	static char szData[256];

#if defined(WIN32) && ( _MSC_VER < 1300 )
	//M$VC++ has "not implemented" uint64_t to float conversion, got error C2520
	int iRelPercentage = int( 100.0f * float( (int64_t)m_ulAccTime ) / float( (int64_t)( m_pkParent ? m_pkParent->m_ulAccTime : m_ulAccTime ) ) );
	snprintf( szData, 256, "%8.3f |   %2d%%    |%8d | ", float( (int64_t)m_ulAccTime ) / float( (int64_t)Timer::GetHeartbeatDivisor() ), iRelPercentage, m_iCount );
#else
	int iRelPercentage = int( 100.0f * float( m_ulAccTime ) / float( m_pkParent ? m_pkParent->m_ulAccTime : m_ulAccTime ) );
	snprintf( szData, 256, "%8.3f |   %2d%%    |%8d | ", float( m_ulAccTime ) / float( Timer::GetHeartbeatDivisor() ), iRelPercentage, m_iCount );
#endif

	*pstrDst += szData;

	for( int i = 0; i < m_iLevel; ++i )
		*pstrDst += "  ";

	*pstrDst += m_strName.m_strData;
	*pstrDst += "\n";

	vector< ProfileData* >::iterator ppkNode    = m_vpkChildren.begin();
	vector< ProfileData* >::iterator ppkNodeEnd = m_vpkChildren.end();

	for( ; ppkNode != ppkNodeEnd; ++ppkNode )
		(*ppkNode)->DumpToString( pstrDst );
}


void ProfileData::Reset()
{
	m_ulAccTime = 0;
	m_iCount    = 0;

	vector< ProfileData* >::iterator ppkNode    = m_vpkChildren.begin();
	vector< ProfileData* >::iterator ppkNodeEnd = m_vpkChildren.end();

	for( ; ppkNode != ppkNodeEnd; ++ppkNode )
		(*ppkNode)->Reset();
}






ProfileManager::ProfileManager() :
	Activator(),
	ProfileData( "root" ),
	ConsoleCmdCallback(),
	RenderEntity()
{
	m_pkActiveNode = this;

	Core::Get()->GetConsole()->RegisterCommand( "showprofile",  this );
	Core::Get()->GetConsole()->RegisterCommand( "clearprofile", this );
	Core::Get()->GetConsole()->RegisterCommand( "resetprofile", this );

	Deactivate();
}


ProfileManager::~ProfileManager()
{
	m_pkActiveNode = 0;
}


void ProfileManager::BeginProfile( const HashString &rstrName )
{
	if( !IsActive() )
		return;

	//Search for child node with same name
	vector< ProfileData* >::iterator ppkNode    = m_pkActiveNode->m_vpkChildren.begin();
	vector< ProfileData* >::iterator ppkNodeEnd = m_pkActiveNode->m_vpkChildren.end();

	for( ; ppkNode != ppkNodeEnd; ++ppkNode )
	{
		if( (*ppkNode)->m_strName == rstrName )
		{			
			m_pkActiveNode = (*ppkNode);

			m_pkActiveNode->m_iCount++;
			m_pkActiveNode->m_ulActivated = Timer::GetHeartbeat();
			
			return;
		}
	}
	
	ProfileData *pkNode = new ProfileData( rstrName, m_pkActiveNode );

	m_pkActiveNode->m_vpkChildren.push_back( pkNode );

	m_pkActiveNode = pkNode;
}


void ProfileManager::EndProfile()
{
	if( !IsActive() || ( m_pkActiveNode == this ) )
		return;

	m_pkActiveNode->m_ulAccTime += ( Timer::GetHeartbeat() - m_pkActiveNode->m_ulActivated );

	if( m_pkActiveNode->m_pkParent )
		m_pkActiveNode = m_pkActiveNode->m_pkParent;
}


void ProfileManager::Dump()
{
	assert( m_pkActiveNode == this );

	std::string strDst;

	DumpToString( &strDst );

	neolog << LogLevel( INFO ) << strDst;
}


void ProfileManager::DumpToString( std::string *pstrDst )
{
	//assert( m_pkActiveNode == this );

	*pstrDst += "PROFILE DATA\n";
	*pstrDst += "  Time   |  Rel %  |  Count  |  Name\n";
	*pstrDst += "------------------------------------------------------------\n";

	vector< ProfileData* >::iterator ppkNode    = m_vpkChildren.begin();
	vector< ProfileData* >::iterator ppkNodeEnd = m_vpkChildren.end();

	m_ulAccTime = 0;

	for( ; ppkNode != ppkNodeEnd; ++ppkNode )
		m_ulAccTime += (*ppkNode)->m_ulAccTime;

	ppkNode = m_vpkChildren.begin();

	for( ; ppkNode != ppkNodeEnd; ++ppkNode )
		(*ppkNode)->DumpToString( pstrDst );
}


void ProfileManager::Clear()
{		
	for( unsigned int i = 0; i < m_vpkChildren.size(); ++i )
		delete m_vpkChildren[i];

	m_vpkChildren.clear();

	m_pkActiveNode = this;
}


void ProfileManager::ConsoleCmd( const HashString &rstrCmd, const HashString &rstrArgs )
{
	if( rstrCmd == "showprofile" )
	{
		if( rstrArgs.length() && ( ( rstrArgs.m_strData[0] == '0' ) || ( rstrArgs == "false" ) ) )
			Deactivate();
		else
			Activate();
	}
	else if( rstrCmd == "clearprofile" )
		Clear();
	else if( rstrCmd == "resetprofile" )
		Reset();
}


bool ProfileManager::Render( Frustum *pkFrustum, bool bForce )
{
	if( !RenderEntity::Render( pkFrustum, bForce ) )
		return false;

	string strData;

	DumpToString( &strData );

	Console::GetDefaultFont()->Printf( 10, 50, "%s", strData.c_str() );

	return true;
}


}; // namespace NeoEngine

