/***************************************************************************
                          radix.cpp  -  Radix sorter
                             -------------------
    begin                : Mon May 13 2002
    copyright            : (C) 2002 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, radix.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "radix.h"

#include <string.h>


namespace NeoEngine
{


class RadixSortTypeDesc
{
	public:

		/*! Number of bytes */
		unsigned int        m_uiSize;

		/*! Signed flag */
		bool                m_bSigned;

		/*! Shift bits */
		unsigned int        m_uiShift;
};


RadixSortTypeDesc g_aeRadixTypes[] = {

	{ 4, true,  2 },
	{ 4, false, 2 },
	{ 8, true,  3 },
	{ 8, false, 3 },
	{ 4, true,  2 }

};




RadixSort::RadixSort( RADIXSORTTYPE eType )
{
	m_puiIndices     = 0;
	m_puiIndices2    = 0;
	m_uiCurrentSize  = 0;
	m_uiPreviousSize = 0;
	m_uiTotalCalls   = 0;
	m_uiHits         = 0;
	m_eType          = eType;

	m_puiHistogram   = new unsigned int[ 256 * g_aeRadixTypes[ eType ].m_uiSize ];
	m_puiOffset      = new unsigned int[ 256 ];

	ResetIndices();
}


RadixSort::~RadixSort()
{
	delete [] m_puiIndices2;
	delete [] m_puiIndices;
	delete [] m_puiHistogram;
	delete [] m_puiOffset;
}


void RadixSort::SetType( RADIXSORTTYPE eType )
{
	if( m_eType == eType )
		return;

	delete [] m_puiHistogram;
	delete [] m_puiOffset;

	m_eType          = eType;
	m_puiHistogram   = new unsigned int[ 256 * g_aeRadixTypes[ eType ].m_uiSize ];
	m_puiOffset      = new unsigned int[ 256 ];
}


bool RadixSort::Resize( unsigned int uiNB )
{
	delete [] m_puiIndices2;
	delete [] m_puiIndices;

	m_puiIndices       = new unsigned int[ uiNB ];
	m_puiIndices2      = new unsigned int[ uiNB ];
	m_uiCurrentSize    = uiNB;

	ResetIndices();

	return true;
}


void RadixSort::ResetIndices()
{
	for( unsigned int i = 0; i < m_uiCurrentSize ; ++i )
		m_puiIndices[i] = i;
}


RadixSort& RadixSort::Sort( const void *pInput, unsigned int uiNB )
{
	if( !pInput || !uiNB )
		return *this;

	if( m_eType == FLOAT32 )
		return Sort( (const float*)pInput, uiNB );

	++m_uiTotalCalls;

	//Resize lists if needed
	if( uiNB != m_uiPreviousSize )
	{
		if( uiNB > m_uiCurrentSize )
			Resize( uiNB );
		else
			ResetIndices();

		m_uiPreviousSize = uiNB;
	}

	// Create histograms (counters). Counters for all passes are created in one run.
	// Pros:	read input buffer once instead of [data type size] times
	// Cons:	Histogram is [data type size]*1Kb instead of 1Kb
	if( CreateHistograms( pInput, uiNB ) )
		return *this; //Already sorted

	// Compute #negative values involved if needed
	unsigned int uiNegativeValues = 0;
	if( g_aeRadixTypes[ m_eType ].m_bSigned )
	{
		// An efficient way to compute the number of negatives values we'll have to deal with is simply to sum the 128
		// last values of the last histogram. Last histogram because that's the one for the Most Significant Byte,
		// responsible for the sign. 128 last values because the 128 first ones are related to positive numbers.
		// NOTE: Since the histogram creation takes system byte ordering into account, MSB is always in last histogram
		unsigned int *puiMSB = &m_puiHistogram[ ( g_aeRadixTypes[ m_eType ].m_uiSize - 1 ) << 8 ];
		for( unsigned int i = 128; i < 256; i++ ) uiNegativeValues += puiMSB[i];	// MSB histogram, 128 for negative part
	}

	// Radix sort, j is the pass number ( 0 = LSB )
	// NOTE: Since the histogram creation takes system byte ordering into account, MSB is always in last histogram
	for( unsigned int j = 0; j < g_aeRadixTypes[ m_eType ].m_uiSize; ++j )
	{
		unsigned int *puiCurCount = &m_puiHistogram[ j << 8 ];

		// Sometimes the fourth (negative) pass is skipped because all numbers are negative and the MSB is 0xFF (for example). This is
		// not a problem, numbers are correctly sorted anyway.
		if( puiCurCount[ *( ((unsigned char*)pInput) + j ) ] != uiNB )
		{
			// Should we care about negative values?
			if( ( j != ( g_aeRadixTypes[ m_eType ].m_uiSize - 1 ) ) || !g_aeRadixTypes[ m_eType ].m_bSigned )
			{
				// Here we deal with positive values only
				// Create offsets
				m_puiOffset[0] = 0;

				for( unsigned int i = 1; i < 256; ++i )
					m_puiOffset[i] = m_puiOffset[i-1] + puiCurCount[i-1];
			}
			else
			{
				// This is a special case to correctly handle negative integers. They're sorted in the right order but at the wrong place.
				unsigned int i = 0;

				// Create biased offsets, in order for negative numbers to be sorted as well
				m_puiOffset[0] = uiNegativeValues; /// First positive number takes place after the negative ones
				for( i = 1; i < 128; ++i )
					m_puiOffset[i] = m_puiOffset[i-1] + puiCurCount[i-1];	// 1 to 128 for positive numbers

				// Fixing the wrong place for negative values
				m_puiOffset[128] = 0;
				for( i = 129; i < 256; ++i )
					m_puiOffset[i] = m_puiOffset[i-1] + puiCurCount[i-1];
			}

			// Perform Radix Sort
			unsigned char *pucInputBytes = (unsigned char*)pInput;
			unsigned int  *puiIndices    = m_puiIndices;
			unsigned int  *puiIndicesEnd = &m_puiIndices[ uiNB ];

			pucInputBytes += j;

			while( puiIndices != puiIndicesEnd )
			{
				unsigned int uiID = *puiIndices++;
				m_puiIndices2[ m_puiOffset[ pucInputBytes[ uiID << g_aeRadixTypes[ m_eType ].m_uiShift ] ]++ ] = uiID;
			}

			// Swap pointers for next pass. Valid indices - the most recent ones - are in m_puiIndices after the swap.
			unsigned int *puiTmp = m_puiIndices;
			m_puiIndices  = m_puiIndices2;
			m_puiIndices2 = puiTmp;
		}
	}

	return *this;
}


RadixSort& RadixSort::Sort( const float *pfInput, unsigned int uiNB )
{
	if( !pfInput || !uiNB )
		return *this;

	++m_uiTotalCalls;

	unsigned int *puiInput = (unsigned int*)pfInput;

	if( uiNB != m_uiPreviousSize )
	{
		if( uiNB > m_uiCurrentSize )
			Resize( uiNB );
		else
			ResetIndices();

		m_uiPreviousSize = uiNB;
	}

	// Floating-point values are always supposed to be signed values, so there's only one code path there.
	// Please note the floating point comparison needed for temporal coherence! Although the resulting asm code
	// is dreadful, this is surprisingly not such a performance hit - well, I suppose that's a big one on first
	// generation Pentiums....We can't make comparison on integer representations because, as Chris said, it just
	// wouldn't work with mixed positive/negative values....
	if( CreateHistograms( pfInput, uiNB ) )
		return *this; //Already sorted

	// Compute #negative values involved if needed
	unsigned int uiNegativeValues = 0;

	// An efficient way to compute the number of negatives values we'll have to deal with is simply to sum the 128
	// last values of the last histogram. Last histogram because that's the one for the Most Significant Byte,
	// responsible for the sign. 128 last values because the 128 first ones are related to positive numbers.
	// NOTE: Since the histogram creation takes system byte ordering into account, MSB is always in last histogram
	unsigned int *puiHistogram = &m_puiHistogram[ 768 ];
	for( unsigned int i = 128; i < 256; ++i )
		uiNegativeValues += puiHistogram[i]; // 768 for last histogram, 128 for negative part

	// Radix sort, j is the pass number ( 0 = LSB, 3 = MSB )
	// NOTE: Since the histogram creation takes system byte ordering into account, MSB is always in last histogram
	for( unsigned int j = 0; j < 4; ++j )
	{
		// Should we care about negative values?
		unsigned int *puiCurCount = &m_puiHistogram[ j << 8 ];

		if( j != 3 )
		{
			// Here we deal with positive values only
			if( puiCurCount[ *(((unsigned char*)pfInput) + j ) ] != uiNB )
			{
				// Create offsets
				m_puiOffset[0] = 0;

				for( unsigned int i = 1; i < 256; ++i )
					m_puiOffset[i] = m_puiOffset[i-1] + puiCurCount[i-1];

				// Perform Radix Sort
				unsigned char *pucInputBytes = (unsigned char*)pfInput;
				unsigned int  *puiIndices    = m_puiIndices;
				unsigned int  *puiIndicesEnd = &m_puiIndices[ uiNB ];

				pucInputBytes += j;

				while( puiIndices != puiIndicesEnd )
				{
					unsigned int uiID = *puiIndices++;

					m_puiIndices2[ m_puiOffset[ pucInputBytes[ uiID << 2 ] ]++ ] = uiID;
				}

				// Swap pointers for next pass. Valid indices - the most recent ones - are in mIndices after the swap.
				unsigned int *puiTmp = m_puiIndices;
				m_puiIndices  = m_puiIndices2;
				m_puiIndices2 = puiTmp;
			}
		}
		else
		{
			unsigned char ucUniqueVal = *(((unsigned char*)pfInput) + j );

			// This is a special case to correctly handle negative values
			if( puiCurCount[ ucUniqueVal ] != uiNB )
			{
    			unsigned int i;
	
				// Create biased offsets, in order for negative numbers to be sorted as well
				m_puiOffset[0] = uiNegativeValues; // First positive number takes place after the negative ones
				for( i = 1; i < 128; ++i )
					m_puiOffset[i] = m_puiOffset[i-1] + puiCurCount[i-1];	// 1 to 128 for positive numbers

				// We must reverse the sorting order for negative numbers!
				m_puiOffset[255] = 0;
				for( i = 0; i < 127; ++i )
					m_puiOffset[254-i] = m_puiOffset[255-i] + puiCurCount[255-i];	// Fixing the wrong order for negative values
				for( i = 128; i < 256; ++i )
					m_puiOffset[i] += puiCurCount[i];							// Fixing the wrong place for negative values

				// Perform Radix Sort
				for( i = 0; i < uiNB; ++i )
				{
					unsigned int uiRadix = puiInput[ m_puiIndices[i] ] >> 24; // Radix byte, same as above. AND is useless here (udword).
					// ### cmp to be killed. Not good. Later.
					if( uiRadix < 128 )
						m_puiIndices2[ m_puiOffset[uiRadix]++ ] = m_puiIndices[i];		// Number is positive, same as above
					else
						m_puiIndices2[ --m_puiOffset[uiRadix] ] = m_puiIndices[i];		// Number is negative, flip the sorting order
				}

				// Swap pointers for next pass. Valid indices - the most recent ones - are in mIndices after the swap.
				unsigned int *puiTmp = m_puiIndices;
				m_puiIndices  = m_puiIndices2;
				m_puiIndices2 = puiTmp;
			}
			else
			{
				// The pass is useless, yet we still have to reverse the order of current list if all values are negative.
				if( ucUniqueVal >= 128 )
				{
					for( unsigned int i = 0; i < uiNB; ++i )
						m_puiIndices2[i] = m_puiIndices[ uiNB - i - 1 ];

					// Swap pointers for next pass. Valid indices - the most recent ones - are in mIndices after the swap.
					unsigned int *Tmp = m_puiIndices;
					m_puiIndices  = m_puiIndices2;
					m_puiIndices2 = Tmp;
				}
			}
		}
	}

	return( *this );
}


bool RadixSort::CreateHistograms( const void *pInput, unsigned int uiNB )
{
	memset( m_puiHistogram, 0, 256 * g_aeRadixTypes[ m_eType ].m_uiSize * sizeof( unsigned int ) );

	//Prepare for temporal coherence
	bool           bAlreadySorted = true; //Optimism...

	//Prepare to count
	unsigned char *pucLoop        = (unsigned char*)pInput;
	unsigned char *pucLoopEnd     = &pucLoop[ uiNB * g_aeRadixTypes[ m_eType ].m_uiSize ];
	unsigned int  *puiIndices     = m_puiIndices;

	//Histograms for all passes
	unsigned int *puiHistogram[8];

	for( unsigned int uiH = 0; uiH < g_aeRadixTypes[ m_eType ].m_uiSize; ++uiH )
		puiHistogram[ uiH ] = &m_puiHistogram[ uiH << 8 ];

	//Read input buffer in previous sorted order and
	//check whether already sorted or not
	switch( m_eType )
	{
		case INT32:
		{
			int *piInput = (int*)pInput;

			int iVal;
			int iPrevVal = *piInput;

			while( pucLoop != pucLoopEnd )
			{
				if( ( iVal = piInput[ *puiIndices++ ] ) < iPrevVal ) { bAlreadySorted = false; break; }
				iPrevVal = iVal;

#if NEBYTEORDER_SYSTEM == NEBYTEORDER_LITTLEENDIAN
				++(puiHistogram[0][ *pucLoop++ ]);
				++(puiHistogram[1][ *pucLoop++ ]);
				++(puiHistogram[2][ *pucLoop++ ]);
				++(puiHistogram[3][ *pucLoop++ ]);
#else
				++(puiHistogram[3][ *pucLoop++ ]);
				++(puiHistogram[2][ *pucLoop++ ]);
				++(puiHistogram[1][ *pucLoop++ ]);
				++(puiHistogram[0][ *pucLoop++ ]);
#endif
			}

			break;
		}

		case UINT32:
		{
			unsigned int *puiInput = (unsigned int*)pInput;

			unsigned int uiVal;
			unsigned int uiPrevVal = *puiInput;

			while( pucLoop != pucLoopEnd )
			{
				if( ( uiVal = puiInput[ *puiIndices++ ] ) < uiPrevVal ) { bAlreadySorted = false; break; }
				uiPrevVal = uiVal;

#if NEBYTEORDER_SYSTEM == NEBYTEORDER_LITTLEENDIAN
				++(puiHistogram[0][ *pucLoop++ ]);
				++(puiHistogram[1][ *pucLoop++ ]);
				++(puiHistogram[2][ *pucLoop++ ]);
				++(puiHistogram[3][ *pucLoop++ ]);
#else
				++(puiHistogram[3][ *pucLoop++ ]);
				++(puiHistogram[2][ *pucLoop++ ]);
				++(puiHistogram[1][ *pucLoop++ ]);
				++(puiHistogram[0][ *pucLoop++ ]);
#endif
			}

			break;
		}

		case INT64:
		{
			int64_t *piInput = (int64_t*)pInput;

			int64_t iVal;
			int64_t iPrevVal = *piInput;

			while( pucLoop != pucLoopEnd )
			{
				if( ( iVal = piInput[ *puiIndices++ ] ) < iPrevVal ) { bAlreadySorted = false; break; }
				iPrevVal = iVal;

#if NEBYTEORDER_SYSTEM == NEBYTEORDER_LITTLEENDIAN
				++(puiHistogram[0][ *pucLoop++ ]);
				++(puiHistogram[1][ *pucLoop++ ]);
				++(puiHistogram[2][ *pucLoop++ ]);
				++(puiHistogram[3][ *pucLoop++ ]);
				++(puiHistogram[4][ *pucLoop++ ]);
				++(puiHistogram[5][ *pucLoop++ ]);
				++(puiHistogram[6][ *pucLoop++ ]);
				++(puiHistogram[7][ *pucLoop++ ]);
#else
				++(puiHistogram[7][ *pucLoop++ ]);
				++(puiHistogram[6][ *pucLoop++ ]);
				++(puiHistogram[5][ *pucLoop++ ]);
				++(puiHistogram[4][ *pucLoop++ ]);
				++(puiHistogram[3][ *pucLoop++ ]);
				++(puiHistogram[2][ *pucLoop++ ]);
				++(puiHistogram[1][ *pucLoop++ ]);
				++(puiHistogram[0][ *pucLoop++ ]);
#endif
			}

			break;
		}

		case UINT64:
		{
			uint64_t *puiInput = (uint64_t*)pInput;

			uint64_t uiVal;
			uint64_t uiPrevVal = *puiInput;

			while( pucLoop != pucLoopEnd )
			{
				if( ( uiVal = puiInput[ *puiIndices++ ] ) < uiPrevVal ) { bAlreadySorted = false; break; }
				uiPrevVal = uiVal;

#if NEBYTEORDER_SYSTEM == NEBYTEORDER_LITTLEENDIAN
				++(puiHistogram[0][ *pucLoop++ ]);
				++(puiHistogram[1][ *pucLoop++ ]);
				++(puiHistogram[2][ *pucLoop++ ]);
				++(puiHistogram[3][ *pucLoop++ ]);
				++(puiHistogram[4][ *pucLoop++ ]);
				++(puiHistogram[5][ *pucLoop++ ]);
				++(puiHistogram[6][ *pucLoop++ ]);
				++(puiHistogram[7][ *pucLoop++ ]);
#else
				++(puiHistogram[7][ *pucLoop++ ]);
				++(puiHistogram[6][ *pucLoop++ ]);
				++(puiHistogram[5][ *pucLoop++ ]);
				++(puiHistogram[4][ *pucLoop++ ]);
				++(puiHistogram[3][ *pucLoop++ ]);
				++(puiHistogram[2][ *pucLoop++ ]);
				++(puiHistogram[1][ *pucLoop++ ]);
				++(puiHistogram[0][ *pucLoop++ ]);
#endif
			}

			break;
		}

		case FLOAT32:
		{
			float *pfInput = (float*)pInput;

			float fVal;
			float fPrevVal = *pfInput;

			while( pucLoop != pucLoopEnd )
			{
				if( ( fVal = pfInput[ *puiIndices++ ] ) < fPrevVal ) { bAlreadySorted = false; break; }
				fPrevVal = fVal;

#if NEBYTEORDER_SYSTEM == NEBYTEORDER_LITTLEENDIAN
				++(puiHistogram[0][ *pucLoop++ ]);
				++(puiHistogram[1][ *pucLoop++ ]);
				++(puiHistogram[2][ *pucLoop++ ]);
				++(puiHistogram[3][ *pucLoop++ ]);
#else
				++(puiHistogram[3][ *pucLoop++ ]);
				++(puiHistogram[2][ *pucLoop++ ]);
				++(puiHistogram[1][ *pucLoop++ ]);
				++(puiHistogram[0][ *pucLoop++ ]);
#endif
			}

			break;
		}

	}

	//If all input values are already sorted, we just have to return
	//and leave the previous list unchanged. That way the routine may
	//take advantage of temporal coherence, for example when used to
	//sort transparent faces
	if( bAlreadySorted )
	{
		m_uiHits++;
		return true;
	}

	//Else there has been an early out and we must finish computing
	switch( g_aeRadixTypes[ m_eType ].m_uiSize )
	{
		case 4:
		{
			while( pucLoop != pucLoopEnd )
			{
				//Create histograms without the previous overhead
#if NEBYTEORDER_SYSTEM == NEBYTEORDER_LITTLEENDIAN
				++(puiHistogram[0][ *pucLoop++ ]);
				++(puiHistogram[1][ *pucLoop++ ]);
				++(puiHistogram[2][ *pucLoop++ ]);
				++(puiHistogram[3][ *pucLoop++ ]);
#else
				++(puiHistogram[3][ *pucLoop++ ]);
				++(puiHistogram[2][ *pucLoop++ ]);
				++(puiHistogram[1][ *pucLoop++ ]);
				++(puiHistogram[0][ *pucLoop++ ]);
#endif
			}

			break;
		}

		case 8:
		{
			while( pucLoop != pucLoopEnd )
			{
				//Create histograms without the previous overhead
#if NEBYTEORDER_SYSTEM == NEBYTEORDER_LITTLEENDIAN
				++(puiHistogram[0][ *pucLoop++ ]);
				++(puiHistogram[1][ *pucLoop++ ]);
				++(puiHistogram[2][ *pucLoop++ ]);
				++(puiHistogram[3][ *pucLoop++ ]);
				++(puiHistogram[4][ *pucLoop++ ]);
				++(puiHistogram[5][ *pucLoop++ ]);
				++(puiHistogram[6][ *pucLoop++ ]);
				++(puiHistogram[7][ *pucLoop++ ]);
#else
				++(puiHistogram[7][ *pucLoop++ ]);
				++(puiHistogram[6][ *pucLoop++ ]);
				++(puiHistogram[5][ *pucLoop++ ]);
				++(puiHistogram[4][ *pucLoop++ ]);
				++(puiHistogram[3][ *pucLoop++ ]);
				++(puiHistogram[2][ *pucLoop++ ]);
				++(puiHistogram[1][ *pucLoop++ ]);
				++(puiHistogram[0][ *pucLoop++ ]);
#endif
			}

			break;
		}
	}

	return false;
}


unsigned int RadixSort::GetUsedMemory() const
{
	unsigned int uiUsedRam = sizeof( RadixSort );

	uiUsedRam += 2 * m_uiCurrentSize * sizeof(unsigned int);

	return uiUsedRam;
}


}; // namespace NeoEngine
