/***************************************************************************
           render.cpp  -  Render device base class and module loader
                             -------------------
    begin                : Wed Oct 31 2001
    copyright            : (C) 2001 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, render.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2001
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "render.h"
#include "renderentity.h"
#include "renderadapter.h"
#include "core.h"
#include "config.h"
#include "console.h"
#include "module.h"
#include "shadow.h"
#include "font.h"
#include "light.h"

using namespace std;


namespace NeoEngine
{


unsigned int RenderEntity::s_uiFrameCount = 1;


bool RenderResolution::operator < ( const RenderResolution &rkRes ) const
{
	return(   ( m_uiWidth <  rkRes.m_uiWidth ) ||
	        ( ( m_uiWidth == rkRes.m_uiWidth ) && ( m_uiHeight <  rkRes.m_uiHeight ) ) ||
	        ( ( m_uiWidth == rkRes.m_uiWidth ) && ( m_uiHeight == rkRes.m_uiHeight ) && ( m_uiBPP <  rkRes.m_uiBPP ) ) ||
	        ( ( m_uiWidth == rkRes.m_uiWidth ) && ( m_uiHeight == rkRes.m_uiHeight ) && ( m_uiBPP == rkRes.m_uiBPP ) && ( m_uiRefreshRate < rkRes.m_uiRefreshRate ) ) );
}



RenderDevice::RenderDevice( FileManager *pkFileManager, InputManager *pkInputManager ) :
	InputDevice( pkInputManager ? pkInputManager : Core::Get()->GetInputManager() ),
	TextureLoader(),
	ConfigCallback(),
	m_pkTexturePool( new TexturePool ),
	m_pkProjection( 0 ),
	m_eCullMode( CULLBACK ),
	m_uiDefaultTextureFilter( Texture::BILINEAR ),
	m_pkFileManager( pkFileManager ? pkFileManager : Core::Get()->GetFileManager() ),
	m_pkShadowGenerator( 0 )
{
	if( Core::Get()->GetConfig() )
		Core::Get()->GetConfig()->RegisterCallback( this );
}


RenderDevice::~RenderDevice()
{
	if( Core::Get()->GetConfig() )
		Core::Get()->GetConfig()->UnregisterCallback( this );

	delete m_pkTexturePool;
	delete m_pkShadowGenerator;
}

void RenderDevice::SetDefaultTextureFiltering( unsigned int uiMode )
{
	m_uiDefaultTextureFilter = uiMode;
}

void RenderDevice::SetDefaultMaxAnisotropy( unsigned int uiMaxAnisotropy )
{
	m_uiDefaultMaxAnisotropy = uiMaxAnisotropy;
}

void RenderDevice::SetPerspectiveProjection()
{
	if( m_pkProjection != (Projection*)&m_kPerspectiveProjection )
	{
		m_pkProjection = &m_kPerspectiveProjection;

		SetProjection( m_pkProjection->m_kMatrix );
	}
}


void RenderDevice::SetPerspectiveProjection( float fFOV, float fZNear, float fZFar )
{
	m_kPerspectiveProjection.m_fFOV   = fFOV;
	m_kPerspectiveProjection.m_fZNear = fZNear;
	m_kPerspectiveProjection.m_fZFar  = fZFar;

	m_pkProjection = &m_kPerspectiveProjection;

	m_pkProjection->BuildProjection( m_kViewport );

	SetProjection( m_pkProjection->m_kMatrix );
}


void RenderDevice::SetOrthographicProjection()
{
	if( m_pkProjection != (Projection*)&m_kOrthographicProjection )
	{
		m_pkProjection = &m_kOrthographicProjection;

		SetProjection( m_pkProjection->m_kMatrix );
	}
}


void RenderDevice::SetOrthographicProjection( float fLeft, float fTop, float fRight, float fBottom, float fZNear, float fZFar )
{
	m_kOrthographicProjection.m_fLeft   = fLeft;
	m_kOrthographicProjection.m_fTop    = fTop;
	m_kOrthographicProjection.m_fRight  = fRight;
	m_kOrthographicProjection.m_fBottom = fBottom;
	m_kOrthographicProjection.m_fZNear  = fZNear;
	m_kOrthographicProjection.m_fZFar   = fZFar;

	m_pkProjection = &m_kOrthographicProjection;

	m_pkProjection->BuildProjection( m_kViewport );

	SetProjection( m_pkProjection->m_kMatrix );
}


void RenderDevice::SetCustomProjection( const Matrix &rkMatrix )
{
	m_kCustomProjection.m_kMatrix = rkMatrix;

	m_pkProjection = &m_kCustomProjection;

	SetProjection( m_pkProjection->m_kMatrix );
}


RenderDevice *Core::CreateRenderDevice( const std::string &rstrName, FileManager *pkFileManager, InputManager *pkInputManager )
{
#if defined(WIN32) && defined(_DEBUG)
	ModulePtr pkModule = s_pkModuleManager->LoadModule( "neodev" + rstrName + "_debug" );
#else
	ModulePtr pkModule = s_pkModuleManager->LoadModule( "neodev" + rstrName );
#endif

	if( !pkModule || !pkModule->IsValid() )
		return false;

	ModSymbol kSymInfo;
	ModSymbol kSymLoad;

	kSymInfo = pkModule->LookupSymbol( "GetVersion" );

	if( !kSymInfo )
		neolog << LogLevel( WARNING ) << "*** render device module [" << rstrName << "] does not export GetVersion symbol" << endl;
	else
	{
		string strName;
		int iMajor;
		int iMinor;
		int iRevision;

		(( void ( NE_CDECL* )( std::string*, int*, int*, int* ))kSymInfo)( &strName, &iMajor, &iMinor, &iRevision );

		neolog << LogLevel( INFO ) << "Creating render device " << rstrName << " version " << iMajor << "." << iMinor << "." << iRevision << endl;
	}

	kSymLoad = pkModule->LookupSymbol( "CreateDevice" );

	if( !kSymLoad )
	{
		neolog << LogLevel( ERROR ) << "*** render device module [" << rstrName << "] does not export CreateDevice method" << endl;
		return 0;
	}

	RenderDevice *pkDevice = (( RenderDevice* ( NE_CDECL* )( FileManager*, InputManager* ))kSymLoad)( pkFileManager, pkInputManager );

	if( pkDevice )
	{
		m_pkRenderDevice = pkDevice;
		m_pkRenderDevice->m_pkModule = pkModule;
	}

	return pkDevice;
}


void Core::DeleteRenderDevice( RenderDevice *pkDevice )
{
	if( pkDevice != m_pkRenderDevice )
	{
		neolog << LogLevel( ERROR ) <<  "*** trying to delete non-current render device" << endl;
		return;
	}

	if( m_pkRenderDevice )
	{
		delete m_pkRenderDevice->m_pkShadowGenerator, m_pkRenderDevice->m_pkShadowGenerator = 0;

		if( m_pkConsole )
		{
			//Make sure font and vertex/polygon buffers gets deleted prior to
			//deleting device and unloading module
			m_pkConsole->SetFont( 0 );
			m_pkConsole->UnloadDefaultFont();
			m_pkConsole->DeleteBuffers();
		}

		//By creating new reference to module, it won't get deleted inside render device dtor,
		//since that is a Bad Thing<tm> - the dtor is inside module space
		ModulePtr pkModule = m_pkRenderDevice->m_pkModule;

		m_pkRenderDevice = 0;
	
		delete pkDevice;
	}
}


RenderDevice *Core::SetRenderDevice( RenderDevice *pkDevice )
{
	RenderDevice *pkOld = m_pkRenderDevice;

	m_pkRenderDevice = pkDevice;

	return( pkOld );
}


}; // namespace NeoEngine
