/***************************************************************************
                   rigidbody.cpp  -  Rigid body physics node
                             -------------------
    begin                : Sat Jan 3 2004
    copyright            : (C) 2004 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, rigidbody.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2004
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "rigidbody.h"

using namespace std;


namespace NeoEngine
{


RigidBody::RigidBody() :
	MassParticle(),
	m_fRestitutionCoeff( 0.5f ),
	m_fFrictionCoeff( 0.8f ),
	m_bManaged( false ),
	m_bResting( false )
{
}


RigidBody::~RigidBody()
{
}


SceneNode *RigidBody::Duplicate()
{
	assert( !"[RigidBody::Duplicate] Not implemented!" );
	return 0;
}


const Vector3d &RigidBody::ApplyForce( const Vector3d &rkForce, const Vector3d &rkPos, bool bLocal )
{
	m_bResting = false;

	if( bLocal )
	{
		Vector3d kWorldForce = m_kRotationMatrix * rkForce;

		m_kTorque += ( m_kRotationMatrix * rkPos ) % kWorldForce;

		return( m_kForce += kWorldForce );
	}
	else
	{
		m_kTorque += ( rkPos - GetWorldTranslation() ) % rkForce;

		return( m_kForce += rkForce );
	}
}


void RigidBody::ApplyImpulse( const Vector3d &rkImpulse, const Vector3d &rkPos, bool bLocal )
{
	m_bResting = false;

	if( bLocal )
	{
		Vector3d kWorldImpulse = m_kRotationMatrix * rkImpulse;

		m_kAngularVelocity += m_kInvInertiaTensor * ( ( m_kRotationMatrix * rkPos ) % kWorldImpulse );
		m_kVelocity        += kWorldImpulse * m_fInvMass;
	}
	else
	{
		m_kAngularVelocity += m_kInvInertiaTensor * ( ( rkPos - GetWorldTranslation() ) % rkImpulse );
		m_kVelocity        += rkImpulse * m_fInvMass;
	}
}


void RigidBody::Update( float fDeltaTime )
{
	if( m_bManaged )
	{
		SceneNode::Update( fDeltaTime );
		return;
	}

	// Updates the translation and velocity
	MassParticle::Update( fDeltaTime );

	m_kAngularVelocity += m_kInvInertiaTensor * ( m_kTorque * fDeltaTime );

	m_kRotation = ( m_kRotation + ( ( Quaternion( m_kAngularVelocity.x, m_kAngularVelocity.y, m_kAngularVelocity.z, 0.0f ) * m_kRotation ) * ( 0.5f * fDeltaTime ) ) ).Normalize();

	m_kRotation.ToMatrix( &m_kRotationMatrix );
	m_kInvInertiaTensor = ( m_kRotationMatrix * m_kInvBodyInertiaTensor );

	Matrix kRotTranspose( m_kRotationMatrix ); kRotTranspose.Transpose();
	m_kInvInertiaTensor = ( m_kInvInertiaTensor * kRotTranspose );
}


void RigidBody::ResetTorque()
{
	m_kTorque.Reset();
}


void RigidBody::SetAngularVelocity( const Vector3d &rkVelocity )
{
	m_kAngularVelocity = rkVelocity;
}


Vector3d RigidBody::GetPointVelocity( const Vector3d &rkPoint, bool bLocal )
{
	if( bLocal )
		return( ( m_kAngularVelocity % ( m_kRotationMatrix * rkPoint ) ) + m_kVelocity );
	else
		return( ( m_kAngularVelocity % ( rkPoint - GetWorldTranslation() ) ) + m_kVelocity );
}


void RigidBody::SetMassiveBlock( const Vector3d &rkDim, float fMass )
{
	SetMass( fMass );

	m_kInvBodyInertiaTensor[0][0] = 1.0f / ( fMass * ( rkDim.y * rkDim.y + rkDim.z * rkDim.z ) / 12.0f );
	m_kInvBodyInertiaTensor[1][1] = 1.0f / ( fMass * ( rkDim.x * rkDim.x + rkDim.z * rkDim.z ) / 12.0f );
	m_kInvBodyInertiaTensor[2][2] = 1.0f / ( fMass * ( rkDim.x * rkDim.x + rkDim.y * rkDim.y ) / 12.0f );

	m_kInvInertiaTensor = ( m_kRotationMatrix * m_kInvBodyInertiaTensor );

	Matrix kRotTranspose( m_kRotationMatrix ); kRotTranspose.Transpose();
	m_kInvInertiaTensor = ( m_kInvInertiaTensor * kRotTranspose );
}


void RigidBody::SetInvBodyInertiaTensor( const Matrix &rkInvBodyInertiaTensor )
{
	m_kInvBodyInertiaTensor = rkInvBodyInertiaTensor;

	m_kInvInertiaTensor = ( m_kRotationMatrix * m_kInvBodyInertiaTensor );

	Matrix kRotTranspose( m_kRotationMatrix ); kRotTranspose.Transpose();
	m_kInvInertiaTensor = ( m_kInvInertiaTensor * kRotTranspose );
}


void RigidBody::PushState()
{
	MassParticle::PushState();

	m_vkStateStack.insert( m_vkStateStack.begin(), RigidBodyState( m_kInvBodyInertiaTensor, m_kInvInertiaTensor, m_kAngularVelocity, m_kTorque, m_kRotationMatrix ) );
}


void RigidBody::PopState( bool bSet )
{
	MassParticle::PopState( bSet );

	if( m_vkStateStack.size() )
	{
		vector< RigidBodyState >::iterator pkState = m_vkStateStack.begin();

		if( bSet )
		{
			m_kInvBodyInertiaTensor = pkState->m_kInvBodyInertiaTensor;
			m_kInvInertiaTensor     = pkState->m_kInvInertiaTensor;
			m_kAngularVelocity      = pkState->m_kAngularVelocity;
			m_kTorque               = pkState->m_kTorque;
			m_kRotationMatrix       = pkState->m_kRotation;
		}

		m_vkStateStack.erase( pkState );
	}
}


void RigidBody::ClearState()
{
	MassParticle::ClearState();

	m_vkStateStack.clear();
}


};
