/***************************************************************************
                          room.cpp  -  description
                             -------------------
    begin                : Tue Oct 30 2001
    copyright            : (C) 2001 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, room.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2001
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "room.h"
#include "logstream.h"

using namespace std;


namespace NeoEngine
{


Room::Room( const HashString &rstrName ) :
	SceneNode( rstrName )
{
}


Room::~Room()
{
	vector< SceneNode* >::iterator ppkNode    = m_vpkGlobalNodes.begin();
	vector< SceneNode* >::iterator ppkNodeEnd = m_vpkGlobalNodes.end();

	for( ; ppkNode != ppkNodeEnd; ++ppkNode )
		delete( *ppkNode );
}


int Room::AttachGlobalNode( SceneNode *pkNode )
{
	pkNode->m_pkRoom = this;

	vector<SceneNode*>::iterator ppkNode = m_vpkGlobalNodes.begin();
	vector<SceneNode*>::iterator ppkEnd  = m_vpkGlobalNodes.end();

	for( ; ppkNode != ppkEnd; ++ppkNode )
		if( *ppkNode == pkNode )
			return 0;

	m_vpkGlobalNodes.push_back( pkNode );

	return m_vpkGlobalNodes.size();
}


int Room::DetachGlobalNode( SceneNode *pkNode )
{
	if( pkNode->m_pkRoom == this )
		pkNode->m_pkRoom = 0;

	vector<SceneNode*>::iterator ppkNode = m_vpkGlobalNodes.begin();
	vector<SceneNode*>::iterator ppkEnd  = m_vpkGlobalNodes.end();

	for( ; ppkNode != ppkEnd; ++ppkNode )
		if( *ppkNode == pkNode )
		{
			m_vpkGlobalNodes.erase( ppkNode );
			return m_vpkGlobalNodes.size();
		}

	return 0;
}


int Room::AttachNode( SceneNode *pkNode, bool bKeepWorldSRT )
{
	pkNode->m_pkRoom = this;

	return SceneNode::AttachNode( pkNode, bKeepWorldSRT );
}


int Room::DetachNode( SceneNode *pkNode )
{
	if( pkNode->m_pkRoom == this )
		pkNode->m_pkRoom = 0;

	return SceneNode::DetachNode( pkNode );
}


bool Room::Render( Frustum *pkFrustum, bool bForce )
{
	if( !SceneNode::Render( pkFrustum, bForce ) )
		return false;

	vector<SceneNode*>::iterator ppkNode = m_vpkGlobalNodes.begin();
	vector<SceneNode*>::iterator ppkEnd  = m_vpkGlobalNodes.end();

	for( ; ppkNode != ppkEnd; ++ppkNode )
		(*ppkNode)->Render( pkFrustum, bForce );

	return true;
}


void Room::Update( float fDeltaTime )
{
	vector<SceneNode*>::iterator ppkNode = m_vpkGlobalNodes.begin();
	vector<SceneNode*>::iterator ppkEnd  = m_vpkGlobalNodes.end();

	for( ; ppkNode != ppkEnd; ++ppkNode )
		(*ppkNode)->Update( fDeltaTime );

	SceneNode::Update( fDeltaTime );
}


void Room::AddGeometry( const PolygonBufferPtr &pkPolygonBuffer, const VertexBufferPtr &pkVertexBuffer, const MaterialPtr &pkMaterial )
{
	neolog << LogLevel( PANIC ) << "[Room::AddGeometry] this room implementation does not allow geometry to be added in runtime" << endl;
	assert( false );
}


void Room::SliceGeometry( const NeoEngine::Vector3d &rkV0, const NeoEngine::Vector3d &rkV1, const NeoEngine::Vector3d &rkV2, std::vector< NeoEngine::Vector3d > *pvkPoints )
{
}


RoomManager::RoomManager()
{
}


RoomManager::~RoomManager()
{
}


ModulePtr RoomManager::LoadModule( const HashString &rstrName )
{
#if defined(WIN32) && defined(_DEBUG)
	HashString strModuleName = "neo" + rstrName.m_strData + "_debug";
#else
	HashString strModuleName = "neo" + rstrName.m_strData;
#endif

	vector< ModulePtr >::iterator ppkModule = m_vpkModules.begin();
	vector< ModulePtr >::iterator ppkEnd    = m_vpkModules.end();

	for( ; ppkModule != ppkEnd; ++ppkModule )
	{
		if( (*ppkModule)->GetName() == strModuleName )
			return( *ppkModule );
	}

	ModulePtr pkModule = Core::Get()->GetModuleManager()->LoadModule( strModuleName );

	if( pkModule )
	{
		ModSymbol kSymInitialize;

		kSymInitialize = pkModule->LookupSymbol( "Initialize" );

		if( !kSymInitialize )
			neolog << LogLevel( WARNING ) << "[RoomManager::LoadModule] room module [" << rstrName << "] does not export Initialize method" << endl;
		else
			(( void ( NE_CDECL* )() )kSymInitialize)();

		m_vpkModules.push_back( pkModule );
	}
	else
		neolog << LogLevel( ERROR ) << "*** unable to load room module [" << rstrName << "]" << endl;

	return pkModule;
}


Room *RoomManager::CreateRoom( const HashString &rstrType )
{
	ModulePtr pkModule = LoadModule( rstrType );

	if( !pkModule )
	{
		neolog << LogLevel( ERROR ) << "[RoomManager::CreateRoom] no module available for room type [" << rstrType << "]" << endl;
		return 0;
	}

	ModSymbol kSymCreate = pkModule->LookupSymbol( "CreateRoom" );

	if( !kSymCreate )
	{
		neolog << LogLevel( ERROR ) << "[RoomManager::CreateRoom] room module [" << pkModule->GetName() << "] does not export CreateRoom method" << endl;
		return 0;
	}

	return( (( Room* ( NE_CDECL* )() )kSymCreate)() );
}


}; // namespace NeoEngine

