/***************************************************************************
       scenenode.cpp  -  Base class for hierarchies in a scene graph
                             -------------------
    begin                : Sat Jan 25 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, scenenode.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "scenenode.h"
#include "sceneentity.h"
#include "aabb.h"
#include "obb.h"
#include "sphere.h"
#include "capsule.h"
#include "frustum.h"
#include "logstream.h"

#include <algorithm>

using namespace std;


namespace NeoEngine
{


SceneNode::SceneNode( const HashString &rstrName, SceneNode *pkParent, BoundingVolume::BOUNDINGVOLUMETYPE eVolume ) :
	UpdateEntity(),
	RenderEntity(),
	HierarchyNode<SceneNode>( rstrName, pkParent ),
	m_pkEntity( 0 ),
	m_bNeedVolumeUpdate( false ),
	m_bIgnoreVolumeUpdate( false ),
	m_bIgnoreVolume( false )
{
	if( eVolume == BoundingVolume::BV_AABB )
		m_pkBoundingVolume = new AABB;
	else if( eVolume == BoundingVolume::BV_OBB )
		m_pkBoundingVolume = new OBB;
	else if( eVolume == BoundingVolume::BV_SPHERE )
		m_pkBoundingVolume = new Sphere;
	else if( eVolume == BoundingVolume::BV_CAPSULE )
		m_pkBoundingVolume = new Capsule;
	else
		m_pkBoundingVolume = 0;

}


SceneNode::SceneNode( SceneNode &rkNode, bool bDuplicateChildren ) :
	UpdateEntity(),
	RenderEntity(),
	HierarchyNode<SceneNode>( rkNode, bDuplicateChildren ),
	m_pkEntity( rkNode.m_pkEntity ? rkNode.m_pkEntity->Duplicate() : 0 ),
	m_pkBoundingVolume( rkNode.GetBoundingVolume() ? rkNode.GetBoundingVolume()->Duplicate() : 0 ),
	m_bNeedVolumeUpdate( rkNode.m_pkEntity ? true : false ),
	m_bIgnoreVolumeUpdate( false ),
	m_bIgnoreVolume( rkNode.m_bIgnoreVolume )
{
	if( m_pkEntity )
		m_pkEntity->m_pkNode   = this;
}


SceneNode::~SceneNode()
{
	std::vector< SceneNodeCallback* >::iterator ppkEnd = m_vpkCallbacks.end();
	for( std::vector< SceneNodeCallback* >::iterator ppkCallback = m_vpkCallbacks.begin(); ppkCallback != ppkEnd; ++ppkCallback )
		(*ppkCallback)->NodeDeleted( this );

	delete m_pkBoundingVolume;
	delete m_pkEntity;
}


bool SceneNode::UpdateWorld()
{
	if( HierarchyNode<SceneNode>::UpdateWorld() )
	{
		//World data is now updated, force update of volume
		NotifyVolumeUpdate( true );

		if( m_pkEntity )
		{
			BoundingVolume *pkEntityVolume = m_pkEntity->GetBoundingVolume();

			if( pkEntityVolume )
			{
				pkEntityVolume->SetRotation( GetWorldRotation() );
				pkEntityVolume->SetTranslation( GetWorldTranslation() );
			}
		}

		return true;
	}

	return false;
}


bool SceneNode::UpdateVolume()
{
	if( m_bNeedVolumeUpdate )
	{
		if( !( m_pkEntity || m_vpkChildren.size() ) )
		{
			m_pkBoundingVolume->SetTranslation( GetWorldTranslation() );
			m_pkBoundingVolume->SetRotation( GetWorldRotation() );

			return true;
		}

		bool bGenerated       = false;
		
		vector< SceneNode* >::iterator ppkChild    = m_vpkChildren.begin();
		vector< SceneNode* >::iterator ppkChildEnd = m_vpkChildren.end();

		//Set flag to ignore child requests to update volume
		m_bIgnoreVolumeUpdate = true;

		BoundingVolume *pkEntityVolume = 0;

		if( m_pkEntity && ( pkEntityVolume = m_pkEntity->GetBoundingVolume() ) )
		{
			m_pkBoundingVolume->Generate( pkEntityVolume );

			//Entity bounding volume is in local coordinate system, transform
			//m_pkBoundingVolume->TranslateWorld( GetWorldTranslation() );
			//m_pkBoundingVolume->RotateWorld( GetWorldRotation() );

			bGenerated = true;
		}

		//Node bounding volume is in world coordinate system
		//Ignore flag makes it safe to get child bounding volumes

		if( !bGenerated && ( ppkChild != ppkChildEnd ) )
		{
			if( !(*ppkChild)->m_bIgnoreVolume )
				m_pkBoundingVolume->Generate( (*ppkChild)->GetBoundingVolume() );
			++ppkChild;
		}

		for( ; ppkChild != ppkChildEnd; ++ppkChild )
			if( !(*ppkChild)->m_bIgnoreVolume )
				m_pkBoundingVolume->Merge( (*ppkChild)->GetBoundingVolume() );


		//Reset flags
		m_bNeedVolumeUpdate   = false;
		m_bIgnoreVolumeUpdate = false;


		//Set SRT data on the entity bounding volume
		if( pkEntityVolume )
		{
			pkEntityVolume->SetRotation( GetWorldRotation() );
			pkEntityVolume->SetTranslation( GetWorldTranslation() );
		}

		return true;
	}

	return false;
}


void SceneNode::Update( float fDeltaTime )
{
	if( m_pkEntity && m_pkEntity->IsActive() )
		m_pkEntity->Update( fDeltaTime );

	unsigned int uiEnd = m_vpkChildren.size();
	vector<SceneNode*>::iterator ppkNode = m_vpkChildren.begin();

	for( unsigned int i = 0; i < uiEnd; ++i, ++ppkNode )
		if( (*ppkNode)->IsActive() )
			(*ppkNode)->Update( fDeltaTime );
}


bool SceneNode::Render( Frustum *pkFrustum, bool bForce )
{
	if( !bForce && !RenderEntity::Render( 0, false ) )
		return false; //We are already rendered

	vector< SceneNode* >::iterator ppkNode    = m_vpkChildren.begin();
	vector< SceneNode* >::iterator ppkNodeEnd = m_vpkChildren.end();

	//If no child nodes, this node bounding volume will be equal (roughly) to the entity bounding volume, don't re-test
	if( m_pkEntity && m_pkEntity->IsActive() && ( !pkFrustum || ( ppkNode == ppkNodeEnd ) || !m_pkEntity->GetBoundingVolume() || m_pkEntity->GetBoundingVolume()->Intersection( pkFrustum ) ) )
		m_pkEntity->Render( pkFrustum, bForce );

	for( ; ppkNode != ppkNodeEnd; ++ppkNode )
	{
		if( (*ppkNode)->IsActive() )
		{
			BoundingVolume *pkVolume = (*ppkNode)->GetBoundingVolume();

			//Bounding volume transformation is in world coordinates, so this is safe
			if( !pkVolume || !pkFrustum || pkVolume->Intersection( pkFrustum ) )
				(*ppkNode)->Render( pkFrustum );
		}
	}

	return true;
}


void SceneNode::SetBoundingVolume( BoundingVolume *pkVolume )
{
	delete m_pkBoundingVolume;

	m_pkBoundingVolume = pkVolume;

	m_bNeedVolumeUpdate = true;
}


void SceneNode::SetEntity( SceneEntity *pkEntity, bool bDeleteOld )
{
	if( pkEntity == m_pkEntity )
		return;

	if( bDeleteOld )
		delete m_pkEntity;

	m_pkEntity          = pkEntity;
	m_bNeedVolumeUpdate = true;

	if( m_pkEntity )
	{
		m_pkEntity->m_pkNode  = this;
		m_pkEntity->SetNode();
	}
}


void SceneNode::PrintHierarchy( unsigned int uiLevel )
{
	for( unsigned int uiSpace = 0; uiSpace < uiLevel; ++uiSpace )
		neolog << LogLevel( DEBUG ) << "  ";

	neolog << m_strName << endl;

	unsigned int uiEnd = m_vpkChildren.size();
	vector<SceneNode*>::iterator ppkNode = m_vpkChildren.begin();

	for( unsigned int i = 0; i < uiEnd; ++i, ++ppkNode )
		(*ppkNode)->PrintHierarchy( uiLevel + 1 );
}


SceneNode *SceneNode::Duplicate()
{
	return new SceneNode( *this );
}


void SceneNode::TraverseNode( BaseVisitor &rkVisitor )
{
	HierarchyNode<SceneNode>::TraverseNode( rkVisitor );

	if( m_pkEntity )
		m_pkEntity->Accept( rkVisitor );
}


bool SceneNode::Intersection( BoundingVolume *pkObj, ContactSet *pkContactSet )
{
	bool bIntersect = false;

	if( m_pkEntity && m_pkEntity->IsActive() )
	{
		UpdateWorld(); //Force entity bounding volume SRT to be set correctly
		
		if( !m_pkEntity->GetBoundingVolume() || pkObj->Intersection( m_pkEntity->GetBoundingVolume() ) )
			bIntersect = m_pkEntity->Intersection( pkObj, pkContactSet, true );
	}

	if( !pkContactSet && bIntersect )
		return true;

	vector< SceneNode* >::iterator ppkNode    = m_vpkChildren.begin();
	vector< SceneNode* >::iterator ppkNodeEnd = m_vpkChildren.end();

	for( ; ppkNode != ppkNodeEnd; ++ppkNode )
		if( (*ppkNode)->IsActive() && (*ppkNode)->GetBoundingVolume()->Intersection( pkObj ) )
			bIntersect |= (*ppkNode)->Intersection( pkObj, pkContactSet );

	return bIntersect;
}


bool SceneNode::Intersection( const Ray &rkRay, ContactSet *pkContactSet )
{
	bool bIntersect = false;

	if( m_pkEntity && m_pkEntity->IsActive() )
	{
		UpdateWorld(); //Force entity bounding volume SRT to be set correctly
		
		if( !m_pkEntity->GetBoundingVolume() || m_pkEntity->GetBoundingVolume()->Intersection( rkRay ) )
			bIntersect = m_pkEntity->Intersection( rkRay, pkContactSet );
	}

	if( !pkContactSet && bIntersect )
		return true;

	vector< SceneNode* >::iterator ppkNode    = m_vpkChildren.begin();
	vector< SceneNode* >::iterator ppkNodeEnd = m_vpkChildren.end();

	for( ; ppkNode != ppkNodeEnd; ++ppkNode )
		if( (*ppkNode)->IsActive() && (*ppkNode)->GetBoundingVolume()->Intersection( rkRay ) )
			bIntersect |= (*ppkNode)->Intersection( rkRay, pkContactSet );

	return bIntersect;
}


void SceneNode::AddCallback( SceneNodeCallback *pkCallback )
{
	vector< SceneNodeCallback* >::iterator ppkCallback = find( m_vpkCallbacks.begin(), m_vpkCallbacks.end(), pkCallback );
	
	if( ppkCallback == m_vpkCallbacks.end() )
		m_vpkCallbacks.push_back( pkCallback );
}


void SceneNode::RemoveCallback( SceneNodeCallback *pkCallback )
{
	vector< SceneNodeCallback* >::iterator ppkCallback = find( m_vpkCallbacks.begin(), m_vpkCallbacks.end(), pkCallback );
	
	if( ppkCallback != m_vpkCallbacks.end() )
		m_vpkCallbacks.erase( ppkCallback );
}


};
