/***************************************************************************
                     shader.cpp  -  Pipeline shader class
                             -------------------
    begin       : Mon May 05 2003
    copyright   : (C) 2003 by Reality Rift Studios
    email       : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, shader.cpp

 The Initial Developer of the Original Code is Matt Holmes.
 Portions created by Matt Holmes are Copyright (C) 2003
 Matt Holmes. All Rights Reserved.

 Contributors: Mattias Jansson (mattias@realityrift.com)
               Cody Russell (cody [at] jhu.edu)

 ***************************************************************************/

#include "shader.h"
#include "file.h"
#include "logstream.h"
#include "strutil.h"

using namespace std;


namespace NeoEngine
{


Shader::SHADERTARGET NE_STATIC Shader::s_eVSTarget = Shader::NOTSUPPORTED;
Shader::SHADERTARGET NE_STATIC Shader::s_eFSTarget = Shader::NOTSUPPORTED;

std::vector< Shader::SHADERTARGET > NE_STATIC Shader::s_veVSPriority;
std::vector< Shader::SHADERTARGET > NE_STATIC Shader::s_veFSPriority;

unsigned int NE_STATIC Shader::s_uiCount = 0;


Shader::Shader( SHADERTYPE eType, FileManager *pkFileManager ) :
	LoadableEntity( pkFileManager ),
	m_eShaderType( eType ),
	m_uiID( ++s_uiCount ),
	m_ppkParams( 0 )
{
}


Shader::~Shader()
{
	ClearParams();
}


void Shader::SetName( const HashString &rstrName )
{
	m_strName = rstrName;
}


void Shader::ClearParams()
{
	vector< ShaderParam* >::iterator ppkParam = m_vpkParams.begin();
	vector< ShaderParam* >::iterator ppkEnd   = m_vpkParams.end();

	for( ; ppkParam != ppkEnd; ++ppkParam )
		delete (*ppkParam);

	delete [] m_ppkParams, m_ppkParams = 0;

	m_vpkParams.clear();
	m_kParams.Clear();
}


void Shader::ParseParams()
{
	string::size_type iParamOff = 0;

	while( ( iParamOff = m_strSource.find( "var", iParamOff ) ) != string::npos )
	{
		string::size_type iTypeOff    = m_strSource.find_first_not_of( " \t", iParamOff + 3 );
		string::size_type iNameOff    = m_strSource.find_first_not_of( " \t", m_strSource.find_first_of( " \t", iTypeOff ) );
		string::size_type iNewlineOff = m_strSource.find_first_of( "\n\r", iTypeOff );
		string::size_type iDeclOff    = m_strSource.find( "c[", iTypeOff );

		iParamOff = iNewlineOff;

		if( iDeclOff >= iNewlineOff )
			continue;

		iDeclOff += 2;

		string strType  = m_strSource.substr( iTypeOff, m_strSource.find_first_of( " \t", iTypeOff ) - iTypeOff );
		string strName  = m_strSource.substr( iNameOff, m_strSource.find_first_of( " \t", iNameOff ) - iNameOff );
		string strIndex = m_strSource.substr( iDeclOff, m_strSource.find_first_not_of( "0123456789", iDeclOff ) - iDeclOff );

		ShaderParam::SHADERPARAMTYPE eType;
		unsigned int uiIndex = atoi( strIndex.c_str() );

		if( ( strType == "float4x4" ) || ( strType == "float3x3" ) )
			eType = ShaderParam::PPT_MATRIX;
		else
			eType = ShaderParam::PPT_VECTOR;

		ShaderParam *pkParam = new ShaderParam;

		pkParam->m_uiIndex = uiIndex;
		pkParam->m_eType   = eType;

		m_kParams.Insert( "c" + strIndex, pkParam );
		m_kParams.Insert( strName, pkParam );
		m_vpkParams.push_back( pkParam );

		neolog << LogLevel( INFO ) << "Parameter: " << strName << " : c" << strIndex << " : " << strType << endl;
	}

	iParamOff = 0;

	while( ( iParamOff = m_strSource.find( "bind", iParamOff ) ) != string::npos )
	{
		string::size_type iNewlineOff = m_strSource.find_first_of( "\n\r", iParamOff );

		vector< string > strSplit;

		Explode( "= \t", m_strSource.substr( iParamOff + 4, iNewlineOff - iParamOff - 4 ), &strSplit );

		iParamOff = iNewlineOff;

		ShaderParam *pkParam = m_kParams.Find( strSplit[0] );

		vector< string > vstrBindPath;

		if( pkParam )
		{
			bool bValid = true;

			Explode( ".", strSplit[1], &vstrBindPath );

			unsigned int uiArgs = vstrBindPath.size();

			if( ( vstrBindPath[0] == "System" ) && ( uiArgs > 1 ) )
			{
				pkParam->m_eType = ShaderParam::PPT_SYSTEM;

				if( ( vstrBindPath[1] == "Matrix" ) && ( uiArgs > 2 ) )
				{
					if( vstrBindPath[2] == "MODEL" )
						pkParam->m_Value.m_eSystemID = ShaderParam::PSP_MODEL_MATRIX;
					else if( vstrBindPath[2] == "VIEW" )
						pkParam->m_Value.m_eSystemID = ShaderParam::PSP_VIEW_MATRIX;
					else if( vstrBindPath[2] == "PROJECTION" )
						pkParam->m_Value.m_eSystemID = ShaderParam::PSP_PROJECTION_MATRIX;
					else if( vstrBindPath[2] == "MODELVIEW" )
						pkParam->m_Value.m_eSystemID = ShaderParam::PSP_MODELVIEW_MATRIX;
					else if( vstrBindPath[2] == "MODELVIEWPROJECTION" )
						pkParam->m_Value.m_eSystemID = ShaderParam::PSP_MODELVIEWPROJ_MATRIX;
					else if( vstrBindPath[2] == "INVERSEMODEL" )
						pkParam->m_Value.m_eSystemID = ShaderParam::PSP_INV_MODEL_MATRIX;
					else
						bValid = false;
				}
				else if( ( vstrBindPath[1] == "Light" ) && ( uiArgs > 3 ) )
				{
					pkParam->m_uiLight = (unsigned int)atoi( vstrBindPath[2].c_str() );

					if( vstrBindPath[3] == "POSITION" )
						pkParam->m_Value.m_eSystemID = ShaderParam::PSP_LIGHT_POSITION;
					else if( vstrBindPath[3] == "DIRECTION" )
						pkParam->m_Value.m_eSystemID = ShaderParam::PSP_LIGHT_DIRECTION;
					else if( vstrBindPath[3] == "POSITION_OBJECTSPACE" )
						pkParam->m_Value.m_eSystemID = ShaderParam::PSP_LIGHT_POSITION_OBJECTSPACE;
					else if( vstrBindPath[3] == "DIRECTION_OBJECTSPACE" )
						pkParam->m_Value.m_eSystemID = ShaderParam::PSP_LIGHT_DIRECTION_OBJECTSPACE;
					else if( ( ( vstrBindPath[3] == "DIFFUSE" ) || ( vstrBindPath[3] == "COLOR" ) ) )
						pkParam->m_Value.m_eSystemID = ShaderParam::PSP_LIGHT_DIFFUSE_COLOR;
					else if( vstrBindPath[3] == "AMBIENT" )
						pkParam->m_Value.m_eSystemID = ShaderParam::PSP_LIGHT_AMBIENT_COLOR;
					else if( vstrBindPath[3] == "SPECULAR" )
						pkParam->m_Value.m_eSystemID = ShaderParam::PSP_LIGHT_SPECULAR_COLOR;
					else if( vstrBindPath[3] == "ATTENUATION" )
						pkParam->m_Value.m_eSystemID = ShaderParam::PSP_LIGHT_ATTENUATION;
					else if( vstrBindPath[3] == "REP_ATTENUATION" )
						pkParam->m_Value.m_eSystemID = ShaderParam::PSP_LIGHT_REP_ATTENUATION;
					else
						bValid = false;
				}
				else if( ( vstrBindPath[1] == "Misc" ) && ( uiArgs > 2 ) )
				{
					if( vstrBindPath[2] == "TOTAL_AMBIENT_COLOR" )
						pkParam->m_Value.m_eSystemID = ShaderParam::PSP_TOTAL_LIGHT_AMBIENT_COLOR;
					else if( vstrBindPath[2] == "EYE_POSITION_OBJECTSPACE" )
						pkParam->m_Value.m_eSystemID = ShaderParam::PSP_EYE_POSITION_OBJECTSPACE;
					else
						bValid = false;
				}
				else
					bValid = false;
			}
			else
				bValid = false;

			if( bValid )
				neolog << LogLevel( DEBUG ) << "Parameter bound: " << strSplit[0] << " = " << strSplit[1] << endl;
			else
				neolog << LogLevel( WARNING ) << "*** Invalid bind statement: [" << strSplit[1] << "] unsupported" << endl;
		}
		else
			neolog << LogLevel( WARNING ) << "*** Invalid bind statement: [" << strSplit[0] << "] not found" << endl;
	}

	unsigned int uiMaxIndex = 0;

	vector< ShaderParam* >::iterator ppkParam = m_vpkParams.begin();
	vector< ShaderParam* >::iterator ppkEnd   = m_vpkParams.end();

	for( ; ppkParam != ppkEnd; ++ppkParam )
		if( uiMaxIndex < (*ppkParam)->m_uiIndex )
			uiMaxIndex = (*ppkParam)->m_uiIndex;

	m_ppkParams = new ShaderParam*[ uiMaxIndex + 1 ];

	ppkParam = m_vpkParams.begin();

	for( ; ppkParam != ppkEnd; ++ppkParam )
		m_ppkParams[ (*ppkParam)->m_uiIndex ] = *ppkParam;
}


bool Shader::LoadNode( unsigned int uiFlags )
{
	if( !m_pkFile || !m_pkFile->IsValid() )
		return false;

	unsigned int  uiSize   = m_pkFile->GetSize();
	char         *pcSource = new char[ uiSize + 1 ];

	m_pkFile->Read( pcSource, uiSize );

	pcSource[ uiSize ] = 0;
	m_strSource        = pcSource;

	delete [] pcSource;

	if( Compile() )
	{
		SetName( File::ExtractBaseFileName( m_pkFile->GetName() ) );
		return true;
	}

	return false;
}


void Shader::UpdateParameter( ShaderParam *pkParameter )
{
	m_vpkUpdatedParams.push_back( pkParameter );
}


void Shader::SetPriority( SHADERTYPE eType, const std::vector< SHADERTARGET > &rvePriority )
{
	if( eType == VERTEXSHADER )
	{
		s_veVSPriority.resize( rvePriority.size() );
		copy( rvePriority.begin(), rvePriority.end(), s_veVSPriority.begin() );
	}
	else
	{
		s_veFSPriority.resize( rvePriority.size() );
		copy( rvePriority.begin(), rvePriority.end(), s_veFSPriority.begin() );
	}
}


std::string Shader::GetTargetAsString( SHADERTARGET eTarget )
{
	switch( eTarget )
	{
		case NEPT_DX9_VS_1_1: return "vs1.1";
		case NEPT_DX9_VS_2_0: return "vs2.0";
		case NEPT_DX9_VS_2_X: return "vs2.x";
		case NEPT_DX9_PS_1_1: return "ps1.1";
		case NEPT_DX9_PS_1_2: return "ps1.2";
		case NEPT_DX9_PS_1_3: return "ps1.3";
		case NEPT_DX9_PS_1_4: return "ps1.4";
		case NEPT_DX9_PS_2_0: return "ps2.0";
		case NEPT_DX9_PS_2_X: return "ps2.x";
		case NEPT_GL_ARBVP_1: return "arbvp1";
		case NEPT_GL_ARBFP_1: return "arbfp1";
		case NEPT_GL_ATIFS:   return "atifs";
		case NEPT_GL_NVRC:    return "nvrc";
		case NEPT_GL_NVRC_2:  return "nvrc2";
		default:              break;
	}

	return "invalidtarget";
}


Shader::SHADERTARGET Shader::GetTargetFromString( const std::string &rstrTarget )
{
	if( rstrTarget == "vs1.1" )  return NEPT_DX9_VS_1_1;
	if( rstrTarget == "vs2.0" )  return NEPT_DX9_VS_2_0;
	if( rstrTarget == "vs2.x" )  return NEPT_DX9_VS_2_X;
	if( rstrTarget == "ps1.1" )  return NEPT_DX9_PS_1_1;
	if( rstrTarget == "ps1.2" )  return NEPT_DX9_PS_1_2;
	if( rstrTarget == "ps1.3" )  return NEPT_DX9_PS_1_3;
	if( rstrTarget == "ps1.4" )  return NEPT_DX9_PS_1_4;
	if( rstrTarget == "ps2.0" )  return NEPT_DX9_PS_2_0;
	if( rstrTarget == "ps2.x" )  return NEPT_DX9_PS_2_X;
	if( rstrTarget == "arbvp1" ) return NEPT_GL_ARBVP_1;
	if( rstrTarget == "arbfp1" ) return NEPT_GL_ARBFP_1;
	if( rstrTarget == "atifs" )  return NEPT_GL_ATIFS;
	if( rstrTarget == "nvrc" )   return NEPT_GL_NVRC;
	if( rstrTarget == "nvrc2" )  return NEPT_GL_NVRC_2;
	return NOTSUPPORTED;
}


};
