/***************************************************************************
                     shadow.cpp  -  Shadow generator base
                             -------------------
    begin                : Wed Apr 30 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, shadow.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "shadow.h"
#include "render.h"
#include "boundingvolume.h"
#include "light.h"
#include "scenenode.h"
#include "shader.h"
#include "config.h"

using namespace std;


namespace NeoEngine
{


StencilShadowGenerator::StencilShadowGenerator( float fShadowLength ) :
	ShadowGenerator()
{
	m_fShadowLength = fShadowLength;
	m_uiMaxVertices = 0;
	m_uiMaxPolygons = 0;
	m_pkVertices    = 0;
	m_pkPolygons    = 0;
	m_uiNumUsedOps  = 0;

	m_pkStencilMat     = new Material( "__shadow_stencilbuf", 0 );
	m_pkFrameBufferMat = new Material( "__stencilshadow", Core::Get()->GetMaterialManager() );

	m_pkStencilMat->m_kZBufferMode.Set( ZBufferMode::LESSNOWRITE );

	m_pkFrameBufferMat->m_kZBufferMode.Set( ZBufferMode::ALWAYSNOWRITE );
	m_pkFrameBufferMat->m_kBlendMode.Set( BlendMode::DECAL );
	m_pkFrameBufferMat->m_kDiffuse.Set( 0.0f, 0.0f, 0.0f, 0.7f );

	if( !Core::Get()->GetRenderDevice()->GetCaps().IsSet( RenderCaps::STENCILBUFFER ) )
	{
		neolog << LogLevel( WARNING ) << "[StencilShadowGenerator] no support for stencil buffer in render device" << endl;

		Deactivate();
	}
	else
	{
		//Register as frame listener
		Core::Get()->GetRenderDevice()->RegisterFrameCallback( FrameCallback::PERLIGHT, this, true );
		Core::Get()->GetRenderDevice()->RegisterFrameCallback( FrameCallback::FRAMEEND, this, true );
	}
}


StencilShadowGenerator::~StencilShadowGenerator()
{
	//Deregister as frame listener
	if( Core::Get()->GetRenderDevice() )
	{
		Core::Get()->GetRenderDevice()->RegisterFrameCallback( FrameCallback::PERLIGHT, this, false );
		Core::Get()->GetRenderDevice()->RegisterFrameCallback( FrameCallback::FRAMEEND, this, false );
	}

	delete [] m_pkPolygons;
	delete [] m_pkVertices;

	vector< ShadowOp* >::iterator ppkOp    = m_vpkOps.begin();
	vector< ShadowOp* >::iterator ppkOpEnd = m_vpkOps.end();

	for( ; ppkOp != ppkOpEnd; ++ppkOp )
		delete( *ppkOp );
}


void StencilShadowGenerator::GenerateShadow( VertexBufferPtr &pkVertexBuffer, PolygonBufferPtr &pkPolygonBuffer, BoundingVolume *pkBoundingVolume, const Matrix &rkTransform, const Matrix &rkInvTransform )
{
	if( !IsActive() )
		return;

	ShadowOp *pkOp = ( m_uiNumUsedOps >= m_vpkOps.size() ) ? new ShadowOp : m_vpkOps[ m_uiNumUsedOps ];

	pkOp->m_pkVertexBuffer   = pkVertexBuffer;
	pkOp->m_pkPolygonBuffer  = pkPolygonBuffer;
	pkOp->m_pkBoundingVolume = pkBoundingVolume;
	pkOp->m_kTransform       = rkTransform;
	pkOp->m_kInvTransform    = rkInvTransform;

	if( !pkPolygonBuffer->GetNormals() )
		pkPolygonBuffer->CalculateNormals( pkVertexBuffer, false );

	if( !pkPolygonBuffer->GetEdges() )
		pkPolygonBuffer->CalculateEdges( pkVertexBuffer );

	if( m_uiNumUsedOps >= m_vpkOps.size() )
		m_vpkOps.push_back( pkOp );

	++m_uiNumUsedOps;
}


void StencilShadowGenerator::FrameEvent( FrameCallback::FRAMECALLBACKTYPE eType, void *pData )
{
	if( eType == FrameCallback::FRAMEEND )
	{
		m_uiNumUsedOps = 0;
		return;
	}

	if( ( eType != FrameCallback::PERLIGHT ) || !m_uiNumUsedOps )
		return;

	Light *pkLight = static_cast< Light* >( pData );

	if( !( pkLight->GetFlags() & Light::CASTSHADOWS ) || ( pkLight->GetType() == Light::AMBIENT ) )
		return;


	RenderPrimitive kPrimitive;

	kPrimitive.m_ePrimitive = RenderPrimitive::TRIANGLES;


	Core::Get()->GetRenderDevice()->SetStencilFunc( RenderDevice::STENCILFUNC_ALWAYS, 1, 0xFFFFFFFF );
	Core::Get()->GetRenderDevice()->SetStencilMask( 0xFFFFFFFF );
	Core::Get()->GetRenderDevice()->SetTargetMask( false );

	Core::Get()->GetRenderDevice()->Clear( RenderDevice::STENCILBUFFER, Color::BLACK, 0.0f, 0 );

	vector< ShadowOp* >           vpkCurOps( m_vpkOps.size(), 0 );
	vector< ShadowOp* >::iterator ppkOp    = m_vpkOps.begin();
	vector< ShadowOp* >::iterator ppkCurOp = vpkCurOps.begin();
	unsigned int                    uiNumOps = 0;
	unsigned int                    uiOp;

	for( uiOp = 0; uiOp < m_uiNumUsedOps; ++uiOp, ++ppkOp )
	{
		if( (*ppkOp)->m_pkBoundingVolume && !( (*ppkOp)->m_pkBoundingVolume->Intersection( pkLight->GetBoundingVolume() ) ) )
			continue;

		*ppkCurOp++ = *ppkOp;

		++uiNumOps;
	}

	ppkCurOp = vpkCurOps.begin();

	for( int iPass = 0; iPass < 2; ++iPass )
	{
		if( iPass == 0 )
		{
			// First pass. Increase stencil value in the shadow
			// with backfacing shadow volume polygons failing z test
			Core::Get()->GetRenderDevice()->SetCullMode( RenderDevice::CULLFRONT );
			Core::Get()->GetRenderDevice()->SetStencilOp( RenderDevice::STENCILOP_KEEP, RenderDevice::STENCILOP_INCR, RenderDevice::STENCILOP_KEEP );
		}
		else
		{
			// Second pass. Decrease stencil value in the shadow
			// with frontfacing shadow volume polygons failing z test
			Core::Get()->GetRenderDevice()->SetCullMode( RenderDevice::CULLBACK );
			Core::Get()->GetRenderDevice()->SetStencilOp( RenderDevice::STENCILOP_KEEP, RenderDevice::STENCILOP_DECR, RenderDevice::STENCILOP_KEEP );
		}

		ppkCurOp = vpkCurOps.begin();

		Core::Get()->GetRenderDevice()->Begin( Core::Get()->GetRenderDevice()->GetViewMatrix(), RenderDevice::NOSORT );
		{
			for( uiOp = 0; uiOp < uiNumOps; ++uiOp, ++ppkCurOp )
			{
				//Calculate and generate primitive buffers first pass
				if( iPass == 0 )
				{
					PolygonBuffer *pkPolygonBuffer = (*ppkCurOp)->m_pkPolygonBuffer;
					VertexBuffer  *pkVertexBuffer  = (*ppkCurOp)->m_pkVertexBuffer;

					Edge       *pkEdges        = pkPolygonBuffer->GetEdges();
					Vector3d   *pkPolyNormals  = pkPolygonBuffer->GetNormals();
					unsigned int  uiNumVertices  = pkVertexBuffer->GetNumElements();
					unsigned int  uiNumPolygons  = pkPolygonBuffer->GetNumElements();
					unsigned int  uiNumEdges     = uiNumPolygons * 3;

					//For completely valid models, we would ideally have worst case scenario of
					// ( uiNumPolygons * 2 + ( ( uiNumEdges / 2 ) * 2 ) )
					//resulting shadow polygons. Worst case scenarion with invalid models (no sharing edges) is
					// ( uiNumPolygons * 2 + uiNumEdges * 2 )
					//Go for inbetween solution and look for overflow
					// ( uiNumPolygons * 2 + ( ( uiNumEdges * 3 ) / 2 ) )

					unsigned int  uiMaxShadowVertices = uiNumVertices * 2;
					unsigned int  uiMaxShadowPolygons = ( uiNumPolygons * 2 + ( ( uiNumEdges * 3 ) / 2 ) );

					if( uiMaxShadowVertices > m_uiMaxVertices )
					{
						delete [] m_pkVertices;

						m_pkVertices = new Vertex[ ( m_uiMaxVertices = uiMaxShadowVertices ) ];
					}

					if( uiMaxShadowPolygons > m_uiMaxPolygons )
					{
						delete [] m_pkPolygons;

						m_pkPolygons = new Polygon [ ( m_uiMaxPolygons = uiMaxShadowPolygons ) ];
					}

					pkVertexBuffer->Lock( Buffer::READ );

					Vertex  *pkShadowVertices = m_pkVertices;
					Polygon *pkShadowPolygons = m_pkPolygons;

					unsigned int uiNumShadowPolygons  = 0;

					unsigned char *pucSrcVertex    = (unsigned char*)pkVertexBuffer->GetVertex();
					unsigned int   uiSrcVertexSize = pkVertexBuffer->GetVertexSize();

					Vector3d kLightPos;
					Vector3d kLightDirection;

					//Calculate shadow vertices
					if( pkLight->GetType() == Light::POINT )
					{
						kLightPos = (*ppkCurOp)->m_kInvTransform * pkLight->GetSceneNode()->GetWorldTranslation();

						Vector3d *pkNearVertex = &pkShadowVertices->m_kPosition;
						Vector3d *pkFarVertex  = &pkShadowVertices->m_kPosition + uiNumVertices;

						// ### FIXME: We assume position offset is 0
						for( unsigned int uiVertex = 0; uiVertex < uiNumVertices; ++uiVertex, pucSrcVertex += uiSrcVertexSize, ++pkNearVertex, ++pkFarVertex )
						{
							*pkNearVertex   = *((Vector3d*)pucSrcVertex);
							*pkFarVertex    = *pkNearVertex + ( *pkNearVertex - kLightPos ).ScaleTo( m_fShadowLength );
						}
					}
					else if( pkLight->GetType() == Light::DIRECTIONAL )
					{
						kLightDirection = -pkLight->GetSceneNode()->GetWorldTransform().GetColumn(2);

						//Only rotation direction vector, NO translation!
						Matrix kInvModelMatrix( (*ppkCurOp)->m_kInvTransform );

						kInvModelMatrix[0][3] = kInvModelMatrix[1][3] = kInvModelMatrix[2][3] =
						kInvModelMatrix[3][0] = kInvModelMatrix[3][1] = kInvModelMatrix[3][2] = 0.0f;

						kLightDirection = kInvModelMatrix * kLightDirection;

						Vector3d kLightVecNear = kLightDirection * 0.02f;
						Vector3d kLightVecFar  = kLightDirection * m_fShadowLength;

						Vector3d *pkNearVertex = &pkShadowVertices->m_kPosition;
						Vector3d *pkFarVertex  = &pkShadowVertices->m_kPosition + uiNumVertices;

						// ### FIXME: We assume position offset is 0
						for( unsigned int uiVertex = 0; uiVertex < uiNumVertices; ++uiVertex, pucSrcVertex += uiSrcVertexSize, ++pkNearVertex, ++pkFarVertex )
						{
							*pkNearVertex = *((Vector3d*)pucSrcVertex) + kLightVecNear;
							*pkFarVertex  = *pkNearVertex + kLightVecFar;
						}
					}
					else
						assert( !"Invalid light type" );


					//Calculate backfacing flag
					Edge       *pkEdge         = pkEdges;
					Vector3d   *pkPolyNormal   = pkPolyNormals;
					unsigned int  uiPoly;

					pucSrcVertex = (unsigned char*)pkVertexBuffer->GetVertex();

					for( uiPoly = 0; uiPoly < uiNumPolygons; ++uiPoly, ++pkPolyNormal )
					{
						if( pkLight->GetType() == Light::POINT )
							kLightDirection = *((Vector3d*)( pucSrcVertex + uiSrcVertexSize * pkEdge->m_usVertices[0] )) - kLightPos;

						int iBackfacing    = ( ( (*pkPolyNormal) * kLightDirection ) < 0.0f ) ? 1 : -1;

						pkEdge->m_iVisible = iBackfacing; ++pkEdge;
						pkEdge->m_iVisible = iBackfacing; ++pkEdge;
						pkEdge->m_iVisible = iBackfacing; ++pkEdge;
					}

					pkVertexBuffer->Unlock();


					pkEdge = pkEdges;
					
					Polygon *pkShadowPolygon = pkShadowPolygons;

					//Loop through all edges, find visible isolated edges and/or visible edges with invisible neighbour polys
					for( unsigned int uiEdge = 0; uiEdge < uiNumEdges; ++uiEdge, ++pkEdge )
					{
						// ### FIXME: What about non-visible edges with no neighbour?

						//Check if visible with invisible neighbour, or if visible and no neighbour edge
						if( ( pkEdge->m_iVisible > 0 ) && ( !pkEdge->m_pkNeighbour || ( pkEdge->m_pkNeighbour->m_iVisible < 0 ) ) )
						{
							//Add two shadow polygons
							if( ( uiNumShadowPolygons + 2 ) > uiMaxShadowPolygons )
							{
								neolog << LogLevel( ERROR ) << "[StencilShadowGenerator::FrameCallback] model invalid, shadow polygon overflow (edge)" << endl;
								break;
							}

							pkShadowPolygon->v[0] = pkEdge->m_usVertices[0];
							pkShadowPolygon->v[1] = pkEdge->m_usVertices[0] + uiNumVertices;
							pkShadowPolygon->v[2] = pkEdge->m_usVertices[1];

							++pkShadowPolygon;

							pkShadowPolygon->v[0] = pkEdge->m_usVertices[1];
							pkShadowPolygon->v[1] = pkEdge->m_usVertices[0] + uiNumVertices;
							pkShadowPolygon->v[2] = pkEdge->m_usVertices[1] + uiNumVertices;

							++pkShadowPolygon;

							uiNumShadowPolygons += 2;
						}
					}

					pkPolygonBuffer->Lock( Buffer::READ );

					Polygon *pkPolygon = pkPolygonBuffer->GetPolygon();

					pkEdge = pkEdges;

					//All edges done, now render frontfacing triangles at infinite distance and backfacing triangles at normal position
					for( uiPoly = 0; uiPoly < uiNumPolygons; ++uiPoly, pkEdge += 3, ++pkPolygon )
					{
						unsigned int uiAdd = ( pkEdge->m_iVisible < 0 ) ? uiNumVertices : 0;

						if( ( uiNumShadowPolygons + 1 ) > uiMaxShadowPolygons )
						{
							neolog << LogLevel( ERROR ) << "[StencilShadowGenerator::FrameCallback] model invalid, shadow polygon overflow (polygon)" << endl;
							break;
						}

						pkShadowPolygon->v[0] = pkPolygon->v[0] + uiAdd;
						pkShadowPolygon->v[1] = pkPolygon->v[1] + uiAdd;
						pkShadowPolygon->v[2] = pkPolygon->v[2] + uiAdd;

						++pkShadowPolygon;
						++uiNumShadowPolygons;
					}

					pkPolygonBuffer->Unlock();


					//Setup render primitive
					(*ppkCurOp)->m_kPrimitive.m_ePrimitive      = RenderPrimitive::TRIANGLES;
					(*ppkCurOp)->m_kPrimitive.m_kModelMatrix    = (*ppkCurOp)->m_kTransform;
					(*ppkCurOp)->m_kPrimitive.m_pkPolygonBuffer = Core::Get()->GetRenderDevice()->CreatePolygonBuffer( Buffer::STATIC | Buffer::NOREADWRITE, uiNumShadowPolygons, pkShadowPolygons, false );
					(*ppkCurOp)->m_kPrimitive.m_pkVertexBuffer  = Core::Get()->GetRenderDevice()->CreateVertexBuffer(  Buffer::STATIC | Buffer::NOREADWRITE, uiNumVertices * 2, &Vertex::s_kDecl, pkShadowVertices );
					(*ppkCurOp)->m_kPrimitive.m_uiNumPrimitives = uiNumShadowPolygons;
					(*ppkCurOp)->m_kPrimitive.m_pkMaterial      = m_pkStencilMat;
				}

				//Render primitive
				Core::Get()->GetRenderDevice()->Render( (*ppkCurOp)->m_kPrimitive, 0 );
			}
		}
		Core::Get()->GetRenderDevice()->End();
	}


	//Render stencil buffer to frame buffer
	Core::Get()->GetRenderDevice()->SetStencilOp( RenderDevice::STENCILOP_KEEP, RenderDevice::STENCILOP_KEEP, RenderDevice::STENCILOP_KEEP );
	Core::Get()->GetRenderDevice()->SetStencilFunc( RenderDevice::STENCILFUNC_NOTEQUAL, 0, 0xFFFFFFFF );
	Core::Get()->GetRenderDevice()->SetCullMode( RenderDevice::CULLBACK );
	Core::Get()->GetRenderDevice()->SetTargetMask( true );

	// ### FIXME: Use current projection mode
	Core::Get()->GetRenderDevice()->SetOrthographicProjection();

	Core::Get()->GetRenderDevice()->Begin( Matrix::IDENTITY, RenderDevice::NOSORT );
	{
		RenderPrimitive kPrimitive;

		kPrimitive.m_ePrimitive      = RenderPrimitive::TRIANGLES;
		kPrimitive.m_pkMaterial      = m_pkFrameBufferMat;
		kPrimitive.m_uiNumPrimitives = 2;
		kPrimitive.m_pkVertexBuffer  = Core::Get()->GetRenderDevice()->CreateVertexBuffer( Buffer::DYNAMIC, 4, &DiffuseVertex::s_kDecl );
		kPrimitive.m_pkPolygonBuffer = Core::Get()->GetRenderDevice()->CreatePolygonBuffer( Buffer::DYNAMIC, 2 );

		kPrimitive.m_pkVertexBuffer->Lock( Buffer::WRITE );

		DiffuseVertex *pkVertex = (DiffuseVertex*)kPrimitive.m_pkVertexBuffer->GetVertex();

		Color32 kColor = m_pkFrameBufferMat->m_kDiffuse;

		pkVertex->m_kPosition.Set( 0.0f, 0.0f, 0.0f );
		pkVertex->m_kColor = kColor; ++pkVertex;

		pkVertex->m_kPosition.Set( 0.0f, 1.0f, 0.0f );
		pkVertex->m_kColor = kColor; ++pkVertex;

		pkVertex->m_kPosition.Set( 1.0f, 1.0f, 0.0f );
		pkVertex->m_kColor = kColor; ++pkVertex;

		pkVertex->m_kPosition.Set( 1.0f, 0.0f, 0.0f );
		pkVertex->m_kColor = kColor;

		kPrimitive.m_pkVertexBuffer->Unlock();

		kPrimitive.m_pkPolygonBuffer->Lock( Buffer::WRITE );

		Polygon *pkPolygon = kPrimitive.m_pkPolygonBuffer->GetPolygon();

		pkPolygon->v[0] = 0;
		pkPolygon->v[1] = 1;
		pkPolygon->v[2] = 2; ++pkPolygon;

		pkPolygon->v[0] = 0;
		pkPolygon->v[1] = 2;
		pkPolygon->v[2] = 3;

		kPrimitive.m_pkPolygonBuffer->Unlock();

		Core::Get()->GetRenderDevice()->Render( kPrimitive, 0 );
	}
	Core::Get()->GetRenderDevice()->End();

	Core::Get()->GetRenderDevice()->SetPerspectiveProjection();

	Core::Get()->GetRenderDevice()->SetStencilFunc( RenderDevice::STENCILFUNC_ALWAYS, 1, 0xFFFFFFFF );

	ppkCurOp = vpkCurOps.begin();

	for( uiOp = 0; uiOp < uiNumOps; ++uiOp, ++ppkCurOp )
	{
		(*ppkCurOp)->m_pkVertexBuffer  = 0;
		(*ppkCurOp)->m_pkPolygonBuffer = 0;

		// This makes buffer be cleaned up next frame or in dtor
		//(*ppkCurOp)->m_kPrimitive.m_pkVertexBuffer  = 0;
		//(*ppkCurOp)->m_kPrimitive.m_pkPolygonBuffer = 0;
	}
}







#if 0

VertexShaderStencilShadowGenerator::VertexShaderStencilShadowGenerator( float fShadowLength, float fOffset ) :
	ShadowGenerator(),
	FrameCallback(),
	m_uiNumUsedOps( 0 ),
	m_fShadowLength( fShadowLength ),
	m_pkVertexShaderPos( 0 ),
	m_pkVertexShaderDir( 0 )
{
	m_pkStencilMat     = new Material( "__shadow_stencilbuf", 0 );
	m_pkFrameBufferMat = new Material( "__stencilshadow", Core::Get()->GetMaterialManager() );

	m_pkStencilMat->m_kZBufferMode.Set( ZBufferMode::LEQUALNOWRITE );

	m_pkFrameBufferMat->m_kZBufferMode.Set( ZBufferMode::ALWAYSNOWRITE );
	m_pkFrameBufferMat->m_kBlendMode.Set( BlendMode::DECAL );
	m_pkFrameBufferMat->m_kColor.Set( 0.0f, 0.0f, 0.0f, 0.7f );

	if( !Core::Get()->GetRenderDevice()->GetCaps().IsSet( RenderCaps::STENCILBUFFER ) )
	{
		neolog << LogLevel( WARNING ) << "[StencilShadowGenerator] no support for stencil buffer in render device" << endl;

		Deactivate();
	}
	else if( !Core::Get()->GetRenderDevice()->GetCaps().IsSet( RenderCaps::VERTEXSHADER ) )
	{
		neolog << LogLevel( WARNING ) << "[VertexShaderStencilShadowGenerator] no support for vertex shaders in render device" << endl;

		Deactivate();
	}
	else
	{
		//Load and compile vertex shaders
		m_pkVertexShaderPos = Core::Get()->GetRenderDevice()->CreateShader( Shader::VERTEXSHADER );

		// NOTE: Light position should be in world space
		string strFrontExtrudeLightPosVP = 

		"void main( in         float4   Position  : POSITION,"
		"           in         float4   Normal    : NORMAL,"
		"           in uniform float4x4 ModelViewProjection,"
		"           in uniform float4x4 InvModel,"
		"           in uniform float4   LightPos,"
		"           in uniform float    ExtrudeLength,"
		"           out        float4   HPosition : POSITION )"
		"{"
		"	float4 LightModelPos = mul( InvModel, LightPos );"
		"	float3 LightDir      = Position.xyz - LightModelPos.xyz;"
		"	LightDir             = normalize( LightDir );"
		"	float ExtrudeFactor  = step( 0.0, -dot( LightDir, Normal.xyz ) ) * ExtrudeLength##offset##;"
		"	LightDir             = LightDir * ExtrudeFactor;"
		"	Position.xyz         = Position.xyz + LightDir;"
		"	HPosition            = mul( ModelViewProjection, Position );"
		"}";

		string strOffset = "";

		if( fabsf( fOffset ) > EPSILON )
		{
			char szOffset[32];

			sprintf( szOffset, " + %.4f", fOffset );

			strOffset = szOffset;
		}

		strFrontExtrudeLightPosVP.replace( strFrontExtrudeLightPosVP.find_first_of( '#' ), 10, strOffset );

		if( !m_pkVertexShaderPos->Compile( &strFrontExtrudeLightPosVP ) )
		{
			neolog << LogLevel( WARNING ) << "[VertexShaderStencilShadowGenerator] unable to compile vertex shader for positional lights" << endl;

			Deactivate();

			return;
		}

		m_pkTransformParamPos    = m_pkVertexShaderPos->GetParameter( "ModelViewProjection" );
		m_pkInvTransformParamPos = m_pkVertexShaderPos->GetParameter( "InvModel"            );
		m_pkLightParamPos        = m_pkVertexShaderPos->GetParameter( "LightPos"            );
		m_pkExtrudeParamPos      = m_pkVertexShaderPos->GetParameter( "ExtrudeLength"       );

		m_pkTransformParamPos->m_eType                = ShaderParam::PPT_SYSTEM;
		m_pkTransformParamPos->m_Value.m_eSystemID    = ShaderParam::PSP_MODELVIEWPROJ_MATRIX;

		m_pkInvTransformParamPos->m_eType             = ShaderParam::PPT_SYSTEM;
		m_pkInvTransformParamPos->m_Value.m_eSystemID = ShaderParam::PSP_INV_MODEL_MATRIX;

		m_pkExtrudeParamPos->m_Value.m_fFloat         = m_fShadowLength;

		m_pkVertexShaderPos->UpdateParameter( m_pkTransformParamPos    );
		m_pkVertexShaderPos->UpdateParameter( m_pkInvTransformParamPos );
		m_pkVertexShaderPos->UpdateParameter( m_pkExtrudeParamPos      );


		m_pkVertexShaderDir = Core::Get()->GetRenderDevice()->CreateShader( Shader::VERTEXSHADER );

		// NOTE: Light direction should be in world space
		string strFrontExtrudeLightDirVP = 

		"void main( in         float4   Position  : POSITION,"
		"           in         float4   Normal    : NORMAL,"
		"           in uniform float4x4 ModelViewProjection,"
		"           in uniform float4x4 InvModel,"
		"           in uniform float4   LightDir,"
		"           in uniform float    ExtrudeLength,"
		"           out        float4   HPosition : POSITION )"
		"{"
		"   LightDir.w           = 0.0f;"
		"	float4 LightModelDir = mul( InvModel, LightDir );"
		"	float ExtrudeFactor  = step( 0.0, -dot( LightModelDir.xyz, Normal.xyz ) ) * ExtrudeLength##offset##;"
		"	LightModelDir        = LightModelDir * ExtrudeFactor;"
		"	Position.xyz         = Position.xyz + LightModelDir.xyz;"
		"	HPosition            = mul( ModelViewProjection, Position );"
		"}";

		strOffset = "";

		if( fabsf( fOffset ) > EPSILON )
		{
			char szOffset[32];

			sprintf( szOffset, " + %.4f", fOffset );

			strOffset = szOffset;
		}

		strFrontExtrudeLightDirVP.replace( strFrontExtrudeLightDirVP.find_first_of( '#' ), 10, strOffset );

		if( !m_pkVertexShaderDir->Compile( &strFrontExtrudeLightDirVP ) )
		{
			neolog << LogLevel( WARNING ) << "[VertexShaderStencilShadowGenerator] unable to compile vertex shader for directional lights" << endl;

			Deactivate();

			return;
		}

		m_pkTransformParamDir    = m_pkVertexShaderDir->GetParameter( "ModelViewProjection" );
		m_pkInvTransformParamDir = m_pkVertexShaderDir->GetParameter( "InvModel"            );
		m_pkLightParamDir        = m_pkVertexShaderDir->GetParameter( "LightDir"            );
		m_pkExtrudeParamDir      = m_pkVertexShaderDir->GetParameter( "ExtrudeLength"       );

		m_pkTransformParamDir->m_eType                = ShaderParam::PPT_SYSTEM;
		m_pkTransformParamDir->m_Value.m_eSystemID    = ShaderParam::PSP_MODELVIEWPROJ_MATRIX;

		m_pkInvTransformParamDir->m_eType             = ShaderParam::PPT_SYSTEM;
		m_pkInvTransformParamDir->m_Value.m_eSystemID = ShaderParam::PSP_INV_MODEL_MATRIX;

		m_pkExtrudeParamDir->m_Value.m_fFloat         = m_fShadowLength;

		m_pkVertexShaderDir->UpdateParameter( m_pkTransformParamDir    );
		m_pkVertexShaderDir->UpdateParameter( m_pkInvTransformParamDir );
		m_pkVertexShaderDir->UpdateParameter( m_pkExtrudeParamDir      );


		//Register as frame listener
		Core::Get()->GetRenderDevice()->RegisterFrameCallback( FrameCallback::PERLIGHT, this, true );
		Core::Get()->GetRenderDevice()->RegisterFrameCallback( FrameCallback::FRAMEEND, this, true );
	}
}


VertexShaderStencilShadowGenerator::~VertexShaderStencilShadowGenerator()
{
	//Deregister as frame listener
	if( Core::Get()->GetRenderDevice() )
	{
		Core::Get()->GetRenderDevice()->RegisterFrameCallback( FrameCallback::PERLIGHT, this, false );
		Core::Get()->GetRenderDevice()->RegisterFrameCallback( FrameCallback::FRAMEEND, this, false );
	}

	vector< ShadowOp* >::iterator ppkOp    = m_vpkOps.begin();
	vector< ShadowOp* >::iterator ppkOpEnd = m_vpkOps.end();

	for( ; ppkOp != ppkOpEnd; ++ppkOp )
		delete( *ppkOp );

	delete m_pkVertexShaderPos;
	delete m_pkVertexShaderDir;
}


void VertexShaderStencilShadowGenerator::GenerateShadow( VertexBufferPtr &pkVertexBuffer, PolygonBufferPtr &pkPolygonBuffer, BoundingVolume *pkBoundingVolume, const Matrix &rkTransform, const Matrix &rkInvTransform )
{
	if( !IsActive() )
		return;

	ShadowOp *pkOp = ( m_uiNumUsedOps >= m_vpkOps.size() ) ? new ShadowOp : m_vpkOps[ m_uiNumUsedOps ];

	pkOp->m_pkVertexBuffer   = pkVertexBuffer;
	pkOp->m_pkPolygonBuffer  = pkPolygonBuffer;
	pkOp->m_pkBoundingVolume = pkBoundingVolume;
	pkOp->m_kTransform       = rkTransform;
	pkOp->m_kInvTransform    = rkInvTransform;

	if( m_uiNumUsedOps >= m_vpkOps.size() )
		m_vpkOps.push_back( pkOp );

	++m_uiNumUsedOps;
}


void VertexShaderStencilShadowGenerator::FrameEvent( FrameCallback::FRAMECALLBACKTYPE eType, void *pData )
{
	if( eType == FrameCallback::FRAMEEND )
	{
		m_uiNumUsedOps = 0;
		return;
	}

	if( ( eType != FrameCallback::PERLIGHT ) || !m_uiNumUsedOps )
		return;

	Light *pkLight = static_cast< Light* >( pData );

	if( !pkLight->GetFlags() & Light::CASTSHADOWS )
		return;

	if( pkLight->GetType() == Light::POINT )
	{
		m_pkLightParamPos->m_Value.m_pkVector = &pkLight->GetSceneNode()->GetWorldTranslation();

		m_pkVertexShaderPos->UpdateParameter( m_pkLightParamPos );

		m_pkStencilMat->m_pkVertexShader = m_pkVertexShaderPos;
	}
	else
	{
		m_pkLightParamDir->m_Value.m_pkVector = &pkLight->m_kDirection;

		m_pkVertexShaderDir->UpdateParameter( m_pkLightParamDir );

		m_pkStencilMat->m_pkVertexShader = m_pkVertexShaderDir;
	}


	RenderPrimitive kPrimitive;

	kPrimitive.m_ePrimitive = RenderPrimitive::TRIANGLES;


	//Core::Get()->GetRenderDevice()->SetStencilFunc( RenderDevice::STENCILFUNC_ALWAYS, 1, 0xFFFFFFFF );
	//Core::Get()->GetRenderDevice()->SetStencilMask( 0xFFFFFFFF );
	//Core::Get()->GetRenderDevice()->SetTargetMask( false );

	Core::Get()->GetRenderDevice()->Clear( RenderDevice::STENCILBUFFER, Color::BLACK, 0.0f, 0 );

	vector< ShadowOp* >           vpkCurOps( m_vpkOps.size(), 0 );
	vector< ShadowOp* >::iterator ppkOp    = m_vpkOps.begin();
	vector< ShadowOp* >::iterator ppkCurOp = vpkCurOps.begin();
	unsigned int                    uiNumOps = 0;
	unsigned int                    uiOp;

	for( uiOp = 0; uiOp < m_uiNumUsedOps; ++uiOp, ++ppkOp )
	{
		if( (*ppkOp)->m_pkBoundingVolume && !( (*ppkOp)->m_pkBoundingVolume->Intersection( pkLight->GetBoundingVolume() ) ) )
			continue;

		*ppkCurOp++ = *ppkOp;

		++uiNumOps;
	}

	ppkCurOp = vpkCurOps.begin();

	//for( int iPass = 0; iPass < 2; ++iPass )
	{
//		if( iPass == 0 )
		{
			// First pass. Increase stencil value in the shadow
			// with frontfacing shadow volume polygons failing z test
			Core::Get()->GetRenderDevice()->SetCullMode( RenderDevice::CULLBACK );
//			Core::Get()->GetRenderDevice()->SetStencilOp( RenderDevice::STENCILOP_KEEP, RenderDevice::STENCILOP_INCR, RenderDevice::STENCILOP_KEEP );
		}
/*		else
		{
			// Second pass. Decrease stencil value in the shadow
			// with backfacing shadow volume polygons failing z test
			Core::Get()->GetRenderDevice()->SetCullMode( RenderDevice::CULLFRONT );
			Core::Get()->GetRenderDevice()->SetStencilOp( RenderDevice::STENCILOP_KEEP, RenderDevice::STENCILOP_DECR, RenderDevice::STENCILOP_KEEP );
		}
*/
		ppkCurOp = vpkCurOps.begin();

		Core::Get()->GetRenderDevice()->Begin( Core::Get()->GetRenderDevice()->GetViewMatrix(), RenderDevice::NOSORT );
		{
			for( uiOp = 0; uiOp < uiNumOps; ++uiOp, ++ppkCurOp )
			{
				//Generate primitives first pass
//				if( iPass == 0 )
				{
					//Setup render primitive
					// ### FIXME: Allow use of triangle strips
					(*ppkCurOp)->m_kPrimitive.m_ePrimitive      = RenderPrimitive::TRIANGLES;
					(*ppkCurOp)->m_kPrimitive.m_kModelMatrix    = (*ppkCurOp)->m_kTransform;
					(*ppkCurOp)->m_kPrimitive.m_kInvModelMatrix = (*ppkCurOp)->m_kInvTransform;
					(*ppkCurOp)->m_kPrimitive.m_pkPolygonBuffer = (*ppkCurOp)->m_pkPolygonBuffer;
					(*ppkCurOp)->m_kPrimitive.m_pkVertexBuffer  = (*ppkCurOp)->m_pkVertexBuffer;
					(*ppkCurOp)->m_kPrimitive.m_uiNumPrimitives = (*ppkCurOp)->m_pkPolygonBuffer->GetNumPolygons();
					(*ppkCurOp)->m_kPrimitive.m_pkMaterial      = m_pkStencilMat;
				}

				//Render primitive
				Core::Get()->GetRenderDevice()->Render( (*ppkCurOp)->m_kPrimitive, 0 );
			}
		}
		Core::Get()->GetRenderDevice()->End();
	}
/*
	//Render stencil buffer to frame buffer
	Core::Get()->GetRenderDevice()->SetStencilOp( RenderDevice::STENCILOP_KEEP, RenderDevice::STENCILOP_KEEP, RenderDevice::STENCILOP_KEEP );
	Core::Get()->GetRenderDevice()->SetStencilFunc( RenderDevice::STENCILFUNC_NOTEQUAL, 0, 0xFFFFFFFF );
	Core::Get()->GetRenderDevice()->SetTargetMask( true );

	Core::Get()->GetRenderDevice()->SetCullMode( RenderDevice::CULLBACK );

	// ### FIXME: Use current projection mode
	Core::Get()->GetRenderDevice()->SetOrthographicProjection();

	Core::Get()->GetRenderDevice()->Begin( Matrix::IDENTITY, RenderDevice::NOSORT );
	{
		RenderPrimitive kPrimitive;

		kPrimitive.m_ePrimitive      = RenderPrimitive::TRIANGLES;
		kPrimitive.m_pkMaterial      = m_pkFrameBufferMat;
		kPrimitive.m_uiNumPrimitives = 2;
		kPrimitive.m_pkVertexBuffer  = Core::Get()->GetRenderDevice()->CreateVertexBuffer( Buffer::DYNAMIC, 4, VertexFormat::NEFVF_XYZC );
		kPrimitive.m_pkPolygonBuffer = Core::Get()->GetRenderDevice()->CreatePolygonBuffer( Buffer::DYNAMIC, 2 );

		kPrimitive.m_pkVertexBuffer->Lock( Buffer::WRITE );

		VertexIterator kVertex( kPrimitive.m_pkVertexBuffer->GetIterator() );

		Color32 kColor = m_pkFrameBufferMat->m_kDiffuse;

		kVertex.GetXYZ()->Set( 0.0f, 0.0f, 0.0f );
		*kVertex.GetColor() = kColor; ++kVertex;

		kVertex.GetXYZ()->Set( 0.0f, 1.0f, 0.0f );
		*kVertex.GetColor() = kColor; ++kVertex;

		kVertex.GetXYZ()->Set( 1.0f, 1.0f, 0.0f );
		*kVertex.GetColor() = kColor; ++kVertex;

		kVertex.GetXYZ()->Set( 1.0f, 0.0f, 0.0f );
		*kVertex.GetColor() = kColor;

		kPrimitive.m_pkVertexBuffer->Unlock();

		kPrimitive.m_pkPolygonBuffer->Lock( Buffer::WRITE );

		Polygon *pkPolygon = kPrimitive.m_pkPolygonBuffer->GetElement();

		pkPolygon->v[0] = 0;
		pkPolygon->v[1] = 1;
		pkPolygon->v[2] = 2; ++pkPolygon;

		pkPolygon->v[0] = 0;
		pkPolygon->v[1] = 2;
		pkPolygon->v[2] = 3;

		kPrimitive.m_pkPolygonBuffer->Unlock();

		Core::Get()->GetRenderDevice()->Render( kPrimitive, 0 );
	}
	Core::Get()->GetRenderDevice()->End();

	Core::Get()->GetRenderDevice()->SetPerspectiveProjection();

	Core::Get()->GetRenderDevice()->SetStencilFunc( RenderDevice::STENCILFUNC_ALWAYS, 1, 0xFFFFFFFF );
*/
	ppkCurOp = vpkCurOps.begin();

	for( uiOp = 0; uiOp < uiNumOps; ++uiOp, ++ppkCurOp )
	{
		(*ppkCurOp)->m_pkVertexBuffer  = 0;
		(*ppkCurOp)->m_pkPolygonBuffer = 0;

		// This makes buffer be cleaned up next frame or in dtor
		//(*ppkCurOp)->m_kPrimitive.m_pkVertexBuffer  = 0;
		//(*ppkCurOp)->m_kPrimitive.m_pkPolygonBuffer = 0;
	}
}

#endif


};
