/***************************************************************************
             skeleton.cpp  -  Skeleton as a hierarchy of bone nodes
                             -------------------
    begin                : Sun Jan 26 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, skeleton.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "skeleton.h"
#include "bone.h"
#include "adaptor.h"
#include "renderprimitive.h"
#include "sphere.h"
#include "scenenode.h"
#include "render.h"
#include "vertex.h"

using namespace std;


namespace NeoEngine
{


Skeleton::Skeleton() :
	SceneNode(),
	SkeletonAnimatorController(),
	m_pkRoot( 0 ),
	m_ppkBones( 0 ),
	m_uiNumBones( 0 ),
	m_bRenderSkeleton( false )
{
	SkeletonAnimatorController::m_pkSkeleton = this;
}


Skeleton::Skeleton( Skeleton &rkSkeleton ) :
	SceneNode( rkSkeleton, false ),
	SkeletonAnimatorController( rkSkeleton ),
	m_pkRoot( 0 ),
	m_ppkBones( new Bone*[ rkSkeleton.m_uiNumBones ] ),
	m_uiNumBones( rkSkeleton.m_uiNumBones ),
	m_bRenderSkeleton( rkSkeleton.m_bRenderSkeleton )
{
	SkeletonAnimatorController::m_pkSkeleton = this;

	unsigned int uiBone;

	for( uiBone = 0; uiBone < m_uiNumBones; ++uiBone )
	{
		m_ppkBones[ uiBone ] = new Bone( rkSkeleton.m_ppkBones[ uiBone ]->m_uiID, rkSkeleton.m_ppkBones[ uiBone ]->GetName() );

		m_ppkBones[ uiBone ]->SetTranslation( rkSkeleton.m_ppkBones[ uiBone ]->GetTranslation() );
		m_ppkBones[ uiBone ]->SetRotation(    rkSkeleton.m_ppkBones[ uiBone ]->GetRotation()    );
	}

	if( rkSkeleton.m_pkRoot )
		m_pkRoot = m_ppkBones[ rkSkeleton.m_pkRoot->m_uiID ];

	for( uiBone = 0; uiBone < m_uiNumBones; ++uiBone )
	{
		if( m_ppkBones[ uiBone ] != m_pkRoot )
			m_ppkBones[ rkSkeleton.m_ppkBones[ uiBone ]->GetParent()->m_uiID ]->AttachNode( m_ppkBones[ uiBone ], false );
	}

	if( m_pkRoot )
		m_pkRoot->NotifyUpdate( true );

	BoneAdaptor *pkAdaptor;

	vector< BoneAdaptor* >::iterator ppkAdaptor    = rkSkeleton.m_vpkAdaptors.begin();
	vector< BoneAdaptor* >::iterator ppkAdaptorEnd = rkSkeleton.m_vpkAdaptors.begin();

	for( ; ppkAdaptor != ppkAdaptorEnd; ++ppkAdaptor )
	{
		m_vpkAdaptors.push_back( ( pkAdaptor = (*ppkAdaptor)->Duplicate( this, m_ppkBones[ (*ppkAdaptor)->GetParent()->m_uiID ] ) ) );

		AttachNode( pkAdaptor->m_pkSceneNode );
	}
}


Skeleton::~Skeleton()
{
	unsigned int uiBone;

	for( uiBone = 0; uiBone < m_uiNumBones; ++uiBone )
		m_ppkBones[uiBone]->DetachFromParent();

	for( uiBone = 0; uiBone < m_uiNumBones; ++uiBone )
		delete m_ppkBones[uiBone];

	delete [] m_ppkBones;
}


void Skeleton::Update( float fDeltaTime )
{
	SkeletonAnimatorController::Update( fDeltaTime );

	m_pkRoot->NotifyUpdate( true );

	SceneNode::Update( fDeltaTime );
}


SceneNode *Skeleton::Duplicate()
{
	return new Skeleton( *this );
}


bool Skeleton::Render( Frustum *pkFrustum, bool bForce )
{
	if( !m_bRenderSkeleton )
		return SceneNode::Render( pkFrustum, bForce );

	RenderPrimitive kPrimitive;
	Sphere kSphere;

	kPrimitive.m_ePrimitive     = RenderPrimitive::LINES;
	kPrimitive.m_pkMaterial     = 0;

	kPrimitive.m_aUnknown[0].m_fData = 1.0f; //Line width

	unsigned int i;

	float *pfScale = new float[ m_uiNumBones ];

	for( i = 0; i < m_uiNumBones; ++i )
	{
		if( !m_ppkBones[i] )
			continue;

		const vector<Bone*> &vpkChildNodes = m_ppkBones[i]->GetChildren();

		float fMaxScale = 0.015f;

		for( unsigned int iChild = 0; iChild < vpkChildNodes.size(); ++iChild )
		{
			Vector3d kDiff  = vpkChildNodes[iChild]->GetTranslation();

			float fScale = 0.08f * kDiff.Len();

			if( fScale > fMaxScale )
				fMaxScale = fScale;
		}

		pfScale[i] = fMaxScale;
	}

	for( i = 0; i < m_uiNumBones; ++i )
	{
		if( !m_ppkBones[i] )
			continue;

		if( GetParent() )
			kPrimitive.m_kModelMatrix = GetParent()->GetWorldTransform() * m_ppkBones[i]->GetWorldTransform();
		else
			kPrimitive.m_kModelMatrix = m_ppkBones[i]->GetWorldTransform();

		const vector<Bone*> &vpkChildNodes = m_ppkBones[i]->GetChildren();

		for( unsigned int iChild = 0; iChild < vpkChildNodes.size(); ++iChild )
		{
			Vector3d kDiff  = vpkChildNodes[iChild]->GetTranslation();
			Vector3d kXAxis( kDiff % Vector3d( 0.0f, 1.0f, 0.0f ) );

			if( kXAxis.Len2() < 0.001f )
				kXAxis.Set( 1.0f, 0.0f, 0.0f );

			kXAxis.ScaleTo( pfScale[i] );

			Vector3d kYAxis = kXAxis % kDiff;

			kYAxis.ScaleTo( pfScale[i] );

			Vector3d kDstDiff = kDiff;

			kDiff.ScaleTo( pfScale[i] );
			kDstDiff.ScaleTo( pfScale[ ((Bone*)vpkChildNodes[iChild])->m_uiID ] );

			VertexBufferPtr pkVertexBuffer = Core::Get()->GetRenderDevice()->CreateVertexBuffer( Buffer::DYNAMIC, 18, &DiffuseVertex::s_kDecl );
			
			kPrimitive.m_pkVertexBuffer = pkVertexBuffer;

			pkVertexBuffer->Lock( Buffer::WRITE );

			DiffuseVertex *pkVertex = (DiffuseVertex*)pkVertexBuffer->GetVertex();

			Vector3d akPoints[5]   = {
		
				( kDiff + kXAxis ) + kYAxis,
				( kDiff - kXAxis ) + kYAxis,
				( kDiff - kXAxis ) - kYAxis,
				( kDiff + kXAxis ) - kYAxis,
				vpkChildNodes[iChild]->GetTranslation() - kDstDiff

			};

			pkVertex->m_kPosition = akPoints[0]; pkVertex->m_kColor.Set( 1.0f, 1.0f, 1.0f, 1.0f ); ++pkVertex;
			pkVertex->m_kPosition = akPoints[1]; pkVertex->m_kColor.Set( 1.0f, 1.0f, 1.0f, 1.0f ); ++pkVertex;

			pkVertex->m_kPosition = akPoints[1]; pkVertex->m_kColor.Set( 1.0f, 1.0f, 1.0f, 1.0f ); ++pkVertex;
			pkVertex->m_kPosition = akPoints[2]; pkVertex->m_kColor.Set( 1.0f, 1.0f, 1.0f, 1.0f ); ++pkVertex;

			pkVertex->m_kPosition = akPoints[2]; pkVertex->m_kColor.Set( 1.0f, 1.0f, 1.0f, 1.0f ); ++pkVertex;
			pkVertex->m_kPosition = akPoints[3]; pkVertex->m_kColor.Set( 1.0f, 1.0f, 1.0f, 1.0f ); ++pkVertex;

			pkVertex->m_kPosition = akPoints[3]; pkVertex->m_kColor.Set( 1.0f, 1.0f, 1.0f, 1.0f ); ++pkVertex;
			pkVertex->m_kPosition = akPoints[0]; pkVertex->m_kColor.Set( 1.0f, 1.0f, 1.0f, 1.0f ); ++pkVertex;

			pkVertex->m_kPosition = akPoints[0]; pkVertex->m_kColor.Set( 1.0f, 1.0f, 1.0f, 1.0f ); ++pkVertex;
			pkVertex->m_kPosition = akPoints[4]; pkVertex->m_kColor.Set( 1.0f, 1.0f, 1.0f, 1.0f ); ++pkVertex;

			pkVertex->m_kPosition = akPoints[1]; pkVertex->m_kColor.Set( 1.0f, 1.0f, 1.0f, 1.0f ); ++pkVertex;
			pkVertex->m_kPosition = akPoints[4]; pkVertex->m_kColor.Set( 1.0f, 1.0f, 1.0f, 1.0f ); ++pkVertex;

			pkVertex->m_kPosition = akPoints[2]; pkVertex->m_kColor.Set( 1.0f, 1.0f, 1.0f, 1.0f ); ++pkVertex;
			pkVertex->m_kPosition = akPoints[4]; pkVertex->m_kColor.Set( 1.0f, 1.0f, 1.0f, 1.0f ); ++pkVertex;

			pkVertex->m_kPosition = akPoints[3]; pkVertex->m_kColor.Set( 1.0f, 1.0f, 1.0f, 1.0f ); ++pkVertex;
			pkVertex->m_kPosition = akPoints[4]; pkVertex->m_kColor.Set( 1.0f, 1.0f, 1.0f, 1.0f ); ++pkVertex;

			pkVertex->m_kPosition = Vector3d::ORIGO; pkVertex->m_kColor.Set( 1.0f, 1.0f, 1.0f, 1.0f ); ++pkVertex;
			pkVertex->m_kPosition = vpkChildNodes[iChild]->GetTranslation(); pkVertex->m_kColor.Set( 1.0f, 1.0f, 1.0f, 1.0f );

			pkVertexBuffer->Unlock();
			
			kPrimitive.m_uiNumPrimitives = 9;

			Core::Get()->GetRenderDevice()->Render( kPrimitive, 0 );
		}

		if( GetParent() )
		{
			kSphere.SetTranslation( ( GetParent()->GetWorldRotation() * m_ppkBones[i]->GetWorldTranslation() ) + GetParent()->GetWorldTranslation() );
			kSphere.SetRotation( GetParent()->GetWorldRotation() * m_ppkBones[i]->GetWorldRotation() );
		}
		else
		{
			kSphere.SetTranslation( m_ppkBones[i]->GetWorldTranslation() );
			kSphere.SetRotation( m_ppkBones[i]->GetWorldRotation() );
		}

		kSphere.m_fRadius = pfScale[i];

		kSphere.RenderOutlines();
	}

	delete [] pfScale;

	return SceneNode::Render( pkFrustum, bForce );
}


void Skeleton::AttachAdaptor( BoneAdaptor *pkAdaptor )
{
	vector< BoneAdaptor* >::iterator ppkAdaptor    = m_vpkAdaptors.begin();
	vector< BoneAdaptor* >::iterator ppkAdaptorEnd = m_vpkAdaptors.begin();

	for( ; ppkAdaptor != ppkAdaptorEnd; ++ppkAdaptor )
		if( (*ppkAdaptor) == pkAdaptor )
			return;

	m_vpkAdaptors.push_back( pkAdaptor );

	AttachNode( pkAdaptor->m_pkSceneNode );
}


void Skeleton::DetachAdaptor( BoneAdaptor *pkAdaptor )
{
	vector< BoneAdaptor* >::iterator ppkAdaptor    = m_vpkAdaptors.begin();
	vector< BoneAdaptor* >::iterator ppkAdaptorEnd = m_vpkAdaptors.begin();

	for( ; ppkAdaptor != ppkAdaptorEnd; ++ppkAdaptor )
		if( (*ppkAdaptor) == pkAdaptor )
		{
			m_vpkAdaptors.erase( ppkAdaptor );
			DetachNode( pkAdaptor->m_pkSceneNode );
			return;
		}
}


};
