/***************************************************************************
                       skin.cpp  -  Skeletal submeshes
                             -------------------
    begin                : Wed Sep 18 2002
    copyright            : (C) 2002 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, skin.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "skin.h"
#include "nemath.h"
#include "vertex.h"
#include "polygon.h"
#include "bone.h"
#include "skeleton.h"
#include "render.h"
#include "util.h"


namespace NeoEngine
{


SkinVertex::~SkinVertex()
{
	delete [] m_pkInfluences;
}


Skin::~Skin()
{
	delete [] m_pkSkinVertices;
}




SkeletalSubMesh::SkeletalSubMesh() :
	SubMesh()
{
	m_pkSkin             = 0;
	m_pkSkeleton         = 0;
	m_bNeedUpdate        = true;
}


SkeletalSubMesh::SkeletalSubMesh( const SkeletalSubMesh &rkSubMesh ) :
	SubMesh( rkSubMesh )
{
	m_pkVertices         = 0;
	m_pkSkin             = rkSubMesh.m_pkSkin;
	m_pkSkeleton         = 0;
	m_bNeedUpdate        = true;
}


SkeletalSubMesh::~SkeletalSubMesh()
{
}


void SkeletalSubMesh::WeightVertices()
{
	if( !m_pkSkeleton || !m_pkSkin )
		return;

	if( !m_pkVertices )
	{
		m_pkSkin->m_pkVertices->Lock( Buffer::READ );

		m_pkVertices = Core::Get()->GetRenderDevice()->CreateVertexBuffer( Buffer::DYNAMIC | Buffer::WRITEPRIORITIZED, m_pkSkin->m_pkVertices->GetNumElements(), m_pkSkin->m_pkVertices->GetVertexFormat(), m_pkSkin->m_pkVertices->GetVertex() );

		m_pkSkin->m_pkVertices->Unlock();
	}

	Vector3d             kFinal;
	SkinVertex          *pkSkinVertex;
	BoneInfluence       *pkInfluence;
	const VertexElement *pkNormalElement   = m_pkVertices->GetVertexFormat()->GetElement( VertexElement::FLOAT3,      VertexElement::NORMAL   );
	const VertexElement *pkTexCoordElement = m_pkVertices->GetVertexFormat()->GetElement( VertexElement::UNKNOWNTYPE, VertexElement::TEXCOORD );
	const VertexElement *pkTangentElement  = m_pkVertices->GetVertexFormat()->GetElement( VertexElement::UNKNOWNTYPE, VertexElement::TANGENT  );
	const VertexElement *pkBinormalElement = m_pkVertices->GetVertexFormat()->GetElement( VertexElement::UNKNOWNTYPE, VertexElement::BINORMAL );

	int iBone        = 0;
	int i            = 0;
	int iEnd         = m_pkVertices->GetNumElements();

	m_pkVertices->Lock( Buffer::WRITE );

	unsigned char   *pucVertex    = (unsigned char*)m_pkVertices->GetVertex();
	unsigned int     uiVertexSize = m_pkVertices->GetVertexSize();

	// ### FIXME: Assuming position is at offset 0 for easy normal access below, and during tangent calculation
	//pucVertex += m_pkVertices->GetVertexFormat()->GetElement( VertexElement::FLOAT3, VertexElement::POSITION )->m_uiOffset;

	//Loop through all vertices and weight them
	for( i = 0; i < iEnd; ++i, pucVertex += uiVertexSize )
	{
		pkSkinVertex = &m_pkSkin->m_pkSkinVertices[i];

		kFinal.Reset();

		pkInfluence = pkSkinVertex->m_pkInfluences;

		for( iBone = 0; iBone < pkSkinVertex->m_iNumInfluences; ++iBone, ++pkInfluence )
			kFinal += ( m_pkSkeleton->m_ppkBones[ pkInfluence->m_iBoneID ]->GetWorldTransform() * pkInfluence->m_kOffset ) * pkInfluence->m_fWeight;
		
		*((Vector3d*)pucVertex) = kFinal;

		if( pkNormalElement )
			((Vector3d*)( pucVertex + pkNormalElement->m_uiOffset ))->Reset();
	}

	m_pkPolygons->Lock( Buffer::READ );

	if( pkNormalElement )
	{
		m_pkPolygons->CalculateNormals( m_pkVertices, true );

		Polygon  *pkPolygon  = m_pkPolygons->GetPolygon();
		Vector3d *pkNormal   = m_pkPolygons->GetNormals();

		int iPolyEnd = m_pkPolygons->GetNumElements();

		pucVertex    = (unsigned char*)m_pkVertices->GetVertex() + pkNormalElement->m_uiOffset;

		for( i = 0; i < iPolyEnd; ++i, ++pkPolygon, ++pkNormal )
		{
			*((Vector3d*)( pucVertex + uiVertexSize * pkPolygon->v[0] )) += *pkNormal;
			*((Vector3d*)( pucVertex + uiVertexSize * pkPolygon->v[1] )) += *pkNormal;
			*((Vector3d*)( pucVertex + uiVertexSize * pkPolygon->v[2] )) += *pkNormal;
		}

		pucVertex    = (unsigned char*)m_pkVertices->GetVertex() + pkNormalElement->m_uiOffset;

		for( i = 0; i < iEnd; ++i, pucVertex += uiVertexSize )
			((Vector3d*)pucVertex)->Normalize();
	}

	if( pkTangentElement || pkBinormalElement )
		m_pkPolygons->CalculateTangents( m_pkVertices, pkNormalElement, pkTexCoordElement, pkTangentElement, pkBinormalElement );

	m_pkPolygons->Unlock();
	m_pkVertices->Unlock();

	m_bChanged = true;
}


SubMesh *SkeletalSubMesh::Duplicate() const
{
	return( new SkeletalSubMesh( *this ) );
}


};
