/***************************************************************************
                 skydome.cpp  -  Skydome class implementation
                             -------------------
    begin                : Sat Dec 14 2002
    copyright            : (C) 2002 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, skydome.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "skydome.h"
#include "vertex.h"
#include "polygon.h"
#include "renderprimitive.h"
#include "render.h"

using namespace std;


namespace NeoEngine
{


SkyDome::SkyDome() :
	SceneEntity()
{
}


SkyDome::~SkyDome()
{
}


void SkyDome::Set( const Quaternion &rkOrientation, float fDistance, float fCurvature, int iSegments, float fTiles )
{
	m_kOrientation = rkOrientation;

	Vector3d kX, kY, kZ;

	//Create the dome
	for( int iPlane = 0; iPlane < NUMPLANES; ++iPlane )
	{
		//Make sure zbuffer mode is set correctly
		//We are rendered in sky mode, which means we will have depth range
		//set to always map to far clip plane
		if( m_apkMaterial[iPlane] )
			m_apkMaterial[iPlane]->m_kZBufferMode.Set( ZBufferMode::LEQUALNOWRITE );

		switch( iPlane )
		{
			case FRONT:
				kX.Set(  1.0f,  0.0f,  0.0f );
				kY.Set(  0.0f,  1.0f,  0.0f );
				kZ.Set(  0.0f,  0.0f,  1.0f );
				break;

			case BACK:
				kX.Set( -1.0f,  0.0f,  0.0f );
				kY.Set(  0.0f,  1.0f,  0.0f );
				kZ.Set(  0.0f,  0.0f, -1.0f );
				break;

			case LEFT:
				kX.Set(  0.0f,  0.0f, -1.0f );
				kY.Set(  0.0f,  1.0f,  0.0f );
				kZ.Set(  1.0f,  0.0f,  0.0f );
				break;

			case RIGHT:
				kX.Set(  0.0f,  0.0f,  1.0f );
				kY.Set(  0.0f,  1.0f,  0.0f );
				kZ.Set( -1.0f,  0.0f,  0.0f );
				break;

			case UP:
				kX.Set(  1.0f,  0.0f,  0.0f );
				kY.Set(  0.0f,  0.0f,  1.0f );
				kZ.Set(  0.0f, -1.0f,  0.0f );
				break;

			default:
				assert( false );
				break;
		}

		//Create dome plane
		float fSize = fDistance * 2.0f;

		int iNumVertices = ( iSegments + 1 ) * ( iSegments + 1 );

		//Create vertices
		if( !m_apkVertexBuffer[iPlane] )
			m_apkVertexBuffer[iPlane] = Core::Get()->GetRenderDevice()->CreateVertexBuffer( Buffer::STATIC, iNumVertices, &TexVertex::s_kDecl );
		else
			m_apkVertexBuffer[iPlane]->AllocateVertices( iNumVertices, &TexVertex::s_kDecl );

		float fOOSeg    = 1.0f / float( iSegments );
		float fHalfSize = fSize * 0.5f;
		float fStep     = fSize * fOOSeg;

		m_apkVertexBuffer[iPlane]->Lock( Buffer::WRITE );

		TexVertex *pkVertex = (TexVertex*)m_apkVertexBuffer[iPlane]->GetVertex();

		Vector3d  kVec;

		float fSphereDist;
		float fRadius =  100.0f - fCurvature;
		float fCamPos = fRadius - 5.0f;

		int iY;
		int iX;

		for( iY = 0; iY < ( iSegments + 1 ); ++iY )
		{
			for( iX = 0; iX < ( iSegments + 1 ); ++iX, ++pkVertex )
			{
				kVec  = kX * ( float( iX ) * fStep - fHalfSize );
				kVec += kY * ( float( iY ) * fStep - fHalfSize );
				kVec += kZ * -fDistance;

				pkVertex->m_kPosition = kVec;

				kVec.Normalize();

				fSphereDist = sqrtf( fCamPos * fCamPos * ( kVec.y * kVec.y - 1.0f ) + fRadius * fRadius ) - ( fCamPos * kVec.y );

				kVec.x *= fSphereDist;
				kVec.z *= fSphereDist;

				pkVertex->m_afTexCoord[0] = kVec.x * 0.01f * fTiles;
				pkVertex->m_afTexCoord[1] = kVec.z * 0.01f * fTiles;
			}
		}

		m_apkVertexBuffer[iPlane]->Unlock();
	}

	//Create polygons
	if( !m_pkPolygonBuffer )
		m_pkPolygonBuffer = Core::Get()->GetRenderDevice()->CreatePolygonBuffer( Buffer::STATIC, iSegments * iSegments * 2 );
	else
		m_pkPolygonBuffer->AllocatePolygons( iSegments * iSegments * 2 );

	m_pkPolygonBuffer->Lock( Buffer::WRITE );

	Polygon *pkPoly = m_pkPolygonBuffer->GetPolygon();
		
	for( int iY = 0; iY < iSegments; ++iY )
		for( int iX = 0; iX < iSegments; ++iX )
		{
			pkPoly->v[0] =   iX       +   iY       * ( iSegments + 1 );
			pkPoly->v[1] = ( iX + 1 ) +   iY       * ( iSegments + 1 );
			pkPoly->v[2] =   iX       + ( iY + 1 ) * ( iSegments + 1 );

			++pkPoly;

			pkPoly->v[0] =   iX       + ( iY + 1 ) * ( iSegments + 1 );
			pkPoly->v[1] = ( iX + 1 ) +   iY       * ( iSegments + 1 );
			pkPoly->v[2] = ( iX + 1 ) + ( iY + 1 ) * ( iSegments + 1 );

			++pkPoly;
		}

	m_pkPolygonBuffer->Unlock();

	m_pkPolygonBuffer->Stripify();
}


void SkyDome::SetMaterial( SKYDOMEPLANE ePlane, MaterialPtr &pkMaterial )
{
	m_apkMaterial[ePlane] = pkMaterial;

	m_apkMaterial[ePlane]->m_kZBufferMode.Set( ZBufferMode::LEQUALNOWRITE );
}


bool SkyDome::Render( Frustum *pkFrustum, bool bForce )
{
	if( !SceneEntity::Render( pkFrustum, bForce ) )
		return false;

	RenderPrimitive kData;

	kData.m_kModelMatrix             = m_kOrientation;

	if( !m_pkPolygonBuffer->IsStripped() )
	{
		kData.m_ePrimitive           = RenderPrimitive::TRIANGLES;
		kData.m_pkPolygonBuffer      = m_pkPolygonBuffer;
		kData.m_uiNumPrimitives      = m_pkPolygonBuffer->GetNumElements();
	}
	else
	{
		kData.m_ePrimitive           = RenderPrimitive::TRIANGLESTRIP;
		kData.m_pkPolygonStripBuffer = m_pkPolygonBuffer->GetStrip();
		kData.m_uiNumPrimitives      = kData.m_pkPolygonStripBuffer->GetNumElements();
	}

	for( int iPlane = 0; iPlane < NUMPLANES; ++iPlane )
	{
		if( !m_apkMaterial[iPlane] )
			continue;

		kData.m_pkMaterial      = m_apkMaterial[iPlane];
		kData.m_pkVertexBuffer  = m_apkVertexBuffer[iPlane];

		Core::Get()->GetRenderDevice()->Render( kData, RenderPrimitive::NOTRANSLATION );
	}

	return true;
}


};

