/***************************************************************************
                     sphere.cpp  -  Sphere bounding volume
                             -------------------
    begin                : Wed Sep 18 2002
    copyright            : (C) 2002 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, sphere.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "sphere.h"
#include "line.h"
#include "plane.h"
#include "aabb.h"
#include "obb.h"
#include "capsule.h"
#include "render.h"
#include "renderprimitive.h"
#include "vertex.h"
#include "contact.h"

#include <limits>

#include <assert.h>

using namespace std;
using namespace NeoEngine;


namespace NeoEngine
{


Sphere::Sphere( float fRadius ) :
	BoundingVolume()
{
	m_fRadius = fRadius;
}


bool Sphere::Intersection( const Vector3d &rkPoint )
{
	return( ( m_kTranslation - rkPoint ).Len2() <= ( m_fRadius * m_fRadius ) );
}


bool Sphere::Intersection( const Line &rkLine )
{
	Vector3d kDiff = m_kTranslation - rkLine.m_kOrigin;
	Vector3d kDir  = rkLine.m_kDirection;

	kDir.Normalize();

	kDiff = kDiff - ( kDir * ( kDiff * kDir ) );

	return( kDiff.Len2() <= ( m_fRadius * m_fRadius ) );
}


bool Sphere::Intersection( const Plane &rkPlane )
{
	return( fabsf( rkPlane.Distance( m_kTranslation ) ) < m_fRadius );
}


bool Sphere::Intersection( const Vector3d &rkV0, const Vector3d &rkV1, const Vector3d &rkV2, ContactSet *pkContactSet, bool bForceTriNormal, const Vector3d &rkNormal )
{
	Vector3d kEdge[3]  = { rkV1 - rkV0, rkV2 - rkV1, rkV0 - rkV2 };
	Vector3d kEdgeOffset[3];
	Vector3d kEdgeNormal[3];
	Vector3d kPos = m_kTranslation;

	Vector3d kPlaneNormal = kEdge[0] % kEdge[1];

	float fRadSqr = m_fRadius * m_fRadius;

	//See if sphere touches plane
	kEdgeOffset[0] = kPos - rkV0;
	
	float fE[3];
	float fS      = kEdgeOffset[0] * kPlaneNormal;
	float fFactor = kPlaneNormal.Len2() * fRadSqr;

	if( fS * fS > fFactor )
		return false;

	// See if point touches prism
	kEdgeNormal[0] = kEdge[0] % kPlaneNormal;
	fE[0]   = kEdgeNormal[0] * kEdgeOffset[0];
	fFactor = kEdgeNormal[0].Len2() * fRadSqr;

	if( ( fE[0] > 0.0f ) && ( fE[0] * fE[0] > fFactor ) )
		return false;

	kEdgeNormal[1] = kEdge[1] % kPlaneNormal;
	kEdgeOffset[1] = kPos - rkV1;
	fE[1]   = kEdgeNormal[1] * kEdgeOffset[1];
	fFactor = kEdgeNormal[1].Len2() * fRadSqr;

	if( ( fE[1] > 0.0f ) && ( fE[1] * fE[1] > fFactor ) )
		return false;

	kEdgeNormal[2] = kEdge[2] % kPlaneNormal;
	kEdgeOffset[2] = kPos - rkV2;
	fE[2]   = kEdgeNormal[2] * kEdgeOffset[2];
	fFactor = kEdgeNormal[2].Len2() * fRadSqr;

	if( ( fE[2] > 0.0f ) && ( fE[2] * fE[2] > fFactor ) )
		return false;

	// Check if projected sphere center is within triangle
	if( ( fE[0] <= 0.0f ) && ( fE[1] <= 0.0f ) && ( fE[2] <= 0.0f ) )
	{
		if( pkContactSet )
		{
			kPlaneNormal.Normalize();

			Contact *pkContact = new Contact;

			pkContact->m_kNormal = kPlaneNormal;

			pkContact->m_pfDepths[0] = kPlaneNormal * kEdgeOffset[0];
			pkContact->m_pkPoints[0] = kPos + kPlaneNormal * -pkContact->m_pfDepths[0];

			pkContact->m_uiNumPoints = 1;

			pkContactSet->Add( pkContact );
		}

		return true;
	}

	// Perform the sphere/edge test.
	if( fE[0] > 0.0f )
	{
		// Project on edge line
		float fT = ( kEdgeOffset[0] * kEdge[0] ) / kEdge[0].Len();

		if( fT < 0.0f )
			fT = 0.0f;
		else if( ( fE[1] <= 0.0f ) && ( fT > 1.0f ) )
			fT = 1.0f;
		
		if( fT <= 1.0f )
		{
			Vector3d kPoint   = rkV0 + ( kEdge[0] * fT );
			Vector3d kDiff    = kPoint - kPos;
			float fDiffSqrLen = kDiff.Len2();

			if( fDiffSqrLen > fRadSqr )
				return false;

			//We have intersection
			if( pkContactSet )
			{
				kPlaneNormal.Normalize();

				float fDenom = kDiff * kPlaneNormal;
				float fDepth = fDenom + sqrtf( ( fDenom * fDenom ) + ( fRadSqr - fDiffSqrLen ) );

				Contact *pkContact = new Contact;

				pkContact->m_kNormal = kPlaneNormal;

				pkContact->m_pkPoints[0] = kPoint;
				pkContact->m_pfDepths[0] = fDepth;

				pkContact->m_uiNumPoints = 1;

				pkContactSet->Add( pkContact );
			}

			return true;
		}
	}

	if( fE[1] > 0.0f )
	{
		// Project on edge line
		float fT = ( kEdgeOffset[1] * kEdge[1] ) / kEdge[1].Len();

		if( fT < 0.0f )
			fT = 0.0f;
		else if( ( fE[2] <= 0.0f ) && ( fT > 1.0f ) )
			fT = 1.0f;
		
		if( fT <= 1.0f )
		{
			Vector3d kPoint = rkV1 + ( kEdge[1] * fT );
			Vector3d kDiff  = kPoint - kPos;
			float fDiffSqrLen = kDiff.Len2();

			if( fDiffSqrLen > fRadSqr )
				return false;

			//We have intersection
			if( pkContactSet )
			{
				kPlaneNormal.Normalize();

				float fDenom = kDiff * kPlaneNormal;
				float fDepth = fDenom + sqrtf( ( fDenom * fDenom ) + ( fRadSqr - fDiffSqrLen ) );

				Contact *pkContact = new Contact;

				pkContact->m_kNormal = kPlaneNormal;

				pkContact->m_pkPoints[0] = kPoint;
				pkContact->m_pfDepths[0] = fDepth;

				pkContact->m_uiNumPoints = 1;

				pkContactSet->Add( pkContact );
			}

			return true;
		}
	}

	if( fE[2] > 0.0f )
	{
		// Project on edge line
		float fT = ( kEdgeOffset[2] * kEdge[2] ) / kEdge[2].Len();

		if( fT < 0.0f )
			fT = 0.0f;
		else if( ( fE[0] <= 0.0f ) && ( fT > 1.0f ) )
			fT = 1.0f;
		
		if( fT <= 1.0f )
		{
			Vector3d kPoint = rkV2 + ( kEdge[2] * fT );
			Vector3d kDiff  = kPoint - kPos;
			float fDiffSqrLen = kDiff.Len2();

			if( fDiffSqrLen > fRadSqr )
				return false;

			//We have intersection
			if( pkContactSet )
			{
				kPlaneNormal.Normalize();

				float fDenom = kDiff * kPlaneNormal;
				float fDepth = fDenom + sqrtf( ( fDenom * fDenom ) + ( fRadSqr - fDiffSqrLen ) );

				Contact *pkContact = new Contact;

				pkContact->m_kNormal = kPlaneNormal;

				pkContact->m_pkPoints[0] = kPoint;
				pkContact->m_pfDepths[0] = fDepth;

				pkContact->m_uiNumPoints = 1;

				pkContactSet->Add( pkContact );
			}

			return true;
		}
	}

	assert( !"This should not happen" );

	return false;
}


bool Sphere::Intersection( const Ray &rkRay, ContactSet *pkContactSet )
{
	Vector3d kSpherePos   = m_kTranslation;
	Vector3d kRayDir      = rkRay.m_kDirection;
	Vector3d kRayStart    = rkRay.m_kOrigin;

	//Impact now between 0 and 1
	Vector3d kOffset      = kRayStart - kSpherePos;
	float fRadSqr         = m_fRadius * m_fRadius;
	float fOffSqr         = kOffset.Len2();

	if( fOffSqr < fRadSqr )
	{
		//Inside at start
		if( pkContactSet )
		{
			Contact *pkContact = new Contact;

			pkContact->m_pkPoints[0] = kRayStart;
			pkContact->m_pfDepths[0] = 0.0f;

			pkContact->m_uiNumPoints = 1;

			pkContactSet->Add( pkContact );
		}

		return true;
	}

	float fRayProj = kRayDir * kOffset;

	if( fRayProj >= 0.0f )
		return false; //Moving away from sphere

	float fA = 1.0f / kRayDir.Len2();
	float fB = 2.0f * fRayProj * fA;
	float fC = ( fOffSqr - fRadSqr ) * fA;

	float fDisc = fB * fB - 4.0f * fC;

	if( fDisc < 0.0f )
		return false;

	float fSqr = sqrtf( fDisc );
	float fT = NeoEngine::MIN( ( fB + fSqr ) * -0.5f, ( fB - fSqr ) * -0.5f );

	if( fT < 0.0f ) { assert( !"This should not happen, ray casting away from sphere" ); }

	if( fT > 1.0f )
		return false;

	if( pkContactSet )
	{
		Contact *pkContact = new Contact;

		pkContact->m_pkPoints[0] = kRayStart + ( kRayDir * fT );
		pkContact->m_pfDepths[0] = fT;

		pkContact->m_uiNumPoints = 1;

		pkContactSet->Add( pkContact );
	}

	return true;
}

/*
bool Sphere::Collision( const Vector3d &rkV0, const Vector3d &rkV1, const Vector3d &rkV2, const Vector3d &rkThisVel, const Vector3d &rkPolyVel, const Vector3d &rkThisAngularVel, const Vector3d &rkPolygonAngularVel, float fMinT, float fMaxT, ContactSet *pkContactSet, const Vector3d &rkNormal )
{
	if( rkPolygonAngularVel != Vector3d::ZERO )
	{
		assert( !"Implement" );
		return false;
	}

	Vector3d kVel = rkThisVel - rkPolyVel;
	Vector3d kStartPos = m_kTranslation + ( kVel * fMinT );
	Vector3d kEndPos   = m_kTranslation + ( kVel * fMaxT );

	Vector3d kEdge[3]  = { rkV1 - rkV0, rkV2 - rkV1, rkV0 - rkV2 };
	Vector3d kEdgeOffset[3];
	Vector3d kEdgeNormal[3];

	Vector3d kPlaneNormal = kEdge[0] % kEdge[1]; //kPlaneNormal.Normalize();

	float fRadSqr = m_fRadius * m_fRadius;

	//See if it touches plane in given t range
	kEdgeOffset[0] = kStartPos - rkV0;
	float fS = ( kEdgeOffset[0]  ) * kPlaneNormal;
	float fE = ( kEndPos - rkV0  ) * kPlaneNormal;
	float fFactor = kPlaneNormal.Len2() * fRadSqr;

	if( ( ( ( fS > 0.0f ) && ( fE > 0.0f ) ) || ( ( fS < 0.0f ) && ( fE < 0.0f ) ) ) &&
	    ( fS * fS > fFactor ) && ( fE * fE > fFactor ) )
		return false; //Sphere does not intersect plane in given t range

	//See if it touches prism in given t range
	kEdgeNormal[0] = kEdge[0] % kPlaneNormal;
	fS = ( kEdgeOffset[0]   ) * kEdgeNormal[0];
	fE = ( kEndPos   - rkV0 ) * kEdgeNormal[0];
	fFactor = kEdgeNormal[0].Len2() * fRadSqr;

	if( ( fS > 0.0f ) && ( fE > 0.0f ) && ( fS * fS > fFactor ) && ( fE * fE > fFactor ) )
		return false;

	kEdgeNormal[1] = kEdge[1] % kPlaneNormal;
	kEdgeOffset[1] = kStartPos - rkV1;
	fS = ( kEdgeOffset[1]   ) * kEdgeNormal[1];
	fE = ( kEndPos   - rkV1 ) * kEdgeNormal[1];
	fFactor = kEdgeNormal[1].Len2() * fRadSqr;

	if( ( fS > 0.0f ) && ( fE > 0.0f ) && ( fS * fS > fFactor ) && ( fE * fE > fFactor ) )
		return false;

	kEdgeNormal[2] = kEdge[2] % kPlaneNormal;
	kEdgeOffset[2] = kStartPos - rkV2;
	fS = ( kEdgeOffset[2]   ) * kEdgeNormal[2];
	fE = ( kEndPos   - rkV2 ) * kEdgeNormal[2];
	fFactor = kEdgeNormal[2].Len2() * fRadSqr;

	if( ( fS > 0.0f ) && ( fE > 0.0f ) && ( fS * fS > fFactor ) && ( fE * fE > fFactor ) )
		return false;

	kPlaneNormal.Normalize();

	//Calculate point of intersection with plane
	float fVelProj   = kVel * kPlaneNormal;
	float fPosProj   = kStartPos * kPlaneNormal;

	float fT0 = (  m_fRadius - fPosProj ) / fVelProj;
	float fT1 = ( -m_fRadius - fPosProj ) / fVelProj;

	float fIntTMin = Core::MIN( fT0, fT1 );

	if( fIntTMin <= 0.0f )
	{
		//Sphere intersects plane at start
		assert( !"Implement intersect-at-start test for sphere/polygon collision" );
	}

	assert( fIntTMin <= fMaxT ); //Otherwise the test for intersect with plane in given range would have failed

	Vector3d kIntPt = kStartPos + ( kVel * fIntTMin );

	kIntPt = kIntPt + ( kPlaneNormal * ( kPlaneNormal * ( rkV0 - kIntPt ) ) );

	if( ( ( kIntPt - rkV0 ) * kEdgeNormal[0] <= 0.0f ) && ( ( kIntPt - rkV1 ) * kEdgeNormal[1] <= 0.0f ) && ( ( kIntPt - rkV2 ) * kEdgeNormal[2] <= 0.0f ) )
	{
		//Collision with face of polygon
		if( pkContactSet )
		{
			Contact *pkData = new Contact;

			pkData->m_fT = fMinT + fIntTMin * ( fMaxT - fMinT );
			pkData->m_vkPoints.push_back( kIntPt );

			pkContactSet->Add( pkData );
		}

		return true;
	}

	//We might have a edge collision, test
	Vector3d kPath    = kEndPos - kStartPos;
	float      fPathSqr = kPath.Len2();
	
	Vector3d kCollPos;
	float      fMinCollTime = 1.0001f;
	bool       bColl        = false;

	Vector3d kTempHit;

	// OPTIMIZE
	//
	// Only check potential edges by check which edges the collision point with plane is "outside"
	// Check MGC source for reference
								
	for( int iEdge = 0; iEdge < 3; ++iEdge )
	{
		Vector3d kVert = ( iEdge == 0 ) ? rkV0 : ( ( iEdge == 1 ) ? rkV1 : rkV2 );

		float fOffsetDotEdge = kEdgeOffset[iEdge] * kEdge[iEdge];
		float fOffsetDotPath = kEdgeOffset[iEdge] * kPath;
		float fOffsetSqr = kEdgeOffset[iEdge].Len2();
		float fEdgeDotPath = kEdge[iEdge] * kPath;
		float fEdgeSqr = kEdge[iEdge].Len2();

		// position of the collision along the edge is given by: xe = kVert + kEdge[iEdge] * s, where s is
		//  in the range [0,1].  position of sphere along its path is given by:
		//  xs = kStartPos + kPath * t, where t is in the range [0,1].  t is time, but s is arbitrary.
		//
		// solve simultaneous equations
		// (1) distance between edge and sphere center must be sphere radius
		// (2) line between sphere center and edge must be perpendicular to edge
		//
		// (1) (xe - xs)*(xe - xs) = m_fRadius*m_fRadius
		// (2) (xe - xs) * kEdge[iEdge] = 0
		//
		// then we do the magic math stuff :)

		float A, B, C, fRoot, fDiscriminant;
		float fRoot1 = 0.0f;
		float fRoot2 = 0.0f;
		float fDivisor;
		
		A = fEdgeDotPath * fEdgeDotPath - fEdgeSqr * fPathSqr;
		B = 2.0f * ( fOffsetDotEdge * fEdgeDotPath - fOffsetDotPath * fEdgeSqr );
		C = fOffsetDotEdge * fOffsetDotEdge + m_fRadius * m_fRadius * fEdgeSqr - fOffsetSqr * fEdgeSqr;

		if( ( A > Core::EPSILON ) && ( A < Core::EPSILON ) )
		{
			// degenerate case, sphere is traveling parallel to edge,
			// but no need to check vertex, other edge will catch this case
			continue; 
		}
		else
		{
			fDiscriminant = B * B - 4.0f * A * C;
			
			if( fDiscriminant > 0.0f )
			{
				fRoot    = sqrtf( fDiscriminant );
				fDivisor = 0.5f / A;
				
				fRoot1 = ( -B + fRoot ) * fDivisor;
				fRoot2 = ( -B - fRoot ) * fDivisor;

				fRoot = Core::MIN( fRoot1, fRoot2 );

				// root1 is a time, check that it's in our currently valid range
				if( ( fRoot < 0.0f ) || ( fRoot >= fMinCollTime ) )
					continue;

				// find sphere and edge positions
				kTempHit = kStartPos + kPath * fRoot;

				// check if hit is between kVert and v1
				fS = ( ( kTempHit - kVert ) * kEdge[iEdge] ) / fEdgeSqr;
				
				if( ( fS >= 0.0f ) && ( fS <= 1.0f ) )
				{
					if( !pkContactSet )
						return true; //We don't need to iterate other edges to find lower t match

					bColl        = true;
					fMinCollTime = fRoot;
					kCollPos     = kVert + kEdge[iEdge] * fS;
					
					continue;
				}
			}
			else
			{
				// discriminant negative, sphere passed edge too far away
				continue;
			}
		}

		//No need to check vertex, other edge will catch this case
	}

	if( bColl )
	{
		Contact *pkData = new Contact;

		pkData->m_fT = fMinT + ( fMaxT - fMinT ) * fMinCollTime;
		pkData->m_vkPoints.push_back( kCollPos );

		pkContactSet->Add( pkData );
	}

	return bColl;
}
*/


void Sphere::Generate( AABB *pkAABB )
{
	SetTranslation( pkAABB->GetTranslation() );

	m_fRadius = pkAABB->GetRadius();
}


void Sphere::Generate( OBB *pkOBB )
{
	SetTranslation( pkOBB->GetTranslation() );

	m_fRadius = pkOBB->GetRadius();
}


void Sphere::Generate( Sphere *pkSphere )
{
	SetTranslation( pkSphere->GetTranslation() );

	m_fRadius = pkSphere->m_fRadius;
}


void Sphere::Generate( Capsule *pkCapsule )
{
	SetTranslation( pkCapsule->GetTranslation() );

	m_fRadius = pkCapsule->m_fLength + pkCapsule->m_fRadius;
}


void Sphere::Generate( VertexBufferPtr &pkVertexBuffer )
{
	AABB kAABB;

	kAABB.Generate( pkVertexBuffer );

	Generate( &kAABB );
}


//Helper method
inline void MergeSpheres( Sphere *pkDst, const Vector3d &rkPos, float fRadius )
{
	Vector3d kCenterDiff = rkPos - pkDst->GetTranslation();

	float fRadiusDiff    = pkDst->m_fRadius - fRadius;
	float fRadiusDiffSqr = fRadiusDiff * fRadiusDiff;

	float fCenterDiffSqr = kCenterDiff.Len2();

	if( fRadiusDiffSqr >= fCenterDiffSqr )
	{
		if( fRadiusDiff < 0.0f )
		{
			pkDst->SetTranslation( rkPos );
			pkDst->m_fRadius = fRadius;
		}
		//else keep current sphere
	}
	else
	{
		float fLen    = sqrtf( fCenterDiffSqr );
		float fFactor = ( fLen - fRadiusDiff ) * 0.5f;

		pkDst->SetTranslation( pkDst->GetTranslation() + ( kCenterDiff * ( fFactor / fLen ) ) );

		pkDst->m_fRadius = ( fFactor + pkDst->m_fRadius );
	}
}


void Sphere::Merge( AABB *pkAABB )
{
	//Treat AABB as sphere with dim as radius
	MergeSpheres( this, pkAABB->GetTranslation(), pkAABB->GetRadius() );
}


void Sphere::Merge( OBB *pkOBB )
{
	//Treat OBB as sphere with dim as radius
	MergeSpheres( this, pkOBB->GetTranslation(), pkOBB->GetRadius() );
}


void Sphere::Merge( Sphere *pkSphere )
{
	MergeSpheres( this, pkSphere->GetTranslation(), pkSphere->m_fRadius );
}


void Sphere::Merge( Capsule *pkCapsule )
{
	Sphere kSphere; kSphere.Generate( pkCapsule );

	MergeSpheres( this, kSphere.GetTranslation(), kSphere.m_fRadius );
}


void Sphere::RenderOutlines( const Color &rkColor ) const
{
	RenderPrimitive kPrimitive;

	kPrimitive.m_ePrimitive      = RenderPrimitive::LINESTRIP;
	kPrimitive.m_uiNumPrimitives = OUTLINEPOINTS - 1;

	kPrimitive.m_kModelMatrix.Set( GetRotation(), GetTranslation() );

	//Line width
	kPrimitive.m_aUnknown[0].m_fData = 1.0f;


	float fTheta    = TWO_PI / float( OUTLINEPOINTS - 1 );
	float fCosTheta = cosf( fTheta ), fSinTheta = sinf( fTheta );
	int   iVert;

	VertexBufferPtr pkBuffer = Core::Get()->GetRenderDevice()->CreateVertexBuffer( Buffer::DYNAMIC, OUTLINEPOINTS, &DiffuseVertex::s_kDecl );

	pkBuffer->Lock( Buffer::WRITE );
	{
		DiffuseVertex *pkVertex     = (DiffuseVertex*)pkBuffer->GetVertex();
		DiffuseVertex *pkLastVertex = pkVertex;

		pkVertex->m_kPosition.Set( m_fRadius, 0.0f, 0.0f );
		pkVertex->m_kColor.Set( 0, (unsigned char)0xFF, 0 ); ++pkVertex;

		for( iVert = 1; iVert < OUTLINEPOINTS - 1; ++iVert, pkLastVertex = pkVertex, ++pkVertex )
		{
			pkVertex->m_kPosition.Set( fCosTheta * pkLastVertex->m_kPosition.x - fSinTheta * pkLastVertex->m_kPosition.z,
			                           0.0f,
			                           fSinTheta * pkLastVertex->m_kPosition.x + fCosTheta * pkLastVertex->m_kPosition.z );
			pkVertex->m_kColor = rkColor;
		}

		pkVertex->m_kPosition.Set( m_fRadius, 0.0f, 0.0f );
		pkVertex->m_kColor.Set( 0, (unsigned char)0xFF, 0 );
	}
	pkBuffer->Unlock();

	kPrimitive.m_pkVertexBuffer = pkBuffer;

	Core::Get()->GetRenderDevice()->Render( kPrimitive );


	VertexBufferPtr pkOldBuffer = pkBuffer;

	pkOldBuffer->Lock( Buffer::READ );
	{
		pkBuffer = Core::Get()->GetRenderDevice()->CreateVertexBuffer( Buffer::STATIC, OUTLINEPOINTS, &DiffuseVertex::s_kDecl, pkOldBuffer->GetVertex() );
	}
	pkOldBuffer->Unlock();

	pkBuffer->Lock( Buffer::WRITE );
	{
		DiffuseVertex *pkVertex = (DiffuseVertex*)pkBuffer->GetVertex();

		for( iVert = 0; iVert < OUTLINEPOINTS; ++iVert, ++pkVertex )
		{
			pkVertex->m_kPosition.y = pkVertex->m_kPosition.z;
			pkVertex->m_kPosition.z = 0.0f;
			pkVertex->m_kColor = rkColor;
		}
	}
	pkBuffer->Unlock();

	kPrimitive.m_pkVertexBuffer = pkBuffer;

	Core::Get()->GetRenderDevice()->Render( kPrimitive );


	pkOldBuffer = pkBuffer;

	pkOldBuffer->Lock( Buffer::READ );
	{
		pkBuffer = Core::Get()->GetRenderDevice()->CreateVertexBuffer( Buffer::STATIC, OUTLINEPOINTS, &DiffuseVertex::s_kDecl, pkOldBuffer->GetVertex() );
	}
	pkOldBuffer->Unlock();

	pkBuffer->Lock( Buffer::WRITE );
	{
		DiffuseVertex *pkVertex = (DiffuseVertex*)pkBuffer->GetVertex();

		for( iVert = 0; iVert < OUTLINEPOINTS; ++iVert, ++pkVertex )
		{
			pkVertex->m_kPosition.z = pkVertex->m_kPosition.x;
			pkVertex->m_kPosition.x = 0.0f;
			pkVertex->m_kColor = rkColor;
		}
	}

	pkBuffer->Unlock();

	kPrimitive.m_pkVertexBuffer = pkBuffer;

	Core::Get()->GetRenderDevice()->Render( kPrimitive );

	DiffuseVertex akAxes[6];
	
	akAxes[0].m_kPosition = Vector3d::ORIGO;   akAxes[0].m_kColor = Color32( 0, (unsigned char)0xFF, 0 );
	akAxes[1].m_kPosition = Vector3d::AXES[0]; akAxes[1].m_kColor = Color32( 0, (unsigned char)0xFF, 0 );
	akAxes[2].m_kPosition = Vector3d::ORIGO;   akAxes[2].m_kColor = Color32( (unsigned char)0xFF, 0, 0 );
	akAxes[3].m_kPosition = Vector3d::AXES[1]; akAxes[3].m_kColor = Color32( (unsigned char)0xFF, 0, 0 );
	akAxes[4].m_kPosition = Vector3d::ORIGO;   akAxes[4].m_kColor = Color32( 0, 0, (unsigned char)0xFF );
	akAxes[5].m_kPosition = Vector3d::AXES[2]; akAxes[5].m_kColor = Color32( 0, 0, (unsigned char)0xFF );

	kPrimitive.m_ePrimitive = RenderPrimitive::LINES;
	kPrimitive.m_uiNumPrimitives = 3;
	kPrimitive.m_pkVertexBuffer = Core::Get()->GetRenderDevice()->CreateVertexBuffer( Buffer::NORMAL, 6, &DiffuseVertex::s_kDecl, akAxes );

	Core::Get()->GetRenderDevice()->Render( kPrimitive );
}


BoundingVolume *Sphere::Duplicate() const
{
	return new Sphere( *this );
}


};
