/***************************************************************************
                  sprite.cpp  -  Quad always facing camera
                             -------------------
    begin                : Tue Jun 4 2002
    copyright            : (C) 2002 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, sprite.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "sprite.h"
#include "vertex.h"
#include "polygon.h"
#include "camera.h"
#include "render.h"
#include "renderprimitive.h"


namespace NeoEngine
{


Sprite::Sprite( const MaterialPtr &pkMaterial, float fSizeX, float fSizeY, const VertexDeclaration *pkFormat ) :
	SceneEntity()
{
	m_fRotation  = 0.0f;
	m_pkMaterial = pkMaterial;

	m_pkVertexBuffer   = Core::Get()->GetRenderDevice()->CreateVertexBuffer(  Buffer::DYNAMIC, 4, pkFormat ? pkFormat : &TexVertex::s_kDecl );
	m_pkPolygonBuffer  = Core::Get()->GetRenderDevice()->CreatePolygonStripBuffer( Buffer::STATIC, 2 );

	SetSize( fSizeX, fSizeY, true );

	m_pkPolygonBuffer->Lock( Buffer::WRITE );

	unsigned short *pucIndices = m_pkPolygonBuffer->GetIndices();

	*pucIndices++ = 0;
	*pucIndices++ = 1;
	*pucIndices++ = 3;
	*pucIndices   = 2;

	m_pkPolygonBuffer->Unlock();
}


Sprite::~Sprite()
{
}


bool Sprite::Render( Frustum *pkFrustum, bool bForce )
{
	if( !SceneEntity::Render( pkFrustum, bForce ) )
		return false;

	//FIXME: use frustum instead of camera!
	//Get active camera
	Camera *pkCamera = Camera::GetActive();

	//Set new rotation
	Quaternion kQuat = pkCamera ? pkCamera->GetWorldRotation() : Quaternion::IDENTITY;

	Matrix kRotMat = kQuat;
	
	Vector3d kAxis = kRotMat.GetColumn(2);

	kQuat *= Quaternion( AxisAngle( kAxis, m_fRotation ) );


	//Render
#ifdef NEOENGINE_THREAD_SAFE
#  ifdef WIN32
	//Ugly as hell, but you can't declare object with constructor as thread-local
	static __declspec( thread ) unsigned char saucPrimBuffer[ sizeof( RenderPrimitive ) ];
	static __declspec( thread ) RenderPrimitive *spkPrimitive = 0;
	spkPrimitive = (RenderPrimitive*)saucPrimBuffer;
#  else
#    error Thread local storage not implemented on platform
#  endif
#else
	static RenderPrimitive  skPrimitive;
	static RenderPrimitive *spkPrimitive = &skPrimitive;
#endif
	
	spkPrimitive->m_kModelMatrix.Set( kQuat, m_pkNode->GetWorldTranslation() );

	spkPrimitive->m_ePrimitive           = RenderPrimitive::TRIANGLESTRIP;
	spkPrimitive->m_pkVertexBuffer       = m_pkVertexBuffer;
	spkPrimitive->m_pkPolygonStripBuffer = m_pkPolygonBuffer;
	spkPrimitive->m_uiNumPrimitives      = spkPrimitive->m_pkPolygonStripBuffer->GetNumElements();
	spkPrimitive->m_pkMaterial           = m_pkMaterial;

	Core::Get()->GetRenderDevice()->Render( *spkPrimitive, 0 );

	//Reset pointers to allow ref counter to clean up
	spkPrimitive->m_pkMaterial           = 0;
	spkPrimitive->m_pkVertexBuffer       = 0;
	spkPrimitive->m_pkPolygonStripBuffer = 0;

	return true;
}


void Sprite::SetSize( float fSizeX, float fSizeY, bool bSetDefaultTexCoords )
{
	m_afSize[0] = fSizeX;
	m_afSize[1] = fSizeY;

	m_pkVertexBuffer->Lock( Buffer::WRITE );

	unsigned char *pucVertex    = (unsigned char*)m_pkVertexBuffer->GetVertex();
	unsigned int   uiVertexSize = m_pkVertexBuffer->GetVertexSize();

	const VertexElement *pkPositionElement = m_pkVertexBuffer->GetVertexFormat()->GetElement( VertexElement::FLOAT3, VertexElement::POSITION );
	const VertexElement *pkTexCoordElement = m_pkVertexBuffer->GetVertexFormat()->GetElement( VertexElement::FLOAT2, VertexElement::TEXCOORD );
	
	((Vector3d*)( pucVertex + pkPositionElement->m_uiOffset ))->Set( -m_afSize[0] * 0.5f,  m_afSize[1] * 0.5f, 0.0f );

	if( bSetDefaultTexCoords && pkTexCoordElement )
	{
		((float*)( pucVertex + pkTexCoordElement->m_uiOffset ))[0] = 0.0f;
		((float*)( pucVertex + pkTexCoordElement->m_uiOffset ))[1] = 0.0f;
	}

	pucVertex += uiVertexSize;

	((Vector3d*)( pucVertex + pkPositionElement->m_uiOffset ))->Set( -m_afSize[0] * 0.5f, -m_afSize[1] * 0.5f, 0.0f );

	if( bSetDefaultTexCoords && pkTexCoordElement )
	{
		((float*)( pucVertex + pkTexCoordElement->m_uiOffset ))[0] = 0.0f;
		((float*)( pucVertex + pkTexCoordElement->m_uiOffset ))[1] = 1.0f;
	}

	pucVertex += uiVertexSize;

	((Vector3d*)( pucVertex + pkPositionElement->m_uiOffset ))->Set(  m_afSize[0] * 0.5f, -m_afSize[1] * 0.5f, 0.0f );

	if( bSetDefaultTexCoords && pkTexCoordElement )
	{
		((float*)( pucVertex + pkTexCoordElement->m_uiOffset ))[0] = 1.0f;
		((float*)( pucVertex + pkTexCoordElement->m_uiOffset ))[1] = 1.0f;
	}

	pucVertex += uiVertexSize;

	((Vector3d*)( pucVertex + pkPositionElement->m_uiOffset ))->Set(  m_afSize[0] * 0.5f,  m_afSize[1] * 0.5f, 0.0f );

	if( bSetDefaultTexCoords && pkTexCoordElement )
	{
		((float*)( pucVertex + pkTexCoordElement->m_uiOffset ))[0] = 1.0f;
		((float*)( pucVertex + pkTexCoordElement->m_uiOffset ))[1] = 0.0f;
	}

	m_pkVertexBuffer->Unlock();
}


}; // namespace NeoEngine
