/***************************************************************************
                          tcp.cpp  -  TCP/IP sockets
                             -------------------
    begin                : Thu Oct 10 2002
    copyright            : (C) 2002 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, tcp.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "tcp.h"
#include "thread.h"
#include "logstream.h"

#ifdef WIN32
#  include <winsock2.h>
#  undef ERROR
#elif defined(POSIX) || defined(__APPLE__)
#  include <sys/types.h>
#  include <sys/socket.h>
#  include <arpa/inet.h>
#  include <netinet/in.h>
#  include <unistd.h>
#  include <fcntl.h>
#  include <netdb.h>
#endif

#ifdef POSIX
#  include <sys/poll.h>
#  include <errno.h>
#elif __APPLE__
#  include <sys/select.h>
#endif

#include <assert.h>

#ifdef __APPLE__
#  define SOCKET_ERROR    -1
#endif /* __APPLE__ */

using namespace std;


namespace NeoEngine
{


class AcceptThread : public ThreadMethod
{
	public:
		virtual void                *ThreadEntry( Thread *pkThread, void *pData );
};




TCPSocket::TCPSocket( SocketCallback *pkCallback ) :
	Socket( pkCallback )
{
	m_iSocket        = socket( PF_INET, SOCK_STREAM, 0 );
	m_pkAcceptThread = 0;
}


TCPSocket::TCPSocket( int iSocket, SocketCallback *pkCallback ) :
	Socket( pkCallback )
{
	m_iSocket        = iSocket;
	m_pkAcceptThread = 0;
}


TCPSocket::~TCPSocket()
{
	if( m_iSocket )
	{
		neolog << LogLevel( DEBUG ) << "[TCPSocket::~TCPSocket] close and shutdown socket" << endl;

		SetBlocking( false );

#ifdef WIN32
		closesocket( m_iSocket );
		shutdown( m_iSocket, SD_BOTH );
#elif defined(POSIX) || defined(__APPLE__)
		close( m_iSocket );
		shutdown( m_iSocket, SHUT_RDWR );
#else
#  error TCPSocket::~TCPSocket not implemented
#endif

		m_iSocket = 0;
	}

	if( m_pkAcceptThread )
	{
		neolog << LogLevel( DEBUG ) << "[TCPSocket::~TCPSocket] deleting accept thread" << endl;

		ThreadMethod *pkMethod = m_pkAcceptThread->GetThreadMethod();

		delete m_pkAcceptThread, m_pkAcceptThread = 0;

		if( pkMethod )
			delete pkMethod;
	}
}


bool TCPSocket::Listen( int iPort, SocketCallback *pkCallback )
{
	assert( m_iSocket );

	if( !m_pkCallback && pkCallback )
		m_pkCallback = pkCallback;

	sockaddr_in saddr;

	saddr.sin_family = AF_INET;

	//Listen to any address
#ifdef WIN32
	saddr.sin_addr.S_un.S_addr = INADDR_ANY;
#elif defined(POSIX) || defined(__APPLE__)
	saddr.sin_addr.s_addr      = INADDR_ANY;
#else
#  error TCPSocket::Listen not implemented
#endif
	saddr.sin_port             = htons( iPort );

	if( bind( m_iSocket, (sockaddr*)&saddr, sizeof( sockaddr_in ) ) )
	{
#ifdef WIN32
		int iError = WSAGetLastError();
		neolog << LogLevel( ERROR ) << "[TCPSocket::Listen] error binding socket[" << iError << "]: " << strerror( iError ) << endl;
#elif defined(POSIX) || defined(__APPLE__)
		neolog << LogLevel( ERROR ) << "[TCPSocket::Listen] error binding socket: " << strerror( errno ) << endl;
#endif
		return false;
	}

	if( listen( m_iSocket, SOMAXCONN ) )
	{
#ifdef WIN32
		int iError = WSAGetLastError();
		neolog << LogLevel( ERROR ) << "[TCPSocket::Listen] error listening to socket[" << iError << "]: " << strerror( iError ) << endl;
#elif defined(POSIX) || defined(__APPLE__)
		neolog << LogLevel( ERROR ) << "[TCPSocket::Listen] error listening to socket: " << strerror( errno ) << endl;
#endif
		return false;
	}

	neolog << LogLevel( DEBUG ) << "[TCPSocket::Listen] starting accept thread" << endl;

	//Fire up accept thread
	m_pkAcceptThread = new Thread( new AcceptThread, this, 0, 0 );

	neolog << LogLevel( DEBUG ) << "[TCPSocket::Listen] accept thread started" << endl;

	return true;
}

bool TCPSocket::Connect( const string &rstrAddr, int iPort )
{
	assert( m_iSocket );

	sockaddr_in saddr;

	//Connect to given host
	hostent *pHost = gethostbyname( rstrAddr.c_str() );

	int  i          = 0;
	bool bConnected = false;

	while( pHost && pHost->h_addr_list[i] )
	{
		saddr.sin_addr   = *(in_addr*)pHost->h_addr_list[i];
		saddr.sin_port   = htons( iPort );
		saddr.sin_family = AF_INET;

		if( !connect( m_iSocket, (sockaddr*)&saddr, sizeof( sockaddr_in ) ) )
		{
			bConnected = true;
			break;
		}

		++i;
	}

	if( !bConnected )
	{
		int *piError = ( int* )malloc( sizeof( int ) );
#ifdef WIN32
		*piError = WSAGetLastError();
		neolog << LogLevel( ERROR ) << "[TCPSocket::Connect] unable to connect to " << rstrAddr << ":" << iPort << " [" << *piError << "]" << endl;
#elif defined(POSIX) || defined(__APPLE__)
		*piError = errno;
		neolog << LogLevel( ERROR ) << "[TCPSocket::Connect] unable to connect " << rstrAddr << ":" << iPort << ": " << strerror( *piError ) << endl;
#endif
		if( m_pkCallback )
		{
			m_pkCallback->SocketEventCallback( this, Socket::ERR, piError );
		}
		free( piError );
		return false;
	}

	if( m_pkCallback )
	{
		m_pkCallback->SocketEventCallback( this, Socket::CONNECTED, 0 );
	}

	return true;
}


void TCPSocket::SetBlocking( bool bBlock )
{
#ifdef WIN32
	unsigned long ulParam = bBlock ? 1 : 0;
	ioctlsocket( m_iSocket, FIONBIO, &ulParam );
#elif defined(POSIX) || defined(__APPLE__)
	long lBlockingMode = fcntl( m_iSocket, F_GETFL );
	if ( bBlock )
	{
		lBlockingMode &= ~O_NONBLOCK;
	}
	else
	{
		lBlockingMode |= O_NONBLOCK;
	}
	if ( fcntl( m_iSocket, F_SETFL, lBlockingMode ) < 0 )
	{
#ifdef WIN32
		int iError = WSAGetLastError();
		neolog << LogLevel( ERROR ) << "*** [TCPSocket::SetBlocking] error set blocking mode[" << iError << "]: " << strerror( iError ) << endl;
#elif defined(POSIX) || defined(__APPLE__)
		neolog << LogLevel( ERROR ) << "*** [TCPSocket::SetBlocking] error set blocking mode: " << strerror( errno ) << endl;
#endif
	}
#else
#  error TCPSocket::SetBlocking not implemented
#endif
}


int TCPSocket::Read( void *pDst, int iBytes )
{
	assert( m_iSocket );

	int iResult = recv( m_iSocket, (char*)pDst, iBytes, 0 );
	if ( iResult < 0 )
	{
#ifdef WIN32
		int iError = WSAGetLastError();
		if ( iError == WSAECONNABORTED )
		{
			neolog << LogLevel( ERROR ) << "[TCPSocket::Read] error reading socket: Socket is closed already" << endl;
		}
		else if ( iError == WSAECONNRESET )
		{
			neolog << LogLevel( ERROR ) << "[TCPSocket::Read] error reading socket: Socket was reseted and must be closed" << endl;
		}
		else 
		{
			neolog << LogLevel( ERROR ) << "[TCPSocket::Read] error reading socket: " << strerror( errno ) << endl;
		}
#elif defined(POSIX) || defined(__APPLE__)
		neolog << LogLevel( ERROR ) << "[TCPSocket::Read] error reading socket: " << strerror( errno ) << endl;
#endif
	}
	else if ( iResult == 0 )
	{
		neolog << LogLevel( ERROR ) << "[TCPSocket::Read] error reading socket: Socket is closed already" << endl;
	}
	return iResult;
}


int TCPSocket::Write( const void *pSrc, int iBytes )
{
	assert( m_iSocket );

	int iResult = send( m_iSocket, (char*)pSrc, iBytes, 0 );
	if ( iResult < 0 )
	{
#ifdef WIN32
		int iError = WSAGetLastError();
		neolog << LogLevel( ERROR ) << "[TCPSocket::Write] error writing socket[" << iError << "]: " << strerror( iError ) << endl;
#elif defined(POSIX) || defined(__APPLE__)
		neolog << LogLevel( ERROR ) << "[TCPSocket::Write] error writing socket: " << strerror( errno ) << endl;
#endif
	}
	else if ( iResult == 0 )
	{
		neolog << LogLevel( ERROR ) << "[TCPSocket::Write] error writing socket: Socket is closed already" << endl;
	}
	return iResult;
}


unsigned int TCPSocket::Poll( int iTimeout )
{
#if defined(WIN32) || defined(__APPLE__)

	fd_set fdread;

	FD_ZERO( &fdread );
	FD_SET( (unsigned int)m_iSocket, &fdread );

	fd_set fderr;

	FD_ZERO( &fderr );
	FD_SET( (unsigned int)m_iSocket, &fderr );

	timeval timeout;

	timeout.tv_sec  = iTimeout / 1000;
	timeout.tv_usec = ( iTimeout % 1000 ) * 1000;

	int iRet = select( m_iSocket + 1, &fdread, 0, &fderr, &timeout );
	if( iRet == SOCKET_ERROR )
	{
		int iError = WSAGetLastError();
		neolog << LogLevel( ERROR ) << "[TCPSocket::Poll] error polling socket[" << iError << "]: " << strerror( iError ) << endl;
		return Socket::INVALID;
	}
	else if ( iRet == 0 )
	{
		return Socket::TIMEOUT;
	}

	unsigned int uiRet = 0;

	if( FD_ISSET( m_iSocket, &fdread ) )
	{
		uiRet |= Socket::DATAIN;
	}
	if( FD_ISSET( m_iSocket, &fderr ) )
	{
		uiRet |= Socket::ERR;
	}
	if( !uiRet )
	{
		uiRet  = Socket::TIMEOUT;
	}

	return uiRet;

#elif POSIX

	pollfd pfd;

	pfd.fd      = m_iSocket;
	pfd.events  = POLLIN | POLLPRI | POLLERR | POLLHUP | POLLNVAL;
	pfd.revents = 0;

	int          iRet     = 0;
	unsigned int uiEvent  = 0;

	if( ( iRet = poll( &pfd, 1, iTimeout ) ) < 0 )
	{
		neolog << LogLevel( ERROR ) << "[TCPSocket::Poll] error polling socket: " << strerror( errno ) << endl;
		return Socket::INVALID;
	}
	else if( !iRet )
	{
		return Socket::TIMEOUT;
	}

	if( pfd.revents & POLLIN )
	{
		uiEvent |= Socket::DATAIN;
	}
	if( pfd.revents & POLLERR )
	{
		uiEvent |= Socket::ERR;
	}
	if( pfd.revents & POLLHUP )
	{
		uiEvent |= Socket::HUP;
	}
	if( pfd.revents & POLLNVAL )
	{
		uiEvent |= Socket::NVAL;
	}

	return uiEvent;

#else
#  error TCPSocket::Poll not implemented
#endif
}





void *AcceptThread::ThreadEntry( Thread *pkThread, void *pData )
{
	TCPSocket *pkTCPSocket = (TCPSocket*)pData;

	int iSocket = pkTCPSocket->m_iSocket;

	sockaddr_in saddr;

	saddr.sin_family = AF_INET;

#if defined(WIN32) || defined(__APPLE__)
	int iAddrSize = sizeof( sockaddr_in );
#elif POSIX
	socklen_t iAddrSize = sizeof( sockaddr_in );
#else
#  error AcceptThread::ThreadEntry not implemented
#endif

	pkThread->m_bStarted = true;

	while( !pkThread->m_bTerminate )
	{
		unsigned int uiResult = pkTCPSocket->Poll( 100 );
		if ( uiResult & Socket::DATAIN )
		{
			int iConnSock = accept( iSocket, (sockaddr*)&saddr, &iAddrSize );
			if( iConnSock < 0 )
			{
#ifdef WIN32
				int iError = WSAGetLastError();
				neolog << LogLevel( ERROR ) << "[AcceptThread::ThreadEntry] error accepting socket[" << iError << "]: " << strerror( iError ) << endl;
#elif defined(POSIX) || defined(__APPLE__)
				neolog << LogLevel( ERROR ) << "[AcceptThread::ThreadEntry] error accepting socket: " << strerror( errno ) << endl;
#endif
				break;
			}
			pkTCPSocket->m_pkCallback->SocketEventCallback( new TCPSocket( iConnSock, pkTCPSocket->m_pkCallback ), Socket::CONNECTION, 0 );
		}
	}

	pkThread->m_bTerminated = true;

	return 0;
}


}; // namespace NeoEngine
