/***************************************************************************
                     terrain.cpp  -  Classes for terrain
                             -------------------
    begin                : Mon Sep 8 2003
    copyright            : (C) 2003-2004 by Cody Russell
    email                : cody `at' jhu.edu
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, terrain.cpp

 The Initial Developer of the Original Code is Cody Russell.
 Portions created by Cody Russell are Copyright (C) 2003-2004
 Cody Russell. All Rights Reserved.

 ***************************************************************************/

#include "aabb.h"
#include "core.h"
#include "render.h"
#include "scenenode.h"
#include "terrain.h"
#include "vertex.h"

using namespace std;
using namespace NeoEngine;

TerrainHeightmap::TerrainHeightmap( float fHScale, float fVScale ) :
	m_pfData( 0 ),
	m_iSize( 0 ),
	m_pkNormalData( 0 ),
	m_fHScale( fHScale ),
	m_fVScale( fVScale )
{
}

TerrainHeightmap::~TerrainHeightmap()
{
	delete [] m_pfData;
}


// Determine the maximum and minimum heights of the entire map
void TerrainHeightmap::PostLoadInit()
{
	m_fMinHeight = m_fMaxHeight = m_pfData[0];

	for( int i = 1; i < m_iSize * m_iSize; i++ )
	{
		if( m_pfData[i] < m_fMinHeight )
			m_fMinHeight = m_pfData[i];

		if( m_pfData[i] > m_fMaxHeight )
			m_fMaxHeight = m_pfData[i];
	}
}


void TerrainHeightmap::Load( const std::string &rstrFilename, int iSize )
{
	ImageData *pkImage = Core::Get()->GetRenderDevice()->LoadImageData( rstrFilename );

	neolog << "Loaded image " << rstrFilename << endl
	       << "Width: "       << pkImage->m_iWidth << endl
	       << "Height: "      << pkImage->m_iHeight << endl
	       << "BPP: "         << pkImage->m_iBPP << endl
	       << "Channels: "    << pkImage->m_iChannels << endl;

	m_iSize = iSize;

	m_pfData = new float[ m_iSize * m_iSize ];

	for( int i = 0; i < m_iSize * m_iSize; i++ )
	{
		m_pfData[i] = pkImage->m_pucData[i] * m_fVScale;
	}

	PostLoadInit();
}

void TerrainHeightmap::Load( ImageData *pkImage )
{
	if( pkImage->m_iWidth != pkImage->m_iHeight )
	{
		neolog << LogLevel( ERROR ) << "[TerrainHeightmap::Load] Image should be square in shape." << endl;
		return;
	}

	if( pkImage->m_iChannels > 1 )
	{
		neolog << LogLevel( ERROR ) << "[TerrainHeightmap::Load] Image has more than 1 color channel." << endl;
		return;
	}

	if( pkImage->m_iBPP > 8 )
	{
		neolog << LogLevel( ERROR ) << "[TerrainHeightmap::Load] Image has greater than 8 bits per pixel." << endl;
		return;
	}

	m_iSize = pkImage->m_iWidth;

	m_pfData = new float[ m_iSize * m_iSize ];

	for( int i = 0; i < m_iSize * m_iSize; i++ )
	{
		m_pfData[i] = pkImage->m_pucData[i] * m_fVScale;
	}

	PostLoadInit();
}


int TerrainHeightmap::GetIndex( int x, int y ) const
{
	if( x >= m_iSize )
	{
		neolog << LogLevel( WARNING ) << "[TerrainHeightmap::GetIndex]: x out of bounds!" << endl;
		x = m_iSize - 1;
	}

	if( y >= m_iSize )
	{
		neolog << LogLevel( WARNING ) << "[TerrainHeightmap::GetIndex]: y out of bounds!" << endl;
		y = m_iSize - 1;
	}

	return ( y * m_iSize + x );
}


float TerrainHeightmap::GetHeight( int x, int y ) const
{
	return m_pfData[ GetIndex( x, y ) ];
}


float TerrainHeightmap::GetWorldHeight( float x, float y )
{
	float fx = x * 2;
	float fy = y * 2;

	int ix0 = TruncFloat( fx );
	int iy0 = TruncFloat( fy );

	fx -= ix0;
	fy -= iy0;

	ix0 = CLAMP( ix0, 0, m_iSize - 1 );
	iy0 = CLAMP( iy0, 0, m_iSize - 1 );

	int ix1 = CLAMP( ix0 + 1, 0, m_iSize - 1 );
	int iy1 = CLAMP( iy0 + 1, 0, m_iSize - 1 );

	float h0 = GetHeight( ix0, iy0 );
	float h1 = GetHeight( ix1, iy0 );
	float h2 = GetHeight( ix0, iy1 );
	float h3 = GetHeight( ix1, iy1 );

#if 0
	neolog << "[TerrainHeightmap::GetWorldHeight( " << x << ", " << y << " )]" << endl;

	neolog << "    GetHeight( " << iMapX0 << ", " << iMapY0 << " ) == " << h0 << endl;
	neolog << "    GetHeight( " << iMapX1 << ", " << iMapY0 << " ) == " << h1 << endl;
	neolog << "    GetHeight( " << iMapX0 << ", " << iMapY1 << " ) == " << h2 << endl;
	neolog << "    GetHeight( " << iMapX1 << ", " << iMapY1 << " ) == " << h3 << endl;

	raise( SIGTRAP );
#endif

	float fLow  = ( h1 * fx ) + ( h0 * ( 1.0f - fx ) );
	float fHigh = ( h3 * fx ) + ( h2 * ( 1.0f - fx ) );

	return ( fHigh * fy ) + ( fLow * ( 1.0f - fy ) );
}


float TerrainHeightmap::GetWorldHeight( int iIndex )
{
	float y = float( iIndex / m_iSize );
	float x = fmodf( (float)iIndex, (float)m_iSize );

	return GetWorldHeight( x, y );
}



const Vector3d& TerrainHeightmap::GetNormal( int x, int y ) const
{
	if( x >= m_iSize ) x = m_iSize - 1;
	if( y >= m_iSize ) y = m_iSize - 1;

	return m_pkNormalData[ y * m_iSize + x];
}


QuadtreeNode::QuadtreeNode()
{
}
 
 
QuadtreeNode::~QuadtreeNode()
{
	if( m_eType != LEAF )
	{
		delete [] m_pkChildren;
	}
}
 
 
void QuadtreeNode::Initialize( TerrainPage *pkTerrain, int x, int y, int iSize, int iDepth )
{
	m_kWorldTranslation.x = float( x * pkTerrain->GetBlockSize() );
	m_kWorldTranslation.y = float( y * pkTerrain->GetBlockSize() );

	Vector3d kDim;

	kDim.x = x * pkTerrain->GetHeightmap()->GetHScale();
	kDim.y = y * pkTerrain->GetHeightmap()->GetHScale();
	kDim.z = pkTerrain->GetHeightmap()->GetVScale();

	if( iDepth > 0 )
	{
		int iChildDepth = iDepth - 1;
		int iChildSize = iSize >> 1;

		int iOffset = iSize >> iDepth;

		m_eType = NODE;

		m_pkChildren = new QuadtreeNode[ 4 ];
		m_pkChildren[ 0 ].Initialize( pkTerrain, x,              y,              iChildSize, iChildDepth );
		m_pkChildren[ 1 ].Initialize( pkTerrain, x + iChildSize, y,              iChildSize, iChildDepth );
		m_pkChildren[ 2 ].Initialize( pkTerrain, x,              y + iChildSize, iChildSize, iChildDepth );
		m_pkChildren[ 3 ].Initialize( pkTerrain, x + iChildSize, y + iChildSize, iChildSize, iChildDepth );
	}
	else
	{
		int iBlockSize = pkTerrain->GetBlockSize();

		m_eType = LEAF;

		m_pkData = pkTerrain->GetBlock( x / iBlockSize, y / iBlockSize );
		SetEntity( m_pkData );
	}
}


bool QuadtreeNode::Render( Frustum *pkFrustum, bool bForce )
{
	if( m_eType == NODE )
	{
		for( int i = 0; i < 4; i++ )
		{
			m_pkChildren[i].Render( pkFrustum, bForce );
		}
	}
	else
	{
		if( m_pkEntity && m_pkEntity->IsActive() && ( !pkFrustum || !m_pkEntity->GetBoundingVolume() || m_pkEntity->GetBoundingVolume()->Intersection( pkFrustum ) ) )
			return m_pkEntity->Render( pkFrustum, bForce );
	}

	return true;
}


void QuadtreeNode::Update( float fDeltaTime )
{
        m_pkData->Update( fDeltaTime );
	neolog << "QuadtreeNode::Update()" << endl;
}


void QuadtreeNode::UpdateVisible( Frustum *pkFrustum )
{
        if( m_eType == NODE )
	{
	        for( int i = 0; i < 4; i++ )
		{
		        m_pkChildren[i].UpdateVisible( pkFrustum );
		}
	}
	else
	{
	        if( m_pkEntity && m_pkEntity->IsActive() && ( !pkFrustum || !m_pkEntity->GetBoundingVolume() || m_pkEntity->GetBoundingVolume()->Intersection( pkFrustum ) ) )
		{
		        m_pkEntity->Update( 0.0f );
		}
	}
}


Quadtree::Quadtree( TerrainPage *pkPage, const NeoEngine::HashString &rstrName ) :
	NeoEngine::Room( rstrName ),
	m_iDepth( pkPage->GetDepth() )
{
	neolog << "Creating Quadtree of depth " << m_iDepth << endl;
	neolog << "Block size == " << pkPage->GetBlockSize() << endl;

	m_pkRoot = new QuadtreeNode();
	m_pkRoot->Initialize( pkPage, 0, 0, pkPage->GetSize(), m_iDepth );

	AttachNode( m_pkRoot );

	neolog << "Node attached" << std::endl;
}


Quadtree::~Quadtree()
{
}


bool Quadtree::Render( Frustum *pkFrustum, bool bForce )
{
	if( !Room::Render( pkFrustum, bForce ) )
		return false;

	if( m_bDirty )
	{
	        m_pkRoot->UpdateVisible( pkFrustum );
		SetDirty( false );
	}

	return m_pkRoot->Render( pkFrustum, bForce );
}


TerrainBlock::TerrainBlock( int iOffsetX, int iOffsetY, int iSize, int iIndex ) :
	m_iOffsetX( iOffsetX ),
	m_iOffsetY( iOffsetY ),
	m_iSize( iSize ),
	m_iBlockIndex( iIndex )
{
}


TerrainPage::TerrainPage( TerrainHeightmap *pkHeightmap, TerrainMaterialFactory *pkMaterialFactory, int iDepth ) :
	m_pkHeightmap( pkHeightmap ),
	m_iBlockSize( pkHeightmap->GetSize() >> iDepth ),
	m_iDepth( iDepth ),
	m_pkMaterialFactory( pkMaterialFactory )
{
}


TerrainPage::~TerrainPage()
{
	delete m_pkHeightmap;
}


TerrainHeightmap *TerrainPage::GetHeightmap()
{
	return m_pkHeightmap;
}


TerrainBlock *TerrainPage::GetBlock( int iIndex )
{
	return m_vpkBlocks[ iIndex ];
}


TerrainBlock *TerrainPage::GetBlock( int x, int y )
{
	return GetBlock( y * GetNumBlocksOnSide() + x );
}


int TerrainPage::GetTotalBlocks()
{
	return m_vpkBlocks.size();
}


int TerrainPage::GetNumBlocksOnSide()
{
	return m_pkHeightmap->GetSize() / m_iBlockSize;
}


int TerrainPage::GetBlockSize()
{
	return m_iBlockSize;
}


int TerrainPage::GetDepth()
{
	return m_iDepth;
}


int TerrainPage::GetSize()
{
	return m_pkHeightmap->GetSize();
}


void TerrainPage::GenerateAABB()
{
#if 0
	std::vector< TerrainBlock* >::iterator ppkIter = m_vpkBlocks.begin();
	std::vector< TerrainBlock* >::iterator ppkEnd  = m_vpkBlocks.end();

	for( ; ppkIter != ppkEnd; ++ppkIter )
	{
		(*ppkIter)->GenerateAABB();
	}
#endif

	neolog << LogLevel( WARNING ) << "[TerrainPage::GenerateAABB]: This function is deprecated." << endl;
}


TerrainManager::TerrainManager()
{
}


TerrainManager::~TerrainManager()
{
	vector< ModulePtr >::iterator ppkModule = m_vpkModules.begin();
	vector< ModulePtr >::iterator ppkEnd    = m_vpkModules.end();

	for( ; ppkModule != ppkEnd; ++ppkModule )
	{
		ModSymbol kSymShutdown;

		kSymShutdown = (*ppkModule)->LookupSymbol( "Shutdown" );

		if( !kSymShutdown )
		{
			neolog << LogLevel( WARNING ) << "[TerrainManager::~TerrainManager] terrain module ["
			       << (*ppkModule)->GetName() << "] does not export Shutdown method" << endl;
		}
		else
		{
			(( void ( NE_CDECL* )() )kSymShutdown)();
		}
	}
}

ModulePtr TerrainManager::LoadModule( const HashString &rstrName )
{
#if defined( WIN32 ) && defined( _DEBUG )
	HashString strModuleName = "neo" + rstrName.m_strData + "_debug";
#else
	HashString strModuleName = "neo" + rstrName.m_strData;
#endif

	vector< ModulePtr >::iterator ppkModule = m_vpkModules.begin();
	vector< ModulePtr >::iterator ppkEnd    = m_vpkModules.end();

	for( ; ppkModule != ppkEnd; ++ppkModule )
	{
		if( (*ppkModule)->GetName() == strModuleName )
			return( *ppkModule );
	}

	ModulePtr pkModule = Core::Get()->GetModuleManager()->LoadModule( strModuleName );

	if( pkModule )
	{
		ModSymbol kSymInitialize;

		kSymInitialize = pkModule->LookupSymbol( "Initialize" );

		if( !kSymInitialize )
		{
			neolog << LogLevel( WARNING ) << "[TerrainManager::LoadModule] Terrain module ["
			       << rstrName << "] does not export Initialize method" << endl;
		}
		else
		{
			(( void ( NE_CDECL* )() )kSymInitialize)();
		}

		m_vpkModules.push_back( pkModule );
	}
	else
	{
		neolog << LogLevel( ERROR ) << "*** Unable to load terrain module [" << rstrName << "]" << endl;
	}

	return pkModule;
}
