/***************************************************************************
        terrainmaterial.h  -  Classes for terrain material generation
                             -------------------
    begin                : Mon Mar 22 2004
    copyright            : (C) 2003-2004 by Cody Russell
    email                : cody `at' jhu.edu
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, terrain.h

 The Initial Developer of the Original Code is Cody Russell.
 Portions created by Cody Russell are Copyright (C) 2003-2004
 Cody Russell. All Rights Reserved.

 ***************************************************************************/

#include "terrain.h"
#include "terrainmaterial.h"
#include "vertex.h"

using namespace std;
using namespace NeoEngine;


TerrainMaterialFactory::TerrainMaterialFactory()
{
}


TerrainMaterialFactory::~TerrainMaterialFactory()
{
}


TerrainMaterialFactory::TerrainMaterialFactory( MaterialPtr &pkMaterial )
{
	m_pkMaterial = pkMaterial;
}


void TerrainMaterialFactory::SetTexCoords( VertexBufferPtr pkVertices, TerrainPage *pkPage )
{
        int iSize = pkPage->GetBlockSize();
        float fHScale = pkPage->GetHeightmap()->GetHScale();

        pkVertices->Lock( Buffer::WRITE );

        for( int y = 0; y <= iSize; y++ )
        {
                for( int x = 0; x <= iSize; x++ )
                {
                        NormalDiffuseTexVertex *pkVertex = (NormalDiffuseTexVertex*) pkVertices->GetVertex( y * ( iSize + 1 ) + x );

                        pkVertex->m_afTexCoord[0] = (float) x / (float) iSize * fHScale;
                        pkVertex->m_afTexCoord[1] = (float) y / (float) iSize * fHScale;
                }
        }

        pkVertices->Unlock();
}


Terrain1TextureFactory::Terrain1TextureFactory( MaterialPtr &pkMaterial ) :
	TerrainMaterialFactory( pkMaterial )
{
}


void Terrain1TextureFactory::SetTexture( const std::string &rstrFilename )
{
        m_strFilename = rstrFilename;
}


void Terrain1TextureFactory::GenerateMaterial( TerrainPage *pkPage )
{
	m_pkMaterial->m_pkTexture = Core::Get()->GetRenderDevice()->LoadTexture( m_strFilename );
}


void TerrainTextureComponent::Set( float fMin, float fOpt, float fMax )
{
        neolog << "[TerrainTextureComponent::Set( " << fMin << ", " << fOpt << ", " << fMax << " )" << endl;

        m_fMin     = fMin;
        m_fOptimal = fOpt;
        m_fMax     = fMax;
}


unsigned char TerrainTextureComponent::CalculateBlendFactor( float f, TerrainHeightmap *pkHeightmap )
{
        float fRet;

        if( f == m_fOptimal )
        {
                //neolog << "CalculateBlendFactor() for " << f << " == OPTIMAL." << endl;
                fRet = 1.0f;
        }
        else if( f <= m_fMin )
        {
                //neolog << "CalculateBlendFactor() for " << f << " is BELOW MINIMUM" << endl;
                fRet = 0.0f;
        }
        else if( f >= m_fMax )
        {
                //neolog << "CalculateBlendFactor() for " << f << " is ABOVE MAXIMUM" << endl;
                fRet = 0.0f;
        }
        else
        {
                if( f < m_fOptimal )
                {
                        fRet = ( ( 1.0f / ( m_fOptimal - m_fMin ) ) * ( f - m_fMin ) );    // ( m_fOptimal - f ) );
                }
                else
                {
                        fRet = ( ( 1.0f / ( m_fMax - m_fOptimal ) ) * ( f - m_fOptimal ) );
                }
        }

        //neolog << "CalculateBlendFactor for " << f << " == " << fRet * 255 << endl;

        return ( unsigned char ) ( fRet * 255 );
}


TexturePtr TerrainBlendMaterialFactory::CalculateBlendmap( TerrainPage *pkPage, int iSize )
{
        ImageData *pkImage = new ImageData;

        pkImage->m_pucData = new unsigned char[ iSize * iSize * 4 ];
        pkImage->m_iWidth  = iSize;
        pkImage->m_iHeight = iSize;

        unsigned char *pucData = pkImage->m_pucData;

	TerrainHeightmap *pkHeightmap = pkPage->GetHeightmap();

        neolog << "Calculating " << iSize << "x" << iSize << " blendmap now." << endl;

        for( int y = 0; y < iSize; y++ )
        {
                for( int x = 0; x < iSize; x++ )
                {
                        float j;

                        j = pkHeightmap->GetWorldHeight( (float)x, (float)y );

                        pucData[0] = m_akRanges[0].CalculateBlendFactor( j, pkHeightmap );
                        pucData[1] = m_akRanges[1].CalculateBlendFactor( j, pkHeightmap );
                        pucData[2] = m_akRanges[2].CalculateBlendFactor( j, pkHeightmap );
                        pucData[3] = m_akRanges[3].CalculateBlendFactor( j, pkHeightmap );

                        pucData += 4;
                }
        }

        TexturePtr pkTexture = Core::Get()->GetRenderDevice()->CreateTexture( "TerrainBlendMap" );

        pkTexture->UploadImage( pkImage, Texture::TEX2D, Texture::RGBA );

        return pkTexture;
}


TerrainBlendMaterialFactory::TerrainBlendMaterialFactory( MaterialPtr &pkMaterial, LIGHTINGTYPE eLightType, TEXTURINGTYPE eTextureType ) :
	TerrainMaterialFactory( pkMaterial ),
        m_eLightingType( eLightType ),
        m_eTexturingType( eTextureType )
{
        m_bCalculateBlendmap = true;
}


void TerrainBlendMaterialFactory::SetTextures( const std::string &rstrFile1, const std::string &rstrFile2, const std::string &rstrFile3, const std::string &rstrFile4 )
{
        m_strFile1 = rstrFile1;
        m_strFile2 = rstrFile2;
        m_strFile3 = rstrFile3;
        m_strFile4 = rstrFile4;
}


void TerrainBlendMaterialFactory::SetBlendmap( const std::string &rstrFile )
{
        m_strBlendmap = rstrFile;

        m_bCalculateBlendmap = false;
}


void TerrainBlendMaterialFactory::SetAmbient( const Vector3d &rkAmbient )
{
        m_kAmbient = rkAmbient;
}


void TerrainBlendMaterialFactory::GenerateMaterial( TerrainPage *pkPage )
{
	TerrainHeightmap *pkHeightmap = pkPage->GetHeightmap();

        int iSize     = pkHeightmap->GetSize();

        float fMin    = pkHeightmap->GetMinHeight();
        float fMax    = pkHeightmap->GetMaxHeight();

        neolog << "[TerrainBlendMaterialFactory::GenerateMaterial]: Min == " << fMin
               << " Max == " << fMax << endl;

        float fDiff = fMin + ( fMax - fMin );

        m_akRanges[0].Set( fMin,          fDiff * 0.15f, fDiff * 0.30f );
        m_akRanges[1].Set( fDiff * 0.25f, fDiff * 0.33f, fDiff * 0.55f );
        m_akRanges[2].Set( fDiff * 0.45f, fDiff * 0.67f, fDiff * 0.79f );
        m_akRanges[3].Set( fDiff * 0.63f, fDiff * 0.87f, fMax );

        if( m_bCalculateBlendmap )
        {
                m_pkMaterial->m_pkTexture = CalculateBlendmap( pkPage, iSize * 2 );
                m_bCalculateBlendmap = false;
        }
        else
        {
                neolog << "Loading blendmap from file." << endl;
                m_pkMaterial->m_pkTexture = Core::Get()->GetRenderDevice()->LoadTexture( m_strBlendmap );
        }

#if 0
        TextureLayer kLayer;
        kLayer.m_pkTexture = Core::Get()->GetRenderDevice()->LoadTexture( m_strFile1 );
        kLayer.m_kBlendMode.Set( BlendMode::SRC_ONE | BlendMode::DEST_ZERO );
        //m_pkMaterial = kLayer;
        //m_pkMaterial->m_vpkTextureLayers.push_back( pkLayer );
#endif

	TextureLayer *pkLayer = new TextureLayer;
	pkLayer->m_pkTexture = Core::Get()->GetRenderDevice()->LoadTexture( m_strFile1 );
	m_pkMaterial->m_vpkTextureLayers.push_back( pkLayer );

        pkLayer->m_pkTexture = Core::Get()->GetRenderDevice()->LoadTexture( m_strFile2 );
        pkLayer->m_kBlendMode.Set( BlendMode::LERP_COLOR );
        m_pkMaterial->m_vpkTextureLayers.push_back( pkLayer );

        pkLayer = new TextureLayer;
        pkLayer->m_pkTexture = Core::Get()->GetRenderDevice()->LoadTexture( m_strFile3 );
        pkLayer->m_kBlendMode.Set( BlendMode::LERP_ALPHA );
        m_pkMaterial->m_vpkTextureLayers.push_back( pkLayer );

#if 0
        pkLayer = new TextureLayer;
        pkLayer->m_pkTexture = Core::Get()->GetRenderDevice()->LoadTexture( m_strFile4 );
        pkLayer->m_kBlendMode.Set( BlendMode::MODULATE );
        m_pkMaterial->m_vpkTextureLayers.push_back( pkLayer );
#endif

        if( !Core::Get()->GetRenderDevice()->GetCaps().IsSet( RenderCaps::FRAGMENTSHADER ) )
        {
                neolog << "Fragment shaders are not supported, falling back to software texturing." << endl;
                m_eTexturingType = TEXTURE_SOFTWARE;
        }

        // Reset the lighting to use lightmaps if we don't have vertex shader support
        if( !Core::Get()->GetRenderDevice()->GetCaps().IsSet( RenderCaps::VERTEXSHADER ) )
        {
                neolog << "Vertex shaders not supported, falling back to software lightmaps." << endl;
                m_eLightingType = LIGHT_SOFTWARE;
                m_eTexturingType = TEXTURE_SOFTWARE;
        }

        neolog << "TerrainPage::TerrainPage: m_eLightingType == " << m_eLightingType << endl;

        if( m_eTexturingType == TEXTURE_HARDWARE )
        {
                if( !( m_pkMaterial->m_pkFragmentShader = Core::Get()->GetRenderDevice()->CreateShader( Shader::FRAGMENTSHADER ) ) )
		{
                        neolog << LogLevel( WARNING ) << "[TerrainPage::TerrainPage]: Could not create fragment shader!  "
                               << "Falling back to software texturing." << endl;

                        m_eTexturingType = TEXTURE_SOFTWARE;
                }
                else
                {
			string strTexturingFP =
				"!!ARBfp1.0\n"
				"#program main\n"
				"#semantic main.IBlendSampler\n"
				"#semantic main.ISamplerR\n"
				"#semantic main.ISamplerG\n"
				"#semantic main.ISamplerB\n"
				"#var float4 IColor : $vin.COLOR : COLOR : 1 : 1\n"
				"#var float2 ITexBlend : $vin.TEXCOORD0 : TEXCOORD0 : 2 : 1\n"
				"#var float2 ITexR : $vin.TEXCOORD1 : TEXCOORD1 : 3 : 1\n"
				"#var float2 ITexG : $vin.TEXCOORD2 : TEXCOORD2 : 4 : 1\n"
				"#var float2 ITexB : $vin.TEXCOORD3 : TEXCOORD3 : 5 : 1\n"
				"#var sampler2D IBlendSampler :  : texunit 0 : 6 : 1\n"
				"#var sampler2D ISamplerR :  : texunit 1 : 7 : 1\n"
				"#var sampler2D ISamplerG :  : texunit 2 : 8 : 1\n"
				"#var sampler2D ISamplerB :  : texunit 3 : 9 : 1\n"
				"#var float4 main : $vout.COLOR : COLOR : -1 : 1\n"
				"TEMP R0;\n"
				"TEMP R1;\n"
				"TEMP R2;\n"
				"TEMP R3;\n"
				"TEX R0, fragment.texcoord[2], texture[2], 2D;\n"
				"TEX R1.xyz, fragment.texcoord[0], texture[0], 2D;\n"
				"MUL R0, R0, R1.y;\n"
				"TEX R2, fragment.texcoord[1], texture[1], 2D;\n"
				"TEX R3, fragment.texcoord[3], texture[3], 2D;\n"
				"MAD R0, R2, R1.x, R0;\n"
				"MAD R0, R3, R1.z, R0;\n"
				"MUL result.color, R0, fragment.color.primary;\n"
				"END\n";

#if 0
                        string strTexturingFP =
                                "!!ARBfp1.0\n"
                                "#program main\n"
                                "#semantic main.IBlendSampler\n"
                                "#semantic main.ISamplerR\n"
                                "#semantic main.ISamplerB\n"
                                "#semantic main.ISamplerG\n"
                                "#semantic main.ISamplerA\n"
                                "#var float4 IColor : $vin.COLOR : COLOR : 1 : 1\n"
                                "#var float2 ITexBlend : $vin.TEXCOORD1 : TEXCOORD1 : 3 : 1\n"
                                "#var float2 ITexR : $vin.TEXCOORD1 : TEXCOORD1 : 3 : 1\n"
                                "#var float2 ITexG : $vin.TEXCOORD2 : TEXCOORD1 : 4 : 1\n"
                                "#var float2 ITexB : $vin.TEXCOORD3 : TEXCOORD1 : 5 : 1\n"
                                "#var float2 ITexA : $vin.TEXCOORD4 : TEXCOORD1 : 6 : 1\n"
                                "#var sampler2D IBlendSampler : TEXUNIT0 : texunit 0 : 7 : 1\n"
                                "#var sampler2D ISamplerR : TEXUNIT1 : texunit 1 : 8 : 1\n"
                                "#var sampler2D ISamplerG : TEXUNIT2 : texunit 2 : 9 : 1\n"
                                "#var sampler2D ISamplerB : TEXUNIT3 : texunit 3 : 10 : 1\n"
                                "#var sampler2D ISamplerA : TEXUNIT4 : texunit 4 : 11 : 1\n"
                                "#var float4 main : $vout.COLOR : COLOR : -1 : 1\n"
                                "TEMP R0;\n"
                                "TEMP R1;\n"
                                "TEMP R2;\n"
                                "TEMP R3;\n"
                                "TEX R0, fragment.texcoord[2], texture[2], 2D;\n"
                                "TEX R1, fragment.texcoord[0], texture[0], 2D;\n"
                                "MUL R0, R0, R1.y;\n"
                                "TEX R2, fragment.texcoord[1], texture[1], 2D;\n"
                                "TEX R3, fragment.texcoord[3], texture[3], 2D;\n"
                                "MAD R0, R2, R1.x, R0;\n"
                                "MAD R0, R3, R1.z, R0;\n"
                                "TEX R2, fragment.texcoord[4], texture[4], 2D;\n"
                                "MAD R0, R2, R1.w, R0;\n"
                                "MUL result.color, R0, fragment.color.primary;\n"
                                "END\n";
#endif

#if 0
                                        "float4 main( float4 IPosition   : POSITION,"
                                        "             float4 IColor      : COLOR,"
                                        "             float2 ITexBlend   : TEXCOORD0,"
                                        "             float2 ITexR       : TEXCOORD1,"
                                        "             float2 ITexG       : TEXCOORD2,"
                                        "             float2 ITexB       : TEXCOORD3,"
                                        "             float2 ITexA       : TEXCOORD4,"
                                        "  uniform sampler2D IBlendSampler,"
                                        "  uniform sampler2D ISamplerR,"
                                        "  uniform sampler2D ISamplerG,"
                                        "  uniform sampler2D ISamplerB,"
                                        "  uniform sampler2D ISamplerA ) : COLOR"
                                        "{"
                                        "    float4 BlendPixel  = tex2D( IBlendSampler, ITexBlend );"
                                        ""
                                        "    float4 TexColorR   = tex2D( ISamplerR, ITexR );"
                                        "    float4 TexColorG   = tex2D( ISamplerG, ITexG );"
                                        "    float4 TexColorB   = tex2D( ISamplerB, ITexB );"
                                        "    float4 TexColorA   = tex2D( ISamplerA, ITexA );"
                                        ""
                                        "    float4 ColorR = ( TexColorR * BlendPixel.r );"
                                        "    float4 ColorG = ( TexColorG * BlendPixel.g );"
                                        "    float4 ColorB = ( TexColorB * BlendPixel.b );"
                                        "    float4 ColorA = ( TexColorA * BlendPixel.a );"
                                        ""
                                        "    return ( ColorR + ColorG + ColorB + ColorA ) * IColor;"
                                        "}";
#endif

                                                                                                                        
                        if( !m_pkMaterial->m_pkFragmentShader->Compile( &strTexturingFP ) )
                        {
                                neolog << LogLevel( WARNING ) << "[TerrainPage::TerrainPage]: Could not compile fragment shader!  "
                                                              << "Falling back to software texturing." << endl;
                                                                                                                        
                                m_pkMaterial->m_pkFragmentShader = 0;
                                                                                                                        
                                m_eTexturingType = TEXTURE_SOFTWARE;
                        }
                        else
                        {
				neolog << "Fragment shader successfully compiled." << endl;

                                /*
                                m_pkBlendMapParam                      = m_pkMaterial->m_pkFragmentShader->GetParameter( "IBlendSampler" );
                                m_pkBlendMapParam->m_Value.m_uiTexture = 0;

                                m_pkTexture1Param                      = m_pkMaterial->m_pkFragmentShader->GetParameter( "ISamplerR" );
                                m_pkTexture1Param->m_Value.m_uiTexture = 1;

                                m_pkTexture2Param                      = m_pkMaterial->m_pkFragmentShader->GetParameter( "ISamplerG" );
                                m_pkTexture2Param->m_Value.m_uiTexture = 2;

                                m_pkTexture3Param                      = m_pkMaterial->m_pkFragmentShader->GetParameter( "ISamplerB" );
                                m_pkTexture3Param->m_Value.m_uiTexture = 3;

                                m_pkTexture4Param                      = m_pkMaterial->m_pkFragmentShader->GetParameter( "ISamplerA" );
                                m_pkTexture4Param->m_Value.m_uiTexture = 4;
                                */
                        }
                }
        }
        else
        {
                // ### TODO: Software texturing goes here
                neolog << LogLevel( ERROR ) << "[TerrainPage::TerrainPage]: Software texturing not yet finished." << endl;
        }

        if( m_eLightingType == LIGHT_HARDWARE )
        {
                // Reset the lighting to use lightmaps if we can't create the vertex shader
                if( !( m_pkMaterial->m_pkVertexShader = Core::Get()->GetRenderDevice()->CreateShader( Shader::VERTEXSHADER ) ) )
                {
                        neolog << LogLevel( WARNING ) << "[TerrainPage::TerrainPage]: Could not create vertex shader!  "
                               << "Falling back to software lightmaps." << endl;

                        m_eLightingType = LIGHT_SOFTWARE;
                }
                else
                {
                        // Setup the vertex shader for lighting
                        string strLightingVP =
#if 1
				"!!ARBvp1.0\n"
				"PARAM c6 = { 10, 0, 0, 0 };\n"
				"#program main\n"
				"#semantic main.ModelViewProjection\n"
				"#semantic main.ILight\n"
				"#semantic main.IAmbient\n"
				"#var float4 IPosition : $vin.POSITION0 : POSITION0 : 0 : 1\n"
				"#var float2 ITexCoords : $vin.TEXCOORD0 : TEXCOORD0 : 1 : 1\n"
				"#var float3 Normal : $vin.NORMAL : NORMAL : 2 : 1\n"
				"#var float4x4 ModelViewProjection :  : c[0], 4 : 3 : 1\n"
				"#var float4 ILight :  : c[4] : 4 : 1\n"
				"#var float4 IAmbient :  : c[5] : 5 : 1\n"
				"#var float4 OPosition : $vout.COLOR0 : COLOR0 : 7 : 1\n"
				"#var float4 OColor : $vout.COLOR0 : COLOR0 : 7 : 1\n"
				"#var float2 OTexCoords : $vout.TEXCOORD0 : TEXCOORD0 : 8 : 1\n"
				"#var float2 OTexCoords1 : $vout.TEXCOORD1 : TEXCOORD1 : 9 : 1\n"
				"#var float2 OTexCoords2 : $vout.TEXCOORD2 : TEXCOORD2 : 10 : 1\n"
				"#bind ModelViewProjection System.Matrix.MODELVIEWPROJECTION\n"
				"#bind IAmbient System.Light.0.AMBIENT\n"
				"TEMP R0;\n"
				"ATTRIB v18 = vertex.normal;\n"
				"ATTRIB v24 = vertex.texcoord[0];\n"
				"ATTRIB v16 = vertex.position;\n"
				"PARAM c5 = program.local[5];\n"
				"PARAM c4 = program.local[4];\n"
				"PARAM c0[4] = { program.local[0..3] };\n"
				"    MOV result.texcoord[0].xy, v24;\n"
				"    DP4 result.position.x, c0[0], v16;\n"
				"    DP4 result.position.y, c0[1], v16;\n"
				"    DP4 result.position.z, c0[2], v16;\n"
				"    DP4 result.position.w, c0[3], v16;\n"
				"    DP3 R0.x, v18.xyzx, c4.xyzx;\n"
				"    ADD result.color.front.primary, R0.x, c5;\n"
				"    MUL R0, v24.xyxx, c6.x;\n"
				"    MOV result.texcoord[2].xy, R0;\n"
				"    MOV result.texcoord[1].xy, R0;\n"
				"END\n";
#else
                                "!!ARBvp1.0\n"
                                "PARAM c6 = { 10, 0, 0, 0 };\n"
                                "#program main\n"
                                "#semantic main.ModelViewProjection\n"
                                "#semantic main.ILight\n"
                                "#semantic main.IAmbient\n"
                                "#var float4 IPosition : $vin.POSITION0 : POSITION0 : 0 : 1\n"
                                "#var float2 ITexCoords : $vin.TEXCOORD0 : TEXCOORD0 : 1 : 1\n"
                                "#var float3 Normal : $vin.NORMAL : NORMAL : 2 : 1\n"
                                "#var float4x4 ModelViewProjection : : c[0], 4 : 3 : 1\n"
                                "#var float4 ILight : : c[4] : 4 : 1\n"
                                "#var float4 IAmbient : : c[5] : 5 : 1\n"
                                "#var float4 OPosition : $vout.POSITION : POSITION : 6 : 1\n"
                                "#var OColor : $vout.COLOR0 : COLOR0 : 7 : 1\n"
                                "#var OTexCoords : $vout.TEXCOORD0 : TEXCOORD0 : 8 : 1\n"
                                "#var OTexCoords1 : $vout.TEXCOORD1 : TEXCOORD1 : 9 : 1\n"
                                "#var OTexCoords2 : $vout.TEXCOORD2 : TEXCOORD2 : 10 : 1\n"
                                "#var OTexCoords3 : $vout.TEXCOORD3 : TEXCOORD3 : 11 : 1\n"
                                "#var OTexCoords4 : $vout.TEXCOORD4 : TEXCOORD4 : 12 : 1\n"
                                "#bind ModelViewProjection System.Matrix.MODELVIEWPROJECTION\n"
//                              "#bind IAmbient System.Light.0.AMBIENT\n"
                                "TEMP R0;\n"
                                "ATTRIB v18 = vertex.normal;\n"
                                "ATTRIB v24 = vertex.texcoord[0];\n"
                                "ATTRIB v16 = vertex.position;\n"
                                "PARAM c5 = program.local[5];\n"
                                "PARAM c4 = program.local[4];\n"
                                "PARAM c0[4] = { program.local[0..3] };\n"
                                "    MOV result.texcoord[0].xy, v24;\n"
                                "    DP4 result.position.x, c0[0], v16;\n"
                                "    DP4 result.position.y, c0[1], v16;\n"
                                "    DP4 result.position.z, c0[2], v16;\n"
                                "    DP4 result.position.w, c0[3], v16;\n"
                                "    DP3 R0.x, v18.xyzx, c4.xyzx;\n"
                                "    ADD result.color.front.primary, R0.x, c5;\n"
                                "    MUL R0, v24.xyxx, c6.x;\n"
                                "    MOV result.texcoord[4].xy, R0;\n"
                                "    MOV result.texcoord[3].xy, R0;\n"
                                "    MOV result.texcoord[2].xy, R0;\n"
                                "    MOV result.texcoord[1].xy, R0;\n"
                                "END\n";
#endif
#if 0
                                "void main( in         float4   IPosition              : POSITION0,"
                                "           in         float2   ITexCoords             : TEXCOORD0,"
                                "           in         float3   Normal                 : NORMAL,"
                                "           in uniform float4x4 ModelViewProjection,"
                                "           in uniform float4   ILight,"
                                "           in uniform float4   IAmbient,"
                                "          out         float4   OPosition              : POSITION,"
                                "          out         float4   OColor                 : COLOR0,"
                                "          out         float2   OTexCoords             : TEXCOORD0,"
                                "          out         float2   OTexCoords1            : TEXCOORD1,"
                                "          out         float2   OTexCoords2            : TEXCOORD2,"
                                "          out         float2   OTexCoords3            : TEXCOORD3,"
                                "          out         float2   OTexCoords4            : TEXCOORD4 )"
                                "{"
                                ""
                                "    OPosition   = mul( ModelViewProjection, IPosition );"
                                "    OColor      = dot( Normal, ILight.rgb ) + IAmbient;"
                                ""
                                "    OTexCoords  = ITexCoords;"
                                "    OTexCoords1 = ITexCoords * ##hscale##;"
                                "    OTexCoords2 = ITexCoords * ##hscale##;"
                                "    OTexCoords3 = ITexCoords * ##hscale##;"
                                "    OTexCoords4 = ITexCoords * ##hscale##;"
                                "}";
#endif

                        // If we can't compile the vertex shader for some reason, default to software lightmaps
                        if( !m_pkMaterial->m_pkVertexShader->Compile( &strLightingVP ) )
                        {
                                neolog << LogLevel( WARNING ) << "[TerrainPage::TerrainPage]: Could not compile vertex shader!  "
                                       << "Falling back to software lightmaps." << endl;

                                m_pkMaterial->m_pkVertexShader = 0;

                                m_eLightingType = LIGHT_SOFTWARE;
                        }
			else
			{
				neolog << "Successfully compiled vertex shader." << endl;
			}

#if 0
                        m_kLight.Set( 0.6f, 0.6f, 0.0f );
                        m_pkLightParam                      = m_pkMaterial->m_pkVertexShader->GetParameter( "ILight" );
                        m_pkLightParam->m_eType             = ShaderParam::PPT_VECTOR;
                        m_pkLightParam->m_Value.m_pkVector  = &m_kLight;

                        m_pkTransformParam                      = m_pkMaterial->m_pkVertexShader->GetParameter( "ModelViewProjection" );
                        m_pkTransformParam->m_eType             = ShaderParam::PPT_SYSTEM;
                        m_pkTransformParam->m_Value.m_eSystemID = ShaderParam::PSP_MODELVIEWPROJ_MATRIX;

                        m_pkMaterial->m_pkVertexShader->UpdateParameter( m_pkTransformParam );
#endif

                        // Set a default light position
                        m_kLight.Set( 0.6f, 0.6f, 0.0f );
                        m_pkLightParam                          = m_pkMaterial->m_pkVertexShader->GetParameter( "ILight" );
                        m_pkLightParam->m_eType                 = ShaderParam::PPT_VECTOR;
                        m_pkLightParam->m_Value.m_pkVector      = &m_kLight;

			neolog << "Updating ILight parameter.." << endl;

                        m_pkMaterial->m_pkVertexShader->UpdateParameter( m_pkLightParam );

                        // Set a default ambient light color
                        m_kAmbient.Set( 0.3f, 0.3f, 0.6f );
                        m_pkAmbientParam                        = m_pkMaterial->m_pkVertexShader->GetParameter( "IAmbient" );
                        m_pkAmbientParam->m_eType               = ShaderParam::PPT_VECTOR;
                        m_pkAmbientParam->m_Value.m_pkVector    = &m_kAmbient;

			neolog << "Updating IAmbient parameter.." << endl;

                        m_pkMaterial->m_pkVertexShader->UpdateParameter( m_pkAmbientParam );
                }
        }

        if( m_eLightingType == LIGHT_SOFTWARE )
        {
        }
}
