/***************************************************************************
     texmatrixgen.cpp  -  Texture coordinate transform matrix generators
                             -------------------
    begin                : Tue Feb 11 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, texmatrixgen.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "texmatrixgen.h"
#include "nemath.h"
#include "timer.h"

using namespace std;


namespace NeoEngine
{


float TextureMatrixGen::s_afSinTable[ TEXMODFUNCTABLE_SIZE ];
float TextureMatrixGen::s_afSquareTable[ TEXMODFUNCTABLE_SIZE ];
float TextureMatrixGen::s_afTriangleTable[ TEXMODFUNCTABLE_SIZE ];
float TextureMatrixGen::s_afSawToothTable[ TEXMODFUNCTABLE_SIZE ];
float TextureMatrixGen::s_afInverseSawToothTable[ TEXMODFUNCTABLE_SIZE ];


void TextureMatrixGenScroll::GenerateMatrix( Matrix *pkMatrix )
{
	float fTimeScale = (float)((int64_t)Timer::GetHeartbeat()) / (float)((int64_t)Timer::GetHeartbeatDivisor());
	float fClampU, fClampV;

	fClampU = -m_afScrollSpeed[0] * fTimeScale;
	fClampV = -m_afScrollSpeed[1] * fTimeScale;

	// clamp so coordinates don't continuously get larger
	fClampU = fClampU - floorf( fClampU );
	fClampV = fClampV - floorf( fClampV );

	float fTransU = (*pkMatrix)[0][0] * fClampU + (*pkMatrix)[0][1] * fClampV;
	float fTransV = (*pkMatrix)[1][0] * fClampU + (*pkMatrix)[1][1] * fClampV;

	(*pkMatrix)[0][3] = fTransU;
	(*pkMatrix)[1][3] = fTransV;
}


TextureMatrixGen *TextureMatrixGenScroll::Duplicate()
{
	TextureMatrixGenScroll *pkGen = new TextureMatrixGenScroll;

	pkGen->m_afScrollSpeed[0] = m_afScrollSpeed[0];
	pkGen->m_afScrollSpeed[1] = m_afScrollSpeed[1];

	return pkGen;
}



void TextureMatrixGenScale::GenerateMatrix( Matrix *pkMatrix )
{
	(*pkMatrix)[0][0] *= m_afScale[0];
	(*pkMatrix)[1][1] *= m_afScale[1];

	(*pkMatrix)[0][3] *= m_afScale[0];
	(*pkMatrix)[1][3] *= m_afScale[1];
}


TextureMatrixGen *TextureMatrixGenScale::Duplicate()
{
	TextureMatrixGenScale *pkGen = new TextureMatrixGenScale;

	pkGen->m_afScale[0] = m_afScale[0];
	pkGen->m_afScale[1] = m_afScale[1];

	return pkGen;
}



void TextureMatrixGenStretch::GenerateMatrix( Matrix *pkMatrix )
{
	float fTimeScale = (float)((int64_t)Timer::GetHeartbeat()) / (float)((int64_t)Timer::GetHeartbeatDivisor());

	float fVal = 1.0f / ( m_fBase + m_pfTable[ int( ( m_fPhase + fTimeScale * m_fFrequency ) * (float)TEXMODFUNCTABLE_SIZE ) & TEXMODFUNCTABLE_MASK ] * m_fAmplitude );

	Matrix kMat;

	kMat[0][0] = fVal;
	kMat[0][3] = 0.5f - 0.5f * fVal;

	kMat[1][1] = fVal;
	kMat[1][3] = 0.5f - 0.5f * fVal;

	*pkMatrix = *pkMatrix * kMat;
}


TextureMatrixGen *TextureMatrixGenStretch::Duplicate()
{
	TextureMatrixGenStretch *pkGen = new TextureMatrixGenStretch;

	pkGen->m_pfTable     = m_pfTable;
	pkGen->m_fBase       = m_fBase;
	pkGen->m_fAmplitude  = m_fAmplitude;
	pkGen->m_fPhase      = m_fPhase;
	pkGen->m_fFrequency  = m_fFrequency;

	return pkGen;
}



void TextureMatrixGenTurbulence::GenerateMatrix( Matrix *pkMatrix )
{
	float fTimeScale = (float)((int64_t)Timer::GetHeartbeat()) / (float)((int64_t)Timer::GetHeartbeatDivisor());

	fTimeScale       = m_fPhase + fTimeScale * m_fFrequency;

	//(*pkMatrix)[0][3] += g_afSinTable[ int( ( ( xyz[0] + xyz[2] ) * 1.0/128 * 0.125 + fTimeScale ) * TEXMODFUNCTABLE_SIZE ) ) & TEXMODFUNCTABLE_MASK ] * m_fAmplitude
	//(*pkMatrix)[1][3] += g_afSinTable[ int( ( ( xyz[1]          ) * 1.0/128 * 0.125 + fTimeScale ) * TEXMODFUNCTABLE_SIZE ) ) & TEXMODFUNCTABLE_MASK ] * m_fAmplitude

	float fU = s_afSinTable[ int( ( ( 1.0f / 128.0f ) * 0.125f + fTimeScale ) * TEXMODFUNCTABLE_SIZE ) & TEXMODFUNCTABLE_MASK ] * m_fAmplitude;
	float fV = s_afSinTable[ int( ( ( 1.0f / 128.0f ) * 0.120f + fTimeScale ) * TEXMODFUNCTABLE_SIZE ) & TEXMODFUNCTABLE_MASK ] * m_fAmplitude;

	float fTransU = (*pkMatrix)[0][0] * fU + (*pkMatrix)[0][1] * fV;
	float fTransV = (*pkMatrix)[1][0] * fU + (*pkMatrix)[1][1] * fV;

	(*pkMatrix)[0][3] = fTransU;
	(*pkMatrix)[1][3] = fTransV;
}


TextureMatrixGen *TextureMatrixGenTurbulence::Duplicate()
{
	TextureMatrixGenTurbulence *pkGen = new TextureMatrixGenTurbulence;

	pkGen->m_fBase       = m_fBase;
	pkGen->m_fAmplitude  = m_fAmplitude;
	pkGen->m_fPhase      = m_fPhase;
	pkGen->m_fFrequency  = m_fFrequency;

	return pkGen;
}



void TextureMatrixGenTransform::GenerateMatrix( Matrix *pkMatrix )
{
	Matrix kMat;

	kMat[0][0] = m_aafRotate[0][0];
	kMat[0][1] = m_aafRotate[0][1];

	kMat[0][3] = m_afTranslate[0];

	kMat[1][0] = m_aafRotate[1][0];
	kMat[1][1] = m_aafRotate[1][1];

	kMat[1][3] = m_afTranslate[1];

	*pkMatrix = *pkMatrix * kMat;
}


TextureMatrixGen *TextureMatrixGenTransform::Duplicate()
{
	TextureMatrixGenTransform *pkGen = new TextureMatrixGenTransform;

	fmemcpy( pkGen->m_aafRotate, m_aafRotate, sizeof( float ) * 4 );

	pkGen->m_afTranslate[0] = m_afTranslate[0];
	pkGen->m_afTranslate[1] = m_afTranslate[1];

	return pkGen;
}



void TextureMatrixGenRotate::GenerateMatrix( Matrix *pkMatrix )
{
	float fTimeScale = (float)((int64_t)Timer::GetHeartbeat()) / (float)((int64_t)Timer::GetHeartbeatDivisor());
	float fRads;
	float fSinValue, fCosValue;

	fRads = -m_fRotateSpeed * fTimeScale;

	int iIndex = int( fRads * ( (float)TEXMODFUNCTABLE_SIZE / TWO_PI ) );

	fSinValue = s_afSinTable[ iIndex & TEXMODFUNCTABLE_MASK ];
	fCosValue = s_afSinTable[ ( iIndex + ( TEXMODFUNCTABLE_SIZE / 4 ) ) & TEXMODFUNCTABLE_MASK ];

	Matrix kMat;

	kMat[0][0] =  fCosValue;
	kMat[0][1] = -fSinValue;
	kMat[0][3] = 0.5f - 0.5f * ( fCosValue - fSinValue );

	kMat[1][0] =  fSinValue;
	kMat[1][1] =  fCosValue;
	kMat[1][3] = 0.5f - 0.5f * ( fSinValue + fCosValue );

	*pkMatrix = *pkMatrix * kMat;
}


TextureMatrixGen *TextureMatrixGenRotate::Duplicate()
{
	TextureMatrixGenRotate *pkGen = new TextureMatrixGenRotate;

	pkGen->m_fRotateSpeed = m_fRotateSpeed;

	return pkGen;
}



TextureMatrixGen *TextureMatrixGenFactory::CreateTextureMatrixGen( const std::string &rstrName )
{
	if( rstrName == "scroll" )
		return new TextureMatrixGenScroll;
	else if( rstrName == "scale" )
		return new TextureMatrixGenScale;
	else if( rstrName == "stretch" )
		return new TextureMatrixGenStretch;
	else if( rstrName == "transform" )
		return new TextureMatrixGenTransform;
	else if( rstrName == "turbulence" )
		return new TextureMatrixGenTurbulence;
	else if( rstrName == "rotate" )
		return new TextureMatrixGenRotate;

	return 0;
}


};
