/***************************************************************************
                      thread.cpp  -  Thread abstraction
                             -------------------
    begin                : Tue Feb 12 2002
    copyright            : (C) 2002 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, thread.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "thread.h"
#include "taskmanager.h"
#include "logstream.h"

#ifdef WIN32
#  include <windows.h>
#elif defined(POSIX) || defined(__APPLE__)
#  include <unistd.h>
#  include <sys/time.h>
#endif

#if 0
//Temporary fix, does macosx have nanosleep?
#if defined(__APPLE__) && defined(HAVE_NANOSLEEP)
#  undef HAVE_NANOSLEEP
#endif
#endif

using namespace std;


namespace NeoEngine
{


#ifdef WIN32
unsigned long __stdcall threadstarter( void *pData )
#elif defined(POSIX) || defined (__APPLE__)
void *threadstarter( void *pData )
#else
#  error threadstarter not implemented
#endif
{
	Thread *pkThread = (Thread*)pData;
	void   *pUserArg = pkThread->GetUserData();

	neolog << LogLevel( DEBUG ) << "[threadstarter] entering thread [" << pkThread->GetThreadID() << "] method" << endl;

	return 
#ifdef WIN32
		(unsigned long)
#endif
		pkThread->GetThreadMethod()->ThreadEntry( pkThread, pUserArg );
}



int Thread::s_iThreadCounter = 0;


Thread::Thread( ThreadMethod *pkMethod, void *pData, ThreadCallback *pkCallback, TaskManager *pkTaskManager )
{
	m_bTerminate     = false;
	m_bTerminated    = false;
	m_bStarted       = false;
	m_pkMethod       = pkMethod;
	m_pkCallback     = pkCallback;
	m_pUserData      = pData;
	m_pkTaskManager  = pkTaskManager;
	m_iThreadID      = s_iThreadCounter++;

	neolog << LogLevel( DEBUG ) << "starting thread [" << m_iThreadID << "]" << endl;

#ifdef WIN32
	CreateThread( 0, 0, threadstarter, this, 0, &m_ulThread );
#elif defined(POSIX) || defined(__APPLE__)
	m_kThread = 0;
	if( pthread_create( &m_kThread, 0, threadstarter, this ) )
		neolog << LogLevel( ERROR ) << "*** unable to create thread" << endl;
#else
#  error Thread::Thread not implemented
#endif

	neolog << LogLevel( DEBUG ) << "thread [" << m_iThreadID << "] started" << endl;

	if( m_pkCallback )
	{
		neolog << LogLevel( DEBUG ) << "thread [" << m_iThreadID << "] callback notification" << endl;
		m_pkCallback->ThreadEvent( m_iThreadID, START );
	}
}


Thread::~Thread()
{
	m_bTerminate = true;

#ifdef WIN32
	if( m_bStarted )
	{
		while( !m_bTerminated )
		{
			neolog << LogLevel( DEBUG ) << "[Thread::~Thread] waiting for thread [" << m_iThreadID << "] to die" << endl;
			SleepEx( 1000, FALSE );

			m_bTerminate = true;
		}
	}
#elif defined(POSIX) || defined(__APPLE__)
	if( m_bStarted )
		neolog << LogLevel( DEBUG ) << "[Thread::~Thread] waiting for thread [" << m_iThreadID << "] to die" << endl;
		
	pthread_join( m_kThread, 0 );
#else
#  error Thread::~Thread not implemented
#endif

	neolog << LogLevel( DEBUG ) << "[Thread::~Thread] thread [" << m_iThreadID << "] killed and joined" << endl;
	
	if( m_pkCallback )
		m_pkCallback->ThreadEvent( m_iThreadID, TERMINATE );
		
	if( m_pkTaskManager )
		m_pkTaskManager->Remove( this );
}




void ThreadMethod::Sleep( int iTime )
{
#ifdef WIN32
	SleepEx( iTime, TRUE );
#elif defined(POSIX) || defined(__APPLE__)

#  ifdef HAVE_NANOSLEEP
	timespec t;

	t.tv_sec  = iTime / 1000;
	t.tv_nsec = (long)( iTime % 1000 ) * 1000000L;

	nanosleep( &t, 0 );
#  elif defined(HAVE_USLEEP)
	usleep( iTime );
#  else
#    error ThreadMethod::Sleep not implemented (no nanosleep or usleep)
#  endif
#else
#  error ThreadMethod::Sleep not implemented
#endif
}



}; // namespace NeoEngine

