/***************************************************************************
                              timer.cpp  -  Timer
                             -------------------
    begin                : Tue Nov 6 2001
    copyright            : (C) 2001 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, timer.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2001
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "timer.h"
#include "core.h"
#include "logstream.h"

#ifdef WIN32
#  include <windows.h>
#elif defined(POSIX) || defined(__APPLE__)
#include <unistd.h>
#include <sys/time.h>
#endif

#if defined( POSIX )
   //Posix-specific includes
#  include <sys/times.h>
#  include <unistd.h>
#  include <time.h>
#endif

#if defined(LINUX) && defined(HAVE_RTC)
#  include <linux/rtc.h>
#endif

#ifdef __APPLE__

#  ifndef HAVE_NANOSLEEP
#    define HAVE_NANOSLEEP
#  endif

#  ifndef HAVE_USLEEP
#    define HAVE_USLEEP
#  endif

#endif


namespace NeoEngine
{


uint64_t Timer::s_uliHeartbeat         = 0;
uint64_t Timer::s_uliDivisor           = 1;
uint64_t Timer::s_uliOldHeartbeatClock = 0;

Timer g_kHeartbeatTimer;


Timer::Timer()
{
	Reset();
}


float Timer::GetDeltaTime( bool bReset )
{
	float fDeltaTime = 0.0f;

#ifdef WIN32

	uint64_t uliCurClock;
	uint64_t uliFrequency;

	QueryPerformanceCounter( (LARGE_INTEGER*)&uliCurClock );
	QueryPerformanceFrequency( (LARGE_INTEGER*)&uliFrequency );

	fDeltaTime = float( double( (__int64)( uliCurClock - m_uliOldClock ) ) / double( (__int64)uliFrequency ) );

#elif defined(POSIX) || defined(__APPLE__)

	static bool  s_bInvalidDetected = false;
	static float s_fLastDeltaTime   = 0.0001f;

	timeval tv; tv.tv_sec = 0; tv.tv_usec = 0;

	if( gettimeofday( &tv, 0 ) < 0 )
		neolog << LogLevel( ERROR ) << "[Timer::GetDeltaTime] gettimeofday returned error" << std::endl;

	uint64_t uliCurClock = uint64_t( tv.tv_sec ) * 1000000LL + (uint64_t)tv.tv_usec;

	if( uliCurClock < m_uliOldClock )
	{
		if( !s_bInvalidDetected )
		{
			neolog << LogLevel( ERROR ) << "[Timer::GetDeltaTime] gettimeofday returned invalid timestamp, uliCurClock ["
			       << neolog.width( 11 ) << uliCurClock << "], m_uliOldClock [" << neolog. width( 11 ) << m_uliOldClock << "] (only be reported once)" << std::endl;
			s_bInvalidDetected = true;
		}

		return s_fLastDeltaTime;
	}

	fDeltaTime = s_fLastDeltaTime = float( double( uliCurClock - m_uliOldClock ) / 1000000.0 );

#else
#  error Timer::GetDeltaTime not implemented
#endif

	if( bReset )
		m_uliOldClock = uliCurClock;

	return fDeltaTime;
}


void Timer::Reset()
{
#ifdef WIN32
	QueryPerformanceCounter( (LARGE_INTEGER*)&m_uliOldClock );
#elif defined(POSIX) || defined(__APPLE__)
	timeval tv;

	gettimeofday( &tv, 0 );

	m_uliOldClock = uint64_t( tv.tv_sec ) * 1000000LL + (uint64_t)tv.tv_usec;
#else
#  error Timer::Reset not implemented
#endif
}


uint64_t Timer::GetHeartbeat()
{
#ifdef WIN32

	uint64_t uliCurClock;

	QueryPerformanceCounter( (LARGE_INTEGER*)&uliCurClock );

#elif defined(POSIX) || defined(__APPLE__)

	timeval tv;

	gettimeofday( &tv, 0 );

	uint64_t uliCurClock = uint64_t( tv.tv_sec ) * 1000000LL + (uint64_t)tv.tv_usec;

#else
#  error Timer::GetHeartbeat not implemented
#endif

	s_uliHeartbeat = uliCurClock - s_uliOldHeartbeatClock;

	return s_uliHeartbeat;
}
		
uint64_t Timer::GetHeartbeatDivisor()
{
	return s_uliDivisor;
}


}; // namespace NeoEngine
