/***************************************************************************
                           udp.cpp  -  UDP sockets
                             -------------------
    begin                : Sun Jun 1 2003
    copyright            : (C) 2003 by Keaton Mullis
    email                : kmullis@zianet.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, udp.cpp

 The Initial Developer of the Original Code is Keaton Mullis.
 Portions created by Keaton Mullis are Copyright (C) 2003
 Keaton Mullis. All Rights Reserved.

 ***************************************************************************/

#include "udp.h"
#include "thread.h"
#include "logstream.h"

#ifdef WIN32
#  include <winsock2.h>
#  undef ERROR
#elif defined(POSIX) || defined(__APPLE__)
#  include <sys/types.h>
#  include <errno.h>
#  ifdef __APPLE__
#    define _BSD_SOCKLEN_T_
#  endif
#  include <sys/socket.h>
#  include <arpa/inet.h>
#  include <netinet/in.h>
#  include <unistd.h>
#  include <fcntl.h>
#  include <netdb.h>
#endif

#ifdef POSIX
#  include <sys/poll.h>
#elif __APPLE__
#  include <sys/select.h>
#endif

#include <assert.h>

#ifdef __APPLE__
#  define SOCKET_ERROR    -1
#endif /* __APPLE__ */

using namespace std;


namespace NeoEngine
{


UDPSocket::UDPSocket( SocketCallback *pkCallback ) :
	Socket( pkCallback )
{
	m_iSocket        = socket( PF_INET, SOCK_DGRAM, 0 );
}


UDPSocket::UDPSocket( int iSocket, SocketCallback *pkCallback ) :
	Socket( pkCallback )
{
	m_iSocket        = iSocket;
}


UDPSocket::~UDPSocket()
{
	if( m_iSocket )
	{
		neolog << LogLevel( DEBUG ) << "[UDPSocket::~UDPSocket] close and shutdown socket" << endl;

		SetBlocking( false );

#ifdef WIN32
		closesocket( m_iSocket );
		shutdown( m_iSocket, SD_BOTH );
#elif defined(POSIX) || defined(__APPLE__)
		close( m_iSocket );
		shutdown( m_iSocket, SHUT_RDWR );
#else
#  error UDPSocket::~UDPSocket not implemented
#endif

		m_iSocket = 0;
	}
}


bool UDPSocket::Listen( int iPort, SocketCallback *pkCallback )
{
	assert( m_iSocket );

	if( !m_pkCallback && pkCallback )
		m_pkCallback = pkCallback;

	sockaddr_in saddr;

	saddr.sin_family = AF_INET;

	//Listen to any address
#ifdef WIN32
	saddr.sin_addr.S_un.S_addr = INADDR_ANY;
#elif defined(POSIX) || defined(__APPLE__)
	saddr.sin_addr.s_addr      = INADDR_ANY;
#else
#  error UDPSocket::Listen not implemented
#endif
	saddr.sin_port             = htons( iPort );

	if( bind( m_iSocket, (sockaddr*)&saddr, sizeof( sockaddr_in ) ) )
	{
		neolog << LogLevel( ERROR ) << "[UDPSocket::Listen] error binding socket: " << strerror( errno ) << endl;
		return false;
	}
	return true;
}

void UDPSocket::SetBlocking( bool bBlock )
{
#ifdef WIN32
	unsigned long ulParam = bBlock ? 1 : 0;
	ioctlsocket( m_iSocket, FIONBIO, &ulParam );
#elif defined(POSIX) || defined(__APPLE__)
	long lBlockingMode = fcntl( m_iSocket, F_GETFL );
	if ( bBlock )
	{
		lBlockingMode &= ~O_NONBLOCK;
	}
	else
	{
		lBlockingMode |= O_NONBLOCK;
	}
	if ( fcntl( m_iSocket, F_SETFL, lBlockingMode ) < 0 )
	{
		neolog << LogLevel( ERROR ) << "*** [TCPSocket::SetBlocking] error set blocking mode: " << strerror( errno ) << endl;
	}
#else
#  error UDPSocket::SetBlocking not implemented
#endif
}


int UDPSocket::Read( void *pDst, int iBytes, string &rstrFromAddr, int &riFromPort )
{
	assert( m_iSocket );

	sockaddr_in sfrom;

#ifdef WIN32
	int size = sizeof( sockaddr_in );
#else
	socklen_t size = sizeof( sockaddr_in );
#endif

	int iRet = recvfrom( m_iSocket, (char*)pDst, iBytes, 0, (sockaddr*)&sfrom, &size );

	rstrFromAddr = inet_ntoa(sfrom.sin_addr);

	riFromPort = ntohs(sfrom.sin_port);

	if( iRet < 0 )
		neolog << LogLevel( ERROR ) << "[UDPSocket::Read] unable to read from socket: " << strerror( errno ) << endl;
	else if( !iRet )
		neolog << LogLevel( WARNING ) << "[UDPSocket::Read] socket closed" << endl;
	else
		return iRet;

	return 0;
}


int UDPSocket::Write( const void *pSrc, int iBytes, string strToAddr, int iToPort )
{
	assert( m_iSocket );

	hostent *pHost = gethostbyname( strToAddr.c_str() );

	sockaddr_in saddr;

	// Write to target
	saddr.sin_family = AF_INET;
	saddr.sin_addr = *(in_addr*)pHost->h_addr_list[ 0 ];
	saddr.sin_port = htons( iToPort );

	int iRet = sendto( m_iSocket, (char*)pSrc, iBytes, 0, (sockaddr*)&saddr, sizeof( sockaddr_in ) );

	if( iRet < 0 )
		neolog << LogLevel( ERROR ) << "[UDPSocket::Write] unable to send to socket: " << strerror( errno ) << endl;
	else
		return iRet;

	return 0;
}


unsigned int UDPSocket::Poll( int iTimeout )
{
#if defined(WIN32) || defined(__APPLE__)

	fd_set fdread;

	FD_ZERO( &fdread );
	FD_SET( (unsigned int)m_iSocket, &fdread );

	fd_set fderr;

	FD_ZERO( &fderr );
	FD_SET( (unsigned int)m_iSocket, &fderr );

	timeval timeout;

	timeout.tv_sec  = iTimeout / 1000;
	timeout.tv_usec = ( iTimeout % 1000 ) * 1000;

	if( SOCKET_ERROR == select( 1, &fdread, 0, &fderr, &timeout ) )
		return Socket::INVALID;

	unsigned int uiRet = 0;

	if( FD_ISSET( m_iSocket, &fdread ) )
		uiRet |= Socket::DATAIN;
	if( FD_ISSET( m_iSocket, &fderr ) )
		uiRet |= Socket::ERR;
	if( !uiRet )
		uiRet  = Socket::TIMEOUT;

	return uiRet;

#elif POSIX

	pollfd pfd;

	pfd.fd      = m_iSocket;
	pfd.events  = POLLIN | POLLPRI | POLLERR | POLLHUP | POLLNVAL;
	pfd.revents = 0;

	int          iRet    = 0;
	unsigned int uiEvent = 0;

	if( ( iRet = poll( &pfd, 1, iTimeout ) ) < 0 )
		return Socket::INVALID;
	else if( !iRet )
		return Socket::TIMEOUT;

	if( pfd.revents & POLLIN )
		uiEvent |= Socket::DATAIN;
	if( pfd.revents & POLLERR )
		uiEvent |= Socket::ERR;
	if( pfd.revents & POLLHUP )
		uiEvent |= Socket::HUP;
	if( pfd.revents & POLLNVAL )
		uiEvent |= Socket::NVAL;

	return uiEvent;

#else
#  error UDPSocket::Poll not implemented
#endif
}


}; // namespace NeoEngine
