/***************************************************************************
            vertexdecl.cpp  -  Flexible vertex format declaration
                             -------------------
    begin                : Thu Jul 31 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, vertexdecl.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2003
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "vertexdecl.h"
#include "util.h"


namespace NeoEngine
{


VertexDeclaration::VertexDeclaration( unsigned int uiNumElements ) :
	m_pkElements( new VertexElement[ uiNumElements ] ),
	m_uiNumElements( uiNumElements )
{
}


VertexDeclaration::VertexDeclaration( const VertexDeclaration &rkDeclaration ) :
	m_pkElements( new VertexElement[ rkDeclaration.m_uiNumElements ] ),
	m_uiNumElements( rkDeclaration.m_uiNumElements )
{
	if( m_uiNumElements )
		fmemcpy( m_pkElements, rkDeclaration.m_pkElements, sizeof( VertexElement ) * m_uiNumElements );
}


VertexDeclaration::~VertexDeclaration()
{
	delete [] m_pkElements;
}


VertexElement *VertexDeclaration::GetElement( unsigned int uiType, unsigned int uiUsage )
{
	for( unsigned int uiElement = 0; uiElement < m_uiNumElements; ++uiElement )
		if( ( ( uiType  == VertexElement::UNKNOWNTYPE  ) || ( m_pkElements[ uiElement ].m_uiType  == uiType  ) ) && 
		    ( ( uiUsage == VertexElement::UNKNOWNUSAGE ) || ( m_pkElements[ uiElement ].m_uiUsage == uiUsage ) ) )
			return &m_pkElements[ uiElement ];

	return 0;
}


const VertexElement *VertexDeclaration::GetElement( unsigned int uiType, unsigned int uiUsage ) const
{
	for( unsigned int uiElement = 0; uiElement < m_uiNumElements; ++uiElement )
		if( ( ( uiType  == VertexElement::UNKNOWNTYPE  ) || ( m_pkElements[ uiElement ].m_uiType  == uiType  ) ) && 
		    ( ( uiUsage == VertexElement::UNKNOWNUSAGE ) || ( m_pkElements[ uiElement ].m_uiUsage == uiUsage ) ) )
			return &m_pkElements[ uiElement ];

	return 0;
}


unsigned int VertexDeclaration::GetVertexSize() const
{
	unsigned int uiSize = 0;

	for( unsigned int uiElement = 0; uiElement < m_uiNumElements; ++uiElement )
		uiSize += ( m_pkElements[ uiElement ].m_uiType < 4 ) ? ( m_pkElements[ uiElement ].m_uiType + 1 ) * 4 : 4;

	return uiSize;
}


VertexDeclaration &VertexDeclaration::operator = ( const VertexDeclaration &rkDeclaration )
{
	delete [] m_pkElements;

	m_uiNumElements = rkDeclaration.m_uiNumElements;

	if( m_uiNumElements )
	{
		m_pkElements = new VertexElement[ m_uiNumElements ];

		fmemcpy( m_pkElements, rkDeclaration.m_pkElements, sizeof( VertexElement ) * m_uiNumElements );
	}
	else
		m_pkElements = 0;

	return( *this );
}


bool VertexDeclaration::operator == ( const VertexDeclaration &rkDeclaration ) const
{
	if( m_uiNumElements != rkDeclaration.m_uiNumElements )
		return false;

	for( unsigned int uiElement = 0; uiElement < m_uiNumElements; ++uiElement )
		if( !( m_pkElements[ uiElement ] == rkDeclaration.m_pkElements[ uiElement ] ) )
			return false;

	return true;
}


VertexDeclaration *VertexDeclaration::Duplicate() const
{
	return new VertexDeclaration( *this );
}


void VertexDeclaration::InterpolateArray( unsigned char *pucFrom, unsigned char *pucTo, unsigned char *pucDest, float fFactor, unsigned int uiNumVertices, std::vector< bool > *pvbMask ) const
{
	unsigned int uiSize = GetVertexSize();

	for( unsigned int uiVertex = 0; uiVertex < uiNumVertices; ++uiVertex, pucFrom += uiSize, pucTo += uiSize, pucDest += uiSize )
	{
		if( pvbMask && !(*pvbMask)[ uiVertex ] )
			continue;

		for( unsigned int uiElement = 0; uiElement < m_uiNumElements; ++uiElement )
		{
			const VertexElement &kElement = m_pkElements[ uiElement ];

			if( kElement.m_uiType < VertexElement::COLOR32 )
			{
				//Float interpolation
				float *pfLast = (float*)( pucFrom + kElement.m_uiOffset );
				float *pfNext = (float*)( pucTo   + kElement.m_uiOffset );
				float *pfDest = (float*)( pucDest + kElement.m_uiOffset );

				// ### FIXME: SIMD anyone? :)
				for( unsigned int uiFloat = 0; uiFloat < ( kElement.m_uiType + 1 ); ++uiFloat, ++pfLast, ++pfNext, ++pfDest )
					*pfDest = *pfLast + ( *pfNext - *pfLast ) * fFactor;

				// OUCH!
				if( kElement.m_uiUsage == VertexElement::NORMAL )
					((Vector3d*)( pucDest + kElement.m_uiOffset ))->Normalize();
			}
			else
			{
				//Color interpolation, each channel separately
				unsigned int uiSrcColorOne = *((unsigned int*)( pucFrom + kElement.m_uiOffset ) );
				unsigned int uiSrcColorTwo = *((unsigned int*)( pucTo   + kElement.m_uiOffset ) );

				// ### FIXME: SIMD anyone? :)
				float fCol[4][2] = { { float(   uiSrcColorOne         & 0xFF ), float(   uiSrcColorTwo         & 0xFF ) },
				                     { float( ( uiSrcColorOne >> 8  ) & 0xFF ), float( ( uiSrcColorTwo >> 8  ) & 0xFF ) },
				                     { float( ( uiSrcColorOne >> 16 ) & 0xFF ), float( ( uiSrcColorTwo >> 16 ) & 0xFF ) },
				                     { float( ( uiSrcColorOne >> 24 ) & 0xFF ), float( ( uiSrcColorTwo >> 24 ) & 0xFF ) } };

				int iFin[4]      = { int( fCol[0][0] + ( fCol[0][1] - fCol[0][0] ) * fFactor ),
				                     int( fCol[1][0] + ( fCol[1][1] - fCol[1][0] ) * fFactor ),
				                     int( fCol[2][0] + ( fCol[2][1] - fCol[2][0] ) * fFactor ),
				                     int( fCol[3][0] + ( fCol[3][1] - fCol[3][0] ) * fFactor ) };

				*((unsigned int*)( pucDest + kElement.m_uiOffset )) = ( iFin[0] & 0xFF ) | ( ( iFin[1] << 8 ) & 0xFF00 ) | ( ( iFin[2] << 16 ) & 0xFF0000 ) | ( ( iFin[3] << 24 ) & 0xFF000000 );
			}
		}
	}
}


}; // namespace NeoEngine
