/***************************************************************************
                        neofcbzip2.cpp  -  description
                             -------------------
    begin                : Mon Apr 29 2002
    copyright            : (C) 2002 by Mattias Jansson
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoFCBZip2, neofcbzip2.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "neofcbzip2.h"
#include "link.h"

#include <neoengine/file.h>
#include <neoengine/package.h>
#include <neoengine/logstream.h>

#include "bzlib.h"

using namespace std;
using namespace NeoEngine;


namespace NeoBZIP2
{

#ifdef BUILD_STATIC
#  define BZIP2EXPORT

	ModuleStatic *g_pkStaticModule = 0;

#else

#  ifdef WIN32
#    define BZIP2EXPORT __declspec(dllexport)
#  else
#    define BZIP2EXPORT
#  endif

extern "C"
{

#endif

	BZIP2EXPORT FileCodec *LoadFileCodec()
	{
		return new NeoBZIP2::FileCodec();
	}

	BZIP2EXPORT int Initialize()
	{
		return 0;
	}

	BZIP2EXPORT int Shutdown()
	{
#ifdef BUILD_STATIC
		delete g_pkStaticModule;
#endif
		return 0;
	}

	BZIP2EXPORT void GetVersion( std::string *pstrName, int *piMajor, int *piMinor, int *piRevision )
	{
		*pstrName   = "NeoFCBzip2";
		*piMajor    = NEOENGINEVERSION_MAJOR;
		*piMinor    = NEOENGINEVERSION_MINOR;
		*piRevision = NEOENGINEVERSION_REVISION;
	}

#ifdef BUILD_STATIC

#else

}

#endif



FileCodec::FileCodec() :
	NeoEngine::FileCodec( "bzip2 compressed file" )
{
	neolog << LogLevel( INFO ) << "Loading bzip2 file codec" << endl << "  Based on the libbzip2 library, copyright (C) 1996-2002, Julian R Seward" << endl << "  For license details, see the LICENSE" << endl;

	m_iFileFlag = Package::FILEFLAG_BZIP2_COMPRESSED;
}


bool FileCodec::IsType( File *pkFile )
{
	string strID;
	char   szID[5] = { 0, 0, 0, 0, 0 };
	int iMajorVersion;
	int iMinorVersion;

	bool bBinary = pkFile->SetBinary( true );
	int  iOffset = pkFile->Tellg();
	
	if( iOffset != 0 )
	{
		pkFile->Seekg( 0, ios_base::beg );
	}
	pkFile->Read( &szID, 4 );
	*pkFile >> iMajorVersion >> iMinorVersion;
	pkFile->Seekg( iOffset, ios_base::beg );
	pkFile->SetBinary( bBinary );

	if( ( strID = szID ) != "NPAC" )
	{
		neolog << LogLevel( ERROR ) << "[Package::LoadNode]: invalid package file ID" << endl;
		return false;
	}

	if( ( iMajorVersion != 0 ) || ( iMinorVersion != 1 ) )
	{
		neolog << LogLevel( ERROR ) << "[Package::LoadNode]: invalid package file version" << endl;
		return false;
	}

	return true;
}


int FileCodec::LoadFileData( File *pkFile, int iSrcSize, void *pData, int *piSize )
{
	unsigned int iDestLen    = (unsigned int)*piSize;
	unsigned int iSrcLen     = iSrcSize;
	unsigned char *pucSrcBuf = new unsigned char[ iSrcLen ];
	int iRet = -1;

	unsigned int uiCurPos = pkFile->Tellg();

	pkFile->Read( pucSrcBuf, iSrcLen );

	pkFile->Seekg( uiCurPos, ios_base::beg );

	int iBZRet;

	int iVerbosity = 0;

	iBZRet = BZ2_bzBuffToBuffDecompress( (char*)pData, &iDestLen, (char*)pucSrcBuf, iSrcLen, 0, iVerbosity );

	if( iVerbosity > 0 )
		printf( "\n" );

	if( iBZRet == BZ_OK )
	{
		iRet = 0;
		*piSize = iDestLen;
	}
	else
		neolog << LogLevel( ERROR ) << "*** Unable to uncompress BZIP2 file: Uncompress failed [" << iBZRet << "]" << endl;

	delete [] pucSrcBuf;

	return iRet;
}



                                                                                                            
                                                                                                            
#ifdef BUILD_STATIC
                                                                                                            

Link::Link()
{
#if defined(WIN32) && defined(_DEBUG)
	g_pkStaticModule = new ModuleStatic( "neofcbzip2_debug" );
#else
	g_pkStaticModule = new ModuleStatic( "neofcbzip2" );
#endif

	g_pkStaticModule->m_Symbols.Insert( "LoadFileCodec", (void**)LoadFileCodec );
	g_pkStaticModule->m_Symbols.Insert( "GetVersion",    (void**)GetVersion    );
	g_pkStaticModule->m_Symbols.Insert( "Initialize",    (void**)Initialize    );
	g_pkStaticModule->m_Symbols.Insert( "Shutdown",      (void**)Shutdown      );

	NeoEngine::Core::GetModuleManager()->InsertModule( g_pkStaticModule );
}


#endif
                                                                                                            

}; // namespace NeoBZIP2
