/***************************************************************************
                 terrain.cpp  -  GeoMipMap terrain implementation
                             -------------------
    begin                : Thu Mar 18 2004
    copyright            : (C) 2004 by Cody Russell
    email                : cody `at' jhu.edu
 ***************************************************************************
 
 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoGeoMip, terrain.cpp

 The Initial Developer of the Original Code is Cody Russell.
 Portions created by Cody Russell are Copyright (C) 2004
 Cody Russell. All Rights Reserved.
 
 ***************************************************************************/

#include "terrain.h"

#include <neoengine/aabb.h>
#include <neoengine/camera.h>
#include <neoengine/scenenode.h>
#include <neoengine/vertex.h>

using namespace NeoEngine;
using namespace std;

namespace NeoGeoMip
{

class MipCameraCallback : public SceneNodeCallback
{
        public:

                virtual void NodeChanged( SceneNode *pkCameraNode )
                {
			Camera *pkCamera = (Camera*) pkCameraNode;
			Quadtree *pkQuadtree = (Quadtree*) pkCamera->GetRoom();

			pkQuadtree->SetDirty( true );
		}

                virtual void NodeDeleted( SceneNode *pkCamera )
                {
                }
};


MipData::MipData( MipTerrainBlock *pkBlock, int lod, int iIncrement, int iSize )
{
        int index = 0;
	int iMapSize;
	int iNumVertices;
	int iBlockSize;
	Polygon *pkPolygon;

	m_iIncrement = iIncrement;

	iMapSize = dynamic_cast< TerrainPage* >( pkBlock->m_pkMipPage )->GetSize();
	iNumVertices = iSize * iSize;
	iBlockSize = pkBlock->GetSize();

	m_pkPolygons = Core::Get()->GetRenderDevice()->CreatePolygonBuffer( Buffer::STATIC, iSize * iSize * 2 );

	m_pkPolygons->Lock( Buffer::WRITE );

	for( int y = 0; y < iBlockSize; y += m_iIncrement )
	{
		for( int x = 0; x < iBlockSize; x += m_iIncrement )
		{
			pkPolygon = m_pkPolygons->GetPolygon( index );
			(*pkPolygon)[0] = x + ( y + m_iIncrement ) * ( iBlockSize + 1 );
			(*pkPolygon)[1] = x + m_iIncrement + ( y + m_iIncrement ) * ( iBlockSize + 1 );
			(*pkPolygon)[2] = x + m_iIncrement + y * ( iBlockSize + 1 );

			pkPolygon = m_pkPolygons->GetPolygon( index + 1 );

			(*pkPolygon)[0] = x + ( y + m_iIncrement ) * ( iBlockSize + 1 );
			(*pkPolygon)[1] = x + m_iIncrement + y * ( iBlockSize + 1 );
			(*pkPolygon)[2] = x + y * ( iBlockSize + 1 );

			index += 2;
                }
        }

	m_pkPolygons->Unlock();
}


MipTerrainBlock::~MipTerrainBlock()
{
	delete m_pfDistanceTable;
}


MipTerrainBlock::MipTerrainBlock( int iOffsetX, int iOffsetY, int iBlockSizeSize, int iIndex, TerrainMaterialFactory *pkMaterialFactory, MipTerrain *pkMipPage ) :
	TerrainBlock( iOffsetX, iOffsetY, iBlockSizeSize, iIndex ),
	m_pkMipPage( pkMipPage ),
	m_pkMaterialFactory( pkMaterialFactory )
{
        int iIncrement;
	int iMaxLOD = dynamic_cast< MipTerrainManager* >( Core::Get()->GetTerrainManager() )->GetMaxDetailLevels();

	m_pfDistanceTable = new float[ iMaxLOD ];

        if( !m_pkVertices )
        {
                m_pkVertices = Core::Get()->GetRenderDevice()->CreateVertexBuffer( Buffer::STATIC, ( m_iSize + 1 ) * ( m_iSize + 1 ), &NormalDiffuseTexVertex::s_kDecl );

//                if( dynamic_cast< TerrainBlendMaterialFactory* >( m_pkMaterialFactory )->GetLighting() == TerrainBlendMaterialFactory::LIGHT_HARDWARE )
                {
                        float fHScale = m_pkMipPage->GetHeightmap()->GetHScale();

                        // Setup the vertex buffer
                        m_pkVertices->Lock( Buffer::WRITE );

                        for( int y = 0; y <= m_iSize; y++ )
                        {
                                for( int x = 0; x <= m_iSize; x++ )
                                {
					int ix, iy;

					ix = m_iOffsetX * m_iSize + x;
					iy = m_iOffsetY * m_iSize + y;

					if( ix == m_pkMipPage->GetHeightmap()->GetSize() )
						ix--;
					if( iy == m_pkMipPage->GetHeightmap()->GetSize() )
						iy--;

                                        unsigned int iOffset = y * ( m_iSize + 1 ) + x;
                                        float fHeight = m_pkMipPage->GetHeightmap()->GetHeight( ix, iy );  // Height is already transformed by VScale

                                        NormalDiffuseTexVertex *pkVertex = (NormalDiffuseTexVertex*) m_pkVertices->GetVertex( iOffset );
                                        pkVertex->m_kPosition.Set( ix * fHScale, fHeight, iy * fHScale );
                                }
                        }

                        m_pkVertices->Unlock();

			m_pkMaterialFactory->SetTexCoords( m_pkVertices, m_pkMipPage );
                }
        }

        // Calulate number of indices per LOD and create all the memory
        for( int lod = 0; lod < iMaxLOD; lod++ )
        {
		MipData *pkData;

		iIncrement = ( lod == 0 ) ? 1 : 2 << ( lod - 1 );

                int size;
                if( lod == 0 )
                {
                        size = m_iSize;
                }
                else
                {
			size = m_iSize >> lod;
                }

		pkData = new MipData( this, lod, iIncrement, size );

		m_vpkData.push_back( pkData );

//		if( dynamic_cast< TerrainBlendMaterialFactory* >( m_pkMaterialFactory )->GetLighting() == TerrainBlendMaterialFactory::LIGHT_HARDWARE )
		{
			pkData->m_pkPolygons->CalculateNormals( m_pkVertices, true );

			m_pkVertices->Lock( Buffer::WRITE );
			pkData->m_pkPolygons->Lock( Buffer::READ );

			Vector3d *pkNormal = pkData->m_pkPolygons->GetNormals();
			const VertexElement *pkNormalElement = m_pkVertices->GetVertexFormat()->GetElement( VertexElement::FLOAT3, VertexElement::NORMAL );
			unsigned char *pucVertex = (unsigned char*) m_pkVertices->GetVertex() + pkNormalElement->m_uiOffset;
			unsigned int uiVertexSize = m_pkVertices->GetVertexSize();

			Polygon *pkPolygon = pkData->m_pkPolygons->GetPolygon( 0 );

			int iPolyEnd = pkData->m_pkPolygons->GetNumElements();
			int iEnd     = m_pkVertices->GetNumElements();
			int j;

			for( j = 0; j < iPolyEnd; ++j, ++pkPolygon, ++pkNormal )
			{
				*((Vector3d*)( pucVertex + uiVertexSize * pkPolygon->v[0] )) += *pkNormal;
				*((Vector3d*)( pucVertex + uiVertexSize * pkPolygon->v[1] )) += *pkNormal;
				*((Vector3d*)( pucVertex + uiVertexSize * pkPolygon->v[2] )) += *pkNormal;
			}

			pucVertex = (unsigned char*) m_pkVertices->GetVertex() + pkNormalElement->m_uiOffset;

			for( j = 0; j < iEnd; ++j, pucVertex += uiVertexSize )
				((Vector3d*)pucVertex)->Normalize();

			pkData->m_pkPolygons->Unlock();
			m_pkVertices->Unlock();
		}
        }

	PrecomputeDistances();

	m_iCurrentLOD = 0;
}


NeoEngine::VertexBufferPtr &MipTerrainBlock::GetVertexBuffer()
{
	return m_pkVertices;
}


NeoEngine::PolygonBufferPtr &MipTerrainBlock::GetPolygonBuffer()
{
	return m_vpkData[ m_iCurrentLOD ]->m_pkPolygons;
}


void MipTerrainBlock::GenerateAABB()
{
	neolog << "MipTerrainBlock::GenerateAABB()" << std::endl;

	AABB *pkAABB = new AABB;

	pkAABB->Generate( m_pkVertices );

	SetBoundingVolume( pkAABB );
}


void MipTerrainBlock::CalculateLOD()
{
	Vector3d v;
	float    fDistSqr;
	int iSize = (int)m_vpkData.size();
	int tmp = m_iCurrentLOD;

	// ### FIXME - use the center of the block, not the world translation
	Vector3d kTrans = m_pkNode->GetWorldTranslation();

	//	neolog << "MipTerrainBlock::CalculateLOD(): World Translation == " << kTrans << endl;

	v = NeoEngine::Camera::GetActive()->GetWorldTranslation() - m_pkNode->GetWorldTranslation(); // - m_kCenter;

	fDistSqr = v.Len2();

	m_iCurrentLOD = 0;

	while( fDistSqr > m_pfDistanceTable[ m_iCurrentLOD ] && m_iCurrentLOD < iSize - 1 )
	{
	  //neolog << "CalculateLOD(): fDistSqr == " << fDistSqr << ", DistanceTable[x] == " << m_pfDistanceTable[ m_iCurrentLOD ] << endl;
		m_iCurrentLOD++;
	}

#if 0
	if( m_iCurrentLOD != tmp )
	{
		neolog << "Switching to LOD " << m_iCurrentLOD << " from " << tmp << endl;
	}
#endif
}


void MipTerrainBlock::UpdateData()
{
        CalculateLOD();
}


MipTerrain::MipTerrain( TerrainHeightmap *pkHeightmap, TerrainMaterialFactory *pkMaterialFactory, int iDepth ) :
        TerrainPage( pkHeightmap, pkMaterialFactory, iDepth ),
	m_fMaxError( 0.015f ),
	m_bCallbackInstalled ( false )
{
	m_pkMaterialFactory->GenerateMaterial( this );

	int iBlocks = GetNumBlocksOnSide();

	for( int iy = 0; iy < iBlocks; iy++ )
	{
		for( int ix = 0; ix < iBlocks; ix++ )
		{
			m_vpkBlocks.push_back( new MipTerrainBlock( ix, iy, m_iBlockSize, iy * iBlocks + ix, pkMaterialFactory, this ) );
		}
	}
}


void MipTerrainBlock::PrecomputeDistances()
{
	int iMaxLOD = dynamic_cast< MipTerrainManager* >( Core::Get()->GetTerrainManager() )->GetMaxDetailLevels();

	TerrainHeightmap *pkHeightmap = m_pkMipPage->GetHeightmap();
 
	for( int iLOD = 0; iLOD < iMaxLOD - 1; iLOD++ )
	{
		float fReal = 0.0f;
		float fLerp = 0.0f;
		float fMax  = 0.0f;

		int iIncrement = m_vpkData[ iLOD ]->m_iIncrement;
 
		int rx = m_iOffsetX * ( GetSize() - 1 );
		int ry = m_iOffsetY * ( GetSize() - 1 );
		int sx = ( m_iOffsetX + 1 ) * ( GetSize() - 1 );
		int sy = ( m_iOffsetY + 1 ) * ( GetSize() - 1 );

		for( int y = ry + iIncrement; y < sy; y += ( iIncrement * 2 ) )
		{
			for( int x = rx; x < ( sx - iIncrement ); x += ( iIncrement * 2 ) )
			{
				fReal = pkHeightmap->GetHeight( x, y );
				fLerp = ( pkHeightmap->GetHeight( x, y - iIncrement ) + pkHeightmap->GetHeight( x, y + iIncrement ) ) / 2.0f;
 
				if( fabs( fReal - fLerp ) > fMax )
					fMax = fabsf( fReal - fLerp );
 
				fReal = pkHeightmap->GetHeight( x + iIncrement, y );
				fLerp = ( pkHeightmap->GetHeight( x, y + iIncrement ) + pkHeightmap->GetHeight( x + ( iIncrement * 2 ), y - iIncrement ) ) / 2.0f;

				if( fabs( fReal - fLerp) > fMax )
					fMax = fabsf( fReal - fLerp );
			}
		}

		for( int y1 = ry; y1 < sy; y1 += ( iIncrement * 2 ) )
		{
			for( int x1 = ( rx + iIncrement ); x1 < ( sx - iIncrement ); x1 += ( iIncrement * 2 ) )
			{
				fReal = pkHeightmap->GetHeight( x1, y1 );
				fLerp = ( pkHeightmap->GetHeight( x1 - iIncrement, y1 ) + pkHeightmap->GetHeight( x1 + iIncrement, y1 ) ) / 2.0f;

				if( fabs( fReal - fLerp ) > fMax )
					fMax = fabsf( fReal - fLerp );
			}
		}
 
		// Compute the distance of this error
		float d = ( 600.0f * fMax ) / 3.0f;    // Divide by max screen error.. make this adjustable?
		m_pfDistanceTable[ iLOD ] = d * d;

		neolog << "For LOD " << iLOD << ", precomputed distance == " << m_pfDistanceTable[ iLOD ] << endl;
	}
}


bool MipTerrainBlock::Render( Frustum *pkFrustum, bool bForce )
{
	if( !SceneEntity::Render( pkFrustum, bForce ) )
		return false;

	static RenderPrimitive  skPrimitive;
	static RenderPrimitive *spkPrimitive = &skPrimitive;

	NeoEngine::PolygonBufferPtr pkPolygons = GetPolygonBuffer();

	if( m_pkNode )
	{
		spkPrimitive->m_kModelMatrix    = m_pkNode->GetWorldTransform();
		spkPrimitive->m_kInvModelMatrix = m_pkNode->GetInverseWorldTransform();
	}

	if( !pkPolygons->IsStripped() )
	{
		spkPrimitive->m_ePrimitive      = RenderPrimitive::TRIANGLES;
		spkPrimitive->m_pkPolygonBuffer = pkPolygons;
		spkPrimitive->m_uiNumPrimitives = spkPrimitive->m_pkPolygonBuffer->GetNumElements();
	}
	else
	{
		spkPrimitive->m_ePrimitive           = RenderPrimitive::TRIANGLESTRIP;
		spkPrimitive->m_pkPolygonStripBuffer = pkPolygons->GetStrip();
		spkPrimitive->m_uiNumPrimitives      = spkPrimitive->m_pkPolygonStripBuffer->GetNumElements();
	}

	spkPrimitive->m_pkVertexBuffer       = m_pkVertices;
	spkPrimitive->m_pkMaterial           = m_pkMaterialFactory->m_pkMaterial;

	Core::Get()->GetRenderDevice()->Render( *spkPrimitive, 0 );

	spkPrimitive->m_pkMaterial           = 0;
	spkPrimitive->m_pkVertexBuffer       = 0;
	spkPrimitive->m_pkPolygonStripBuffer = 0;

	return true;
}


TerrainPage *MipTerrainManager::CreateTerrain( TerrainHeightmap *pkHeightmap, TerrainMaterialFactory *pkMaterialFactory )
{
        MipTerrain *pkMip = new MipTerrain( pkHeightmap, pkMaterialFactory, 2 );

	NeoEngine::Camera *pkCamera = NeoEngine::Camera::GetActive();

	if( !pkCamera )
	{
	        neolog << LogLevel( ERROR ) << "ERROR!!  It appears that MipTerrainManager::CreateTerrain() was called before" << endl
		       << "a camera was created.  The camera object MUST be created before a MipTerrain may be created!!" << endl;
	}

	MipCameraCallback *pkCallback = new MipCameraCallback();

	//	pkCallback->SetUserData( pkMip );

	pkCamera->AddCallback( pkCallback );

	return pkMip;
}


}; // namespace NeoGeoMip
