/***************************************************************************
                         neoicdds.cpp  -  description
                             -------------------
    begin                : Fri Apr 09 2004
    copyright            : (C) 2004 by Patrick Levin
    email                : p.levin@numlock.de
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoICDDS, neoicdds.cpp

 The Initial Developer of the Original Code is Patrick Levin.
 Portions created by Patrick Levin are Copyright (C) 2004
 Numlock Software GmbH. All Rights Reserved.

 ***************************************************************************/

#include "neoicdds.h"
#include "link.h"
#include "dds_defines.h"

#include <neoengine/file.h>
#include <neoengine/util.h>
#include <neoengine/logstream.h>
#include <string.h>

using namespace std;
using namespace NeoEngine;

#define CODECVERSION_MAJOR		0
#define CODECVERSION_MINOR		1
#define CODECVERSION_REVISION	0

// DDS cubemap surfaces must include all faces. This macro combines the according flags.
#define DDS_CUBEMAP_FLAGS ( \
	DDSCAPS2_CUBEMAP_POSITIVEX | DDSCAPS2_CUBEMAP_NEGATIVEX | \
	DDSCAPS2_CUBEMAP_POSITIVEY | DDSCAPS2_CUBEMAP_NEGATIVEY | \
	DDSCAPS2_CUBEMAP_POSITIVEZ | DDSCAPS2_CUBEMAP_NEGATIVEZ )

namespace NeoDDS
{

#ifdef BUILD_STATIC
#  define DDSEXPORT

ModuleStatic *g_pkStaticModule = 0;

#else

#  ifdef WIN32
#    define DDSEXPORT __declspec(dllexport)
#  else
#    define DDSEXPORT
#endif

extern "C"
{

#endif

	DDSEXPORT ImageCodec *LoadImageCodec() {
		
		vector<string> vstrExtensions;

		vstrExtensions.push_back( "dds" );
		vstrExtensions.push_back( "DDS" );

		ImageCodec *pkCodec = new NeoDDS::ImageCodec( vstrExtensions );

		return pkCodec;
	}

	DDSEXPORT int Initialize() {
		return 0;
	}

	DDSEXPORT int Shutdown()
	{
#ifdef BUILD_STATIC
		delete g_pkStaticModule;
#endif
		return 0;
	}

	DDSEXPORT void GetVersion( std::string *pstrName, int *piMajor, int *piMinor, int *piRevision ) {
		*pstrName   = "NeoICDDS";
		*piMajor    = CODECVERSION_MAJOR;
		*piMinor    = CODECVERSION_MINOR;
		*piRevision = CODECVERSION_REVISION;
	}

#ifdef BUILD_STATIC

#else

}

#endif

/*!	\namespace Private	Local namespace that contains internal classes.
**/
namespace Private
{	

	/// Cobine color vector from components.
	static unsigned int inline CombineColor ( const int r, const int g, const int b, const int a = 255 )
	{
		return r | ( g << 8 ) | ( b << 16 ) | ( a << 24 );
	}

	/*! Internal DDS decoder base class
	 *
	 *	Each supported format is assigned a decoder class that
	 *	handles stream decoding and must be derived from this
	 *	class.
	**/
	class DDSImageDecoder {
	
	protected:

		ImageData	*m_pkImage; //!< destination image

		/*! Decode one image line
		 *
		 *	\param	pkFile	File stream to decode from.
		 *	\param	nLine	Number of line.
		 *	\param	nPixels	Number of pixels to be decoded.
		**/
		virtual bool DecodeLine( File *pkFile, const int nLine, const int nPixels ) {
			return false;
		}
		
	public:

		/*! Constructor
		 *
		 *	\param	pkImage	Output image object.
		**/
		inline DDSImageDecoder( ImageData *pkImage ) : m_pkImage( pkImage )
		{
		}

		/// Destructor
		virtual inline ~DDSImageDecoder() {}

		/// Set another image.
		void inline SetImage( ImageData *pkImage )
		{
			m_pkImage = pkImage;
		}

		/*! Return image size for decoded format
		 *
		 *	\param	nWidth	Image width.
		 *	\param	nHeight	Image height.
		 *	\return	Image size in bytes.
		**/
		virtual unsigned int GetImageSize( const int nWidth, const int nHeight ) const = 0;

		/*! Initialize destination image to correct format
		 *
		 *	\param	nWidth		Image width.
		 *	\param	nHeight		Image height.
		 *	\param	rkFormat	Pixelformat to prepare from.
		**/
		virtual bool	PrepareOutput( int nWidth, int nHeight, DDPIXELFORMAT &rkFormat ) = 0;

		/*! Decode input stream
		 *
		 *	\param	pkFile	
		 *	\param	nWidth	source width
		 *	\param	nHeight	source height
		**/
		virtual bool inline Decode( File *pkFile, const int nWidth, const int nHeight ) {

			for( int nLine = 0; nLine < nHeight; ++nLine )
			{
				// decode each line
				if( !DecodeLine( pkFile, nLine, nWidth ) )
				{
					// failed to read from stream
					return false;
				}
			}

			return true;
		}
	};

	/////////////////////////////////////////////////////////////////////////////
	//		   			DECODERS FOR UNCOMPRESSED DDS FILES					   //
	/////////////////////////////////////////////////////////////////////////////

	/*! Base class for handling uncompressed DDS files.
	 *
	 *	Handles basic DDS converting.
	**/
	class DDSUncompressed : public DDSImageDecoder
	{
	protected:

		unsigned int	m_MaskR;	//!< Bit mask for red color component extraction.
		unsigned int	m_MaskG;	//!< Bit mask for green color component extraction.
		unsigned int	m_MaskB;	//!< Bit mask for blue color component extraction.
		unsigned int	m_MaskA;	//!< Bit mask for alpha color component extraction.

		unsigned int	m_ShiftR;	//!< Bit shift for red color component extraction.
		unsigned int	m_ShiftG;	//!< Bit shift for red color component extraction.
		unsigned int	m_ShiftB;	//!< Bit shift for red color component extraction.
		unsigned int	m_ShiftA;	//!< Bit shift for red color component extraction.

		unsigned int	m_UpShiftR;	//!< Bit shift for red color component scaling.
		unsigned int	m_UpShiftG;	//!< Bit shift for green color component scaling.
		unsigned int	m_UpShiftB;	//!< Bit shift for blue color component scaling.
		unsigned int	m_UpShiftA;	//!< Bit shift for alpha color component scaling.

		int		m_PixelSize;		//!< Size of a single pixel in bytes.
		void*	m_LineBuffer;		//!< Temporary array.

		/// Find bit shift
		void GetBitShift (const unsigned int nValue, unsigned int &nShift) const {
			
			unsigned int nLow	= 0;
			unsigned int nMaskLow  = 1;
			
			// note: could be implemented with a single x86 assembly command
			//		 C is used to keep it portable
			while( nValue != 0 && !( nValue & nMaskLow ) )
			{
				nMaskLow <<= 1;
				++nLow;
			}

			nShift = nLow;
		}

		/// Find component size
		void GetUpShift (const unsigned int nValue, const unsigned int nDown, unsigned int &nShift) const {
			
			unsigned int nHigh	= 31;			// maximum shift value: MSB for 32 bit integers
			unsigned int nMask = 1 << nHigh;

			// note: could be implemented with a single x86 assembly command
			//		 C is used to keep it portable
			while( nHigh && !( nValue & nMask ) )
			{
				nMask >>= 1;
				--nHigh;
			}

			nShift = nHigh ? 8 - ( nHigh - nDown + 1 ) : 0;
		}

		/// Convert a line of pixels
		virtual void ConvertLine( const int nLine, const int nPixels ) = 0;

		// read and convert uncompressed pixels
		virtual bool DecodeLine( File *pkFile, const int nLine, const int nPixels )
		{

			pkFile->Read( m_LineBuffer, nPixels*m_PixelSize );

			ConvertLine(nLine, nPixels);

			return true;
		}

		/// Get shift values
		void inline GetShiftValues()
		{
			GetBitShift( m_MaskR, m_ShiftR );
			GetBitShift( m_MaskG, m_ShiftG );
			GetBitShift( m_MaskB, m_ShiftB );
			GetBitShift( m_MaskA, m_ShiftA );
			GetUpShift( m_MaskR, m_ShiftR, m_UpShiftR );
			GetUpShift( m_MaskG, m_ShiftG, m_UpShiftG );
			GetUpShift( m_MaskB, m_ShiftB, m_UpShiftB );
			GetUpShift( m_MaskA, m_ShiftA, m_UpShiftA );
		}

	public:

		/// Constructor
		inline DDSUncompressed( ImageData *pkImage ) : DDSImageDecoder( pkImage ), m_LineBuffer( 0 )
		{
			memset( &m_MaskR, 0, 12 * sizeof( unsigned int ) );
		}

		/// Deconstructor
		/// Frees line buffer.
		virtual inline ~DDSUncompressed()
		{
			delete [] m_LineBuffer;
		}

		/*! Prepare output data
		**/
		bool	PrepareOutput( int nWidth, int nHeight, DDPIXELFORMAT &rkFormat )
		{
			
			// assign pixel masks

			m_MaskR = rkFormat.dwRBitMask;
			m_MaskG = rkFormat.dwGBitMask;
			m_MaskB = rkFormat.dwBBitMask;
			m_MaskA = rkFormat.dwRGBAlphaBitMask;

			GetShiftValues();

			m_PixelSize = ( rkFormat.dwRGBBitCount + 7 ) >> 3;

			m_LineBuffer = new unsigned char[ m_PixelSize * nWidth ];

			m_pkImage->m_iWidth = nWidth;
			m_pkImage->m_iHeight= nHeight;			
			
			if( rkFormat.dwFlags & DDPF_RGB )
			{
				// 24 or 32 bit RGB image
				if( m_MaskA || ( rkFormat.dwFlags & DDPF_ALPHAPIXELS ) || ( rkFormat.dwRGBBitCount > 24 ) )
				{
					m_pkImage->m_iChannels = 4;
					m_pkImage->m_eFormat = Texture::RGBA;
				}
				else
				{
					m_pkImage->m_iChannels = 3;
					m_pkImage->m_eFormat = Texture::RGB;
				}
			}
			else
			{
				// alpha channel only
				m_pkImage->m_iChannels = 1;
				m_pkImage->m_eFormat = Texture::ALPHA;
			}
			
			m_pkImage->m_iBPP	 = m_pkImage->m_iChannels * 8;

			m_pkImage->m_pucData = new unsigned char[ m_pkImage->m_iChannels * nWidth * nHeight ];

			return true;
		}
	};

	/*! Class that decodes uncompressed 8 bit DDS files.
	**/
	class DDSUncompressed8 : public DDSUncompressed
	{
	protected:

		/// Extract color components and compose to match library format.
		void ConvertLine( const int nLine, const int nPixels ) 
		{
			unsigned char *pucLine = static_cast< unsigned char* >( m_LineBuffer );

			unsigned char *pucData = m_pkImage->m_pucData + nLine * m_pkImage->m_iChannels * 
				m_pkImage->m_iWidth;
			
			for ( int nPixel = 0, nOffset = 0; nPixel < nPixels; ++nPixel, nOffset += 3 )
			{
				const unsigned short nValue = pucLine[ nPixel ];
				
				const int r = ( nValue & m_MaskR ) >> m_ShiftR;
				const int g = ( nValue & m_MaskG ) >> m_ShiftG;
				const int b = ( nValue & m_MaskB ) >> m_ShiftB;

				pucData[nOffset + 0] = r << m_UpShiftR;
				pucData[nOffset + 1] = g << m_UpShiftG;
				pucData[nOffset + 2] = b << m_UpShiftB;				
			}
		}
	
	public:

		/// Constructor
		inline DDSUncompressed8( ImageData *pkImage) : DDSUncompressed( pkImage )
		{
		}

		/// Return image size for 8 bit images
		unsigned int inline GetImageSize( const int nWidth, const int nHeight ) const
		{
			return nWidth * nHeight;
		}
	};

	/*! Class that decodes uncompressed 15 and 16 bit DDS files.
	**/
	class DDSUncompressed16 : public DDSUncompressed
	{
	protected:

		/// Extract color components and compose to match library format.
		void ConvertLine( const int nLine, const int nPixels ) 
		{
			unsigned short *pucLine = static_cast< unsigned short* >( m_LineBuffer );

			unsigned char *pucData = m_pkImage->m_pucData + nLine * m_pkImage->m_iChannels * 
				m_pkImage->m_iWidth;
			
			for ( int nPixel = 0, nOffset = 0; nPixel < nPixels; ++nPixel, nOffset += 3 )
			{
				const unsigned short nValue = pucLine[ nPixel ];
				
				const int r = ( nValue & m_MaskR ) >> m_ShiftR;
				const int g = ( nValue & m_MaskG ) >> m_ShiftG;
				const int b = ( nValue & m_MaskB ) >> m_ShiftB;

				pucData[nOffset + 0] = r << m_UpShiftR;
				pucData[nOffset + 1] = g << m_UpShiftG;
				pucData[nOffset + 2] = b << m_UpShiftB;
				
				if ( m_MaskA != 0 )
				{
					if( m_UpShiftA < 7 )
						pucData[nOffset + 3] = ( ( nValue & m_MaskA ) >> m_ShiftA ) << m_UpShiftA;
					else
						pucData[nOffset + 3] = ( nValue & m_MaskA ) ? 255 : 0;

					++nOffset;
				}
			}
		}
	
	public:

		/// Constructor
		inline DDSUncompressed16( ImageData *pkImage) : DDSUncompressed( pkImage )
		{
		}

		/// Return image size for 15 and 16 bit images
		unsigned int inline GetImageSize( const int nWidth, const int nHeight ) const
		{
			return nWidth * nHeight * 2;
		}
	};

	/*! Class that decodes uncompressed 24 bit DDS files.
	**/
	class DDSUncompressed24 : public DDSUncompressed
	{
	protected:

		/// Extract color components and compose to match library format
		void ConvertLine (const int nLine, const int nPixels)
		{
			unsigned char *pucLine = static_cast< unsigned char* >( m_LineBuffer );

			unsigned char *pucData = m_pkImage->m_pucData + nLine * m_pkImage->m_iChannels *
				m_pkImage->m_iWidth;

			for( int nPixel = 0, nOffset = 0; nPixel < nPixels; ++nPixel, nOffset += 3 )
			{
				unsigned int nValue = *pucLine++;
				nValue |= *pucLine++ << 8;
				nValue |= *pucLine++ << 16;
				
				const int r = ( nValue & m_MaskR ) >> m_ShiftR;
				const int g = ( nValue & m_MaskG ) >> m_ShiftG;
				const int b = ( nValue & m_MaskB ) >> m_ShiftB;

				pucData[nOffset + 0] = r << m_UpShiftR;
				pucData[nOffset + 1] = g << m_UpShiftG;
				pucData[nOffset + 2] = b << m_UpShiftB;
			}
		}
	
	public:

		/// Constructor
		DDSUncompressed24( ImageData *pkImage ) : DDSUncompressed( pkImage )
		{
		}

		/// Return image size for 24 bit images
		unsigned int inline GetImageSize( const int nWidth, const int nHeight ) const
		{
			return nWidth * nHeight * 3;
		}
	};


	/*! This class decodes uncompressed 32 bit DDS files.
	**/
	class DDSUncompressed32 : public DDSUncompressed
	{
	protected:

		/// Extract color components and compose to match library format
		void ConvertLine( const int nLine, const int nPixels )
		{
			unsigned int *puiLine = static_cast< unsigned int* >( m_LineBuffer );

			unsigned int *puiData = reinterpret_cast< unsigned int* > ( m_pkImage->m_pucData )
				+ nLine * m_pkImage->m_iWidth;

			for( int nPixel = 0; nPixel < nPixels; ++nPixel )
			{
				unsigned int nValue = puiLine[nPixel];
				
				const int r = ( nValue & m_MaskR ) >> m_ShiftR;
				const int g = ( nValue & m_MaskG ) >> m_ShiftG;
				const int b = ( nValue & m_MaskB ) >> m_ShiftB;

				int nAlpha = (m_MaskA != 0) ? (nValue & m_MaskA) >> m_ShiftA : 255;
				nAlpha <<= m_UpShiftA;

				puiData[nPixel] = CombineColor( r << m_UpShiftR, g << m_UpShiftG, b << m_UpShiftB, nAlpha );
			}
		}
	
	public:

		/// Constructor
		DDSUncompressed32( ImageData* pkImage ) : DDSUncompressed( pkImage )
		{
		}

		/// Return image size for 32 bit images
		unsigned int inline GetImageSize( const int nWidth, const int nHeight ) const
		{
			return nWidth * nHeight * 4;
		}
	};

	/*! This class decodes uncompressed alpha channel-only DDS files.
	**/
	class DDSUncompressedAlpha : public DDSUncompressed
	{
	protected:

		/// Set alpha channel
		void ConvertLine ( const int nLine, const int nPixels )
		{
			unsigned char *pucLine = static_cast< unsigned char* >( m_LineBuffer );

			unsigned char *pucData = m_pkImage->m_pucData + nLine * m_pkImage->m_iWidth;

			memcpy(pucData, pucLine, nPixels);
		}
	
	public:

		/// Constructor
		DDSUncompressedAlpha( ImageData* pkImage ) : DDSUncompressed( pkImage )
		{
		}

		/// Return image size for alpha-only images
		unsigned int inline GetImageSize( const int nWidth, const int nHeight ) const
		{
			return nWidth * nHeight;
		}
	};

	/////////////////////////////////////////////////////////////////////////////
	//		   				COMPRESSED DDS FILE DECODERS					   //
	/////////////////////////////////////////////////////////////////////////////

	/*! S3TC decoder base class,
	**/
	class DXTDecoder : public DDSImageDecoder
	{
	protected:

		/// S3TC RGB color block
#ifdef WIN32
#	pragma pack(push)
#	pragma pack(2)
#endif
		struct NEOENGINE_ATTRIBUTE_PACKED DXTColorBlock {
			unsigned short b : 5;
			unsigned short g : 6;
			unsigned short r : 5;
		};
#ifdef WIN32
#pragma pack(pop)
#endif

		/// Line buffering
		virtual bool ReadLine( File *pkFile, const int nBlocks ) = 0;

		/// Decode DXT block
		virtual void DecodeBlock( const int nBlockX, const int nBlockY ) = 0;

	public:

		/// Constructor
		inline DXTDecoder( ImageData *pkImage) : DDSImageDecoder( pkImage )
		{
		}

		/// Destructor
		virtual inline ~DXTDecoder()
		{
		}

		/// Return image size for DXT1 textures.
		unsigned int inline	GetImageSize( const int nWidth, const int nHeight ) const
		{
			// formula as from the DirectX(tm) SDK
			return std::max( 1, nWidth/4 ) * std::max( 1, nHeight/4 ) * 8;
		}

		/// Decode S3TC compressed blocks
		bool inline Decode( File *pkFile, const int nWidth, const int nHeight )
		{
			for( int nBlockY = 0, nBlocksY = nHeight/4, nBlocksX = nWidth/4; nBlockY < nBlocksY; ++nBlockY )
			{
				if( !ReadLine( pkFile, nBlocksX ) )
				{
					// failed to read blocks
					return false;
				}

				for( int nBlockX = 0; nBlockX < nBlocksX; ++nBlockX )
				{
					DecodeBlock( nBlockX, nBlockY );
				}
			}

			return true;
		}

	};


	/*! DXT1 decoder base class.
	**/
	class DXT1Decoder : public DXTDecoder
	{
	protected:

		/// DXT1 Block structure
#ifdef WIN32
#	pragma pack(push)
#	pragma pack(4)
#endif

		struct NEOENGINE_ATTRIBUTE_PACKED DXT1Block
		{
			union {
				DXTColorBlock	c0;		//!< DXT color 1
				unsigned short	v0;
			};
			union {
				DXTColorBlock	c1;		//!< DXT color 2
				unsigned short	v1;
			};
			unsigned char		code[4];//!< code bits
		};

#ifdef WIN32
#	pragma pack(pop)
#endif

		/// pre-calculated code masks
		static const unsigned int CODEMASK[4];
		static const unsigned int CODESHIFT[4];

		DXT1Block	*m_LineBuffer;

		/// Line buffering
		virtual bool ReadLine ( File *pkFile, const int nBlocks )
		{
			pkFile->Read( m_LineBuffer, nBlocks*8 );
			return true;
		}

		/// Decode color block (e.g. extract block palette)
		void DecodeColor ( const DXT1Block &rBlock, unsigned int *puiColor, bool bSetOpaque = true ) const 
		{
			int r0, g0, b0, r1, g1, b1;

			const int alpha = bSetOpaque ? 255 : 0;

			// expand block base colors (two 16 bit color vectors)
			r0 = rBlock.c0.r << 3; r1 = rBlock.c1.r << 3;
			g0 = rBlock.c0.g << 2; g1 = rBlock.c1.g << 2;
			b0 = rBlock.c0.b << 3; b1 = rBlock.c1.b << 3;

			// this is a little macro that implements the S3TC color interpolation ramp
			#define	DECODE( i, j ) ( ( ( 2 * ( i ) ) + ( j ) ) / 3 )

			puiColor[0] = CombineColor( r0, g0, b0, alpha );
			puiColor[1] = CombineColor( r1, g1, b1, alpha );
			
			int rx, gx, bx;

			// extrapolate packed colors

			rx = DECODE( r0, r1 );
			gx = DECODE( g0, g1 );
			bx = DECODE( b0, b1 );

			puiColor[2] = CombineColor( rx, gx, bx, alpha );

			rx = DECODE( r1, r0 );
			gx = DECODE( g1, g0 );
			bx = DECODE( b1, b0 );

			puiColor[3] = CombineColor( rx, gx, bx, alpha );

			// fourth color vector is interpolated from the two first (explicit) color vectors
			puiColor[4] = CombineColor( ( r0 + r1 ) >> 1, ( g0 + g1 ) >> 1, ( b0 + b1 ) >> 1, alpha );

			#undef DECODE
		}

		/// Decode DXT1 block. Each block encodes 16 pixels (4 x 4)
		virtual void DecodeBlock( const int nBlockX, const int nBlockY )
		{
			const DXT1Block &rBlock = m_LineBuffer[nBlockX];

			// each DXT1 block contains an encoded 5 color palette.
			static unsigned int auiColor[5];

			DecodeColor( rBlock, auiColor );

			// I use index mapping to avoid an if-clause inside the decoding loop.
			// Color codes for opaque blocks map directly to color indices.
			int aiIndex[4] = { 0, 1, 2, 3 };

			if( rBlock.v0 < rBlock.v1 )
			{
				// block contains transparent pixels - let code "2" map to color 4 and make color 3
				// fully transparent.
				aiIndex[2] = 4;
				auiColor[3]= 0;
			}

			unsigned int *puiData = reinterpret_cast< unsigned int* >
				( m_pkImage->m_pucData ) + nBlockY * 4 * m_pkImage->m_iWidth + nBlockX * 4;

			for (int y = 0; y < 4; ++y, puiData += m_pkImage->m_iWidth)
			{				
				for (int x = 0; x < 4; ++x)
				{
					// find color index for this pixel
					int nCode = rBlock.code[y] & CODEMASK[x];
					nCode >>= CODESHIFT[x];

					// unmap color vector
					puiData[x] = auiColor[ aiIndex[ nCode ] ];
				}
			}
		}

	public:

		/// Constructor
		inline DXT1Decoder( ImageData *pkImage) : DXTDecoder( pkImage ), m_LineBuffer( 0 )
		{
		}

		/// Destructor
		virtual inline ~DXT1Decoder()
		{			
			delete [] m_LineBuffer;
		}

		/// Prepare output image
		bool PrepareOutput( int nWidth, int nHeight, DDPIXELFORMAT & )
		{
			// allocate line buffer
			m_LineBuffer = new DXT1Block[ nWidth / 4 ];
			
			// always create 32 bit image (as long as there is no direct support within NeoEngine)

			m_pkImage->m_iWidth		= nWidth;
			m_pkImage->m_iHeight	= nHeight;

			m_pkImage->m_eFormat	= Texture::RGBA;
			m_pkImage->m_iChannels	= 4;
			m_pkImage->m_iBPP		= 32;
			m_pkImage->m_pucData	= new unsigned char[ nWidth * nHeight * 4 ];

			return true;
		}
	};


	/*! DXT3 decoder base class. DXT3 is the same as DXT1, except that an explicit 4 bit alpha
	 *	channel is present.
	**/
	class DXT3Decoder : public DXT1Decoder {
	
	protected:

		/// Explicit alpha values
		struct DXT3AlphaBlock {
			unsigned char	nAlpha[8];	//!< uncompressed alpha bits
		};

		/// Line buffering
		bool inline ReadLine( File *pkFile, const int nBlocks )
		{
			// two blocks of 64 bits each are read
			pkFile->Read( m_LineBuffer, nBlocks * 16 );
			return true;
		}

		/// Decode alpha value.
		virtual unsigned int DecodeAlpha( const int x, const int y, const DXT1Block &rBlock ) const
		{
			// cast to alpha block
			const DXT3AlphaBlock &rAlpha = reinterpret_cast< const DXT3AlphaBlock& >( rBlock );

			// get index
			const int nIndex = (y * 4 + x) / 2;
			unsigned int nValue = rAlpha.nAlpha[nIndex];

			// switch between upper and lower nibble
			if (x & 1)
			{
				nValue &= 0xf0;
			}
			else
			{
				nValue &= 0x0f;
				nValue <<= 4;
			}

			return nValue;
		}
	
		/// Decode DXT3 block
		virtual void DecodeBlock(const int nBlockX, const int nBlockY) {

			// grab alpha information
			const DXT1Block &rAlpha = m_LineBuffer[ nBlockX * 2 ];
			// RGB data is preceeded by explicit alpha values
			const DXT1Block &rBlock = m_LineBuffer[ nBlockX * 2 + 1];

			// get image data offset
			unsigned int *puiData = reinterpret_cast< unsigned int* > ( m_pkImage->m_pucData )
				+ nBlockY * 4 * m_pkImage->m_iWidth + nBlockX * 4;

			unsigned int auiColor[5];

			// decode palette, set alpha values to zero 
			DecodeColor( rBlock, auiColor, false );
				
			for( int y = 0; y < 4; ++y, puiData += m_pkImage->m_iWidth )
			{
				for( int x = 0; x < 4; ++x )
				{
					int nCode = rBlock.code[y] & CODEMASK[x];
					nCode >>= CODESHIFT[x];

					// decode alpha bits
					unsigned int nAlpha = DecodeAlpha(x, y, rAlpha);

					// transparency codes are ignored, as alpha information is
					// directly available
					puiData[x] = auiColor[nCode] | (nAlpha << 24);
				}
			}
		}

	public:

		/// Constructor
		DXT3Decoder( ImageData* pkImage ) : DXT1Decoder( pkImage )
		{
		}

		/// Destructor - releases line buffer if valid
		virtual ~DXT3Decoder() {
		}

		/// Prepare output image
		bool PrepareOutput ( int nWidth, int nHeight, DDPIXELFORMAT & )
		{
			// allocate line buffer
			m_LineBuffer = new DXT1Block[ nWidth / 2 ];

			// always create 32 bit image (as long as there is no direct support within NeoEngine)

			m_pkImage->m_iWidth		= nWidth;
			m_pkImage->m_iHeight	= nHeight;

			m_pkImage->m_eFormat	= Texture::RGBA;
			m_pkImage->m_iChannels	= 4;
			m_pkImage->m_iBPP		= 32;
			m_pkImage->m_pucData	= new unsigned char[ nWidth * nHeight * 4 ];

			return true;
		}

		/// Return image size for DXT3 textures
		unsigned int inline  GetImageSize( const int nWidth, const int nHeight ) const
		{
			// formula as from the DirectX(tm) SDK
			return std::max( 1, nWidth / 4 ) * std::max( 1, nHeight / 4 ) * 16;
		}
	};


	/*! DXT5 decoder class.
	 *
	 *	Derived from DXT3Decoder as only alpha decoding differs from DXT3.
	 *	Alpha values are interpolated.
	**/
	class DXT5Decoder : public DXT3Decoder {
	protected:

		/// DXT5 Alpha code block
		struct DXT5AlphaBlock {
			unsigned char	nAlpha0;		//!< Explicit alpha
			unsigned char	nAlpha1;		//!< Explicit alpha
			unsigned char	nBits[6];		//!< Code bits (3 bits per code ->  6x8 = 48 bits = 16 codes)
		};

		/// Decode alpha interpolated value
		unsigned int DecodeAlpha( const int x, const int y, const DXT1Block &rBlock ) const
		{
			// cast to alpha block
			const DXT5AlphaBlock &rAlpha = reinterpret_cast< const DXT5AlphaBlock& >( rBlock );

			unsigned int nAlpha;
			int			 a0 = rAlpha.nAlpha0;
			int			 a1 = rAlpha.nAlpha1;
			int			 i = ( y * 4 + x) * 3;	// bit index
			int			 j = i / 8;				// byte index
			int			 nCode = rAlpha.nBits[j];
			
			i -= j * 8;		// get number of bits to shift
			nCode >>= i;	// shift to bit offset
			nCode &= 7;		// mask upper bits away
			i = 8 - i;		// find out how many bits are available in this byte

			if( i == 1 )
			{
				// need two more bits from next byte
				nCode |= ( rAlpha.nBits[ j + 1 ] & 3 ) << 1;
			}
			else if( i == 2 )
			{
				// need one more bit from next byte
				nCode |= ( rAlpha.nBits[ j + 1 ] & 1 ) << 2;
			} 
			else
			{
				// there are enough bits in this byte - code is valid
			}

			if( nCode == 0 )
			{
				nAlpha = a0;
			}
			else if( nCode == 1 )
			{
				nAlpha = a1;
			}
			else if( rAlpha.nAlpha0 > rAlpha.nAlpha1 )
			{
				// Use interpolation variant A)
				// instead of a switch statement, I put the pattern into a single formula,
				// which adds 2 more ops (minus and assignment) but keeps the code clean
				i = 8 - nCode;				
				nAlpha = ( a0 * i + a1 * ( 7 - i ) ) / 7;
			}
			else if ( nCode < 6 )
			{
				// Use interpolation variant B)
				i = 6 - nCode;
				nAlpha = ( a0 * i + a1 * ( 5 - i ) ) / 5;
			} else
			{
				// Either opaque or fully transparent
				nAlpha = ( nCode == 6 ) ? 0 : 255;
			}

			return nAlpha;
		}

	public:

		/// Constructor
		DXT5Decoder( ImageData *pkImage ) : DXT3Decoder( pkImage )
		{
		}
	};

	const unsigned int DXT1Decoder::CODEMASK[4]		= { 0x03, 0x0c, 0x30, 0xc0 };
	const unsigned int DXT1Decoder::CODESHIFT[4]	= { 0, 2, 4, 6 };

	/// Decoder factory: create decoder instance based on format.
	static DDSImageDecoder *CreateDecoder( ImageData *pkImage, const DDSURFACEDESC2 &rkSurfaceDesc)
	{
		if(0 == pkImage)
		{
			// no image
			return 0;
		}

		if( rkSurfaceDesc.ddpfPixelFormat.dwFlags & DDPF_FOURCC )
		{
			// compressed formats
			switch( rkSurfaceDesc.ddpfPixelFormat.dwFourCC )
			{
				case FOURCC_DXT1:
					return new DXT1Decoder( pkImage );
				case FOURCC_DXT3:
					return new DXT3Decoder( pkImage );
				case FOURCC_DXT5:
					return new DXT5Decoder( pkImage );
				default:
					return 0;
			}
		}
		else
		{
			// uncompressed formats
			if( rkSurfaceDesc.ddpfPixelFormat.dwFlags & DDPF_RGB )
			{
				switch( rkSurfaceDesc.ddpfPixelFormat.dwRGBBitCount )
				{
				case 8:
					return new DDSUncompressed8( pkImage );
				case 15:
				case 16:
					return new DDSUncompressed16( pkImage );
				case 24:
					return new DDSUncompressed24( pkImage );
				case 32:
					return new DDSUncompressed32( pkImage );
				default:
					return 0;
				}
			}
			else if( ( rkSurfaceDesc.ddpfPixelFormat.dwFlags & DDPF_ALPHA ) && 
				rkSurfaceDesc.ddpfPixelFormat.dwAlphaBitDepth == 8 ) 
			{
				// decode alpha channel
				return new DDSUncompressedAlpha( pkImage );
			}
			else
			{
				return 0;
			}
		}
	}
}

/*! Verify a DirectDraw(tm) sourface description from a DDS file header.
 *
 *	\param	pkSurfaceDesc	Surface description to be verified.
 *	\param	bLogError		If true, errors will be logged.
 *	\return <code>true</code>, if the header is valid.
**/
static bool inline VerifyDescription( const DDSURFACEDESC2 &rkSurfaceDesc, bool bLogError = true )
{
	if( rkSurfaceDesc.dwSize != sizeof( DDSURFACEDESC2 ) )
	{
		if( bLogError )
			neolog << LogLevel( ERROR ) << "*** Invalid surface descriptor: Size mismatch" << endl;
		// invalid surface descriptor size
		return false;
	}

	if( ( rkSurfaceDesc.dwFlags & DDSDEFAULTFLAGS ) != DDSDEFAULTFLAGS )
	{
		if( bLogError )
			neolog << LogLevel( ERROR ) << "*** Invalid surface descriptor: Mandatory flag(s) missing" << endl;
		// mandantory flags missing
		return false;
	}

	if( ( rkSurfaceDesc.ddsCaps.dwCaps & DDSCAPS_COMPLEX ) &&
		( rkSurfaceDesc.ddsCaps.dwCaps2 &  DDSCAPS2_CUBEMAP ) &&
		( rkSurfaceDesc.ddsCaps.dwCaps2 & DDS_CUBEMAP_FLAGS ) != DDS_CUBEMAP_FLAGS )
	{
		if( bLogError )
			neolog << LogLevel( ERROR ) << "*** Invalid cubemap surface: missing face(s) detected" << endl;
		return false;
	}

	// seems to be valid
	return true;
}

/*! Verify a DirectDraw(tm) pixel format from a DDS file header.
 *
 *	\param	rkPixelFormat	Pixel format to be verified.
 *	\param	bLogError		If true, errors will be logged.
 *	\return <code>true</code>, if the header is valid.
**/
static bool VerifyPixelFormat( const DDPIXELFORMAT &rkPixelFormat, bool bLogError = true )
{
	if(rkPixelFormat.dwSize != sizeof( DDPIXELFORMAT ) )
	{
		if( bLogError )
			neolog << LogLevel( ERROR ) << "*** Invalid pixel format" << endl;
		// invalid pixel format size
		return false;
	}
	
	// file seems to be a valid DDS file; test for supported formats

	if( rkPixelFormat.dwFlags & DDPF_FOURCC )
	{
		// check for supported FOURCC format
		switch( rkPixelFormat.dwFourCC )
		{
			case FOURCC_DXT1:			// DXT1
			case FOURCC_DXT3:			// DXT3
			case FOURCC_DXT5:			// DXT5
				break;
			default:
				if( bLogError )
					neolog << LogLevel( ERROR ) << "*** Unsupported FOURCC: [" << std::hex <<
						((char*)&rkPixelFormat.dwFourCC)[0] <<
						((char*)&rkPixelFormat.dwFourCC)[1] <<
						((char*)&rkPixelFormat.dwFourCC)[2] <<
						((char*)&rkPixelFormat.dwFourCC)[3] << std::dec << "]" << endl;
				// unsupported format (includes DXT with pre-multiplied alpha)
				return false;
		}
	}
	else if( rkPixelFormat.dwFlags & DDPF_RGB )
	{
		// verify bit depth
		switch( rkPixelFormat.dwRGBBitCount )
		{
			case 8:
			case 15:
			case 16:
			case 24:
			case 32:
				break;
			default:
				if( bLogError )
					neolog << LogLevel( ERROR ) << "*** Unsupported bit depth: ["
					<< rkPixelFormat.dwRGBBitCount << "]" << endl;
				// unsupported bit depth
				return false;
		}
	}
	else if( !( rkPixelFormat.dwFlags & DDPF_ALPHA ) )
	{
		if( bLogError )
			neolog << LogLevel( ERROR ) << "*** Unsupported pixel format" << endl;
		// unable to determine format
		return false;
	}

	return true;
}

ImageCodec::ImageCodec( const vector<string> &rvstrExtensions ) :
	NeoEngine::ImageCodec( "Microsoft(r) DirectDraw(tm) Surface (DDS)", rvstrExtensions )
{
	neolog << LogLevel( INFO ) << "Loading DDS image codec" << endl << "  Copyright 2004 Numlock Software GmbH" << endl;
}


bool ImageCodec::IsType( File *pkFile ) 
{
	bool bBinary = pkFile->SetBinary( true );
	int  iOffset = pkFile->Tellg();

	pkFile->Seekg( 0, ios_base::beg );

	int iSize = pkFile->GetSize();

	if ( iSize < sizeof( DDSHEADER ) + 1 )
	{		
		// cannot possibly be a valid DDS file -> to small
		return false;
	}
	
	DDSHEADER kHeader;
	
	pkFile->Read( &kHeader, sizeof( DDSHEADER ) );

	if ( kHeader.dwMagic != DDSMAGIC )
	{		
		// invalid marker
		return false;
	}

	if ( !VerifyDescription( kHeader.sDesc, false ) )
	{
		// invalid descriptor
		return false;
	}

	if ( !VerifyPixelFormat( kHeader.sDesc.ddpfPixelFormat, false ) )
	{
		// unsupported format
		return false;
	}
	
	// file seems to be a valid DDS file
	return true;
}


ImageData *ImageCodec::LoadImage( File *pkFile )
{
	bool bBinary = pkFile->SetBinary( true );
	int  iOffset = pkFile->Tellg();
	
	pkFile->Seekg( 0, ios_base::beg );			
	
	int iSize = pkFile->GetSize();

	if ( iSize < sizeof( DDSHEADER ) + 1 )
	{
		// cannot possibly be a valid DDS file -> to small
		neolog << LogLevel( ERROR ) << "*** Invalid DDS: File is too small" << endl;		
		return 0;
	}

	DDSHEADER kHeader;
	
	pkFile->Read( &kHeader, sizeof( DDSHEADER ) );

	if ( kHeader.dwMagic != DDSMAGIC )
	{
		// invalid marker
		neolog << LogLevel( ERROR ) << "*** Invalid DDS: Unknown header magic" << endl;		
		return 0;
	}

	if ( !VerifyDescription( kHeader.sDesc ) )
	{
		// invalid descriptor
		return 0;
	}

	if ( !VerifyPixelFormat( kHeader.sDesc.ddpfPixelFormat ) )
	{
		// unsupported format
		return 0;
	}

	ImageData *pkImage = new ImageData;

	Private::DDSImageDecoder *pkDecoder = Private::CreateDecoder( pkImage, kHeader.sDesc );

	if( pkDecoder == 0 )
	{
		neolog << LogLevel( ERROR ) << "*** No suitable sub-format decoder available" << endl;
		delete pkImage;
		delete pkDecoder;
		return 0;
	}

	int iWidth = kHeader.sDesc.dwWidth;
	int iHeight= kHeader.sDesc.dwHeight;

	pkDecoder->PrepareOutput( iWidth, iHeight, kHeader.sDesc.ddpfPixelFormat );

	pkDecoder->Decode( pkFile, iWidth, iHeight );

	// process mip-maps
	if ( kHeader.sDesc.dwFlags & DDSD_MIPMAPCOUNT )
	{
		int iMipMaps = pkImage->m_iMipMaps = kHeader.sDesc.dwMipMapCount - 1;
		pkImage->m_pkMipMaps = new ImageData[ iMipMaps];

		iWidth >>= 1;
		iHeight >>= 1;

		for( int iMipMap = 0; iMipMap < iMipMaps; ++iMipMap, iWidth >>= 1, iHeight >>= 1 )
		{
			pkDecoder->SetImage( pkImage->m_pkMipMaps + iMipMap );
			pkDecoder->PrepareOutput( iWidth, iHeight, kHeader.sDesc.ddpfPixelFormat );
			pkDecoder->Decode( pkFile, iWidth, iHeight );
		}
	}

	// todo: cube-maps

	// attach decoder
	pkImage->m_pkCodec = this;

	pkFile->SetBinary( bBinary );
	
	return pkImage;
}


void ImageCodec::FreeImage(ImageData *pkImageData) {
	delete pkImageData;
}

#ifdef BUILD_STATIC


Link::Link()
{
#if defined(WIN32) && defined(_DEBUG)
	g_pkStaticModule = new ModuleStatic( "neoicdds_debug" );
#else
	g_pkStaticModule = new ModuleStatic( "neoicdds" );
#endif

	g_pkStaticModule->m_Symbols.Insert( "LoadImageCodec", (void**)LoadImageCodec );
	g_pkStaticModule->m_Symbols.Insert( "GetVersion",     (void**)GetVersion     );
	g_pkStaticModule->m_Symbols.Insert( "Initialize",     (void**)Initialize     );
	g_pkStaticModule->m_Symbols.Insert( "Shutdown",       (void**)Shutdown       );

	NeoEngine::Core::GetModuleManager()->InsertModule( g_pkStaticModule );
}


#endif


}; // namespace NeoDDS
