/***************************************************************************
                         neoictga.cpp  -  description
                             -------------------
    begin                : Wed Apr 17 2002
    copyright            : (C) 2002 by Mattias Jansson
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoICTGA, neoictga.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "neoictga.h"
#include "link.h"

#include <neoengine/file.h>
#include <neoengine/util.h>
#include <neoengine/logstream.h>

#include <string.h>

using namespace std;
using namespace NeoEngine;


namespace NeoTGA
{

#ifdef BUILD_STATIC
#  define TGAEXPORT

ModuleStatic *g_pkStaticModule = 0;

#else

#  ifdef WIN32
#    define TGAEXPORT __declspec(dllexport)
#  else
#    define TGAEXPORT
#endif

extern "C"
{

#endif

	TGAEXPORT ImageCodec *LoadImageCodec()
	{
		vector<string> vstrExtensions;

		vstrExtensions.push_back( "tga" );
		vstrExtensions.push_back( "TGA" );

		ImageCodec *pkCodec = new NeoTGA::ImageCodec( vstrExtensions );

		return pkCodec;
	}

	TGAEXPORT int Initialize()
	{
		return 0;
	}

	TGAEXPORT int Shutdown()
	{
#ifdef BUILD_STATIC
		delete g_pkStaticModule;
#endif
		return 0;
	}

	TGAEXPORT void GetVersion( std::string *pstrName, int *piMajor, int *piMinor, int *piRevision )
	{
		*pstrName   = "NeoICTGA";
		*piMajor    = NEOENGINEVERSION_MAJOR;
		*piMinor    = NEOENGINEVERSION_MINOR;
		*piRevision = NEOENGINEVERSION_REVISION;
	}

#ifdef BUILD_STATIC

#else

}

#endif


File &operator >> ( File &rkFile, Header &rkHeader )
{
	return rkFile
		>> rkHeader.m_ucIDLength
		>> rkHeader.m_ucColorMapType
		>> rkHeader.m_ucDataType
		>> rkHeader.m_usColorMapOrigin
		>> rkHeader.m_usColorMapLength
		>> rkHeader.m_ucColorMapEntrySize
		>> rkHeader.m_usImageXOrigin
		>> rkHeader.m_usImageYOrigin
		>> rkHeader.m_usImageWidth
		>> rkHeader.m_usImageHeight
		>> rkHeader.m_ucImageBPP
		>> rkHeader.m_ucImageDescriptor;
}



ImageCodec::ImageCodec( const vector<string> &rvstrExtensions ) :
	NeoEngine::ImageCodec( "TARGA Image File", rvstrExtensions )
{
	neolog << LogLevel( INFO ) << "Loading TGA image codec" << endl << "  Copyright 2002-2003 Reality Rift Studios" << endl;
}


bool ImageCodec::IsType( File *pkFile )
{
	bool bBinary = pkFile->SetBinary( true );
	int  iOffset = pkFile->Tellg();
	
	pkFile->Seekg( 0, ios_base::beg );
		
	Header kHeader;
	
	*pkFile >> kHeader;

	pkFile->Seekg( iOffset, ios_base::beg );
	pkFile->SetBinary( bBinary );
	
	if( !kHeader.m_usImageWidth )
		return false;	
	
	if( !kHeader.m_usImageHeight )
		return false;	
	
	if( ( kHeader.m_ucImageBPP != 8 )  &&
	    ( kHeader.m_ucImageBPP != 15 ) &&
	    ( kHeader.m_ucImageBPP != 16 ) &&
	    ( kHeader.m_ucImageBPP != 24 ) &&
	    ( kHeader.m_ucImageBPP != 32 ) )
		return false;	

	//if( kHeader.m_ucImageDescriptor & 0x8 ) //Supposed to be zero
	//	return false;

	int iOrigin = kHeader.m_ucImageDescriptor & Header::TGAORG_MASK;

	if( ( iOrigin != Header::TGAORG_TOPLEFT     ) &&
	    ( iOrigin != Header::TGAORG_TOPRIGHT    ) &&
	    ( iOrigin != Header::TGAORG_BOTTOMLEFT  ) &&
	    ( iOrigin != Header::TGAORG_BOTTOMRIGHT ) )
		return false;

	return true;
}


ImageData *ImageCodec::LoadImage( File *pkFile )
{
	bool bBinary = pkFile->SetBinary( true );
	int  iOffset = pkFile->Tellg();
	
	pkFile->Seekg( 0, ios_base::beg );
		
	Header kHeader;
	
	*pkFile >> kHeader;

	ImageData *pkImage = 0;

	switch( kHeader.m_ucDataType )
	{
		case Header::TGAFMT_COLMAP_UNCOMP:
		case Header::TGAFMT_COLMAP_COMP:
		{
			pkImage = ReadColMap( pkFile, kHeader );
			break;
		}

		case Header::TGAFMT_UNMAP_UNCOMP:
		case Header::TGAFMT_UNMAP_COMP:
		{
			pkImage = ReadUnmap( pkFile, kHeader );
			break;
		}

		case Header::TGAFMT_BW_UNCOMP:
		case Header::TGAFMT_BW_COMP:
		{
			pkImage = ReadBW( pkFile, kHeader );
			break;
		}
	
		case Header::TGAFMT_NO_DATA:
		default:
		{
			neolog << LogLevel( ERROR ) << "*** Unable to load TGA file: Invalid data format [" << kHeader.m_ucDataType << "]" << endl;
			break;
		}
	}

	pkFile->Seekg( iOffset, ios_base::beg );
	pkFile->SetBinary( bBinary );
	
	return pkImage;
}


void ImageCodec::FreeImage( ImageData *pkImageData )
{
	delete pkImageData;
}


void ImageCodec::WriteImage( ImageData *pkImage, File *pkFile )
{
        neolog << "WriteImage(): TGA image writing not yet implemented." << endl;
}


ImageData *ImageCodec::ReadColMap( File *pkFile, Header &rkHeader )
{
	char           aucID[256];
	unsigned int   uiPaletteBits;
	unsigned int   uiPaletteSize;
	unsigned char *pucPalette;

	if( rkHeader.m_ucImageBPP != 8 )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to load TGA file: Invalid image, bpp [" << rkHeader.m_ucImageBPP << "] not supported for color map data" << endl;
		return 0;
	}

	if( rkHeader.m_ucIDLength && !pkFile->Read( aucID, rkHeader.m_ucIDLength ) )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to load TGA file: Failed to read image descriptor" << endl;
		return 0;
	}


	switch( rkHeader.m_ucColorMapEntrySize )
	{
		case 24:
		{
			uiPaletteBits = 24;
			uiPaletteSize = rkHeader.m_usColorMapLength * 3;
			break;
		}

		case 16:
		case 32:
		{
			uiPaletteBits = 32;
			uiPaletteSize = rkHeader.m_usColorMapLength * 3;
			break;
		}

		default:
		{
			//We should *never* reach this point
			neolog << LogLevel( ERROR ) << "*** Unable to load TGA file: Invalid color map entry size [" << rkHeader.m_ucColorMapEntrySize << "]" << endl;
			return 0;
		}
	}


	pucPalette = new unsigned char[ uiPaletteSize ];

	if( rkHeader.m_usColorMapOrigin )
		neolog << LogLevel( WARNING ) << "*** Color map offset [" << rkHeader.m_usColorMapOrigin << "] not supported for TGA files" << endl;
		

	if( rkHeader.m_ucColorMapEntrySize != 16 )
	{
		if( !pkFile->Read( pucPalette, uiPaletteSize ) )
		{
			neolog << LogLevel( ERROR ) << "*** Unable to load TGA file: Failed to read palette" << endl;
			delete [] pucPalette;
			return 0;
		}
	}
	else
	{
		unsigned char *pucCur = pucPalette;

		//Extract channels
		for( unsigned int i = 0; i < uiPaletteSize; i += 4, pucCur += 4 )
		{
			unsigned short usPixel = 0;

			if( !( *pkFile >> usPixel ) )
			{
				neolog << LogLevel( ERROR ) << "*** Unable to load TGA file: Failed to read palette entry [" << ( i / 4 ) << "]" << endl;
				delete [] pucPalette;
				return 0;
			}

			pucCur[3] = ( usPixel & 0x8000 ) >> 12;
			pucCur[0] = ( usPixel & 0xFC00 ) >> 7;
			pucCur[1] = ( usPixel & 0x03E0 ) >> 2;
			pucCur[2] = ( usPixel & 0x001F ) << 3;
		}
	}


	unsigned char *pucImage = new unsigned char[ rkHeader.m_usImageWidth * rkHeader.m_usImageHeight ];


	if( rkHeader.m_ucDataType == Header::TGAFMT_COLMAP_COMP )
	{
		if( !UncompressStream( pkFile, pucImage, rkHeader.m_usImageWidth, rkHeader.m_usImageHeight, 1 ) )
		{
			delete [] pucImage;
			delete [] pucPalette;
			return 0;
		}
	}
	else
	{
		if( !pkFile->Read( pucImage, rkHeader.m_usImageWidth * rkHeader.m_usImageHeight ) )
		{
			neolog << LogLevel( ERROR ) << "*** Unable to load TGA file: Failed to read image data" << endl;
			delete [] pucImage;
			delete [] pucPalette;
			return 0;
		}
	}


	unsigned int   uiBytesPerPixel = ( rkHeader.m_ucColorMapEntrySize == 24 ) ? 3 : 4;


	//Create image data
	ImageData *pkData = new ImageData;

	pkData->m_eFormat   = ( rkHeader.m_ucColorMapEntrySize == 24 ) ? Texture::RGB : Texture::RGBA;
	pkData->m_iWidth    = rkHeader.m_usImageWidth;
	pkData->m_iHeight   = rkHeader.m_usImageHeight;
	pkData->m_iChannels = ( rkHeader.m_ucColorMapEntrySize == 24 ) ? 3 : 4;
	pkData->m_iBPP      = pkData->m_iChannels * 8;
	pkData->m_pkCodec   = this;
	pkData->m_pucData   = new unsigned char[ rkHeader.m_usImageWidth * rkHeader.m_usImageHeight * uiBytesPerPixel ];


	//Map data
	unsigned char *pucCur         = 0;
	unsigned char *pucDest        = pkData->m_pucData;
	int            iSrcStep       = 0;
	int            iSrcLineStep   = 0;

	int iOrigin = rkHeader.m_ucImageDescriptor & Header::TGAORG_MASK;

	//We know origin is valid, it is checked in IsType method
	switch( iOrigin )
	{
		default:
		case Header::TGAORG_TOPLEFT:
		{
			pucCur        = pucImage;
			iSrcStep      = 1;
			iSrcLineStep  = rkHeader.m_usImageWidth;
			break;
		}

		case Header::TGAORG_TOPRIGHT:
		{
			pucCur        = pucImage + ( rkHeader.m_usImageWidth - 1 );
			iSrcStep      = -1;
			iSrcLineStep  = rkHeader.m_usImageWidth;
			break;
		}

		case Header::TGAORG_BOTTOMLEFT:
		{
			pucCur        = pucImage + ( rkHeader.m_usImageWidth * ( rkHeader.m_usImageHeight - 1 ) );
			iSrcStep      = 1;
			iSrcLineStep  = -rkHeader.m_usImageWidth;
			break;
		}

		case Header::TGAORG_BOTTOMRIGHT:
		{
			pucCur        = pucImage + ( ( rkHeader.m_usImageWidth * rkHeader.m_usImageHeight ) - 1 );
			iSrcStep      = -1;
			iSrcLineStep  = -rkHeader.m_usImageWidth;
			break;
		}
	}


	for( unsigned int y = 0; y < rkHeader.m_usImageHeight; ++y )
	{
		unsigned char *pucLineStart = pucCur;

		for( unsigned int x = 0; x < rkHeader.m_usImageWidth; ++x )
		{
			fmemcpy( pucDest, &pucPalette[ (*pucCur) * uiBytesPerPixel ], uiBytesPerPixel );

			pucDest += uiBytesPerPixel;
			pucCur  += iSrcStep;
		}

		pucCur = pucLineStart + iSrcLineStep;
	}


	delete [] pucImage;
	delete [] pucPalette;

	return pkData;
}


ImageData *ImageCodec::ReadUnmap( File *pkFile, Header &rkHeader )
{
	unsigned char uiBytesPerPixel;
	unsigned char aucID[256];

	if( rkHeader.m_ucImageBPP == 8 )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to load TGA file: Invalid image, bpp [" << rkHeader.m_ucImageBPP << "] not supported for unmapped data" << endl;
		return 0;
	}

	if( rkHeader.m_ucIDLength && !pkFile->Read( aucID, rkHeader.m_ucIDLength ) )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to load TGA file: Failed to read image descriptor" << endl;
		return 0;
	}


	if( rkHeader.m_ucImageBPP == 15 )
		uiBytesPerPixel = 2;
	else
		uiBytesPerPixel = ( rkHeader.m_ucImageBPP >> 3 );


	//Is this correct?
	if( rkHeader.m_usColorMapLength * rkHeader.m_ucColorMapEntrySize )
	{
		neolog << LogLevel( WARNING ) << "*** Color map data present in unmapped TGA image, untested" << endl;
		pkFile->Seekg( ( rkHeader.m_usColorMapLength * rkHeader.m_ucColorMapEntrySize >> 3 ), ios_base::cur );
	}


	unsigned char *pucImage = new unsigned char[ rkHeader.m_usImageWidth * rkHeader.m_usImageHeight * uiBytesPerPixel ];

	if( rkHeader.m_ucDataType == Header::TGAFMT_UNMAP_COMP )
	{
		if( !UncompressStream( pkFile, pucImage, rkHeader.m_usImageWidth, rkHeader.m_usImageHeight, uiBytesPerPixel ) )
		{
			delete [] pucImage;
			return 0;
		}
	}
	else
	{
		if( !pkFile->Read( pucImage, rkHeader.m_usImageWidth * rkHeader.m_usImageHeight * uiBytesPerPixel ) )
		{
			neolog << LogLevel( ERROR ) << "*** Unable to load TGA file: Unable to read image data" << endl;
			delete [] pucImage;
			return 0;
		}
	}


	unsigned int uiDestBytesPerPixel;

	ImageData *pkData = new ImageData;

	pkData->m_iWidth    = rkHeader.m_usImageWidth;
	pkData->m_iHeight   = rkHeader.m_usImageHeight;
	pkData->m_pkCodec   = this;

	switch( uiBytesPerPixel )
	{
		case 2:
		case 3:
		{
			pkData->m_eFormat   = Texture::RGB;
			pkData->m_iBPP      = 24;
			pkData->m_iChannels = 3;
			uiDestBytesPerPixel = 3;
			break;
		}

		default:
		case 4:
		{
			pkData->m_eFormat   = Texture::RGBA;
			pkData->m_iBPP      = 32;
			pkData->m_iChannels = 4;
			uiDestBytesPerPixel = 4;
			break;
		}
	}

	pkData->m_pucData   = new unsigned char[ rkHeader.m_usImageWidth * rkHeader.m_usImageHeight * uiDestBytesPerPixel ];



	//Map data
	unsigned char *pucCur         = 0;
	unsigned char *pucDest        = pkData->m_pucData;
	int            iSrcStep       = 0;
	int            iSrcLineStep   = 0;

	int iOrigin = rkHeader.m_ucImageDescriptor & Header::TGAORG_MASK;

	//We know origin is valid, it is checked in IsType method
	switch( iOrigin )
	{
		default:
		case Header::TGAORG_TOPLEFT:
		{
			pucCur        = pucImage;
			iSrcStep      = uiBytesPerPixel;
			iSrcLineStep  = rkHeader.m_usImageWidth * uiBytesPerPixel;
			break;
		}

		case Header::TGAORG_TOPRIGHT:
		{
			pucCur        = pucImage + ( ( rkHeader.m_usImageWidth - 1 ) * uiBytesPerPixel );
			iSrcStep      = -uiBytesPerPixel;
			iSrcLineStep  = rkHeader.m_usImageWidth * uiBytesPerPixel;
			break;
		}

		case Header::TGAORG_BOTTOMLEFT:
		{
			pucCur        = pucImage + ( rkHeader.m_usImageWidth * ( rkHeader.m_usImageHeight - 1 ) * uiBytesPerPixel );
			iSrcStep      = uiBytesPerPixel;
			iSrcLineStep  = -rkHeader.m_usImageWidth * uiBytesPerPixel;
			break;
		}

		case Header::TGAORG_BOTTOMRIGHT:
		{
			pucCur        = pucImage + ( ( ( rkHeader.m_usImageWidth * rkHeader.m_usImageHeight ) - 1 ) * uiBytesPerPixel );
			iSrcStep      = -uiBytesPerPixel;
			iSrcLineStep  = -rkHeader.m_usImageWidth * uiBytesPerPixel;
			break;
		}
	}


	for( unsigned int y = 0; y < rkHeader.m_usImageHeight; ++y )
	{
		unsigned char *pucLineStart = pucCur;

		if( uiBytesPerPixel == 2 )
		{
			//Expand to 24 bit
			unsigned short usPixel;

			for( unsigned int x = 0; x < rkHeader.m_usImageWidth; ++x )
			{
				usPixel    = *((unsigned short*)pucCur);

				*pucDest++ = ( usPixel & 0x7C00 ) >> 7;
				*pucDest++ = ( usPixel & 0x03E0 ) >> 2;
				*pucDest++ = ( usPixel & 0x001F ) << 3;

				pucCur  += iSrcStep;
			}
		}
		else
		{
			for( unsigned int x = 0; x < rkHeader.m_usImageWidth; ++x )
			{
				*pucDest++ = *( pucCur + 2 );
				*pucDest++ = *( pucCur + 1 );
				*pucDest++ = *( pucCur + 0 );

				if( uiDestBytesPerPixel == 4 )
					*pucDest++ = *( pucCur + 3 );

				pucCur  += iSrcStep;
			}
		}

		pucCur = pucLineStart + iSrcLineStep;
	}


	delete [] pucImage;

	return pkData;
}


ImageData *ImageCodec::ReadBW( File *pkFile, Header &rkHeader )
{
	unsigned char aucID[256];

	if( rkHeader.m_ucImageBPP != 8 )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to load TGA file: Invalid image, bpp [" << rkHeader.m_ucImageBPP << "] not supported for bw data" << endl;
		return 0;
	}

	if( rkHeader.m_ucIDLength && !pkFile->Read( aucID, rkHeader.m_ucIDLength ) )
	{
		neolog << LogLevel( ERROR ) << "*** Unable to load TGA file: Failed to read image descriptor" << endl;
		return 0;
	}

	//Is this correct?
	if( rkHeader.m_usColorMapLength * rkHeader.m_ucColorMapEntrySize )
	{
		neolog << LogLevel( ERROR ) << "*** Color map data present in bw image, untested" << endl;
		pkFile->Seekg( ( rkHeader.m_usColorMapLength * rkHeader.m_ucColorMapEntrySize >> 3 ), ios_base::cur );
	}


	unsigned char *pucImage = new unsigned char[ rkHeader.m_usImageWidth * rkHeader.m_usImageHeight ];

	if( rkHeader.m_ucDataType == Header::TGAFMT_BW_COMP )
	{
		if( !UncompressStream( pkFile, pucImage, rkHeader.m_usImageWidth, rkHeader.m_usImageHeight, 1 ) )
		{
			delete [] pucImage;
			return 0;
		}
	}
	else
	{
		if( !pkFile->Read( pucImage, rkHeader.m_usImageWidth * rkHeader.m_usImageHeight ) )
		{
			neolog << LogLevel( ERROR ) << "*** Unable to load TGA file: Failed to read image data" << endl;
			delete [] pucImage;
			return 0;
		}
	}


	ImageData *pkData = new ImageData;

	pkData->m_iWidth    = rkHeader.m_usImageWidth;
	pkData->m_iHeight   = rkHeader.m_usImageHeight;
	pkData->m_pkCodec   = this;
	pkData->m_eFormat   = Texture::ALPHA;
	pkData->m_iChannels = 1;
	pkData->m_iBPP      = 8;
	pkData->m_pucData   = new unsigned char[ rkHeader.m_usImageWidth * rkHeader.m_usImageHeight ];


	//Map data
	unsigned char *pucCur         = 0;
	unsigned char *pucDest        = pkData->m_pucData;
	int            iSrcStep       = 0;
	int            iSrcLineStep   = 0;

	int iOrigin = rkHeader.m_ucImageDescriptor & Header::TGAORG_MASK;

	//We know origin is valid, it is checked in IsType method
	switch( iOrigin )
	{
		default:
		case Header::TGAORG_TOPLEFT:
		{
			pucCur        = pucImage;
			iSrcStep      = 1;
			iSrcLineStep  = rkHeader.m_usImageWidth;
			break;
		}

		case Header::TGAORG_TOPRIGHT:
		{
			pucCur        = pucImage + ( rkHeader.m_usImageWidth - 1 );
			iSrcStep      = -1;
			iSrcLineStep  = rkHeader.m_usImageWidth;
			break;
		}

		case Header::TGAORG_BOTTOMLEFT:
		{
			pucCur        = pucImage + ( rkHeader.m_usImageWidth * ( rkHeader.m_usImageHeight - 1 ) );
			iSrcStep      = 1;
			iSrcLineStep  = -rkHeader.m_usImageWidth;
			break;
		}

		case Header::TGAORG_BOTTOMRIGHT:
		{
			pucCur        = pucImage + ( ( rkHeader.m_usImageWidth * rkHeader.m_usImageHeight ) - 1 );
			iSrcStep      = -1;
			iSrcLineStep  = -rkHeader.m_usImageWidth;
			break;
		}
	}


	for( unsigned int y = 0; y < rkHeader.m_usImageHeight; ++y )
	{
		unsigned char *pucLineStart = pucCur;

		for( unsigned int x = 0; x < rkHeader.m_usImageWidth; ++x )
		{
			*pucDest++  = *pucCur;

			pucCur     += iSrcStep;
		}

		pucCur = pucLineStart + iSrcLineStep;
	}


	delete [] pucImage;

	return pkData;
}


bool ImageCodec::UncompressStream( File *pkFile, unsigned char *pucDest, unsigned int uiWidth, unsigned int uiHeight, unsigned int uiBytesPerPixel )
{
	unsigned int  uiBytesRead  = 0;
	unsigned int  uiSize       = 0;
	unsigned int  uiRunLen     = 0;
	
	unsigned char ucHeader     = 0;
	unsigned char aucColor[4];

	uiSize = uiWidth * uiHeight * uiBytesPerPixel;

	while( uiBytesRead < uiSize )
	{
		pkFile->Read( &ucHeader, 1 );

		if( ucHeader & 0x80 )
		{
			ucHeader ^= 0x80;

			if( !pkFile->Read( aucColor, uiBytesPerPixel ) )
			{
				neolog << LogLevel( ERROR ) << "*** Unable to uncompress TGA stream: Failed to read data" << endl;
				return false;
			}

			uiRunLen = ( ucHeader + 1 ) * uiBytesPerPixel;

			for( unsigned int i = 0; i < uiRunLen; i += uiBytesPerPixel )
			{
				for( unsigned int c = 0; c < uiBytesPerPixel; ++c )
					*pucDest++ = aucColor[c];
			}

			uiBytesRead += uiRunLen;
		}
		else
		{
			uiRunLen = ( ucHeader + 1 ) * uiBytesPerPixel;

			if( !pkFile->Read( pucDest, uiRunLen ) )
			{
				neolog << LogLevel( ERROR ) << "*** Unable to uncompress TGA stream: Failed to read data" << endl;
				return false;
			}

			pucDest     += uiRunLen;
			uiBytesRead += uiRunLen;
		}
	}

	return true;
}



#ifdef BUILD_STATIC


Link::Link()
{
#if defined(WIN32) && defined(_DEBUG)
	g_pkStaticModule = new ModuleStatic( "neoictga_debug" );
#else
	g_pkStaticModule = new ModuleStatic( "neoictga" );
#endif

	g_pkStaticModule->m_Symbols.Insert( "LoadImageCodec", (void**)LoadImageCodec );
	g_pkStaticModule->m_Symbols.Insert( "GetVersion",     (void**)GetVersion     );
	g_pkStaticModule->m_Symbols.Insert( "Initialize",     (void**)Initialize     );
	g_pkStaticModule->m_Symbols.Insert( "Shutdown",       (void**)Shutdown       );

	NeoEngine::Core::GetModuleManager()->InsertModule( g_pkStaticModule );
}


#endif


}; // namespace NeoTGA
