/***************************************************************************
                  area.cpp  -  Object with background material
                             -------------------
    begin                : Sun Nov 17 2002
    copyright            : (C) 2002 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoWTK, area.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "area.h"

#include <neoengine/render.h>
#include <neoengine/renderprimitive.h>
#include <neoengine/vertex.h>

#ifdef HAVE_NEOCHUNKIO
#  include <neochunkio/material.h>
#  include <neochunkio/stdstring.h>

using namespace NeoChunkIO;

#endif

using namespace std;
using namespace NeoEngine;



namespace NeoWTK
{


Area::Area( Object *pkParent, Area *pkObject ) :
	Object( pkParent, pkObject )
{
	m_pkRenderPrim = 0;

	if( !pkObject )
	{
		m_aiNumTiles[0] = 1;
		m_aiNumTiles[1] = 1;

		m_eTilingMode  = STRETCH;
	}
	else
	{
		m_pkBackground = pkObject->m_pkBackground;
		m_eTilingMode  = pkObject->m_eTilingMode;
	
		m_aiNumTiles[0] = pkObject->m_aiNumTiles[0];
		m_aiNumTiles[1] = pkObject->m_aiNumTiles[1];

		m_akUV[0]       = pkObject->m_akUV[0];
		m_akUV[1]       = pkObject->m_akUV[1];
		m_akUV[2]       = pkObject->m_akUV[2];
	}
}


Area::~Area()
{
	delete m_pkRenderPrim;
}


void Area::SetBackground( const MaterialPtr &pkMat )
{
	m_pkBackground = pkMat;

	m_pkBackground->m_kZBufferMode.Set( ZBufferMode::ALWAYSNOWRITE );

	// New material, force recalculation of vertex buffer (texture coords)
	if( m_pkVertexBuffer )
		SetVertexBuffer();
}


const Coord &Area::SetSize( const Coord &rkSize )
{
	if( rkSize == m_kSize )
		return m_kSize;

	Object::SetSize( rkSize );

	//Log( LOGLEVEL_DEBUG, "[Area::SetSize] [%s] size set to [ %d %d ]\n", GetName().c_str(), m_kSize.x, m_kSize.y );

	//Update vertex buffer
	if( m_pkVertexBuffer )
		SetVertexBuffer();

	return m_kSize;
}


void Area::SetVertexBuffer()
{
	if( !m_pkVertexBuffer )
		m_pkVertexBuffer  = Core::Get()->GetRenderDevice()->CreateVertexBuffer( Buffer::DYNAMIC, ( m_aiNumTiles[0] + 1 ) * ( m_aiNumTiles[1] + 1 ), &TexVertex::s_kDecl );

	m_pkVertexBuffer->Lock( Buffer::WRITE );

	float fStep[2];
	float fOONumTiles[2];

	fOONumTiles[0] = 1.0f / float( m_aiNumTiles[0] );
	fOONumTiles[1] = 1.0f / float( m_aiNumTiles[1] );

	fStep[0] = float( m_kSize.x ) / Core::Get()->GetRenderDevice()->GetWidth();
	fStep[1] = float( m_kSize.y ) / Core::Get()->GetRenderDevice()->GetHeight();

	fStep[0] *= fOONumTiles[0];
	fStep[1] *= fOONumTiles[1];

	TexVertex *pkVertex = (TexVertex*)m_pkVertexBuffer->GetVertex();

	float fOOWidth  = 1.0f / float( m_pkBackground->m_pkTexture->GetWidth()  );
	float fOOHeight = 1.0f / float( m_pkBackground->m_pkTexture->GetHeight() );

	float fX = 0.0f;
	float fY = 0.0f;

	float fU = ( float( m_akUV[0].x ) + 0.5f ) * fOOWidth;
	float fV = ( float( m_akUV[0].y ) + 0.5f ) * fOOHeight;

	float fUVStep[2][2];

	float fUOffset = ( m_akUV[1].x > 0 ) ? -1.0f : ( ( m_akUV[1].x < 0 ) ? -1.0f : 0.0f );
	float fVOffset = ( m_akUV[1].y > 0 ) ? -1.0f : ( ( m_akUV[1].y < 0 ) ? -1.0f : 0.0f );

	fUVStep[0][0] = ( float( m_akUV[1].x ) + fUOffset ) * fOOWidth  * fOONumTiles[0];
	fUVStep[0][1] = ( float( m_akUV[1].y ) + fVOffset ) * fOOHeight * fOONumTiles[0];

	fUOffset = ( m_akUV[2].x > 0 ) ? -1.0f : ( ( m_akUV[2].x < 0 ) ? -1.0f : 0.0f );
	fVOffset = ( m_akUV[2].y > 0 ) ? -1.0f : ( ( m_akUV[2].y < 0 ) ? -1.0f : 0.0f );

	fUVStep[1][0] = ( float( m_akUV[2].x ) + fUOffset ) * fOOWidth  * fOONumTiles[1];
	fUVStep[1][1] = ( float( m_akUV[2].y ) + fVOffset ) * fOOHeight * fOONumTiles[1];

	for( int iY = 0; iY < m_aiNumTiles[1] + 1; ++iY, fY += fStep[1] )
	{
		float fUSave = fU;
		float fVSave = fV;

		for( int iX = 0; iX < m_aiNumTiles[0] + 1; ++iX, fX += fStep[0], ++pkVertex, fU += fUVStep[0][0], fV += fUVStep[0][1] )
		{
			pkVertex->m_kPosition.Set( fX, fY, 0.0f );
			pkVertex->m_afTexCoord[0] = fU;
			pkVertex->m_afTexCoord[1] = fV;
		}

		fX = 0.0f;

		fU = fUSave + fUVStep[1][0];
		fV = fVSave + fUVStep[1][1];
	}

	m_pkVertexBuffer->Unlock();
}


void Area::SetPolygonBuffer()
{
	if( !m_pkPolygonBuffer )
		m_pkPolygonBuffer = Core::Get()->GetRenderDevice()->CreatePolygonBuffer( Buffer::DYNAMIC, m_aiNumTiles[0] * m_aiNumTiles[1] * 2 );

	m_pkPolygonBuffer->Lock( Buffer::WRITE );

	Polygon *pkPoly = m_pkPolygonBuffer->GetPolygon();

	int iYOffset     = 0;
	int iYNextOffset = m_aiNumTiles[0] + 1;

	for( int iY = 0; iY < m_aiNumTiles[1]; ++iY )
	{
		for( int iX = 0; iX < m_aiNumTiles[0]; ++iX )
		{
			(*pkPoly)[0] = iX     + iYOffset;
			(*pkPoly)[1] = iX     + iYNextOffset;
			(*pkPoly)[2] = iX + 1 + iYOffset;

			++pkPoly;

			(*pkPoly)[0] = iX + 1 + iYOffset;
			(*pkPoly)[1] = iX     + iYNextOffset;
			(*pkPoly)[2] = iX + 1 + iYNextOffset;

			++pkPoly;
		}

		iYOffset     += m_aiNumTiles[0] + 1;
		iYNextOffset += m_aiNumTiles[0] + 1;
	}

	m_pkPolygonBuffer->Stripify();

	m_pkPolygonBuffer->Unlock();
}


void Area::SetTilingMode( TILINGMODE eMode, int iX, int iY )
{
	bool bUpdateBuffers = false;

	if( ( m_eTilingMode != eMode ) || ( m_aiNumTiles[0] != iX ) || ( m_aiNumTiles[1] != iY ) )
		bUpdateBuffers = true;

	m_eTilingMode  = eMode;
	m_aiNumTiles[0] = iX;
	m_aiNumTiles[1] = iY;

	if( bUpdateBuffers && m_pkVertexBuffer )
		SetVertexBuffer();

	if( bUpdateBuffers && m_pkPolygonBuffer )
		SetPolygonBuffer();
}


void Area::SetUVs( const Coord &rkOrigin, const Coord &rkRight, const Coord &rkBottom )
{
	m_akUV[0] = rkOrigin;
	m_akUV[1] = rkRight;
	m_akUV[2] = rkBottom;

	//Update vertex buffer
	if( m_pkVertexBuffer )
		SetVertexBuffer();
}


bool Area::Render( Frustum *pkFrustum, bool bForce )
{
	if( m_pkBackground )
	{
		if( !m_pkVertexBuffer )
			SetVertexBuffer();

		if( !m_pkPolygonBuffer )
			SetPolygonBuffer();

		//Render background
		if( !m_pkRenderPrim )
		{
			m_pkRenderPrim = new RenderPrimitive;

			m_pkRenderPrim->m_pkMaterial      = m_pkBackground;
			m_pkRenderPrim->m_pkVertexBuffer  = m_pkVertexBuffer;

			if( m_pkPolygonBuffer->IsStripped() )
			{
				m_pkRenderPrim->m_pkPolygonStripBuffer = m_pkPolygonBuffer->GetStrip();
				m_pkRenderPrim->m_ePrimitive           = RenderPrimitive::TRIANGLESTRIP;
				m_pkRenderPrim->m_uiNumPrimitives      = m_pkRenderPrim->m_pkPolygonStripBuffer->GetNumElements();
			}
			else
			{
				m_pkRenderPrim->m_pkPolygonBuffer      = m_pkPolygonBuffer;
				m_pkRenderPrim->m_ePrimitive           = RenderPrimitive::TRIANGLES;
				m_pkRenderPrim->m_uiNumPrimitives      = m_pkPolygonBuffer->GetNumElements();
			}
		}

		m_pkRenderPrim->m_kModelMatrix.SetTranslation( Vector3d( float( GetWorldPosition().x ) / Core::Get()->GetRenderDevice()->GetWidth(), float( GetWorldPosition().y ) / Core::Get()->GetRenderDevice()->GetHeight(), 0.0f ) );

		Core::Get()->GetRenderDevice()->Render( *m_pkRenderPrim );
	}

	return Object::Render( pkFrustum );
}


unsigned int Area::ProcessMsg( Msg *pkMsg )
{
	return Object::ProcessMsg( pkMsg );
}


Object *Area::Duplicate( Object *pkParent, Object *pkObject )
{
	if( !pkObject )
		pkObject = new Area( pkParent, this );
	else
	{
		Area *pkArea = dynamic_cast< Area* >( pkObject );

		pkArea->m_pkBackground = m_pkBackground;
		pkArea->m_eTilingMode  = m_eTilingMode;
	
		pkArea->m_aiNumTiles[0] = m_aiNumTiles[0];
		pkArea->m_aiNumTiles[1] = m_aiNumTiles[1];

		pkArea->m_akUV[0]       = m_akUV[0];
		pkArea->m_akUV[1]       = m_akUV[1];
		pkArea->m_akUV[2]       = m_akUV[2];
	}

	return Object::Duplicate( pkParent, pkObject );
}


#ifdef HAVE_NEOCHUNKIO


int AreaChunk::ParseData( unsigned int uiFlags, FileManager *pkFileManager )
{
	Chunk *pkChunk = 0;
	int    iRet    = 0;

	if( !m_pkObject )
		m_pkObject = new Area( 0, 0 );

	Area *pkArea = dynamic_cast< Area* >( m_pkObject );

	if( ( iRet = ObjectChunk::ParseData( uiFlags, pkFileManager ) ) <= 0 )
		return iRet;

	if( ( pkChunk = FindChunk( "", NeoChunkIO::ChunkType::MATERIAL, -1 ) ) )
		pkArea->SetBackground( dynamic_cast< MaterialChunk* >( pkChunk )->m_pkMaterial );

	Area::TILINGMODE eTilingMode = Area::STRETCH;
	int iTilesX = 1;
	int iTilesY = 1;

	if( ( pkChunk = FindChunk( "tiling", NeoChunkIO::ChunkType::STRING, -1 ) ) )
	{
		std::string &strTiling = dynamic_cast< StringChunk* >( pkChunk )->m_strData;

		if( strTiling == "stretch" )
			eTilingMode = Area::STRETCH;
	}

	if( ( pkChunk = FindChunk( "tiles", ChunkType::COORD, -1 ) ) )
	{
		iTilesX = dynamic_cast< CoordChunk* >( pkChunk )->m_kCoord.x;
		iTilesY = dynamic_cast< CoordChunk* >( pkChunk )->m_kCoord.y;
	}

	pkArea->SetTilingMode( eTilingMode, iTilesX, iTilesY );

	Coord akUV[3];

	if( ( pkChunk = FindChunk( "origin", ChunkType::COORD, -1 ) ) )
		akUV[0] = dynamic_cast< CoordChunk* >( pkChunk )->m_kCoord;

	if( ( pkChunk = FindChunk( "right", ChunkType::COORD, -1 ) ) )
		akUV[1] = dynamic_cast< CoordChunk* >( pkChunk )->m_kCoord;

	if( ( pkChunk = FindChunk( "bottom", ChunkType::COORD, -1 ) ) )
		akUV[2] = dynamic_cast< CoordChunk* >( pkChunk )->m_kCoord;

	pkArea->SetUVs( akUV[0], akUV[1], akUV[2] );

	return 1;
}


#endif


}; /*! namespace NeoWTK */

