/***************************************************************************
                    area.h  -  Object with background material
                             -------------------
    begin                : Sun Nov 17 2002
    copyright            : (C) 2002 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoWTK, area.h

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#ifndef __NEOWTKAREA_H
#define __NEOWTKAREA_H


#include "base.h"
#include "object.h"

#include <neoengine/vertexbuffer.h>
#include <neoengine/polygonbuffer.h>
#include <neoengine/material.h>

#ifdef HAVE_NEOCHUNKIO
#  include <neochunkio/complex.h>
#  include "chunktype.h"
#endif


/**
  * \file area.h
  * An area is a simple object with a single material background
  */


// External classes
namespace NeoEngine
{
	class RenderPrimitive;
};


namespace NeoWTK
{


/**
  * \class Area
  * \brief Object with background material
  * \author Mattias Jansson (mattias@realityrift.com)
  */	
class NEOWTK_API Area : public Object
{
	public:

		/**
		* \enum TILINGMODE
		* \brief Tiling modes for area UV coords
		*/
		enum TILINGMODE
		{
		  STRETCH                                     = 0
		};


	protected:

		/*! Background */
		NeoEngine::MaterialPtr                        m_pkBackground;

		/*! Vertex buffer */
		NeoEngine::VertexBufferPtr                    m_pkVertexBuffer;

		/*! Polygon buffer */
		NeoEngine::PolygonBufferPtr                   m_pkPolygonBuffer;

		/*! Render op object */
		NeoEngine::RenderPrimitive                   *m_pkRenderPrim;

		/*! Tiling mode */
		TILINGMODE                                    m_eTilingMode;

		/*! Number of tiles in each direction */
		int                                           m_aiNumTiles[2];

		/*! UV coord vectors */
		Coord                                         m_akUV[3];


		/**
		* Setup vertex buffer
		*/
		void                                          SetVertexBuffer();

		/**
		* Setup polygon buffer
		*/
		void                                          SetPolygonBuffer();



	public:



		/**
		* \param pkParent                             Parent object
		* \param pkObject                             Reference object to copy values from
		*/
		                                              Area( Object *pkParent, Area *pkObject = 0 );

		/**
		*/
		virtual                                      ~Area();

		/**
		* Set size
		* \param rkSize                               Size
		* \return                                     New size
		*/
		virtual const Coord                          &SetSize( const Coord &rkSize );

		/**
		* Process message
		* \param pkMsg                                Message object
		* \return                                     Message specific return code, 0 indicates message was unprocessed
		*/
		unsigned int                                  ProcessMsg( Msg *pkMsg );

		/**
		* Render object, recurses on children
		* \param pkFrustum                            Currently ignored
		* \param bForce                               Force rendering
		* \return                                     true if rendered, false if not
		*/
		virtual bool                                  Render( NeoEngine::Frustum *pkFrustum = 0, bool bForce = false );

		/**
		* Set background material
		* \param pkMat                                New material
		*/
		virtual void                                  SetBackground( const NeoEngine::MaterialPtr &pkMat );

		/**
		* Set tiling parameters
		* \param eMode                                Tiling mode
		* \param iX                                   Tiles in horizontal direction
		* \param iY                                   Tiles in vertical direction
		*/
		virtual void                                  SetTilingMode( TILINGMODE eMode, int iX = 1, int iY = 1 );

		/**
		* Set UV coordinates for area corners, forming rectangle (possibly skewed) in texture space
		* \param rkOrigin                             Coordinate in texture for top-left corner of area
		* \param rkRight                              Coordinates in texture space (pixels) for top right corner
		* \param rkBottom                             Coordinates in texture space (pixels) for bottom left corner
		*/
		virtual void                                  SetUVs( const Coord &rkOrigin, const Coord &rkRight, const Coord &rkBottom );

		/**
		* \return                                     Background material
		*/
		virtual NeoEngine::MaterialPtr               &GetBackground() { return m_pkBackground; }

		/**
		* Duplicate and/or copy this object
		* \param pkParent                             Parent object to attach to
		* \param pkObject                             Object to copy data into, if null will allocate new object. Will only duplicate children if this object is null or has no children
		* \return                                     New object that is duplicate of this object
		*/
		virtual Object                               *Duplicate( Object *pkParent = 0, Object *pkObject = 0 );
};


#ifdef HAVE_NEOCHUNKIO


/**
  * \class AreaChunk
  * \brief Base chunk for area objects
  * \author Mattias Jansson (mattias@realityrift.com)
  */
class NEOWTK_API AreaChunk : public ObjectChunk
{
	public:

		/**
		* Initialize chunk
		* \param usType                            Chunk type
		* \param rstrType                          Chunk type as string
		* \param rstrID                            Chunk ID string
		*/
		                                           AreaChunk( unsigned short usType, const NeoEngine::HashString &rstrType, const NeoEngine::HashString &rstrID = "" ) : ObjectChunk( usType, rstrType, rstrID ) {}

		/**
		* Deallocate data and subchunks
		*/
		virtual                                   ~AreaChunk() {}

		/**
		* Parse chunk data
		* \param uiFlags                           Parse flags
		* \param pkFileManager                     File manager
		* \return                                  <0 if error, >0 if successful (0 reserved)
		*/
		virtual int                                ParseData( unsigned int uiFlags, NeoEngine::FileManager *pkFileManager );

		/**
		* Allocate new chunk
		* \param usType                            Type identifier
		* \param rstrType                          Type identifier as string
		* \param rstrID                            ID string
		* \return                                  New chunk
		*/
		static NeoChunkIO::Chunk                  *Allocator( unsigned short usType, const NeoEngine::HashString &rstrType, const NeoEngine::HashString &rstrID ) { return new AreaChunk( usType, rstrType, rstrID ); }
};


#endif /*! HAVE_NEOCHUNKIO */


}; /*! namespace NeoWTK */


#endif
