/***************************************************************************
               button.cpp  -  A clickable button with two states
                             -------------------
    begin                : Tue Jan 7 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoWTK, button.cpp

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#include "button.h"
#include "borderarea.h"
#include "textarea.h"
#include "msg.h"

#ifdef HAVE_NEOCHUNKIO
#  include <neochunkio/chunk.h>

using namespace NeoChunkIO;

#endif

using namespace std;
using namespace NeoEngine;


namespace NeoWTK
{


Button::Button( Object *pkParent, Button *pkObject, bool bCreateAreas ) :
	Object( pkParent, pkObject )
{
	m_pkState[0] = 0;
	m_pkState[1] = 0;

	if( bCreateAreas )
	{
		//Create state objects
		if( !pkObject )
		{
			TextArea *pkCenter = new TextArea( 0 );
			pkCenter->SetName( "Center" );
			m_pkState[0] = new BorderArea( this, pkCenter );

			pkCenter = new TextArea( 0 );
			pkCenter->SetName( "Center" );
			m_pkState[1] = new BorderArea( this, pkCenter );
		}
		else
		{
			m_pkState[0] = dynamic_cast< BorderArea* >( m_vpkChildren[0] );
			m_pkState[1] = dynamic_cast< BorderArea* >( m_vpkChildren[1] );
		}

		m_pkState[0]->Activate();
		m_pkState[1]->Deactivate();
	}

	m_eState = UP;
}


Button::~Button()
{
	//State border area objects are deleted by object dtor, they are children to us
}


const Coord &Button::SetSize( const Coord &rkSize )
{
	if( m_pkState[0] )
		m_pkState[0]->SetSize( rkSize );

	if( m_pkState[1] )
		m_pkState[1]->SetSize( rkSize );

	return Object::SetSize( rkSize );
}


bool Button::Render( Frustum *pkFrustum, bool bForce )
{
	return Object::Render( pkFrustum, bForce );
}


BorderArea *Button::GetStateObject( STATE eState )
{
	return m_pkState[ eState ];
}


unsigned int Button::ProcessMsg( Msg *pkMsg )
{
	if( pkMsg->m_eID == Msg::MOUSE_DOWN )
	{
		Coord kMouse( pkMsg->m_uiData[2], pkMsg->m_uiData[3] );

		if( pkMsg->m_uiData[1] || ( HitTest( kMouse ) != OUTSIDE ) )
		{
			m_eState = DOWN;

			m_pkState[ UP ]->Deactivate();
			m_pkState[ DOWN ]->Activate();

			//Log( LOGLEVEL_DEBUG, "[button] [%s] pressed\n", GetName().c_str() );

			//non-zero value indicates reason was mouse leaving area
			if( !pkMsg->m_uiData[1] )
			{
				Msg *pkButtonMsg = new Msg( Msg::BUTTON_DOWN, this, 0 );

				ProcessMsg( pkButtonMsg );

				delete pkButtonMsg;
			}
		}
	}	
	else if( pkMsg->m_eID == Msg::MOUSE_UP )
	{
		Coord kMouse( pkMsg->m_uiData[2], pkMsg->m_uiData[3] );

		if( pkMsg->m_uiData[1] || ( HitTest( kMouse ) != OUTSIDE ) )
		{
			m_eState = UP;

			m_pkState[ DOWN ]->Deactivate();
			m_pkState[ UP ]->Activate();

			//Log( LOGLEVEL_DEBUG, "[button] released [%s]\n", GetName().c_str() );

			//non-zero value indicates reason was mouse leaving area
			if( !pkMsg->m_uiData[1] )
			{
				Msg *pkButtonMsg = new Msg( Msg::BUTTON_UP, this, 0 );

				ProcessMsg( pkButtonMsg );

				delete pkButtonMsg;
			}
		}
	}

	return Object::ProcessMsg( pkMsg );
}


Object *Button::Duplicate( Object *pkParent, Object *pkObject )
{
	if( !pkObject )
		pkObject = new Button( pkParent, this );
	else
	{
		Object::Duplicate( pkParent, pkObject );

		Button *pkButton = dynamic_cast< Button* >( pkObject );

		pkButton->m_pkState[0] = dynamic_cast< BorderArea* >( pkButton->m_vpkChildren[0] );
		pkButton->m_pkState[1] = dynamic_cast< BorderArea* >( pkButton->m_vpkChildren[1] );

		pkButton->m_pkState[0]->SetSize( GetSize() );
		pkButton->m_pkState[1]->SetSize( GetSize() );
	}

	return pkObject;
}





#ifdef HAVE_NEOCHUNKIO


int ButtonChunk::ParseData( unsigned int uiFlags, FileManager *pkFileManager )
{
	Chunk *pkChunk = 0;
	int      iRet    = 0;

	if( !m_pkObject )
		m_pkObject = new Button( 0, 0, false );

	Button *pkButton = dynamic_cast< Button* >( m_pkObject );

	if( ( iRet = ObjectChunk::ParseData( uiFlags | ObjectChunk::NOCHILDREN | ObjectChunk::NOATTRIBUTES, pkFileManager ) ) <= 0 )
		return iRet;

	string strAreaName[] = { "Up", "Down" };

	for( int iState = 0; iState < 2; ++iState )
	{
		if( ( pkChunk = FindChunk( strAreaName[ iState ], NeoWTK::ChunkType::BORDERAREA, -1 ) ) )
		{
			BorderAreaChunk *pkAreaChunk = dynamic_cast< BorderAreaChunk* >( pkChunk );

			pkButton->AttachObject( pkAreaChunk->m_pkObject );

			pkButton->m_pkState[ iState ] = dynamic_cast< BorderArea* >( pkButton->m_vpkChildren[ iState ] );
			pkButton->m_pkState[ iState ]->SetName( strAreaName[ iState ] );
			
			pkAreaChunk->m_pkObject = 0;
		}
	}

	if( !( uiFlags & ObjectChunk::NOCHILDREN ) )
		AttachChildren();

	if( !( uiFlags & ObjectChunk::NOATTRIBUTES ) )
		SetAttributes();

	pkButton->m_pkState[ Button::UP   ]->Activate();
	pkButton->m_pkState[ Button::DOWN ]->Deactivate();

	return 1;
}


#endif


}; /*! namespace NeoWTK */

