/***************************************************************************
                  button.h  -  A clickable button with two states
                             -------------------
    begin                : Tue Jan 7 2003
    copyright            : (C) 2003 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoWTK, button.h

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#ifndef __NEOWTKBUTTON_H
#define __NEOWTKBUTTON_H


#include "base.h"
#include "object.h"

#ifdef HAVE_NEOCHUNKIO
#  include <neochunkio/complex.h>
#  include "chunktype.h"
#endif


/**
  * \file button.h
  * A clickable button with two states
  */


namespace NeoWTK
{


// External classes
class BorderArea;


// Forward declarations
class ButtonChunk;


/**
  * \class Button
  * \brief A clickable button with two states
  * \author Mattias Jansson (mattias@realityrift.com)
  */
class NEOWTK_API Button : public Object
{
#ifdef HAVE_NEOCHUNKIO
	friend class ButtonChunk;
#endif

	public:

		/**
		* \enum STATE
		* \brief Button state identifiers
		*/
		enum STATE
		{
		  /*! Not pressed */
		  UP                                          = 0,

		  /*! Pressed */
		  DOWN                                        = 1
		};


	protected:

		/*! Button state objects */
		BorderArea                                   *m_pkState[2];

		/*! Current state */
		STATE                                         m_eState;


	public:

		/**
		* \param pkParent                             Parent object
		* \param pkObject                             Reference object to copy values from
		* \param bCreateAreas                         Create child areas if true
		*/
		                                              Button( Object *pkParent, Button *pkObject = 0, bool bCreateAreas = true );

		/**
		*/
		virtual                                      ~Button();

		/**
		* Set size for both state children
		* \param rkSize                               Size
		* \return                                     New size
		*/
		virtual const Coord                          &SetSize( const Coord &rkSize );

		/**
		* Render object, recurses on children
		* \param pkFrustum                            Currently ignored
		* \param bForce                               Force rendering
		* \return                                     true if rendered, false if not
		*/
		virtual bool                                  Render( NeoEngine::Frustum *pkFrustum = 0, bool bForce = false );

		/**
		* \return                                     Borderarea object for state
		*/
		virtual BorderArea                           *GetStateObject( STATE eState );

		/**
		* Process message
		* \param pkMsg                                Message object
		* \return                                     Message specific return code, 0 indicates message was unprocessed
		*/
		virtual unsigned int                          ProcessMsg( Msg *pkMsg );

		/**
		* Duplicate and/or copy this object
		* \param pkParent                             Parent object to attach to
		* \param pkObject                             Object to copy data into, if null will allocate new object
		* \return                                     New object that is duplicate of this object
		*/
		virtual Object                               *Duplicate( Object *pkParent = 0, Object *pkObject = 0 );
};


#ifdef HAVE_NEOCHUNKIO


/**
  * \class ButtonChunk
  * \brief Base chunk for button objects
  * \author Mattias Jansson (mattias@realityrift.com)
  */
class NEOWTK_API ButtonChunk : public ObjectChunk
{
	public:

		/**
		* Initialize chunk
		* \param usType                               Chunk type
		* \param rstrType                             Chunk type as string
		* \param rstrID                               Chunk ID string
		*/
		                                              ButtonChunk( unsigned short usType, const NeoEngine::HashString &rstrType, const NeoEngine::HashString &rstrID = "" ) : ObjectChunk( usType, rstrType, rstrID ) {}

		/**
		* Deallocate data and subchunks
		*/
		virtual                                      ~ButtonChunk() {}

		/**
		* Parse chunk data
		* \param uiFlags                              Parse flags
		* \param pkFileManager                        File manager
		* \return                                     <0 if error, >0 if successful (0 reserved)
		*/
		virtual int                                   ParseData( unsigned int uiFlags, NeoEngine::FileManager *pkFileManager );

		/**
		* Allocate new chunk
		* \param usType                               Type identifier
		* \param rstrType                             Type identifier as string
		* \param rstrID                               ID string
		* \return                                     New chunk
		*/
		static NeoChunkIO::Chunk                     *Allocator( unsigned short usType, const NeoEngine::HashString &rstrType, const NeoEngine::HashString &rstrID ) { return new ButtonChunk( usType, rstrType, rstrID ); }
};


#endif /*! HAVE_NEOCHUNKIO */


}; /*! namespace NeoWTK */


#endif
