/***************************************************************************
                    coord.h  -  Class for a 2D coordinate
                             -------------------
    begin                : Sun Nov 17 2002
    copyright            : (C) 2002 by Reality Rift Studios
    email                : mattias@realityrift.com
 ***************************************************************************

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with
 the License. You may obtain a copy of the License at 
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 for the specific language governing rights and limitations under the
 License.

 The Original Code is the NeoEngine, NeoWTK, coord.h

 The Initial Developer of the Original Code is Mattias Jansson.
 Portions created by Mattias Jansson are Copyright (C) 2002
 Reality Rift Studios. All Rights Reserved.

 ***************************************************************************/

#ifndef __NEOWTKCOORD_H
#define __NEOWTKCOORD_H


#include "base.h"

#ifdef HAVE_NEOCHUNKIO
#  include <neochunkio/complex.h>
#  include "chunktype.h"
#endif


/**
  * \file coord.h
  * Simple class for a 2D coordinate
  */


namespace NeoWTK
{


/**
  * \class Coord
  * \brief 2D integer coordinate data
  * \author Mattias Jansson (mattias@realityrift.com)
  */
class NEOWTK_API Coord
{
	public:

		/*! X coordinate */
		int                            x;

		/*! Y coordinate */
		int                            y;

		/**
		* Reset to 0
		*/
		                               Coord() : x(0), y(0) {}

		/**
		* \param iX                    X coord
		* \param iY                    Y coord
		*/
		                               Coord( int iX, int iY ) : x(iX), y(iY) {}

		/**
		* Set components
		* \param iX                    X coord
		* \param iY                    Y coord
		*/
		inline void                    Set( int iX, int iY ) { x = iX; y = iY; }

		/**
		* Add two coords
		* \param rkCoord               Coordinate to add with
		* \return                      Coordinate that is sum of arguments
		*/
		inline Coord                   operator + ( const Coord &rkCoord ) const { return Coord( x + rkCoord.x, y + rkCoord.y ); }

		/**
		* Subract two coords
		* \param rkCoord               Coordinate to subtract with
		* \return                      Coordinate that is difference of arguments
		*/
		inline Coord                   operator - ( const Coord &rkCoord ) const { return Coord( x - rkCoord.x, y - rkCoord.y ); }

		/**
		* Scale coord
		* \param fScale                Scale factor
		* \return                      Scaled coordinate
		*/
		inline Coord                   operator * ( float fScale ) const { return Coord( int( float( x ) * fScale ), int( float( y ) * fScale ) ); }

		/**
		* Add two coords and store
		* \param rkCoord               Coordinate to add with
		* \return                      Coordinate that is sum of arguments (this)
		*/
		inline Coord                  &operator += ( const Coord &rkCoord ) { x += rkCoord.x; y += rkCoord.y; return( *this ); }

		/**
		* Subtract two coords and store
		* \param rkCoord               Coordinate to subtract with
		* \return                      Coordinate that is difference of arguments (this)
		*/
		inline Coord                  &operator -= ( const Coord &rkCoord ) { x -= rkCoord.x; y -= rkCoord.y; return( *this ); }

		/**
		* Scale coord
		* \param fScale                Scale factor
		* \return                      Scaled coordinate (this)
		*/
		inline Coord                  &operator *= ( float fScale ) { x = int( float( x ) * fScale ); y = int( float( y ) * fScale ); return( *this ); }

		/**
		* Negate coord
		* \return                      Negated coordinate
		*/
		inline Coord                   operator - () const { return Coord( -x, -y ); }

		/**
		* Compare two coords
		* \param rkCoord               Coordinate to compare with
		* \return                      true if equal, false if not
		*/
		inline bool                    operator == ( const Coord &rkCoord ) const { return( ( x == rkCoord.x ) && ( y == rkCoord.y ) ); }

		/**
		* Compare two coords
		* \param rkCoord               Coordinate to compare with
		* \return                      false if equal, true if not
		*/
		inline bool                    operator != ( const Coord &rkCoord ) const { return( ( x != rkCoord.x ) || ( y != rkCoord.y ) ); }
};


#ifdef HAVE_NEOCHUNKIO


/**
  * \class CoordChunk
  * \brief Coordinate data chunk
  * \author Mattias Jansson (mattias@realityrift.com)
  */
class NEOWTK_API CoordChunk : public NeoChunkIO::ComplexChunk
{
	public:

		/*! Object */
		Coord                                         m_kCoord;


		/**
		* Initialize chunk
		* \param usType                               Chunk type
		* \param rstrType                             Chunk type as string
		* \param rstrID                               Chunk ID string
		*/ 
		                                              CoordChunk( unsigned short usType, const NeoEngine::HashString &rstrType, const NeoEngine::HashString &rstrID = "" ) : NeoChunkIO::ComplexChunk( usType, rstrType, rstrID ) {}
		
		/**
		* Deallocate data and subchunks
		*/
		virtual                                      ~CoordChunk() {}

		/**
		* Parse chunk data
		* \param uiFlags                              Parse flags
		* \param pkFileManager                        File manager
		* \return                                     <0 if error, >0 if successful (0 reserved)
		*/
		virtual int                                   ParseData( unsigned int uiFlags, NeoEngine::FileManager *pkFileManager );

		/**
		* Allocate new chunk
		* \param usType                               Type identifier
		* \param rstrType                             Type identifier as string
		* \param rstrID                               ID string
		* \return                                     New chunk
		*/
		static NeoChunkIO::Chunk                     *Allocator( unsigned short usType, const NeoEngine::HashString &rstrType, const NeoEngine::HashString &rstrID ) { return new CoordChunk( usType, rstrType, rstrID ); }
};


#endif /*! HAVE_NEOCHUNKIO */


}; /*! namespace NeoWTK */


#endif
